#!/bin/sh
#
# build-user-javadoc outdir [getscript ...]
#
# Generate the user level JavaDoc for Enhydra.  This copies the source files
# from modules into an isolated tree and runs javadoc on them.  This creates
# a limited, easier to understand, view of the java class hierarchy.  This
# must be run from the top level Enhydra dir $(ROOT).  The files should
# are copied by executing the module/build/get-user-doc-srcs in each
# module.  This script output a tar stream of the files to include, with
# the names rooted at the top of the package hierarchy.
#
# Arguments:
#   o -doctitle title - Overrides default documentation title.
#   o -header header - Overrides default header.
#   o -list-omitted - Generate the omitted files list for all of
#     enhydra in tmp/non-user-doc-srcs.txt.
#   o outdir - The directory to write the javadoc.
#   o getscript - Explict list of get scripts to use.  If not specified,
#     then modules/*/build/get-user-doc-srcs is used.
#
# Also creates a file tmp/non-user-doc-srcs.txt that lists the java
# files that were not included in the generated documentation.  This
# is useful for discovering files that have been omitted.
#

set -e

# Config
ROOT="@ROOT@"
JAVADOC="@JAVADOC@"
JDKVERSION="@JDKVERSION@"
JDKDIR="@JDKDIR@"
ENHYDRA_VERSION="@ENHYDRA_VERSION@"
JAVADOC_MEMORY="-J-mx64m"
TAR=tar
export TAR

# Formatting stuff.

DOCTITLE="Enhydra Application Server ${ENHYDRA_VERSION} API"
HEADER="Enhydra ${ENHYDRA_VERSION} API"

# Output
non_user_doc=tmp/non-user-doc-srcs.txt

# Tmp files
TMPDIR="$ROOT/tmp/user-doc-srcs"
pkgs=${TMPDIR}/pkgs.tmp
user_java_srcs=$TMPDIR/user_java_srcs.tmp
all_java_srcs=$TMPDIR/all_java_srcs.tmp

if [ "X$OSTYPE" = "Xcygwin32" ] ; then
    PS=\;
else
    PS=:
fi

#
# Parse arguments.
#
list_omittted=NO
MODULES=""
while [ $# -gt 0 ] ;do
    case $1 in 
        -doctitle)
            shift
            DOCTITLE="$1"
            shift;;
        -header)
            shift
            HEADER="$1"
            shift;;
	-modules)
            shift
            MODULES="$1"
            shift;;
        -list-omitted)
            list_omittted=YES
            shift;;
        --)
            shift
            break;;
        -*)
            usage "invalid option \"$1\""
            break;;
         *)
            break;;
    esac
done
if [ $# = 0 ] ;then
    echo 'build-user-javadoc outdir [module ...]' >&2
    exit 1
fi

outdir=$1
shift

# Assemble list of module names.
if [ $# = 0 ] ; then
    getscripts=modules/*/build/get-user-doc-srcs
else
    getscripts="$*"
fi

#
# Create termporary module heirarchy containing only the source we want to
# document.
#
rm -rf $TMPDIR
mkdir -p $TMPDIR

#
# Copy java files to tmp source directory.
#
rm -f $pkgfiles
if [ "$MODULES" != "" ]
then
    for mod in $MODULES
    do 
	getscripts=`cd $mod && ls modules/*/build/get-user-doc-srcs`    
	for getscript in $getscripts
	do
	    
	    echo "running $mod/$getscript"
	    cd $mod && $getscript | (cd $TMPDIR && tar -xf -)
	    cd $ROOT
	done
    done
else	
    for getscript in $getscripts ; do
	echo $getscript
	echo running $getscript
	$getscript | (cd $TMPDIR && tar -xf -)
    done
fi

#
# Generate list of packages from source files.
# sed commands, in order of appearance:
#  1) Remove class name and .java
#  2) Remove leading `./' if it exists.
#  2) Change file `/' separators to package `.' separators.
(cd $TMPDIR && find . -name '*.java') | \
    sed -e 's%/[^/]*\.*java$%%' -e 's%^\./%%g' -e 's%/%.%g' | sort -u >$pkgs

#
# This is bit of a hack to pick up the ENHYDRA_CLASSPATH without the modules
# directories set to anyplace useful.  This keeps us from picking up
# the real source.  Pass in a new location for the MODULES variable and output
# ENHYDRA_CLASSPATH
#
tmpmake="$TMPDIR/Makefile.tmp"
cat >$tmpmake <<EOF
ROOT = .
getpath: outpath

include \${ROOT}/config.mk

outpath:
	@echo "\${ENHYDRA_CLASSPATH}"
EOF
CLASSPATH=${TMPDIR}${PS}`@MAKE@ --no-print-directory -f $tmpmake MODULES=$TMPDIR`

# Setup for output.
rm -rf $outdir
mkdir -p $outdir

# Javadoc is run differently for 1.1 and 1.2
if [ "X$JDKVERSION" = "X1.1" ] ; then
    # Java 1.1 javadoc needed some stuff from the java doc.
    JDOCGUIDE="$JDKDIR/docs/api/API_users_guide.html"
    cp -f $JDOCGUIDE ${outdir}
    (cd $JDKDIR/docs/api && tar -cf - images) | \
	 (cd $outdir && tar -xf -)

    # javadoc 1.1 didn't include runtime classes if user set CLASSPATH
    CLASSPATH="$JDKDIR/lib/classes.zip$PS$CLASSPATH"
    $JAVADOC ${JAVADOC_MEMORY} -d "$outdir" -sourcepath "$CLASSPATH" -version `cat $pkgs`
    export CLASSPATH
else
    export CLASSPATH
    $JAVADOC ${JAVADOC_MEMORY} -d "$outdir" -sourcepath "$CLASSPATH" -version \
        -splitindex -doctitle "$DOCTITLE" -windowtitle "$DOCTITLE" \
        -header "$HEADER" `cat $pkgs`
fi

#
# Generate list of undocumented *.java files.
#
if [ "X$list_omittted" = "XYES" ] ; then
    (cd modules && find . -name '*.java') | sed -e 's|^\./||' | sort -u >$all_java_srcs
    (cd $TMPDIR && find . -name '*.java') | sed -e 's|^\./||' | sort -u >$user_java_srcs
    (diff $user_java_srcs $all_java_srcs || exit 0) |
        sed -e '/^</d' -e '/^[-0-9]/d' -e 's/^> //' >$non_user_doc
    echo "Note: See $non_user_doc for sources omitted from user-level JavaDoc"
fi
