#!/bin/sh

#
# check_tree.sh
# Andy John
# 3/99
# www.enhydra.org
#
# This shell script verifies that the classes directory of a built Enhydra
# tree is in a consistant state.
#
# Because several jar files are created by the Enhydra tree, there are
# several sub-classes directories. At the end of the build each one is
# made into it's own jar file.
#
# However, because compiling one Java class may invisibly lead to the
# compiling of other classes elsewhere in the tree, sometimes this
# means that classes get put in the wrong classes subdirectory.
# This leads to subtle runtime problems, because classes end up getting
# loaded from the wrong jar files.
#
# If you type "make" from the top of the tree, and it builds sucessfully
# then this problem will not occur. The tree is set up so it builds in
# the right order to avoid the problem. But if you build half the tree,
# then do a "make" down in the tree it might occur. 
#
# This is not a common problem, but it's hard to detect. That is where
# this script comes in. It will detect if there are classes in the
# wrong place, then offer to delete them. You should delete them.
# Then do a "make" from the top of the tree to recompile the classes
# and they will end up in the right place.
#
# If you have questions, join the Enhydra mailing list and ask your
# question there, or email me directly: andy@enhydra.org.
#
#

if [ ! -d classes ]
then
    echo "No classes directory. Run me at the top of a built Enhydra tree."
    exit 1
fi

cd classes

check_dir () {
    if [ -z "$3" ]
    then
        BAD=`find $1 -type f | grep -v $2`
    else
        BAD=`find $1 -type f | grep $2`
    fi
    if [ ! -z "$BAD" ]
    then
        echo $BAD
        echo
        echo "These classes are in the wrong subdirectory, ok to delete? (y/n)"
        read X
        if [ "$X" = "y" ]
        then
            for f in $BAD
            do
                rm -f $f
            done
            echo
            echo "$1 subdirectory cleaned."
            echo "Do a \"make\" from the top of the tree to rebuild."
        fi
    fi
}


check_dir welcome '^welcome/com/lutris/appserver/welcome'
check_dir admin '^admin/com/lutris/appserver/admin'
check_dir debugger '^debugger/com/lutris/appserver/debugger'

check_dir enhydra '^enhydra/com/lutris/appserver/welcome' x
check_dir enhydra '^enhydra/com/lutris/appserver/admin' x
check_dir enhydra '^enhydra/com/lutris/appserver/debugger' x


