/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: MultiClassLoaderTests.java,v 1.15.8.1 2000/10/19 17:58:53 jasona Exp $
 */

package com.lutris.classloader.tests;

import com.lutris.classloader.*;
import java.util.Vector;
import java.net.URL;
import java.net.MalformedURLException;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.FileOutputStream;
import java.util.Properties;
import java.util.Enumeration;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import com.lutris.logging.LogChannel;
import com.lutris.logging.StandardLogger;
import com.lutris.util.FatalExceptionError;

public class MultiClassLoaderTests {
    // Relative location of our classes
    private static final String classDir
        = "../../../../../../../classes/enhydra";

    // Relative location of our package
    private static final String packageDir
        = classDir + "/com/lutris/classloader/tests";

    private static final String cpString = "/usr/local/java";
    private static final String[] cpStringArray = new String[]{"/usr/local/java"};

    private static final File cpFile = new File("/usr/local/java");
    private static final File[] cpFileArray = new File[]{new File("/usr/local/java")};

    private static final URL cpURL;
    private static final URL[] cpURLArray;

    private static final String SYSTEM_CLASSPATH_PROP = "java.class.path";

    private static final String TEST_CLASS_NAME = "com.lutris.classloader.tests.TestClass";
    private static final String TEST_CLASS2_NAME = "com.lutris.classloader.tests.TestClass2";

    private static final String[] lutrisCP = {
        classDir
    };

    private static final String LOG_FILE = "loader.log";

    private static final String TEST_FACILITY = "test";
    private static final String TEST_LOG_LEVEL = "TEST";

    // System class loader (or null for 1.1)
    private static ClassLoader sysClassLoader;
    
    static {
        // If the parent has getSystemClassLoader, its 1.2 and we use that
        // classloader.
        try {
            Method getSysCL = 
                ClassLoader.class.getMethod("getSystemClassLoader", new Class[0]);
            sysClassLoader = (ClassLoader)getSysCL.invoke(ClassLoader.class, new Object[0]);
        } catch (NoSuchMethodException except) {
            // 1.1
            sysClassLoader = null;
        } catch (InvocationTargetException except) {
            throw new FatalExceptionError(except);
        } catch (IllegalAccessException except) {
            throw new FatalExceptionError(except);
        }
    }

    // Filter that forces the test classes to be loaded by the
    // MultiClassLoader.
    private static final MultiClassLoader.ClassFilter mustLoadTestClassesFilter
        = new MultiClassLoader.ClassFilter() {
                public int loadCheck(String className) {
                    if (className.equals(TEST_CLASS_NAME)
                        || className.equals(TEST_CLASS_NAME)) {
                        return MUST_LOAD;
                    } else {
                        return NORMAL_LOAD;
                    }
                }
            };

    // Count of errors.
    private static int errorCnt = 0;

    /**
     * Both print and log a message.  Having messages in the log makes it
     * easier to use for debugging test failures.
     */
    private static void testMsg(String msg) {
        System.out.println(msg);
        log(msg);
    }

    /*
     * Generate a message and up the error count.
     */
    private static void testError(String msg) {
        String emsg = "Error: " + msg;
        System.out.println(emsg);
        log(emsg);
        errorCnt++;
    }

    /**
     * Generate a URL from a String, URL or File.
     */
    private static URL mkURL(Object obj) {
        // URL file must end in a slash.
        try {
            if (obj instanceof URL) {
                URL objUrl = (URL)obj;
                return new URL(objUrl.getProtocol(), objUrl.getHost(),
                               objUrl.getFile() + "/");
            } else if (obj instanceof File) {
                return new URL("file", "", ((File)obj).getPath() + "/");
            } else {
                return new URL("file", "", (String)obj + "/");
            }
        } catch (MalformedURLException except) {
            throw new FatalExceptionError(except);
        }
    }

    /*
     * Check the class path in an class loader against an * expected value.
     * @param testId Id for reporting errors.
     * @param loader Class loader to check.
     * @param expected Expected class path, maybe a String, URL, File,
     *  or an array of one of these..
     */
    private static void checkClassPath(String testId,
                                       MultiClassLoader loader,
                                       Object expected) {
        URL[] expectedURLs;
        // Convert expected to an array of URLs.
        if (expected.getClass().isArray()) {
            Object[] objs = (Object[])expected;
            expectedURLs = new URL[objs.length];
            for (int i = 0; i < objs.length; i++) {
                expectedURLs[i] = mkURL(objs[i]);
            }
        } else {
            expectedURLs = new URL[1];
            expectedURLs[0] = mkURL(expected);
        }

        URL[] actual = loader.getClassPath();
        if (actual.length != expectedURLs.length) {
            testError(testId + ": class path length (" + actual.length
                      + ") not expected length (" + expectedURLs.length + ")");
            return;
        }
        for (int i = 0; i < actual.length; i++) {
            if (!actual[i].equals(expectedURLs[i])) {
                testError(testId + ": class path entry (" + actual[i]
                      + ") not expected (" + expectedURLs[i] + ")");
            }
        }
    }

    /**
     * Test constructors for void, String, File, URL, and Vector.
     */
    private static void testConstructors() {
	testMsg("\ntest-1: Begin testing constructors.");

	// Test void constructor
	MultiClassLoader loader = new MultiClassLoader(getLogChannel());
        checkClassPath("test-1.1", loader, new String[0]);

	// Test constructor String and String[]
        loader = new MultiClassLoader(cpString, getLogChannel());
        checkClassPath("test-1.2", loader, cpString);

	loader = new MultiClassLoader(cpStringArray, getLogChannel());
        checkClassPath("test-1.3", loader, cpStringArray);

	// Test constructor for File and File[]
	loader = new MultiClassLoader(cpFile, getLogChannel());
        checkClassPath("test-1.4", loader, cpFile);

	loader = new MultiClassLoader(cpFileArray, getLogChannel());
        checkClassPath("test-1.5", loader, cpFileArray);

	// Test constructor for URL and URL[]
        try {
            loader = new MultiClassLoader(cpURL, getLogChannel());
            checkClassPath("test-1.6", loader, cpURL);
            loader = new MultiClassLoader(cpURLArray, getLogChannel());
            checkClassPath("test-1.7", loader, cpURLArray);
	} catch (Throwable err) {
	    testError("test-1.8: URL constructor tests unsuccessful: " + err);
            err.printStackTrace();
	}
    }

    /**
     * Test clear/set/getClassPath for Strings, Files, and URLs.
     */
    private static void testSetClassPath() {
	testMsg("\ntest-2: Begin testing clear/set/getClassPath methods.");

	// Test setClassPath and getClassPath for String and String[]
	MultiClassLoader loader = new MultiClassLoader(getLogChannel());
	loader.setClassPath(cpString);
        checkClassPath("test-2.1", loader, cpString);

	loader.setClassPath(cpStringArray);
        checkClassPath("test-2.2", loader, cpStringArray);

	// Test setClassPath and getClassPath for File
	loader = new MultiClassLoader(getLogChannel());
	loader.setClassPath(cpFile);
        checkClassPath("test-2.3", loader, cpFile);

	loader.setClassPath(cpFileArray);
        checkClassPath("test-2.4", loader, cpFileArray);

	// Test setClassPath and getClassPath for URL
	loader = new MultiClassLoader(getLogChannel());
	loader.setClassPath(cpURL);
        checkClassPath("test-2.5", loader, cpURL);

	loader.setClassPath(cpURLArray);
        checkClassPath("test-2.6", loader, cpURLArray);

	// Test clearClassPath and getClassPath
	loader.clearClassPath();
        checkClassPath("test-2.7", loader, new String[0]);
    }

    /**
     * Test addClassPath for Strings, Files, and URLs.
     */
    private static void testAddClassPath() {
	testMsg("\ntest-3: Begin testing addClassPath methods.");

	// Test addClassPath and getClassPath for String and String[]
	MultiClassLoader loader = new MultiClassLoader(getLogChannel());
	loader.addClassPath(cpString);
        checkClassPath("test-3.1", loader, cpString);

	loader = new MultiClassLoader(getLogChannel());
	loader.addClassPath(cpStringArray);
        checkClassPath("test-3.2", loader, cpStringArray);

	// Test addClassPath and getClassPath for File
	loader = new MultiClassLoader(getLogChannel());
	loader.addClassPath(cpFile);
        checkClassPath("test-3.3", loader, cpFile);

	loader = new MultiClassLoader(getLogChannel());
	loader.addClassPath(cpFileArray);
        checkClassPath("test-3.4", loader, cpFileArray);

	// Test addClassPath and getClassPath for URL
	loader = new MultiClassLoader(getLogChannel());
	loader.addClassPath(cpURL);
        checkClassPath("test-3.5", loader, cpURL);

	loader = new MultiClassLoader(getLogChannel());
	loader.addClassPath(cpURLArray);
        checkClassPath("test-3.6", loader, cpURLArray);
    }

    /**
     * Test getSystemClassPath.
     */
    private static void testGetSystemClassPath() {
        testMsg("\ntest-4: Begin testing getSystemClassPath method.");

        String oldSystemCP = setSystemClassPath(cpString);
        try {
            URL[] sysCP = MultiClassLoader.getSystemClassPath();
            String cp1 = sysCP[0].getFile();
            String cp2 = System.getProperty("java.class.path");
            if (cp1 == null || cp2 == null || !equals(cp1, cp2)) {
                testError("test-4.1: getSystemClassPath unsuccessful: wrong classpath returned");
            }
        } finally {
            setSystemClassPath(oldSystemCP);
        }
    }

    /**
     * Check if classloader and classname are correct.
     */
    private static void checkLoadedClass(Object obj,
                                         ClassLoader expectedLoader,
                                         String expectedName,
                                         String errorMsg) {
        Class cl = obj.getClass();
        if (cl.getClassLoader() != expectedLoader) {
            testError(errorMsg + ": wrong classloader");
        }
        if (!cl.getName().equals(expectedName)) {
            testError("test-4.2: loadClass with path unsuccessful: wrong class");
        }
    }

    /**
     * Test loadClass with .class file.
     */
    private static void testLoadClass() {
	testMsg("\ntest-4: Begin testing loadClass method with class.");

	try {
	    // Try loading with the class path set
	    MultiClassLoader loader = new MultiClassLoader(getLogChannel());
	    loader.setClassPath(lutrisCP);
	    Object tc1 = loader.loadClass(TEST_CLASS_NAME).newInstance();
            checkLoadedClass(tc1, loader, TEST_CLASS_NAME,
                             "test-4.4: loadClass with path unsuccessful");

	    // Try fallback to system loader.
            String oldSystemCP = setSystemClassPath(cpString);
            try {
                loader = new MultiClassLoader(getLogChannel());
                tc1 = loader.loadClass(TEST_CLASS_NAME).newInstance();
                checkLoadedClass(tc1, sysClassLoader, TEST_CLASS_NAME,
                                 "test-4.5: loadClass system loader unsuccessful");
            } finally {
                setSystemClassPath(oldSystemCP);
            }
	} catch (Throwable err) {
	    testError("test-4.6: testLoadClass unsuccessful: " + err);
            err.printStackTrace();
	}
    }

    /**
     * Test loadClass with .jar file.
     */
    private static void testLoadJar() {
	testMsg("\ntest-5: Begin testing loadClass method with jars.");

	try {
	    // Try loading with the class path set
	    String testCP = packageDir + "/TestClass.jar";
	    MultiClassLoader loader = new MultiClassLoader(getLogChannel());
	    loader.setClassPath(testCP);
	    Object tc1 = loader.loadClass(TEST_CLASS_NAME).newInstance();
            checkLoadedClass(tc1, loader, TEST_CLASS_NAME,
                             "test-5.1: loadClass with path unsuccessful");

	    // Try fallback to system loader
            String oldSystemCP = setSystemClassPath(testCP);
            try {
                loader = new MultiClassLoader(getLogChannel());
                tc1 = loader.loadClass(TEST_CLASS_NAME).newInstance();
                checkLoadedClass(tc1, sysClassLoader, TEST_CLASS_NAME,
                                 "test-5.2: loadClass with system loader unsuccessful");
            } finally {
                setSystemClassPath(oldSystemCP);
            }
        } catch (Throwable err) {
	    testError("test-5.3: testLoadJar unsuccessful. " + err);
            err.printStackTrace();
	}
    }

    /**
     * Test getResourceAsStream with directory.
     */
    private static void testGetResourceDir() {
	testMsg("\ntest-6: Begin testing getResourceAsStream method with directory.");

        InputStream tc3 = null;
        Class c3 = null;
        int a3 = 0;
	try {
	    // Try loading with system loader
	    String testCP = classDir;
            String oldSystemCP = setSystemClassPath(testCP);
            try {
                MultiClassLoader loader = new MultiClassLoader(getLogChannel());
                loader.setClassPath(testCP);
                InputStream tc1 = 
                    loader.getResourceAsStream("com/lutris/classloader/tests/TestClass.class");

                Class c1 = tc1.getClass();
                int a1 = 0;
                try {
                    a1 = tc1.available();
                } catch (Throwable err) {
                    testError("test-6.1: loadClass unsuccessful. " + err);
                    err.printStackTrace();
                }
                if ((tc1 == null) || (a1 < 400)) {
                    testError("test-6.2: getResourceAsStream with path unsuccessful: available=" + a1);
                }
            } finally {
                setSystemClassPath(oldSystemCP);
            }

	    // Try loading with the class path empty
            oldSystemCP = setSystemClassPath(".:/");
            try {
                if (false) {
                    //FIXME Implement or delete.
                    MultiClassLoader loader = new MultiClassLoader(getLogChannel());
                    InputStream tc2 = 
                        loader.getResourceAsStream("/users/kristen/lutris/LBS/Harmony/classes/lutris/com/lutris/classloader/tests/TestClass.class");
                    Class c2 = tc2.getClass();
                    int a2 = tc2.available();

                    // Try loading with the Class getResourceAsStream
                    Class fclass = Class.forName("java.io.File");
                    File somefile = new File(packageDir + "/TestClass.class");
                    tc3 = fclass.getResourceAsStream(somefile.getAbsolutePath());
                    c3 = tc3.getClass();
                    try {
                        a3 = tc3.available();
                    } catch (Throwable err) {
                        testError("test-6.3: loadClass unsuccessful. " + err);
                        err.printStackTrace();
                    }
                }
            } finally {
                setSystemClassPath(oldSystemCP);
            }

	} catch (Throwable err) {
	    testError("test-6.4: loadClass unsuccessful. " + err);
            err.printStackTrace();
	}
    }
    /**
     * Test getResourceAsStream with .jar file.
     */
    private static void testGetResourceJar() {
	testMsg("\ntest-7: Begin testing getResourceAsStream method with jar.");

        String testCP = packageDir + "/TestClass.jar";
        String oldSystemCP = setSystemClassPath(testCP);
	try {
	    // Try loading with the class path set
	    MultiClassLoader loader = new MultiClassLoader(getLogChannel());
	    loader.setClassPath(testCP);
	    InputStream tc1 = 
                loader.getResourceAsStream("com/lutris/classloader/tests/TestClass.class");
	    Class c1 = tc1.getClass();
            int a1 =  tc1.available();

	    if ((tc1 == null) || (a1 < 400)) {
		testError("test-7.1: getResourceAsStream with path unsuccessful.");
	    }
	} catch (Throwable err) {
	    testError("test-7.2: loadClass unsuccessful. " + err);
            err.printStackTrace();
        } finally {
            setSystemClassPath(oldSystemCP);
        }
    }

    /**
     * Test shouldReload and hasBeenModified.
     */
    private static void testModifiedCheck() {
	testMsg("\ntest-8: Begin testing shouldReload method.");
	try {
            // The time on the file in the jar doesn't change, so
            // shouldReload is never true.
	    String testCP = packageDir + "/TestClass.jar";
	    MultiClassLoader loader = new MultiClassLoader(getLogChannel());
            loader.addClassFilter(mustLoadTestClassesFilter);
	    loader.setClassPath(testCP);
            if (loader.shouldReload()) {
                testError("test-8.1: shouldReload true");
            }
            loader.loadClass(TEST_CLASS_NAME);
            if (loader.shouldReload()) {
                testError("test-8.2: shouldReload true");
            }
	    Process p = Runtime.getRuntime().exec("touch " + testCP);
	    p.waitFor();
            if (p.exitValue() != 0) {
                testError("test-8.3: touch exited non-zero: " + testCP);
            }
            if (loader.shouldReload()) {
                testError("test-8.4: shouldReload true");
            }
            if (loader.shouldReload()) {
                testError("test-8.5: shouldReload true");
            }

            // Test standalone file.
            loader = new MultiClassLoader(getLogChannel());
            loader.addClassFilter(mustLoadTestClassesFilter);
	    testCP = classDir;
	    loader.setClassPath(testCP);
	    testCP += "/com/lutris/classloader/tests/TestClass.class";
            if (loader.shouldReload()) {
                testError("test-8.6: shouldReload true");
            }
            loader.loadClass(TEST_CLASS_NAME);
            if (loader.shouldReload()) {
                testError("test-8.7: shouldReload true");
            }
	    p = Runtime.getRuntime().exec("touch " + testCP);
	    p.waitFor();
            if (p.exitValue() != 0) {
                testError("test-8.8: touch exited non-zero: " + testCP);
            }
            if (!loader.shouldReload()) {
                testError("test-8.9: shouldReload false");
            }
            if (!loader.shouldReload()) {
                testError("test-8.10: shouldReload false");
            }
	} catch (Throwable err) {
	    testError("test-8.11: shouldReload unsuccessful. " + err);
            err.printStackTrace();
	    errorCnt++;
	}
    }

    /**
     * Test filters and passing to secondary.
     */
    private static void testFilter() {
	testMsg("\ntest-9: Begin testing filtering.");
	try {
	    MultiClassLoader loader1 = new MultiClassLoader(lutrisCP, getLogChannel());
	    MultiClassLoader loader2 = new MultiClassLoader(null, loader1, getLogChannel());
            loader2.setClassPath(lutrisCP);
            loader2.addClassFilter(new MultiClassLoader.ClassFilter() {
                public int loadCheck(String className) {
                    return className.equals(TEST_CLASS2_NAME)
                        ? NORMAL_LOAD : DONT_LOAD;
                }
            });
            Class c1 = loader2.loadClass(TEST_CLASS_NAME);
            Class c2 = loader2.loadClass(TEST_CLASS2_NAME);
            if (c1.getClassLoader() != loader1) {
                testError("test-9.1: wrong class loader for " + TEST_CLASS_NAME);
            }
            if (c2.getClassLoader() != loader2) {
                testError("test-9.2: wrong class loader for " + TEST_CLASS2_NAME);
            }

	} catch (Throwable err) {
	    testError("test-9.3: testFilter unsuccessful. " + err);
            err.printStackTrace();
	}
    }

    public static boolean equals(String path1, String path2) {
	if (path1.equals(path2)) {
	    return true;
	} else {
	    String path1b = path1 + "/";
	    String path2b = path2 + "/";
	    if (path1b.equals(path2) || path2b.equals(path1)) {
		return true;
	    }
	}
	return false;
    }

    public static void main(String args[]) {
        testConstructors();
        testSetClassPath();
        testAddClassPath();
        testGetSystemClassPath();
        testLoadClass();
        testLoadJar();
        testGetResourceDir();
        testGetResourceJar();
        testModifiedCheck();
        testFilter();
	testMsg("\nDONE: There were " + errorCnt + " errors.\n");
        if (errorCnt > 0) {
            System.exit(1);
        }
    }

    private static LogChannel logChannel;
    private static int logLevel;

    static {
        File logFile = new File(LOG_FILE);
	StandardLogger logger = new StandardLogger(false);
	try {
            // FIXME: Deleting here causes file to endup empty, so just zero.
            new FileOutputStream(logFile).close();
	    logger.configure(logFile,
                             new String[]{TEST_LOG_LEVEL, "CLASSLOAD"},
                             new String[]{});
	} catch (IOException ioe) {
            throw new FatalExceptionError(ioe);
	}
	logChannel = logger.getChannel(TEST_FACILITY);
	logLevel = logChannel.getLevel(TEST_LOG_LEVEL);
        try {
            cpURL = new URL("file:///usr/local/java");
            cpURLArray = new URL[]{new URL("file:///usr/local/java")};
        } catch (MalformedURLException mue) {
            throw new FatalExceptionError(mue);
        }
    }

    public static void log(String message) {
	logChannel.write(logLevel, message);
    }

    public static LogChannel getLogChannel() {
	return logChannel;
    }

    /**
     * Modify the system class path and return the old class path.
     */
    private static String setSystemClassPath(String cp) {
        Properties sysProps = System.getProperties();
        String oldCP = (String)sysProps.get(SYSTEM_CLASSPATH_PROP);

        Properties props = new Properties(sysProps);
        props.put(SYSTEM_CLASSPATH_PROP, cp);
        System.setProperties(props);
        return oldCP;
    }

    /**
     * Print the classpath.
     */
    private static void printClassPath(MultiClassLoader loader) {
        testMsg("Classpath = ");
        URL[] urls = loader.getClassPath();
        for (int i = 0; i < urls.length; i++) {
            testMsg("    " + urls[i]);
        }
    }
}
