/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HttpCookieManager.java,v 1.10.12.1 2000/10/19 17:58:53 jasona Exp $
 */





package com.lutris.http;
import java.util.*;
import javax.servlet.http.Cookie;

/**
 * This object manages cookies defined in http request and
 * response headers.
 *
 * @author	Kyle Clark
 * @version	$Revision: 1.10.12.1 $
 * @since       Harmony2.0
 */
public class HttpCookieManager {

    /**
     * The set of cookies managed by this object.
     */
    private Vector cookies;

    /**
     * Constructs a new cookie manager.
     */
    public HttpCookieManager () {
        cookies = new Vector();
    }

    /**
     * Returns the string representation of all the cookies being managed
     * by this object as an http cookie request header.
     * The format of the header is:
     * <pre>
     *    NAME1=VALUE1; NAME2=VALUE2; ....
     * </pre>
     *
     * @return the "Cookie" header as it would appear in an http request header.
     */
    public String getRequestHeader () {
        return getRequestHeader(null, null);
    }

    /**
     * Returns the string representation of the cookies being managed
     * by this object as an http cookie request header
     * The format of the header is:
     * <pre>
     *    NAME1=VALUE1; NAME2=VALUE2; ....
     * </pre>
     *
     * @param path
     *   Only return the cookies that are valid in the specified
     *   path. If path is null then path is ignored.
     * @param domain
     *   Only return the cookies that are valid in the specified
     *   domain.  If null, the domain is ignored.
     * @return the "Cookie" header as it would appear in an http request
     *   or null if this object contains no cookies that apply
     *   to the path and domain contraints.
     */
    public String getRequestHeader(String path, String domain) {
        // FIX: use string buffer
        String header = null;
        Enumeration e = cookies.elements();
        while (e.hasMoreElements()) {
            Cookie c = (Cookie)e.nextElement();
            boolean matchPath = true;
            boolean matchDomain = true;
            if (path != null) {
                if (c.getPath() != null)
                    matchPath = c.getPath().startsWith(path);
                else
                    matchPath = true;
            }
            if (domain != null) {
                // FIX: read spec, need to do this correctly
                if (c.getDomain() != null)
                    matchDomain = c.getDomain().endsWith(domain);
                else
                    matchDomain = true;
            }
            if (matchPath && matchDomain) {
                if (header == null)
                    header = c.getName() + "=" + c.getValue();
                else
                    header += "; " + c.getName() + "=" + c.getValue();
            }
        }            
        return header;
    }

    /**
     * Returns the string representation of the cookies being managed
     * by this object as an http cookie response ("Set-Cookie")
     * header.  The format of the header is a comma (,) separated
     * list of:
     * <pre>
     *   name=value [;EXPIRES=dateValue] [;DOMAIN=domainName]
     *              [;PATH=path] [;SECURE]
     * </pre>
     *
     * @return the string representation of the cookies being managed
     * by this object as an http cookie response header ("Set-Cookie").
     * null is returned if there aren't any cookies
     * being managed by this cookie manager.
     */
    public String getResponseHeader() {
        String header = null;
        Enumeration e = cookies.elements();
        while (e.hasMoreElements()) {
            Cookie c = (Cookie)e.nextElement();
            if (header == null)
                header = c.toString();
            else
                header += ", " + c.toString();
        }            
        return header;
    }

    /**
     * Merges the cookies contained in the specified http response header
     * into the set of cookies being managed by this object.
     *
     * @param header the Set-Cookie http response header to merge.
     */
    public synchronized void mergeHeader(String header) {
        if (header == null) {
            return;
        }
        Vector newCookies = HttpCookieParser.parseResponseHeader(header);
        Cookie [] merge = new Cookie[cookies.size()];
        Cookie [] mergeNew = new Cookie[newCookies.size()];
        cookies.copyInto(merge);
        newCookies.copyInto(mergeNew);

        //
        // Remove from the original set any cookies that are
        // redefined in the new set.
        //
        for (int idx=0; idx<merge.length; idx++) {
            for (int idxNew=0; idxNew<mergeNew.length; idxNew++) {
                if (testEquality(merge[idx], mergeNew[idxNew])) {
                    merge[idx] = null;
                    break;
                }
            }
        }
 
        //
        // Reconstruct the cookie set.
        //
        cookies = new Vector();
        for (int idx=0; idx<merge.length; idx++) {
            if (merge[idx] != null)
                cookies.addElement(merge[idx]);
        }
        for (int idx=0; idx<mergeNew.length; idx++) { 
            cookies.addElement(mergeNew[idx]);
        }
    }

    /**
     * Returns an array of the cookies being managed.
     *
     * @return enumeration of javax.servlet.http.Cookie
     *   objects.
     * @see javax.servlet.http.Cookie
     */
    public Cookie[] getCookies() {
        Cookie[] cookieTab = new Cookie[cookies.size()];
        cookies.copyInto(cookieTab);
        return cookieTab;
    }

    /**
     * Removes all cookies from this manager.
     */
    public synchronized void clear() {
        cookies.setSize(0);
    }

    /**
     * Returns the string representation of the headers/cookies being managed
     * by this object as an http cookie response header (i.e. Set-Cookie
     * header).
     *
     * @return the string representation of the cookies being managed
     * by this object as an http cookie response header (i.e. Set-Cookie
     * header).
     */
    public String toString() {
        return getResponseHeader();
    }

    /**
     * Compares two cookies.  If the name, path, and domain are the
     * same then the cookies are equal equal.
     *
     * @return true if the cookies are equal.
     */

    private boolean testEquality(Cookie c1, Cookie c2) {
        String name1 = c1.getName();
        String name2 = c2.getName();
        if ((name1 == null) && (name2 != null))
            return false;
        if ((name1 != null) && (name2 == null))
            return false;
        if ((name1 != null) && (name2 != null) && !name1.equals(name2))
            return false;

        String path1 = c1.getPath();
        String path2 = c2.getPath();
        if ((path1 == null) && (path2 != null))
            return false;
        if ((path1 != null) && (path2 == null))
            return false;
        if ((path1 != null) && (path2 != null) && !path1.equals(path2))
            return false;

        String domain1 = c1.getDomain();
        String domain2 = c2.getDomain();
        if ((domain1 == null) && (domain2 != null))
            return false;
        if ((domain1 != null) && (domain2 == null))
            return false;
        if ((domain1 != null) && (domain2 != null) && !domain1.equals(domain2))
            return false;

        return true;
    }
}
