/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HttpCookieParser.java,v 1.22.2.2.2.1 2000/10/19 17:58:53 jasona Exp $
 */





package com.lutris.http;
import javax.servlet.http.Cookie;
import java.util.*;
import java.text.*;
import java.math.*;

/**
 * This class takes a cookie header and converts it into a
 * set of http cookies.  It can be used to parse request and
 * response cookie headers as well as format response cookies.
 *
 * @author Kyle Clark
 * @version $Revision: 1.22.2.2.2.1 $
 * @since Harmony1.0
 */
public class HttpCookieParser {

    private static final String EXPIRES = "expires";
    private static final String MAXAGE = "max-age";
    private static final String PATH = "path";
    private static final String DOMAIN = "domain";
    private static final String SECURE = "secure";

    /**
     * Object for formatting dates.
     */
    private static final SimpleDateFormat dateFormatter =
        new SimpleDateFormat("EEEE, dd-MMM-yyyy HH:mm:ss z", Locale.US);

    /*
     * Class constructor.
     */
    static {
	dateFormatter.setTimeZone(TimeZone.getTimeZone("GMT"));
    }

    /**
     * Parses an http cookie request header and returns a vector
     * of Cookie objects defined in the header.  This method
     * should only be used with the <b>request</b> cookie header "Cookie:".
     *
     * @param httpReqCookieHeader
     *   The http cookie request header to parse.
     * @return
     *   vector containing the cookies defined in the header.
     */
    public static Vector parseRequestHeader(String httpReqCookieHeader) {
        Cookie cookie;
        Vector result = new Vector();

        // Cookie are separated by ';'
        StringTokenizer cookieTokens =
            new StringTokenizer(httpReqCookieHeader.trim(), ";");
        while (cookieTokens.hasMoreTokens()) {
            // Name is separated from value by '='
            StringTokenizer t =
                new StringTokenizer(cookieTokens.nextToken(), "=");
            String name = t.nextToken().trim();
            if (t.hasMoreTokens()) {
                // Name has a value
                String value = t.nextToken().trim();
                result.addElement(new Cookie(name, value));
            }
        }
        return result;
    }

    /**
     * Parses an http cookie request header and returns the first
     * cookie that matches the specified name.  This method
     * should only be used with the <b>request</b> cookie header "Cookie:".
     *
     * @param httpReqCookieHeader
     *   The http cookie request header to parse.
     * @param name
     *   The cookie name of interest.
     * @return
     *   The first cookie that matches.  null if no matching cookies
     *   are found.
     */
    public static Cookie parseRequestHeader(String httpReqCookieHeader,
                                            String name) {
        // Cookie are separated by ';'
        StringTokenizer cookieTokens =
            new StringTokenizer(httpReqCookieHeader.trim(), ";");
        while (cookieTokens.hasMoreTokens()) {
            // Name is separated from value by '='
            StringTokenizer t =
                new StringTokenizer(cookieTokens.nextToken(), "=");
            String tokenName = t.nextToken().trim();
            if (t.hasMoreTokens() && tokenName.equals(name)) {
                // if name has a value
                String value = t.nextToken().trim();
                return new Cookie(name, value);
            }
        }
        return null;
    }


    /**
     * Convert an old cookie expires data to maximum age; assuming now
     * as the start.
     *
     * @param dateStr The string expiry date.
     * @return The max age, in seconds.
     */
    public static int expiresToMaxAge(String dateStr) {
        Date date = dateFormatter.parse(dateStr.trim(), new ParsePosition(0));
        Date current = new Date();
        return (int)(current.getTime()-date.getTime()) * 1000;
    }


    /**
     * Convert amaximum age to a old cookie expires data; assuming now
     * as the start.
     *
     * @param The max age, in seconds.
     * @return The string expiry date.
     */
    public static String maxAgeToExpires(int maxAge) {
        long maxInMillis = (long)maxAge * 1000;
        Date date = new Date(new Date().getTime()+maxInMillis);
        StringBuffer dateStr = new StringBuffer();
        dateFormatter.format(date, dateStr, new FieldPosition(0));
        return dateStr.toString();
    }

    /**
     * Parses an http cookie response header and returns a vector
     * of Cookie objects defined in the header.  This method
     * only works with http <b>response</b> headers, (i.e. Set-Cookie).
     *
     * @param httpResponseCookieHeader
     *   The http cookie response header from which the cookies
     *   are constructed.
     * @return
     *   Vector containing the cookies defined in the header.
     */
    public static Vector parseResponseHeader(String httpResponseCookieHeader) {
        Vector result = new Vector();
        httpResponseCookieHeader = httpResponseCookieHeader.trim();
        // Cookies are separated by ','
        StringTokenizer cookieTokens =
            new StringTokenizer(httpResponseCookieHeader, ",");
        while (cookieTokens.hasMoreTokens()) {
            // Cookie fields are separated by ';'
            StringTokenizer tokens =
                new StringTokenizer(cookieTokens.nextToken(), ";");
            Cookie cookie = null;
            while (tokens.hasMoreTokens()) {
                // field name is separated from value by '='
                StringTokenizer t =
                    new StringTokenizer(tokens.nextToken(), "=");
                String name = t.nextToken().trim();
                if (t.hasMoreTokens()) {
                    String value = t.nextToken().trim();
                    if (cookie != null) {
                        if (name.equalsIgnoreCase(EXPIRES)) {
                            cookie.setMaxAge(expiresToMaxAge(value));
                        } else if (name.equalsIgnoreCase(DOMAIN)) {
                            cookie.setDomain(value);
                        } else if (name.equalsIgnoreCase(PATH)) {
                            cookie.setPath(value);
                        }
                    } else {      // New cookie
                        cookie = new Cookie(name, value);
                        result.addElement(cookie);
                    }
                } else if ((cookie != null) && name.equalsIgnoreCase(SECURE)) {
                    cookie.setSecure(true);
                }
            }
        }
        return result;
    }


    public static Cookie parseCookieString(String cookieString) {
        cookieString = cookieString.trim();
        // Cookie fields are separated by ';'
        StringTokenizer tokens = new StringTokenizer(cookieString, ";");
        Cookie cookie = null;
        while (tokens.hasMoreTokens()) {
            // field name is separated from value by '='
            StringTokenizer t =
                new StringTokenizer(tokens.nextToken(), "=");
            String name = t.nextToken().trim();
            if (t.hasMoreTokens()) {
                String value = t.nextToken().trim();
                if (cookie != null) {
                    if (name.equalsIgnoreCase(EXPIRES)) {
                        cookie.setMaxAge(expiresToMaxAge(value));
                    } else if (name.equalsIgnoreCase(DOMAIN)) {
                        cookie.setDomain(value);
                    } else if (name.equalsIgnoreCase(PATH)) {
                        cookie.setPath(value);
                    } else if (name.equalsIgnoreCase(MAXAGE) ) {
                        int maxAge = 0;
                        try {
                            maxAge = Integer.parseInt(value);
                        } catch (Exception e) {
                        }
                        cookie.setMaxAge(maxAge);
                    }
                } else {      // New cookie
                    cookie = new Cookie(name, value);
                }
            } else if ((cookie != null) && name.equalsIgnoreCase(SECURE)) {
                cookie.setSecure(true);
            }
        }
        return cookie;
    }


    /**
     * Append a cookie keyword/value pair, with any necessary quoting of
     * the value.
     *
     * @param buf String to append cookie arguments to.
     * @param key Keyword to append.
     * @param value Value to append, with quoting.
     */
    //FIXME: Turns out, old cookies didn't support quoted strings,
    // So we need to check the version number.  Right now, just don't
    // do quoting.
    private static void appendCookieArg(StringBuffer buf, String key, String value) {
        if (buf.length() > 0) {
            buf.append("; ");
        }
        buf.append(key);
        buf.append("=");
        buf.append(value);
    }

    /**
     * Append a cookie keyword/value pair, with any necessary quoting of
     * the value.  Adjust case of keyword if its a version zero cookie.
     * This is should not be used on the cookie name, as case is honored
     * there.
     *
     * @param version Cookie version.
     * @param buf String to append cookie arguments to.
     * @param key Keyword to append.
     * @param value Value to append, with quoting.
     */
    private static void appendCookieArg(int version, StringBuffer buf, String key, String value) {
        //FIXME:  Don't think this is really needed; just put in as a possible
        //hack for the page-reload problem.
        if (version == 0) {
            // Version 0 appears to like lower case keywords.
            key = key.toLowerCase();
        }
        appendCookieArg(buf, key, value);
    }

    /**
     * Format a Cookie for a response header.
     *
     * @param The cookie to format.
     * @return The value for the header (name is not included).
     */
    public static String formatResponseCookie(Cookie cookie) {
        StringBuffer buf = new StringBuffer();
        int version = cookie.getVersion();

        //FIXME: only 0.0 is supported; just got to figure out quoting.
        if (version > 0) {
            throw new Error("HttpCookieParser: only version 0 cookies supported right now");
        }

        appendCookieArg(buf, cookie.getName(), cookie.getValue());
        if ((version > 0) && (cookie.getComment() != null)) {
            appendCookieArg(version, buf, "Comment", cookie.getComment());
        }
        if (cookie.getDomain() != null) {
            appendCookieArg(version, buf, "Domain", cookie.getDomain());
        }
        if (cookie.getMaxAge() >= 0) {
            if (version > 0) {
                appendCookieArg(version, buf, "Max-Age", Integer.toString(cookie.getMaxAge()));
            } else {
                appendCookieArg(version, buf, "Expires", maxAgeToExpires(cookie.getMaxAge()));
            }
        }
        if (cookie.getPath() != null) {
            appendCookieArg(version, buf, "Path", cookie.getPath());
        }
        if (cookie.getDomain() != null) {
            appendCookieArg(version, buf, "Domain", cookie.getDomain());
        }
        return buf.toString();
    }
}
