/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: MultipartMimeInput.java,v 1.8.12.1 2000/10/19 17:58:53 jasona Exp $
 */




package com.lutris.mime;
import	java.io.*;
import	com.lutris.util.*;

/**
 * Presents an incoming Multipart MIME message as a series of
 * distinct MultipartMimeInputStream streams.
 */
public class MultipartMimeInput
{
    /**
     * The current section of Multipart input.  This is maintained so that
     * it can be invalidated if a request is made to go to the next 
     * section.
     */
    protected MultipartMimeInputStream	currentStream;
    
    /**
     * An input stream that does exact pattern searching using an efficient
     * string matching algorithm.
     */
    protected BMByteSearchStream	inputSource;
    
    /**
     * The string that represents a newline for this instance of input.
     * This is necessary because some web browsers (typically on Unix
     * hosts) send LF instead of the Mime-required CR+LF.  
     */
    protected String			newlineString="\r\n";
    
    /**
     * The boundary string which separates different sections of the 
     * multipart Mime input.
     */
    protected String			inputSeparator;
    
    /**
     * Pre-compiled Boyer-Moore pattern search object for the Mime
     * boundary string.
     */
    protected BMByteSearch		inputPattern;
    
    /**
     * True if the end of the input source has been reached and no
     * more multipart sections are available to be read.
     */
    protected boolean			atEOF = false;
    
    /**
     * Constructs a new MultipartMimeInput object from an input source 
     * of type <code>InputStream</code> and a <code>ContentHeader</code>
     * object.  The <code>ContentHeader</code> object will usually be a
     * <code>Content-Type</code> header and its value <i>must</i> begin
     * with <code>"multipart/"</code> to indicate multipart Mime input.
     * Other Mime types will cause a <code>MimeException</code> to be
     * thrown.  In addition, a parameter called <code>"boundary"</code>
     * <i>must</i> exist in the header, since multipart Mime input
     * is split using a boundary string passed in the
     * <code>Content-Type</code> header.
     * 
     * @param source	Input stream from which Mime input will be read.
     * @param contentHeader <code>ContentHeader</code> object containing
     * 			a <code>boundary</code> field and Mime type to
     * 			be used in scanning the input.
     * @exception MimeException Thrown if an illegal header or Mime
     * 			type is encountered while processing the 
     * 			input.
     */
    public
    MultipartMimeInput(InputStream source, ContentHeader contentHeader)
    throws MimeException
    {
	atEOF = false;
	String value = contentHeader.getValue();
	if (value == null) {
	    throw new MimeException("Missing content header value.",
				    MimeException.INVALID_HEADER);
	}
	if (!value.toLowerCase().trim().startsWith("multipart/")) {
	    throw new MimeException("Illegal mime type.",
				    MimeException.INVALID_MIME_TYPE);
	}
	String boundary = contentHeader.getParameter("boundary");
	if (boundary == null) {
	    throw new MimeException("Missing boundary parameter.",
				    MimeException.INVALID_HEADER);
	}
	
	//
	// Skip the first boundary.
	// 
	int skipped=0;
	try {
	    inputSource = new BMByteSearchStream(source, boundary, 2000);
	    skipped = inputSource.skipPattern();
	} catch (IOException e) {
	    throw new MimeException("Error while reading to first boundary: "+
				    e.toString(), MimeException.GENERIC);
	}
	if (skipped < 1) {
	    throw new MimeException("Boundary pattern missing in input " +
				    "stream.", MimeException.GENERIC);
	}
	
	//
	// Skip newline after first boundary.  Also determine whether
	// "\n\r" (correct), or "\n" (incorrect) is being used.  We're
	// nice and handle both.  The catch is that the user must be
	// consistently right or consistently buggy, not both.
	// 
	try {
	    switch (inputSource.read()) {
		case '\r':
		    if (inputSource.read() != '\n') {
			// This should be rare, but we'll try to accept it.
			newlineString = "\r";
		    } else {
			// Newline sequence is RFC2045-compliant (CR+LF).
			newlineString="\r\n";
		    }
		    break;
		case '\n':
		    // Input was sent by an off-standard unix browser. 
		    // Uses LF only.  We'll be nice and accept the input
		    // as long as "LF only" is done consistently.  Using
		    // CR+LF later will cause spurious CR's to be appended
		    // to input sections.
		    newlineString="\n";	// Incorrect but accept anyway.
		    break;
		default:
		    // If the boundary pattern is not followed immediately
		    // by a newline in any know form, then we just give up
		    // and return nothing.
		    atEOF = true;	// Corrupt stream - no valid parts.
		    throw new MimeException("Missing newline after boundary.",
					    MimeException.INVALID_HEADER);
	    }
	} catch (IOException e) {
	    throw new MimeException("Missing newline after boundary:" +
				    e.toString(),
				    MimeException.INVALID_HEADER);
	}
	inputSeparator = newlineString + "--" + boundary;
	inputPattern = new BMByteSearch(inputSeparator);
	inputSource.setPattern(inputPattern);
    }
    
    /**
     * Returns the next section of the Multipart MIME stream as a
     * MultipartMimeInputStream object.  Returns <code>null</code> if
     * the end of the input source has been reached.
     * <P>
     * <B>Note:</B><P>
     * Since Multipart MIME data flows in a single stream, calling
     * this method causes the previously returned MultipartMimeInputStream
     * to be automatically closed and its input to be skipped.
     * 
     * @return		A MultipartMimeInputStream object for the next
     * 			section of the multipart message.
     * @exception MimeException If an error occurs while skipping to
     * 			the next section of input.
     */
    public
    MultipartMimeInputStream nextPart()
    throws MimeException
    {
	try {
	    if (currentStream != null) {
		currentStream.close();
		if (currentStream.lastPart) atEOF = true;
	    }
	    if (atEOF) return null;
	    currentStream = new MultipartMimeInputStream(inputSource,
							 inputPattern);
	} catch (MimeEOFException e) {
	    return null;
	} catch (IOException ioe) {
	    throw new MimeException("IO Error between parts: " +
				    ioe.toString(), MimeException.GENERIC);
	}
	return currentStream;
    }
    
    /**
     * Closes the input source and the current
     * <code>MultipartMimeInputStream</code> object.  No more parts
     * will be returned by <code>nextPart</code>.
     * 
     * @exception MimeException If an error occurs while closing the
     * 		input stream.
     */
    public
    void close()
    throws MimeException
    {
	try {
	    currentStream.close();
	    currentStream = null;
	    inputSource.close();
	    inputSource = null;
	    atEOF = true;
	} catch (IOException ioe) {
	    throw new MimeException("MultipartMimeInput: " +
				    "IO Error during close: " + ioe.toString(),
				    MimeException.GENERIC);
	}
    }
}
