/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: MultipartMimeInputStream.java,v 1.8.2.1.2.1 2000/10/19 17:58:53 jasona Exp $
 */




package	com.lutris.mime;
import	com.lutris.util.*;
import	java.io.*;
import	java.util.*;

/**
 * Presents the current section of a multipart MIME stream as a distinct
 * InputStream object for which the end of the section is the end of the
 * stream.  MIME headers for the current section are extracted, parsed,
 * and made available via the <code>getMimeHeaders()<code> method.  The
 * stream itself begins at the first byte following the Mime header
 * section.  Closing an instance of <code>MultipartMimeInputStream</code>
 * causes it to skip data on the underlying input stream until the next
 * section is found, or end-of-input is reached.
 */
public class MultipartMimeInputStream extends InputStream
{
    /**
     * Boyer-Moore streaming pattern searcher that represents the input
     * source.  It searches its InputStream, up to a specified pattern,
     * returning the read data until the pattern is reached.
     */
    private BMByteSearchStream		inputSource;

    /**
     * The stream is closed if <code>true</code> or not closed if
     * <code>false</code>.
     */
    private boolean			closed;

    /**
     * Precomputed separator pattern search object.
     */
    private BMByteSearch		searchPattern;

    /**
     * Precomputed BM search pattern for "\n".  This should only
     * ever be computed once.
     */
    private static BMByteSearch		newlinePattern = null;

    /**
     * Byte to read into for single-byte read() operation.
     */
    private byte[]			readByte = new byte[1];

    /**
     * EOF indicated to prevent skipping end marker twice.
     */
    private boolean			atEOF = false;

    /**
     * Array of raw, unparsed Mime headers.
     */
    private String[]			rawHeaders = null;

    /**
     * Array of unparseable headers.
     */
    private String[]			garbageHeaders = null;

    /**
     * Table of all parsed headers.
     */
    private Hashtable			headers = null;

    /**
     * Boolean that indicates that this input stream was the
     * last part of the total multipart input.  This field is only to
     * be accessed by the MultipartMimeInput object.
     */
    boolean lastPart = false;

    /**
     * Creates a MultipartFormStream object from the given pattern search
     * stream object, separator, and end of line pattern.  This constructor
     * may only be called internally from within
     * <code>MultipartMimeInput</code> or its derived subclasses.
     * Other packages never create instances of this class directly.
     *
     * @param source		Input stream that can perform a
     * 				pattern serarch on the raw input stream.
     * @param sep		Separator pattern.
     * @exception IOException	If an I/O error occurs.
     * @exception MimeEOFException If at EOF and the current
     * 				section does not exist.
     */
    protected
    MultipartMimeInputStream(BMByteSearchStream source, BMByteSearch sep)
    throws IOException, MimeEOFException
    {
	int rawPos=0, garbagePos=0;
	int n;
	byte[] lineBuf = new byte[2048];
	inputSource = source;
	searchPattern = sep;

	// Create hash table for headers.
	headers = new Hashtable();

	// Only create the newline pattern once.
	if (newlinePattern == null) {
	    newlinePattern = new BMByteSearch("\n");
	}
	inputSource.setPattern(newlinePattern);

	//
	// Read and process the MIME header up to and including
	// the terminating blank line.
	//
        outer: while ((n = readAll(lineBuf)) > 0) {
	    if (lineBuf[n-1] == '\r') --n;  // Handle CR+LF case.
	    if (n <= 0) break outer; 	    // ..and recheck length.
	    String line = new String(bytesToChars(lineBuf,0,n));
	    String[] newRaw = new String[rawPos + 1];
	    for (int i=0; i<rawPos; i++) {
		newRaw[i] = rawHeaders[i];
		rawHeaders[i] = null;
	    }
	    newRaw[rawPos++] = line;
	    rawHeaders = newRaw; newRaw = null;
	    try {
		MimeHeader header = new MimeHeader(line);
		if ((header.getValue() == null)
		    || (header.getHeaderType() == null))
		{
		    // Not a parseable Mime header.  Save under garbage
		    // collection key in case caller wants to try to parse it.
		    String[] newGarbage = new String[garbagePos + 1];
		    for (int i=0; i<garbagePos; i++) {
			newGarbage[i] = garbageHeaders[i];
			garbageHeaders[i] = null;
		    }
		    newGarbage[garbagePos++] = line;
		    garbageHeaders = newGarbage; newGarbage = null;
		} else {
		    String name = header.getHeaderType().toLowerCase();
		    MimeHeader[] hdrs =(MimeHeader[]) headers.get(name);
		    int num = (hdrs == null) ? 0 : hdrs.length;
		    MimeHeader[] newHdrs = new MimeHeader[num+1];
		    for (int i=0; i<num; i++) {
			newHdrs[i] = hdrs[i];
			hdrs[i] = null;
		    }
		    newHdrs[num] = header;
		    hdrs = null;
		    headers.put(name, newHdrs);
		}
	    } catch (Exception e) {
		throw new IOException("Error while reading " +
			"MultipartMimeInputStream: " + e.toString());
	    }
	} /* outer */
	// Now, begin reading the stream.
	inputSource.setPattern(searchPattern);
    }

    /**
     * Reads the next byte of data from this input stream.  The value byte
     * is returned as an <code>int</code> in the range 0 to 255.  If no
     * byte is available because the end of the stream has been reached,
     * the value -1 is returned.  This method blocks until input data is
     * available, the end of the stream is detected, or an exception is
     * thrown
     *
     * @return			The next byte of data, or -1 if the end
     * 				of stream is reached.
     * @exception IOException	If an I/O error occurs.
     */
    public
    int read()
    throws IOException
    {
	int len = read(readByte, 0, 1);
	if (len != 1) return -1;
	return ((int)readByte[0] + 0x100) & 0xff;
    }

    /**
     * Reads up to <code>buffer.length</code> bytes of data from this input
     * stream into an array of bytes.  This method blocks until some input
     * is available
     *
     * @param buffer		The buffer into which data are read.
     * @return			The number of bytes actually read, or -1
     * 				if there are no more bytes because the
     * 				end of stream has been reached.
     * @exception IOException	If an I/O error occurs.
     */
    public
    int read(byte[] buffer)
    throws IOException
    {
	return(read(buffer, 0, buffer.length));
    }

    /**
     * Reads <code>length</code> bytes of data from this input stream into
     * an array of bytes.  This method blocks until some input is available.
     *
     * @param buffer		The buffer into which data are read.
     * @param offset		The start offset of the data.
     * @param length		The maximum number of bytes read.
     * @return			The total number of bytes read into the
     * 				buffer, or -1 if there are no more bytes
     * 				because the end of stream has been reached.
     * @exception IOException	If an I/O error occurs.
     */
    public
    int read(byte[] buffer, int offset, int length)
    throws IOException
    {
	checkOpen();
	int ret = inputSource.readTo(buffer, offset, length);
	switch (ret) {
	    case BMByteSearchStream.AT_PATTERN:
		inputSource.skipPattern();
		if (inputSource.peekAheadString(2).equals("--")) lastPart=true;
		inputSource.setPattern(newlinePattern);
		inputSource.skipPattern();
		atEOF = true;
		return(-1);
	    case BMByteSearchStream.EOF:
		return(-1);
	    default:
		if (ret < 0) return -1;
		return(ret);
	}
    }

    /**
     * Skips over and discards <code>n</code> bytes of data from this
     * input stream.  The <code>skip</code> method may, for a variety of
     * reasons, end up skipping over some smaller number of bytes,
     * possibly <code>0</code>.  The actual number of bytes skipped is
     * returned.
     *
     * @param num		The number of bytes to be skipped.
     * @return			The actual number of bytes skipped.
     *
     * @exception IOException	If an I/O error occurs.
     */
    public
    long skip(long num)
    throws IOException
    {
	checkOpen();
	byte[]	buf = new byte[1024];
	long	count=0, n;
	while (count < num) {
	    if (num - count > 1024)
		n = this.read(buf, 0, 1024);
	    else
		n = this.read(buf, 0, (int)(num - count));
	    if (n < 0) break;
	    count += n;
	}
	return count;
    }

    /**
     * Returns the number of bytes that can be read from this stream
     * without blocking.
     *
     * @return 			The number of bytes that can be read
     * 				from this stream without blocking.
     *
     * @exception IOException If an I/O error occurs.
     */
    public
    int available()
    throws IOException
    {
	checkOpen();
	return(inputSource.available());
    }

    /**
     * Skips all remaining bytes on this stream and closes it.  Further
     * operations on this stream, other than <code>close()</code> will
     * cause an <code>IOException</code> to be thrown.
     *
     * @exception IOException If an I/O error occurs.
     */
    public
    void close()
    throws IOException
    {
	if (!atEOF) {
	    inputSource.setPattern(searchPattern);
	    inputSource.skipPattern();
	    if (inputSource.peekAheadString(2).equals("--")) lastPart=true;
	    inputSource.setPattern(newlinePattern);
	    inputSource.skipPattern();
	}
	atEOF = true;
	closed = true;
    }

    /**
     * Returns a single <code>MimeHeader</code> object associated with
     * a given Mime header name.  If the selected header name is associated
     * with more than one value, then the last instance received from the input
     * stream is returned.
     *
     * @param	headerName	The name of the Mime header to return.
     * @return	The value of the header, or <code>null</code> if not found.
     */
    public
    MimeHeader getHeader(String headerName)
    {
	MimeHeader[] hdrs =
	    (MimeHeader[]) headers.get(headerName.toLowerCase());
	if (hdrs == null) return null;
	if (hdrs.length < 1) return null;
	return(hdrs[hdrs.length - 1]);
    }

    /**
     * Returns a array of type <code>MimeHeader</code> containing all values
     * associated with a given Mime header name.
     *
     * @param	headerName	The name of the Mime header to return.
     * @return	All values associated with the header, or
     * 		<code>null</code> if not found.
     */
    public
    MimeHeader[] getHeaders(String headerName)
    {
	MimeHeader[] hdrs =
	    (MimeHeader[]) headers.get(headerName.toLowerCase());
	if ((hdrs == null) || (hdrs.length < 1)) return null;
	return hdrs;
    }

    /**
     * Returns an array of <code>String</code> containing each Mime
     * header as it was read from the header of this section.  All
     * whitespace and punctuation is left intact.
     *
     * @return		Array of raw Mime headers.
     */
    public
    String[] getRawHeaders()
    {
	return rawHeaders;
    }

    /**
     * Returns an array of all Mime header lines that could not be
     * parsed in the normal "name: value" fashion.
     *
     * @return		Array of raw Mime headers.
     */
    public
    String[] getGarbageHeaders()
    {
	return garbageHeaders;
    }

    // ------ Private Utility Routines Below -------

    /**
     * Reads into a byte array until the array is full, EOF is reached,
     * or an I/O error occurs.  This private method is intended
     * only for reading individual lines of MIME headers.  If the
     * buffer is not long enough to hold an entire line from the input
     * then the remainder bytes are skipped.
     *
     * @param buffer		The buffer to read into.
     * @return			The number of bytes read.
     * @exception IOException	If an I/O error occurs.
     * @exception MimeEOFException If EOF reached in the MIME header.
     */
    private
    int readAll(byte[] buffer)
    throws IOException, MimeEOFException
    {
	String la;
	int pos = 0;
	int len = buffer.length;
	int n=0;

	while (pos < len) {
	    switch (n = inputSource.readTo(buffer, pos, len - pos)) {
		case BMByteSearchStream.EOF:
		    throw new MimeEOFException("EOF");
		case BMByteSearchStream.AT_PATTERN:
		    inputSource.skipPattern();
		    if (inputSource.peekAheadString(2).equals("--"))
			lastPart=true;
		    return pos;
		default:
		    pos += n;
		    break;
	    }
	}
	//
	// If the input line overflows the buffer, then the overflow
	// part is discarded.  Care should be taken to use a buffer
	// large enough to hold any reasonable Mime header line.
	//
	inputSource.skipPattern();
	return pos;
    }

    /**
     * Converts a byte array into a raw ASCII character array with
     * characters in the range \u0000 to \u00ff by twos-complement
     * signed-to-unsigned conversion.
     *
     * @param buffer		The byte array to convert.
     * @param off		Offset in buffer where conversion begins.
     * @param len		Number of bytes to convert.
     * @return			New character array containing the
     * 				converted bytes
     */
    private static final
    char[] bytesToChars(byte[] buffer, int off, int len)
    {
	char[] cbuf = new char[len];
	int src, dst;
	for (dst=0,src=off; dst < len; dst++,src++) {
	    cbuf[dst] = (char) (((int)(buffer[src]) + 0x100) & 0xff);
	}
	return cbuf;
    }

    /**
     * Used by routines that fail after closure to verify that the
     * current stream is open.  If closed, this method throws an
     * <code>IOException</code> to indicate an operation attempted
     * on a closed stream.
     *
     * @exception IOException	Thrown if the stream is closed.
     */
    private
    void checkOpen()
    throws IOException
    {
	if (!closed) return;
	throw new IOException("Operation on a closed stream.");
    }

}
