/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: Tester.java,v 1.11.12.1 2000/10/19 17:58:53 jasona Exp $
 */

package com.lutris.testutil;
import java.lang.reflect.Array;

/**
 * Class use to verify the results of tests.
 *
 * @version	$Revision: 1.11.12.1 $
 * @author	Mark Diekhans
 */
public class Tester {
    /**
     * Count of the number of failures that have occured.
     */
    public int errorCnt = 0;
  
    private static final String wrongTypeMsg =
        "Result object class does not match expected class";
    private static final String notEqualMsg =
        "Result value does not match expected value";
    private static final String resultNullMsg =
        "Result value null, expected non-null";
    private static final String expectNullMsg =
        "Result value non-null, expected null";
    private static final String arrayLenMsg =
        "Result array length not the same as expected length";
    
    private static final String bigBar = 
        "===================================================================";
    static final String smallBar = 
        "-------------------------------------------------------------------";
    
    /*
     * Class used to return a description of a failure.
     */
    class FailDesc {
        String msg;
        String resultStr;
        String expectStr;
        String arrayIdxStr;
        
        FailDesc(String msg,
                 String resultStr,
                 String expectStr) {
            this.msg = msg;
            this.resultStr = resultStr;
            this.expectStr = expectStr;
            arrayIdxStr = "";
        }
    }
  
    /*
     * Print out a failure message.
     */
    private void printFailure(String testId,
                              FailDesc failDesc) {
        System.err.println(bigBar);
        System.err.print("Test failure: " + testId);
        if (failDesc.arrayIdxStr.length () > 0) {
            System.err.print(" @ array index " + failDesc.arrayIdxStr);
        }
        System.err.println();
        System.err.println(failDesc.msg);
        System.err.println(smallBar);
        if (failDesc.resultStr != null) {
            System.err.println("Result: ");
            System.err.println(failDesc.resultStr);
            System.err.println(smallBar);
        }
        if (failDesc.expectStr != null) {
            System.err.println("Expected: ");
            System.err.println(failDesc.expectStr);
            System.err.println(smallBar);
        }
    }
    

    /*
     * Generate a type name for a class to use in error messages.
     */
    private String getTypeName(Class classObj) {
        if (classObj.isArray()) {
            return getTypeName(classObj.getComponentType()) + " []";
        } else {
            return classObj.getName();
        }
    }
    
    /*
     * Compare arrays of any depth.
     */
    private FailDesc compareArrays(Object result,
                                   Object expect) {
        if (Array.getLength(result) != Array.getLength(expect)) {
            return new FailDesc(arrayLenMsg,
                                "array length is " +
                                Array.getLength(result),
                                "array length is " +
                                 Array.getLength(expect));
        }
        for (int idx = 0; idx < Array.getLength(result); idx++) {
            FailDesc failDesc = compare (Array.get(result,idx),
                                         Array.get(expect, idx));
            if (failDesc != null) {
                // Add our array index to message
                failDesc.arrayIdxStr = "[" + idx + "]" + failDesc.arrayIdxStr;
                return failDesc;
            }
        }
        return null;
    }

    /*
     * Compare Throwable objects.
     */
    private FailDesc compareThrowable(Throwable result,
                                      Throwable expect) {
        if (result.getClass() != expect.getClass()) {
            return new FailDesc ("Wrong exception occured",
                                 result.getClass().getName () + ": " +
                                 result.getMessage(),
                                 expect.getClass().getName());
        }

        String resultMsg = ((Throwable) result).getMessage();
        String expectMsg = ((Throwable) expect).getMessage();
        if (!resultMsg.equals(expectMsg)) {
            return new FailDesc("Exception message wrong",
                                resultMsg,
                                expectMsg);
        }
        return null;
    }
    
    /*
     * Compare result and expected objects, which maybe arrays.  Returns null
     * if ok, description otherwise.
     */
    private FailDesc compare(Object result,
                             Object expect) {
        if ((result == null) || (expect == null)) {
            if (result != null) {
                return new FailDesc(resultNullMsg,
                                    null, null);
            } else if (expect != null) {
                return new FailDesc(expectNullMsg,
                                    null, null);
            }
            return null;  // Are the same
        }
        
        if (result.getClass() != expect.getClass()) {
            return new FailDesc(wrongTypeMsg,
                                getTypeName(result.getClass()),
                                getTypeName(expect.getClass()));
        }
        
        if (result.getClass().isArray()) {
            return compareArrays(result,
                                 expect);
        } else if (!result.equals(expect)) {
            return new FailDesc (notEqualMsg,
                                 result.toString(),
                                 expect.toString());
        }
        return null;  // Are the same
    }
    
    
    /**
     * Compare an object representing the result of a test to an expected
     * value.  Print an error message with details of the values of the
     * object if they don't compare.
     *
     * @param testID A description and id number that identify the test step.
     *  Printed on failures.
     * @param result An object representing the result.  If the object is
     *  derived from type Throwable, then its assumed that an exception
     *  occured.
     * @param expect An object to compare to the result of the test step.
     * @return <CODE>true</CODE> if test succeeded, <CODE>false</CODE>
     *  if it failed.  Generally not used unless test process needs to
     *  have flow altered on failure.
     */
    public boolean check(String testId,
                         Object result,
                         Object expect) {
        if (result instanceof Throwable) {
            unexpectedException(testId,
                                (Throwable)result);
            return false;
        } else {
            FailDesc failDesc = compare(result,
                                        expect);
            if (failDesc != null) {
                printFailure(testId,
                             failDesc);
                errorCnt++;
                return false;
            }
        }
        return true;
    }
    
    
    /**
     * Compare an object representing an exception that should have occured
     * to an result object.
     *
     * @param testID A description and id number that identify the test step.
     *  Printed on failures.
     * @param result An object representing the result.  If its not an
     *  exception, this is a failure.  Often set to null if an exception did
     *  not occur.
     * @param expect The expected exception.
     * @return <CODE>true</CODE> if test succeeded, <CODE>false</CODE>
     *  if it failed.  Generally not used unless test process needs to
     *  have flow altered on failure.
     */
    public boolean checkExcept(String testId,
                               Object result,
                               Throwable expect) {
        if (!(result instanceof Throwable)) {
            printFailure(testId,
                         new FailDesc("No exception occured",
                                      null,
                                      expect.getClass().getName()));
            errorCnt++;
            return false;
        } else {
            FailDesc failDesc = compareThrowable((Throwable)result,
                                                 expect);
            if (failDesc != null) {
                printFailure(testId,
                             failDesc);
                errorCnt++;
                return false;
            }
        }
        return true;
    }
    
    /**
     * Generate an error message on an unexpected exception. 
     *
     * @param testID A description and id number that identify the test step.
     *  Printed on failures.
     * @param except Expection.
     */
    public void unexpectedException(String testId,
                                    Throwable except) {
        printFailure(testId,
                     new FailDesc("Unexpected exception",
                                  null, null));
        System.err.println(except.getClass().getName() + ": " +
                           except.getMessage());
        except.printStackTrace(System.err);
        System.err.println(smallBar);
        errorCnt++;
    }

    /**
     * Generate a general failure message for cases not covered by
     * above functions.
     */
    public void failure(String testId,
                        String msg) {
        System.err.println(bigBar);
        System.err.println("Test failure: " + testId);
        System.err.println(msg);
        System.err.println(smallBar);
        errorCnt++;
    }
}
