/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: KeywordValueTableTests.java,v 1.11.8.1 2000/10/19 17:58:53 jasona Exp $
 */





package com.lutris.util.tests;
import	com.lutris.util.*;
import	com.lutris.testFramework.*;

/**
 * Tests for KeywordValue Table.
 *
 * @version	$Revision: 1.11.8.1 $
 * @author	Mark Diekhans
 * @since	Jolt1.0
 */
public class KeywordValueTableTests {
    /*
     * Merge two keyword parts into a single keyword.
     */
    static String mergeKey (String part1,
                            String part2) {
        if ((part1 == null) || (part1.length () == 0)) {
            return part2;
        }
        return part1 + "." + part2;
    }

    /*
     * Given a base keyword, run a series of tests.  The base has keywords
     * appended to it to form a path.  This is done so the same test can
     * be used against a single level table and a nested table.
     */
     static void basicTests (String baseDesc,
                             String baseKey,
                             Tester tester) {
         KeywordValueTable keyTab = new KeywordValueTable ();

         //------------------------------------------------------------------
         // Empty table
         try {
             tester.check (baseDesc + " #1.1",
                           keyTab.keys (),
                           new String [0]);
         } catch (Throwable except) {
             tester.unexpectedException (baseDesc + " #1.2",
                                         except);
         }

         //------------------------------------------------------------------
         // Key with string value.
         String key1 = mergeKey (baseKey,
                                 "key1");
         try {
             String [] key1Paths = KeywordParser.parse (key1);
             String [] expect = {key1Paths [0]};

             keyTab.set (key1,
                         "value1");
             tester.check (baseDesc + " #2.1",
                           keyTab.keys (),
                           expect);
         } catch (Throwable except) {
             tester.unexpectedException (baseDesc + " #2.2",
                                         except);
         }
         

         //------------------------------------------------------------------
         // Verify full path.
         try {
             String [] expect = {key1};
             tester.check (baseDesc + " #3.1",
                           keyTab.leafKeys (),
                           expect);
         } catch (Throwable except) {
             tester.unexpectedException (baseDesc + " #3.2",
                                         except);
         }
         

         //------------------------------------------------------------------
         // Check value.
         try {
             tester.check (baseDesc + " #4.1",
                           (String) keyTab.get (key1),
                           "value1");
         } catch (Throwable except) {
             tester.unexpectedException (baseDesc + " #4.2",
                                         except);
         }

         //------------------------------------------------------------------
         // Check value using getString.
         try {
             tester.check (baseDesc + " #5.1",
                           keyTab.getString (key1),
                           "value1");
         } catch (Throwable except) {
             tester.unexpectedException (baseDesc + " #5.2",
                                         except);
         }

         //------------------------------------------------------------------
         // Second key, string value.
         String key2 = mergeKey (baseKey,
                                 "key2");
         try {
             keyTab.set (key2,
                         "value2");
             String [] expect  = {key1, key2};
             String [] result;

             // Try both ways to test interfaces.
             if (baseKey != null) {
                 result = (String []) keyTab.leafKeys ();
             } else {
                 result = (String []) keyTab.keys ();
             }
             
             tester.check (baseDesc + " #6.1",
                           Sorter.sort (result),
                           expect);
         } catch (Throwable except) {
             tester.unexpectedException (baseDesc + " #6.2",
                                         except);
         }

         //------------------------------------------------------------------
         // Check values.
         try {
             String result [] = new String [2];
             result [0] = (String) keyTab.get (key1);
             result [1] = (String) keyTab.get (key2);
         
             String [] expect = {"value1", "value2"};
             tester.check (baseDesc + " #7.1",
                           Sorter.sort ((String []) result),
                           expect);
         } catch (Throwable except) {
             tester.unexpectedException (baseDesc + " #7.2",
                                         except);
         }

         //------------------------------------------------------------------
         // Float value.
         String key3 = mergeKey (baseKey,
                                 "key3");
         try {
             keyTab.set (key3,
                         new Float (100.001));
             String [] expect;
             if (baseKey == null) {
                 expect = new String [] {key1, key2, key3};
             } else {
                 expect = new String [] {KeywordParser.parse (key1) [0]};
             }
             tester.check (baseDesc + " #8.1",
                           Sorter.sort ((String []) keyTab.keys ()),
                           expect);
         } catch (Throwable except) {
             tester.unexpectedException (baseDesc + " #8.2",
                                         except);
         }

         //------------------------------------------------------------------
         // Try full key paths.
         try {
             keyTab.set (key3,
                         new Float (100.001));
         
             String [] expect = {key1, key2, key3};
             tester.check (baseDesc + " #9.1",
                           Sorter.sort ((String []) keyTab.leafKeys ()),
                           expect);
         } catch (Throwable except) {
             tester.unexpectedException (baseDesc + " #9.2",
                                         except);
         }

         //------------------------------------------------------------------
         // Check value.
         try {
             tester.check (baseDesc + " #10.1",
                           keyTab.get (key3),
                           new Float (100.001));
         } catch (Throwable except) {
             tester.unexpectedException (baseDesc + " #10.2",
                                         except);
         }

         //------------------------------------------------------------------
         // Second key, string value.
         String key4 = mergeKey (baseKey,
                                 "key4");
         try {
             keyTab.set (key4,
                         new Float (100.002));
             String [] expect  = {key1, key2, key3, key4};
             tester.check (baseDesc + " #11.1",
                           Sorter.sort ((String []) keyTab.leafKeys ()),
                           expect);

         } catch (Throwable except) {
             tester.unexpectedException (baseDesc + " #11.2",
                                         except);
         }
         
         //------------------------------------------------------------------
         // Check values.
         try {
             Float result [] = new Float [2];
             result [0] = (Float) keyTab.get (key3);
             result [1] = (Float) keyTab.get (key4);
         
             Float [] expect = {new Float (100.001),
                                new Float (100.002)};
             tester.check (baseDesc + " #12.1",
                           Sorter.sort ((Float []) result),
                           expect);
         } catch (Throwable except) {
             tester.unexpectedException (baseDesc + " #12.2",
                                         except);
         }

         //------------------------------------------------------------------
         // Section access, but only if a base key is creating nested tables.
         if (baseKey != null) {
             try {
                 KeywordValueTable section = keyTab.getSection (baseKey);
                 String [] expect = {"key1", "key2", "key3", "key4"};
                 
                 tester.check (baseDesc + " #13.1",
                               Sorter.sort (section.keys ()),
                               expect);
             } catch (Throwable except) {
                 tester.unexpectedException (baseDesc + " #13.2",
                                             except);
             }
             
         }

         //------------------------------------------------------------------
         // Test updating values.
         try {
             keyTab.set (key1, "new-value1");

             tester.check (baseDesc + " #14.1",
                           keyTab.get (key1),
                            "new-value1");
         } catch (Throwable except) {
             tester.unexpectedException (baseDesc + " #14.2",
                                         except);
         }

         //------------------------------------------------------------------
         try {
             keyTab.set (key2, new Double (3.14159));

             tester.check (baseDesc + " #15.1",
                           keyTab.get (key2),
                           new Double (3.14159));
         } catch (Throwable except) {
             tester.unexpectedException (baseDesc + " #15.2",
                                         except);
         }

         //------------------------------------------------------------------
         // Converting a type to a section.
         try {
             String key2b = KeywordParser.concat (key2,
                                                  "subkey2.subkey3.subkey4");
             keyTab.set (key2b, new Double (6248.1));

             tester.check (baseDesc + " #16.1",
                           keyTab.get (key2b),
                           new Double (6248.1));
         } catch (Throwable except) {
             tester.unexpectedException (baseDesc + " #16.2",
                                         except);
         } 
         
         //------------------------------------------------------------------
         try {
             tester.check (baseDesc + " #17.1",
                           keyTab.get (key2).getClass (),
                           keyTab.getClass ());
         } catch (Throwable except) {
             tester.unexpectedException (baseDesc + " #17.2",
                                         except);
         } 

         //------------------------------------------------------------------
         try {
             tester.check (baseDesc + " #18.1",
                           new Boolean (keyTab.containsKey (key1)),
                           new Boolean (true));
         } catch (Throwable except) {
             tester.unexpectedException (baseDesc + " #18.2",
                                         except);
         }

         
         //------------------------------------------------------------------
         try {
             tester.check (baseDesc + " #19.1",
                           new Boolean (keyTab.containsKey ("fookey1")),
                           new Boolean (false));
         } catch (Throwable except) {
             tester.unexpectedException (baseDesc + " #19.2",
                                         except);
         }

         
         //------------------------------------------------------------------
         // Deleting fields.
         try {
             tester.check (baseDesc + " #20.1",
                           new Boolean (keyTab.containsKey (key3)),
                           new Boolean (true));
             keyTab.remove (key3);
             tester.check (baseDesc + " #20.2",
                           new Boolean (keyTab.containsKey (key3)),
                           new Boolean (false));
             keyTab.remove (key3);
             tester.check (baseDesc + " #20.3",
                           new Boolean (keyTab.containsKey (key3)),
                           new Boolean (false));
         } catch (Throwable except) {
             tester.unexpectedException (baseDesc + " #20.4",
                                         except);
         }

         //------------------------------------------------------------------
         try {
             String key4b1 = KeywordParser.concat (key4,
                                                   "subkey5");
             String key4b2 = KeywordParser.concat (key4,
                                                   "subkey5.subkey6");
             String key4b3 = KeywordParser.concat (key4,
                                                   "subkey5.subkey6.subkey7");
             keyTab.set (key4b3,
                         "Long key here");
             tester.check (baseDesc + " #21.1",
                           new Boolean (keyTab.containsKey (key4b3)),
                           new Boolean (true));
             tester.check (baseDesc + " #21.2",
                           new Boolean (keyTab.containsKey (key4b2)),
                           new Boolean (true));
             tester.check (baseDesc + " #21.3",
                           new Boolean (keyTab.containsKey (key4b1)),
                           new Boolean (true));

             keyTab.remove (key4b2);
             keyTab.remove (key4b2);  // No-op

             tester.check (baseDesc + " #21.4",
                           new Boolean (keyTab.containsKey (key4b3)),
                           new Boolean (false));
             tester.check (baseDesc + " #21.5",
                           new Boolean (keyTab.containsKey (key4b2)),
                           new Boolean (false));
             tester.check (baseDesc + " #21.6",
                           new Boolean (keyTab.containsKey (key4b1)),
                           new Boolean (true));
         } catch (Throwable except) {
             tester.unexpectedException (baseDesc + " #21.7",
                                         except);
         }

         //------------------------------------------------------------------
         // Access to missing fields.
         try {
             tester.check (baseDesc + " #22.1",
                           keyTab.get ("nokey"),
                           null);
         } catch (Throwable except) {
             tester.unexpectedException (baseDesc + " #22.2",
                                         except);
         }

         try {
             tester.check (baseDesc + " #22.3",
                           keyTab.get ("nokey", "Foo!"),
                           "Foo!");
         } catch (Throwable except) {
             tester.unexpectedException (baseDesc + " #22.4",
                                         except);
         }

     }

    /**
     * Test error handling.
     */
    static void errorTests (String baseDesc,
                            String baseKey,
                            Tester tester) {
        KeywordValueTable keyTab = new KeywordValueTable ();

        //------------------------------------------------------------------
        // Non-existant field
        String key2a = mergeKey (baseKey,
                                 "key2");
        String key2b = mergeKey (baseKey,
                                 "key2.subkey2");
        int level = 0;
        try {
            level = KeywordParser.parse (key2b).length - 2;
            keyTab.set (key2a,
                        new Double (1.2));
            tester.check (baseDesc + " #2.1",
                          keyTab.get (key2a),
                          new Double (1.2));
            keyTab.get (key2b);
            throw new Throwable (baseDesc + ": #2.2 should have failed");
        } catch (Throwable except) {
            String msg = "keyword specifies a non-leaf component that is " +
                "not a KeywordValueTable: " + key2b + " (component #" +
                level + ")";
            tester.checkExcept (baseDesc + " #2.3",
                                except,
                                new KeywordValueException (msg));
        }

        //------------------------------------------------------------------
        // Illegal keywords.
        String key3 = mergeKey (baseKey,
                                "");
        try {
            keyTab.set (key3,
                        new Double (1.2));
            throw new Throwable (baseDesc + ": #3.1 should have failed");
        } catch (Throwable except) {
            String msg = "keyword component must be single word or words " +
                "separated by '.': \"" + key3 + "\"";
            tester.checkExcept (baseDesc + " #3.2",
                                except,
                                new KeywordValueException (msg));
        }

        //------------------------------------------------------------------
        String key4 = ".";
        try {
            keyTab.set (key4,
                        new Double (1.2));
            throw new Throwable (baseDesc + ": #4.1 should have failed");
        } catch (Throwable except) {
            String msg =
                "keyword should not start with a '.' separator: \".\"";
            tester.checkExcept (baseDesc + " #4.2",
                                except,
                                new KeywordValueException (msg));
        }

        //------------------------------------------------------------------
        String key5 = ".a";
        try {
            keyTab.set (key5,
                        new Double (1.2));
            throw new Throwable (baseDesc + ": #5.1 should have failed");
        } catch (Throwable except) {
            String msg = "keyword component must be single word or words " +
                "separated by '.': \".a\"";
            tester.checkExcept (baseDesc + " #5.2",
                                except,
                                new KeywordValueException (msg));
        }

        //------------------------------------------------------------------
        String key6 = "";
        try {
            keyTab.set (key6,
                        new Double (1.2));
            throw new Throwable (baseDesc + ": #6.1 should have failed");
        } catch (Throwable except) {
            String msg = "keyword component must be single word or words " +
                "separated by '.': \"\"";
            tester.checkExcept (baseDesc + " #6.2",
                                except,
                                new KeywordValueException (msg));
        }

        //------------------------------------------------------------------
        String key7 = mergeKey (baseKey,
                                "a..b");
        try {
            keyTab.set (key7,
                        new Double (1.2));
            throw new Throwable (baseDesc + ": #7.1 should have failed");
        } catch (Throwable except) {
            String msg = "keyword component must be single word or words " +
                "separated by '.': \"" + key7 + "\"";
            tester.checkExcept (baseDesc + " #7.2",
                                except,
                                new KeywordValueException (msg));
        }

        //------------------------------------------------------------------
        String key8 = mergeKey (baseKey,
                                "102B");
        try {
            keyTab.set (key8,
                        new Double (1.2));
            throw new Throwable (baseDesc + ": #8.1 should have failed");
        } catch (Throwable except) {
            String msg = "keyword component must be a legal Java identifier " +
                "component \"102B\": \"" + key8 + "\"";
            tester.checkExcept (baseDesc + " #8.2",
                                except,
                                new KeywordValueException (msg));
        }

        //------------------------------------------------------------------
        String key9 = mergeKey (baseKey,
                                "A#b");
        try {
            keyTab.set (key9,
                        new Double (1.2));
            throw new Throwable (baseDesc + ": #9.1 should have failed");
        } catch (Throwable except) {
            String msg = "keyword component must be a legal Java identifier " +
                "component \"A#b\": \"" + key9 + "\"";
            tester.checkExcept (baseDesc + " #9.2",
                                except,
                                new KeywordValueException (msg));
        }

        //------------------------------------------------------------------
        String key10 = mergeKey (baseKey,
                                 "A.");
        try {
            keyTab.set (key10,
                        new Double (1.2));
            throw new Throwable (baseDesc + ": #10.1 should have failed");
        } catch (Throwable except) {
            String msg = "keyword component must be single word or words " +
                "separated by '.': \"" + key10 + "\"";
            tester.checkExcept (baseDesc + " #10.2",
                                except,
                                new KeywordValueException (msg));
        }
    }

        
        
    public static void runTests () {
        Tester tester = new Tester ();
        
        basicTests ("Simple table",
                    null,
                    tester);
        basicTests ("One level nested table",
                    "Top1",
                    tester);
        basicTests ("Two level nested table",
                    "Top1.mid1",
                    tester);
        errorTests ("Simple error tests",
                    null,
                    tester);
        errorTests ("One level error tests",
                    "Top1",
                    tester);
        errorTests ("Two level error tests",
                    "Top1.mid1",
                    tester);

        if (tester.errorCnt > 0) {
            System.err.println ("\n **** " + tester.errorCnt +
                                " keyword/value table tests failed ****\n");
            System.exit (1);
        } else {
            System.out.println ("\nAll keyword/value table tests succeeded :-)\n");
        }
    }

    public static void main (String args []) {
        runTests();
    }

}
