/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * $Id: DOTreeNode.java,v 1.10.14.1 2000/12/01 22:27:39 chip Exp $
 *-----------------------------------------------------------------------------
 */



package com.lutris.dods.gui.tree;

import com.lutris.dods.store.doa.*;
import java.util.Vector;
import java.util.ResourceBundle;
/**
 * This class provides a wrapper for DataObjects and DOPackages in the treeview.
 *
 * @author  Philip Martin
 * @version $Revision: 1.10.14.1 $
 */
public class DOTreeNode {
    /**
     * The DataObject in this node.
     */
    protected DataObject DO;
    /**
     * All of the Objects which descend from this node in the tree.
     */
    protected Vector children;
    /**
     * true when this is a wrapper for a DOPackage.
     */
    protected boolean folder;
    /**
     * This constructor creates a wrapper for the given DataObject which is
     * not a folder.
     */
    public DOTreeNode(DataObject d) {
        this(d, false);
    }
    /**
     * This Constructor creates a wrapper for the given DataObject which is
     * either a Folder or not.
     */
    public DOTreeNode(DataObject d, boolean isFolder) {
        DO = d;
        folder = isFolder;
        children = new Vector();

        if(DO instanceof DOPackage) {
//System.err.println("DOTreeNode_ctor");
            DOPackage DOP = (DOPackage)DO;
//System.err.println("DOP_getDataObjects"+DOP.getDataObjects().size());
            for(int i = 0; i < DOP.getDataObjects().size(); i++) {
                DataObject daob = (DataObject)DOP.getDataObjects().elementAt(i);
                if(daob instanceof DOPackage) {
                    children.addElement(new DOTreeNode(daob, true));
                }
                else {
                    children.addElement(new DOTreeNode(daob));
                }
            }
        }
    }
    /**
     * @return the DataObject contained in this wrapper.
     */
    public DataObject getDataObject() {
        return DO;
    }
    /**
     * @return the vector containing the node's children.
     */
    public Vector getChildren() {
        if(folder) {
            return children;
        }
        return null;
    }
    /**
     * The String format for these nodes is just the name of the dataobject
     * it is for.
     * @return the name of the DataObject in this node.
     */
    public String toString() {
        return DO.getName();
    }
    /**
     * @return !folder.
     */
    public boolean isLeaf() {
        return !folder;
    }

    /**
     * This method looks through all of the tree which descends from this
     * node and finds the node containing the given DataObject in its children.
     * @return The node containing the given DataObject as a child or null.
     */
    public DOTreeNode getParentWithChildContaining(DataObject data) {
        for(int i = 0; i < children.size(); i++) {
            DOTreeNode dotn = (DOTreeNode)children.elementAt(i);
            if(dotn.getDataObject().equals(data)) {
                return this;
            }
        }
        for(int i = 0; i < children.size(); i++) {
            DOTreeNode dotn = (DOTreeNode)children.elementAt(i);
            DOTreeNode returnvalue = dotn.getParentWithChildContaining(data);
            if(returnvalue != null) {
                return returnvalue;
            }
        }
        return null;
    }


  public DOTreeNode getNodeContaining(Object dataObject) {
    DOTreeNode node = null;
    if (getDataObject().equals(dataObject)) {
      node = this;
    } else {
      if (getChildren() != null) {
        int count = getChildren().size();
        for (int i = 0 ; i < count ; i++) {
          node = (DOTreeNode) getChildren().elementAt(i);
          if (node.getDataObject().equals(dataObject)) {
            break;
          } else {
            node = node.getNodeContaining(node);
            if (node != null) {
              break;
            }
          }
        }
      }
    }
    return node;
  }


    /**
     * This method updates the order of the children so that all of the leaves
     * are before all of the folders.  It then calls sortChildren to put them
     * in alphabetical order.
     */
    public void updateChildrenOrder() {
        Vector newChildren = new Vector();
        for(int i = 0; i < children.size(); i++) {
            DOTreeNode insert = (DOTreeNode)children.elementAt(i);
            if(insert.isLeaf()){
                newChildren.insertElementAt(insert, 0);
            }
            else {
                newChildren.addElement(insert);
                insert.updateChildrenOrder();
            }
        }
        children = newChildren;
        sortChildren();
    }
    /**
     * This method puts the children in alpha order without disturbing
     * The separation of leaves and Folders.
     */
    private void sortChildren() {
        if(children.size() < 2) {
            return;
        }
        Vector newChildren = new Vector();
        DOTreeNode insert = (DOTreeNode)children.elementAt(0);
        int i = 0;
        while(insert.isLeaf()) {
            if(newChildren.size()==0) {
                newChildren.addElement(insert);
                i++;
                if(i==children.size()) {
                    children = newChildren;
                    return;
                }
                insert = (DOTreeNode)children.elementAt(i);
                continue;
            }
            if(newChildren.size()==1) {
                DataObject setData = ((DOTreeNode)newChildren.elementAt(0)).getDataObject();
                if(insert.getDataObject().getName().compareTo(setData.getName()) < 0) {
                    newChildren.insertElementAt(insert, 0);
                }
                else {
                    newChildren.addElement(insert);
                }
                i++;
                if(i==children.size()) {
                    children = newChildren;
                    return;
                }
                insert = (DOTreeNode)children.elementAt(i);
                continue;
            }
            for(int j = 1; j < newChildren.size(); j++) {
                DataObject setData1 = ((DOTreeNode)newChildren.elementAt(j-1)).getDataObject();
                DataObject setData2 = ((DOTreeNode)newChildren.elementAt(j)).getDataObject();
                if(insert.getDataObject().getName().compareTo(setData1.getName())<0) {
                    newChildren.insertElementAt(insert, 0);
                    i++;
                    if(i==children.size()) {
                        children = newChildren;
                        return;
                    }
                    insert = (DOTreeNode)children.elementAt(i);
                    break;
                }
                else if(insert.getDataObject().getName().compareTo(setData2.getName())<0) {
                    newChildren.insertElementAt(insert, j);
                    i++;
                    if(i==children.size()) {
                        children = newChildren;
                        return;
                    }
                    insert = (DOTreeNode)children.elementAt(i);
                    break;
                }
                else if(j == newChildren.size()-1) {
                    newChildren.addElement(insert);
                    i++;
                    if(i==children.size()) {
                        children = newChildren;
                        return;
                    }
                    insert = (DOTreeNode)children.elementAt(i);
                }
            }
        }
        int leafCount = newChildren.size();
        while(!insert.isLeaf()) {
            if(newChildren.size()==leafCount) {
                newChildren.addElement(insert);
                i++;
                if(i==children.size()) {
                    children = newChildren;
                    return;
                }
                insert = (DOTreeNode)children.elementAt(i);
                continue;
            }
            if(newChildren.size()==leafCount+1) {
                DataObject setData = ((DOTreeNode)newChildren.elementAt(leafCount)).getDataObject();
                if(insert.getDataObject().getName().compareTo(setData.getName())<0) {
                    newChildren.insertElementAt(insert, leafCount);
                }
                else {
                    newChildren.addElement(insert);
                }
                i++;
                if(i==children.size()) {
                    children = newChildren;
                    return;
                }
                insert = (DOTreeNode)children.elementAt(i);
                continue;
            }
            for(int j = leafCount + 1; j < newChildren.size(); j++) {
                DataObject setData1 = ((DOTreeNode)newChildren.elementAt(j-1)).getDataObject();
                DataObject setData2 = ((DOTreeNode)newChildren.elementAt(j)).getDataObject();
                if(insert.getDataObject().getName().compareTo(setData1.getName())<0) {
                    newChildren.insertElementAt(insert, leafCount);
                    i++;
                    if(i==children.size()) {
                        children = newChildren;
                        return;
                    }
                    insert = (DOTreeNode)children.elementAt(i);
                    break;
                }
                else if(insert.getDataObject().getName().compareTo(setData2.getName())<0) {
                    newChildren.insertElementAt(insert, j);
                    i++;
                    if(i==children.size()) {
                        children = newChildren;
                        return;
                    }
                    insert = (DOTreeNode)children.elementAt(i);
                    break;
                }
                else if(j == newChildren.size()-1) {
                    newChildren.addElement(insert);
                    i++;
                    if(i==children.size()) {
                        children = newChildren;
                        return;
                    }
                    insert = (DOTreeNode)children.elementAt(i);
                }
            }
        }
    }
    /**
     * This method returns true if this node or any node that descends
     * from this node has the given node as a child.
     */
    public boolean contains(DOTreeNode node) {
        boolean returnValue = false;
        if(children.contains(node)) {
            return true;
        }
        int i = 0;
        while(returnValue == false && i < children.size()) {
            DOTreeNode node2 = (DOTreeNode)children.elementAt(i);
            returnValue = node2.contains(node);
            i++;
        }
        return returnValue;
    }
    /**
     * This method returns true if this node or any node that descends
     * from this node has the given DataObject in a node as a child.
     */
    public boolean contains(DataObject data) {
        boolean returnValue = false;
        if(DO.equals(data)) {
            returnValue = true;
        }
        int i = 0;
        while(returnValue == false && i < children.size()) {
            DOTreeNode node2 = (DOTreeNode)children.elementAt(i);
            returnValue = node2.contains(data);
            i++;
        }
        return returnValue;
    }
}
