/*
 * -----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement:
 * This product includes Enhydra software developed by Lutris
 * Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 * -----------------------------------------------------------------------------
 * $Id:
 * -----------------------------------------------------------------------------
 */
package com.lutris.dods.gui.tree;

import com.lutris.dods.store.doa.DOPackage;
import com.lutris.dods.store.doa.DataObject;
import java.util.Vector;
import javax.swing.tree.TreePath;
import javax.swing.tree.DefaultMutableTreeNode;

/**
 * This class provides a wrapper for displaying DOPackages in a treeview.
 * 
 * @author  Paul Mahar
 * @version $Revision: 1.2 $
 */
public class PackageTreeNode extends DefaultMutableTreeNode {
    private DOPackage exclude = null;

    /**
     * Invoking this construtor will cause this node to be created as
     * well as all childs that belong to it.
     *
     *
     * @param thisPackage
     * The package display in a tree.
     */
    public PackageTreeNode(DOPackage thisPackage) {
	this(thisPackage, null);
    }

    /**
     * Invoking this construtor will cause this node to be created as
     * well as all child nodes that are not equal to a given excluded package.    
     *
     *
     * @param thisPackage
     * The package to display in a tree.
     * @param ex
     * The package to exclude when populating child nodes.
     *  
     */
    public PackageTreeNode(DOPackage thisPackage, DOPackage ex) {
	super(thisPackage, true);

	exclude = ex;

	populate();
    }

    /**
     * Create the child nodes.
     */
    private void populate() {
	DOPackage thisPackage = (DOPackage) this.getUserObject();
	Vector    dataObjects = thisPackage.getDataObjects();
	int       count = dataObjects.size();

	for (int i = 0; i < count; i++) {
	    if (dataObjects.elementAt(i) instanceof DOPackage) {
		DOPackage newPackage = (DOPackage) dataObjects.elementAt(i);

		if ((exclude == null) || (!exclude.equals(newPackage))) {
		    PackageTreeNode newNode = null;

		    newNode = new PackageTreeNode(newPackage, exclude);

		    add(newNode);
		} 
	    } 
	} 
    } 

    /**
     * Find a node containing a given user object. The found node may be this
     * node or and descendents of this node.
     *
     * @param userObject
     * The object to look for.
     *
     * @return
     * A node containing the userObject.
     */
    public PackageTreeNode getNodeContaining(Object userObject) {
	PackageTreeNode node = null;

	if (getUserObject().equals(userObject)) {
	    node = this;
	} else {
	    int count = getChildCount();

	    for (int i = 0; i < count; i++) {
		node = (PackageTreeNode) getChildAt(i);

		if (node.getUserObject().equals(userObject)) {
		    break;
		} else {
		    node = node.getNodeContaining(node);

		    if (node != null) {
			break;
		    } 
		} 
	    } 
	} 

	return node;
    } 

    /**
     * Get a tree path for this node.
     * 
     * @return
     * A tree path you can use to make selections in a JTree.
     */
    public TreePath getTreePath() {
	return new TreePath(getPath());
    } 

    /**
     * Get the string to show in a tree.
     *
     * @return
     * This text is displayed by the default cell renderer.
     */
    public String toString() {
	DOPackage thisPackage = (DOPackage) this.getUserObject();

	return thisPackage.getName();
    } 

}


