/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 */

package com.lutris.dods.store.doml;

import java.io.InvalidClassException;
import java.io.File;
import java.io.FileInputStream;
import java.io.ObjectInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import org.w3c.dom.*;

/**
 * This class is mostly an example use of the
 * <b>com.lutris.dods.project.doml</b> package. It
 * contains just a main method, which can be used for converting an existing
 * DOA file into a DOML file, or a DOML file of any format into a nicely
 * humanly-readable, indented, but possibly incorrect DOML file.
 *
 * This file refers to both the old style dods package
 * <code>dods.data</code> and the new package <code>
 * com.lutris.dods.store.doa</code>
 *
 * @author Jason Abbott (jason@lutris.com)
 */
public class DOMLFactory {

  private static void printWarnings(DOMLWarning warnings[]) {
    for (int i = 0; i<warnings.length; i++) {
      System.err.println("Warning: " + warnings[i].getMessage());
      Throwable chainedException = warnings[i].getCause();
      if (chainedException != null)
      System.err.println("         " + chainedException.toString());
    }
  }

  public static com.lutris.dods.store.doa.DODSProject readDOML(File file) {
    com.lutris.dods.store.doa.DODSProject project = null;
    DOMLProjectFactory projFactory = new DOMLProjectFactory();

    try {
      project = projFactory.createProject(file);
      if (projFactory.hasWarnings())
      printWarnings(projFactory.getWarnings());
    } catch (Exception e) {
      System.err.println("ERROR: Failed to read DOML file " +
                 file.getAbsolutePath() + ", " + e.toString());
      e.printStackTrace();
    }
    if (project == null) {
      System.err.println("ERROR: DOML file " +
                 file.getAbsolutePath() +
                 " creates null project");
    }

    return project;
  }


  private static void writeDOA(
     com.lutris.dods.store.doa.DODSProject project,
     File                                  outFile) {

    try {
      FileOutputStream   fOutput = new FileOutputStream(outFile);
      ObjectOutputStream oOutput = new ObjectOutputStream(fOutput);

      oOutput.writeObject(project);
    } catch (Exception e) {
      System.err.println("Error: Trouble handling file " +
                 outFile.getAbsolutePath());
      System.err.println("       " + e.toString());
      e.printStackTrace();
    }
  }


  public static void writeDOML(com.lutris.dods.store.doa.DODSProject project,
                               File        outFile,
                               boolean     indent) {
    DOMLDocumentFactory docFactory = new DOMLDocumentFactory();
    Document             doc = null;

    try {
      doc = docFactory.createDocument(project);
    } catch (DOMLException e) {
      System.err.println(
         "ERROR: Failed to create document, " + e.toString());
      e.printStackTrace();
    }

    try {
      FileOutputStream out = new FileOutputStream(outFile);
      DOMLWriter      writer = new DOMLWriter(out, false, indent, false);
      writer.print(doc);
    } catch (Exception e) {
      System.err.println(
        "ERROR: Failed to write file "
        + outFile.getAbsolutePath()
        + ", " +   e.toString());
      e.printStackTrace();
    }
  }

  /**
   * This method is mostly an example use of the
   * <b>com.lutris.dods.project.doml</b> package,
   * which shows how to convert an existing DOA file into a DOML file, or
   * a DOML file of any format into a nicely humanly-readable, indented,
   * but possibly incorrect DOML file.
   * <p>
   * Usage:
   * <dl>
   *  <dt><code>java com.lutris.dods.project.doml.DOMLFactory
   *      file.doa</code>
   *  <dd> converts a DOA file into a DOML file of the same name
   *  <dt><code>java com.lutris.dods.project.doml.DOMLFactory -i
   *      file.doml file2.doml</code>
   *  <dd> converts a DOML file into a humanly readable DOML file, with
   *       possibly illegal DOML.DTD structure.
   *  <dt><code>java com.lutris.dods.project.doml.DOMLFactory
   *      file.doml</code>
   *  <dd> converts a DOML file into a DOA file.
   * </dl>
   *
   * @param args
   *  This String array should contain one and only one element, the name of
   *  the file to parse.
   */
  public static void main(String[] args) {
    boolean indent = false;
    int     currentArg = 0;
    com.lutris.dods.store.doa.DODSProject project = null;

    if (args == null || args.length == 0) {
      System.err.println("Usage: DOMLFactory [-i] <readFile> [<writeFile>]");
      System.exit(1);
    }


    if (args[currentArg].equals("-i")) {
      currentArg++;
      indent = true;
    }

    File fromFile = new File(args[currentArg]);
    currentArg++;

    String fromType = com.lutris.dods.util.Util.getExtension(fromFile);
    if (fromType == null) {
      System.err.println("ERROR: Can't handle file " +
                 fromFile.getAbsolutePath() +
                 ", no \".\" in the file name");
      System.exit(1);
    }

    File   toFile = null;
    String toType = "doml";

    if (args.length > currentArg) {
      toFile = new File(args[currentArg]);
      toType = com.lutris.dods.util.Util.getExtension(toFile);
      if (toType == null) {
        System.err.println("ERROR: Can't handle file " +
                   toFile.getAbsolutePath() +
                   ", no \".\" in the file name");
        System.exit(1);
      }
      currentArg++;
    } else {
      String fileBase = com.lutris.dods.util.Util.getBase(fromFile);
      if (fromType.equals("doml")) {
        toType = "doa";
      }
      toFile = new File(fromFile.getParent(), fileBase + "." + toType);
    }

    if (fromType.equals("doa")) {
      System.err.println(
          "ERROR: Use the dods.doml.Upgrade class to convert DOA files to DOML files");
      System.exit(1);
    } else if (fromType.equals("doml")) {
      project = readDOML(fromFile);
    } else {
      System.err.println("ERROR: File " + fromFile.getAbsolutePath() +
                 " format " + fromType + " unknown");
      System.exit(1);
    }

    if (toType.equals("doa")) {
      writeDOA(project, toFile);
    } else if (toType.equals("doml")) {
      writeDOML(project, toFile, indent);
    } else {
      System.err.println("ERROR: File " + toFile.getAbsolutePath() +
                 " format " + toType + " unknown");
      System.exit(1);
    }
    System.out.println("Done");
  }
}
