
/*-----------------------------------------------------------------------------
 * Enhydra Java Application Server
 * Copyright 1997-2000 Lutris Technologies, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes Enhydra software developed by Lutris
 *      Technologies, Inc. and its contributors.
 * 4. Neither the name of Lutris Technologies nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY LUTRIS TECHNOLOGIES AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL LUTRIS TECHNOLOGIES OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 */

package com.lutris.dods.store.doml;

import com.lutris.dods.store.doa.*;
import com.lutris.dods.exceptions.*;

import java.io.*;
import java.util.*;

import org.w3c.dom.*;

import org.apache.xerces.parsers.DOMParser;

import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXParseException;

/**
 * This class is used to create DODS Project trees out of DOML files.
 *
 * @author Jason Abbott (jason@lutris.com)
 */
public class DOMLProjectFactory {
    
  private Hashtable dataObjectHash = null;
  private Hashtable dataObjectExtensionPendingHash = null;
  private Hashtable dataObjectReferencePendingHash = null;
  private Vector    warningVector = null;

  private void registerWarning(DOMLWarning warning) {
    if (warningVector == null) {
      warningVector = new Vector();
    }

    warningVector.addElement(warning);
  }

  private void placeAttributesIntoTableDataObject(DataObject   table,
                                                  NamedNodeMap attributes) {
    Node attr = null;

    attr = attributes.getNamedItem("dbTableName");
    try {
        if (attr == null)
            table.setDBTableName(table.getName());
        else
            table.setDBTableName(attr.getNodeValue());
    } catch (DataObjectNameException done) {
        registerWarning(new DOMLWarning(done,
                                         "ERROR: Illegal database name " +
                                         table.getName()));
    }

    attr = attributes.getNamedItem("isFinal");
    if (attr != null && attr.getNodeValue().equals("true"))
      table.setFinal(true);
    else
      table.setFinal(false);

    attr = attributes.getNamedItem("isLazyLoading");
    if (attr != null && attr.getNodeValue().equals("true"))
        table.setLazyLoading(true);
    else
        table.setLazyLoading(false);

    attr = attributes.getNamedItem("isIndex");
    if (attr != null && attr.getNodeValue().equals("true"))
        table.setIndex(true);
    else
        table.setIndex(false);

    attr = attributes.getNamedItem("isAbstract");
    if (attr != null && attr.getNodeValue().equals("true"))
        table.setAbstract(true);
    else
        table.setAbstract(false);

    attr = attributes.getNamedItem("isEJB");
    if (attr != null && attr.getNodeValue().equals("true")) {
        /* Will be implemented in the future */
        // table.setEJB(true);
        registerWarning(new DOMLWarning("EJB Support not yet functional, table " + table.getName()));
    } else {
        /* Will be implemented in the future */
        // table.setEJB(false);
    }

    attr = attributes.getNamedItem("isView");
    if (attr != null && attr.getNodeValue().equals("true")) {
        /* Will be implemented in the future */
        // table.setView(true);
        registerWarning(new DOMLWarning("View Support not yet functional, table " + table.getName()));
    } else {
        /* Will be implemented in the future */
        // table.setView(false);
    }

    attr = attributes.getNamedItem("caching");
    if (attr == null || attr.getNodeValue().equals("none")) {
        table.setFullCaching(false);
        table.setCaching(false);
    } else if (attr.getNodeValue().equals("partial")) {
        table.setFullCaching(false);
        table.setCaching(true);
    } else {
        table.setFullCaching(true);
        table.setCaching(true);
    }

    attr = attributes.getNamedItem("extensionOf");
    if (attr != null && attr.getNodeValue().length() > 0) {
        DataObject extension = (DataObject)
            dataObjectHash.get(attr.getNodeValue());
        if (extension != null) {
            table.setExtensionOf(extension);
        } else {
            addPendingExtensionWatch(attr.getNodeValue(), table);
        }
    }
  }

  private void placeNodeIntoTableDataObject(Node         node,
                                            DataObject   table,
                                            NamedNodeMap attributes)  {
    /* We store the table for future lookup by ref objects/extensions */
    String hashindex = table.getPackageName() + "." + table.getName();
    if (table.getPackageName() == null ||
          table.getPackageName().length() == 0) {
      hashindex = table.getName();
    }

    watchTable(hashindex, table);

    if (attributes != null) {
      placeAttributesIntoTableDataObject(table, attributes);
    }

    NodeList columns = node.getChildNodes();
    if (columns != null) {
      for (int i = 0; i<columns.getLength(); i++) {
        Node columnNode = columns.item(i);
        placeNodeIntoColumnDataObject(columnNode, table);
      }
    }
  }

  private void watchTable(String     name,
                          DataObject table) {
    dataObjectHash.put(name, table);

    if (dataObjectExtensionPendingHash != null &&
          dataObjectExtensionPendingHash.containsKey(name)) {
      Vector vector = (Vector) dataObjectExtensionPendingHash.get(name);
      Enumeration e = vector.elements();
      while (e.hasMoreElements()) {
        DataObject extendedTable = (DataObject) e.nextElement();
        extendedTable.setExtensionOf(table);
      }
      dataObjectExtensionPendingHash.remove(name);
    }

    if (dataObjectReferencePendingHash != null &&
          dataObjectReferencePendingHash.containsKey(name)) {
      Vector vector = (Vector) dataObjectReferencePendingHash.get(name);
      Enumeration e = vector.elements();
      while (e.hasMoreElements()) {
        Attribute attribute = (Attribute) e.nextElement();
        attribute.setObjectRef(true);
        attribute.setObjectRefDO(table);
      }
      dataObjectReferencePendingHash.remove(name);
    }
  }

  private void addPendingExtensionWatch(String     name,
                                        DataObject table) {
    Vector vector = null;
    if (dataObjectExtensionPendingHash == null) {
      dataObjectExtensionPendingHash = new Hashtable();
    }
    if (dataObjectExtensionPendingHash.containsKey(name)) {
      vector = (Vector) dataObjectExtensionPendingHash.get(name);
    } else {
      vector = new Vector();
    }
    vector.addElement(table);
    dataObjectExtensionPendingHash.put(name, vector);
  }

  private void addPendingReferenceWatch(String    name,
                                        Attribute attr) {
    Vector vector = null;
    if (dataObjectReferencePendingHash == null) {
      dataObjectReferencePendingHash = new Hashtable();
    }
    if (dataObjectReferencePendingHash.containsKey(name)) {
      vector = (Vector) dataObjectReferencePendingHash.get(name);
    } else {
      vector = new Vector();
    }
    vector.addElement(attr);
    dataObjectReferencePendingHash.put(name, vector);
  }

  private void placeNodeIntoColumnDataObject(Node       columnNode,
                                             DataObject table) {
    if (columnNode.getNodeType() == Node.TEXT_NODE) {
      /* This isn't a data node, this is just whitespace padding */
      return;
    }

    NamedNodeMap columnAttributes = columnNode.getAttributes();
    Node         idColAttr = null;
    String       columnName = null;

    if (columnAttributes != null) {
      idColAttr = columnAttributes.getNamedItem("id");
    }

    if (idColAttr != null) {
      columnName = idColAttr.getNodeValue();
      if (columnName.length() == 0) {
          columnName = null;
      }
    }

    if (columnName == null) {
      registerWarning(new DOMLWarning("Unnamed column in table " +
                                       table.getPackageName() + "." +
                                       table.getName()));
      return;
    }

    Attribute attribute = new Attribute(columnName);
    table.addAttribute(attribute);

    Node colAttr = columnAttributes.getNamedItem("size");
    if (colAttr != null) {
      attribute.setSize(colAttr.getNodeValue());
    }

    colAttr = columnAttributes.getNamedItem("isIndex");
    if (colAttr != null && colAttr.getNodeValue().equals("true")) {
      attribute.setIndex(true);
    } else {
      attribute.setIndex(false);
    }

    colAttr = columnAttributes.getNamedItem("usedForQuery");
    if (colAttr != null && colAttr.getNodeValue().equals("true")) {
      attribute.setUsedForQuery(true);
    } else {
      attribute.setUsedForQuery(false);
    }

    colAttr = columnAttributes.getNamedItem("isConstant");
    if (colAttr != null && colAttr.getNodeValue().equals("true")) {
      attribute.setConstant(true);
    } else {
      attribute.setConstant(false);
    }

    colAttr = columnAttributes.getNamedItem("isUnique");
    if (colAttr != null && colAttr.getNodeValue().equals("true")) {
      /* IsUnique not yet defined in the current DODS implementation */
      // attribute.setUnique(true);
      registerWarning(new DOMLWarning("isUnique() not yet defined in DODs implementation, column " + table.getPackageName() + "." + table.getName() + "." + columnName));
    } else {
      /* IsUnique not yet defined in the current DODS implementation */
      // attribute.setUnique(false);
    }

    Node childNode = columnNode.getFirstChild();
    while (childNode != null) {

      String       nodeType = childNode.getNodeName();
      NamedNodeMap childAttrs = childNode.getAttributes();

      if (childNode.getNodeType() == Node.TEXT_NODE) {
        /* Ignore text modes, they're white space */
      } else if (nodeType == null) {
        registerWarning(new DOMLWarning("column sub-element has no node name for column " + table.getPackageName() + "." + table.getName() + "." + columnName));
      } else if (nodeType.equals("error")) {
        Node textNode = childNode.getFirstChild();
        if (textNode != null) {
          String errorText = textNode.getNodeValue();
          if (errorText != null) {
            attribute.setErrorText(errorText);
            attribute.setError(true);
          }
        }
      } else if (nodeType.equals("javadoc")) {
        Node textNode = childNode.getFirstChild();
        if (textNode != null) {
          String javadoc = textNode.getNodeValue();
          if (javadoc != null) {
            attribute.setJavadoc(javadoc);
          }
        }
      } else if (nodeType.equals("referenceObject")) {
        Node refObjNode = childAttrs.getNamedItem("reference");
        if (refObjNode != null &&
              refObjNode.getNodeValue().length() > 0) {
          DataObject refObj = (DataObject)
              dataObjectHash.get(refObjNode.getNodeValue());
          if (refObj != null) {
            attribute.setObjectRef(true);
            attribute.setObjectRefDO(refObj);
          } else {
            addPendingReferenceWatch(refObjNode.getNodeValue(),
                                     attribute);
          }
        }
        Node constraintNode = childAttrs.getNamedItem("constraint");
        if (constraintNode != null &&
                constraintNode.getNodeValue().equals("true")) {
          attribute.setRefConstraint(true);
        } else {
          attribute.setRefConstraint(false);
        }
      } else if (nodeType.equals("type")) {
        Node chdAt = childAttrs.getNamedItem("canBeNull");
        if (chdAt != null && chdAt.getNodeValue().equals("true")) {
          attribute.setCanBeNull(true);
        } else {
          attribute.setCanBeNull(false);
        }
        chdAt = childAttrs.getNamedItem("size");
        if (chdAt != null && chdAt.getNodeValue().length() > 0) {
          attribute.setSize(chdAt.getNodeValue());
        }
        chdAt = childAttrs.getNamedItem("dbType");
        if (chdAt != null && chdAt.getNodeValue().length() > 0) {
            attribute.setDBType(chdAt.getNodeValue());
        } else {
            attribute.setDBType("VARCHAR");
        }
        chdAt = childAttrs.getNamedItem("javaType");
        if (chdAt != null && chdAt.getNodeValue().length() > 0) {
            attribute.setJavaType(chdAt.getNodeValue());
        } else {
            attribute.setJavaType("String");
        }
      } else if (nodeType.equals("initialValue")) {
        Node textNode = childNode.getFirstChild();
        if (textNode != null) {
           String initialValue = textNode.getNodeValue();
           if (initialValue != null) {
              attribute.setInitialValue(initialValue);
           }
        }
      } else {
          registerWarning(new DOMLWarning ("column sub-element " + nodeType + " is an unknown node type for column " + table.getPackageName() + "." + table.getName() + "." + columnName));
      }
      childNode = childNode.getNextSibling();
    }

  }

  private void placeNodeIntoPackageDataObject(Node      node,
                                              DOPackage dopackage)
    throws DOMLException {
    NodeList elements = node.getChildNodes();
    String   childPackageName = dopackage.getPackageName();

    if (childPackageName == null || childPackageName.length() == 0)
        childPackageName = dopackage.getName();
    else
        childPackageName = childPackageName + "." + dopackage.getName();

    if (elements != null) {
      for (int i = 0; i<elements.getLength(); i++) {
        Node element = elements.item(i);

        if (element.getNodeType() == Node.TEXT_NODE) {
          /* we ignore text at a package level, it's whitespace */
          continue;
        }

        NamedNodeMap elementAttrs = element.getAttributes();
        String       id = null;

        if (elementAttrs != null) {
          Node idAttr = elementAttrs.getNamedItem("id");
          if (idAttr != null)
            id = idAttr.getNodeValue();
          if (id != null && childPackageName != null) {
            if (id.startsWith(childPackageName)) {
              id = id.substring(childPackageName.length() + 1);
            }
          }
        }

        if (id == null) {
          registerWarning(new DOMLWarning("null id for " +
                                           element.getNodeName() +
                                           " node in " +
                                           childPackageName));
          continue;
        }

        if (element.getNodeName().equals("package")) {
          DOPackage subPackage = new DOPackage(id);
          try {
              subPackage.setPackageName(childPackageName);
          } catch (DataObjectNameException done) {
              throw new DOMLException(done,
                                       "DOML package " + id +
                                       " has poor package name");
          }
          placeNodeIntoPackageDataObject(element, subPackage);
          dopackage.addDOPackage(subPackage);
        } else {
          DataObject table = new DataObject(id);
          try {
            table.setPackageName(childPackageName);
          } catch (DataObjectNameException done) {
            throw new DOMLException(done,
                                     "DOML non-package " + id +
                                     " has poor package name");
          }
          placeNodeIntoTableDataObject(element, table, elementAttrs);
          dopackage.addDataObject(table);
        }
      }
    }
  }

  /**
   * Create a DOML Project producting factory. A single DOML Project
   * Factory can be used repeatedly for multiple calls to
   * <b>createProject</b>.
   */
  public DOMLProjectFactory()
  {
  }

  /**
   * Create a DODS DODSProject tree from the given DOM Document (which should
   * conform to the DOML.DTD)
   *
   * @param file
   *  The File that contains a DOML DOM Document.
   * @return
   *  A DODSProject data structure.
   * @exception FileNotFoundException
   *  Thrown if the given file could not be found.
   * @exception IOException
   *  Thrown if there is a problem reading the file.
   * @exception SAXException
   *  Thrown if there is a problem parsing the file as a DOM object
   * @exception DOMLException
   *  Thrown if there is a problem parsing the file as a DOML object
   */
  public DODSProject createProject(File file)
          throws FileNotFoundException,
                 IOException,
                 SAXException,
                 DOMLException {
    FileInputStream fis = new FileInputStream(file);

    // Create an input source for the document we want to parse.
    // First line provides the actual input stream, the second
    // line specifies the system identifier for error reporting.
    InputSource input = new InputSource(fis);
    input.setSystemId(file.getCanonicalPath());

    // Create a new DOM parser with a standard error handler.
    DOMParser parser = new DOMParser();

    // Parse the document.
    parser.parse(input);

    // Retrieve the document directly from the parser.
    Document doc = parser.getDocument();

    return createProject(doc);
  }

  /**
   * Create a DODS DODSProject tree from the given DOM Document (which should
   * conform to the DOML.DTD)
   *
   * @param doc
   *  The Document that contains a DOML DOM Document.
   * @return
   *  A DODSProject data structure.
   * @exception DOMLException
   *  Thrown if there is a problem parsing the file as a DOML object
   */
  public DODSProject createProject(Document doc)
          throws DOMLException  {
    dataObjectExtensionPendingHash = null;
    dataObjectReferencePendingHash = null;
    warningVector = null;
    dataObjectHash = new Hashtable();

    NodeList elements = doc.getElementsByTagName("database");

    if (elements.getLength() == 0) {
      throw new DOMLException("No database elements in file");
    }
    if (elements.getLength() > 1) {
      registerWarning(new DOMLWarning("Only one database allowed per file"));
    }
    Node dbNode = elements.item(0);
    if (dbNode == null) {
      throw new DOMLException("Top level database element is corrupt.");
    }

    Node projectNode = dbNode.getFirstChild();
    while (projectNode != null &&
             projectNode.getNodeName().equals("package") == false) {
      projectNode = projectNode.getNextSibling();
    }
    if (projectNode == null) {
      throw new DOMLException("Top level database element contains no project element.");
    }

    NamedNodeMap nnm = projectNode.getAttributes();
    if (nnm == null) {
      throw new DOMLException("Top database.project element contains no attributes.");
    }

    Node projectNameNode = nnm.getNamedItem("id");
    if (projectNameNode == null) {
      throw new DOMLException("Top database.project element contains no id attribute.");
    }

    DODSProject project = new DODSProject(projectNameNode.getNodeValue());

    nnm = dbNode.getAttributes();
    if (nnm != null) {
      Node databaseNode = nnm.getNamedItem("database");
      if (databaseNode != null) {
        if (databaseNode.getNodeValue() != null) {
          project.setDatabase(databaseNode.getNodeValue());
        }
      }
    }

    placeNodeIntoPackageDataObject(projectNode, project);

    if (dataObjectReferencePendingHash != null) {
      Enumeration e = dataObjectReferencePendingHash.keys();
      while (e.hasMoreElements()) {
        String name = (String) e.nextElement();
        Vector vector = (Vector)
            dataObjectReferencePendingHash.get(name);
        Enumeration v = vector.elements();
        while (v.hasMoreElements()) {
          Attribute attr = (Attribute) v.nextElement();
          registerWarning(new DOMLWarning("Reference for table " +
                                           name + " by attribute " +
                                           attr.getName() +
                                           " unresolved."));
        }
      }
    }

    if (dataObjectExtensionPendingHash != null) {
      Enumeration e = dataObjectExtensionPendingHash.keys();
      while (e.hasMoreElements()) {
        String name = (String) e.nextElement();
        Vector vector = (Vector)
            dataObjectExtensionPendingHash.get(name);
        Enumeration v = vector.elements();
        while (v.hasMoreElements()) {
          DataObject table = (DataObject) v.nextElement();
          registerWarning(new DOMLWarning("Extension for table " +
                                           name + " by table " +
                                           table.getPackageName() +
                                           "." + table.getName() +
                                           " unresolved."));
        }
      }
    }

    return project;
  }

  /**
   * Determine if the recent call to <b>createProject</b> produced parsing
   * warnings relative to the current implementation of DODs and the DOML
   * parser.
   *
   * @return
   *  the boolean true if parsing the document caused warnings.
   */
  public boolean hasWarnings()
  {
      return (warningVector != null);
  }

  /**
   * Fetch the array of warnings produced by the recent call to
   * <b>createProject</b>.
   *
   * @return
   *  the DOMLWarning array, which is guarenteed to be non-null, but might
   *  be length zero.
   */
  public DOMLWarning[] getWarnings() {
    int arraySize = (warningVector == null) ? 0 : warningVector.size();
    DOMLWarning retVal[] = new DOMLWarning[arraySize];
    if (warningVector != null) {
      warningVector.copyInto(retVal);
    }
    return retVal;
  }

}


