/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: CoreDO.java,v 1.15.12.1 2000/10/19 17:59:05 jasona Exp $
 */





package	com.lutris.appserver.server.sql;
import com.lutris.appserver.server.LBS;
import java.sql.*;
import java.io.*;

/**
 * Base class from which the data objects are derived.
 *
 * @version $Revision: 1.15.12.1 $
 * @author Kyle Clark
 */
abstract public class CoreDO implements Transaction, Serializable {

    /**
     * The 'oid' and 'version' column names are made public so
     * they can be modified by StandardDatabaseManager.
     * These could be maintained in the LogicalDatabase but that
     * would require extensive modifications to DBTransaction and
     * DBQuery.
     * This means that you are currently limited to using the
     * same column names for all databases in applications using
     * multiple databases.
     */
    private static String oidColumnName = "oid";
    private static String versionColumnName = "version";

    public static boolean versioning = true;

   /**
    * Disable use of the version column.
    *
    * @author    Jay Gunter
    */
    public static void disableVersioning() {
	versioning = false;
    }

    // The object id associated with this object.
    private ObjectId oId;

    // The version of this object.
    private int version = 0;

    // The next version of this object.
    private int newVersion = 0;

    // True if this object exists in the database.
    protected boolean persistent = false;

    /**
     * Public constructor.
     */
    public CoreDO() {
        persistent = false;
    }

    /**
     * Public constructor.
     * @param rs a result set.
     * @exception SQLException if an error occurs while instantiating this
     * object from the result set.
     * @exception ObjectIdException if the object id for this object is
     * invalid.
     */
    public CoreDO(ResultSet rs) throws SQLException, ObjectIdException {
        persistent = true;
        setOId(new ObjectId(rs.getBigDecimal(getOIdColumnName(), 0)));
	if ( versioning )
       	    setVersion(rs.getInt(getVersionColumnName()));
    }

    /**
     * Returns the object identifier column name.
     * @return the object identifier column name.
     */
    public static String getOIdColumnName() {
	return oidColumnName;
    }

    /**
     * Sets the object identifier column name.
     * @param oidColumnName Name of object identifier column.
     */
    public static void setOIdColumnName(String oidColumnName) {
	CoreDO.oidColumnName = oidColumnName;
    }

    /**
     * Returns the version column name.
     * @return the version column name.
     */
    public static String getVersionColumnName() {
	return versionColumnName;
    }

    /**
     * Sets the version column name.
     * @param versionColumnName Name of version column.
     */
    public static void setVersionColumnName(String versionColumnName) {
	CoreDO.versionColumnName = versionColumnName;
    }
    /**
     * Returns this object's identifier.
     * @return this object's identifier.
     */
    public ObjectId getOId() {
	return oId;
    }

    /**
     * Sets this object's identifier.
     * @param oId this object's identifier.
     */
    protected void setOId(ObjectId oId) {
	this.oId = oId;
    }

    /**
     * Sets this object's version.
     * @param version the object's version.
     */
    protected void setVersion(int version) {
        this.version = version;
    }

    /**
     * Returns this object's version.
     * @return this object's version.
     */
    protected int getVersion() {
        return this.version;
    }

    /**
     * Sets this object's new version number.
     * @param newVersion this object's next version.
     */
    protected void setNewVersion(int newVersion) {
        this.newVersion = newVersion;
    }

    /**
     * Returns this object's new version.
     * @return this object's new version.
     */
    protected int getNewVersion() {
        return this.newVersion;
    }

    /**
     * Returns true if this object is persistent in the database
     * (i.e. it has been inserted and not deleted)
     * @return true if this data object has been inserted into
     *   the database.
     */
    public boolean isPersistent() {
        return this.persistent;
    }

    /**
     * Sets the persistent state for this object.
     * @param persistent true if this object is present in
     *   persistent store.
     */
    public void setPersistent(boolean persistent) {
        this.persistent = persistent;
    }

    /**
     * Returns the statement that can be used to insert this
     * object into the database.
     *
     * @param conn the database connection.
     * @return the statement used to insert this object into the
     *   database.  null if the object cannot be inserted.
     * @exception java.sql.SQLException If an error occurs.
     */
    public abstract PreparedStatement getInsertStatement(DBConnection conn)
            throws SQLException;

    /**
     * Returns the statement that can be used to update this
     * object in the database.
     *
     * @param conn the database connection.
     * @return the statement used to update this object in the
     *   database.  null if the object cannot be updated.
     * @exception java.sql.SQLException If an error occurs.
     */
    public abstract PreparedStatement getUpdateStatement(DBConnection conn)
            throws SQLException;

    /**
     * Returns the statement that can be used to delete this
     * object from the database.
     *
     * @param conn the database connection.
     * @return the statement used to delete this object from the
     *   database.  null if the object cannot be deleted.
     * @exception java.sql.SQLException If an error occurs.
     */
    public abstract PreparedStatement getDeleteStatement(DBConnection conn)
            throws SQLException;

    /**
     * Inserts this object into the database.
     *
     * @param conn the database connection.
     * @exception java.sql.SQLException if a database access error occurs.
     * @exception DBRowUpdateException If a version error occurs.
     */
    public synchronized void executeInsert(DBConnection conn)
            throws SQLException, DBRowUpdateException {
        PreparedStatement stmt;
        if (persistent) {
            executeUpdate(conn);
        } else {
            stmt = getInsertStatement(conn);
            if (stmt != null) {
                conn.executeUpdate(stmt, "executeInsert");
            }
            persistent = true;
        }
    }

    /**
     * Updates the persistent state.
     *
     * @param success true if the transaction succeeded
     *   and this object was successfully inserted into the database.
     */
    public void finalizeInsert(boolean success) {
        persistent = success;
        if (versioning) {
	    if (success) {
		setVersion(getNewVersion());
	    } else {
		setNewVersion(getVersion());
	    }
	}
    }

    /**
     * Updates the contents of this object in the database.
     *
     * @param conn the database connection.
     * @exception java.sql.SQLException If a database access error occurs.
     * @exception DBRowUpdateException If a version error occurs.
     */
    public synchronized void executeUpdate(DBConnection conn)
            throws SQLException, DBRowUpdateException {
        boolean mustUpdate = (getNewVersion() == getVersion());
        if (mustUpdate && versioning) {
            setNewVersion(getVersion() + 1);
        }
        PreparedStatement stmt = getUpdateStatement(conn);
        if (stmt != null) {
            if ((conn.executeUpdate(stmt, "execute update") == 0)
                    && mustUpdate) {
		// Zero rows were affected by the update.
		// This usually occurs when the version number does not match
		// the version number for this oId in the database table.
		// We verify that this is the case, and throw an appropriate
		// exception.
                try {
		    String table = this.getTableName();
		    stmt = conn.prepareStatement(
			    "select " + versionColumnName +
			    " from " + table +
			    " where " + oidColumnName + " = " + getOId() );
		    ResultSet rs = stmt.executeQuery();
		    if ( false == rs.next() )
			throw new DBRowUpdateException(
			    "Update failed: Table " +  table +
			    " contains no row with id=" + getOId() );
		    int v = rs.getInt( 1 );
		    if ( getVersion() != v )
			throw new DBRowUpdateException(
			    "Update failed: Table " +  table +
			    " id=" + getOId() +
			    " has version " + v +
			    " where object has version " + getVersion() +
			    ( versioning ?
				"." :
				", and versioning is disabled!" ) );
		    throw new SQLException(
			    "Update failed, but Table " +  table +
			    " id=" + getOId() +
			    " does exist with version=" + v );
                } catch ( DBRowUpdateException re ) {
		    throw re;
                } catch ( SQLException se ) {
		    throw se;
                } catch ( Exception e ) {
                    throw new SQLException(
			"Error determining cause of update failure.",
			e.getMessage() );
                }
            }
        }
    }

    /**
     * Return the name of the table whose rows represent these objects.
     * This method should be overridden by derived classes.
     *
     * @return the name of the table.
     * @see #executeUpdate
     */
    protected String getTableName() {
      // FIXME: really should be a specific error
      throw new Error("getTableName() not implemented. DO classes need to be regenerated");
    }

    /**
     * The version number of this object is set to the next
     * version for this object if
     * this object was successfully updated in the database.
     *
     * @param success true if the transaction succeeded
     *   and this object was successfully updated in the database.
     * @see #getNewVersion
     */
    public void finalizeUpdate(boolean success) {
        if (versioning) {
	    if (success) {
		setVersion(getNewVersion());
	    } else {
		setNewVersion(getVersion());
	    }
	}
    }

    /**
     * Deletes this object from the database.
     *
     * @param conn Database connection.
     * @exception java.sql.SQLException If a database access error
     *  occurs.
     */
    public void executeDelete(DBConnection conn)
            throws SQLException {
        PreparedStatement stmt = getDeleteStatement(conn);
        if (stmt != null) {
            if (conn.executeUpdate(stmt, "execute delete") == 0) {
                throw new SQLException("no match", "02000", 11111111);
            }
        }
    }

    /**
     * Currently does nothing.
     *
     * @param success true if the transaction succeeded
     *   and this object was successfully deleted from the
     *   database.
     */
    public void finalizeDelete(boolean success) {
        if (persistent) {
            persistent = !success;
        }
    }

}

