/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: StandardDatabaseManager.java,v 1.20.12.1 2000/10/19 17:59:05 jasona Exp $
 */





package com.lutris.appserver.server.sql;

import com.lutris.util.*;
import com.lutris.appserver.server.Enhydra;
import com.lutris.appserver.server.sql.*;
import com.lutris.appserver.server.sql.standard.StandardLogicalDatabase;
import com.lutris.appserver.server.sql.oracle.OracleLogicalDatabase;
import com.lutris.appserver.server.sql.informix.InformixLogicalDatabase;
import com.lutris.appserver.server.sql.sybase.SybaseLogicalDatabase;
import com.lutris.appserver.server.sql.msql.MsqlLogicalDatabase;
import com.lutris.appserver.server.sql.postgresql.PostgreSQLLogicalDatabase;
import java.sql.*;
import java.util.Date;
import java.util.Hashtable;
import java.util.Enumeration;


/**
 * The standard database manager implementation. A database manager
 * manages logical databases. It provides a single object from which
 * database connections, object ids (OIDs), transactions and queries
 * can be obtained. The configuration file specifies what logical
 * database to create.
 * <P>
 * The configuration data is specified as follows:
 * <UL>
 * <LI> <B><CODE>DatabaseManager.Databases</CODE></B> -
 *	A list of logical SQL database names.
 * <LI> <B><CODE>DatabaseManager.DefaultDatabase</CODE></B> -
 *	The default logical database used by this application. Optional,
 *	if not specified, then the first entry in
 *	"DatabaseManager.Databases" is used.
 * <LI> <B><CODE>DatabaseManager.Debug</CODE></B> -
 *	Specify true to enable Query and Transaction logging, false to
 *	disable it. Optional, false if not specified.
 * </UL>
 * For each logical database, there is a set of entry names in the form
 * <B><CODE>DatabaseManager.DB.<I>dbname</I></CODE></B>.
 * Where <CODE><I>dbname</I></CODE> is one of the listed logical databases.
 * <P>
 * <B><CODE>DatabaseManager.DB.<I>dbname</I>.ClassType</CODE></B> -
 * This is an optional field which specifies the class of the logical
 * database implementation or a symbolic name if one on the standard types
 * are selected. This is recommended because although JDBC abstracts the
 * data access, the functionality of each database is slightly different
 * and this parameter allows for optimised usage. The following are
 * standard types:
 * <UL>
 * <LI> <I>Oracle</I> - For optimized Oracle 7/8 usage.
 * <LI> <I>Informix</I> - For optimized Informix usage.
 * <LI> <I>Sybase</I> - For optimized Sybase usage.
 * <LI> <I>Msql</I> - For optimized Microsoft MSQL usage.
 * <LI> <I>Standard</I> - For all other JDBC databases. <B>(Default)</B>
 * <LI> <I>class</I> - For vendor supplied database classes.
 * </UL>
 * <P>
 * Note that since a single SQL user is used to access the database by
 * the entire application, connections maybe kept open, thus saving this
 * overhead on each request.  Connections are opened as desired until
 * the maximum configured is reached.
 * <P>
 * If a thread needs to process a transaction, it requests a connection.
 * If none are available, then a new connection is created up to the
 * configured maximum.  A thread is queued waiting for a connection to be
 * returned if a new connection can't be created.
 * <P>
 * Note also that multiple logical names may map to the same actual
 * database, making it easy to migrate databases if necessary and to
 * provide access through multiple users.
 *
 * @version	$Revision: 1.20.12.1 $
 * @since	LBS1.8
 * @author	Paul Morgan
 * @author	Kyle Clark
 */
public class StandardDatabaseManager implements DatabaseManager {

    /**
     * Table of named logical databases.
     */
    private Hashtable logicalDatabases = new Hashtable ();

    /**
     * Default logical database.
     */
    private LogicalDatabase defaultLogicalDatabase = null;

    /**
     * Controls debugging for Transactions and Queries.
     */
    protected boolean debug = false;


    /**
     * Creates a new <code>DatabaseManager</code> object and configures
     * the logical databases defined in the config file.
     *
     * @param config
     *  The configuration data for logical databases.
     * @exception ConfigException
     *  If there is an error in the configuration file.
     * @exception DatabaseManagerException
     *  If a logical database name is specified twice in the configuration file.
     * @exception SQLException
     *  If a SQL error occurs.
     */
    public StandardDatabaseManager(Config config)
        throws ConfigException, DatabaseManagerException, SQLException {

        String[] databases = config.getStrings ("Databases");
        if (databases == null)
            return;

	/**
	 * Configure each logical database.
	 */
	Config dbConfig;
        for (int idx = 0; idx < databases.length; idx++) {
	    String dbName = databases[idx];
	    try {
		dbConfig = (Config)config.getSection("DB." + dbName);
	    } catch (KeywordValueException except) {
		throw new ConfigException("No DatabaseManager.DB." + dbName + " defined in config file.");
	    }

	    if (logicalDatabases.get (dbName) != null) {
		throw new DatabaseManagerException
		("duplicate logical database name: \"" + dbName + "\"");
	    }

	    LogicalDatabase logicalDatabase =
		loadLogicalDatabase (dbName, dbConfig);

	    logicalDatabases.put (dbName, logicalDatabase);
        }

	/**
	 * Set default database if supplied.
	 */
        String defaultDB = config.getString ("DefaultDatabase", databases[0]);
        setDefaultDatabase(defaultDB);

	/**
	 * Set debug logging of Queries and Transactions.
	 */
        boolean debugLogging = config.getBoolean ("Debug", false);
	setDebugLogging(debugLogging);

	/**
	 * Set the oid and version column names in CoreDO.
	 * Note that these could be configured for each logical
	 * database but this would require some complex extensions
	 * to DBQuery and DBTransaction in order to ensure that
	 * the correct values were always being used without
	 * creating race conditions.
	 * This means that the limitation is currently that all
	 * databases accessed by your application must use the
	 * same column names.
	 */
        String oidColumnName = config.getString("ObjectIdColumnName", null);
	if (oidColumnName != null) {
            CoreDO.setOIdColumnName(oidColumnName);
	}
        String versionColumnName = config.getString("VersionColumnName", null);
	if (versionColumnName != null) {
            CoreDO.setVersionColumnName(versionColumnName);
	}
    }


    /**
     * Actually load the specified logical database. This method provides
     * an easy way to override the default behavour.
     *
     * @return
     *   The logical database.
     * @exception DatabaseManagerException
     *   if an error occurs creating the logical database.
     */
    public LogicalDatabase loadLogicalDatabase (String dbName,
						Config dbConfig)
	throws DatabaseManagerException {

	LogicalDatabase lDB = null;
	try {
	    String dbClassName = dbConfig.getString ("ClassType", "Standard");
	    if (dbClassName.equals ("Standard")) {
		lDB = new StandardLogicalDatabase (dbName, dbConfig);
	    } else if (dbClassName.equals ("Oracle")) {
		lDB = new OracleLogicalDatabase (dbName, dbConfig);
	    } else if (dbClassName.equals ("Informix")) {
		lDB = new InformixLogicalDatabase (dbName, dbConfig);
	    } else if (dbClassName.equals ("Sybase")) {
		lDB = new SybaseLogicalDatabase (dbName, dbConfig);
	    } else if (dbClassName.equals ("Msql")) {
		lDB = new MsqlLogicalDatabase (dbName, dbConfig);
	    } else if (dbClassName.equals ("PostgreSQL")) {
		lDB = new PostgreSQLLogicalDatabase (dbName, dbConfig);
	    } else {
		// Must be a custom class. Make instance and call init
		// method to configure.  Also we must use the application
		// class loader.
		ClassLoader appClassLoader =
		    Enhydra.getApplication().getClass().getClassLoader();
		Class dbClass = appClassLoader.loadClass(dbClassName);
		lDB = (LogicalDatabase) dbClass.newInstance ();
		lDB.init (dbName, dbConfig);
	    }
	} catch (Exception except) {
	    throw new DatabaseManagerException ("Could not create logical database " + dbName, except);
	}

	return lDB;
    }


    /**
     * Allocate a connection to a thread. The connection should be returned
     * to the allocator by calling its
     * <a href=com.lutris.appserver.server.sql.DBConnection#release>
     * release()</a> function.  A thread will wait if
     * no connections are available.
     * Interupted exceptions are converted to
     * errors.
     * N.B. Can't be synchronized, as connection allocator
     * <CODE>allocate</CODE> may wait.
     *
     * @param dbName
     *   Logical name of the database to allocate a connection to.
     * @return
     *   The allocated connection object.
     * @exception DatabaseManagerException
     *   If a nonexistent logical database name is supplied.
     * @exception SQLException
     *   If a SQL error occures.
     */
    public DBConnection allocateConnection (String dbName)
        throws DatabaseManagerException, SQLException {

        LogicalDatabase logicalDatabase = findLogicalDatabase (dbName);
        return logicalDatabase.allocateConnection();
    }


    /**
     * Allocate a connection to a thread. The connection should be returned
     * to the allocator by calling its
     * <a href=com.lutris.appserver.server.sql.DBConnection#release>
     * release()</a> function.  A thread will wait if
     * no connections are available.
     * Interupted exceptions are converted to
     * errors.  The connection is allocated from the default logical
     * database.
     *
     * @return
     *   The allocated connection object.
     * @exception DatabaseManagerException
     *   If no default logical database has been set.
     * @exception SQLException
     *   If a SQL error occurs.
     * @see
     *   #setDefaultDatabase
     */
    public DBConnection allocateConnection ()
        throws DatabaseManagerException, SQLException {

        if (defaultLogicalDatabase == null)
            throw new DatabaseManagerException ("Default logical database " +
                                                 "has not been specified.");
        return defaultLogicalDatabase.allocateConnection ();
    }


    /**
     * Allocate an object id from the specified logical database.
     *
     * @param name
     *   Logical name of the database from which to obtain an object id.
     * @return The allocated unique OID
     * @exception DatabaseManagerException
     *   If a nonexistent logical database name is supplied.
     * @exception ObjectIdException
     *   If a problem (e.g. SQL error) occured in obtaining the OID.
     */
    public ObjectId allocateObjectId (String dbName)
        throws DatabaseManagerException, ObjectIdException {

        LogicalDatabase logicalDatabase = findLogicalDatabase (dbName);
        return logicalDatabase.allocateObjectId();
    }


    /**
     * Allocate an object id from the specified logical database.
     *
     * @return The allocated connection object.
     * @exception DatabaseManagerException
     *   If a nonexistent default logical database has been set.
     * @exception ObjectIdException
     *   If a problem (e.g. SQL error) occured in obtaining the OID.
     * @see
     *   #setDefaultDatabase
     */
    public ObjectId allocateObjectId ()
        throws DatabaseManagerException, ObjectIdException {

        if (defaultLogicalDatabase == null)
            throw new DatabaseManagerException ("Default logical database " +
                                                 "has not been specified.");
        return defaultLogicalDatabase.allocateObjectId ();
    }


    /**
     * Create a transaction object for the specified logical database.
     *
     * @param dbName
     *   Logical name of the database from which to obtain a transaction.
     * @return The transaction
     * @exception DatabaseManagerException
     *   If a nonexistent or invalid logical database name is supplied.
     * @exception SQLException
     *   If a problem occured creating the transaction.
     */
    public DBTransaction createTransaction (String dbName)
        throws DatabaseManagerException, SQLException {

        LogicalDatabase logicalDatabase = findLogicalDatabase (dbName);
        return logicalDatabase.createTransaction ();
    }


    /**
     * Create a transaction object for the default logical database.
     *
     * @param dbName
     *   Logical name of the database from which to obtain a transaction.
     * @return The transaction
     * @exception DatabaseManagerException
     *   If a nonexistent default logical database has been set.
     * @exception SQLException
     *   If a problem occured creating the transaction.
     * @see
     *   #setDefaultDatabase
     */
    public DBTransaction createTransaction ()
        throws DatabaseManagerException, SQLException {

        if (defaultLogicalDatabase == null)
            throw new DatabaseManagerException ("Default logical database " +
                                                 "has not been specified.");
        return defaultLogicalDatabase.createTransaction ();
    }


    /**
     * Create a query object for the specified logical database.
     *
     * @param dbName
     *   Logical name of the database from which to obtain a query.
     * @return The query
     * @exception DatabaseManagerException
     *   If a nonexistent or invalid logical database name is supplied.
     * @exception SQLException
     *   If a problem occured creating the query.
     */
    public DBQuery createQuery (String dbName)
        throws DatabaseManagerException, SQLException {

        LogicalDatabase logicalDatabase = findLogicalDatabase (dbName);
        return logicalDatabase.createQuery ();
    }


    /**
     * Create a query object for the default logical database.
     *
     * @param dbName
     *   Logical name of the database from which to obtain a query.
     * @return The query
     * @exception DatabaseManagerException
     *   If a nonexistent default logical database has been set.
     * @exception SQLException
     *   If a problem occured creating the query.
     * @see
     *   #setDefaultDatabase
     */
    public DBQuery createQuery ()
        throws DatabaseManagerException, SQLException {

        if (defaultLogicalDatabase == null)
            throw new DatabaseManagerException ("Default logical database " +
                                                 "has not been specified.");
        return defaultLogicalDatabase.createQuery ();
    }


    /**
     * Find the named logical database in hash table.
     *
     * @param dbName Logical name of the database to locate.
     * @exception DatabaseManagerException If a nonexistant logical database
     *  name is supplied.
     */
    private LogicalDatabase findLogicalDatabase (String dbName)
        throws DatabaseManagerException {

        LogicalDatabase logicalDatabase =
	    (LogicalDatabase) logicalDatabases.get (dbName);
        if (logicalDatabase == null) {
            throw new DatabaseManagerException
                ("unknown logical database name: \"" +
                 dbName + "\"");
        }
        return logicalDatabase;
    }


    /**
     * Set the default logical database. This should only be called
     * after the logical database
     * (<A HREF=com.lutris.appserver.server.sql.LogicalDatabase></A>
     * has been established.
     *
     * @param dbName The default logical database.
     * @exception DatabaseManagerException
     *   if the logical database name is invalid or not found.
     */
    public void setDefaultDatabase (String dbName)
        throws DatabaseManagerException {

        defaultLogicalDatabase = findLogicalDatabase (dbName);
    }

    /**
     * Shutdown the database manager. All logical databases will be
     * shutdown and all connections closed.
     */
    public void shutdown () {

	for (Enumeration keys = logicalDatabases.keys ();
		keys.hasMoreElements();) {

	    LogicalDatabase logicalDatabase = (LogicalDatabase)
		logicalDatabases.get ((String)keys.nextElement());

	    logicalDatabase.shutdown();
	}
    }


    //====================================================================
    // The following are primarily for management purposes...
    //====================================================================

    /**
     * Returns the list of managed logical databases.
     *
     * @return	List of logical database names.
     */
    public String[] getLogicalDatabaseNames () {

	String[] names = new String[logicalDatabases.size()];
	int idx = 0;
        Enumeration keys = logicalDatabases.keys();
        while (keys.hasMoreElements()) {
            names[idx++] = (String) keys.nextElement();
        }
	return names;
    }


    /**
     * Returns a description of the logical database type.
     *
     * @param dbName
     *   The logical database name.
     * @return
     *   A text description of the logical database type.
     * @exception DatabaseManagerException
     *   If a nonexistent logical database name is supplied.
     */
    public String getType(String dbName)
	throws DatabaseManagerException {

        LogicalDatabase logicalDatabase = findLogicalDatabase (dbName);
        return logicalDatabase.getType();
    }


    /**
     * Gets the number of requests made to the database since startup time.
     *
     * @param	The name of the logical database.
     * @exception DatabaseManagerException
     *   If a nonexistent logical database name is supplied.
     * @return	The number of database requests since the server started.
     */
    public long getRequestCount (String dbName)
        throws DatabaseManagerException {

        LogicalDatabase logicalDatabase = findLogicalDatabase (dbName);
        return logicalDatabase.getRequestCount();
    }


    /**
     * Gets the number of currently active connections.
     *
     * @param	dbName The name of the logical database.
     * @exception DatabaseManagerException
     *   If a nonexistent logical database name is supplied.
     * @return	The number of currently active connections.
     */
    public int getActiveConnectionCount (String dbName)
        throws DatabaseManagerException {

        LogicalDatabase logicalDatabase = findLogicalDatabase (dbName);
        return logicalDatabase.getActiveConnectionCount();
    }


    /**
     * Gets the maximum number of concurent connections that existed
     * at any time since this object was created, or
     * <CODE>resetMaxConnectionCount()</CODE> was called.
     * This is a historical highwater mark.
     * If you do not implement this feature, return -1.
     *
     * @param	dbName The name of the logical database.
     * @exception DatabaseManagerException
     *   If a nonexistent logical database name is supplied.
     * @return The highwater mark for number of connections, or -1.
     */
    public int getMaxConnectionCount (String dbName)
        throws DatabaseManagerException {

        LogicalDatabase logicalDatabase = findLogicalDatabase (dbName);
        return logicalDatabase.getMaxConnectionCount();
    }


    /**
     * Gets the time when the maximum refered to by
     * <CODE>maxConnectionCount()</CODE> occured.
     *
     * @param	dbName The name of the logical database.
     * @exception DatabaseManagerException
     *   If a nonexistent logical database name is supplied.
     * @return The Date of when the maximum number of connections occured.
     */
    public Date getMaxConnectionCountDate (String dbName)
        throws DatabaseManagerException {

        LogicalDatabase logicalDatabase = findLogicalDatabase (dbName);
        return logicalDatabase.getMaxConnectionCountDate();
    }


    /**
     * Reset the maximum connection count. See
     * <CODE>maxConnectionCount()</CODE>. The highwater mark should be
     * reset to the current number of connections.
     *
     * @param	dbName The name of the logical database.
     * @exception DatabaseManagerException
     *   If a nonexistent logical database name is supplied.
     */
    public void resetMaxConnectionCount (String dbName)
        throws DatabaseManagerException {

        LogicalDatabase logicalDatabase = findLogicalDatabase (dbName);
        logicalDatabase.resetMaxConnectionCount();
	return;
    }


    /**
     * Turn debugging on or off.
     *
     * @param	condition on of off.
     */
    public void setDebugLogging (boolean condition) {
	debug = condition;
    }
}
