/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: InformixDBConnection.java,v 1.9.12.1 2000/10/19 17:59:05 jasona Exp $
 */





package	com.lutris.appserver.server.sql.informix;

import com.lutris.appserver.server.LBS;
import com.lutris.appserver.server.sql.*;
import com.lutris.appserver.server.sql.standard.*;
import com.lutris.logging.*;
import java.sql.*;
import java.util.*;

/**
 * An Informix database connection.
 *
 * @see		DBConnection
 * @author	Kyle Clark
 * @since	LBS1.7
 * @version	$Revision: 1.9.12.1 $
 */
public class InformixDBConnection extends StandardDBConnection {

    /**
     * The amound time to wait on a query or transaction.
     */
    private int lockModeWait = 0;

    /**
     * The logical database for this connection.
     */
    protected InformixLogicalDatabase logicalDatabase;

    /**
     * Create a connection to an Informix database.
     *
     * @param url JDBC URL of database.
     * @param user SQL user name.
     * @param password SQL password.
     * @param maxPreparedStatements  Maximum number of preparse statements.
     *  A value of less than zero queries JDBC for the value.
     * @param connectionAllocator The connection allocator that this
     *  connection belongs to.
     * @param logging Specifying <CODE>true</CODE> enables SQL logging.
     * @param generation A generation number used to drop old connection
     *  when they are released.
     * @exception java.sql.SQLException If a connection can't be established.
     */
    protected InformixDBConnection(InformixConnectionAllocator connectionAllocator,
                                   String url, String user, String password,
                                   int maxPreparedStatements,
                                   boolean logging, int generation)
        throws SQLException {
        super(connectionAllocator, url, user, password, maxPreparedStatements,
              logging, generation);
        this.logicalDatabase = connectionAllocator.logicalDatabase;
    }

    /**
     * Specify if you prefer to wait for locks instead of generating
     * errors.
     *
     * @param on
     *   True if you prefer to wait, false otherwise.
     * @exception java.sql.SQLException
     *   If a database access error occurs. 
     */
    public synchronized void setLockModeToWait(boolean on)
        throws SQLException
    {
        if (on) {
            setLockModeToWait(-1);
        } else {
            setLockModeToWait(0);
        }
    }

    /**
     * Specify if you prefer to wait for locks for the specified
     * amount of time before an exception is thrown.
     *
     * @param seconds
     *   The number of seconds to wait.  If seconds is == 0
     *   the waiting is disabled.  If seconds is less than
     *   0 then wait indefinitly.
     * @exception java.sql.SQLException
     *   If a database access error occurs. 
     */
    public synchronized void setLockModeToWait(int seconds)
        throws SQLException
    {
        if (lockModeWait != seconds) {
            logDebug("set lock mode to wait " + seconds);            
            if (seconds > 0) {
                execute("SET LOCK MODE TO WAIT " + seconds);
            } else if (seconds == 0) {
                execute("SET LOCK MODE TO NOT WAIT");
            } else {
                execute("SET LOCK MODE TO WAIT");
            }
            lockModeWait = seconds;
        }
    }

}
