/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: Application.java,v 1.24.14.1 2000/10/19 17:59:06 jasona Exp $
 */




package com.lutris.appserver.server;

import com.lutris.appserver.server.httpPresentation.*;
import com.lutris.appserver.server.session.SessionManager;
import com.lutris.appserver.server.session.SessionException;
import com.lutris.appserver.server.sql.DatabaseManager;
import com.lutris.appserver.server.*;
import com.lutris.logging.LogChannel;
import com.lutris.util.Config;
import org.enhydra.xml.xmlc.XMLCFactory;

import java.io.IOException;
import javax.servlet.*;
import javax.servlet.http.*;

/**
 * Interface used to define a Lutris server application.  Each application
 * defines a class that implements this interface.  The specific application
 * class serves as an entry and control point for the application.  This
 * class is responsible for initializing any application specific resources,
 * including instance of standard services that are needed.  The class
 * must have a constructor with a configuration file argument.
 * <P>
 * The class is instantiated from a presentation manager and is used to
 * start and stop the application.  This structure allows applications to be
 * started on demand by URL reference, as is required for applications
 * running under servlets in HTTPD virtual machines.  If remote threads are
 * part of the application, they are normally started from this object.
 * <P>
 * The application is in one of several states, documented below.  The
 * application state is used to determine what actions are take when a
 * request is made for the application.
 * <P>
 * The following states are possible:
 * <UL>
 * <LI> <CODE>STOPPED</CODE> - The application is not running and maybe
 *      started.  This is the initial state and the state normally
 *      entered on a <CODE>shutdown()</CODE>.
 * <LI> <CODE>RUNNING</CODE> - The application is currently running.
 * <LI> <CODE>INCOMPLETE</CODE> - The application is not completely running
 *      and the restartup() method maybe called to attempt to bring it into the
 *      <CODE>RUNNING</CODE> state.  This state is normally used when a
 *      remote server that is required by the application is not available
 *      when <CODE>startup()</CODE> is called.  Partial initialization will
 *      take place, but the request that caused the startup will fail.
 *      A subsequent requests will results in calls to <CODE>restartup()</CODE>
 *      until the application can be fully started.  This eliminates startup
 *      order dependencies when initializing a distributed application.
 *      The state may also be entered by the application when a remote server
 *      fails.
 * <LI> <CODE>DEAD</CODE> - The application is in an fatal error state and
 *      can not be restarted.
 * <LI> <CODE>HALTED</CODE> - shutdown was called on the application and
 *      the application is not designed so that in can be restarted.
 * </UL>
 * The state is under control of the application. Its is used by the
 * server to know what actions are allowed, but never modified by
 * the server. The simplest application enters the <CODE>RUNNING</CODE> state
 * when <CODE>startup()</CODE> is called and the <CODE>STOPPED</CODE> state
 * when <CODE>shutdown()</CODE> is called.
 * <P>
 * The server provides synchronization on state changes to the application;
 * however multiple requests threats may enter the application at the same
 * time.
 * <P>
 * Two configuration parameters are required to start an application.  Under
 * servlets, these are specified as init args for the presentation
 * servlet.  They are:
 * <UL>
 * <LI> <CODE>appClass</CODE> - This is the class name of the application
 *      class.  If not specified, the presentation servlet will run
 *      without an application object.
 * <LI> <CODE>appConfig</CODE> - An arbitrary string to pass to the
 *      <CODE>startup(0</CODE> method.  It normally contains the path or URL
 *      of a Config or properties file containing the application configuration.
 *      If not specified, <CODE>null</CODE> is passed to <CODE>startup</CODE>.
 * </UL>
 * 
 * @version	$Revision: 1.24.14.1 $
 * @author	Mark Diekhans
 */
public interface Application {

    /**
     * Stopped state code.
     */
    public static final int STOPPED = 1;

    /**
     * Running state code.
     */
    public static final int RUNNING = 2;

    /**
     * Incomplete state code.
     */
    public static final int INCOMPLETE = 3;

    /**
     * Dead state code.
     */
    public static final int DEAD = 4;

    /*
     * Halted state code.
     */
    public static final int HALTED = 5;

    /**
     * Get the application state.
     *
     * @return The application's state code.
     */
    public int getState();

    /**
     * Application Data accessible through Jolt Fields.
     */
    public ApplicationData getApplicationData();

    /**
     * Set the application symbolic name.
     *
     * @param name The application's name.
     */
    public void setName(String name);

    /**
     * Get the application symbolic name.  
     *
     * @return The symbolic name.
     */
    public String getName();

    /**
     * Get the application's config object.
     *
     * @return The config object.
     */
    public Config getConfig();

    /**
     * Start the application.  The default method sets the state to
     * <CODE>RUNNING</CODE>.
     * 
     * @param appConfig Application configuration object.
     * @exception ApplicationException If an error occurs starting the
     *	application.
     */
    public void startup(Config appConfig) throws ApplicationException;

    /**
     * Continue the startup up process with the application is in the
     * <CODE>INCOMPLETE</CODE> state.  The default method generates an
     * error, as the application should never be in the <CODE>INCOMPLETE</CODE>
     * state if it doesn't implement this method.
     * 
     * @param appConfig
     *   The same <CODE>appConfig</CODE> object that was passed to
     *   <CODE>startup</CODE>.
     * @exception ApplicationException
     *   If an error occurs starting the application.
     */
    public void restartup(Config appConfig) throws ApplicationException;

    /**
     * Shutdown the application.  The default method sets the state to
     * <CODE>STOPPED</CODE>.
     */
    public void shutdown();

    /**
     * Request preprocessing method.  All requests for URLs associated
     * with this application go through this function.  It may do any
     * preprocessing desired, including handling the request or generating
     * a page redirect.  It can be used to force login, allocate session
     * objects, restrict access, etc.<P>
     *
     * Page redirect requests also go through this method, so care must be
     * taken not to generate an endless page redirect loop.  ErrorHandler
     * presentation objects do not go through this method if they are invoked
     * from within the presentation manager.  However, direct URL requests to
     * ErrorHandler presentation objects will go through this function. <P>
     *
     * Note: unlike the method servletRequestPreprocessor(), this method
     * is commonly used in most applications. It is the only place that
     * the application has where all requests come through. So you could,
     * for example, put a check for HTTP basic authentication here, and that
     * would automatically protect your entire application. Or, if you
     * were so inclined, you could have a counter that counts the total
     * pages served by your application.
     *
     * @param comms
     *   Object containing request, response and redirect objects.
     * @return
     *  true if the request has been handled, false if normal request
     *  handling should continue.
     * @exception Exception
     *   Any exception can be thrown.
     *   PageRedirectExceptions are handled as with presentation object.
     *   Other exceptions are handled by searching for an error handler
     *   starting at the requested presentation object.
     */
    public boolean requestPreprocessor(HttpPresentationComms comms) 
            throws Exception;

    
    /**
     * Request post-processing method.  All requests for URLs associated
     * with this application go through this function.  It may do any
     * post-processing desired (e.g. saving sessions to persistent store). <P>
     *
     * @param comms
     *   Object containing request, response and redirect objects.
     * @exception Exception
     *   Any exception can be thrown.
     *   PageRedirectExceptions are handled as with presentation object.
     *   Other exceptions are handled by searching for an error handler
     *   starting at the requested presentation object.
     */
    public void requestPostProcessor(HttpPresentationComms comms) 
            throws Exception;


    /**
     *  This preprocessor hook is called before the request/response pair
     *  is processed by the HttpPresentationServlet. It gives the
     *  application a chance to work with the raw request/response pair
     *  before they are wrapped by HttpPresentationRequest and
     *  HttpPresentationResponse objects. Return false to indicate
     *  that the request should continue to be processed normally
     *  or true to indicate that this method has handled the request,
     *  and all processing of the request should stop, in which case
     *  it will not be passed on to the presentation manager etc...
     *  The default method in StandardApplication always returns false.<P>
     *
     *  <B>Almost all applications should not override this!</B> The
     *  version provided by StandardApplication simply returns false.
     *  Currently the debugger is the only application that used this,
     *  to support the flow-through debugging of non-Enhydra servlets,
     *  it needs to pass the raw request/response pair to the target
     *  servlet. 
     *
     *  @param servlet
     *    The servlet we are running in.
     *  @param context
     *    The ServletContext that was used to initialize our servlet.
     *  @param request
     *    The incomming request object.
     *  @param response
     *    The incomming response object.
     *  @return
     *    True if this method handled the request, in which case no
     *    further action will be taken. Or false if normal processing
     *    should continue.
     *  @exception ServletException
     *    You are allowed to throw the same exceptions that the servlet's
     *    service() method throws.    
     *  @exception IOException
     *    You are allowed to throw the same exceptions that the servlet's
     *    service() method throws.    
     */
    public boolean servletRequestPreprocessor(Servlet servlet,
                                   ServletContext context,
                                   HttpServletRequest request,
                                   HttpServletResponse response)
          throws ServletException, IOException;


    /** 
     * Set the <CODE>LogChannel</CODE> associated with this application.
     *
     * @param The <CODE>LogChannel</CODE> to write to.
     */
    public void setLogChannel(LogChannel chan);

    /**
     * Get the <CODE>LogChannel</CODE> associated with this application.
     *
     * @return The log channel or <CODE>null</CODE> if this application
     * does not have one.
     */
    public LogChannel getLogChannel();

    /**
     * Get the <CODE>SessionManager</CODE> associated with this application.
     *
     * @return The session manager or <CODE>null</CODE> if the application
     *   doesn't have a <CODE>SessionManager</CODE>.
     */
    public SessionManager getSessionManager();

    /**
     * Get the <CODE>DatabaseManager</CODE> associated with this application.
     *
     * @return
     *   The database manager or <CODE>null</CODE> if the application
     *   doesn't have a <CODE>DatabaseManager</CODE>.
     */
    public DatabaseManager getDatabaseManager();

    /**
     * Get the <CODE>HttpPresentationManager</CODE> associated with this
     * application.
     *
     * @return
     *   The presentation manager used by this application.
     */
    public HttpPresentationManager getHttpPresentationManager();

    /**
     * Called to tell the application about the presentation manager
     * instance. This is necessary because the Presentation Manager is
     * created in parallel with the application by the
     * HttpPresentationServlet.
     * 
     * @param pm
     *   The presentation manager responsible for running this application.
     */
    public void setHttpPresentationManager(HttpPresentationManager pm);

    /**
     * Get the XMLC factory object being used by the application.
     * If one was not configured, a default one is created on the
     * first call to this method.
     */
    public XMLCFactory getXMLCFactory();

    /**
     * Set the XMLC factory based on the recload and recompilation options.
     */
    public void setXMLCFactory(boolean enableReload,
                               boolean enableRecompilation);
}
