/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: Enhydra.java,v 1.7.14.1 2000/10/19 17:59:06 jasona Exp $
 */





package com.lutris.appserver.server;

import java.util.Hashtable;
import com.lutris.appserver.server.*;
import com.lutris.appserver.server.session.SessionManager;
import com.lutris.appserver.server.sql.DatabaseManager;
import com.lutris.logging.*;

/**
 * The Enhydra class provides static methods that allow each application
 * to conviently get at their application object.  It also provides
 * access <P>
 *
 * This class implements dynamically scoped global variables, accessed
 * via static methods. Which variable is accessed (and thus which
 * Application object is returned) depends on the flow of control of
 * the program before the access (the call stack). <P>
 *
 * Currently a hashtable keyed on the threads themselves is used to
 * store the different Application, in Java 1.2 threads will have a field
 * already available for storing client data. <P>
 *
 * Normal usage of this class is to call <CODE>register()</CODE> when
 * a thread enters your application, then <CODE>unRegister()</CODE>
 * when it leaves. It is very important that every <CODE>register()</CODE>
 * is matched with an <CODE>unRegister()</CODE>, or else it will be
 * a memory leak, threads that re-enter the application will have
 * a stale Application (until the first <CODE>register()</CODE>), and
 * an extra pointer to the Application will be left around, preventing
 * garbage collection. <P>
 *
 * If no thread is passed in (the normal usage), the current thread is
 * used. If your application creates new threads, these will not initially be
 * associated with any Application. These new threads can be passed in to
 * <CODE>register()</CODE>, to assocaiate them with an Application.
 * You may call <CODE>register()</CODE> before calling <CODE>start()</CODE>
 * on the thread. <P>
 *
 * To ensure that every <CODE>register()</CODE> is balanced with an
 * <CODE>unRegister()</CODE>, the following code is recommended:
 * <PRE>
 * Enhydra.register(application);
 * try {
 *     doSomeWork(); 
 * } finally {
 *     Enhydra.unRegister();
 * }
 * </PRE>
 * This will ensure that the <CODE>unRegister()</CODE> is called even if
 * an exception is thrown, but it will not interfere with the exception
 * processing.
 * 
 * @version	$Revision: 1.7.14.1 $
 * @author	Paul Morgan
 * @author	Andy John
 * @since	LBS1.8
 */
public class Enhydra {
    // One application per thread.
    private static Hashtable applications = new Hashtable();
    
    /**
     * Prevent instantiation
     */
    protected Enhydra() {
    }


    /**
     *  Associate the given application with the current thread. 
     *  After this call, calls to <CODE>getApplication()</CODE>
     *  made by this thread will return <CODE>app</CODE>.
     *  Be sure to call <CODE>unRegister()</CODE> when you are done
     *  with the thread. This must be called before the any calls to
     *  <CODE>getApplication()</CODE>.
     * 
     * @param app The Application object to associate with the 
     * current thread. This Application will be returned by calls made
     * by this thread to <CODE>getApplication()</CODE>.
     */
    public static void register(Application app) {
        applications.put(Thread.currentThread(), app);
    }

    /**
     *  Associate the given application with the given thread. 
     *  After this call, calls to <CODE>getApplication()</CODE>
     *  made by the specified thread thread will return <CODE>app</CODE>.
     *  Be sure to call <CODE>unRegister()</CODE> when you are done
     *  with the thread. This must be called before the thread calls
     *  <CODE>getApplication()</CODE>.
     * 
     * @param thread
     *   the thread to associate with <CODE>app</CODE>.
     * @param app
     *   the Application object to associate with the 
     *   specified thread. This Application will be returned by calls made
     *   by this thread to <CODE>getApplication()</CODE>.
     */
    public static void register(Thread thread, Application app) {
        applications.put(thread, app);
    }


    /**
     * Remove the association between the current thread and it's Application.
     * This must be called when the thread is done using this Enhydra class.
     * The current thread should not call <CODE>getApplication()</CODE>
     * after this.
     */
    public static void unRegister() {
        unRegister(Thread.currentThread());
    }


    /**
     * Remove the association between the specified thread and it's Application.
     * This must be called when the thread is done using this Enhydra class.
     * The thread should not call <CODE>getApplication()</CODE>
     * after this.
     * 
     * @param thread
     *   the thread to operate on.
     */
    public static void unRegister(Thread thread) {
        if (applications.remove(thread) == null) {
            throw new Error("Thread was not registered: " + thread.getName());
        }
    }


    /**
     * Return the application object for current application.
     * Returns null if there is no Application currently associated
     * with the current thread.
     *
     * @return The application object.
     */
    public static Application getApplication() {
        return (Application) applications.get(Thread.currentThread());
    }


    /**
     * Return the database manager object for current application.
     * Returns null if there is no database manager associated
     * with the application or there is no application associated
     * with the current thread.
     *
     * @return The database manager object if available else null.
     */
    public static DatabaseManager getDatabaseManager() {
        Application app = getApplication();
        if (app != null) {
	    return app.getDatabaseManager();
	} else {
	    return null;
	}
    }
    

    /**
     * Return the session manager object for current application.
     * Returns null if there is no session manager associated
     * with the application or there is no application associated
     * with the current thread.
     *
     * @return The session manager object if available else null.
     */
    public static SessionManager getSessionManager() {
        Application app = getApplication();
        if (app != null) {
	    return app.getSessionManager();
	} else {
	    return null;
	}
    }


    /**
     * Return the LogChannel in the current Application object, or
     * a log channel to the facility "Enhydra" if there is no current
     * Application set.
     * 
     * @return The LogChannel for the current application, or a
     *   generic LogChannel.
     */
    public static LogChannel getLogChannel() {
        Application app = getApplication();
        LogChannel chan = null;
        if (app != null)
            chan = app.getLogChannel();
        if (chan != null)
            return chan;
        Logger logger = Logger.getCentralLogger();
        if (logger == null)
            return null;  // This should never happen.
        return logger.getChannel("Enhydra");
    }
}
