/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: StandardApplication.java,v 1.62.8.2 2000/11/16 09:09:55 markd Exp $
 */

package com.lutris.appserver.server;

import java.net.*;
import com.lutris.util.*;
import com.lutris.logging.*;
import com.lutris.appserver.server.httpPresentation.*;
import com.lutris.appserver.server.session.*;
import com.lutris.appserver.server.sql.*;
import com.lutris.classloader.MultiClassLoader;
import java.io.IOException;
import javax.servlet.*;
import javax.servlet.http.*;
import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.xmlc.EnhydraXMLCLogger;
import org.enhydra.xml.xmlc.reloading.RecompilationHandler;
import org.enhydra.xml.xmlc.reloading.StandardRecompilationHandler;
import org.enhydra.xml.xmlc.reloading.XMLCReloadingFactory;



import org.enhydra.servlet.connectionMethods.EnhydraDirector.EnhydraDirectorConnection;


/**
 * Default application implementation. This class provides standard start/stop
 * services.  This can be used as a base class to derive Application objects.
 *
 * @version	$Revision: 1.62.8.2 $
 * @author	Mark Diekhans
 */
public abstract class StandardApplication implements Application {

    /**
     * Determines if a cookie with the session id is sent to the client.
     */
    protected boolean sendCookieForNewSession = true;

    /**
     * The name of the application; defaults to the unqualified
     * class name.
     */
    protected String appName = null;

    /**
     * Current state of the application.  Only modified by the derived
     * application.
     */
    protected int state = STOPPED;

    /**
     * Session manager for all application sessions. A session manager
     * manages the creation, deletion and assocation of sessions with
     * a user.
     */
    protected SessionManager sessionManager;

    /**
     * Presentation manager instance for application. The presentation
     * manager handles incomming http requests.
     */
    protected HttpPresentationManager presentationManager;

    /**
     * Database manager instance for application. The database manager
     * manages logical database and is responsible for allocating
     * connection, create transactions, etc.
     */
    protected DatabaseManager databaseManager;

    /**
     * This applications config object. Every application has a configuration
     * object that is created from the application configuration file.
     *
     * @see Config
     */
    protected Config config;

    /**
     * Default URL used for application if defined in Config file. This
     * is used to redirect the user to the preferred start page if a
     * URL equal to the root of this application is given. By setting
     * this to index.html, it behaves in the same way as a http server.
     */
    protected String defaultUrl;

    /**
     * The log channel for this application to write to.
     */
    protected LogChannel logChannel;

    /**
     * Application Data accessable through Jolt Fields.
     */
    protected ApplicationData data = new ApplicationData ();

    /**
     * XMLC Factory.
     */
    protected XMLCFactory xmlcFactory;

    /**
     * A constructor with no arguments is required.
     */
    public StandardApplication() {
    }


    /**
     * Get the application state.
     * N.B. Purposely not syncronized for speed.
     *
     * @return
     *   The application's state code.
     */
    public int getState() {
        return state;
    }

    /**
     * Tells the application if a cookie is used to bind new sessions
     * to a client. If no cookie is set, session ID's must be
     * communicated by means of URL rewriting.
     * @param flag indicates whether to send a cookie or not.
     */
    public void setCookieForNewSession(boolean flag) {
        sendCookieForNewSession = flag;
    }

    /**
     * Get the application's config object.
     *
     * @return
     *   The application's config object.
     */
    public Config getConfig() {
        return config;
    }


    /**
     * Set the application name.
     *
     * @param name The new name for the application.
     */
    public void setName(String name) {
        appName = name;
    }


    /**
     * Get the application symbolic name.  This is normally the
     * package name of this application class. <P>
     * N.B. Purposely not syncronized for speed.
     *
     * @return
     *   The symbolic name.
     */
    public String getName() {
        return appName;
    }


    /**
     * Set the <CODE>LogChannel</CODE> associated with this application.
     *
     * @param The <CODE>LogChannel</CODE> to write to.
     */
    public void setLogChannel(LogChannel chan) {
        logChannel = chan;
    }


    /**
     * Get the <CODE>LogChannel</CODE> associated with this application.
     *
     * @return The log channel or <CODE>null</CODE> if this application
     * does not have one.
     */
    public LogChannel getLogChannel() {
        return logChannel;
    }


    /**
     * Start the application.  The default method sets the state to
     * <CODE>RUNNING</CODE>.
     *
     * @param appConfig
     *   Application configuration object.
     * @exception ApplicationException
     *   If an error occurs starting the application.
     */
    public synchronized void startup(Config appConfig)
	throws ApplicationException {

        if (state == RUNNING)
            return;  // Handle multiple threads doing a startup.

        printCopyrightHeader();

	/**
	 * Read config object.
	 */
	Config sessionConfig, databaseConfig;
	try {
	    if (appConfig.containsKey("Application")) {
		defaultUrl = appConfig.getString("Application.DefaultUrl");
	    }
	    sessionConfig = (Config)appConfig.getSection("SessionManager");
	    databaseConfig = (Config)appConfig.getSection("DatabaseManager");
	} catch (KeywordValueException except) {
	    throw new ApplicationException(except);
	}

	/**
	 * Create session manager...
	 */
	sessionManager = createSessionManager(sessionConfig);

	/**
	 * Create database manager if required...
	 */
	if (databaseConfig != null) {
	    databaseManager = createDatabaseManager(databaseConfig);
	}

	/**
	 * Remember config object...
	 */
	config = appConfig;

        state = RUNNING;
    }


    public void printCopyrightHeader() {
        /**
         *  Print out a copyright notice. In most cases this will
         *  fulfill a third party's copyright obligations when using Enhydra.
         */
        System.out.println("Enhydra Java Application Server");
        System.out.println("Copyright 1997-2000 Lutris Technologies, Inc.");
        System.out.println("All rights reserved.");
        if (logChannel != null) {
            logChannel.write(Logger.INFO, "Enhydra Java Application Server");
            logChannel.write(Logger.INFO, "Copyright 1997-2000 Lutris Technologies, Inc.");
            logChannel.write(Logger.INFO, "All rights reserved.");
        }
    }


    /**
     * Continue the startup up process with the application is in the
     * <CODE>INCOMPLETE</CODE> state.  The default method generates an
     * error, as the application should never be in the <CODE>INCOMPLETE</CODE>
     * state if it doesn't implment this method.
     *
     * @param appConfig
     *   The same <CODE>appConfig</CODE> object that was passed to
     *   <CODE>startup</CODE>.
     * @exception ApplicationException
     *   If an error occurs restarting the application.
     */
    public synchronized void restartup(Config appConfig)
            throws ApplicationException {
        throw new ApplicationException("no support for INCOMPLETE state");
    }


    /**
     * Shutdown the application.  The default method sets the state to
     * <CODE>STOPPED</CODE>.
     */
    public synchronized void shutdown() {
        if (state == STOPPED)
            return;  // Handle multiple threads doing a shutdown.

        if (sessionManager != null) {
            sessionManager.shutdown();
            sessionManager = null;
        }
        if (databaseManager != null) {
            databaseManager.shutdown();
            databaseManager = null;
        }

        state = STOPPED;
    }


    /**
     * Create the session manager to be used by this application. By
     * default this will return the <EM>Lutris</EM> standard session manager.
     * This method can be overwritten to create a custom session
     * manager.
     *
     * @param sessionMgrConfig
     *   Configuration object containing "SessionManager" keys.
     *   This can be <CODE>null</CODE> to use defaults.
     * @exception ApplicationException
     *   If an error occurs in creating a session manager.
     * @return
     *   The session manager.
     */
    protected SessionManager createSessionManager(Config sessionMgrConfig)
	throws ApplicationException {

        if (sessionMgrConfig == null) {
            sessionMgrConfig = new Config();
        }
	try {
	    return new StandardSessionManager(this, sessionMgrConfig, logChannel);
	} catch (Exception except) {
	    throw new ApplicationException(except);
	}
    }


    /**
     * Create the database manager to be used by this application. By
     * default this will return the <EM>Lutris</EM>
     * <CODE>StandardDatabaseManager</CODE>.
     * This method can be overwritten to create a custom session
     * manager.
     *
     * @param databaseMgrConfig
     *   Configuration object containing "DatabaseManager" keys.
     * @exception ApplicationException
     *   If an error occurs in creating a database manager.
     */
    protected DatabaseManager createDatabaseManager(Config databaseMgrConfig)
	throws ApplicationException {

	try {
	    return new StandardDatabaseManager(databaseMgrConfig);
	} catch (Exception except) {
	    throw new ApplicationException(except);
	}
    }


    /**
     * Create a new session for application. By default this will create
     * the <EM>Lutris</EM> <CODE>StandardSession</CODE>.
     * This method can be overwritten to create a custom session.
     *
     * @param comms
     *   Object containing request, response and redirect objects.
     *   <I>Warning: comms.session and comms.sessionData will be null.
     *   they are filled in after this method returns.</I>
     * @return
     *   The session.
     * @exception ApplicationException
     *   If an error occurs in creating the session.
     */
    protected Session createSession(HttpPresentationComms comms)
                      throws ApplicationException {
        try {
	    Session s = null;
	    String saTok = null;
	    try  {
		saTok = comms.request.getHeader(":aff:");
	    } catch (HttpPresentationException hpe) { saTok = null; }
	    if (saTok == null) {
		s = sessionManager.createSession();
	    } else {
		s = sessionManager.createSession(saTok);
		if (logChannel != null) {
		    logChannel.write(Logger.DEBUG,
				 "SessionKey created for EnhydraDirector");
		}
	    }
	    return s;
        } catch (SessionException e) {
            throw new ApplicationException(e);
        }
    }


    /**
     * Default method used by <CODE>requestPreprocessor</CODE>
     * to ensure that a session exists and initialize the
     * <CODE>session</CODE> and <CODE>sessionData</CODE> fields
     * in the <CODE>HttpPresentationComms</CODE> object.
     * New sessions are only created on requests to presentation
     * objects, not requests for other types of files (such as gifs).
     * This avoids allocating multiple sessions when a browser makes
     * multiple requests for HREFs.  If the session already exist, it
     * is alwasy set in <CODE>comms</CODE><P>
     *
     * This normally looks up the session using a session key from
     * a cookie.
     *
     * @param comms
     *   Object containing request, response and redirect objects.
     * @exception ApplicationException
     *   If an error occurs setting up the session.
     */
    protected void ensureSession(HttpPresentationComms comms)
        throws ApplicationException {

        try {
            if (presentationManager.isPresentationRequest(comms.request)) {
                // FIX - verify OK
                comms.session = StandardAppUtil.getRequestSession(comms);
            }
            if (comms.session != null &&
		sessionManager.sessionExists(comms.session.getSessionKey())) {
                /*
                 * InitializeNewSession() sets up comms.sessionData, so
                 * that it may be used if an application defines it's own
                 * version of that method (after calling the super method,
                 * it has a normal comms object to work with).
                 * Therefore, if we did not call initializeNewSession(),
                 * we have to initialize comms.sessionData here.
                 */
                comms.sessionData = comms.session.getSessionData();
            } else {

                if (presentationManager.isPresentationRequest(comms.request)) {
                    initializeNewSession(comms);
                }
            }
        } catch (HttpPresentationException except) {
            throw new ApplicationException(except);
        } catch (SessionException except) {
            throw new ApplicationException(except);
        }
    }

    /**
     * Calls application session manager's <i>passivateSession</i> method
     * for the session associated with this request.
     *
     * @param comms
     *   Object containing request, response and redirect objects.
     * @exception ApplicationException
     *   If the session manager was unable to "passivate" the session.
     * @see com.lutris.appserver.server.session.SessionManager.passivateSession
     */
    public void requestPostProcessor(HttpPresentationComms comms)
        throws ApplicationException {
        try {
            if (comms.session != null) {
                comms.application.getSessionManager().
                    passivateSession(Thread.currentThread(),
                                     comms.session.getSessionKey());
            }
        } catch (SessionException e) {
            throw new ApplicationException(e);
        }
    }


    /**
     *  This is called when no valid session is found in an incoming request.
     *  A new session is created and added to <CODE>comms</CODE>. If
     *  sending session cookies is enabled, a cookie is created and
     *  added to the response, so future requests will be associated
     *  with the newly created session. If sending session cookies is
     *  disabled, it is the application's responsibility to include
     *  the session ID in URLs to associate further requests with the
     *  session.<P>
     *
     *  <I>Warning:</I> At first <CODE>comms.session</CODE>
     *  and <CODE>comms.sessionData</CODE> are null. This method initializes
     *  them both.
     *
     * @param comms
     *   Object containing request, response and redirect objects. Session
     *   and SessionData objects are initialized.
     * @exception ApplicationException
     *   If an error occurs setting up the session.
     * @see #setCookieForNewSession
     */
    protected void initializeNewSession(HttpPresentationComms comms)
                   throws ApplicationException {
        comms.session = createSession(comms);
        if (sendCookieForNewSession) {
            StandardAppUtil.bindSessionToClient(comms);
        }
        comms.sessionData = comms.session.getSessionData();
    }


    /**
     * Returns the specified url rewritten to include the sessionId
     *
     * @param url String to urlencode
     * @return the param url rewritten to contain the sessionId
     */
    public String encodeUrl(String url, String sessionKey)
      throws ApplicationException{

      return StandardAppUtil.encodeUrl(url, sessionKey);
    }


    /**
     * Returns the specified url rewritten to include the sessionId
     *
     * @param url String to urlencode
     * @return the param url rewritten to contain the sessionId
     * @deprecated Use encodeUrl(String,String)
     */
    public String encodeUrl(String url, String sessionKey, String appName)
      throws ApplicationException{

      return StandardAppUtil.encodeUrl(url, sessionKey);
    }


    /**
     * Application Data accessable through Jolt Fields.
     */
    public ApplicationData getApplicationData () {
	return data;
    }


    /**
     * Default application request preprocessor.  This method sets up
     * a session data structure if a cookie is not set for the application.
     * It doesn't require a login however.  The unauthenticated session is
     * useful for persistence in simple applications.  It is possible to
     * later force a login.  If the URL is not a reference to a presentation
     * object, then a session will not be allocated if it does not already
     * exist.  This prevents allocation of multiple sessions when a browser
     * is requesting other objects, such as gifs, in parallel.  Thus, after
     * this function is called, it is possible for session to be null. <P>
     *
     * N.B. Purposely not syncronized for speed/concurrency.
     *
     * @param comms
     *   Object containing request, response and redirect objects.
     * @return
     *   Always returns false, as request handling is not done.
     * @exception Exception
     *   May throw any exception, as with a presentation object;
     *   including page redirects.
     */
    public boolean requestPreprocessor(HttpPresentationComms comms)
            throws Exception {

	/**
	 * Remap / ---> "defaultUrl" if defined.
	 */
	if (defaultUrl != null) {
	    String ap = comms.request.getApplicationPath();
	    if (!ap.endsWith("/"))
		ap += "/";
	    String page = comms.request.getPresentationObjectPath();
	    if (!page.endsWith("/"))
		page += "/";
	    if (page.equals(ap)) {
                // Don't want two slashes in a row.
                String target = comms.request.getHttpServletRequest().getContextPath();
                if (target.equals("/")) {
                    target = "";
                }
                target = "";

                if (defaultUrl.startsWith("/")) {
                    target += ap + defaultUrl.substring(1);
                } else {
                    target += ap + defaultUrl;
                }
                // get query string and add it to target
                String query = comms.request.getQueryString();
                if (query != null) {
                    target += "?" + query;
                }
                throw new ClientPageRedirectException(target);
            }
	}

        // Only need the session if we haven't redirected to a new page.
        ensureSession(comms);

        return false;	// Proceed normally.
    }


    /**
     * Get the <CODE>SessionManager</CODE> associated with this application.
     *
     * @return
     *   The session manager or <CODE>null</CODE> if the application
     *   doesn't have a <CODE>SessionManager</CODE>.
     */
    public SessionManager getSessionManager() {
        return sessionManager;
    }


    /**
     * Get the <CODE>DatabaseManager</CODE> associated with this application.
     *
     * @return
     *   The database manager or <CODE>null</CODE> if the application
     *   doesn't have a <CODE>DatabaseManager</CODE>.
     */
    public DatabaseManager getDatabaseManager() {
        return databaseManager;
    }


    /**
     * Get the <CODE>HttpPresentationManager</CODE> associated with this
     * application.
     *
     * @return The presentation manager.
     */
    public HttpPresentationManager getHttpPresentationManager() {
        return presentationManager;
    }


    /**
     * Tells the application about the Presentation Manager running it.
     * This is necessary because the presentation manager is created
     * before the application.
     *
     * @param pm The HTTP presentation manager running this application.
     */
    public void setHttpPresentationManager(HttpPresentationManager pm) {
        presentationManager = pm;
    }



    /**
     * This is a hook that allows applications to act on requests before
     * they enter the Enhydra framework. This implementation always returns
     * false, so the request is processed normally. <B>Do not override this
     * unless you absolutly have to.</B> Currently the only application
     * that needs to is the debugger, because it needs to hand off requests
     * to other non-Enhydra servlets. <P>
     *
     * In contrast, the method requestPreprocessor() is commonly overridden.
     * It provides applications with a centeral place to put code that has
     * to be run on every request before it is sent to the presentation
     * objects.
     *
     *  @param servlet
     *    The servlet we are running in.
     *  @param context
     *    The ServletContext that was used to initialize our servlet.
     *  @param request
     *    The incomming request object.
     *  @param response
     *    The incomming response object.
     *  @return
     *    True if this method handled the request, in which case no
     *    further action will be taken. Or false if normal processing
     *    should continue.
     *  @exception ServletException
     *    You are allowed to throw the same exceptions that the servlet's
     *    service() method throws.
     *  @exception IOException
     *    You are allowed to throw the same exceptions that the servlet's
     *    service() method throws.
     */
    public boolean servletRequestPreprocessor(Servlet servlet,
                                   ServletContext context,
                                   HttpServletRequest request,
                                   HttpServletResponse response)
                   throws ServletException, IOException {
        return false;
    }

    /**
     * Get the XMLC factory object being used by the application.
     */
    public XMLCFactory getXMLCFactory() {

        return xmlcFactory;
    }

    /**
     * Set the XMLC factory based on the recload and recompilation options.
     */
    public void setXMLCFactory(boolean enableReload,
                               boolean enableRecompilation) {
        if (enableReload || enableRecompilation) {
            RecompilationHandler recompiler = null;
            if (enableRecompilation) {
                recompiler = new StandardRecompilationHandler();
            }
            xmlcFactory =
                new XMLCReloadingFactory(recompiler,
                                         (MultiClassLoader)presentationManager.getAppClassLoader(),
                                         new EnhydraXMLCLogger(logChannel));
        } else {
            xmlcFactory = new XMLCStdFactory(presentationManager.getAppClassLoader(),
                                             new EnhydraXMLCLogger(logChannel));
        }
    }
}
