/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: AppletUtils.java,v 1.10.14.1 2000/10/19 17:59:06 jasona Exp $
 */





package com.lutris.appserver.server.httpPresentation;

import com.lutris.util.*;
import com.lutris.http.*;
import com.lutris.applet.LBSConnection;
import java.util.*;
import java.io.*;
import java.lang.reflect.*;


/**
 *   This is the counterpart to com.lutris.applet.LBSConnection. This class
 *   creates <APPLET ...> tags with extra initialization parameters.
 *   LBSConnection reads those extra initialization parameters and uses them
 *   to establish a connection back to the Mutliserver.
 *   This file is used by a presentation object, while LBSConnection is 
 *   included in the applet's jar file or classes directory. This way it
 *   does not need to include the com.lutris.appserver.server.* classes.
 *
 * @see com.lutris.applet.LBSConnection.
 */
public class AppletUtils {

    /**
     * Private constructor to prevent instantiation; use the static methods.
     */
    private AppletUtils() {
    }


    /**
     *
     *  This assumes that if cookies are being used, the name of the cookie
     *  is the application name. This is true if the StandardApplication is
     *  being used.
     *
     *  @param comms
     *   HTTP communications object. Contains objects and interfaces to read
     *   the request and send a response.
     *  @param targetURL
     *   The URL of the presentation object the applet should connect to.
     *   For example: "/package/Presentaion.po". The server name and
     *   presentation path will automatically be added.
     *  @param options
     *   A list of various applet options to be included in the applet tag.
     *   For example: "code=myApplet.class width=400 height=100". These
     *   three are required options.
     *  @param parameterNames
     *   An array of parameter names, for example: {"X", "Y", "maxIterations"}.
     *   This is optional.
     *  @param parameterValues
     *   An array of parameter values, for example: {"24", "19", "2000"}.
     *   Either this an <CODE>parameterNames</CODE> must be null, or they
     *   must have the same number of elements. <CODE>parameterNames[i]</CODE>
     *   is matched up with <CODE>parameterValues[i]</CODE>.
     *   This is optional.
     *  @param alternateHtml Any HTML to display if the browser does not
     *   support applets. This is optional, and may be null.
     *  @return  
     *   A string of the form "<APPLET code=... > <PARAM NAME=... VALUE=...>
     *   .....   </APPLET>".
     */
    public static String createAppletTag(
                             HttpPresentationComms comms,
                             String targetURL,
                             String options,
                             String[] parameterNames,
                             String[] parameterValues,
                             String alternateHtml) {
        /*
         * Check input.
         */
        if (targetURL == null)
            targetURL = ""; // This is an error, but ignore for now.
        if (options == null)
            options = ""; // This is an error, but ignore for now.
        if (alternateHtml == null)
            alternateHtml = "";
        /*
         *  Start off with the APPLET tag and the options.
         */
        StringBuffer result = new StringBuffer(); 
        result.append("\n<APPLET ");
        result.append(options);
        result.append(">\n");
        /*
         *  Add any user-supplied parameters.
         */
        if ((parameterNames != null) && (parameterValues != null))
            for (int i=0; i<parameterNames.length; i++) {
                result.append("<PARAM NAME=");
                result.append(parameterNames[i]);
                result.append(" VALUE=");
                result.append(parameterValues[i]);
                result.append(">\n");
            }
        /*
         *  Figure out the cookie settings.
         */
        String cookieName = comms.application.getName(); 
        String cookieValue = comms.session.getSessionKey();
        /*
         *  Build up a complete URL for the target.
         */
        String server = comms.request.getServerName();
        int port = comms.request.getServerPort();
        String pp = "/";
        try {
            pp = comms.request.getApplicationPath();
        } catch (HttpPresentationException e) {
        }
        if (!pp.startsWith("/"))
            pp = "/" + pp;
        if (!pp.endsWith("/"))
            pp = pp + "/";
        if (targetURL.startsWith("/"))
            targetURL = targetURL.substring(1);
        String fullURL = "http://" + server;
        if (port != 80)
            fullURL += ":" + port;
        fullURL += pp + targetURL;
        /*
         *  Add the target and the session cookie parameter. 
         *  The applet will need this
         *  when it issues requests back to the server, to access the current
         *  session.
         *  See StandardAppUtil.bindSessionToClient().
         */ 
        result.append("<PARAM NAME=");
        result.append(LBSConnection.nameParamName);
        result.append(" VALUE=");
        result.append(cookieName);
        result.append(">\n");
        result.append("<PARAM NAME=");
        result.append(LBSConnection.valueParamName);
        result.append(" VALUE=");
        result.append(cookieValue);
        result.append(">\n");
        result.append("<PARAM NAME=");
        result.append(LBSConnection.targetParamName);
        result.append(" VALUE=");
        result.append(fullURL);
        result.append(">\n");
        /*
         *  Finish off with the alternate html, then close the tag.
         */
        result.append(alternateHtml);
        result.append("\n</APPLET>\n");
        /*
         *  All done.
         */
        return result.toString(); 
    }


}
