/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HttpPresentationIOException.java,v 1.6.14.1 2000/10/19 17:59:06 jasona Exp $
 */





package com.lutris.appserver.server.httpPresentation;
import java.io.IOException;
import com.lutris.util.*;


/**
 * IOException derived class that is thrown when the presentation manager
 * encounters an I/O error when talking to a client.  This allows the methods
 * to determine how the error should be handled, even though it was generated
 * by a call that occured in presentation code.
 */
public class HttpPresentationIOException extends IOException {
    /**
     * Static method that does the approriate conversion from an IOException
     * to a HttpPresentationIOException If the exception is already an
     * HttpPresentationIOException, it is simple rethrown.  If it is another type of
     * IOException, it is converted to a HttpPresentationIOException.
     *
     * @param except The exception to rethrow or encapsulate.
     * @return An exception ready to throw.
     */
    public static IOException rethrow(IOException except)
            throws IOException {
        if (except instanceof HttpPresentationIOException) {
            return except;
        } else {
            return new HttpPresentationIOException(except);
        }
    }


    /**
     * Flag that indicates that this came from an a I/O operation
     * rather than a generated exception.
     */
    private boolean trueIOException = false;

    /**
     * Construct a new exception from an existing IOException.
     *
     * @param except The exception to encapsulate.
     */
    public HttpPresentationIOException(IOException except) {
        super(except.getMessage());
        trueIOException = true;
    }

    /**
     * Construct a new exception with a specific message.
     *
     * @param msg The error message
     */
    public HttpPresentationIOException(String msg) {
        super(msg);
    }

    /**
     * Did this exception come from an a I/O operation
     * rather than a generated exception.
     * @return True if its an actual I/O exception. 
     */
    public boolean isTrueIOException() {
        return trueIOException;
    }

    /**
     * Given a exception or error, determine if its generated by a HTTP
     * client I/O exception.  This handles ChyainedExceptions that might
     * contain a client I/O exception.
     *
     * @param except Exception to check.
     * @return <CODE>true</CODE> if this is cause by the client socket,
     *  <CODE>false</CODE> if not.
     */
    public static boolean isClientIOException(Throwable except) {
        Throwable scan = except;
        while (scan != null) {
            if ((scan instanceof HttpPresentationIOException)
                && ((HttpPresentationIOException)scan).isTrueIOException()) {
                return true;
            }
            if (scan instanceof ChainedException) {
                scan = ((ChainedException)scan).getCause();
            } else {
                return false;
            }
        }
        return false;
    }

}
