/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HttpPresentationManager.java,v 1.67.2.1.4.1 2000/10/19 17:59:06 jasona Exp $
 */





package com.lutris.appserver.server.httpPresentation;

import com.lutris.http.*;
import com.lutris.appserver.server.*;
import com.lutris.logging.*;
import java.io.*;

import com.lutris.appserver.server.session.*;
import javax.servlet.*;
import javax.servlet.http.*;


/**
 * Presentation manager class.  Handles loading and execution of 
 * presentation objects.
 *
 * @version	$Revision: 1.67.2.1.4.1 $
 * @author	Mark Diekhans
 */
public class HttpPresentationManager {

    /**
     * Maximum number of times to go through page redirect/error
     * handling loops.
     */
    private final int maxRedirectErrorLoops = 25;

    /**
     * Presentation and resource loader/cache.
     */
    private PresentationLoader loader;

    /**
     * Application that this presentation manager is associate with.
     */
    private Application application;

    /**
     * Applications log channel.  No logging is done if null.
     */
    private LogChannel logChannel;
    
    /**
     * Is DEBUG logging enabled?
     */
    private boolean debugLoggingEnabled;

    /**
     * Request id incremented on each request.
     */
    private long nextRequestId = 0;


    /**
     *  The Servlet we are running in. Will be an instance of
     *  <CODE>
     *  com.lutris.appserver.server.httpPresenation.HttpPresentationServlet
     *  </CODE>.
     */
    private Servlet servlet;

    /**
     *  The ServletContext that was used to initialize the Servlet we are
     *  running in. This is needed to, for example, get the other servlets 
     *  that are our peers in the server.
     */
    private ServletContext servletContext;


    /**
     * Construct a HTTP presentation manager.
     *
     * @param appPresentationPrefix Prefix added to the file name portion
     *   of the URL.  The resulting names is used to search for classes and
     *   files on the class path.
     * @param presApplication Application object that this presentation is
     *   associated with.
     * @param applicationClassLoader
     *   the class loader to use for the application.
     * @param cacheClasses Enables or disables caching of presentation object
     *   classes in memory.
     * @param cacheFiles Enables or disables caching of files (html, gif, etc)
     *   that are servered as part of the application.
     */
    public HttpPresentationManager(String appPresentationPrefix,
                                   Application presApplication,
                                   ClassLoader applicationClassLoader,
                                   boolean cacheClasses,
                                   boolean cacheFiles)
            throws HttpPresentationException {

        application = presApplication;
        logChannel = application.getLogChannel();
        if (logChannel != null) {
            debugLoggingEnabled = logChannel.isEnabled(Logger.DEBUG);
        }
        loader = new PresentationLoader(appPresentationPrefix,
                                        applicationClassLoader,
                                        cacheClasses,
                                        cacheFiles,
                                        logChannel);
    }

   



    /**
     * The method is invoked to fulfill an HTTP request.  This function is
     * thread-safe and does not need synchronization.  It is expected that the
     * protocol layer invoking this function provide multi-threading to handle
     * requests.
     *
     * The alogrithm for handling a request is:
     * <PRE>
     * presObj = requestUrl
     * while request-not-handled {
     *    if not error-handler-redirect
     *        invoke-application-request-preprocessor presObj
     *    load-presentation-object-class
     *    invoke-presentation-object
     *    if not error-handler-redirect
     *        invoke-application-request-postprocessor presObj
     *    if (client-side-page-redirect-exception) {
     *        send-redirect-to-client
     *        break
     *    }
     *    if (server-side-page-redirect-exception) {
     *        (Not implemented yet)
     *        set-request-to-redirectUrl
     *        presObj = redirectUrl
     *        continue
     *    }
     *    if (401-unauthorized-exception) {
     *        set-header-and-response-code
     *        break
     *    }
     *    if (other-exception) {
     *        presObj = find-error-handler
     *    }
     * }
     * </PRE>
     *
     * @param request Request object for the protocol invoking the manager.
     * @param response Response object for the protocol invoking the manager.
     * @exception HttpPresentationException All exceptions will be encapsulated
     *   in an exception of this class.
     */
    public void Run(HttpPresentationRequest request,
                    HttpPresentationResponse response)
        throws HttpPresentationException, IOException {

        long requestId;

        synchronized (this) {
            requestId = nextRequestId++;
        }


        //FIXME: Error handle has gotten very complex and needs to be
        // revisited (again).
        HttpPresentationComms comms =
		new HttpPresentationComms (request, response, application);

        /*
         * Loop calling the presentation and then PageRedirect or ErrorHandlers
         * until the request is handled.  This is a state machine that is either
         * loading request page or an error handler.  Server side redirect can 
         * cause a state transition back to the request page state.
         * The application can also handle the request.
         */

        /*
         * N.B. getPresentationObjectPathInfo may return null on an invalid
         * URL or empty, at least with servlet runner and using just the
         * name of the application.  Handled by runPresObj.
         */
	String presObjPath = request.getPresentationObjectRelativePath();

        if (presObjPath == null) {
            // If presObj is null, it could be an invalid URL that just 
	    // references the application.  See what happens.
            presObjPath = "/";
        }

        String requestedPrseObjPath = presObjPath;
        int loopCount = 0;

        // When this variable is not-null, an exception is being handled.
        Throwable exception = null;

        presLoop:
        while (true) {
            logPresObjRun(comms, presObjPath, exception, 
			  requestedPrseObjPath, requestId);
            loopCount++;
            try {
                if (exception == null) {
                    if (runRequestPreprocessor(comms, requestId, 
                                               presObjPath)) {
                        break presLoop;  // Finished!
                    }
                }
                runPresentationObj(comms, presObjPath, exception, application);
                break presLoop;  // Finished!`
            } catch (PageRedirectException except) {
                presObjPath = PageRedirect.handler(comms, presObjPath, except, logChannel, requestId);
                if (presObjPath == null) {
                    break presLoop;  // Finished!
                }
                exception = null; // Server redirected, No longer handling an exception.
            } catch (PageUnauthorizedException except) {
                ExceptionHandler.sendUnauthorizedPage(comms,
                                                      logChannel,
                                                      requestId,
						      application,
                                                      presObjPath,
						      except);

                break presLoop;  // Finished!
            } catch (Throwable except) {
                logChannel.write(Logger.DEBUG, "RID:" + requestId
                                 + ": exception running presentation"
                                 + presObjPath,
                                 except);
                // Don't execute error handler on client disconnect.
                if (HttpPresentationIOException.isClientIOException(except)) {
                    // We don't except this to actually display anything, just
                    // log.
                    ExceptionHandler.logDisplayException(comms, logChannel, requestId,
                                                         application, except, request.getPathInfo());
                    // We always run the post processor
                    try {
                        runRequestPostProcessor(comms, requestId, presObjPath);
                    } catch (Throwable e) {
                        // FIX - OK? - should log details of error.
                        logChannel.write(Logger.DEBUG, "RID:" + requestId
                                         + ": Request Post Proecessor exception: "
                                         + presObjPath + ": " + e);
                    }
                    return;  // Finished, but don't break to second flush.
                }

                // Look for the next error handler.
                String exceptpresObjPath = presObjPath;
                presObjPath = 
		    ExceptionHandler.getErrorHandler(presObjPath,
						     requestedPrseObjPath,
						     (exception != null),
						     loader,
						     logChannel,
						     requestId);
                if (presObjPath == null) {
                    // Top of tree, no more places to check.
                    ExceptionHandler.processUnhandledException(comms,
                                                               logChannel,
                                                               requestId,
                                                               application,
                                                               requestedPrseObjPath,
                                                               except);
                    break presLoop;  // Finished!
                }
                exception = except;  // Now handling this exception.
            }
            if (loopCount > maxRedirectErrorLoops) {
                ExceptionHandler.maxRedirectsReached(comms,
                                                     logChannel,
                                                     requestId,
                                                     maxRedirectErrorLoops,
                                                     application,
                                                     presObjPath);
                break presLoop;  // Finished!
            }
        }


        try {
            response.flush();
            runRequestPostProcessor(comms, requestId, presObjPath);
        } catch (Throwable except) {
            // We don't except this to actually display anything, just
            // log.
            ExceptionHandler.logDisplayException(comms, logChannel, requestId, application, except,
                                                 request.getPathInfo());
        }
    }


    /**
     * Log the execution of a presentation object.
     */
    private void logPresObjRun(HttpPresentationComms comms,
                               String presObjPath,
                               Throwable exception,
                               String requestedPrseObjPath,
                               long requestId) throws HttpPresentationException {
        if (debugLoggingEnabled) {
            if (exception != null) {
                logChannel.write(Logger.DEBUG, "RID:" + requestId
                                 + ": presentation run exception handler: " 
				 + presObjPath + " for:\n"
                                 + "    " + exception.getClass().getName() 
				 + ": " + exception.getMessage()
                                 + "\nwhile accessing: " 
				 + requestedPrseObjPath);
            } else if (comms.request.getMethod().equals("HEAD")) {
                logChannel.write(Logger.DEBUG, "RID:" + requestId
                                 + ": presentation head : " + presObjPath);
            } else {
                logChannel.write(Logger.DEBUG, "RID:" + requestId
                                 + ": presentation run: " + presObjPath);
            }
        }
    }


    /**
     * This is an internal use only hook, see the description in
     * Application.java.  The call is passed through to the application.
     */
    public boolean servletRequestPreprocessor(Servlet me, 
                                   ServletContext context,
                                   HttpServletRequest request,
                                   HttpServletResponse response) 
                   throws ServletException, IOException {
        if (application != null)
            return application.servletRequestPreprocessor(
                                    me, context, request, response);
        else
            return false;
    }



    /**
     * Run the request preprocessor.
     */
    private boolean runRequestPreprocessor(HttpPresentationComms comms,
                                           long requestId,
                                           String presObjPath) throws Exception {
        if (application.requestPreprocessor(comms)) {
            if (debugLoggingEnabled) {
                logChannel.write(Logger.DEBUG, "RID:" + requestId
                                 + ": application.requestPreprocessor finished request: "
                                 + presObjPath);
            }
            return true;
        }
        return false;
    }

    /**
     * Run the request post-processor.
     */
    private void runRequestPostProcessor(HttpPresentationComms comms,
                                         long requestId,
                                         String presObjPath)
        throws Exception {
        application.requestPostProcessor(comms);
    }

    /**
     * Load and run a presentation object (including error handlers).
     * A return without exception indicates that the presentation has been
     * processed.  Will rethrow an exception if one is being handled and
     * a handler is not found.
     *
     * @param comms
     *   comms object.
     * @param urlPath
     *   the path of the url.
     * @param exception
     *   if not null, it indicates that an exception or error is being handled.
     * @param application
     *   the application
     */
    private void runPresentationObj(HttpPresentationComms comms,
                                    String urlPath,
                                    Throwable exception,
                                    Application application) throws Throwable {

        HttpPresentation presObj;
        comms.exception = exception;

        // Get the presentation object.  If a class is not found and we
        // where handling an error, rethrow that error.
        try {
            presObj = loader.loadPresentation(urlPath); 
        } catch (ClassNotFoundException noClass) {
            if (exception != null) {
                throw exception;  // ErrorHandler was not found.
            } else {
                throw noClass;
            }
        }
        presObj.run(comms);
    }


    /**
     * Determine if the PO cache is enabled.
     *
     * @return <code>true</code> is the cache is enabled.
     */
    public boolean isPOCacheEnabled() {
        return loader.isPOCacheEnabled();
    }

    /**
     * Return the number of entries in the PO cache.
     *
     * @return the number of entries in the cache or 0 is disabled.
     */
    public int sizeofPOCache() {
        return loader.sizeofPOCache();
    }

    /**
     * Determine if the file resource class cache is enabled.
     *
     * @return <code>true</code> is the cache is enabled.
     */
    public boolean isResourceCacheEnabled() {
        return loader.isResourceCacheEnabled();
    }

    /**
     * Return the number of entries in the resource cache.
     *
     * @return the number of entries in the cache or 0 is disabled.
     */
    public int sizeofResourceCache() {
        return loader.sizeofResourceCache();
    }

    /**
     * Get a file associated with the application.  The file is located
     * in the same manner as presentation objects are located.  That is,
     * prepending the <CODE>presentationPrefix</CODE> to the specified
     * path and searching the class path for a directory or JAR containing
     * the file.
     *
     * @param appfileName The file name relative to the
     *  to the application's <CODE>presentationPrefix</CODE>.
     * @return An input stream associated with the file.
     * @exception IOException If the file can not be found or opened.
     */
    public InputStream getAppFileAsStream(String appFileName)
            throws IOException, HttpPresentationException {
        return loader.getAppFileAsStream(appFileName);
    }

    /**
     * Determine if a request URL references a presentation object.
     *
     * @param request The request for the presentation.
     * @return true if the URL has a presentation object MIME type.
     */
    public boolean isPresentationRequest(HttpPresentationRequest request)
            throws HttpPresentationException {
        return loader.isPresentationRequest(request);
    }


    /**
     *  Looks up the session object (if any) that would be used to 
     *  process the request. This is will not used normally, because the
     *  session is give to the application's preprocessor method and the
     *  presentation objects. Also, it is not normal to have a raw
     *  ServletRequest. The debugger uses this to examine the session data
     *  before and after each request. Consider this an internal use onlyu
     *  method.
     *
     *  @param request
     *    The (raw) request that would be sent in to this application.
     *  @return
     *    The session object that would be associated with the request.
     *    Returns null if none is found; a new session is not created.
     */
    public Session getSession(ServletRequest request) {
        if (application == null)
            return null;
        Session s = null;
        try {
            s = StandardAppUtil.getRequestSession(request, application);
        } catch (ApplicationException e) {
        }
        return s;
    }



    /*
     *  The next 3 functions were added to that applications can
     *  get their neghbor servlets. In the beginning we were prepared
     *  to leave servlets behind and move on to a new api if they
     *  didn't catch on. They caught on. So in the future we could
     *  simplify the archetecture some by not hiding the servlets
     *  from the applications with an abstraction layer.
     */


    /**
     *  Notify the presentation manager which servlet we are running in.
     *  This will be done immediatly after this object is created, if
     *  this ojbject is created by an HttpPresentationServlet (currently
     *  the only creator). In the future, if this class is used in non-servlet
     *  situations, this need not be called.
     *
     *  @param servlet
     *    The servlet we are running in.
     *  @param servletContext 
     *    The ServletContext used to initialize the servlet we are running in.
     */
    public void setServletAndContext(Servlet servlet, 
                                     ServletContext servletContext) {
        this.servlet = servlet;
        this.servletContext = servletContext;
    }


    /**
     *  Get the servlet we are running in. Currently this class is only
     *  used in a servlet (HttpPresentationServlet), so this value will
     *  always be set. But in the future this class may be used in other
     *  non-servlet enviornments, in which case this method will return
     *  null.
     *
     *  @return     
     *    The servlet we are running in, if any.
     */
    public Servlet getServlet() {
        return this.servlet;
    }


    /**
     *  Get the servlet context used to initialize the servlet we are running
     *  in. Currently this class is only used in a servlet
     *  (HttpPresentationServlet), so this value will always be set. But in
     *  the future this class may be used in other non-servlet enviornments,
     *  in which case this method will return null.
     *
     *  @return     
     *    The servlet context used to initialize the servlet we are
     *    running in, if any.
     */
    public ServletContext getServletContext() {
        return this.servletContext;
    }

    /**
     * Flush the presentation object and resource caches.
     */
    public void flushCache() {
        loader.flushCache();
    }

    /** Add a new mime type to extension mapping. */
    public void addMimeType(String mimeType, String extension) {
        loader.addMimeType(mimeType, extension);
    }

    /**
     * Get the application class loader.
     *
     * @return The application class loader.
     */
    public ClassLoader getAppClassLoader() {
        return loader.getAppClassLoader();
    }
}
