/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HttpPresentationRequest.java,v 1.24.14.1 2000/10/19 17:59:06 jasona Exp $
 */





package com.lutris.appserver.server.httpPresentation;

import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;
import java.util.Enumeration;


/**
 * Object passed to presentation objects that is used to access HTTP request
 * data.
 */
public interface HttpPresentationRequest {
    /**
     * Returns the original HttpServletRequest.
     */
    public HttpServletRequest getHttpServletRequest();

    /**
     * Returns the size of the request entity data, or -1 if not known. Same
     * as the CGI variable CONTENT_LENGTH.
     */
    public int getContentLength()
	throws HttpPresentationException;

    /**
     * Returns the Internet Media Type of the request entity data, or null if
     * not known. Same as the CGI variable CONTENT_TYPE.
     */
    public String getContentType()
	throws HttpPresentationException;

    /**
     * Returns the protocol and version of the request as a string of
     * the form <code>&lt;protocol&gt;/&lt;major version&gt;.&lt;minor
     * version&gt</code>.  Same as the CGI variable SERVER_PROTOCOL.
     */
    public String getProtocol()	
	throws HttpPresentationException;

    /**
     * Returns the scheme of the URL used in this request, for example
     * "http", "https", or "ftp".  Different schemes have different
     * rules for constructing URLs, as noted in RFC 1738.  The URL used
     * to create a request may be reconstructed using this scheme, the
     * server name and port, and additional information such as URIs.
     */
    public String getScheme();

    /**
     * Returns the host name of the server that received the request.
     * Same as the CGI variable SERVER_NAME.
     */
    public String getServerName();

    /**
     * Returns the port number on which this request was received.
     * Same as the CGI variable SERVER_PORT.
     */
    public int getServerPort();

    /**
     * Returns the IP address of the agent that sent the request. Same as the
     * CGI variable REMOTE_ADDR.
     */
    public String getRemoteAddr()
	throws HttpPresentationException;

    /**
     * Returns the fully qualified host name of the agent that sent the
     * request. Same as the CGI variable REMOTE_HOST.
     */
    public String getRemoteHost()
	throws HttpPresentationException;

    /**
     * Returns an input stream for reading the request body.
     */
    public HttpPresentationInputStream getInputStream()
	throws HttpPresentationException;

    /**
     * Returns a string containing the lone value of the specified query
     * parameter, or null if the parameter does not exist. Presentation
     * writers should use this method only when they are sure that there is
     * only one value for the parameter.  If the parameter has (or could have)
     * multiple values, then use getParameterValues. If a multiple valued
     * parameter name is passed as an argument, the return value is
     * implementation dependent.
     *
     * @param name the name of the parameter whose value is required.
     * @see HttpPresentationRequest#getParameterValues
     */
    public String getParameter(String name)
	throws HttpPresentationException;

    /**
     * Returns the values of the specified query parameter for the request as
     * an array of strings, or a 0 length array if the named parameter does
     * not exist.
     *
     * @param name the name of the parameter whose value is required.
     */
    public String[] getParameterValues(String name)
	throws HttpPresentationException;

    /**
     * Returns the parameter names for this request as an enumeration
     * of strings, or an empty enumeration if there are no parameters.
     */
    public Enumeration getParameterNames()
	throws HttpPresentationException;

    //Omitted: public Object getAttribute(String name);

    /**
     * Returns the method with which the request was made. The returned
     * value can be "GET", "HEAD", "POST", or an extension method. Same
     * as the CGI variable REQUEST_METHOD.
     */
    public String getMethod()
	throws HttpPresentationException;

    /**
     * Returns the request URI as a URL.
     */
    public String getRequestURI()
	throws HttpPresentationException;

    /**
     * Returns the presentation URI.
     */
    public String getPresentationURI()
        throws HttpPresentationException;

    /**
     * Returns the part of the request URI that refers to the application
     * object being invoked. Analogous to the CGI variable SCRIPT_NAME.
     *
     * @deprecated This method was named in a confusing manner; it
     *  returns the application, not presentation object path. Use
     *  <A HREF="getApplicationPath.html">getApplicationPath()</A>.
     */
    public String getPresentationPath()
	throws HttpPresentationException;

    /**
     * Returns the part of the request URI that refers to the presentation
     * object being invoked.
     */
    public String getPresentationObjectPath()
        throws HttpPresentationException;

    /**
     * Returns the part of the request URI after the presentation
     * manager servlet, upto and including the presentation object .po,
     * but not any path info.
     */
    public String getPresentationObjectRelativePath()
        throws HttpPresentationException;

    /**
     * Returns the part of the request URI that refers to the application.
     * Analogous to the CGI variable SCRIPT_NAME.
     */
    public String getApplicationPath()
	throws HttpPresentationException;

    /**
     * Returns optional extra path information following the presentation
     * path, but immediately preceding the query string. Returns null if
     * not specified. Same as the CGI variable PATH_INFO.
     */
    public String getPathInfo()
	throws HttpPresentationException;

    /**
     * Returns extra path information translated to a real path. Returns
     * null if no extra path information specified. Same as the CGI variable
     * PATH_TRANSLATED.  
     */
    public String getPathTranslated()
	throws HttpPresentationException;

    /**
     * Returns the query string part of the presentation URI, or null if none.
     * Same as the CGI variable QUERY_STRING.
     */
    public String getQueryString()
	throws HttpPresentationException;

    /**
     * Returns the name of the user making this request, or null if not
     * known.  The user name is set with HTTP authentication.  Whether
     * the user name will continue to be sent with each subsequent
     * communication is browser-dependent.  Same as the CGI variable
     * REMOTE_USER.
     */
    public String getRemoteUser()
	throws HttpPresentationException;

    /**
     * Returns the authentication scheme of the request, or null if none.
     * Same as the CGI variable AUTH_TYPE.
     */
    public String getAuthType()
	throws HttpPresentationException;

    /**
     * Gets the array of cookies found in this request.
     *
     * @return The array of cookies found in this request.
     */
    public Cookie[] getCookies()
	throws HttpPresentationException;

    /**
     * Indicates whether client submitted their session id through a cookie
     * @return true if client submitted their sessionId via a cookie, 
     *         false otherwise
     */
    public boolean isRequestedSessionIdFromCookie()
	throws HttpPresentationException; 

    /*
     * set the flag indicating whether the sessionId came from a cookie
     * @param isFromCookie boolean indicating whether sessionId came 
     *        from cookie
     */
    public void setRequestedSessionIdFromCookie(boolean isFromCookie) 
	throws HttpPresentationException; 

    /**
     * Indicates whether client submitted their sessionId through a 
     * rewritten url
     * @return true if client submitted their sessionId via a rewritten url, 
     *         false otherwise
     */
    public boolean isRequestedSessionIdFromUrl()
	throws HttpPresentationException; 

   
    /*
     * set the flag indicating whether the sessionId came from a url
     * @param isFromUrl boolean indicating whether sessionId came from url
     */
    public void setRequestedSessionIdFromUrl(boolean isFromUrl) 
	throws HttpPresentationException; 

    /**
     * Returns the value of a header field, or null if not known.
     * The case of the header field name is ignored.
     * @param name the case-insensitive header field name
     */
    public String getHeader(String name)
	throws HttpPresentationException; 

    /**
     * Returns the value of an integer header field, or -1 if not found.
     * The case of the header field name is ignored.
     * @param name the case-insensitive header field name
     */
    public int getIntHeader(String name)
	throws HttpPresentationException;

    /**
     * Returns the value of a date header field, or -1 if not found.
     * The case of the header field name is ignored.
     * @param name the case-insensitive header field name
     */
    public long getDateHeader(String name)
	throws HttpPresentationException;

    /**
     * Returns an enumeration of strings representing the header names
     * for this request. Some server implementations do not allow headers
     * to be accessed in this way, in which case this method will return null.
     */
    public Enumeration getHeaderNames()
	throws HttpPresentationException;

    /**
     * Get the URI path for a file in the application.  This converts a path
     * to the file part of the URL.  It adds in the reference to application
     * servlet.
     *
     * @param file File with in the application.  Currently this must
     *  be a path relative to the presentation prefix.
     * @return The file path portion of the URL, starting with
     * a <CODE>/</CODE>.
     */
    public String getAppFileURIPath(String file)
	throws HttpPresentationException;
}
