/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: AbstractFilterableServlet.java,v 1.8.14.1 2000/10/19 17:59:06 jasona Exp $
 */





package com.lutris.appserver.server.httpPresentation.servlet;

import java.util.Vector;
import java.io.IOException;
import javax.servlet.ServletException;
import javax.servlet.Servlet;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import org.enhydra.servlet.filter.Filter;


/**
 *  This class demonstrates the responsibilities of a
 *  <CODE>FilterableServlet</CODE>. Because most servlets will need to 
 *  extend the base class <CODE>HttpServlet</CODE>, they will not be able
 *  to extend this class. Therefore this class is not expected to be used,
 *  it is merely an instructional reference. That is why it does not implement
 *  the full <CODE>Servlet</CODE> interface, just the extensions in 
 *  <CODE>FilterableServlet</CODE> <CODE>service()</CODE> method.
 *
 * @see com.lutris.appserver.server.httpPresentation.servlet.FilterableServlet
 */
public abstract class AbstractFilterableServlet
                      extends HttpServlet
                      implements FilterableServlet {

    /**
     *  The list of current filters.
     */
    private Vector filters = new Vector();


    public AbstractFilterableServlet() {
    }


    /**
     *  Add a filter to the end of the list of filters.
     *
     *  @param filter
     *    The filter to add.
     */
    public void addFilter(Filter filter) {
        synchronized (filters) {
            filters.addElement(filter);
        }
    }

    /**
     *  Remove a filter from the list of current filters. 
     *  Silently fails if the filter is not in the list.
     *
     *  @param filter
     *    The filter to remove.
     */
    public void removeFilter(Filter filter) {
        synchronized (filters) {
            filters.removeElement(filter);
        }
    }


    /**
     *  Handle a service request. Use the first filter to wrap the request,
     *  response, and servlet (this) objects. This will return new objects.
     *  Then use the second filter to wrap those, again returning new
     *  objects. Then use the fourth filter to wrap those etc....
     *  Then finally call <CODE>service()</CODE> on the outermost results.
     *
     *  @param request
     *    The original servlet request object.
     *  @param response
     *    The original servlet response object.
     */
    public void service(HttpServletRequest request,
                        HttpServletResponse response) 
                throws ServletException, IOException {
        /*
         * This is synchronized for safety, but in real life you wouldn't want
         * to do this as it would introduce too much slowdown in such a
         * critical bottlneck area of the code. You would be prepared for the
         * array to change out from under you. Adding/removing filters is
         * a very rare operation, so it's ok.
         */
        synchronized (filters) {
            int size = filters.size();
            if (size == 0) {
                serviceDirect(request, response);
            } else {
		HttpPresentationServletGlue glue =
			new HttpPresentationServletGlue(this);
		HttpServletRequest reqResult = request;
		HttpServletResponse resResult = response;
                for (int i=0; i<size; i++) {
                    Filter f = (Filter) filters.elementAt(i);
                    reqResult = f.wrapRequest(reqResult, resResult);
                    resResult = f.wrapResponse(reqResult, resResult);
                }
                glue.service(reqResult, resResult);
            } 
        }
    }



    /**
     *  Once the filters are done, you don't want to call service(), or it
     *  would result in an infinite loop of filter processing. So you call this
     *  instead. It does the real work that service() would normally do, 
     *  without applying the filters. <P>
     *
     *  This is left an abstract method because this is where you would do
     *  your real work if you subclassed this. However this class is 
     *  most likely never going to be extended; it is meant to be used as
     *  a code reference to see how filters work.
     *
     *  @param request
     *    The request object.
     *  @param response
     *    The response object.
     */
    public abstract void serviceDirect(HttpServletRequest request,
                                       HttpServletResponse response)
                         throws ServletException, IOException;


}

