/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HttpPresentationServlet.java,v 1.63.2.3.4.2 2000/10/19 17:59:06 jasona Exp $
 */





package com.lutris.appserver.server.httpPresentation.servlet;

import java.util.Vector;
import java.io.*;
import javax.servlet.*;
import javax.servlet.http.*;
import com.lutris.appserver.server.httpPresentation.*;
import com.lutris.appserver.server.session.Session;
import com.lutris.appserver.server.*;
import com.lutris.classloader.MultiClassLoader;
import com.lutris.Lutris;
import com.lutris.util.*;
import com.lutris.logging.*;
import org.enhydra.servlet.servletManager.*;
import org.enhydra.servlet.ServletContainer;
import org.enhydra.servlet.filter.Filter;
import com.lutris.multiServer.MultiServer;


/**
 * Presentation server implemented as a servlet. Translates servlet
 * requests into HttpPresentationManager requests.  There is a one
 * to one correspondence between an instance of this servlet and
 * an Enhydra application.  An instance of the application is created by
 * the servlet.
 *
 * The servlet requires a single init parameter `configFile', that
 * contains the path to the application's configuration file.  This
 * file must define the following fields:
 * <UL>
 * <LI> <CODE>server.classPath</CODE> - The class path for the application,
 *      as a comma seperated list.  This should not contain the class path
 *      for the Java runtime classes or the Lutris classes.
 * <LI> <CODE>server.appClass</CODE> - The full class name of the application
 *      class.  This must implement
 *      <CODE>com.lutris.appserver.server.Application</CODE>.
 * <LI> <CODE>server.presentationPrefix</CODE> - This is the path that is used as
 *      a prefixed for all URL references withing the application.  It must use
 *      a `/' seperator (as with URLs).
 * <LI> <CODE>server.autoReload</CODE> - A flag to indicate whether or not
 *      the application should be automatically reloaded when any class or
 *      jar file changes in the application's classpath.  <B><I>NOTE: THIS IS
 *      A DEBUGGING OPTION AND MAY SLOW DOWN THE APPLICATION!</I></B>
 * <LI>

 * </UL> <P>
 *
 * This servlet also implements the <CODE>ListenableServlet</CODE>
 * interface. This allows other objects to register with this servlet and
 * be notified every time <CODE>service</CODE> is called. These remote
 * listeners may ignore the notification, examine the request, or even
 * replace the request and/or response with extended versions. This is
 * used by the debugging and monitoring servlets.
 *
 *
 * @version	$Revision: 1.63.2.3.4.2 $
 * @author	Mark Diekhans
 * @since	1.3
 */
//FIXME:  Explain the presentationPrefix better; give example.
//FIXME: Integrate URLEncodeServlet
public class HttpPresentationServlet
             extends HttpServlet
             implements FilterableServlet {

    /**
     * Log level for DOM statistics.
     */
    public static final String XMLC_DOM_STATS_LOG_LEVEL = "XMLC_DOM_STATS";

    protected ServletContext context;

    /*
     * Application configuration definitions.
     */
    private static final String APP_CONFIG_INIT_PARAM_NAME =
	    ServletContainer.CONF_FILE;
    private static final String APP_CLASS_PATH_FIELD =
	    ServletContainer.SERVER + "." + ServletContainer.CLASS_PATH;
    private static final String APP_APP_CLASS_FIELD =
	    ServletContainer.SERVER + "." + ServletContainer.APP_CLASS;
    private static final String APP_PRESENTATION_PREFIX_FIELD =
	    ServletContainer.SERVER + "." + ServletContainer.PRESENTATION_PREFIX;
    private static final String APP_AUTO_RELOAD_FIELD =
	    ServletContainer.SERVER + "." + ServletContainer.AUTO_RELOAD;
    private static final String APP_LOGFILE =
            ServletContainer.SERVER + "." + MultiServer.LOG_FILE;
    private static final String APP_LOG_TO_FILE =
            ServletContainer.SERVER + "." + MultiServer.LOG_TO_FILE;
    private static final String APP_LOG_TO_STDERR =
            ServletContainer.SERVER + "." + MultiServer.LOG_TO_STDERR;

    /*
     *  Where to log information to.
     *  If we are running in the multiserver, the central logger will already
     *  be created and initialzied. In any other server, we will need to
     *  instantiate the central logger. The first application (servlet)
     *  to be instantiated in a given class loader will configure logging;
     *  any additional applications (servlets) created in the same class
     *  loader will use the existing logger (ignoring the settings in their
     *  config file).
     */
    private LogChannel logChannel;

    /*
     *  If we create the central logger, we save a copy of it for later
     *  initialization. If we didn't create it, this will be null.
     */
    private StandardLogger standardLogger;

    /*
     *  If we create the central logger, we will need to initialize it later
     *  when init() is called.
     */
    private boolean needToConfigureLogChannel = false;

    /*
     *  Application and its configuration information.
     */
    Application application = null;
    Config appConfig = null;
    String presentationPrefix;
    boolean cacheClasses;
    boolean cacheFiles;
    boolean autoReload;  // Reload application when its classes are modified?

    HttpPresentationManager presentationManager = null;

    /**
     * If not null, log DOM statistics here after each DOM write.
     */
    private PrintWriter domStatsLogWriter;

    /**
     * Create a new HttpPresentationServlet.
     */
    public HttpPresentationServlet() {
        super();

        /**
         *  If the central logger exists, use it. That means we are
         *  probably running in the MultiServer. If it does not yet exist,
         *  then create it (configure it later).
         */
        if (Logger.getCentralLogger() == null) {
            standardLogger = new StandardLogger(true);
            needToConfigureLogChannel = true;
        }
        logChannel = Logger.getCentralLogger().getChannel("Multiserver");
    }

    /**
     * Throw a servlet exceptionm logging the message if possible.
     */
    private void throwServletException(String msg, Throwable cause)
        throws ServletException {
        if (logChannel != null) {
            logChannel.write(Logger.ERROR, msg, cause);
        } else {
            // Desprite
            System.err.println("Error: " + msg);
            cause.printStackTrace();
        }
        throw new ServletException(msg + "\ncaused by " + cause.getClass()
                                   + ": " + cause.getMessage());
    }

    /**
     * Generate an error for a field not being specified in a config file.
     */
    private void missingConfField(String field, String file)
            throws HttpPresentationException {
        throw new HttpPresentationException("field \"" + field +
            "\" not specified in application config file \"" + file + "\"");
    }

    /**
     * Open and read the application's configuration file and initialize the
     * application, but don't start it yet. This is done on in service so it
     * part of the standard state-change mechanism.
     */
    private void initApplication(ServletConfig servletConfig,
                                 MultiClassLoader appClassLoader)
            throws ServletException {

	try {
	    /*
	     * If this servlet is being run by the MultiServer, then it
	     * will be passed a special context object that can be used
	     * to get extra information about the context the servlet is
	     * running in. If it is a special context, use it, but we may
	     * not assume that it is, or depend on it.
	     */
	    context = servletConfig.getServletContext();
	    EnhydraServletContext lbsContext = null;
	    if (context instanceof
		org.enhydra.servlet.servletManager.EnhydraServletContext)
		lbsContext = (EnhydraServletContext)context;

	    // Get configuration file and parameters.
	    String configFileName = servletConfig.getInitParameter(APP_CONFIG_INIT_PARAM_NAME);
	    if (configFileName == null) {
		throw new HttpPresentationException("init parameter \"" +
						    APP_CONFIG_INIT_PARAM_NAME +
						    "\" not specified for Enhydra servlet");
	    }
	    File configFile = new File(configFileName);
	    ConfigFile configF = new ConfigFile(configFile);
	    appConfig = configF.getConfig();

	    /*
	     *  First thing is to configure logging (if we created the central
	     *  logger). That way we can log errors asap. This will only happen
	     *  if we are the first application to load and we are not in the
	     *  multiserver.
	     */
	    if (needToConfigureLogChannel) {
		configureLogChannel(configFileName);
            }

	    //FIXME: The null checks below are never used, as get  returns an error.

	    String[] appClassPath = appConfig.getStrings(APP_CLASS_PATH_FIELD);
	    if (appClassPath == null) {
		missingConfField(APP_CLASS_PATH_FIELD, configFileName);
	    }
	    String appClass = appConfig.getString(APP_APP_CLASS_FIELD);
	    if (appClass == null) {
		missingConfField(APP_APP_CLASS_FIELD, configFileName);
	    }
	    presentationPrefix = appConfig.getString(APP_PRESENTATION_PREFIX_FIELD);
	    if (presentationPrefix == null) {
		missingConfField(APP_PRESENTATION_PREFIX_FIELD, configFileName);
	    }
	    autoReload = appConfig.getBoolean(APP_AUTO_RELOAD_FIELD, false);

	    // Add application's class path to class loader.
	    appClassLoader.setClassPath(appClassPath);

	    // Config of presentation manager
	    cacheClasses = appConfig.getBoolean("PresentationManager.CacheClasses",
						true);
	    cacheFiles = appConfig.getBoolean("PresentationManager.CacheFiles",
					      false);

	    // FIXME: Left in to help with configuration debugging
	    // until logging is available.
	    logChannel.write(Logger.DEBUG, "Loaded Enhydra application: " + appClass);
	    logChannel.write(Logger.DEBUG,
			     "    presentation prefix: " + presentationPrefix);
	    String cp = "    Class path: ";
	    for (int i = 0; i < appClassPath.length; i++) {
		if (i > 0)
		    cp += ", ";
		cp += appClassPath[i];
	    }
	    logChannel.write(Logger.DEBUG, cp);

	    // Load actual application object.
	    Class appClassObj = appClassLoader.loadClass(appClass);
	    application = (Application)appClassObj.newInstance();

	    /*
	     *  Register the application object with this thread.
	     *  We do this so that all calls into the Application object
	     *  can count the Enhydra class to sucessfully return the
	     *  Application object.
	     */
	    Enhydra.register(application);
	    try {
		/*
		 * If we are being run by a ServletManager, then use the extra
		 * context information. Otherwise, use default values.
		 */
		if (lbsContext != null) {
		    // Being run by a ServletManager (MultiServer).
		    application.setLogChannel(lbsContext.getLogChannel());
		    application.setName(lbsContext.getName());
		} else {
		    // Not run by ServletManager (MultiServer). Name = class name.
		    String name = appClass;
		    int idx = name.lastIndexOf('.');
		    if (idx >= 0)
			name = name.substring(idx);
		    application.setLogChannel(
			  Logger.getCentralLogger().getChannel(name));
		    application.setName(name);
		}
	    } finally {
		// Be sure the thread is unregistered no matter what.
		Enhydra.unRegister();
	    }
	    setPresentationManager(context, appClassLoader);

            // Set XMLC Factory (needs classLoader set).
            boolean xmlcReload = appConfig.getBoolean("Server.XMLC.AutoReload",
                                                      false);
            boolean xmlcRecomp = appConfig.getBoolean("Server.XMLC.AutoRecompilation",
                                                      false);
            application.setXMLCFactory(xmlcReload, xmlcRecomp);
            
            if (application.getLogChannel().isEnabled(XMLC_DOM_STATS_LOG_LEVEL)) {
                domStatsLogWriter = application.getLogChannel().getLogWriter(XMLC_DOM_STATS_LOG_LEVEL);
            }

            // Tell the presentation manager about addititonal
            // extension/mime-type mappings.
            Config mimeTypes = appConfig.getConfig("Server.MimeType");
            if (mimeTypes != null) {
                String extensions[] = mimeTypes.keys();
                for (int idx = 0; idx < extensions.length; idx++) {
                    presentationManager.addMimeType(mimeTypes.getString(extensions[idx]),
                                                    extensions[idx]);
                }
            }
        } catch(Exception except) {
            throwServletException("Initialization of application failed",
                                  except);
        }
    }

    /*
     * Standard servlet init function, creates the presentation manager
     * we are going to use.
     *
     * @param config Object containing the servlet's startup
     *		     configuration and initialization parameters.
     * @exception ServletException if a servlet exception has occurred.
     */
    public synchronized void init(ServletConfig config) throws ServletException {
        super.init(config);  // You must do this when overriding this method.

        /*
         * If this servlet is being run by the MultiServer, then it
         * will be passed a special context object that can be used
         * to get extra information about the context the servlet is
         * running in. If it is a special context, use it, but we may
         * not assume that it is, or depend on it.
         */
        ServletContext context = config.getServletContext();
        EnhydraServletContext lbsContext = null;
        if (context instanceof
                  org.enhydra.servlet.servletManager.EnhydraServletContext) {
            lbsContext = (EnhydraServletContext) context;
            // Use the application's log channel, rather than our generic one.
            logChannel = lbsContext.getLogChannel();
        }

        /*
         * If we can get the class loader used to load us, then do so.
         * Otherwise, create a new loader. This loader will be used
         * to load the presentation objects.
         */
        MultiClassLoader appClassLoader;
        if (lbsContext != null)
            appClassLoader = lbsContext.getClassLoader();
        else
	    // Not using MultiServer, so must create class loader from scratch
            appClassLoader = new MultiClassLoader(logChannel);

	initApplication(config, appClassLoader);

        // This call is added for the 3.0 release.
        // in the past ensure app is running was called by the servletmanager
        // after a cast of the servlet variable.
        ensureAppIsRunning();
    }

    /**
     * Sets the presentation manager for the application.  First creates a
     * new presentation manager with the presentationPrefix, application,
     * cacheClasses and cacheFiles fields (so these parameters must already
     * be initialized).  Then, the presentation manager is set in the
     * application.
     *
     * @param context The servlet context for this application.
     * @param appClassLoader The class loader for this application.
     * @exception ServletException if a servlet exception has occurred.
     */
    private void setPresentationManager(ServletContext context,
                                        ClassLoader appClassLoader)
	    throws ServletException {
        try {
            presentationManager = new HttpPresentationManager(presentationPrefix,
								application,
								appClassLoader,
								cacheClasses,
								cacheFiles);
            presentationManager.setServletAndContext(this, context);
        } catch(Exception except) {
            throwServletException("Initialization of presentation manager failed",
                                  except);
        }

        /*
	 * Tell application about the presentation manager running it.
         *
         *  Register the application object with this thread.
         *  We do this so that all calls into the Application object
         *  can count the Enhydra class to sucessfully return the
         *  Application object.
         */
        Enhydra.register(application);
        try {
            application.setHttpPresentationManager(presentationManager);
        } finally {
            // Be sure the thread is unregistered no matter what.
            Enhydra.unRegister();
        }
    }

    /**
     * Attempt to change the application to the running state.
     */
    private synchronized void changeToRunningState()
            throws ApplicationException {

        // We may or may not be registered here..
        boolean isRegistered = (Enhydra.getApplication() != null);
        if (!isRegistered) {
            Enhydra.register(application);
        }
        try {
            switch (application.getState()) {
              case Application.STOPPED:
                application.startup(appConfig);
                break;
              case Application.RUNNING:
                break;
              case Application.INCOMPLETE:
                application.restartup(appConfig);
                break;
              case Application.DEAD:
                throw new ApplicationException("The application " +
                                               application.getName() +
                                               " is dead");
              case Application.HALTED:
                throw new ApplicationException("The application " +
                                               application.getName() +
                                               " is halted");

              default:
                throw new ApplicationException("The application " +
                                               application.getName() +
                                               " in invalid state");
            }
        } finally {
            if (!isRegistered) {
                Enhydra.unRegister();
            }
        }
    }



    private void configureLogChannel(String configFile)
                 throws HttpPresentationException {

        String defaultLogFile = "/tmp/enhydra.log";
        String[] defaultToFile = {"EMERGENCY", "ALERT", "CRITICAL", "ERROR",
                                  "WARNING", "NOTICE", "INFO"};
        String[] defaultToStderr = {"EMERGENCY", "ALERT", "CRITICAL", "ERROR",
                                    "WARNING", "NOTICE", "INFO"};

        String logFile = null;
        try {
            logFile = appConfig.getString(APP_LOGFILE, null);
        } catch (ConfigException e) {
            System.out.println("Warning: error parsing key " +
                APP_LOGFILE + " from log file " +
                configFile + ". Using default log options.");
            logFile = defaultLogFile;
        }
        if (logFile == null) {
            System.out.println("Warning: no key " + APP_LOGFILE +
                " was found in config file " + configFile +
                 ". Using default log options.");
            logFile = defaultLogFile;
        }

        String[] toFile = null;
        try {
            toFile = appConfig.getStrings(APP_LOG_TO_FILE, null);
        } catch (ConfigException e) {
            System.out.println("Warning: Error parsing key " +
                APP_LOG_TO_FILE + " from log file " + configFile +
                ". Using default log options.");
            toFile = defaultToFile;
        }
        if (toFile == null) {
            System.out.println("Warning: no key " +
                APP_LOG_TO_FILE + " found in log file " + configFile +
                ". Using default log options.");
            toFile = defaultToFile;
        }

        String[] toStderr = null;
        try {
            toStderr = appConfig.getStrings(APP_LOG_TO_STDERR, null);
        } catch (ConfigException e) {
            System.out.println("Warning: error parsing key " +
                    APP_LOG_TO_STDERR +
                    " from log file " + configFile +
                    ". Using default log options.");
            toStderr = defaultToStderr;
        }
        if (toStderr == null) {
            System.out.println("Warning: no key " + APP_LOG_TO_STDERR +
                    " from log file " + configFile +
                    ". Using default log options.");
            toStderr = defaultToStderr;
        }

        File theLogFile = new File(logFile);
        try {
            if (standardLogger != null)
                standardLogger.configure(theLogFile, toFile, toStderr);
        } catch (IOException ioe) {
            System.out.println(
                   "Warning: Unable to begin logging to the file " + logFile +
                   ":\n" + ioe.getMessage());
            return;
        }

        /*
         * Done bootstrapping. From here on out we may use logging.
         */
        logChannel.write(Logger.INFO, "Logging initialized");
    }


    /**
     * This is an HTTP-specific version of the Servlet.service method
     * that translates that request into a request to a presentation
     * manager.  This object is method is not synchronized, the thread
     * is used to handle to entire request. <P>
     *
     * As part of implementing the FiterableServlet interface, the real work
     * is done here in serviceDirect(), while calls to service() result in
     * filters being applied and called. The end of the filter chain is a
     * glue servlet that directly calls this method.
     *
     * @param req encapsulates the request to the servlet
     * @param resp encapsulates the response from the servlet
     * @exception ServletException if the request could not be handled
     * @exception IOException if detected when handling the request
     */
    public void serviceDirect (HttpServletRequest req, HttpServletResponse resp)
            throws ServletException, IOException {

        /*
         * Register the application object with this thread.
         * We do this so that all the calls into the Application object
         * can count on the Enhydra class to return the Application object.
         * This is important not so much for the Application object, but
         * for any utility classes it may call.
         */
        Enhydra.register(application);
        try {
            ensureAppIsRunning();
            presentationManager.Run(new ServletHttpPresentationRequest(req),
                                    new ServletHttpPresentationResponse(resp, domStatsLogWriter));
        } catch(HttpPresentationException except) {
            throwServletException("An unhandled error occured while servicing an application request",
                                  except);
        } finally {
            // Be sure the thread is unregistered no matter what.
            Enhydra.unRegister();
        }
    }

    /**
     * This method is public for use by the ServletManager
     * to run the application startup() method after init().
     *
     * @exception ServletException
     *	If any error occurs on startup.
     */
    public void ensureAppIsRunning() throws ServletException {
	checkAutoReload();
        // If an application is associated,
        // then make sure it's in the run state.
        if ((application != null) &&
            (application.getState() != Application.RUNNING)) {
            try {
                changeToRunningState();
                context.setAttribute(ServletManager.SESSION_MANAGER_KEY,application.getSessionManager());
            } catch (Throwable except) {
                throwServletException("Unable to change application to running state",
                                      except);
            }
        }
    }

    /**
     * Return information specific to this servlet.
     */
    public String getServletInfo() {
        return Lutris.getEnhydraLongName() + "/" + Lutris.getEnhydraVersion();
    }

    /**
     * This method instantiates the HttpPresentationRequest
     * object that the presentation manager expects when
     * servicing a request.
     *
     * @param req
     *   The servlet request object.
     * @param resp
     *   The servlet response object.
     * @return The HttpPresentationRequest object used
     *   by the presentation manager.
     */
    protected HttpPresentationRequest getHttpPresentationRequest(
            HttpServletRequest req,
            HttpServletResponse resp) {
        return new ServletHttpPresentationRequest(req);
    }

    /**
     * This method instantiates the HttpPresentationResponse
     * object that the presentation manager expects when
     * servicing a request.
     *
     * @param req
     *   The servlet request object.
     * @param resp
     *   The servlet response object.
     * @return The HttpPresentationResponse object used
     *   by the presentation manager.
     */
    protected HttpPresentationResponse getHttpPresentationResponse(
        HttpServletRequest req,
        HttpServletResponse resp) {
        return new ServletHttpPresentationResponse(resp, domStatsLogWriter);
    }

    /**
     * Returns the application being run by this instance of
     * the servlet.
     */
    public Application getApplication() {
        return application;
    }

    /**
     * Returns the application configuration init parameter.
     */
    protected Config getAppConfigInitParam() {
        return appConfig;
    }

    /**
     * Returns the application configuration init parameter name.
     */
    protected String getAppConfigInitParamName() {
        return APP_CONFIG_INIT_PARAM_NAME;
    }


    // ----------------- FilterableServlet methods --------------


    private Vector filters = new Vector();


    /**
     *  Add a filter to the end of the list of filters.
     *
     *  @param filter
     *    The filter to add.
     */
    public void addFilter(Filter filter) {
        synchronized (filters) {
            filters.addElement(filter);
        }
    }

    /**
     *  Remove a filter from the list of current filters.
     *  Silently fails if the filter is not in the list.
     *
     *  @param filter
     *    The filter to remove.
     */
    public void removeFilter(Filter filter) {
        synchronized (filters) {
            filters.removeElement(filter);
        }
    }


    /**
     *  Handle a service request. Use the first filter to wrap the request,
     *  response, and servlet (this) objects. This will return new objects.
     *  Then use the second filter to wrap those, again returning new
     *  objects. Then use the fourth filter to wrap those etc....
     *  Then finally call <CODE>service()</CODE> on the outermost results.
     *
     *  @param request
     *    The original servlet request object.
     *  @param response
     *    The original servlet response object.
     */
    public void service(HttpServletRequest request,
                        HttpServletResponse response)
                throws ServletException, IOException {
        /*
         *  This hook was added to support the debugger. Almost all
         *  applications will not use this.
         */
        if (presentationManager.servletRequestPreprocessor(this,
                                   context, request, response)) {
            // The request has been completly processed. Nothing left to do.
            return;
	}

        /*
         *  Route the request/response through the list of filters (if any)
         *  that have been applied to this application. Normally there will
         *  not be any. The debugger, for example, installs a filter to
         *  implement the request logging and tracing.
         */
	HttpPresentationServletGlue glue =
		new HttpPresentationServletGlue(this);
        /*
         *  Technically this should be synchronized on the filters vector,
         *  so it doesn't change out from under us while we iterate through
         *  the vector. But that would kill performance. This is a major
         *  bottleneck area (all requests come through here), so we make the
         *  tradeoff that a filter might be skipped if a request comes through
         *  at the same time someone adds or removes a filter (adding and
         *  removing filters is a very rare occurance).
         *
         *  The filters are here to support debugging, and possibly monitoring
         *  and other applications in the future. Normally the filters
         *  vector will be empty.
         */
        int size = filters.size();
	HttpServletRequest reqResult = request;
	HttpServletResponse resResult = response;
        if (size == 0) {
            serviceDirect(request, response);
        } else {
            for (int i=0; i<size; i++) {
                Filter f = null;
                try {
                    f = (Filter) filters.elementAt(i);
                } catch (Throwable t) {
                    // Array might have changed out from under us.
                }
                if (f != null) {
                    reqResult = f.wrapRequest(reqResult, resResult);
                    resResult = f.wrapResponse(reqResult, resResult);
		}
            }
            glue.service(reqResult, resResult);
        }
    }


    /**
     *  Looks up the session object (if any) that would be used to
     *  process the request. This is will not used normally, because the
     *  session is give to the application's preprocessor method and the
     *  presentation objects. Also, it is not normal to have a raw
     *  ServletRequest. The debugger uses this to examine the session data
     *  before and after each request. Consider this an internal use only
     *  method.
     *
     *  @param request
     *    The (raw) request that would be sent in to this application.
     *  @return
     *    The session object that would be associated with the request.
     *    Returns null if none is found; a new session is not created.
     */
    public Session getSession(ServletRequest request) {
        if (application == null)
            return null;
        Enhydra.register(application);
        Session s = null;
        try {
            s = presentationManager.getSession(request);
        } finally {
            Enhydra.unRegister();
        }
        return s;
    }

    /**
     * Check if the application should be automatically reloaded and, if so,
     * call <CODE>initApplication</CODE> with a new instance of the class
     * loader.  The application should be reloaded if the
     * <CODE>autoReload</CODE> flag is <CODE>true</CODE> and any class has
     * changed in the application's classpath.
     *
     * @exception ServletException if a servlet exception has occurred.
     */
    private void checkAutoReload() throws ServletException {
	if (autoReload) {
	    MultiClassLoader appClassLoader =
		(MultiClassLoader) application.getClass().getClassLoader();
	    if (appClassLoader.shouldReload()) {
		application.shutdown();
		initApplication(getServletConfig(),
		    new MultiClassLoader(logChannel));
	    }
	}
    }

    /**
     * Destroys the servlet.  Called by the servlet manager when the servlet
     * is shutdown.
     */
    public synchronized void destroy() {
        application.shutdown();
    }
}



// FIX: It appears that throwing an error in a servlet doesn't get
//      displayed in the client, we might want to deal with this
//      explictly.
