/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: PagedSessionHandle.java,v 1.2.14.1 2000/10/19 17:59:06 jasona Exp $
 */

package com.lutris.appserver.server.session;
import com.lutris.appserver.server.user.User;
import com.lutris.util.*;
import com.lutris.logging.*;
import java.util.*;

/**
 * A lightweight handle to a session which has been paged to disk.
 * Referenced to the session manger and user associated with the
 * session are not paged.  Instead they are heald by this object
 * until the session and corresponding session data are read
 * back from disk.
 *
 * @see         PagedSessionHome
 * @version	$Revision: 1.2.14.1 $
 * @author	Kyle Clark
 */
class PagedSessionHandle {

    /**
     * Keep track if the session object that is paged to disk
     * is new.
     */
    private boolean sessionIsNew;

    /**
     * Keep track of time session was created.
     */
    private long sessionTimeCreate;

    /**
     * The store interface.
     */
    private PersistentStore store;

    /**
     * The session that is being paged.
     */
    private PagedSession session;

    /**
     * The session key.
     */
    private String sessionKey;

    /**
     * Reference to transient data that should be
     * held by the handle.
     */
    private Object[] transientData;
     
    /**
     * Creates an instance of PagedSessionHandle.
     *
     * @param config
     *   configuration settings required by this object.
     * @param session
     *   the session represented by this object.
     */
    PagedSessionHandle(PagedSession session, PersistentStore store) {
        this.session = session;
        this.store = store;

        // Convenience - performance
        this.sessionIsNew = session.isNew();
        this.sessionTimeCreate = session.getTimeCreated();
        this.sessionKey = session.getSessionKey();

    }

    /**
     * Pages the session to the file system.
     *
     * @exception SessionException if an error occurs.
     */
    synchronized void write() throws SessionException {
        try {
            transientData = session.getTransientData();
            store.store(sessionKey, session);
            // delete reference to session
            // can be cleaned up via garbage collector.
            session = null; 
        } catch (Exception e) {
            throw new SessionException("Unable to write session to persistent store.", e);
        }
    }

    /**
     * Reads the session from the file system.  The session is
     * deleted from the file system.  May return null.
     *
     * @return the session.
     * @exception SessionException if an error occurs.
     */
    synchronized PagedSession read() throws SessionException {
        if (session != null) {
            return session;
        }
        try {
            session = (PagedSession)store.remove(sessionKey);
            session.restoreTransientData(transientData);
            return session;
        } catch (Exception e) {
            throw new SessionException("Unable to read session from persistent store.", e);
        }
    }

    /**
     * Deletes the paged session.
     *
     * @exception SessionException if an error occurs.
     */
    synchronized void delete() throws SessionException {
        try {
            store.delete(sessionKey);
        } catch (Exception e) {
            throw new SessionException("Unable to delete session from persistent store.", e);
        }
    }

    /**
     * Returns true if the session referenced by this handle is new.
     *
     * @return
     *   true if the session referenced by this handle is new.
     */
    boolean isNew() {
        return sessionIsNew;
    }

    /**
     * Returns the time the session referenced by this handle
     * was created.
     *
     * @return
     *   the time the session referenced by this handle
     *   was created.
     */
    long getTimeCreated() {
        return sessionTimeCreate;
    }

    /**
     * Returns the session key associated with this object.
     *
     * @return
     *   the session key.
     */
    String getSessionKey() {
        return sessionKey;
    }
     
}

