/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: DBUtil.java,v 1.4.14.1 2000/10/19 17:59:06 jasona Exp $
 */

package com.lutris.appserver.server.session.persistent;

import com.lutris.appserver.server.Enhydra;
import com.lutris.appserver.server.session.*;
import com.lutris.appserver.server.sql.*;
import com.lutris.logging.*;
import java.util.*;
import java.sql.*;

/**
 * Utilitity class for performint database operations.
 *
 * @see         StandardSession
 * @see         StandardSessionManager
 * @version	$Revision: 1.4.14.1 $
 * @author	Kyle Clark
 */
class DBUtil {

    /**
     * Saves a session to the database.
     *
     * @param session
     *   the session to save.
     * @param dbName the database that should be accessed.  May be
     *   null in which case the default database is used.
     * @return the number of rows that were affected
     *   by this update.
     * @exception SessionException
     *   if an error occurs.
     */
    static int dbUpdate(PersistentSession session, String dbName)
        throws SessionException {
        PersistentSessionDO sessionDO = new PersistentSessionDO(session);
        try {
	    DBTransaction t = createTransaction(dbName);
            try {
                t.update(sessionDO);
                t.commit();
            } catch (Exception e) {
                t.rollback();
		throw e;
            } finally {
                t.release();
            }
        } catch (Exception e) {
            throw new SessionException(e);
        }
        return sessionDO.getTransactionRowCount();
    }

    /**
     * Deletes a session from the database.
     *
     * @param sessionKey
     *   the key that identifies the session that will be deleted.
     * @param dbName the database that should be accessed.  May be
     *   null in which case the default database is used.
     * @exception SessionException
     *   if an error occurs.
     */
    static void dbDelete(String sessionKey, String dbName)
        throws SessionException {
        try {
	    DBTransaction t = createTransaction(dbName);
            SessionDeleteDO s = new SessionDeleteDO(sessionKey);
            try {
                t.delete(s);
                t.commit();
            } catch (Exception e) {
                t.rollback();
		throw e;
            } finally {
                t.release();
            }
        } catch (Exception e) {
            throw new SessionException(e);
        }
    }

    /**
     * Inserts a session into the database.
     *
     * @param session
     *   the session that will be inserted.
     * @param dbName the database that should be accessed.  May be
     *   null in which case the default database is used.
     * @exception SessionException
     *   if an error occurs.
     */
    static void dbInsert(PersistentSession session, String dbName)
        throws SessionException {
        try {
            PersistentSessionDO sessionDO = new PersistentSessionDO(session);
	    DBTransaction t = createTransaction(dbName);
            try {
                t.insert(sessionDO);
                t.commit();
            } catch (Exception e) {
                t.rollback();
		throw e;
            } finally {
                t.release();
            }
        } catch (Exception e) {
            throw new SessionException(e);
        }
    }

    /**
     * Executes a query of the database.
     *
     * @param query
     *   the object containg the query statement.
     * @param dbName the database that should be queried.  May be
     *   null in which case the default database is used.
     * @exception SessionException
     *   if an erro occurs.
     */
    static Object dbQuery(Query query, String dbName) throws SessionException {
        try {
            DBQuery q = createQuery(dbName);
            try {
                q.query(query);
                return q.next();
            } catch (Exception e) {
                throw e;
            } finally {
                q.release();
            }
        } catch (Exception e) {
            throw new SessionException(e);
        }
    }

    /**
     * Creates a transaction for the specified DB name.
     *
     * @param dbName the database name.  May be null in
     *   which case the default database is used.
     * @exception DatabaseManagerException if an error occurs.
     * @exception SQLException if an error occurs.
     */
    static DBTransaction createTransaction(String dbName)
        throws DatabaseManagerException, SQLException {
        DBTransaction t;
        if (dbName != null) {
            t = Enhydra.getDatabaseManager().createTransaction(dbName);
        } else {
            t = Enhydra.getDatabaseManager().createTransaction();
        }
        return t;
    }

    /**
     * Creates a query for the specified DB name.
     *
     * @param dbName the database name.  May be null in
     *   which case the default database is used.
     * @exception DatabaseManagerException if an error occurs.
     * @exception SQLException if an error occurs.
     */
    static DBQuery createQuery(String dbName)
        throws DatabaseManagerException, SQLException {
        DBQuery q;
        if (dbName != null) {
            q = Enhydra.getDatabaseManager().createQuery(dbName);
        } else {
            q = Enhydra.getDatabaseManager().createQuery();
        }
        return q;
    }

}




