/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: PersistentSessionDO.java,v 1.4.14.1 2000/10/19 17:59:06 jasona Exp $
 */

package com.lutris.appserver.server.session.persistent;

import com.lutris.appserver.server.sql.*;
import com.lutris.appserver.server.session.*;
import com.lutris.appserver.server.user.*;
import java.util.*;
import java.io.*;
import java.sql.*;
import java.math.*;

/**
 * Database interface for persistent session.
 *
 * @version $Revision: 1.4.14.1 $
 * @author  Kyle Clark
 */
public class PersistentSessionDO implements Transaction {

    /**
     * The session represented by this object.
     */
    private PersistentSession session;

    /**
     * Holds the transaction status.
     */
    private boolean success = false;

    /**
     * Holds the transaction row count.
     */
    private int rowCount = 0;

    /**
     * @param session the session for which this data object
     *   will provide a database interface.
     */
    protected PersistentSessionDO(PersistentSession session) {
        this.session = session;
    }

    /**
     * @param rs
     *   Result set from which to instantiate the session.
     * @param loader
     *   The class loader to use when reconstructing the session
     *   from serialized data.
     * @exception SQLException
     *   If a database connection error occurs.
     * @exception IOException
     *   If the serialized session cannot be re-instantiated from the
     *   persistent data.
     * @exception ClassNotFoundException
     *   If a class associated with the serialized session cannot
     *   be found.
     */
    public PersistentSessionDO(ResultSet rs, ClassLoader loader)
        throws SQLException, IOException, ClassNotFoundException {
        InputStream in = rs.getBinaryStream("data");
        LoaderObjectInputStream objIn =
            new LoaderObjectInputStream(in, loader);
        session = (PersistentSession)objIn.readObject();
        objIn.close();
    }

    /**
     * Inserts the session into the database.
     *
     * @param conn Database connection.
     * @exception java.sql.SQLException If a database access error
     *  occurs or if the session data could not be serialized.
     */
    public void executeInsert(DBConnection conn)
        throws SQLException {
        if (session == null) {
            return;
        }
        String sql = "insert into " + PersistentSessionHome.dbTableName
            + " values (?, ?, ?, ?, ?, ?, ?, ?, ?)";
        PreparedStatement stmt = conn.prepareStatement(sql);
        int param = 1;
        stmt.setString(param++, session.getSessionKey());
        if (session.isNew()) {
            stmt.setString(param++, "1");
        } else {
            stmt.setString(param++, "0");
        }
        stmt.setBigDecimal(param++, new BigDecimal((double)session.getTimeCreated()));
        stmt.setBigDecimal(param++, new BigDecimal((double)session.getTimeLastUsed()));
        stmt.setBigDecimal(param++, new BigDecimal((double)session.getTimeExpires()));
        stmt.setBigDecimal(param++, new BigDecimal((double)session.getMaxIdleTime()));
        stmt.setBigDecimal(param++, new BigDecimal((double)session.getMaxNoUserIdleTime()));
        User user = session.getUser();
        if (user != null) {
            stmt.setString(param++, user.getName());
        } else {
            stmt.setNull(param++, Types.VARCHAR);
        }
        try {
            InputStream in = getBinaryInputStream();
            stmt.setBinaryStream(param++, in, in.available());
            rowCount = conn.executeUpdate(stmt, sql);
            in.close();
        } catch (IOException e) {
            throw new SQLException(e.getMessage());
        }
    }

    /**
     * If this object's <code>executeInsert</code> method was
     * called then <code>finalizeInsert</code> is called with
     * the status of the database transaction.  This method
     * allows the data object to perform any post processing
     * if the transaction succeeded or failed.
     *
     * @param success true if the transaction succeeded
     *   and this object was successfully inserted into the database.
     */
    public void finalizeInsert(boolean success) {
        this.success = success;
        if (!success) {
            this.rowCount = 0;
        }
    }

    /**
     * Method to update contents of object in database.
     *
     * @param conn Database connection.
     * @exception java.sql.SQLException If a database access error
     *  occurs.
     */
    public void executeUpdate(DBConnection conn)
        throws SQLException {
        String sql =
            "update " + PersistentSessionHome.dbTableName 
            + " set isNew = ?, timeCreated = ?, timeLastUsed = ?, "
            + " timeExpires = ?, maxIdleTime = ?, maxNoUserIdleTime = ?, "
            + " userName = ?, data = ? where sessionKey = ?";
        PreparedStatement stmt = conn.prepareStatement(sql);
        int param = 1;
        if (session.isNew()) {
            stmt.setString(param++, "1");
        } else {
            stmt.setString(param++, "0");
        }
        stmt.setBigDecimal(param++, new BigDecimal((double)session.getTimeCreated()));
        stmt.setBigDecimal(param++, new BigDecimal((double)session.getTimeLastUsed()));
        stmt.setBigDecimal(param++, new BigDecimal((double)session.getTimeExpires()));
        stmt.setBigDecimal(param++, new BigDecimal((double)session.getMaxIdleTime()));
        stmt.setBigDecimal(param++, new BigDecimal((double)session.getMaxNoUserIdleTime()));
        User user = session.getUser();
        if (user != null) {
            stmt.setString(param++, user.getName());
        } else {
            stmt.setNull(param++, Types.VARCHAR);
        }
        try {
            InputStream in = getBinaryInputStream();
            stmt.setBinaryStream(param++, in, in.available());
            stmt.setString(param++, session.getSessionKey());
            rowCount = conn.executeUpdate(stmt, sql);
            in.close();
        } catch (IOException e) {
            throw new SQLException(e.getMessage());
        }
    }

    /**
     * If this object's <code>executeUpdate</code> method was
     * called then <code>finalizeUpdate</code> is called with
     * the status of the database transaction.
     * For instance the data object may want to
     * increment its version number once it has successfully
     * been commited to the database.  
     *
     * @param success true if the transaction succeeded
     *   and this object was successfully updated in the database.
     */
    public void finalizeUpdate(boolean success) {
        this.success = success;
        if (!success) {
            this.rowCount = 0;
        }
    }

    /**
     * Method to delete an object from the database.
     *
     * @param conn Database connection.
     * @exception java.sql.SQLException If a database access error
     *  occurs.
     */
    public void executeDelete(DBConnection conn)
        throws SQLException {
        String sql = "delete from " + PersistentSessionHome.dbTableName 
            +  " where sessionKey = ?";
        PreparedStatement stmt = conn.prepareStatement(sql);
        stmt.setString(1, session.getSessionKey());
        rowCount = conn.executeUpdate(stmt, sql);
    }

    /**
     * If this object's <code>executeDelete</code> method was
     * called then <code>finalizeDelete</code> is called with
     * the status of the database transaction.
     *
     * @param success true if the transaction succeeded
     *   and this object was successfully deleted from the
     *   database.
     */
    public void finalizeDelete(boolean success) {
        this.success = success;
        if (!success) {
            this.rowCount = 0;
        }
    }

    /**
     * Returns an input stream containing the serialized
     * session data.
     *
     * @return the input stream.
     */
    private InputStream getBinaryInputStream() throws IOException {
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        ObjectOutputStream objOut = new ObjectOutputStream(out);
        objOut.writeObject(session);
        objOut.flush();
        ByteArrayInputStream in = new ByteArrayInputStream(out.toByteArray());
        objOut.close();
        return in;
    }

    /**
     * Returns the session for which this data object is
     * acting as a data base interface.
     *
     * @return the session
     */
    PersistentSession getSession() {
        return session;
    }

    /**
     * Returns the transaction status.
     *
     * @return the transaction status.
     */
    boolean getTransactionStatus() {
        return success;
    }

    /**
     * Returns the transaction count - i.e. the number
     * of rows affected in the transaction.
     */
    int getTransactionRowCount() {
        return rowCount;
    }
}


// Internal class that can load objects with the loader we specify.
class LoaderObjectInputStream extends ObjectInputStream {

    private ClassLoader loader;

    public LoaderObjectInputStream(InputStream in, ClassLoader loader)
	throws IOException, StreamCorruptedException {
        super(in);
        this.loader = loader;
    }

    /**
     * Subclasses may implement this method to allow classes to be
     * fetched from an alternate source. 
     *
     * The corresponding method in ObjectOutputStream is
     * annotateClass.  This method will be invoked only once for each
     * unique class in the stream.  This method can be implemented by
     * subclasses to use an alternate loading mechanism but must
     * return a Class object.  Once returned, the serialVersionUID of the
     * class is compared to the serialVersionUID of the serialized class.
     * If there is a mismatch, the deserialization fails and an exception
     * is raised. <p>
     *
     * By default the class name is resolved relative to the class
     * that called readObject. <p>
     *
     * @exception ClassNotFoundException If class of
     * a serialized object cannot be found.
     * @since     JDK1.1
     */
    protected Class resolveClass(ObjectStreamClass v)
	throws IOException, ClassNotFoundException {
        if (loader != null) {
            return loader.loadClass(v.getName());
        } else {
            return super.resolveClass(v);
        }
    }

}
