/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: DebugRequest.java,v 1.12.14.1 2000/10/19 17:59:07 jasona Exp $
 */





package com.lutris.http.debug;

import java.util.Enumeration;
import javax.servlet.http.Cookie;

/**
 * Object used to access HTTP request data for debugging purposes.
 *
 * @author	Paul Morgan
 * @since	LBS1.7
 * @version	$Revision: 1.12.14.1 $
 */
public interface DebugRequest {
    /**
     * Returns the size of the request entity data, or -1 if not known.
     *
     * @exception Exception if an error occurs.
     */
    public int getContentLength() throws Exception;

    /**
     * Returns the Internet Media Type of the request entity data, or null
     * if not known.
     *
     * @exception Exception if an error occurs.
     */
    public String getContentType() throws Exception;

    /**
     * Returns the protocol and version of the request as a string of
     * the form <code>&lt;protocol&gt;/&lt;major version&gt;.&lt;minor
     * version&gt</code>.
     *
     * @exception Exception if an error occurs.
     */
    public String getProtocol() throws Exception;

    /**
     * Returns the scheme of the URL used in this request, for example
     * "http", "https", or "ftp".  Different schemes have different
     * rules for constructing URLs, as noted in RFC 1738.  The URL used
     * to create a request may be reconstructed using this scheme, the
     * server name and port, and additional information such as URIs.
     */
    public String getScheme();

    /**
     * Returns the host name of the server that received the request.
     */
    public String getServerName();

    /**
     * Returns the port number on which this request was received.
     */
    public int getServerPort();

    /**
     * Returns the IP address of the agent that sent the request.
     *
     * @exception Exception if an error occurs.
     */
    public String getRemoteAddr() throws Exception;

    /**
     * Returns the fully qualified host name of the agent that sent the
     * request.
     *
     * @exception Exception if an error occurs.
     */
    public String getRemoteHost() throws Exception;

    /**
     * Returns the values of the specified query parameter for the request as
     * an array of strings, or a 0 length array if the named parameter does
     * not exist.
     *
     * @param name
     *   the name of the parameter whose value is required.
     * @exception Exception if an error occurs.
     */
    public String[] getParameterValues(String name) throws Exception;

    /**
     * Returns the parameter names for this request as an array of
     * strings, or a 0 length array if no parameters exist.
     */
    public Enumeration getParameterNames() throws Exception;

    /**
     * Returns the method with which the request was made. The returned
     * value can be "GET", "HEAD", "POST", or an extension method.
     *
     * @exception Exception if an error occurs.
     */
    public String getMethod() throws Exception;

    /**
     * Returns the request URI.
     *
     * @exception Exception if an error occurs.
     */
    public String getRequestURI() throws Exception;

    /**
     * Returns the part of the request URI that refers to the presentation
     * object or servlet being invoked.
     *
     * @exception Exception if an error occurs.
     */
    //FIXME:  Don't know where PM got this, as nothing implements this.
    // public String getPath() throws Exception;

    /**
     * Returns optional extra path information following the presentation
     * path, but immediately preceding the query string. Returns null if
     * not specified. Same as the CGI variable PATH_INFO.
     *
     * @exception Exception if an error occurs.
     */
    public String getPathInfo() throws Exception;

    /**
     * Returns extra path information translated to a real path. Returns
     * null if no extra path information specified.
     *
     * @exception Exception if an error occurs.
     */
    public String getPathTranslated() throws Exception;

    /**
     * Returns the query string part of the presentation URI, or null if none.
     *
     * @exception Exception if an error occurs.
     */
    public String getQueryString() throws Exception;

    /**
     * Returns the name of the user making this request, or null if not
     * known.  The user name is set with HTTP authentication.
     *
     * @exception Exception if an error occurs.
     */
    public String getRemoteUser() throws Exception;

    /**
     * Returns the authentication scheme of the request, or null if none.
     *
     * @exception Exception if an error occurs.
     */
    public String getAuthType() throws Exception;

    /**
     * Returns an array of cookies for this request, or a 0
     * length array if no cookies exist.
     *
     * @exception Exception if an error occurs.
     * @return The array of cookies found in this request.
     */
    public Cookie[] getCookies() throws Exception;

    /**
     * Returns the values of the specified header. This should really
     * return an array of strings, however Sun's servlet classes do
     * not provide this functionality and thus implementing it would
     * be mute.
     *
     * @param name
     * 	 the case-insensitive header field name
     * @exception Exception if an error occurs.
     */
    public String getHeader(String name) throws Exception;

    /**
     * Returns the strings representing the header names for this request.
     * This does not have to return cookies. getCookies() can be used.
     *
     * @exception Exception if an error occurs.
     */
    public Enumeration getHeaderNames() throws Exception;

    /**
     * Returns the number of bytes associated with the request.
     */
    public int getTotalBytes();
}
