/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HttpTestClient.java,v 1.1.2.2.2.1 2000/10/19 17:59:08 jasona Exp $
 */

package org.enhydra.httpServerTest.client;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;
import java.util.Iterator;
import java.io.ByteArrayOutputStream;
import java.io.FileOutputStream;
import java.io.File;

import com.simscomputing.testbed.TestCaseCollection;
import com.simscomputing.testbed.TestResultCollection;
import com.simscomputing.testbed.TestFailedException;
import com.simscomputing.testbed.TestResult;

/**
 * This class represents a test case collection for the HttpServerTest module.
 * A test case collection is a class that executes any number of individual
 * test cases.  The class is derived from TestCaseCollection and the TestBed
 * module provides the framework to run and log all test cases within a 
 * collection.
 * @see com.simscomputing.testbed.TestCaseCollection
 *
 * This class reads individual test case parameters from an XML configuration 
 * file, HttpTestClientConfig.dtd.
 * @see org.enhydra.httpServerTest.client.HttpTestClientConfig
 * @see org.enhydra.httpServerTest.client.HttpTestClientConfigParser
 *
 * Test case collections are run from the following classes:
 * @see com.simscomputing.testbed.gui.Main
 * @see com.simscomputing.testbed.Main
 *
 * @author Mike Ward
 **/
public class HttpTestClient
    extends TestCaseCollection {

    /**
     * The URI of the XML configuration file.
     **/
    public static final String HTTP_CLIENT_TEST_CONFIG_FILE = 
	"file:" + Main.USER_DEFINED_CONFIG_FILE;
    
    /**
     * The hashtable key for the single user GET request test case.  This key
     * must match the configuration attribute 'testName' of the Test element.
     * @see org.enhydra.httpServerTest.client.HttpTestClientConfig
     **/
    public static final String SINGLE_USER_GET_REQUEST_KEY =
	"SingleUserGetRequest";

    /**
     * The hashtable key for the single user POST request test case.  This key
     * must match the configuration attribute 'testName' of the Test element.
     * @see org.enhydra.httpServerTest.client.HttpTestClientConfig
     **/
    public static final String SINGLE_USER_POST_REQUEST_KEY =
	"SingleUserPostRequest";

    /**
     * A hashtable of test case configuration parameters.  Each test case
     * can have 0 or more configuration sections, they are indexed in the 
     * hashtable as <TEST_CASE_NAME>_<CONFIG_CNT>.
     **/
    private Hashtable testSuiteParams = null;

    /**
     * A boolean flag to indicate if any errors were encountered while
     * setting up the collection.  Currently, the only errors encountered
     * during setup will be configuration parser errors.
     **/
    private boolean setupErrorFound = false;

    /**
     * The error message associated with setup.
     **/
    private String setupErrorString = null;
    
    /**
     * This function is called before any test cases are executed by the 
     * TestBed module.  It calls on the XML configuration parser to
     * parse and return a hashtable of test parameters.  If the XML
     * file fails to validate against HttpTestClientConfig.dtd the TestBed
     * will exit with an error on what failed during parsing.  Parsing 
     * errors will appear as a test case assertion failure.
     * @see org.enhydra.httpServerTest.client.HttpTestClientConfigParser
     **/
    public void setupCollection() {
	try {
	    testSuiteParams = 
		HttpTestClientConfigParser.parse(HTTP_CLIENT_TEST_CONFIG_FILE);
	} catch (Exception e) {
	    setupErrorString = e.toString();
	    setupErrorFound = true;
	}
    }

    /**
     * Called after each test case is run.  Allows the opertunity for
     * a test case to log all standard output.  Any error occuring while
     * opening and/or writing to the log file will cause this function to
     * quitely exit.
     *
     * @param testCaseCollectionName
     *        The current class name of test case collection
     * @param testCaseName
     *        The current name of the test case within the test case 
     *        collection.
     * @param outStream
     *        The collected output from the test case that just finished 
     *        executing.
     **/
    public void logStdOut(String testCaseCollectionName, 
			  String testCaseName,
			  ByteArrayOutputStream outStream) {
	try {
	    File dir = new File(org.enhydra.httpServerTest.client.Main.USER_DEFINED_LOG_DIR);
	    dir.mkdirs();
	    File outFile = new File(dir, 
				    testCaseCollectionName + "." +
				    testCaseName + ".out");
	    
	    FileOutputStream fileOutputStream = new FileOutputStream(outFile);
	    fileOutputStream.write(outStream.toByteArray());
	    fileOutputStream.flush();
	    fileOutputStream.close();
	} catch (Exception e) {
	}
    }

    /**
     * Called after each test case is run.  Allows the opertunity for
     * a test case to log all standard error.  Any error occuring while
     * opening and/or writing to the log file will cause this function to
     * quitely exit.
     *
     * @param testCaseCollectionName
     *        The current class name of test case collection
     * @param testCaseName
     *        The current name of the test case within the test case 
     *        collection.
     * @param errStream
     *        The collected output from the test case that just finished 
     *        executing.
     **/
    public void logStdError(String testCaseCollectionName, 
			    String testCaseName,
			    ByteArrayOutputStream errStream) {
	try {
	    File dir = new File(org.enhydra.httpServerTest.client.Main.USER_DEFINED_LOG_DIR);
	    dir.mkdirs();
	    File outFile = new File(dir,
				    testCaseCollectionName + "." +
				    testCaseName + ".err");
	    
	    FileOutputStream fileOutputStream = new FileOutputStream(outFile);
	    fileOutputStream.write(errStream.toByteArray());
	    fileOutputStream.flush();
	    fileOutputStream.close();
	} catch (Exception e) {
	}
    }

    /**
     * This test case spawns one or more SingleUserGetRequestThreads and 
     * logs all aserts thrown during thread execution.  If a setup error
     * is found, this test case will log the error with an assertion and
     * exit.  If no setup error is found this test case will continually 
     * loop until all configuration sections have been executed.  Failed 
     * assertions are not reported until all configuration sections are 
     * complete.
     * @see org.enhydra.httpServerTest.client.SingleUserGetRequestThread
     **/
    public void runTestSingleUserGetRequest() {
	if (setupErrorFound) {
	    assert(false, "Error parsing " + HTTP_CLIENT_TEST_CONFIG_FILE +
		   ": " + setupErrorString);
	    return;
	}
	    
	int testCaseCnt = 0;

	Vector threadResultsVector = new Vector();

	while (true) {
	    HttpTestClientConfig singleUserGetRequestParams = 
		getTestCaseConfig(SINGLE_USER_GET_REQUEST_KEY, testCaseCnt);
	    if (singleUserGetRequestParams == null) {
		break;
	    }
	    
	    ThreadGroup singleUserGetRequestThreadGroup = 
		new ThreadGroup(SINGLE_USER_GET_REQUEST_KEY);
	    ThreadResultsCollector[] threadResultsArray = 
		new ThreadResultsCollector[singleUserGetRequestParams.numThreads];
	    for (int i = 0; i < singleUserGetRequestParams.numThreads; i++) {
		try {
		    threadResultsArray[i] = (ThreadResultsCollector)
			new ThreadResultsCollectorImpl();
		    SingleUserGetRequestThread singleUserGetRequestThread =
			new SingleUserGetRequestThread(singleUserGetRequestParams.id + ":ThreadNum_" + i, singleUserGetRequestParams.cloneInstance(), threadResultsArray[i]);
		    new Thread(singleUserGetRequestThreadGroup,
			       singleUserGetRequestThread).start();
		} catch (IllegalArgumentException e) {
		    assert(false, "ERROR: IllegalArgumentException creating SingleUserGetRequestThread: " + e);
		} catch (SecurityException e) {
		    assert(false, "ERROR: SecurityException creating SingleUserGetRequestThread: " + e);
		} catch (Exception e) {
		    assert(false, "ERROR: Exception thrown in SingleUserGetRequestThread: " + e);
		}
	    }

	    waitForActiveThreads(singleUserGetRequestThreadGroup);
	    addFailedAssertsToResultsVector(threadResultsVector,
					    threadResultsArray);

	    testCaseCnt++;
	}
	checkAsserts(threadResultsVector);
    }

    /**
     * This test case spawns one or more SingleUserPostRequestThreads and 
     * logs all aserts thrown during thread execution.  If a setup error
     * is found, this test case will log the error with an assertion and
     * exit.  If no setup error is found this test case will continually 
     * loop until all configuration sections have been executed.  Failed 
     * assertions are not reported until all configuration sections are 
     * complete.
     * @see org.enhydra.httpServerTest.client.SingleUserPostRequestThread
     **/
    public void runTestSingleUserPostRequest() {
	if (setupErrorFound) {
	    assert(false, "Error parsing " + HTTP_CLIENT_TEST_CONFIG_FILE +
		   ": " + setupErrorString);
	    return;
	}
	    
	int testCaseCnt = 0;

	Vector threadResultsVector = new Vector();

	while (true) {
	    HttpTestClientConfig singleUserPostRequestParams = 
		getTestCaseConfig(SINGLE_USER_POST_REQUEST_KEY, testCaseCnt);
	    if (singleUserPostRequestParams == null) {
		break;
	    }

	    ThreadGroup singleUserPostRequestThreadGroup = 
		new ThreadGroup(SINGLE_USER_POST_REQUEST_KEY);
	    ThreadResultsCollector[] threadResultsArray = 
		new ThreadResultsCollector[singleUserPostRequestParams.numThreads];
     
	    for (int i = 0; i < singleUserPostRequestParams.numThreads; i++) {
		try {
		    threadResultsArray[i] = (ThreadResultsCollector)
			new ThreadResultsCollectorImpl();
		    SingleUserPostRequestThread singleUserPostRequestThread =
			new SingleUserPostRequestThread(singleUserPostRequestParams.id + ":ThreadNum_" + i, singleUserPostRequestParams.cloneInstance(), threadResultsArray[i]);
		    new Thread(singleUserPostRequestThreadGroup,
			       singleUserPostRequestThread).start();
		} catch (IllegalArgumentException e) {
		    assert(false, "ERROR: IllegalArgumentException creating SingleUserPostRequestThread: " + e);
		} catch (SecurityException e) {
		    assert(false, "ERROR: SecurityException creating SingleUserPostRequestThread: " + e);
		} catch (Exception e) {
		    assert(false, "ERROR: Exception thrown in SingleUserPostRequestThread: " + e);
		}
	    }

	    waitForActiveThreads(singleUserPostRequestThreadGroup);
	    addFailedAssertsToResultsVector(threadResultsVector,
					    threadResultsArray);

	    testCaseCnt++;
	}
	checkAsserts(threadResultsVector);
    }

    /**
     * This function will check for the hashtable key 
     * <testCaseKey>_<testCaseCnt> in testSuiteParams.  If the key exists
     * the configuration section for the test is returned.  Otherwise null
     * is returned.
     *
     * @param testCaseKey
     *        The name of the current test case being executed.
     * @param testCaseCnt
     *        The number of previously read configuration sections for this
     *        test.
     * @return HttpTestClientConfig
     *         The configuration section for the test case, or null if not
     *         found.
     **/
    protected HttpTestClientConfig getTestCaseConfig(String testCaseKey,
						     int testCaseCnt) {
	String testCaseHash = testCaseKey + "_" + testCaseCnt;
	if (testSuiteParams.containsKey(testCaseHash)) {
	    return (HttpTestClientConfig)testSuiteParams.get(testCaseHash);
	} else {
	    return null;
	}
    }

    /**
     * This function continually loops until all thread in threadGroup
     * have finished executing.  If an exception is thrown while looping,
     * the test fails and an assertion is logged.
     *
     * @param threadGroup
     *        The thread group to watch.
     **/
    protected void waitForActiveThreads(ThreadGroup threadGroup) {
	if (threadGroup == null) {
	    return;
	}

	while(threadGroup.activeCount() >= 1) {
	    // This conditional check is used to manually close the 
	    // SocketTimeout thread spawned by the HTTPClient package.
	    // For one reason or another this thread refuses to die on
	    // its own!!
	    if (threadGroup.activeCount() == 1) {
		Thread[] activeThreadArray = new Thread[1];
		threadGroup.enumerate(activeThreadArray, true);
		if (activeThreadArray[0] != null &&
		    activeThreadArray[0].getName() != null &&
		    activeThreadArray[0].getName().equals("SocketTimeout")) {
		    try {
			activeThreadArray[0].stop();
		    } catch (SecurityException e) {
			assert(false, "ERROR: SecurityException caught while trying to stop SocketTimeout thread. " + e);
		    }
		}
	    }

	    try {
		Thread.sleep((long)100);
	    } catch (InterruptedException e) {
		assert(false, "ERROR: InterruptedException caught while waiting for all Threads in ThreadGroup " + threadGroup.getName() + " to complete: " + e);
	    }
	}
    }

    /**
     * This function adds any reported failed asserts to the results vector.
     * Failed asserts are logged to the ThreadResultsCollector by each child
     * thread spawned within a single test case.  A test case is considered
     * failed if the resultsVector has one or more failed asserts when 
     * all configuration sections for that test case have completed.
     *
     * @param resultsVector
     *        The vector that holds all failed asserts.
     * @param resultsArray
     *        The array of results reported by the child threads spawned
     *        within a test case.
     **/
    protected void addFailedAssertsToResultsVector(Vector resultsVector,
						   ThreadResultsCollector[] resultsArray) {
	for (int i = 0; i < resultsArray.length; i++) {
	    if (resultsArray[i].failedAssertsState()) {
		resultsVector.addElement(resultsArray[i]);
	    }
	}
    }

    /**
     * This function checks the results from a test case and formats and logs
     * any failed assertions.  If no failed assertions are found, the test
     * case passed and the function returns with no action.
     *
     * @param resultsVector
     *        The results vector from a test case.
     **/
    protected void checkAsserts(Vector resultsVector) {

	ThreadResultsCollector[] resultsArray = 
	    new ThreadResultsCollector[resultsVector.size()];
	resultsVector.copyInto(resultsArray);

	StringBuffer failedAssertBuffer = new StringBuffer();

	for (int i = 0; i < resultsArray.length; i++) {
	    if (resultsArray[i].failedAssertsState()) {
		Enumeration failedAsserts = 
		    resultsArray[i].getFailedAsserts();
		while (failedAsserts.hasMoreElements()) {
		    if (failedAssertBuffer.length() == 0) {
			failedAssertBuffer.append("\nReported Thread Errors");
			failedAssertBuffer.append("\n----------------------");
		    }
		    failedAssertBuffer.append("\n\t" + 
				       (String)failedAsserts.nextElement());
		}
	    }
	}

	if (failedAssertBuffer.length() > 0) {
	    assert(false, failedAssertBuffer.toString());
	}
    }
}
