/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HttpTestClientUtils.java,v 1.1.2.3.2.1 2000/10/19 17:59:08 jasona Exp $
 */

package org.enhydra.httpServerTest.client;

import org.enhydra.httpServerTest.common.HttpResponseObject;
import org.enhydra.httpServerTest.common.HttpServerTestUtils;

import HTTPClient.CookieModule;
import HTTPClient.Cookie;
import HTTPClient.HTTPConnection;
import HTTPClient.NVPair;

import java.net.URLEncoder;
import java.util.Enumeration;
import java.util.Vector;

/**
 * This class is a set of utilities that are used by client threads to help
 * construct requests information used for issueing http request and validating
 * response objects.
 *
 * @author Mike Ward
 **/
public class HttpTestClientUtils {
    
    /**
     * This function constructs a URI string from the HttpTestClientConfig
     * object.  The string consists of the HttpTestClientConfig file element
     * followed by any CGI parameters declared.  If a redirected to file 
     * attribute is declared, this value is used in place of the file 
     * attribute.
     *
     * @see org.enhydra.httpServerTest.client.HttpTestClientConfig
     *
     * @param params
     *        The configuration object to construct URI string from.
     * @return String
     *         The URI string.
     **/
    public static String getURI(HttpTestClientConfig params) {

	StringBuffer uri = null;
	if (params.redirectedToFile == null) {
	    uri = new StringBuffer(params.file);
	} else {
	    uri = new StringBuffer(params.redirectedToFile);
	}

	Enumeration cgiParamKeys = params.cgiParams.keys();
	while(cgiParamKeys.hasMoreElements()) {
	    String cgiParamName = (String)cgiParamKeys.nextElement();
	    String cgiParamValue = (String)params.cgiParams.get(cgiParamName);
	    HttpTestClientUtils.addCgiParam(uri,
					    cgiParamName,
					    cgiParamValue);
	}
	if (uri.length() == 0) {
	    return "/";
	} else {
	    return uri.toString();
	}
    }

    /**
     * This function returns a set of NVPairs representing all declared
     * CGI arguments in the configuration object.  These arguments are
     * sent as part an http request.
     *
     * @see org.enhydra.httpServerTest.client.SingleUserGetRequestThread
     * @see org.enhydra.httpServerTest.client.SingleUserPostRequestThread
     * @see HTTPClient.HTTPConnection
     * @see HTTPClient.NVPair
     *
     * @param params
     *        The configuration object holding declared CGI arguments.
     * @return NVPair[]
     *         Name-Value pair array holding CGI arguments.
     **/
    public static NVPair[] getCgiArgs(HttpTestClientConfig params) {

	Enumeration cgiArgKeys = params.cgiArgs.keys();
	Vector cgiArgVector = new Vector();
	while(cgiArgKeys.hasMoreElements()) {
	    String cgiArgName = (String)cgiArgKeys.nextElement();
	    String cgiArgValue = (String)params.cgiArgs.get(cgiArgName);
	    cgiArgVector.add(new NVPair(cgiArgName, cgiArgValue));
	}
	NVPair[] cgiArgArray = new NVPair[cgiArgVector.size()];
	cgiArgVector.copyInto(cgiArgArray);
	return cgiArgArray;
    }

    /**
     * This function returns a set of NVPairs representing all declared
     * request headers in the configuration object.  These headers are
     * sent as part an http request.
     *
     * @see org.enhydra.httpServerTest.client.SingleUserGetRequestThread
     * @see org.enhydra.httpServerTest.client.SingleUserPostRequestThread
     * @see HTTPClient.HTTPConnection
     * @see HTTPClient.NVPair
     *
     * @param params
     *        The configuration object holding declared request headers.
     * @return NVPair[]
     *         Name-Value pair array holding request headers.
     **/
    public static NVPair[] getHeaders(HttpTestClientConfig params) {

	Enumeration headerKeys = params.headers.keys();
	Vector headerVector = new Vector();
	while(headerKeys.hasMoreElements()) {
	    String headerName = (String)headerKeys.nextElement();
	    String headerValue = (String)params.headers.get(headerName);
	    headerVector.add(new NVPair(headerName, headerValue));
	}
	NVPair[] headerArray = new NVPair[headerVector.size()];
	headerVector.copyInto(headerArray);
	return headerArray;
    }

    /**
     * This function adds all declared cookies to the CookieModule in
     * the context of the current connection.  All cookies in the current
     * connection context are automatically sent by HTTPConnection when
     * the request is made.  The cookie's parameters are set to:
     *   Name: declared cookie name
     *   Value: declared cookie value
     *   Path: '/'
     *   Exp Date: 'null'
     *   Secure: 'false'
     *
     * @see org.enhydra.httpServerTest.client.SingleUserGetRequestThread
     * @see org.enhydra.httpServerTest.client.SingleUserPostRequestThread
     * @see HTTPClient.HTTPConnection
     * @see HTTPClient.CookieModule
     *
     * @param con
     *        The current http connection.
     * @param params
     *        The configuration object holding declared cookies.
     **/
    public static void addCookies(HTTPConnection con,
				  HttpTestClientConfig params) {
	
	Enumeration cookieKeys = params.cookies.keys();
	while(cookieKeys.hasMoreElements()) {
	    String cookieName = (String)cookieKeys.nextElement();
	    String cookieValue = (String)params.cookies.get(cookieName);
	    Cookie cookie = new Cookie(cookieName, 
				       cookieValue,
				       params.host,  // domain
				       "/",  // path
				       null,  // Exp Date
				       false); // Secure
	    CookieModule.addCookie(cookie, con.getContext());
	}
    }

    /**
     * This function is used to clear all cookies for the current connection
     * context.  This is usually called between requests because the default
     * behavior of HTTPConnection is to act as a browser would and resend all
     * relevent cookies with each request.  The client side request threads
     * typically don't want this behavior and reconstruct the cookies 
     * explicitly before each request.
     *
     * @see org.enhydra.httpServerTest.client.SingleUserGetRequestThread
     * @see org.enhydra.httpServerTest.client.SingleUserPostRequestThread
     * @see HTTPClient.HTTPConnection
     * @see HTTPClient.CookieModule
     *
     * @param con
     *        The current http connection.
     **/
    public static void removeCookies(HTTPConnection con) {

	CookieModule.discardAllCookies(con.getContext());
    }


    /**
     * This function is used as a post initialization for session information.
     * If session tracking is required, the session key is unknown until the
     * first request has been made.  This function will add a session cookie
     * or a session CGI parameter to the configuration object so that 
     * subsequent request will send a valid session key.  This function will
     * also add the session key to the given response and golden objects so 
     * that a comparison of these objects will return true.
     *
     * @see org.enhydra.httpServerTest.client.BasicTestThread
     * @see org.enhydra.httpServerTest.common.HttpResponseObject
     *
     * @param con
     *        The current http connection, used to retrieve the current
     *        connection context so only relevant cookies are checked for
     *        session keys.
     * @param params
     *        The configuration parameters.  This object is populated with
     *        the valid session key if session tracking is enabled.
     * @param golden
     *        This is the golden response object.  This object is used by
     *        client threads as a comparison object to evaluate the validity
     *        of an http response.  If session tracking is enabled, this object
     *        is populated with the valid session key.
     * @param resp
     *        This is the curent response object.  This object represents the
     *        response sent back from the server and is used by the client
     *        threads to test against the golden object.  If session tracking
     *        is enabled, this object is populated with the valid session key.
     **/
    public static void populateSessionInfo(HTTPConnection con,
					   HttpTestClientConfig params,
					   HttpResponseObject golden,
					   HttpResponseObject resp) {

	if (resp == null) {
	    return;
	}

	Cookie[] cookies = CookieModule.listAllCookies(con.getContext());
	Cookie sessionCookie = null;
	for (int i = 0; i < cookies.length; i++) {
	    if (cookies[i].getName().equals(HttpServerTestUtils.HTTP_COOKIE_SESSION_KEY)) {
		sessionCookie = cookies[i];
		break;
	    }
	}

	if (params.sessionTracking.equals(HttpTestClientConfig.DEFAULT_SESSION_TRACKING) ||
	    golden.session != HttpResponseObject.DEFAULT_SESSION) {
	    // Session tracking not useed or has been handed a value from
	    // configuration file
	    return;
	} else if (params.sessionTracking.equals(HttpTestClientConfig.SESSION_TRACKING_BY_COOKIE)) {
	    if (sessionCookie != null) {
		setCookieTrackingSessionInfo(sessionCookie,
					     params,
					     golden,
					     resp);
	    }
	} else if (params.sessionTracking.equals(HttpTestClientConfig.SESSION_TRACKING_BY_CGI_PARAM)) {
	    if (sessionCookie != null) {
		setCgiParamTrackingSessionInfo(sessionCookie,
					       params,
					       golden,
					       resp);
	    }
	}
    }

    /**
     * This function is a helper function for 
     * HttpTestClientUtils.populateSessionInfo().  It initializes the 
     * configuration, golden, and response objects to reflect session
     * tracking with a cookie.
     *
     * @param sessionCookie
     *        The current session cookie associated with the http connection.
     * @param params
     *        The configuration parameters.  This object is populated with
     *        the valid session key as a cookie.
     * @param golden
     *        This is the golden response object.  This object is used by
     *        client threads as a comparison object to evaluate the validity
     *        of an http response.  A session cookie is added to this object.
     * @param resp
     *        This is the curent response object.  This object represents the
     *        response sent back from the server and is used by the client
     *        threads to test against the golden object.  A session cookie
     *        is added to this object.
     **/
    public static void setCookieTrackingSessionInfo(Cookie sessionCookie,
						   HttpTestClientConfig params,
						   HttpResponseObject golden,
						   HttpResponseObject resp) {
	golden.session = sessionCookie.getValue();
	golden.sessionIdFromCookie = true;
	golden.sessionIdFromURL = false;
	golden.sessionIdInvalid = false;

	resp.session = sessionCookie.getValue();
	resp.sessionIdFromCookie = true;
	resp.sessionIdFromURL = false;
	resp.sessionIdInvalid = false;

	// Add session info to both the received response and the 
	// golden response.
	golden.cookies.put(HttpServerTestUtils.HTTP_COOKIE_SESSION_KEY,
			   sessionCookie.getValue());
	resp.cookies.put(HttpServerTestUtils.HTTP_COOKIE_SESSION_KEY,
			 sessionCookie.getValue());
	// Add to thread params so session cookie gets sent for now on
	// and appears when testing equivalence of response object
	params.cookies.put(HttpServerTestUtils.HTTP_COOKIE_SESSION_KEY,
			   sessionCookie.getValue());
    }

    /**
     * This function is a helper function for 
     * HttpTestClientUtils.populateSessionInfo().  It initializes the 
     * configuration, golden, and response objects to reflect session
     * tracking with a CGI parameter.
     *
     * @param sessionCookie
     *        The current session cookie associated with the http connection.
     * @param params
     *        The configuration parameters.  This object is populated with
     *        the valid session key as a CGI parameter.
     * @param golden
     *        This is the golden response object.  This object is used by
     *        client threads as a comparison object to evaluate the validity
     *        of an http response.  A session CGI parameter is added to this 
     *        object.
     * @param resp
     *        This is the curent response object.  This object represents the
     *        response sent back from the server and is used by the client
     *        threads to test against the golden object.  A session CGI
     *        parameter is added to this object.
     **/
    public static void setCgiParamTrackingSessionInfo(Cookie sessionCookie,
						   HttpTestClientConfig params,
						   HttpResponseObject golden,
						   HttpResponseObject resp) {
	golden.session = sessionCookie.getValue();
	golden.sessionIdFromCookie = false;
	golden.sessionIdFromURL = true;
	golden.sessionIdInvalid = false;

	resp.session = sessionCookie.getValue();
	resp.sessionIdFromCookie = false;
	resp.sessionIdFromURL = true;
	resp.sessionIdInvalid = false;

	// Add session info to both the received response and the 
	// golden response.
	golden.cgiParams.put(HttpServerTestUtils.CGI_PARAM_SESSION_KEY,
			     sessionCookie.getValue());
	resp.cgiParams.put(HttpServerTestUtils.CGI_PARAM_SESSION_KEY,
			   sessionCookie.getValue());
	// Add to thread params so session param gets sent for now on
	// and appears when testing equivalence of response object
	params.cgiParams.put(HttpServerTestUtils.CGI_PARAM_SESSION_KEY,
			     sessionCookie.getValue());

	golden.requestURI = HttpTestClientUtils.getURI(params);
	resp.requestURI = HttpTestClientUtils.getURI(params);
	if (params.redirectedToFile == null) {
	    golden.pathInfo = 
		golden.requestURI.substring(params.file.length());
	    resp.pathInfo = 
		resp.requestURI.substring(params.file.length());
	} else {
	    golden.pathInfo = 
		golden.requestURI.substring(params.redirectedToFile.length());
	    resp.pathInfo = 
		resp.requestURI.substring(params.redirectedToFile.length());
	}
    }

    /**
     * This funtion is used as a post initialization for the path translated
     * element.  This, like session cookies, can not be determined by the
     * client until the first request has been made.  This function 
     * initializes the golden object with the path translated information
     * found in the response object.
     *
     * @param params
     *        The configuration parameters.  This object is populated with
     *        the valid session key as a CGI parameter.
     * @param golden     
     *        This is the golden response object.  This object is used by
     *        client threads as a comparison object to evaluate the validity
     *        of an http response.  This object is update with the path
     *        translated information from the reponse.  This value can only
     *        be set once.
     * @param resp
     *        This is the curent response object.  This object represents the
     *        response sent back from the server and is used by the client
     *        threads to test against the golden object.  The path translated
     *        information is retrieved from this object.
     **/
    protected static void populatePathTranslated(HttpTestClientConfig params,
						 HttpResponseObject golden,
						 HttpResponseObject resp) {
	if (resp == null) {
	    return;
	}

	if (golden.pathTranslated == null ||
	    golden.pathTranslated.equals("null")) {
	    golden.pathTranslated = resp.pathTranslated;
	    if (params.sessionTracking.equals(params.SESSION_TRACKING_BY_CGI_PARAM) &&
		resp.session != null) {
		golden.pathTranslated += (";" + 
					  HttpServerTestUtils.CGI_PARAM_SESSION_KEY +
					  "=" + resp.session);
		resp.pathTranslated += (";" + 
					HttpServerTestUtils.CGI_PARAM_SESSION_KEY +
					"=" + resp.session);
	    }
		
	}
    }

    /**
     * This function appends a CGI parameter to the current URI string buffer.
     *
     * @param uri
     *        The current URI string buffer.
     * @param name
     *        The name of the CGI parameter to add.
     * @param value
     *        The value of the CGI parameter to add.
     **/
    public static void addCgiParam(StringBuffer uri,
				   String name,
				   String value) {
	if (uri == null ||
	    name == null ||
	    value == null) {
	    return;
	}

	String encodedName = URLEncoder.encode(name);
	String encodedValue = URLEncoder.encode(value);

	int paramIndex = 0;
	for (; paramIndex < uri.length(); paramIndex++) {
	    if (uri.charAt(paramIndex) == ';') {
		break;
	    }
	}
	if (paramIndex < uri.length()) {
	    uri.insert(paramIndex, ";" + encodedName + "=" + encodedValue);
	} else {
	    uri.append(";" + encodedName + "=" + encodedValue);
	}
    }

    /**
     * This function appends a CGI argument to the current URI string buffer.
     *
     * @param uri
     *        The current URI string buffer.
     * @param name
     *        The name of the CGI argument to add.
     * @param value
     *        The value of the CGI argument to add.
     **/
    public static void addCgiArgGetMethod(StringBuffer url,
					  String name,
					  String value) {
	if (url == null ||
	    name == null ||
	    value == null) {
	    return;
	}

	String encodedName = URLEncoder.encode(name);
	String encodedValue = URLEncoder.encode(value);

	int paramIndex = 0;
	for (; paramIndex < url.length(); paramIndex++) {
	    if (url.charAt(paramIndex) == '?') {
		break;
	    }
	}
	if (paramIndex < url.length()) {
	    url.insert(++paramIndex, encodedName + "=" + encodedValue + "&");
	} else {
	    url.append("?" + encodedName + "=" + encodedValue);
	}
    }
}
