/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: SingleUserPostRequestThread.java,v 1.6.2.4.2.1 2000/10/19 17:59:08 jasona Exp $
 */

package org.enhydra.httpServerTest.client;

import org.enhydra.httpServerTest.common.HttpResponseObjectParser;
import org.enhydra.httpServerTest.common.HttpResponseObject;
import org.enhydra.httpServerTest.common.HttpServerTestUtils;

import org.xml.sax.InputSource;

import HTTPClient.ModuleException;
import HTTPClient.HTTPConnection;
import HTTPClient.HTTPResponse;
import HTTPClient.NVPair;
import HTTPClient.Codecs;
import HTTPClient.ProtocolNotSuppException;

import java.io.IOException;
import java.io.StringReader;
import java.net.URLEncoder;
import java.util.Hashtable;
import java.util.Enumeration;

/**
 * This class extends BaseTestThread and implements POST requests based on
 * the configuration parameters received.
 * @see org.enhydra.httpServerTest.client.BaseTestThread
 * @see org.enhydra.httpServerTest.client.HttpTestClientConfig
 *
 * For every POST request issued, this class will compare the response against
 * an expected response and report an error if differences are found.  This 
 * class will also log response times to System.out and record any
 * errors encountered while establishing connections.  Errors are collected
 * and reported when all threads have completed.
 * @see org.enhydra.httpServerTest.client.ThreadResultsCollector
 * @see org.enhydra.httpServerTest.client.HttpTestClient
 * 
 * @author Mike Ward
 **/
public class SingleUserPostRequestThread 
    extends BaseTestThread {
    
    /**
     * A glodal reference to the current HTTP connection.
     * @see HTTPClient.HTTPConnection
     **/
    private HTTPConnection httpConnection = null;

    /**
     * A glodal reference to the current HTTP response.
     * @see HTTPClient.HTTPResponse
     **/
    private HTTPResponse httpResponse = null;

    /**
     * Constructor initializes all member variables for this thread by calling
     * super().
     *
     * @param id
     *        A String identification for this thread.
     * @param params
     *        The thread's configuration parameters; can not be null.
     * @param threadResultsCollector
     *        The result collector for this thread; can not be null.
     * @exception IllegalArgumentException
     *            Thrown if a parameter does not meet the above constraints.
     **/
    public SingleUserPostRequestThread(String id, 
				HttpTestClientConfig params,
				ThreadResultsCollector threadResultsCollector)
	throws IllegalArgumentException {
	super(id, params, threadResultsCollector);
    }
    
    /**
     * Overloaded method to properly initialize the golden response method
     * attribute.
     *
     * @return boolean
     *         The success state of set function.
     **/
    protected boolean setGoldenResponseMethod() {
	goldenResponse.method = HttpServerTestUtils.POST_REQUEST_TAG;
	return true;
    }

    /**
     * This function reads in all specified file uploads and computes an
     * MD5 Hash Key for each file.  These hash keys are are used to validate
     * that the server properly received all the files and is able to compute
     * and return the same keys in response.
     *
     * @return boolean
     *         The success state of set function.
     **/
    protected boolean setGoldenResponseMD5HashKey() {
	if (params.fileUploads == null ||
	    params.fileUploads.size() == 0) {
	    return super.setGoldenResponseMD5HashKey();
	} else {
	    Enumeration files = params.fileUploads.elements();
	    while (files.hasMoreElements()) {
		String currentFile = (String)files.nextElement();
		String hashKey = null;
		try {
		    hashKey = 
			HttpServerTestUtils.calculateFileHash(currentFile);
		} catch (Exception e) {
		    threadResultsCollector.assert(false, "ERROR: Unable to compute hash key for file.  Filename: " + currentFile + " -- Exception: " + e, this);
		    return false;
		}
		if (hashKey == null) {
		    threadResultsCollector.assert(false, "ERROR: Unable to open file for reading.  Filename: " + currentFile, this);
		    return false;
		}
		goldenResponse.md5HashKey.addElement(hashKey);
	    }
	    return true;
	}
    }

    /**
     * This function constructs a URL based on the configuration parameters
     * for this thread.  The URL is formatted for a POST method request and
     * a connection is established to the server.  All CGI arguments, CGI
     * parameter, header, cookie, and file uploads are sent and the response
     * object is initialized.
     *
     * @param requestCnt
     *        An int used to create a unique context for the new connection.
     * @exception Exception
     *            Thrown if an error is encountered during this process.  All
     *            errors will be logged in the run() method of this thread.
     **/
    protected void sendRequest(int requestCnt) 
	throws Exception {

	try {
	    String uri = HttpTestClientUtils.getURI(params);
	    NVPair[] cgiArgs = HttpTestClientUtils.getCgiArgs(params);
	    NVPair[] custHeaders = HttpTestClientUtils.getHeaders(params);
	    HttpTestClientUtils.addCookies(httpConnection, params);

	    if (params.fileUploads == null ||
		params.fileUploads.size() == 0) {
		httpResponse = httpConnection.Post(uri,
		 				   cgiArgs,
			 			   custHeaders);
	    } else {
		NVPair[] files = new NVPair[params.fileUploads.size()];
		for (int i = 0; i < params.fileUploads.size(); i++) {
		    files[i] = new NVPair(HttpServerTestUtils.POST_FILE_UPLOAD_PREFIX + i, 
					  (String)params.fileUploads.elementAt(i));
		}
		NVPair[] boundaryHeader = new NVPair[1];
		byte[] requestBytes = Codecs.mpFormDataEncode(cgiArgs,
							      files,
							      boundaryHeader);
		NVPair[] headers = new NVPair[boundaryHeader.length +
					     custHeaders.length];
		int i = 0;
		for (; i < boundaryHeader.length; i++) {
		    headers[i] = boundaryHeader[i];
		}
		for (int j = 0; j < custHeaders.length; j++) {
		    headers[i++] = custHeaders[j];
		}
		
		httpResponse = httpConnection.Post(uri,
						   requestBytes,
						   headers);
	    }
	} catch (ModuleException e) {
	    threadResultsCollector.assert(false, "ERROR: ModuleException while issueing request: " + e, this);
	    throw e;
	} catch (IOException e) {
	    threadResultsCollector.assert(false, "ERROR: IOException while establishing connection: " + e, this);
	    throw e;
	}
    }

    /**
     * This function receives the open connection extablished in the 
     * sendRequest() method and reads the response sent from the server.
     * The response is returned as a string and is expected to be an XML
     * structure that is valid against HttpResponseObject.dtd.
     *
     * @return String
     *         The response string received from the server.
     * @exception Exception
     *            Thrown if an error is encountered during this process.  All
     *            errors will be logged in the run() method of this thread.
     **/
    protected String readResponse() 
	throws Exception {
	
	byte[] responseData = null;
	// Begin reading from response
	try {
	    if (httpResponse.getStatusCode() >= 300) {
		threadResultsCollector.assert(false, "Response Error: " + httpResponse.getReasonLine(), this);
		throw new Exception();
	    } else {
		responseData = httpResponse.getData();
	    }
	} catch (ModuleException e) {
	    threadResultsCollector.assert(false, "ERROR: ModuleException while reading response: " + e, this);
	    throw e;
	} catch (IOException e) {
	    threadResultsCollector.assert(false, "ERROR: IOException while reading response: " + e, this);
	    throw e;
	}

	return new String(responseData);
    }
    
    /**
     * The run method for this thread executes X number of requests that
     * delay for Y milliseconds between each request (where X and Y are
     * specified through configuration parameters).
     * @see org.enhydra.httpServerTest.client.HttpTestClientConfig
     *
     * For each request made, the time for completion is logged to System.out
     * or any error are reported to the thread results collector.
     * @see org.enhydra.httpServerTest.client.ThreadResultsCollector
     *
     * Each response received is validated agains the goldenResponse and
     * any errors are reported to the thread results collector.  An error
     * of this type will not cause this thread to exit but will be logged
     * as an asert error.  An error that occures while establishing or
     * reading from the connection will cause this thread to exit, the
     * error will also be logged as an assert.  Failure to initialize the
     * golden response will also cause this thread to exit.
     **/
    public void run() {
	boolean initSuccess = initGoldenResponse();
	if (!initSuccess) {
	    return;
	}

	if (!params.protocol.equals(HttpTestClientConfig.DEFAULT_PROTOCOL)) {
	    HTTPConnection.setKeyStorePassword(params.keyStorePassword);
	    HTTPConnection.setKeyStoreLocation(params.keyStoreLocation);
	    HTTPConnection.addCustomX509TrustManager();
	}

	try {
	    httpConnection = new HTTPConnection(params.protocol,
						params.host,
						params.port);
	} catch (ProtocolNotSuppException e) {
	    threadResultsCollector.assert(false, 
					  "Invalid protocol specified: "
					  + e,  this);
	    return;
	}

	Long contextTime = new Long(System.currentTimeMillis());
	String context = contextTime.toString() + 
	    "_" + getId();
	httpConnection.setContext(context);
	httpConnection.setAllowUserInteraction(false);
	    
	for (int i = 0; i < params.requestsPerThread; i++) {

	    long requestStartTime = System.currentTimeMillis();
	    String responseString = null;
	    try {
		sendRequest(i);
		responseString = readResponse();
	    } catch (Exception e) {
		goldenResponse = null;
		httpConnection = null;
		httpResponse = null;
		return;
	    }

	    System.out.println("SingleUserPostRequestThread ID:\"" + 
			       getId() + "\" Request Number: \"" +
			       i + "\" received response in: \"" +
			       (System.currentTimeMillis() - 
				requestStartTime) + " msec\"");
	    
	    StringReader reader = new StringReader(responseString);
	    InputSource inputSource = new InputSource(reader);

	    HttpResponseObject httpResponseStruct = null;
	    try {
		httpResponseStruct = 
		    HttpResponseObjectParser.parse(inputSource);
	    } catch (Exception e) {
		threadResultsCollector.assert(false, "Error parsing response: " + e,  this);
		continue;
	    }

	    if (i == 0) {
		HttpTestClientUtils.populateSessionInfo(httpConnection, 
							params,
							goldenResponse,
							httpResponseStruct);
		HttpTestClientUtils.populatePathTranslated(params,
							   goldenResponse,
							   httpResponseStruct);
	    }
	    HttpTestClientUtils.removeCookies(httpConnection);

	    if (!httpResponseStruct.equals(goldenResponse)) {
		threadResultsCollector.assert(false, "Invalid response received:\n\t" + httpResponseStruct.toString() + "\nGolden Response:\n\t" + goldenResponse.toString(), this);
	    }

	    try {
		Thread.sleep(params.delayBetweenRequests);
	    } catch (InterruptedException e) {
		threadResultsCollector.assert(false, "InterruptedException caught while sleeping between requests: " + e, this);
	    } catch (SecurityException e) {
		threadResultsCollector.assert(false, "SecurityException caught while resuming thread execution: " + e, this);
	    }
	    
	}
	goldenResponse = null;
	httpConnection = null;
	httpResponse = null;
    }
}
