/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HttpResponseObject.java,v 1.3.2.1.2.1 2000/10/19 17:59:08 jasona Exp $
 */

package org.enhydra.httpServerTest.common;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;
import java.util.StringTokenizer;

/**
 * This class is the data structure that is shared between the client and 
 * server tests.  The life cycle of this class begins when a client thread's
 * run() method is invoked.  The first step in each of the 
 * SingleUserGetRequestThread and SingleUserPostRequestThread is to initialize
 * a golden response object of this type from the thread's supplied 
 * configuration data.  On success of this initialization, each of these 
 * threads construct an HTTP request also based on its configuration data.
 * The request is sent to the server which then generates a new object of
 * this type populated with all the request information sent from the client.
 * This object's toString() method is then called by the server and the 
 * output, a valid XML structure based on the data populated by the server and
 * validated againt HttpResponseObject.dtd, is returned to the client.  The
 * client then parses the response (the XML structure) and does an equal()
 * comparison between the golden response object initialized earlier and the
 * resulting object (from parsing the XML response).  If these objects are
 * equal, the process begins again starting with issueing a new request.  If
 * these objects are not equal, an assertion is logged by the client thread
 * and, again, the process is repeated with issueing a new request.
 *
 * This structure represents the data returned from the servlet request api
 * and the relationship between its getter methods and the data members held
 * here pretty much match one-to-one.
 * @see org.enhydra.httpServerTest.common.HttpResponseObjectParser
 * @see org.enhydra.httpServerTest.server.application.presentation.BasicTest
 * @see org.enhydra.httpServerTest.client.BaseTestThread
 * @see org.enhydra.httpServerTest.client.SingleUserGetRequestThread
 * @see org.enhydra.httpServerTest.client.SingleUserPostRequestThread
 * 
 * @author Mike Ward
 **/
public class HttpResponseObject {

    /**
     * Static XML header tag.
     * @see HttpResponseObject.dtd for details
     **/
    public static final String XML_HEADER_STRING = "<?xml version=\"1.0\"?>";

    /**
     * This holds the location of the DTD used to validate the XML structure
     * returned by this class's toString() method.  This value should be set
     * before calling toString().
     **/
    public static String HTTP_RESPONSE_OBJECT_DTD = null;

    /**
     * Static XML document type tag.  This is the DTD reference used for 
     * validity checks.
     * @see HttpResponseObject.dtd for details
     **/
    public static String XML_DOCTYPE_STRING_BEGIN = 
	"<!DOCTYPE HttpResponseObject SYSTEM \"";
    public static String XML_DOCTYPE_STRING_END = "\">";

    /**
     * Thes are common syntax tags used for constructing XML elements.
     * @see HttpResponseObject.dtd for details
     **/
    public static final String BEGIN_ELEMENT_TAG = "<";
    public static final String END_ELEMENT_TAG = "</";
    public static final String END_EMPTY_ELEMENT_TAG = "/>";
    public static final String END_PCDATA_ELEMENT_TAG = ">";

    /**
     * Static HttpResponseObject element tag.
     * @see HttpResponseObject.dtd for details
     **/
    public static final String HTTP_RESPONSE_OBJ_ELEMENT = 
	"HttpResponseObject";

    /**
     * Static HttpRequestIngo element tag and its attributes.
     * @see HttpResponseObject.dtd for details
     **/
    public static final String HTTP_REQUEST_INFO_ELEMENT = 
	"HttpRequestInfo";
    public static final String HTTP_REQUEST_INFO_AUTH_TYPE_ATTR = 
	"authType";
    public static final String HTTP_REQUEST_INFO_METHOD_ATTR = 
	"method";
    public static final String HTTP_REQUEST_INFO_PATH_INFO_ATTR = 
	"pathInfo";
    public static final String HTTP_REQUEST_INFO_PATH_TRANSLATED_ATTR = 
	"pathTranslated";
    public static final String HTTP_REQUEST_INFO_QUERY_STRING_ATTR = 
	"queryString";
    public static final String HTTP_REQUEST_INFO_REMOTE_USER_ATTR = 
	"remoteUser";
    public static final String HTTP_REQUEST_INFO_REQUEST_URI_ATTR = 
	"requestURI";
    public static final String HTTP_REQUEST_INFO_SERVLET_PATH_ATTR = 
	"servletPath";
    public static final String HTTP_REQUEST_INFO_SESSION_ATTR = 
	"session";
    public static final String HTTP_REQUEST_INFO_SESSION_ID_FROM_COOKIE_ATTR = 
	"sessionIdFromCookie";
    public static final String HTTP_REQUEST_INFO_SESSION_ID_FROM_URL_ATTR =
	"sessionIdFromURL";
    public static final String HTTP_REQUEST_INFO_SESSION_ID_INVALID_ATTR = 
	"sessionIdInvalid";

    /**
     * Static CgiParam element tag and its attributes.
     * @see HttpResponseObject.dtd for details
     **/
    public static final String CGI_PARAM_ELEMENT = "CgiParam";
    public static final String CGI_PARAM_NAME_ATTR = "name";
    public static final String CGI_PARAM_VALUE_ATTR = "value";

    /**
     * Static CgiArg element tag and its attributes.
     * @see HttpResponseObject.dtd for details
     **/
    public static final String CGI_ARG_ELEMENT = "CgiArg";
    public static final String CGI_ARG_NAME_ATTR = "name";
    public static final String CGI_ARG_VALUE_ATTR = "value";

    /**
     * Static Header element tag and its attributes.
     * @see HttpResponseObject.dtd for details
     **/
    public static final String HEADER_ELEMENT = "Header";
    public static final String HEADER_NAME_ATTR = "name";
    public static final String HEADER_VALUE_ATTR = "value";

    /**
     * Static Cookie element tag and its attributes.
     * @see HttpResponseObject.dtd for details
     **/
    public static final String COOKIE_ELEMENT = "Cookie";
    public static final String COOKIE_NAME_ATTR = "name";
    public static final String COOKIE_VALUE_ATTR = "value";

    /**
     * Static FileUpload element tag and its attributes.
     * @see HttpResponseObject.dtd for details
     **/
    public static final String FILE_UPLOAD_ELEMENT = "FileUpload";
    public static final String FILE_UPLOAD_MD5_HASH_KEY_ATTR = "md5HashKey";

    /**
     * Static default or implied values for element attributes.
     * @see HttpResponseObject.dtd for details
     **/
    public static final String DEFAULT_SESSION = null;

    public static final boolean DEFAULT_SEESION_ID_FROM_COOKIE = false;
    public static final boolean DEFAULT_SESSION_ID_FROM_URL = false;
    public static final boolean DEFAULT_SESSION_ID_INVALID = false;

    /**
     * Response object variables.
     **/
    public String authType = null;
    public String method = null;
    public String pathInfo = null;
    public String pathTranslated = null;
    public String queryString = null;
    public String remoteUser = null;
    public String requestURI = null;
    public String servletPath = null;
    public String session = DEFAULT_SESSION;
    public boolean sessionIdFromCookie = DEFAULT_SEESION_ID_FROM_COOKIE;
    public boolean sessionIdFromURL = DEFAULT_SESSION_ID_FROM_URL;
    public boolean sessionIdInvalid = DEFAULT_SESSION_ID_INVALID;

    public Hashtable cgiParams = new Hashtable();
    public Hashtable cgiArgs = new Hashtable();
    public Hashtable headers = new Hashtable();
    public Hashtable cookies = new Hashtable();
    public Vector md5HashKey = new Vector();


    /**
     * This function overrides the default toString method to generate a 
     * valid XML structure based on data held within this object.  The
     * generated structure is valid against HttpResponseObject.dtd and
     * is the stream that is sent back from the server to the client.
     *
     * @return String
     *         The XML string representing this object.
     **/
    public String toString() {
	StringBuffer xmlBuffer = new StringBuffer();
	
	String[] cgiParamStrings = getCgiParamStrings();
	for (int i = 0; i < cgiParamStrings.length; i++) {
	    xmlBuffer.append(cgiParamStrings[i]);
	}

	String[] cgiArgStrings = getCgiArgStrings();
	for (int i = 0; i < cgiArgStrings.length; i++) {
	    xmlBuffer.append(cgiArgStrings[i]);
	}

	String[] headerStrings = getHeaderStrings();
	for (int i = 0; i < headerStrings.length; i++) {
	    xmlBuffer.append(headerStrings[i]);
	}

	String[] cookieStrings = getCookieStrings();
	for (int i = 0; i < cookieStrings.length; i++) {
	    xmlBuffer.append(cookieStrings[i]);
	}

	String[] md5Strings = getFileUploadStrings();
	for (int i = 0; i < md5Strings.length; i++) {
	    xmlBuffer.append(md5Strings[i]);
	}

	String[] httpRequestInfoStrings = getHttpRequestInfoStrings();
	xmlBuffer.insert(0, httpRequestInfoStrings[0]);
	xmlBuffer.append(httpRequestInfoStrings[1]);

	String[] httpResponseObjectStrings = getHttpResponseObjectStrings();
	xmlBuffer.insert(0, httpResponseObjectStrings[0]);
	xmlBuffer.append(httpResponseObjectStrings[1]);

	xmlBuffer.insert(0, XML_DOCTYPE_STRING_END);
	xmlBuffer.insert(0, HTTP_RESPONSE_OBJECT_DTD);
	xmlBuffer.insert(0, XML_DOCTYPE_STRING_BEGIN);

	xmlBuffer.insert(0, XML_HEADER_STRING);

	return xmlBuffer.toString();
    }

    /**
     * This function overrides the default equals method to provide a detailed
     * comparison of each data member held within this object.  If any 
     * differences are found, a 'false' value is immediately returned.  If
     * all data members are determined to be equal, a 'true' value is
     * returned.
     *
     * @param object
     *        The object to compare against.
     * @return boolean
     *         The equality status of all data members.
     **/
    public boolean equals(Object object) {

	if (object == null) {
	    return false;
	}

	if (!(object instanceof HttpResponseObject)) {
	    return false;
	}

	HttpResponseObject httpResponseObject = (HttpResponseObject)object;
	
	if (!testStringNulls(authType, httpResponseObject.authType)) {
	    if (authType == null ||
		httpResponseObject.authType == null ||
		!authType.equals(httpResponseObject.authType)) {
		return false;
	    }
	}
	if (!testStringNulls(method, httpResponseObject.method)) {
	    if (method == null ||
		httpResponseObject.method == null ||
		!method.equals(httpResponseObject.method)) {
		return false;
	    }
	}
	if (!testStringNulls(pathInfo, httpResponseObject.pathInfo)) {
	    if (pathInfo == null ||
		httpResponseObject.pathInfo == null ||
		!testPathInfo(pathInfo, httpResponseObject.pathInfo)) {
		return false;
	    }
	}
	if (!testStringNulls(pathTranslated, httpResponseObject.pathTranslated)) {
	    if (pathTranslated == null ||
		httpResponseObject.pathTranslated == null ||
		!pathTranslated.equals(httpResponseObject.pathTranslated)) {
		return false;
	    }
	}
	if (!testStringNulls(queryString, httpResponseObject.queryString)) {
	    if (queryString == null ||
		httpResponseObject.queryString == null ||
		!testQueryString(queryString, 
				 httpResponseObject.queryString)) {
		return false;
	    }
	}
	if (!testStringNulls(remoteUser, httpResponseObject.remoteUser)) {
	    if (remoteUser == null ||
		httpResponseObject.remoteUser == null ||
		!remoteUser.equals(httpResponseObject.remoteUser)) {
		return false;
	    }
	}
	if (!testStringNulls(requestURI, httpResponseObject.requestURI)) {
	    if (requestURI == null ||
		httpResponseObject.requestURI == null ||
		!testRequestURI(requestURI, httpResponseObject.requestURI)) {
		return false;
	    }
	}
	if (!testStringNulls(servletPath, httpResponseObject.servletPath)) {
	    if (servletPath == null ||
		httpResponseObject.servletPath == null ||
		!servletPath.equals(httpResponseObject.servletPath)) {
		return false;
	    }
	}
	if (!testStringNulls(session, httpResponseObject.session)) {
	    if (session == null ||
		httpResponseObject.session == null ||
		!session.equals(httpResponseObject.session)) {
		return false;
	    }
	}
	if (!sessionIdFromCookie == httpResponseObject.sessionIdFromCookie) {
	    return false;
	}
	if (!sessionIdFromURL == httpResponseObject.sessionIdFromURL) {
	    return false;
	}
	if (!sessionIdInvalid == httpResponseObject.sessionIdInvalid) {
	    return false;
	}
	if (!(cgiParams == null && httpResponseObject.cgiParams == null)) {
	    if (!testHashtables(cgiParams, httpResponseObject.cgiParams)) {
		return false;
	    }
	}
	if (!(cgiArgs == null && httpResponseObject.cgiArgs == null)) {
	    if (!testHashtables(cgiArgs, httpResponseObject.cgiArgs)) {
		return false;
	    }
	}
	if (!(headers == null && httpResponseObject.headers == null)) {
	    if (!testHashtables(headers, httpResponseObject.headers)) {
		return false;
	    }
	}
	if (!(cookies == null && httpResponseObject.cookies == null)) {
	    if (!testHashtables(cookies, httpResponseObject.cookies)) {
		return false;
	    }
	}
	if (!(md5HashKey == null && httpResponseObject.md5HashKey == null)) {
	    if (!testVectors(md5HashKey, httpResponseObject.md5HashKey)) {
		return false;
	    }
	}
	return true;
    }

    /**
     * This is a convenience function fro testing the equality of
     * two possibly null strings.  This function considers all strings
     * actaully equal to null, empty strings, and strings holding the value
     * 'null' to be considered null.  The purpose of this function is to 
     * deal with the case that a string, when sent over an http connection,
     * often gets the string value of 'null' applied to it when parsed by
     * the SAX parser.
     *
     * @param string1
     *        The first string used for testing.
     * @param string2
     *        The second string used for testing.
     * @return boolean
     *         The result of testing the two strings for nulls.  Returns true
     *         if both strings are determined to be null.
     **/
    public boolean testStringNulls(String string1,
				   String string2) {
	if (string1 == null &&
	    string2 == null) {
	    return true;
	} else if (string1 == null &&
		   (string2.equals("null") ||
		    string2.equals(""))) {
	    return true;
	} else if (string2 == null &&
		   (string1.equals("null") ||
		    string1.equals(""))) {
	    return true;
	} else {
	    return false;
	}
    }

    /**
     * This is a conviencce function to test the equality of two hashtables.
     * This function will consider a null hashtable and an empty hashtable
     * equal.  If both the supplied hashtables are populated, this function
     * will loop through all keys in one hashtable and check that the second
     * hashtable contains each key and that each key has the same value
     * associated with it.
     *
     * @param hashTable1
     *        The first hashtable used for testing.
     * @param hashTable2
     *        The second hashtable used for testing.
     * @return boolean
     *         The result of the two hashtable test.  Returns true if both
     *         hashtables are determined to be equal, false otherwise.
     **/
    public boolean testHashtables(Hashtable hashTable1,
				  Hashtable hashTable2) {
	if (hashTable1 != null &&
	    hashTable1.size() == 0 &&
	    hashTable2 == null) {
	    return true;
	} else if (hashTable2 != null &&
		   hashTable2.size() == 0 &&
		   hashTable1 == null) {
	    return true;
	} else if (hashTable1.size() != hashTable2.size()) {
	    return false;
	}

	Enumeration hashTable1Keys = hashTable1.keys();
	while (hashTable1Keys.hasMoreElements()) {
	    String key = (String)hashTable1Keys.nextElement();
	    if (!hashTable2.containsKey(key)) {
		return false;
	    } else if (!(hashTable2.get(key)).equals(hashTable2.get(key))) {
		return false;
	    }	       
	}
	return true;
    }

    /**
     * This is a conviencce function to test the equality of two vectors.
     * This function will consider a null vector and an empty vector
     * equal.  If both the supplied vectors are populated, this function
     * will loop through all objects held in one vector and check that the
     * second vector contains each object.
     *
     * @param vector1
     *        The first vector used for testing.
     * @param vector2
     *        The second vector used for testing.
     * @return boolean
     *         The result of the two vector test.  Returns true if both
     *         vectors are determined to be equal, false otherwise.
     **/
    public boolean testVectors(Vector vector1,
			       Vector vector2) {
	if (vector1 != null &&
	    vector1.size() == 0 &&
	    vector2 == null) {
	    return true;
	} else if (vector2 != null &&
		   vector2.size() == 0 &&
		   vector1 == null) {
	    return true;
	} else if (vector1.size() != vector2.size()) {
	    return false;
	}

	Enumeration vector1Elements = vector1.elements();
	while (vector1Elements.hasMoreElements()) {
	    String element = (String)vector1Elements.nextElement();
	    if (!vector2.contains(element)) {
		return false;
	    }
	}
	return true;
    }

    /**
     * This is a convenience function to split the path information strings
     * into a token set and then check that each token in the first string
     * is also in the second string.  Any mismatch will return false.
     *
     * @param pathInfoString1
     *        The first path information string used for testing.
     * @param pathInfoString2
     *        The second path information string used for testing.
     * @param boolean
     *         The result of the two path information string test.  Returns 
     *         true if both strings are determined to be equal, false 
     *         otherwise.
     **/
    public boolean testPathInfo(String pathInfoString1,
				String pathInfoString2) {
	StringTokenizer pathInfoStringTokens1 = 
	    new StringTokenizer(pathInfoString1, ";", false);
	StringTokenizer pathInfoStringTokens2 = 
	    new StringTokenizer(pathInfoString2, ";", false);

	return testStringTokens(pathInfoStringTokens1,
				pathInfoStringTokens2);
    }

    /**
     * This is a convenience function to split the request URI strings
     * into a token set and then check that each token in the first string
     * is also in the second string.  Any mismatch will return false.
     *
     * @param uriString1
     *        The first request URI string used for testing.
     * @param uriString2
     *        The second request URI string used for testing.
     * @param boolean
     *         The result of the two request URI string test.  Returns 
     *         true if both strings are determined to be equal, false 
     *         otherwise.
     **/
    public boolean testRequestURI(String uriString1,
				  String uriString2) {
	StringTokenizer uriStringTokens1 = 
	    new StringTokenizer(uriString1, ";", false);
	StringTokenizer uriStringTokens2 = 
	    new StringTokenizer(uriString2, ";", false);

	return testStringTokens(uriStringTokens1,
				uriStringTokens2);
    }

    /**
     * This is a convenience function to split the query string strings
     * into a token set and then check that each token in the first string
     * is also in the second string.  Any mismatch will return false.
     *
     * @param queryString1
     *        The first query string string used for testing.
     * @param queryString2
     *        The second query string string used for testing.
     * @param boolean
     *         The result of the two query strings string test.  Returns 
     *         true if both strings are determined to be equal, false 
     *         otherwise.
     **/
    public boolean testQueryString(String queryString1,
				   String queryString2) {

	StringTokenizer queryStringTokens1 = 
	    new StringTokenizer(queryString1, "%26", false);
	StringTokenizer queryStringTokens2 = 
	    new StringTokenizer(queryString2, "%26", false);

	return testStringTokens(queryStringTokens1,
				queryStringTokens2);
    }

    /**
     * This is a conviencce function to test the equality of two token sets.
     * If both the supplied token sets are populated, this function
     * will loop through all objects held in one token set and check that the
     * second token set contains each object.
     *
     * @param tokens1
     *        The first token set used for testing.
     * @param tokens2
     *        The second token set used for testing.
     * @return boolean
     *         The result of the two token sets test.  Returns true if both
     *         token sets are determined to be equal, false otherwise.
     **/
    public boolean testStringTokens(StringTokenizer tokens1,
				    StringTokenizer tokens2) {
	if (tokens1.countTokens() != 
	    tokens2.countTokens()) {
	    return false;
	}
	
	String[] array1 = 
	    new String[tokens1.countTokens()];
	String[] array2 = 
	    new String[tokens2.countTokens()];
	int i = 0;
	int j = 0;
	while(tokens1.hasMoreTokens()) {
	    array1[i] = tokens1.nextToken();
	    array2[i] = tokens2.nextToken();
	    i++;
	}

	for (i = 0; i < array1.length; i++) {
	    for (j = 0; j < array2.length; j++) {
		if (array1[i].equals(array2[j])) {
		    break;
		}
	    }
	    if (j == array2.length) {
		return false;
	    }
	}
	return true;
    }

    /**
     * This is a convenience function to format and return valid XML 
     * representing all the CGI parameters held in this object.  The XML
     * returned is defined in HttpResponseObject.dtd.
     *
     * @return String[]
     *         An array of CGI parameters in an XML format.
     **/
    public String[] getCgiParamStrings() {
	if (cgiParams == null) {
	    return new String[0];
	}

	String[] cgiParamStrings = new String[cgiParams.size()];

	Enumeration keys = cgiParams.keys();
	int i = 0;
	while (keys.hasMoreElements()) {
	    String currentKey = (String)keys.nextElement();
	    cgiParamStrings[i++] = 
		BEGIN_ELEMENT_TAG +
		CGI_PARAM_ELEMENT + " " +
		CGI_PARAM_NAME_ATTR + "=\"" + currentKey + "\" " +
		CGI_PARAM_VALUE_ATTR + "=\"" + 
		cgiParams.get(currentKey) + "\"" +
		END_EMPTY_ELEMENT_TAG;
	}
	return cgiParamStrings;
    }

    /**
     * This is a convenience function to format and return valid XML 
     * representing all the CGI arguments held in this object.  The XML
     * returned is defined in HttpResponseObject.dtd.
     *
     * @return String[]
     *         An array of CGI arguments in an XML format.
     **/
    public String[] getCgiArgStrings() {
	if (cgiArgs == null) {
	    return new String[0];
	}

	String[] cgiArgStrings = new String[cgiArgs.size()];

	Enumeration keys = cgiArgs.keys();
	int i = 0;
	while (keys.hasMoreElements()) {
	    String currentKey = (String)keys.nextElement();
	    cgiArgStrings[i++] = 
		BEGIN_ELEMENT_TAG +
		CGI_ARG_ELEMENT + " " +
		CGI_ARG_NAME_ATTR + "=\"" + currentKey + "\" " +
		CGI_ARG_VALUE_ATTR + "=\"" + 
		cgiArgs.get(currentKey) + "\"" +
		END_EMPTY_ELEMENT_TAG;
	}
	return cgiArgStrings;
    }

    /**
     * This is a convenience function to format and return valid XML 
     * representing all the headers held in this object.  The XML
     * returned is defined in HttpResponseObject.dtd.
     *
     * @return String[]
     *         An array of headers in an XML format.
     **/
    public String[] getHeaderStrings() {
	if (headers == null) {
	    return new String[0];
	}

	String[] headerStrings = new String[headers.size()];

	Enumeration keys = headers.keys();
	int i = 0;
	while (keys.hasMoreElements()) {
	    String currentKey = (String)keys.nextElement();
	    headerStrings[i++] = 
		BEGIN_ELEMENT_TAG +
		HEADER_ELEMENT + " " +
		HEADER_NAME_ATTR + "=\"" + currentKey + "\" " +
		HEADER_VALUE_ATTR + "=\"" + 
		headers.get(currentKey) + "\"" +
		END_EMPTY_ELEMENT_TAG;
	}
	return headerStrings;
    }
	    
    /**
     * This is a convenience function to format and return valid XML 
     * representing all the cookies held in this object.  The XML
     * returned is defined in HttpResponseObject.dtd.
     *
     * @return String[]
     *         An array of cookies in an XML format.
     **/
    public String[] getCookieStrings() {
	if (cookies == null) {
	    return new String[0];
	}

	String[] cookieStrings = new String[cookies.size()];

	Enumeration keys = cookies.keys();
	int i = 0;
	while (keys.hasMoreElements()) {
	    String currentKey = (String)keys.nextElement();
	    cookieStrings[i++] = 
		BEGIN_ELEMENT_TAG +
		COOKIE_ELEMENT + " " +
		COOKIE_NAME_ATTR + "=\"" + currentKey + "\" " +
		COOKIE_VALUE_ATTR + "=\"" + 
		cookies.get(currentKey) + "\"" +
		END_EMPTY_ELEMENT_TAG;
	}
	return cookieStrings;
    }

    /**
     * This is a convenience function to format and return valid XML 
     * representing all the file uploads held in this object.  The XML
     * returned is defined in HttpResponseObject.dtd.
     *
     * @return String[]
     *         An array of file uploads in an XML format.
     **/
    public String[] getFileUploadStrings() {
	if (md5HashKey == null) {
	    return new String[0];
	}

	String[] md5Strings = new String[md5HashKey.size()];

	Enumeration elements = md5HashKey.elements();
	int i = 0;
	while (elements.hasMoreElements()) {
	    String currentElement = (String)elements.nextElement();
	    md5Strings[i++] = 
		BEGIN_ELEMENT_TAG +
		FILE_UPLOAD_ELEMENT + " " +
		FILE_UPLOAD_MD5_HASH_KEY_ATTR + "=\"" + 
		currentElement + "\"" +
		END_EMPTY_ELEMENT_TAG;
	}
	return md5Strings;
    }
	
    /**
     * This is a convenience function to format and return valid XML 
     * representing the HttpRequestInfo element and all its attributes.
     * The XML returned is defined in HttpResponseObject.dtd.  Position
     * zero in the returned array represents the start tag for this
     * element.  Position one in the returned array represents the stop
     * tag for this element.
     *
     * @return String[]
     *         An array on length two holding the start and stop tags for
     *         this element and all its attributes.
     **/
    public String[] getHttpRequestInfoStrings() {
	String[] httpRequestInfoStrings = new String[2];
	
	StringBuffer requestBuffer = new StringBuffer();
	requestBuffer.append(BEGIN_ELEMENT_TAG +
			     HTTP_REQUEST_INFO_ELEMENT + " ");
	requestBuffer.append(HTTP_REQUEST_INFO_AUTH_TYPE_ATTR + "=\"" +
			     authType + "\" ");
	requestBuffer.append(HTTP_REQUEST_INFO_METHOD_ATTR + "=\"" +
			     method + "\" ");
	requestBuffer.append(HTTP_REQUEST_INFO_PATH_INFO_ATTR + "=\"" +
			     pathInfo + "\" ");
	requestBuffer.append(HTTP_REQUEST_INFO_PATH_TRANSLATED_ATTR + "=\"" +
			     pathTranslated + "\" ");
	requestBuffer.append(HTTP_REQUEST_INFO_QUERY_STRING_ATTR + "=\"" +
			     queryString + "\" ");
	requestBuffer.append(HTTP_REQUEST_INFO_REMOTE_USER_ATTR + "=\"" +
			     remoteUser + "\" ");
	requestBuffer.append(HTTP_REQUEST_INFO_REQUEST_URI_ATTR + "=\"" +
			     requestURI + "\" ");
	requestBuffer.append(HTTP_REQUEST_INFO_SERVLET_PATH_ATTR + "=\"" +
			     servletPath + "\" ");
	requestBuffer.append(HTTP_REQUEST_INFO_SESSION_ATTR + "=\"" +
			     session + "\" ");
	requestBuffer.append(HTTP_REQUEST_INFO_SESSION_ID_FROM_COOKIE_ATTR + 
			     "=\"" + sessionIdFromCookie + "\" ");
	requestBuffer.append(HTTP_REQUEST_INFO_SESSION_ID_FROM_URL_ATTR + 
			     "=\"" + sessionIdFromURL + "\" ");
	requestBuffer.append(HTTP_REQUEST_INFO_SESSION_ID_INVALID_ATTR + 
			     "=\"" + sessionIdInvalid + "\"");
	requestBuffer.append(END_PCDATA_ELEMENT_TAG);

	httpRequestInfoStrings[0] = requestBuffer.toString();
	httpRequestInfoStrings[1] = END_ELEMENT_TAG + 
	    HTTP_REQUEST_INFO_ELEMENT + END_PCDATA_ELEMENT_TAG;

	return httpRequestInfoStrings;
    }

    /**
     * This is a convenience function to format and return valid XML 
     * representing the HttpResponseObject element and all its attributes.
     * The XML returned is defined in HttpResponseObject.dtd.  Position
     * zero in the returned array represents the start tag for this
     * element.  Position one in the returned array represents the stop
     * tag for this element.
     *
     * @return String[]
     *         An array on length two holding the start and stop tags for
     *         this element and all its attributes.
     **/
    public String[] getHttpResponseObjectStrings() {
	String[] httpResponseObjectStrings = new String[2];
	
	httpResponseObjectStrings[0] = BEGIN_ELEMENT_TAG +
	    HTTP_RESPONSE_OBJ_ELEMENT + END_PCDATA_ELEMENT_TAG;
	httpResponseObjectStrings[1] = END_ELEMENT_TAG + 
	    HTTP_RESPONSE_OBJ_ELEMENT + END_PCDATA_ELEMENT_TAG;
	return httpResponseObjectStrings;
    }
}

