/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HttpResponseObjectParser.java,v 1.3.2.1.2.2 2000/10/19 17:59:08 jasona Exp $
 */

package org.enhydra.httpServerTest.common;
                    
import java.util.Hashtable;
import java.util.Vector;
import java.io.File;
import java.io.FileReader;

import org.xml.sax.InputSource;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;

/**
 * This class uses the Xerces SAX parser to parse the http response
 * and populate an HttpResponseObject object.  This class implements a 
 * factory mechanism for retrieving a single response object from the
 * input XML.  Any errors encountered while parsing the XML file or 
 * stream will cause an exception to be thrown.
 * @see org.enhydra.httpServerTest.client.HttpTestClientConfig
 * @see org.apache.xerces.parsers.SAXParser
 *
 * This parser implementation will assume validation against 
 * HttpResponseObject.dtd.
 *
 * @author Mike Ward
 **/
public class HttpResponseObjectParser
    extends DefaultHandler {

    /** 
     * Default parser name.  This value can be supplied if a different SAX
     * parser is desired.
     * @see org.apache.xerces.parsers.SAXParser
     **/
    public static final String DEFAULT_PARSER_NAME = 
	"org.apache.xerces.parsers.SAXParser";

    /**
     * The response object that is populated based on the input XML.
     **/
    private HttpResponseObject httpResponseObject = null;

    /**
     * A boolean flag to indicate if any errors were found while parsing 
     * the XML stream.  This is a type of 'workaround' since some
     * of the SAX callback functions do not allow exceptions to be thrown.
     **/
    private boolean parseErrorFound = false;

    /**
     * The error string associated with a parse error.  This is the final 
     * message that is reported in the Exception thrown from the parse()
     * functions.  This acts as a message buffer used by the 'error()' and 
     * 'fatalError()' callback methods.
     **/
    private String parseErrorMessage = null;

    /**
     * Private constructor.  Prevents object from being instantiated from
     * outside the factory implementation.  The constructor initializes a new
     * HttpResponseObject.
     **/
    private HttpResponseObjectParser() {
	httpResponseObject = new HttpResponseObject();
    }

    /**
     * Static factory method for retrieving a response object using the 
     * default Xerces SAX parser.  This function just calls 
     * 'parse(inputSource, parserName) with the value of 'DEFAULT_PARSER_NAME'.
     *
     * @param uri
     *        The location of the XML configuration file.  This function will
     *        throw an exception if the uri does not begin with 'file:' or
     *        the file is not found.
     * @return HttpResponseObject
     *         The response object populated with data from XML source.
     * @exception Exception
     *            Thrown if an error is encoutered while parsing or the uri
     *            does not meet the above stated specifications.
     **/
    public static HttpResponseObject parse(String uri) 
	throws Exception {
	if (!uri.startsWith("file:")) {
	    throw new Exception("Only a URI of type \"file:\" is currently supported. URI: \"" + uri + "\"");
	}
	File file = new File(uri.substring(uri.indexOf(':') + 1));
	FileReader fileReader = new FileReader(file);
	return parse(new InputSource(fileReader), DEFAULT_PARSER_NAME);
    }

    /**
     * Static factory method for retrieving a response object using the 
     * default Xerces SAX parser.  This function just calls 
     * 'parse(inputSource, parserName) with the value of 'DEFAULT_PARSER_NAME'.
     *
     * @param inputSource
     *        The input source for the XML data to parse.
     * @return HttpResponseObject
     *         The response object populated with data from XML source.
     * @exception Exception
     *            Thrown if an error is encoutered while parsing
     **/
    public static HttpResponseObject parse(InputSource inputSource) 
	throws Exception {
	return parse(inputSource, DEFAULT_PARSER_NAME);
    }
    
    /**
     * Static factory method for retrieving a response object using the 
     * default Xerces SAX parser.  This function instantiates a
     * HttpResponseObjectParser object and registers that instance as the
     * content and error handler for the supplied SAX parser.  Any errors
     * encountered while parsing cause an exception to be thrown.
     *
     * @param inputSource
     *        The input source of the XML data to parse.
     * @param parserName
     *        The supplied class name of a preferred SAX parser.  This will
     *        override the use of the DEFAULT_PARSER_NAME.
     * @return HttpResponseObject
     *         The response object populated with data from XML source.
     * @exception Exception
     *            Thrown if an error is encoutered while parsing.
     **/
    public static HttpResponseObject parse(InputSource inputSource,
					   String parserName) 
	throws Exception {

	HttpResponseObjectParser httpResponseObjectParser = 
	    new HttpResponseObjectParser();

	XMLReader parser = 
	    (XMLReader)Class.forName(parserName).newInstance();
	parser.setContentHandler(httpResponseObjectParser);
	parser.setErrorHandler(httpResponseObjectParser);
	parser.setFeature("http://xml.org/sax/features/validation", true);
	
	parser.parse(inputSource);

	if (httpResponseObjectParser.parseErrorFound) {
	    throw new Exception(httpResponseObjectParser.parseErrorMessage);
	}

	return httpResponseObjectParser.httpResponseObject;
    }

    /** 
     * Callback function invoked when a new element is found.  This function
     * implements a switch based on the element type retrieved.  The possible
     * set of elements can be found in HttpResponseObject as static final
     * Strings matching the pattern <elemenet type>_ELEMENT.
     * 
     * @see org.apache.xerces.parsers.SAXParser
     * @see org.enhydra.httpServerTest.client.HttpResponseObject
     *
     * @param uri
     *        This argument is not used here.
     * @param local
     *        This argument is not used here.
     * @param elementName
     *        The name of the current element.
     * @param attributes
     *        The attributes associated with this element.
     **/
    public void startElement(String uri, 
			     String local, 
			     String elementName, 
			     Attributes attributes) {

	if (elementName.equals(HttpResponseObject.HTTP_REQUEST_INFO_ELEMENT)) {
	    setHttpRequestInfoAttributes(attributes);
	} else if (elementName.equals(HttpResponseObject.CGI_PARAM_ELEMENT)) {
	    setCgiParamAttributes(attributes);
	} else if (elementName.equals(HttpResponseObject.CGI_ARG_ELEMENT)) {
	    setCgiArgAttributes(attributes);
	} else if (elementName.equals(HttpResponseObject.HEADER_ELEMENT)) {
	    setHeaderAttributes(attributes);
	} else if (elementName.equals(HttpResponseObject.COOKIE_ELEMENT)) {
	    setCookieAttributes(attributes);
	} else if (elementName.equals(HttpResponseObject.FILE_UPLOAD_ELEMENT)) {
	    setFileUploadAttributes(attributes);
	}
    }

    /**
     * This function is invoked when an element of type 'HttpRequestInfo' is 
     * encountered.  The list of possible 'HttpRequestInfo' element attributes 
     * can be found in HttpResponseObject as static final Strings matching 
     * the pattern TEST_PARAMS_<attribute name>_ATTR.
     *
     * @see org.enhydra.httpServerTest.client.HttpResponseObject
     *
     * @param attributes
     *        The attributes found with this element.
     **/
    protected void setHttpRequestInfoAttributes(Attributes attributes) {
	for (int i = 0; i < attributes.getLength(); i++) {
	    String attributeName = attributes.getQName(i);
	    if (attributeName.equals(HttpResponseObject.HTTP_REQUEST_INFO_AUTH_TYPE_ATTR)) {
		httpResponseObject.authType = attributes.getValue(i);
	    } else if (attributeName.equals(HttpResponseObject.HTTP_REQUEST_INFO_METHOD_ATTR)) {
		httpResponseObject.method = attributes.getValue(i);
	    } else if (attributeName.equals(HttpResponseObject.HTTP_REQUEST_INFO_PATH_INFO_ATTR)) {
		httpResponseObject.pathInfo = attributes.getValue(i);
	    } else if (attributeName.equals(HttpResponseObject.HTTP_REQUEST_INFO_PATH_TRANSLATED_ATTR)) {
		httpResponseObject.pathTranslated = attributes.getValue(i);
	    } else if (attributeName.equals(HttpResponseObject.HTTP_REQUEST_INFO_QUERY_STRING_ATTR)) {
		httpResponseObject.queryString = attributes.getValue(i);
	    } else if (attributeName.equals(HttpResponseObject.HTTP_REQUEST_INFO_REMOTE_USER_ATTR)) {
		httpResponseObject.remoteUser = attributes.getValue(i);
	    } else if (attributeName.equals(HttpResponseObject.HTTP_REQUEST_INFO_REQUEST_URI_ATTR)) {
		httpResponseObject.requestURI = attributes.getValue(i);
	    } else if (attributeName.equals(HttpResponseObject.HTTP_REQUEST_INFO_SERVLET_PATH_ATTR)) {
		httpResponseObject.servletPath = attributes.getValue(i);
	    } else if (attributeName.equals(HttpResponseObject.HTTP_REQUEST_INFO_SESSION_ATTR)) {
		httpResponseObject.session = attributes.getValue(i);
	    } else if (attributeName.equals(HttpResponseObject.HTTP_REQUEST_INFO_SESSION_ID_FROM_COOKIE_ATTR)) {
		httpResponseObject.sessionIdFromCookie = 
		    getBooleanAttribute(attributes.getValue(i));
	    } else if (attributeName.equals(HttpResponseObject.HTTP_REQUEST_INFO_SESSION_ID_FROM_URL_ATTR)) {
		httpResponseObject.sessionIdFromURL = 
		    getBooleanAttribute(attributes.getValue(i));
	    } else if (attributeName.equals(HttpResponseObject.HTTP_REQUEST_INFO_SESSION_ID_INVALID_ATTR)) {
		httpResponseObject.sessionIdInvalid = 
		    getBooleanAttribute(attributes.getValue(i));
	    }
	}
    }

    /**
     * This function is invoked when an element of type 'CgiParam' is 
     * encountered.  The list of possible 'CgiParam' element attributes 
     * can be found in HttpResponseObject as static final Strings matching 
     * the pattern TEST_PARAMS_<attribute name>_ATTR.
     *
     * @see org.enhydra.httpServerTest.client.HttpResponseObject
     *
     * @param attributes
     *        The attributes found with this element.
     **/
    protected void setCgiParamAttributes(Attributes attributes) {
	if (attributes.getLength() != 2) {
	    return;
	}
	
	String attributeName = null;
	String attributeValue = null;	
	
	for (int i = 0; i < 2; i++) {
	    if ((attributes.getQName(i)).equals(HttpResponseObject.CGI_PARAM_NAME_ATTR)) {
		attributeName = attributes.getValue(i);
	    } else if ((attributes.getQName(i)).equals(HttpResponseObject.CGI_PARAM_VALUE_ATTR)) {
		attributeValue = attributes.getValue(i);
	    }
	}
	httpResponseObject.cgiParams.put(attributeName, attributeValue);
    }

    /**
     * This function is invoked when an element of type 'CgiArg' is 
     * encountered.  The list of possible 'CgiArg' element attributes 
     * can be found in HttpResponseObject as static final Strings matching 
     * the pattern TEST_PARAMS_<attribute name>_ATTR.
     *
     * @see org.enhydra.httpServerTest.client.HttpResponseObject
     *
     * @param attributes
     *        The attributes found with this element.
     **/
    protected void setCgiArgAttributes(Attributes attributes) {
	if (attributes.getLength() != 2) {
	    return;
	}
	
	String attributeName = null;
	String attributeValue = null;	
	
	for (int i = 0; i < 2; i++) {
	    if ((attributes.getQName(i)).equals(HttpResponseObject.CGI_ARG_NAME_ATTR)) {
		attributeName = attributes.getValue(i);
	    } else if ((attributes.getQName(i)).equals(HttpResponseObject.CGI_ARG_VALUE_ATTR)) {
		attributeValue = attributes.getValue(i);
	    }
	}
	httpResponseObject.cgiArgs.put(attributeName, attributeValue);
    }

    /**
     * This function is invoked when an element of type 'Header' is 
     * encountered.  The list of possible 'Header' element attributes 
     * can be found in HttpResponseObject as static final Strings matching 
     * the pattern TEST_PARAMS_<attribute name>_ATTR.
     *
     * @see org.enhydra.httpServerTest.client.HttpResponseObject
     *
     * @param attributes
     *        The attributes found with this element.
     **/
    protected void setHeaderAttributes(Attributes attributes) {
	if (attributes.getLength() != 2) {
	    return;
	}
	
	String attributeName = null;
	String attributeValue = null;	
	
	for (int i = 0; i < 2; i++) {
	    if ((attributes.getQName(i)).equals(HttpResponseObject.HEADER_NAME_ATTR)) {
		attributeName = attributes.getValue(i);
	    } else if ((attributes.getQName(i)).equals(HttpResponseObject.HEADER_VALUE_ATTR)) {
		attributeValue = attributes.getValue(i);
	    }
	}
	httpResponseObject.headers.put(attributeName, attributeValue);
    }

    /**
     * This function is invoked when an element of type 'Cookie' is 
     * encountered.  The list of possible 'Cookie' element attributes 
     * can be found in HttpResponseObject as static final Strings matching 
     * the pattern TEST_PARAMS_<attribute name>_ATTR.
     *
     * @see org.enhydra.httpServerTest.client.HttpResponseObject
     *
     * @param attributes
     *        The attributes found with this element.
     **/
    protected void setCookieAttributes(Attributes attributes) {
	if (attributes.getLength() != 2) {
	    return;
	}
	
	String attributeName = null;
	String attributeValue = null;	
	
	for (int i = 0; i < 2; i++) {
	    if ((attributes.getQName(i)).equals(HttpResponseObject.COOKIE_NAME_ATTR)) {
		attributeName = attributes.getValue(i);
	    } else if ((attributes.getQName(i)).equals(HttpResponseObject.COOKIE_VALUE_ATTR)) {
		attributeValue = attributes.getValue(i);
	    }
	}
	httpResponseObject.cookies.put(attributeName, attributeValue);
    }
    
    /**
     * This function is invoked when an element of type 'FileUpload' is 
     * encountered.  The list of possible 'FileUpload' element attributes 
     * can be found in HttpResponseObject as static final Strings matching 
     * the pattern TEST_PARAMS_<attribute name>_ATTR.
     *
     * @see org.enhydra.httpServerTest.client.HttpResponseObject
     *
     * @param attributes
     *        The attributes found with this element.
     **/
    protected void setFileUploadAttributes(Attributes attributes) {
	for (int i = 0; i < attributes.getLength(); i++) {
	    String attributeName = attributes.getQName(i);
	    if (attributeName.equals(HttpResponseObject.FILE_UPLOAD_MD5_HASH_KEY_ATTR)) {
		httpResponseObject.md5HashKey.addElement(attributes.getValue(i));
	    }
	}
    }

    /**
     * This is a convenience function to retrieve a boolean value from a 
     * String.  False is returned for any string not matching 'true'.
     *
     * @param attributeValue
     *        The string value to convert.
     * @return boolean
     *         The converted string value.
     **/
    protected boolean getBooleanAttribute(String attributeValue) {
	return Boolean.valueOf(attributeValue).booleanValue();
    }
	
    /** 
     * Warning callback function.  Called by the SAX parser if a warning error
     * is encountered while parsing.  This implementation only prints a 
     * warning to standard error and returns.
     *
     * @see org.apache.xerces.parsers.SAXParser
     *
     * @param e
     *        The SAXParseException that was thrown to cause the warning.
     **/
    public void warning(SAXParseException e) {
        System.err.println("WARNING: " +
                           getLocationString(e)+": " +
                           e.getMessage());
    }

    /** 
     * Error callback function.  Called by the SAX parser if a non-fatal error
     * is encountered while parsing.  This function reports the error to the
     * global class variables 'parseErrorFound' and 'parseErrorMessage' to 
     * cause an exception to be thrown from the 'parse()' function.  This
     * parser implementation considers all errors fatal except warnings.  This
     * is done because a test case can not run with an incomplete set of
     * configuration parameters.
     *
     * @see org.apache.xerces.parsers.SAXParser
     *
     * @param e
     *        The SAXParseException that was thrown to cause the error.
     **/
    public void error(SAXParseException e) {
	parseErrorFound = true;
        parseErrorMessage = "ERROR: " +
	    getLocationString(e)+": " +
	    e.getMessage();
    }

    /** 
     * Fatal Error callback function.  Called by the SAX parser if a fatal 
     * error is encountered while parsing.  This function reports the error 
     * to the global class variables 'parseErrorFound' and 'parseErrorMessage'
     * to cause an exception to be thrown from the 'parse()' function.  This
     * parser implementation considers all errors fatal except warnings.  This
     * is done because a test case can not run with an incomplete set of
     * configuration parameters.
     *
     * @see org.apache.xerces.parsers.SAXParser
     *
     * @param e
     *        The SAXParseException that was thrown to cause the fatal error.
     * @exception SAXException
     *            This is not used in this implementation but is required for
     *            the proper method signiture.  The global error flag and
     *            message buffer is used instead.
     **/
    public void fatalError(SAXParseException e) 
	throws SAXException {
	parseErrorFound = true;
        parseErrorMessage = "FATAL ERROR: " +
	    getLocationString(e)+": " +
	    e.getMessage();
    }

    /** 
     * Returns a string representing the line:column that the exception was
     * encountered.  This is only used from error reporting purposes.
     *
     * @param e
     *        The SAXParseException that was thrown to cause the error.
     * @return String
     *         The line:column that the error occured.
     **/
    private String getLocationString(SAXParseException ex) {
        StringBuffer str = new StringBuffer();

        String systemId = ex.getSystemId();
        if (systemId != null) {
            int index = systemId.lastIndexOf('/');
            if (index != -1) 
                systemId = systemId.substring(index + 1);
            str.append(systemId);
        }
        str.append(':');
        str.append(ex.getLineNumber());
        str.append(':');
        str.append(ex.getColumnNumber());

        return str.toString();

    }
}
