/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: BasicTestApp.java,v 1.2.2.1.2.1 2000/10/19 17:59:08 jasona Exp $
 */

package org.enhydra.httpServerTest.server.application;

import com.lutris.appserver.server.ApplicationException;
import com.lutris.appserver.server.StandardApplication;
import com.lutris.appserver.server.httpPresentation.HttpPresentationComms;
import com.lutris.util.Config;
import com.lutris.util.ConfigException;
import com.lutris.util.KeywordValueException;

import java.io.File;
/**
 * This is the basic application class object for the http server test.
 * This class is very limited in its functionality and simply provides 
 * static references to a temporary directory location, setable through the
 * appication configuration file, and a counter.  The temorary directory
 * is used for caching of files that are uploaded.  The counter is used
 * to ensure that each thread of this application, handling uploaded files,
 * will create a unique sub-directory, under the specified temporary 
 * directory, to write to.  This allows each thread to only account for
 * uploaded files that occured on the current request being handled.
 *
 * @author Mike Ward
 **/
public class BasicTestApp extends StandardApplication {

    /**
     * This is the reference key used to lookup the validation DTD for
     * HttpResponseObject.  This value is specified in the application
     * configuration file as 
     * 'Application.HttpResponseObjectDTD = <location of DTD>
     **/
    public static final String HTTP_RESPONSE_OBJECT_DTD = 
	"HttpResponseObjectDTD";
    
    /**
     * This is the reference key used to lookup the temporary base
     * directory in the application configuration file.  The expected
     * key is 'Application.TemporaryBaseDir = <some tmp dir>'.
     **/
    public static final String TEMP_BASE_DIR = "TemporaryBaseDir";
    
    /**
     * This is the reference key used to grab the application specific
     * section of the configuration file.  All application specific
     * configuration variables begin with this tag 'Application.<var name>'.
     **/
    public static final String APPLICATION_TAG = "Application";

    /**
     * This holds the location of the DTD used to validate the 
     * HttpResponseObject when read from the application configuration file.
     **/
    public static String httpResponseObjectDTD = null;

    /**
     * This holds the value of the temporary base direstory when read from
     * the application configuration file.
     **/
    public static String temporaryBaseDir = null;

    /**
     * This is a static counter used by the presentation objects to ensure
     * a unique sub-directory under the specified temporary base directory.
     * The directories created using this count variable are used to cache
     * files uploaded from the client.
     **/
    protected static Long fileCacheDirCnt = new Long(0);

    /**
     * This is the global Config object that is shared by the 
     * whole application. If there was an error reading the file, this
     * will be null.
     *
     * @see com.lutris.util.Config
     **/
    public static Config config = null;

    /**
     * This is the startup function called when this application is loaded.
     * This function calls super.startup(appConfig) and then attempts to
     * read the temporary base directory from the supplied configuration
     * parameters.  If this configuration setting is not found, this function
     * will throw an excepion.
     * 
     * @param appConfig
     *        The configuration object for this application.
     * @exception ApplicationException
     *            Thrown if an error occurs while retrieving the temporary
     *            base directory.
     **/
    public void startup(Config appConfig) 
	throws ApplicationException {
        super.startup(appConfig);
        //  Here is where you would read application-specific settings from
        //  your config file.
        config = appConfig;
	    try {
		Config appConfigSection = config.getConfig(APPLICATION_TAG);
		temporaryBaseDir = appConfigSection.getString(TEMP_BASE_DIR);
		if (!temporaryBaseDir.endsWith(File.pathSeparator)) {
		    temporaryBaseDir += File.separator;
		}
		httpResponseObjectDTD = 
		    appConfigSection.getString(HTTP_RESPONSE_OBJECT_DTD);
	    } catch (ConfigException e) {
		throw new ApplicationException(e.toString());
	    } catch (KeywordValueException e) {
		throw new ApplicationException(e.toString());
	    }
    }

    /**
     * This is a static function to retrieve the next sequence of the conter
     * variable held in this class.  This function synchronizes on the 
     * fileCacheDirCnt and increments this count each time this function is
     * called and then returns the new count.
     *
     * @return long
     *         The current, incremented counter.
     **/
    public static long getFileCacheDirCnt() {
	synchronized (fileCacheDirCnt) { 
	    fileCacheDirCnt = new Long(fileCacheDirCnt.longValue() + (long)1);
	    return fileCacheDirCnt.longValue();
	}
    }

    /**
     * This is an optional function, used only by the Multiserver's graphical
     * administration. This bit of HTML appears in the status page for this
     * application. You could add extra status info, for example
     * a list of currently logged in users.
     *
     * @return HTML that is displayed in the status page of the Multiserver.
     **/
    public String toHtml() {
        return "This is <I>BasicTestApp</I>";
    }
}

