/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: BasicTest.java,v 1.3.2.1.2.1 2000/10/19 17:59:08 jasona Exp $
 */

package org.enhydra.httpServerTest.server.application.presentation;

import org.enhydra.httpServerTest.server.application.BasicTestApp;
import org.enhydra.httpServerTest.common.HttpResponseObject;
import org.enhydra.httpServerTest.common.HttpServerTestUtils;
import org.enhydra.httpServerTest.server.common.FileUploadMangler;

import com.lutris.appserver.server.httpPresentation.HttpPresentationException;

import com.lutris.appserver.server.httpPresentation.HttpPresentation;
import com.lutris.appserver.server.httpPresentation.HttpPresentationRequest;
import com.lutris.appserver.server.httpPresentation.HttpPresentationComms;
import com.lutris.appserver.server.httpPresentation.HttpPresentationException;
import com.lutris.appserver.server.session.SessionException;
import com.lutris.appserver.server.session.SessionManager;
import com.lutris.appserver.server.session.Session;

import HTTPClient.NVPair;
import HTTPClient.Codecs;

import java.util.Random;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.net.URLEncoder;
import java.io.DataInputStream;
import java.io.File;

import javax.servlet.http.Cookie;

/**
 * This class is the generic server side test presentation object that client
 * side test threads talk to.  This class respondes to every request comming in
 * by constructing an HttpResponseObject and returning it in the response as
 * an XML stream.  The response object returned contains information about
 * all data received via the request.  This presentation object properly
 * handles GET and POST method request and multi-part mime encodings.
 *
 * @see org.enhydra.httpServerTest.common.HttpResponseObject
 *
 * @author Mike Ward
 **/
public class BasicTest
    implements HttpPresentation {

    /**
     * This holds the directory, unique to this request, that any uploaded
     * files are stored.
     **/
    private File fileCacheDir = null;

    /**
     * The run method for this presentation object.  This method runs through
     * the remainder of the functions held in this class to initialize the
     * response object based on information pulled from the request.  Once
     * the response object is populated, this method calls toString() on it
     * and returns the generated XML structure to the client.
     *
     * @param comms
     *        The comms objects for this request.  All request, response,
     *        and session information are available from this object.
     **/
    public void run(HttpPresentationComms comms) 
        throws HttpPresentationException {

	HttpPresentationRequest request = comms.request;
	Session session = comms.session;

	HttpResponseObject httpResponseObject = new HttpResponseObject();
	httpResponseObject.HTTP_RESPONSE_OBJECT_DTD = 
	    BasicTestApp.httpResponseObjectDTD;

	populateRequestInfo(httpResponseObject, request);
	if (request.getMethod().equals(HttpServerTestUtils.GET_REQUEST_TAG) ||
	    request.getContentType().equals(HttpServerTestUtils.POST_URL_ENCODED_CONTENT_TYPE)) {
	    // Standard request, url-encoded
	    populateCgiArgsFromGet(httpResponseObject, request);
  	} else {
	    // Multi-part mime request, file uploads expected
  	    populateCgiArgsFromPost(httpResponseObject, request);
  	    populateFileUploadFromPost(httpResponseObject);
	}
	populateCgiParams(httpResponseObject, request);
	populateHeaders(httpResponseObject, request);
	populateCookies(httpResponseObject, request);
	populateSessionInfo(httpResponseObject, comms);
	
        comms.response.writeHTML(httpResponseObject.toString());
    }

    /**
     * This function runs through the HttpPresentationRequest API and 
     * populates the response object with retrieved information.  The 
     * request API used here is a sub-set of the HttpPresentationRequest
     * API that is common with a stardard servlet request object.  Each
     * of the requst method calls are individually wrapped in a try/catch
     * block so that an error thrown from one will not cause the remaining
     * calls to be skipped.
     *
     * @param httpResponseObject
     *        The response object to populate with request data.
     * @param request
     *        The HttpPresentationRequest object to pull data from.
     **/
    protected void populateRequestInfo(HttpResponseObject httpResponseObject,
				       HttpPresentationRequest request) {
	try {
	    httpResponseObject.authType = request.getAuthType();
	} catch (HttpPresentationException e) {
	}

	try {
	    httpResponseObject.method = request.getMethod();
	} catch (HttpPresentationException e) {
	}

	try {
	    httpResponseObject.pathInfo = request.getPathInfo();
	} catch (HttpPresentationException e) {
	}
	
	try {
	    httpResponseObject.pathTranslated = request.getPathTranslated();
	} catch (Exception e) {
	}

	try {
	    String queryString = request.getQueryString();
	    if (queryString != null) {
		httpResponseObject.queryString = 
		    URLEncoder.encode(queryString);
	    } else {
		httpResponseObject.queryString = null;
	    }
	} catch (HttpPresentationException e) {
	}

	try {
	    httpResponseObject.remoteUser = request.getRemoteUser();
	} catch (HttpPresentationException e) {
	}

	try {
	    httpResponseObject.requestURI = request.getRequestURI();
	} catch (HttpPresentationException e) {
	}
	
	try {
	    httpResponseObject.servletPath = request.getApplicationPath();
	} catch (HttpPresentationException e) {
	}
    }

    /**
     * This function will populate the response object with all CGI parameters
     * submitted with the request.
     *
     * @param httpResponseObject
     *        The response object to populate with request data.
     * @param request
     *        The HttpPresentationRequest object to pull data from.
     **/
    protected void populateCgiParams(HttpResponseObject httpResponseObject,
				     HttpPresentationRequest request) {
	StringTokenizer paramTokens = null;
	try {
	    paramTokens = 
		new StringTokenizer(request.getPathInfo(), ";", false);
	} catch (HttpPresentationException e) {
	    return;
	}

	while (paramTokens.hasMoreTokens()) {
	    String currentToken = paramTokens.nextToken();
	    StringTokenizer currentNameValuePair = 
		new StringTokenizer(currentToken, "=", false);
	    if (currentNameValuePair.countTokens() == 2) {
		String paramName = currentNameValuePair.nextToken();
		String paramValue = currentNameValuePair.nextToken();
		httpResponseObject.cgiParams.put(paramName, paramValue);
	    }
	}
    }

    /**
     * This function will populate the response object with all CGI arguments
     * submitted with the request.  This method assumes a stardard GET or
     * url-encoded POST request (ie. no multi-part mime).
     *
     * @param httpResponseObject
     *        The response object to populate with request data.
     * @param request
     *        The HttpPresentationRequest object to pull data from.
     **/
    protected void populateCgiArgsFromGet(HttpResponseObject httpResponseObject,
					  HttpPresentationRequest request) {
	try {
	    Enumeration cgiArgs = request.getParameterNames();
	    while (cgiArgs.hasMoreElements()) {
		String cgiArgName = (String)cgiArgs.nextElement();
		String cgiArgValue = 
		    request.getParameter(cgiArgName);
		httpResponseObject.cgiArgs.put(cgiArgName, cgiArgValue);
	    }
	} catch (HttpPresentationException e) {	    
	}
    }

    /**
     * This function will populate the response object with all http headers
     * submitted with the request.  This function will not include any
     * 'default' headers applied by HTTPConnection.
     *
     * @see HTTPClient.HTTPConnection
     * @see org.enhydra.httpServerTest.common.HttpServerTestUtils
     *
     * @param httpResponseObject
     *        The response object to populate with request data.
     * @param request
     *        The HttpPresentationRequest object to pull data from.
     **/
    protected void populateHeaders(HttpResponseObject httpResponseObject,
				   HttpPresentationRequest request) {
	try {
	    Enumeration headers = request.getHeaderNames();
	    while (headers.hasMoreElements()) {
		String headerName = (String)headers.nextElement();
		if (!HttpServerTestUtils.isDefaultHeader(headerName)) {
		    String headerValue = 
			request.getHeader(headerName);
		    httpResponseObject.headers.put(headerName, headerValue);
		}
	    }
	} catch (HttpPresentationException e) {	    
	}

    }

    /**
     * This function will populate the response object with all http cookies
     * submitted with the request.
     *
     * @param httpResponseObject
     *        The response object to populate with request data.
     * @param request
     *        The HttpPresentationRequest object to pull data from.
     **/
    protected void populateCookies(HttpResponseObject httpResponseObject,
				   HttpPresentationRequest request) {
	try {
	    Cookie[] cookies = request.getCookies();
	    for (int i = 0; i < cookies.length; i++) {
		httpResponseObject.cookies.put(cookies[i].getName(),
					       cookies[i].getValue());
	    }
	} catch (HttpPresentationException e) {
	}
    }

    /**
     * This function will populate the response object with all CGI arguments
     * submitted with the request.  This method assumes a multi-part mime
     * encoding from a POST request.
     *
     * @see HTTPClient.Codecs
     *
     * @param httpResponseObject
     *        The response object to populate with request data.
     * @param request
     *        The HttpPresentationRequest object to pull data from.
     **/
    protected void populateCgiArgsFromPost(HttpResponseObject httpResponseObject,
					  HttpPresentationRequest request) {
	try {
	    byte[] body = new byte[request.getContentLength()];
	    new DataInputStream(request.getInputStream()).readFully(body);
	    String contentTypeAndBoundary = request.getContentType();
	
	    initFileCacheDir();

	    NVPair[] cgiArgs = 
		Codecs.mpFormDataDecode(body, 
					contentTypeAndBoundary,
					fileCacheDir.toString(),
					new FileUploadMangler());
	    for (int i = 0; i < cgiArgs.length; i++) {
		if (!cgiArgs[i].getName().startsWith(HttpServerTestUtils.POST_FILE_UPLOAD_PREFIX)) {
		    httpResponseObject.cgiArgs.put(cgiArgs[i].getName(),
						   cgiArgs[i].getValue());
		}
	    }
	} catch (Exception e) {
	}
    }

    /**
     * This function will continually loop until a unique temporary sub-
     * directory is found.  This is accomplished by using the static data 
     * members and methods held in the BasicTestApp class.  Once a unique
     * path is found, the directories are created and the global fileCacheDir
     * is initialized.
     **/
    protected void initFileCacheDir() {

	String fileCacheDirString = BasicTestApp.temporaryBaseDir + 
	    BasicTestApp.getFileCacheDirCnt() + File.separator;
	fileCacheDir = new File(fileCacheDirString);
	    
	while (true) {
	    if (fileCacheDir.exists()) {
		fileCacheDirString = BasicTestApp.temporaryBaseDir + 
		    BasicTestApp.getFileCacheDirCnt() + File.separator;
		fileCacheDir = new File(fileCacheDirString);
	    } else {
		break;
	    } 
	}
	fileCacheDir.mkdirs();
    }

    /**
     * This function will loop through all files found under the unique 
     * directory created by initFileCacheDir() and compute an MD5 hashkey
     * for each individual file.  The response object is then populated with
     * each hashkey.
     *
     * @param httpResponseObject
     *        The response object to populate with file upload data.
     **/
    protected void populateFileUploadFromPost(HttpResponseObject httpResponseObject) {
	String[] uploadedFiles = fileCacheDir.list();
	for (int i = 0; i < uploadedFiles.length; i++) {
	    String currentFile = fileCacheDir.toString() + 
		File.separator + uploadedFiles[i];
	    String hashKey = null;
	    try {
		hashKey = 
		    HttpServerTestUtils.calculateFileHash(currentFile);
	    } catch (Exception e) {
	    }
	    if (!(hashKey == null)) {
		httpResponseObject.md5HashKey.addElement(hashKey);
	    }
	}
	cleanupFiles();
    }

    /**
     * This function is called after all file MD5 hashkeys have been computed.
     * This function will recursively delete all uploaded files and temporary
     * directories created while caching uploaded files.
     **/
    public void cleanupFiles() {
	String[] uploadedFiles = fileCacheDir.list();
	for (int i = 0; i < uploadedFiles.length; i++) {
	    String currentFile = fileCacheDir.toString() + 
		File.separator + uploadedFiles[i];
	    File tmpFile = new File(currentFile);
	    tmpFile.delete();
	}
	fileCacheDir.delete();
    }

    /**
     * This function will populate the response object with all session
     * information submitted with the request.
     *
     * @param httpResponseObject
     *        The response object to populate with request data.
     * @param request
     *        The HttpPresentationRequest object to pull data from.
     **/
    protected void populateSessionInfo(HttpResponseObject httpResponseObject,
				       HttpPresentationComms comms) {
	    
	try {
	    httpResponseObject.sessionIdFromCookie = 
		comms.request.isRequestedSessionIdFromCookie();
	    httpResponseObject.sessionIdFromURL = 
		comms.request.isRequestedSessionIdFromUrl();
	} catch (HttpPresentationException e) {
	}

	if (httpResponseObject.sessionIdFromCookie ||
	    httpResponseObject.sessionIdFromURL) {
	    httpResponseObject.sessionIdInvalid = false;
	    httpResponseObject.session = comms.session.getSessionKey();
	} else {
	    httpResponseObject.sessionIdInvalid = true;
	    httpResponseObject.session = HttpResponseObject.DEFAULT_SESSION;
	}
    }
}
