/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: FileLines.java,v 1.7.12.1 2000/10/19 17:59:08 jasona Exp $
 */





package com.lutris.appserver.server.jolt.joltc;

import java.util.*;
import java.io.*;

/**
 * Class used to assemble file of generated code.
 * Handles tracking of filename/linenumbers and doing indentation.
 */
class FileLines {
    /*
     * Amount to adjust indent by.
     */
    private static final int INDENT_INCR = 4;
    
    private StringBuffer code = new StringBuffer();
    private int currentIndent = 0;    // Current indentation.
    private String indentPrefix = "";
    private Vector lines = new Vector();

    // Previous location is used when append without a file name/line num
    // being specified.
    FileLine prevLine;

    /**
     * Construction a new file lines buffer.
     */
    public FileLines(String startingFile) {
        setIndentPrefix();

        // Use starting file for lines before first append with line numbers.
        prevLine= new FileLine("", startingFile, 0);
    }

    /**
     * Construction a new file lines buffer.
     */
    public FileLines(File startingFile) {
        this(startingFile.getName());
    }

    /**
     * Set the indention prefix.
     */
    private void setIndentPrefix() {
        StringBuffer prefixBuf = new StringBuffer();
        for (int cnt = 0; cnt < currentIndent; cnt++) {
            prefixBuf.append(' ');
        }
        indentPrefix = prefixBuf.toString();
    }


    /**
     * Increase indentation.
     */
    public void incrIndent() {
        currentIndent += INDENT_INCR;
        setIndentPrefix();
    }

    /**
     * Decrease indentation.
     */
    public void decrIndent() {
        currentIndent -= INDENT_INCR;
        setIndentPrefix();
    }

    /**
     * Add a line, remembering it as the previous line.
     */
    private void addLine(String line,
                         String srcFileName,
                         int srcLineNum) {
        prevLine = new FileLine(indentPrefix + line, srcFileName, srcLineNum);
        lines.addElement(prevLine);
    }

    /**
     * Append a string to the buffer, handling indentation and
     * line number mapping.  A newline is appended to the end of the
     * string.
     */
    public void append(String str,
                       String srcFileName,
                       int srcLineNum) {
        str += "\n";

        int idx = 0;
        int nlIdx;

        while ((idx < str.length()) &&
               ((nlIdx = str.indexOf('\n', idx)) >= 0)) {
            addLine(str.substring(idx, nlIdx+1), srcFileName, srcLineNum);
            srcLineNum++; 
            idx = nlIdx + 1;
        }
    }

    /**
     * Append with out a location.  Used for code that that does not
     * come from a source file.
     */
    public void append(String str) {
        String srcFileName = prevLine.srcFileName;
        // Line number is Not incremented by this routine.
        int srcLineNum = prevLine.srcLineNum;
        str += "\n";

        int idx = 0;
        int nlIdx;

        while ((idx < str.length()) &&
               ((nlIdx = str.indexOf('\n', idx)) >= 0)) {
            addLine(str.substring(idx, nlIdx+1), srcFileName, srcLineNum);
            idx = nlIdx + 1;
        }
    }

    /**
     * Append another FileLines to this one.
     */
    public void append(FileLines srcFile) {
        for (int idx = 0; idx < srcFile.lines.size(); idx++) {
            prevLine = (FileLine)srcFile.lines.elementAt(idx);
            lines.addElement(prevLine);
        }
    }

    /**
     * Get current number of lines.
     */
    int size() {
        return lines.size();
    }

    /**
     * Get code as a string.
     */
    public String toString() {
        StringBuffer buf = new StringBuffer();
        for (int idx = 0; idx < lines.size(); idx++) {
            buf.append(((FileLine)lines.elementAt(idx)).line);
        }
        return buf.toString();
    }
    
    /**
     * Write code to a file.
     */
    public void write(FileOutputStream outStream)
            throws java.io.IOException {
        PrintWriter out = new PrintWriter(outStream);
        for (int idx = 0; idx < lines.size(); idx++) {
            out.print(((FileLine)lines.elementAt(idx)).line);
        }
        out.flush();
    }
    
    /*
     * Print class (for debugging).
     */
    public void dbgPrint() {
        for (int idx = 0; idx < lines.size(); idx++) {
            System.err.println(idx+":"+(FileLine)lines.elementAt(idx));
        }
    }

    /**
     * Build map of generated line numbers to a string of "file:linenum".
     */
    String[] getLineNumMap() {
        String[] map = new String[lines.size()+1];
        map[0] = "???";
        for (int idx = 0; idx < lines.size(); idx++) {
            FileLine loc = (FileLine)lines.elementAt(idx);
            map[idx+1] = loc.srcFileName + ":" + loc.srcLineNum;
        }
        return map;
    }
}
