/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: JoltCodeGen.java,v 1.12.12.1 2000/10/19 17:59:08 jasona Exp $
 */





package com.lutris.appserver.server.jolt.joltc;

import com.lutris.appserver.server.jolt.parser.*;
import java.io.*;
import java.util.*;

/**
 * Class that converts a Jolt HTML page to the components of the final
 * class  This doesn't create the actual source file; it calls the parser
 * and the assembles the components into in-memory strings.
 *
 * @version	$Revision: 1.12.12.1 $
 * @author	Mark Diekhans
 * @since	2.0
 */
class JoltCodeGen {
    /*
     * Current source file and line number.
     */
    String curSrcFile;
    
    /*
     * Used to generate unique field names.
     */
    private int fieldNum; 

    /*
     * Table of HTMLDEFs methods names.  Also used to detect
     * recursive calls.  Has a true value if being called,
     * a false value otherwise.
     */
    private Hashtable htmlDefMethods = new Hashtable();
    

    /*
     * Objects that will contain the assembled code.
     */
    protected FileLines importDefs;    // Definitions of imports from <JOLT>.
    protected FileLines classDefs;     // Definitions of classes from <JOLT>.
    protected FileLines htmlStrDefs;   // Definitions of HTML string sections
    protected FileLines runMethod;     // Method to generate page.
    protected FileLines pageCatches;   // Catch code to include.
    protected FileLines pageFinally;   // Finally code to include.

    /**
     * Generate a Java string constant, quoting `\' and `"' and converting
     * newlines.
     *
     * @param text Text to quote.
     * @return The quoted string.
     */
    private String createJavaStringConst(String text) {
        StringBuffer str = new StringBuffer();
        int len = text.length ();

        str.append ("\"");
        for (int idx = 0; idx < len; idx++) {
            char textChar = text.charAt (idx);
            switch (textChar) {
              case '\\':
              case '"':
                str.append ('\\');
                str.append (textChar);
                break;
              case '\n':
                str.append ("\\n");
                break;
              case '\015':
		// Hack to work around the fact that JavaCC does not
		// correctly convert ^M\n into a single unicode newline
		// character.
		if (idx < len - 1 && text.charAt (idx + 1) == '\n')
		    break;
		else {
		    str.append (textChar);
		}
              default:
                str.append (textChar);
                break;
            }
        }
        str.append ("\"");
        return str.toString();
    }
    
    /**
     * Define a field in the generated class containing a string of
     * text for later reference.
     *
     * @param text Text to store in class.
     * @param lineNum The source line number.
     * @return The name of the field in the generated class.
     */
    private String createTextField(String text, int lineNum) {
        String fieldName = "$text" + fieldNum;
        fieldNum++;

        htmlStrDefs.append ("static final String " + fieldName + " = " +
                            createJavaStringConst(text) + ";",
                            curSrcFile, lineNum);
        return fieldName;
    }
    
    /**
     * Output `if' part of a boolean conditional to runMethod, incrementing
     * indent
     */
    private void beginBooleanConditional(ASTBooleanConditional node,
                                         int lineNum) {
        String notOp;
        if (node.getOperator().equalsIgnoreCase("ifeq")) {
            notOp = "";
        } else if (node.getOperator().equalsIgnoreCase("ifneq")) {
            notOp = "!";
        } else {
            throw new Error("unknown operator: " + node.getOperator());
        }
        runMethod.append("if (" + notOp +
                         "this.fieldCompare(\"" + 
                         node.getField() + "\", \"" +
                         node.getValue() + "\")) {",
                         curSrcFile, lineNum);
        runMethod.incrIndent();
    }

    /**
     * Output `if' part of a defined conditional to runMethod, incrementing
     * indent
     */
    private void beginDefinedConditional(ASTDefinedConditional node,
                                         int lineNum) {
        String notOp;
        if (node.getOperator().equalsIgnoreCase("ifdef")) {
            notOp = "";
        } else if (node.getOperator().equalsIgnoreCase("ifndef")) {
            notOp = "!";
        } else {
            throw new Error("unknown operator: " + node.getOperator());
        }
        runMethod.append("if ("+ notOp +
                         "this.fieldExists(\"" +
                         node.getField() + "\")) {",
                         curSrcFile, lineNum);
        runMethod.incrIndent();
    }
   
    private int loopIndexNumber = 0;
    /**
     * Output `if' part of a method conditional to runMethod, incrementing
     * indent
     */
    private void beginMethodConditional(ASTMethodConditional node,
                                        int lineNum) {
        String notOp;
        boolean isloop = false;
        if (node.getOperator().equalsIgnoreCase("ifcall")) {
            notOp = "";
        } else if (node.getOperator().equalsIgnoreCase("ifncall")) {
            notOp = "!";
        } else if (node.getOperator().equalsIgnoreCase("ifloop")) {
            notOp = "";
            isloop = true;
        } else if (node.getOperator().equalsIgnoreCase("ifnloop")) {
            notOp = "!";
            isloop = true;
        } else {
            throw new Error("unknown operator: " + node.getOperator());
        }
        if (isloop) {
            String loopIndexName = "$loop" + loopIndexNumber++;
            runMethod.append("for (int " + loopIndexName + " = 0;" + notOp +
                         node.getMethodName() + "(this, " + loopIndexName +
                         "); ++" + loopIndexName + ") {", curSrcFile, lineNum);
        } else {
            runMethod.append("if (" + notOp +
                         node.getMethodName() + "(this)) {",
                         curSrcFile, lineNum);
      }
        runMethod.incrIndent();
    }
   
    /**
     * Output closing part of any conditional to runMethod, decrementing
     * indent
     */
    private void endConditional() {
        runMethod.decrIndent();
        runMethod.append("}");
    }

    /**
     * Generate any conditionals for a node.
     * @return true if there is a conditional and endConditional
     *  should be called.
     */
    private boolean processConditionals(SimpleNode node, int lineNum) {
        Node child;
        int numChildren = node.jjtGetNumChildren();
        
        for (int idx = 0; idx < numChildren; idx++) {
            child = node.jjtGetChild(idx);
            if (child instanceof ASTBooleanConditional) {
                beginBooleanConditional((ASTBooleanConditional)child,
                                        lineNum);
                return true;
            }
            if (child instanceof ASTDefinedConditional) {
                beginDefinedConditional((ASTDefinedConditional)child,
                                        lineNum);
                return true;
            }
            if (child instanceof ASTMethodConditional) {
                beginMethodConditional((ASTMethodConditional)child,
                                       lineNum);
                return true;
            }
        }
        return false;
    }

    /**
     * Search for a ASTJoltContents child node.
     * @return null if contents node not found or is zero length.  
     */
    private String findContents(SimpleNode node) {
        Node child;
        int numChildren = node.jjtGetNumChildren();

        for (int idx = 0; idx < numChildren; idx++) {
            child = node.jjtGetChild(idx);
            if (child instanceof ASTJoltContents) {
                String contents = ((ASTJoltContents)child).getContents();
                if (contents.length() > 0) {
                    return contents;
                }
            }
        }
        return null;
    }

    /**
     * Search for a ASTJoltArgList child node.
     * @return the node, or null not found.
     */
    private ASTArgList findArgList(SimpleNode node) {
        Node child;
        int numChildren = node.jjtGetNumChildren();

        for (int idx = 0; idx < numChildren; idx++) {
            child = node.jjtGetChild(idx);
            if (child instanceof ASTArgList) {
                return (ASTArgList)child;
            }
        }
        return null;
    }

    /*
     * Code generator for a HtmlSection node; HTML not inside of 
     * a JOLT tag.
     */
    private void processHtmlSection(ASTHtmlSection node) {
        String contents = node.getContents();
        if (contents.length() > 0) {
            int lineNum = node.getBeginLine();
            runMethod.append ("this.append(" +
                              createTextField(node.getContents(), lineNum) +
                              ");", curSrcFile, lineNum);
        }
    }

    /*
     * Code generator for a Javadef node.
     */
    private void processJavadef(ASTJavadef node) {
        String contents = findContents(node);
        if (contents != null) {
            classDefs.append(contents, curSrcFile, node.getBeginLine());
        }
    }

    /*
     * Start a HTMLDEF method definition.
     */
    private void startHtmlDefMethod(String methodName) {
        runMethod.append("private void " + methodName + "(JoltPage page)");
        runMethod.append("        throws Exception {");
        runMethod.incrIndent();
    }

    /*
     * Finish a HTMLDEF method definiton.
     */
    private void finishHtmlDefMethod(String methodName) {
        runMethod.incrIndent();
        runMethod.append("}");
    }

    /*
     * Code generator for a Htmldef node.
     */
    private void processHtmldef(ASTHtmldef node)
            throws java.io.FileNotFoundException,
                   java.io.IOException,
                   JoltCompileException {
        String methodName = node.getMethodName();

        Boolean defined = (Boolean)htmlDefMethods.get(methodName);
        if (defined != null) {
            throw new JoltCompileException("duplicate definition of HTMLDEF " + methodName,
                                           curSrcFile, node.getBeginLine());
        }
        htmlDefMethods.put(methodName, new Boolean(true));

        // Use call stack as stack of methods being generated;
        // push current method.
        FileLines holdMethod = runMethod;
        runMethod = new FileLines(curSrcFile);
        try {
            startHtmlDefMethod(methodName);
            processChildren(node);
            finishHtmlDefMethod(methodName);
            classDefs.append(runMethod);
        } finally {
            runMethod = holdMethod;
        }
        htmlDefMethods.put(methodName, new Boolean(false));
    }

    /*
     * Setup arguments for a java method call.
     */
    private void setUpJavaCallArgs(ASTAnyCall node) {
        ASTArgList argList = findArgList(node);
        int lineNum = node.getBeginLine();
        runMethod.append("this.callArgs.clear();", curSrcFile, lineNum);
        if (argList == null) {
            return; // Not args
        }

        int numArgs = argList.jjtGetNumChildren();
        for (int idx = 0; idx < numArgs; idx++) {
            ASTArg arg = (ASTArg)argList.jjtGetChild(idx);
            runMethod.append("this.callArgs.put(" +
                             createJavaStringConst(arg.getName()) +
                             ", " +
                             createJavaStringConst(arg.getValue()) +
                             ");", curSrcFile, lineNum);
        }
    }

    /*
     * Generate a call of a java method.
     */
    private void generateJavaCall(ASTAnyCall node, String contentsConst) {
        setUpJavaCallArgs(node);
        int lineNum = node.getBeginLine();
        runMethod.append("this.tagContents = " + contentsConst + ";",
                         curSrcFile, lineNum);
        runMethod.append(node.getMethodName() + "(this);",
                         curSrcFile, lineNum);;
    }

    /*
     * Setup arguments for a HTML method call.
     */
    private void setUpHtmlCallArgs(ASTAnyCall node, String methodName) {
        ASTArgList argList = findArgList(node);
        int lineNum = node.getBeginLine();
        runMethod.append("this.callArgs.clear();", curSrcFile, lineNum);
        if (argList == null) {
            return; // Not args
        }

        int numArgs = argList.jjtGetNumChildren();
        for (int idx = 0; idx < numArgs; idx++) {
            ASTArg arg = (ASTArg)argList.jjtGetChild(idx);
            runMethod.append("this.data.set(\"" + methodName + "." +
                             arg.getName() + "\", " +
                             createJavaStringConst(arg.getValue()) +
                             ");", curSrcFile, lineNum);
        }
    }

    /*
     * Generate a call of a HTML method.
     */
    private void generateHtmlCall(ASTAnyCall node, String contentsConst)
            throws JoltCompileException {
        String methodName = node.getMethodName();
        int lineNum = node.getBeginLine();

        setUpHtmlCallArgs(node, methodName);

        if (((Boolean)htmlDefMethods.get(methodName)).booleanValue()) {
            throw new JoltCompileException("recursive call of HTML method " + methodName,
                                           curSrcFile, lineNum);
        }
        runMethod.append("this.data.set(\"" + methodName +
                         ".tagContents\", " + contentsConst + ");",
                         curSrcFile, lineNum);
        runMethod.append(methodName + "(this);",
                         curSrcFile, lineNum);;
        runMethod.append("this.data.remove(\"" + methodName + "\");",
                         curSrcFile, lineNum);
    }

    /*
     * Code generator for an AnyCall node.
     */
    private void processAnyCall(ASTAnyCall node)
            throws JoltCompileException {
        int lineNum = node.getBeginLine();
        boolean cond = processConditionals(node, lineNum);

        String contents = findContents(node);
        String contentsConst;
        if (contents != null) {
            contentsConst = createTextField(contents, lineNum);
        } else {
            contentsConst = "\"\"";
        }
        if (htmlDefMethods.get(node.getMethodName()) != null) {
            generateHtmlCall(node, contentsConst);
        } else {
            generateJavaCall(node, contentsConst);
        }
        if (cond) {
            endConditional();
        }
    }

    /*
     * Code generator for a JoltHtml node.
     */
    private void processJoltHtml(ASTJoltHtml node)
            throws java.io.FileNotFoundException,
                   java.io.IOException,
                   JoltCompileException {
        boolean cond =  processConditionals(node, node.getBeginLine());
        processChildren(node);
        if (cond) {
            endConditional();
        }
    }

    /*
     * Code generator for a JavaImport node.
     */
    private void processJavaImport(ASTJavaImport node) {
        String contents = findContents(node);
        if (contents != null) {
            importDefs.append(contents, curSrcFile, node.getBeginLine());
        }
    }

    /*
     * Code generator for an Include node.
     */
    private void processInclude(ASTInclude node)
            throws java.io.FileNotFoundException,
                   java.io.IOException,
                   JoltCompileException {
        parse(new File(node.getFilename()));
    }

    /*
     * Code generator for an Source node.
     */
    private void processSource(ASTSource node)
            throws java.io.FileNotFoundException,
                   java.io.IOException,
                   JoltCompileException {
        boolean cond =  processConditionals(node, node.getBeginLine());
        runMethod.append("this.sourceFile(" + 
                         createJavaStringConst(node.getFilename()) + ");");
        if (cond) {
            endConditional();
        }
    }

    /*
     * Code generator for a JavaCatch  node.
     */
    private void processJavaCatch(ASTJavaCatch node) {
        String contents = findContents(node);
        if (contents == null) {
            // Always generate catch
            contents = "";
        }
        int lineNum = node.getBeginLine();
        pageCatches.append("catch(" + node.getException() + " exception) {",
                           curSrcFile, lineNum);
        pageCatches.append (contents, curSrcFile, lineNum);
        pageCatches.append ("}", curSrcFile, lineNum);
    }

    /*
     * Code generator for a JavaFinally node.
     */
    private void processJavaFinally(ASTJavaFinally node) {
        String contents = findContents(node);
        if (contents != null) {
            pageFinally.append (contents, curSrcFile, node.getBeginLine());
        }
    }

    /*
     * Code generator for a JoltField node.
     */
    private void processJoltField(ASTJoltField node)
            throws JoltCompileException {
        StringBuffer statement = new StringBuffer();
        String encoding = node.getEncoding();
        String defaultVal = node.getDefault();
        String fieldName = node.getFieldName();

        // Parser generates #ALL, page expects "" for all fields.
        if (fieldName.equals("#ALL")) {
            fieldName = "";
        }

        statement.append("appendFieldValue(\"" + fieldName + "\", ");

        if ((encoding == null) || (encoding.equalsIgnoreCase("html"))) {
            statement.append("JoltPage.HTML_ENCODING");
        } else if (encoding.equalsIgnoreCase("javascript")) {
            statement.append("JoltPage.JAVASCRIPT_ENCODING");
        } else if (encoding.equalsIgnoreCase("raw")) {
            statement.append("JoltPage.NO_ENCODING");
        } else {
            throw new JoltCompileException("invalid encoding for field " + 
                                           node.getFieldName() + ": \"" +
                                           encoding + "\"",
                                           curSrcFile, node.getBeginLine());
        }

        statement.append(", ");
        if (defaultVal != null) {
            statement.append(createJavaStringConst(defaultVal));
        } else {
            statement.append("null");
        }
        statement.append(");");
        runMethod.append(statement.toString(), curSrcFile,
                         node.getBeginLine());
    }
    
    /**
     * Process all children of a node, with the exception of the
     * nodes that are explitly accessed.
     */
    private void processChildren(SimpleNode node)
            throws java.io.FileNotFoundException,
                   java.io.IOException,
                   JoltCompileException,
                   JoltCompileException {
        Node child;
        int numChildren = node.jjtGetNumChildren();

        for (int idx = 0; idx < numChildren; idx++) {
            child = node.jjtGetChild(idx);
            if (child instanceof ASTJoltField) {
                processJoltField((ASTJoltField)child);
            } else if (child instanceof ASTHtmlSection) {
                processHtmlSection((ASTHtmlSection)child);
            } else if (child instanceof ASTJavadef) {
                processJavadef((ASTJavadef)child);
            } else if (child instanceof ASTHtmldef) {
                processHtmldef((ASTHtmldef)child);
            } else if (child instanceof ASTAnyCall) {
                processAnyCall((ASTAnyCall)child);
            } else if (child instanceof ASTJoltHtml) {
                processJoltHtml((ASTJoltHtml)child);
            } else if (child instanceof ASTJavaImport) {
                processJavaImport((ASTJavaImport)child);
            } else if (child instanceof ASTInclude) {
                processInclude((ASTInclude)child);
            } else if (child instanceof ASTSource) {
                processSource((ASTSource)child);
            } else if (child instanceof ASTJavaCatch) {
                processJavaCatch((ASTJavaCatch)child);
            } else if (child instanceof ASTJavaFinally) {
                processJavaFinally((ASTJavaFinally)child);
            } else if (child instanceof ASTJoltParsedContents) {
                processChildren((SimpleNode)child);
            } else if ((child instanceof ASTArg)
                       || (child instanceof ASTArgList)
                       || (child instanceof ASTBooleanConditional)
                       || (child instanceof ASTDefinedConditional)
                       || (child instanceof ASTMethodConditional)) {
                // Skip nodes that are accessed explictly
            } else {
                throw new Error("unknown child type: " + child.getClass().getName());
            }
        }
    }
 
    /**
     * Parse the page, storing generated code components in this object.
     */
    private void parse(File srcFile)
            throws java.io.FileNotFoundException,
                   java.io.IOException,
                   JoltCompileException {
        JoltParser joltParser;
        ASTStart startNode;
        String oldSrcFile = curSrcFile;
        curSrcFile = srcFile.getPath();

        FileInputStream inStream = new FileInputStream(srcFile);
        try {
            joltParser = new JoltParser(new FileInputStream(srcFile));
        } finally {
            inStream.close();
        }
        try {
            startNode = joltParser.Start();
            processChildren(startNode);
        } catch (ParseException except) {
            throw new JoltCompileException(curSrcFile, except);
        }
        curSrcFile = oldSrcFile;
    }

    /**
     * Constructor.  Parse the specified file.
     * @exception java.io.FileNotFoundException If the source file or
     *  an include file was not found.
     * @exception java.io.IOException If an error occured reading the source
     *  or an included file was not found.
     * @exception JoltCompileException
     */
    protected JoltCodeGen(File srcFile)
            throws java.io.FileNotFoundException,
                   java.io.IOException,
                   JoltCompileException {
        importDefs = new FileLines(srcFile);
        classDefs = new FileLines(srcFile);
        htmlStrDefs = new FileLines(srcFile);
        runMethod = new FileLines(srcFile);
        pageCatches = new FileLines(srcFile);
        pageFinally = new FileLines(srcFile);
        parse(srcFile);
    }
}  
