/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: JoltCompile.java,v 1.34.12.1 2000/10/19 17:59:08 jasona Exp $
 */





package com.lutris.appserver.server.jolt.joltc;

import java.util.*;
import java.io.*;

/**
 * Command line program to compile a JoltHTML into Java source for a
 * Jolt Presentation Object.
 *
 * <PRE>
 *    JoltCompile [-d destdir] [-j javadir] [-k] \
 *                src.jhtml packageName
 * </PRE>
 * <UL>
 * <LI> -d destdir - Destination root for java class files.  The default is to
 * end up in the javadir.
 * <LI> -j javadir - Directory to put generate java files in.
 * If not specified, the current directory is used.
 * <LI> -k - Keep intermediate java files.
 * </UL>
 */
public class JoltCompile {
    /*
     * Arguments.
     */
    File destDir;
    File javaDir;
    boolean keepJavaSrc;
    File joltSrc;
    String packageName;
    String className;
    String[] lineNumMap;

    /*
     * Temp generated file.  Always purged at end unless keep is specified.
     */
    File javaSrc = null;

    // FIX: Create a utility to read an entire string into a file.
    /**
     * Read an JoltHTML page into a string.
     *
     * @param pagePath File path of the JoltHTML file to read.
     * @return A string containing the page.
     * @throws java.io.IOException If a read error occurs.
     */
    static String readHtmlPage (File pagePath)
        throws java.io.IOException {
        BufferedReader input =
            new BufferedReader (new FileReader (pagePath));
        String line;
        StringBuffer page = new StringBuffer ();
        
        try {
            while ((line = input.readLine ()) != null) {
                page.append (line);
                page.append ("\n");
            }
        } finally {
            input.close ();
        }
        return page.toString ();
    }
    
    /*
     * Convert a JoltHTML page to Java.
     */
    private void compilePageToJava()
            throws java.io.FileNotFoundException,
                   java.io.IOException,
                   JoltCompileException {

        // Get the filename now so we can use it in the line mapping.
        if (javaDir == null) {
            //FIX: Javadoc says below new should work with a null dir
            //but it gets an exception.
            javaDir  = new File(".");
        }
        javaSrc = new File(javaDir, className + ".java");

        // Compile page to java code.
        FileLines javaCode = JoltToJava.compile(packageName, className,
                                                joltSrc, javaSrc);
        lineNumMap = javaCode.getLineNumMap();

        // Create java source file.
        javaDir.mkdirs();
        FileOutputStream out = new FileOutputStream(javaSrc);
        try {
            javaCode.write(out);
        } finally {
            out.close ();
        }
    }

    /*
     * Delete generate java src if it exists and keep not specified.
     */
    private void cleanupJavaSrcFile() {
        if (!keepJavaSrc && (javaSrc != null)) {
            javaSrc.delete();
        }
    }

    /**
     * Usage message and exit.
     */
    private void usage(String msg) {
        if (msg != null) {
            System.err.println(msg);
        }
        System.err.println ("joltc [-d destdir] [-j javadir] [-k] [-s] src.jhtml packageName");
        System.exit(1);
    }

    /**
     * Check for an required argument to an option existing
     * and only being there specified once.
     */
    private void optionArgCheck(String[] args, int idx, Object value) {
        if (value != null) {
            usage(args[idx] + " specified multiple times");
        }
        if (idx == args.length-1) {
            usage(args[idx] + " requires an argument");
        }
    }


    /**
     * Parse arguments and save in object.
     */
    private void parseArgs(String[] args) {
        int idx = 0;
        destDir = null;
        javaDir = null;
        keepJavaSrc = false;
        while ((idx < args.length) && (args[idx].indexOf('-') == 0)) {
            if (args[idx].equals("-d")) {
                optionArgCheck(args, idx, destDir);
                destDir = new File(args[++idx]);
            } else if (args[idx].equals("-j")) {
                optionArgCheck(args, idx, javaDir);
                javaDir = new File(args[++idx]);
            } else if (args[idx].equals("-k")) {
                keepJavaSrc = true;
            } else if (args[idx].equals("-jc")) {
                keepJavaSrc = true;
            } else {
                usage("illegal argument \"" + args[idx] + "\"");
            }
            idx++;
        }
        if (idx != args.length-2) {
            usage("wrong number of arguments");
        }
        joltSrc = new File(args[idx++]);
        // Allow class or file syntax on name.
        packageName = args[idx].replace('/', '.');

        className = joltSrc.getName();
        int extIdx = className.lastIndexOf('.');
        if (extIdx > 0) {
            className = className.substring(0, extIdx);
        }
    }

    /**
     * Do the compilation.
     */
    private int doComp(String[] args) 
            throws java.io.FileNotFoundException,
                   java.io.IOException,
                   JoltCompileException {
        
        parseArgs(args);
        
        // Convert page to a java source file.
        compilePageToJava();
        
        // Setup filtering to map line numbers on errors.
        JavaErrorFileFilter filter =
            new JavaErrorFileFilter(javaSrc, lineNumMap);

        // Get arguments and call javac.
        Vector javacArgs = new Vector();
        if (destDir != null) {
            javacArgs.addElement("-d");
            javacArgs.addElement(destDir.getPath());
        }
        javacArgs.addElement(javaSrc.getPath());
        String[] argArray = new String[javacArgs.size()];
        javacArgs.copyInto(argArray);

        Javac javac = new Javac();
        int code = javac.javac(argArray, filter);
        return code;
    }

    /*
     * Do compilation and handle errors.
     */
    private void compileJolt(String[] args) {
        int code = 0;
        try {
            code = doComp(args);
        } catch (JoltCompileException except) {
            System.err.println(except.getMessage());
            cleanupJavaSrcFile();
            System.exit (1);
        } catch (FileNotFoundException except) {
            System.err.println(except.getClass().getName() + ": " +
                               except.getMessage ());
            cleanupJavaSrcFile();
            System.exit (1);
        } catch (java.io.IOException except) {
            System.err.println(except.getClass().getName() + ": " +
                               except.getMessage ());
            cleanupJavaSrcFile();
            System.exit (1);
        } catch (Exception except) {
            System.err.println("Fatal Error: " + except.getMessage ());
            except.printStackTrace(System.err);
            cleanupJavaSrcFile();
            System.exit(100);
        }
        cleanupJavaSrcFile();
        System.exit(code);
    }

    /**
     * Program entry for command line JoltHTML compiler.
     */
    public static void main (String[] args) {
        new JoltCompile().compileJolt(args);
    }
}


/**
 * Class to filter output, looking for errors lines that have the
 * generated java file in them and mapping the filename/linenumber bac
 * to the jolt source.
 */
class JavaErrorFileFilter implements OutputFilter {
    private String javaFile;
    private String[] lineNumMap;

    /**
     * Construact a filter.
     */
    JavaErrorFileFilter(File javaFile,
                        String[] lineNumMap) {
        this.javaFile = javaFile.getPath();
        this.lineNumMap = lineNumMap;
    }
    
    /**
     * Add notice that mapping of a line failed.
     */
    private String mappingFailed(String line, int id) {
        return "*** Notice: joltc was unable to map filename and line number for (" +
            id + "):\n" + line;
    }

    /**
     * Filter a line, mapping filename/linenumber if neccessary.
     */
    public String filterLine(String line) {
        if (!line.startsWith(javaFile))
            return line;  // Line doesn't start with file name.
        int idx = javaFile.length();

        // Extract and convert line number.
        if (line.charAt(idx) != ':')
            return mappingFailed(line, 1);
        idx++;
        int idx2 = line.indexOf(':', idx);
        if (idx2 < 0) 
            return mappingFailed(line, 2);
        idx2++;
        int lineNum;
        try {
            lineNum = Integer.parseInt(line.substring(idx, idx2-1));
        } catch (NumberFormatException except) {
            return mappingFailed(line, 3);
        }

        
        // Look up line number.
        String srcLoc = (lineNum < lineNumMap.length) ?
            lineNumMap[lineNum] : lineNumMap[lineNumMap.length-1];
        
        // Assemble new line
        return srcLoc + ":" + line.substring(idx2);
    }
}
