/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: JoltPage.java,v 1.58.12.1 2000/10/19 17:59:08 jasona Exp $
 */





package com.lutris.appserver.server.jolt.joltpo;

import java.lang.reflect.Method;
import java.io.*;
import java.util.*;
import com.lutris.util.*;
import com.lutris.http.*;
import com.lutris.appserver.server.httpPresentation.*;
import com.lutris.appserver.server.Application;
import com.lutris.appserver.server.ApplicationData;
import com.lutris.appserver.server.session.*;


/**
 * Base class of a Jolt-HTML page.  Each page extends this class.  It is also
 * passed to each of the page component objects as an argument.  
 * 
 * @author	Mark Diekhans
 * @version	$Revision: 1.58.12.1 $
 */
public abstract class JoltPage implements HttpPresentation {
    
    // FIX: Add method to clear header.

    /**
     * Use HTML encoding.
     */
    protected static final int HTML_ENCODING = 1;

    /**
     * Use JavaScript encoding.
     */
    protected static final int JAVASCRIPT_ENCODING = 2;

    /**
     * Use no encoding.
     */
    protected static final int NO_ENCODING = 3;

    /**
     * Current HTML page's content.
     */
    private StringBuffer pageHtml = new StringBuffer();

    /**
     * Request object that contains the data from the HTTP request.
     */
    public JoltRequest request;

    /**
     * Response object used to set cookies and other headers.
     */
    public JoltResponse response;

    /**
     * Table used to set Jolt fields, contains CGI args, etc.
     */
    public JoltPageData data = null;

    /**
     * On Jolt calls, contains the arguments specified in the
     * &lt;JOLT CALL&gt; tag.
     */
    public JoltCallArgs callArgs = new JoltCallArgs();

    /**
     * On Jolt calls, contains the body of the tag between
     * &lt;JOLT CALL&gt; and &lt;/JOLT&gt;.
     */
    public String tagContents = null;

    /**
     * Application object that this presentation is associated with.
     * If null, no application is associated.
     */
    public Application application;

    /**
     * Session object that this presentation is associated with.
     * If null, no session has been associated by the application.
     */
    public Session session;

    /**
     * If an exception/error occured, this field is set to the exception
     * object. This is only set for and used by `ErrorHandler' presentations.
     */
    public Throwable exception;

    /**
     * Size of buffer use by sourceFile.
     */
    private static final int SOURCE_BUFFER_SIZE = 4096;

    /**
     * CGI arguments, stored as a hash table.
     */
    protected Hashtable queryTable = new Hashtable();

    /*
     * The HttpPresentationComms object. Used by JoltDebug.
     * FIX: This is purposely not documented and is a bit of a hack.
     * We really need a debug routine that just works off of the
     * publicly available data in JoltRequest or perhaps we just need to
     * export this field and live dangerously.
     */
    protected HttpPresentationComms presentationComms;


    /**
     * Keep track if data has been flushed.
     */
    private boolean flushed = false;


    /**
     * Append HTML to the current page.
     *
     * @param html string to append.
     */
    public void append(String html) {
        pageHtml.append(html);
    }


    /**
     * Append HTML to the current page.
     *
     * @param html string buffer to append.
     */
    public void append(StringBuffer html) {
        pageHtml.append(html);
    }


    /**
     * Append HTML to the current page with carrage return.
     *
     * @param html string to append.
     */
    public void appendln(String html) {
        pageHtml.append(html);
        pageHtml.append("\n");
    }


    /**
     * Append HTML to the current page with carrage return.
     *
     * @param html string buffer to append.
     */
    public void appendln(StringBuffer html) {
        pageHtml.append(html);
        pageHtml.append("\n");
    }


    /**
     * Determine if a specified field is defined.
     *
     * @param keyword Key of the object in the page data table to check.
     * @return If the field exists in the page data table, <CODE>true</CODE>,
     *  if not <CODE>false</CODE>.
     * @exception HttpPresentationException If keyword is invalid.
     */
    protected boolean fieldExists (String keyword)
            throws HttpPresentationException {
        try {
            return data.containsKey(keyword);
        } catch(KeywordValueException except) {
            throw new HttpPresentationException(except);
        }
    }
  

    /**
     * Determine if a specified field's value compares to a string.
     *
     * @param keyword Key of the object in the page data table.
     * @param str String to compare
     * @return If the field's string representation is is equal to
     *  <CODE>str</CODE> <CODE>true</CODE> is returned, if they are not
     *  equal or <CODE>false</CODE> if the strings are not equal or the
     *  field does not exist.
     * @exception HttpPresentationException If keyword is invalid.
     */
    protected boolean fieldCompare (String keyword, String str)
	    throws HttpPresentationException {
        try {
            Object obj = data.get(keyword, null);
            if (obj == null)
                return false;
            return str.equals(obj.toString());
        } catch(KeywordValueException except) {
            throw new HttpPresentationException(except);
        }
    }
  

    /**
     * Clears the internal buffer of HTML appended since the last
     * <CODE>flush()</CODE> or since the beginning of the page. Note
     * that it does not effect HTTP headers. It can be used effectively
     * to replace the dynamically generated page with an HTML error.
     */
    public void clear() {
        pageHtml.setLength(0);
    }
    

    /**
     * Write all currently buffered HTML to the client.  This is normally
     * not necessary, as the data is all written when the page <CODE>run</CODE>
     * method completes.  This is provide mainly to allow a stream of
     * JavaScript to be written to the client while some long-running task
     * is waiting to complete.  The first time this function is called, any
     * response cookies or headers that have been set are sent to the client.
     * Attempts to set response cookies or headers after this function has 
     * been called will result in an error.
     */
    public void flush() throws HttpPresentationException {
        flushed = true;
        response.write(pageHtml.toString());
        pageHtml.setLength(0);
    }

    /**
     * Format an field. If object has a toHtml() method, then it
     * is called else toString() is called.
     *
     * @param fieldObj Field object to format.
     * @param encoding Type of encoding to use.
     * @exception KeywordValueException If an error occurs accessing the
     *  KeywordValueTable
     */
    private void formatField(Object fieldObj, int encoding)
       	    throws KeywordValueException {
	try {
	    // If object has a toHtml() method then call it,
	    // else substitue toString().
	    Class objClass = fieldObj.getClass();
	    Method toHtmlMethod = objClass.getMethod("toHtml", null);
	    pageHtml.append(toHtmlMethod.invoke(fieldObj, null));
        } catch(Exception except) {
            switch (encoding) {
            case HTML_ENCODING:
                pageHtml.append(HtmlEncoder.encode(fieldObj.toString()));
                break;
            case JAVASCRIPT_ENCODING:
                pageHtml.append(JavaScriptEncoder.encode(fieldObj.toString()));
                break;
            case NO_ENCODING:
            default:
                pageHtml.append(fieldObj.toString());
                break;
            }
	}
    }
    

    /**
     * Function to append a field from keyword/value table. If the value is
     * a section (recursive key/value table), Then its formatted as HTML.
     * In addition, if any object implements a toHtml() method then this
     * is invoked to generate HTML output.
     * 
     * @param field Name of field to append.
     * @param encoding Type of encoding to use.  One of 
     *    <CODE>HTML_ENCODING</CODE>, <CODE>JAVASCRIPT_ENCODING</CODE>, or
     *    <CODE>NO_ENCODING</CODE>.
     * @param defaultValue The default value to use, or <CODE>null<CODE> if
     *    no default.
     * @exception Exception JoltHTML page objects may throw any exception.
     */
    protected void appendFieldValue(String field, int encoding,
                                    String defaultValue)
            throws Exception {
        if (field.length() == 0) {
            formatField (data, encoding);  // The whole thing.
        } else {
            Object value = data.get(field);
            if (value == null) {
                value = defaultValue;
            }
            if (value == null) {
                throw new HttpPresentationException(
                   "Jolt field " + field + " is not defined in page data");
            }
            formatField(value, encoding);
        }
    }


    /**
     * Sources a file found by the class loader, appending it to the output
     * buffer.
     *
     * @param appFileName The file to source.  It may either be relative to the
     *  path of this presentation object, or absolute, which makes it relative
     *  to the application's <CODE>presentationPrefix</CODE>.
     * @exception IOException If the file can not be found or opened or an
     *  I/O error occurs.
     */
    protected void sourceFile(String appFileName)
            throws IOException, HttpPresentationException {
        File appFile = new File(appFileName);
        if (!appFile.isAbsolute()) {
            File poFile = new File(request.getPathInfo());
            appFile = new File(poFile.getParent(), appFileName);
        }
        InputStream input = application.getHttpPresentationManager().getAppFileAsStream(appFile.getPath());
        try {
            InputStreamReader reader = new InputStreamReader(input);
            char [] buffer = new char[SOURCE_BUFFER_SIZE];
            while (true) {
                int readLen = reader.read(buffer, 0, SOURCE_BUFFER_SIZE);
                if (readLen < 0) {
                    break;
                }
                pageHtml.append(buffer, 0, readLen);
            }
        } finally {
            input.close();
        }
    }


    /**
     * Parse CGI arguments into a keyword/value table and save in the
     * page data object.  In certain cases, this behavior might not be
     * desirable, such as the argument names not being legal keywords.
     * In this case, the Jolt object can override this method to do
     * what ever it wants.
     *
     * @param queryTable Hashtable contain CGI arguments and values.
     * @exception Exception JoltHTML page objects may throw any exception.
     */
    protected void parseCgiArgs() throws Exception {
        data.set(JoltPageData.CGI_ARGS, HttpUtil.buildQueryTable(queryTable));
    }

    /**
     * Methods defined by the derived class that executes the contents
     * of the page.
     * @exception Exception JoltHTML page objects may throw any exception.
     */
    protected abstract void runPage() throws Exception;

    

    /**
     * Do all setup before running a page.
     *
     * @param comms HTTP communications object.
     * @exception Exception JoltHTML page objects may throw any exception
     *  and this routine calls methods that can be overridden.
     */
    private void setup(HttpPresentationComms comms) throws Exception {
        /*
         * If we get a handle on the input stream and directly read and
         * parse the data, then the comms.request.getParameter* methods
         * are invalidated and return null when accessed by the presentation
         * objects.  This is essentially what is happening if we use
         * the above code.  To fix, if we use the getParameter* methods to
         * create the queryTable, then the getParamter* methods can
         * subsequently be used by the presentation objects and by error
         * handlers.  Hence the following code:
         */
        Enumeration e = comms.request.getParameterNames();
        while (e.hasMoreElements()) {
            String name = e.nextElement().toString();
            String [] val = comms.request.getParameterValues(name);
            if (val == null) {
                queryTable.put(name, "");
            } else if (val.length == 1) {
                queryTable.put(name, val[0]);
            } else {
                queryTable.put(name, val);
            }
        }

        flushed = false;
        pageHtml.setLength(0);
        data = new JoltPageData();
        request = new JoltRequest(comms.request, queryTable);
        response = new JoltResponse(comms.response);

        application = comms.application;
        session = comms.session;
        exception = comms.exception;
        presentationComms = comms;

        parseCgiArgs();

	// Attach session data...
        // If SessionManager.createSession() fails then we are probably handling
        // an exception and there may be not session available.....
        if ((session == null) && (exception == null)) {
            throw new HttpPresentationException("Session reference passed to JoltPage is null");
        }
        SessionData sessionData = null;
        if (session != null) {
            sessionData = session.getSessionData();
        }
        if (sessionData != null) {
            data.set("session", sessionData);
        }

	// Attach application data...
        ApplicationData applicationData = application.getApplicationData();
        if (applicationData != null) {
            data.set("application", applicationData);
        }
    }

    /**
     * Abstract methods that run all code on the page and output the results
     * to the HTTP client.
     *
     * @param comms HTTP communications object.
     * @exception Exception JoltHTML page objects may throw any exception.
     */
    public void run (HttpPresentationComms comms) throws Exception {
        try {
            if (comms.request.getMethod().equals("HEAD")) {
                response = new JoltResponse(comms.response);
                response.writeHeaders();
            } else {
                setup(comms);
                // Execute the page...
                runPage();
                
                // Setting the content length inproves performance on many browsers.
                if (!flushed) {
                    comms.response.setContentLength(pageHtml.length());
                }
            
                response.write(pageHtml.toString());
            }
        } finally {
            // Clean up to mimimize memory if object cached for reuse.
            pageHtml.setLength(0);
            data = null;
            request = null;
            response = null;
            application = null;
            session = null;
            exception = null;
            presentationComms = null;

        }
    }
}
