/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: JoltSim.java,v 1.15.4.1 2000/10/19 17:59:08 jasona Exp $
 */





package lutris.jolt.joltsim;
import lutris.jolt.presentation.*;
import lutris.jolt.jhtmlPresentation.*;
import lutris.jolt.jhtml.page.*;
import lutris.keywordValue.*;
import lutris.http.*;
import java.io.*;

/**
 * Class used to simulate the Jolt presentation environment.  This is used
 * for the basic debugging of presentation screens outside of a full
 * server environment.  Currently only GET requests are supported.
 */
public class JoltSim {
    /**
     * Run the specified Presentation
     *
     * @param presObj Name of presentation object.
     * @param pageArgs HTTP page arguments to pass to the page.
     * @param headerOutputStream Output stream receive HTTP response header.
     * @param bodyOutputStream Output stream receive HTTP response body.
     * @exception Exception Presentations are allowed to throw any exception.
     */
    public static void run (String presObj,
                            HttpPageArgs pageArgs,
                            DataOutputStream headerOutputStream,
                            DataOutputStream bodyOutputStream)
            throws java.lang.Exception {
        SimComms comms = new SimComms (pageArgs,
                                       headerOutputStream,
                                       bodyOutputStream);
        String presClass = presObj + ".Presentation";
        ((HttpPresentation) Class.forName (presClass).newInstance()).run (comms);
    }
    
    /**
     * Generate a usage message and exit.
     *
     * @param msg Message describing the specific error.
     */
    static void usage (String msg) {
        String umsg = 
            "Usage: joltsim [options] presobj\n" +
            "   -h headerout - File to write the HTTP reponse header to or specify `-'\n" +
            "                  to discard.  Default is stdout.\n" +
            "   -b bodyout -   File to write the HTTP reponse body to or specify `-' to\n" +
            "                  discard.  Default is stdout.\n" +
            "   -p argsfile -  Specify file containing the pageArgs to pass to the page,\n" +
            "                  including the 'http.' section.  The file should consist of\n" +
            "                  lines of:\n" +
            "                       fieldname: value\n" +
            "                  Blank lines and lines where the first non-blank character\n" +
            "                  is '#' are skipped.  Lines can't currently be continued\n" + 
            "                  or contain newlines.  Maybe specified multiple times.\n" +
            "   -a arg -       Specify a single page argument.  The arg string is in the\n" +
            "                  form \"fieldname: value\". This option maybe specified\n" +
            "                  multiple times and in conjunction with -p.\n" +
            "   presobj -      The fully qualified package of the presentation object to\n" +
            "                  run.  It must be on the current CLASSPATH.";
        System.err.println (msg);
        System.err.println (umsg);
        System.exit (1);
    }

    /**
     * Parse a `fieldname: value' string.
     *
     * @param entry Entry to parse.
     * @param from File name or argument option where this value was obtained.
     * @return array of keyword and value.
     */
    static String [] parseFieldValue (String entry,
                                      String from) {
        String [] result = new String [2];
        int colon = entry.indexOf (':');
        if (colon < 0) {
            System.err.println ("No `:' in keyword/value entry \"" +
                                entry + "\", from " + from);
            System.exit (1);
        }
        result [0] = entry.substring (0, colon).trim();
        result [1] = entry.substring (colon+1, entry.length()).trim();
        return result;
    }

    /**
     * Parse a keyword/value file.
     * 
     * @param filename The name of the file to parse.
     * @param pageArgs Keyword/value table to add the arguments to.
     * @exception FileNotFoundException If file does not exist.
     * @exception SecurityException Should not be thrown.
     * FIX:  There has to be a better way to read strings.
     */
    static void parsePageArgs (String filename,
                               KeywordValueTable pageArgs)
        throws FileNotFoundException,
               IOException,
               KeywordValueException {
        FileInputStream infile = new FileInputStream (filename);
        StringBuffer inbuf = new StringBuffer ();
        int inbyte = 0;
        int lineNum = 0;
        do {
            inbyte = infile.read();
            if ((inbyte < 0) || (inbyte == '\n')) {
                lineNum++;
                String line = inbuf.toString().trim();
                if ((line.length() > 0) && (line.charAt(0) != '#')) {
                    String [] pair = parseFieldValue (line, filename + ":" + lineNum);
                    pageArgs.set (pair[0], pair[1]);
                }
                inbuf.setLength(0);
            } else {
                inbuf.append((char) inbyte);
            }
        } while (inbyte >= 0);
    }

    /**
     * Issue an error about a option requiring an argument.
     *
     * @param option Option to complain about.
     */
    static void missingArg (String option) {
        usage ("option " + option + " requires an argument");
    } 


    /**
     * Command line entry for running a presentation object in a test environment.
     *
     * @param args Command line arguments to program.
     * @exception Presentations can throw any exception.
     * @see lutris.jolt.joltsim.JoltSim
     */
    public static void main (String [] args)
        throws Exception {
        String presObj = null;
        HttpPageArgs pageArgs = new HttpPageArgs();
        String headerFile = null;
        String bodyFile = null;

        int idx;
        for (idx = 0; (idx < args.length) && (args[idx].indexOf('-') == 0); idx++) {
            if (args[idx].equals ("-h")) {
                idx++;
                if (idx == args.length) {
                    missingArg ("-h");
                }
                if (args[idx].equals ("-")) {
                    headerFile = "/dev/null";
                } else {
                    headerFile = args[idx];
                }
            } else if (args[idx].equals ("-b")) {
                idx++;
                if (idx == args.length) {
                    missingArg ("-b");
                }
                if (args[idx].equals ("-")) {
                    bodyFile = "/dev/null";
                } else {
                    bodyFile = args[idx];
                }
            } else if (args[idx].equals ("-p")) {
                idx++;
                if (idx == args.length) {
                    missingArg ("-p");
                }
                parsePageArgs (args [idx], pageArgs);
            } else if (args[idx].equals ("-a")) {
                idx++;
                if (idx == args.length) {
                    missingArg ("-a");
                }
                String [] pair = parseFieldValue (args [idx], "-a");
                pageArgs.set (pair[0], pair[1]);
            } else {
                usage ("unknown option \"" + args [idx] + "\"");
            }
        }
        
        if (idx != args.length-1) {
            usage ("Wrong # of arguments:");
        }
        presObj = args [idx];

        // Setup output files
        DataOutputStream headerOutputStream, bodyOutputStream;

        if (headerFile == null) {
            headerOutputStream = new DataOutputStream (new FileOutputStream (FileDescriptor.out));
        } else {
            headerOutputStream = new DataOutputStream (new FileOutputStream (headerFile));
        }
        if (bodyFile == null) {
            bodyOutputStream = new DataOutputStream (new FileOutputStream (FileDescriptor.out));
        } else {
            bodyOutputStream = new DataOutputStream (new FileOutputStream (bodyFile));
        }

        run (presObj, pageArgs, headerOutputStream, bodyOutputStream);
    }
}

