/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 */
package org.enhydra.kelp.common.wizard;

// Kelp imports
import org.enhydra.kelp.common.Constants;
import org.enhydra.kelp.common.KelpTool;
import org.enhydra.kelp.common.node.OtterNode;
import org.enhydra.kelp.common.node.OtterProject;
import org.enhydra.kelp.common.capture.Writer;
import org.enhydra.kelp.common.capture.WriteListener;

// Standard imports
import java.io.PrintWriter;
import java.io.File;
import java.util.StringTokenizer;

/**
 *
 *
 * @author Paul Mahar
 */
abstract public class AbstractBuildTool extends KelpTool {
    private boolean              echo = true;
    private OtterProject         project = null;
    private OtterNode[]          nodes = null;
    private PrintWriter          echoWriter = null;

    /**
     * Create a CompileThread with listeners ready to
     * recieve output from XMLC.
     *
     *
     * @param listeners
     * A listener implements an onWrite method that is called
     * whenever XMLC writes output. The writes are captured
     * for display in a wizard.
     */
    public AbstractBuildTool() {
        this(new WriteListener[0]);
    }

    public AbstractBuildTool(WriteListener[] listeners) {
        this(listeners, false);
    }

    public AbstractBuildTool(WriteListener[] listeners, boolean buffer) {
        super();
        Writer capture = new Writer();

        capture.setLineBuffered(buffer);
        for (int i = 0; i < listeners.length; i++) {
            capture.addWriteListener(listeners[i]);
        }
        echoWriter = new PrintWriter(capture, false);
    }

    /**
     * Loop through all selected nodes files and build them.
     * Write out exceptions to any registered listeners.
     */
    abstract public void make();

    /**
     * Determine what files were created and write out the
     * build results to any registered listeners.
     */
    abstract public void printResults();

    /**
     * Get the HTML source files that are to be compiled.
     */
    public OtterNode[] getNodes() {
        return nodes;
    }

    /**
     * Set the HTML source files to compile.
     */
    public void setNodes(OtterNode[] n) {
        nodes = n;
    }

    /**
     * Get the IDE project that contains the HTML source files to compile.
     */
    public OtterProject getProject() {
        return project;
    }

    /**
     * Set the IDE project that contains the HTML source files to compile.
     */
    public void setProject(OtterProject bp) {
        project = bp;
    }

    public boolean isEcho() {
        return echo;
    }

    public void setEcho(boolean b) {
        echo = b;
    }

    public PrintWriter getEchoWriter() {
        return echoWriter;
    }

    /**
     * Run the XMLC for all selected HTML files.
     */
    public void run() {
        setFresh(true);
        make();
    }

    public synchronized void runInCurrentThread() {
        setFresh(true);
        make();
    }


    public void printExceptions() {
        StringTokenizer tokenizer = null;
        int             eCount = 0;
        int             totalCount = getNodes().length;

        for (int i = 0; i < totalCount; i++) {
            if (getNodes()[i].isSelected()
                    && getNodes()[i].getException() != null) {
                eCount++;
            }
        }
        if (eCount > 0) {
            echoWriter.println(new String());
            echoWriter.println("Exception encountered:");
            for (int i = 0; i < totalCount; i++) {
                if (getNodes()[i].getException() != null) {
                    echoWriter.print(Constants.TAB4);
                    echoWriter.print(getNodes()[i].getFilePath());
                    echoWriter.println(':');
                    tokenizer =
                        new StringTokenizer(getNodes()[i].getException().getMessage(),
                                            Constants.NEWLINE);
                    while (tokenizer.hasMoreTokens()) {
                        echoWriter.println(tokenizer.nextToken());
                    }
                }
            }
        }
    }

    /**
     * Sleep the thread for a given number of seconds.
     *
     *
     * @param seconds
     * How many seconds to sleep.
     */
    public void sleepInSeconds(double seconds) {
        try {
            Thread.sleep((int) (seconds * 1000));
        } catch (InterruptedException e) {}
    }

}
