/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: ActKeyValidation.java,v 1.3.4.1 2000/10/19 17:59:08 jasona Exp $
 */
 
package com.lutris.multiServer;

import com.lutris.util.Config;
import com.lutris.logging.LogChannel;

/**
 * This interface is provided so that some sort of key checking or
 * validation on the right-to-run of a servlet is performed.
 *
 * Note that this isn't the most secure implementation of this schema;
 * as the wily hacker who is trying to circumvent checking can
 * just compile up their own version of whatever calls instantiations 
 * of this interface and put it before the original one in the classpath.
 *
 * This class can also be used as an initializer for a check or a set of
 * checks that will be called in the performance of an application to
 * see what level of survice the runner of the application has subscribed
 * to.
 *
 * For instance, if the following class if defined:
 *
 * <pre>
 * class Checkpoint implements com.lutris.multiServer.ActKeyValidation {
 *    ...
 *    boolean hasMinimalServiceContract() ...
 *    boolean hasExpensiveServiceContract() ...
 * }
 * </pre>
 *
 * Then other portions of the application can just call into that class's
 * <code>hasMinimalServiceContract</code> and 
 * <code>hasExpensiveServiceContract</code> methods to find out if 
 * extra-charge portions of the application are allowed to run.
 *
 * @see org.enhydra.servlet.multiServer.MultiServer#readConfigFile
 * @author Jason Abbott (jason@lutris.com)
 */

public interface ActKeyValidation {

    /**
     * Classes of this instance will only be instanciated and then have
     * this one function called to validate the currently running servlet's
     * right-to-run. If the validation routine is to do anything short of
     * halting the servlet then it should not throw an exception.
     *
     * @param config
     *		The Config structure of the configuration file. This will
     *		probably contain information you need in order to validate
     *		this servlet (such as an activation key or serial number).
     * @param logChannel
     *		This LogChannel is where any error or warning should be
     *		written to.
     * @exception ActKeyValidationException
     *		This Exception can be thrown if the servlet is detected as
     *          running an illegal copy. This will probably result in the
     *          catcher of this exception in halting the servlet.
     */
    public void validate(Config     config,
			 LogChannel logChannel)
        throws ActKeyValidationException;
}
