/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: ConfEditor.java,v 1.26.2.1.2.1 2000/10/19 17:59:08 jasona Exp $
 */





package com.lutris.appserver.admin.business;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Properties;
import java.util.StringTokenizer;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import com.lutris.util.Config;
import com.lutris.util.ConfigFile;
import com.lutris.util.ConfigException;
import com.lutris.util.KeywordValueTable;
import com.lutris.util.KeywordValueException;
import com.lutris.multiServer.MultiServer;

import org.enhydra.servlet.ServletContainer;
import org.enhydra.servlet.servletManager.ServletStatus;
import org.enhydra.servlet.servletManager.ServletManager;
import org.enhydra.servlet.connectionMethods.*;

import javax.servlet.ServletException;


/**
 * Methods used by the Enhydra administration application to read and
 * modify application's config files and servlet properties. Multiple
 * servlets may be edited at once, so the servlet id must be passed in
 * to all functions. <P>
 *
 * To use this class, simply call one of the get methods first. The
 * first get will trigger the reading of the config file (if the servlet
 * is an application), and the fetching of the servlet properties from
 * the ServletManager. <BR>
 * Then call the get or set methods in any order. All changes made via
 * the set methods are stored in memory in this class. No changes are
 * made to the application or the MultiServer until <CODE>save()</CODE>
 * is called. <BR>
 * When <CODE>save()</CODE> is called, the application's config file is
 * rewritten to reflect any changes made. If no values were changed,
 * no rewriting occurs. Calling a set method with the same value that was
 * returned from a get method does not count as a change. If the servlet
 * in question was not an Enhydra application, there is no config file for it,
 * so obviously it is not rewritten. If any changes to the servlet manager
 * properties (class name, class path, doc root, init args or description),
 * were made, then the old servlet is deleted from the MultiServer's
 * ServletManager, and the new servlet is added, reusing the same id. <P>
 *
 * The getXXXPairs methods return subsets of the key/value pairs found
 * in the applications config file. They return null for servlets that
 * are not Enhydra applications. The method <CODE>getSessionPairs()</CODE>
 * returns the keys that start with "SessionManager.". The method
 * <CODE>getDatabasePairs</CODE> returns the keys that start with
 * "DatabaseManager.". The method <CODE>getOtherPairs()</CODE> returns
 * keys that do not start with either ""SessionManager.",
 * ""DatabaseManager." or "Server.". Each of these is designed to return
 * keys found on one tab of the modify screen. The method
 * <CODE>get()</CODE> is general purpose, and can get any key. <P>
 *
 * Although there are several ways to get values from they config file,
 * there is only one method for modifying the values: <CODE>set()</CODE>.
 * NOTE: keys can not be created or delted, they can only have their
 * values changed.
 *
 *
 * @version	$Revision: 1.26.2.1.2.1 $
 * @author	Andy John
 */
public class ConfEditor {
    /**
     * Private constructor to prevent instantiation.
     */
    private void ConfEditor() {}

    // Stores one ConfigEditorEntry object for each servlet being modified.
    static private Hashtable allFiles = new Hashtable();

    /**
     *  Returns a Vector of KeyValuePair objects. Only keys that start
     *  with "SessionManager." are returned.
     *
     * @param servletID Which servlet to work on.
     * @return The SessionManager keys and values found in the application's
     * config file, or null if the servlet is not an Enhydra application.
     */
    synchronized static public Vector getSessionPairs(String servletID) {
        return getPairs(servletID, "SessionManager", null, null, null);
    }


    /**
     * Returns a Vector of KeyValuePair objects. Only keys that start
     * with "DatabaseManager." are returned.
     *
     * @param servletID Which servlet to work on.
     * @return The DatabaseManager keys and values found in the application's
     * config file, or null if the servlet is not an Enhydra application.
     */
    synchronized static public Vector getDatabasePairs(String servletID) {
        return getPairs(servletID, "DatabaseManager", null, null, null);
    }



    /**
     * Returns a Vector of KeyValuePair objects. Only keys that do not
     * start with "SessionManager.", "DatabaseManager." or "Server." are
     * returned.
     *
     * @param servletID Which servlet to work on.
     * @return The "other" keys and values found in the application's
     * config file, or null if the servlet is not an Enhydra application.
     */
    synchronized static public Vector getOtherPairs(String servletID) {
        return getPairs(servletID, null,
                        "Server", "SessionManager", "DatabaseManager");
    }



    /*
     * Fetch all keys that start with yes, but do not start with
     * any of the nos. They can be null.
     */
    static private Vector getPairs(String servletID,
                      String yes, String no1, String no2, String no3) {
        ConfEditorEntry ce = getEntry(servletID);
        if (ce == null)
            return null;
        if (ce.config == null)
            // Not an application. So no config file.
            return null;
        Config subsection = null;
        if (yes != null) {
            try {
                subsection = (Config) ce.config.getSection(yes);
            } catch (KeywordValueException kve) {
            }
        } else {
            subsection = ce.config;
        }
        if (subsection == null)
            return new Vector();
        Vector res = new Vector();
        String[] keys = subsection.leafKeys();
        for (int i=0; i<keys.length; i++) {
            String key = keys[i];
            if ((no1 != null) && key.startsWith(no1 + "."))
                continue;
            if ((no2 != null) && key.startsWith(no2 + "."))
                continue;
            if ((no3 != null) && key.startsWith(no3 + "."))
                continue;
            String value = null;
            boolean isArray = false;
            try {
                isArray = subsection.isArray(key);
                if (isArray)
                    value = Utils.joinArray(subsection.getStrings(key));
                else
                    value = subsection.getString(key);
            } catch (ConfigException e) {
                continue;
            }
            if (yes != null)
                key = yes + "." + key;
            res.addElement(new KeyValuePair(key, value, isArray));
        }
        return res;
    }


    /**
     *  Returns true if the value of the key is an array of strings, or
     *  false if it is a singe string. This determines how the string
     *  should be presented to the user. If the key is not found,
     *  or the servlet is not an Enhydra application, false is returned;
     *  use <CODE>get()</CODE> to tell if a key is present or not.
     *
     * @param servletID Which servlet to work on.
     * @param key The key to get.
     * @return True if the key is an array, else false.
     */
    synchronized static public boolean isArray(String servletID, String key) {
        ConfEditorEntry ce = getEntry(servletID);
        if (ce == null)
            return false;
        if (ce.config == null)
            // Not an application. So no config file.
            return false;
        boolean result = false;
        try {
            result = ce.config.isArray(key);
        } catch (ConfigException e) {}
        return result;
    }


    /**
     *  Get an arbitrary key from the application's config file.
     *  Returns null if the key was not found, or if the servlet is not
     *  an Enhydra application. <P>
     *
     *  If the value of the key is an array, then the array is joined
     *  together into one string with newlines. It should be presented
     *  to the user in a multi-line text entry box. Use the method
     *  <CODE>isArray()</CODE> to tell if the key is an array or not.
     *
     * @param servletID Which servlet to work on.
     * @param key The key to get.
     * @return The value of the key, or null.
     */
    synchronized static public String get(String servletID, String key) {
        ConfEditorEntry ce = getEntry(servletID);
        if (ce == null)
            return null;
        if (ce.config == null)
            // Not an application. So no config file.
            return null;
        String value = null;
        try {
            if (ce.config.isArray(key))
                value = Utils.joinArray(ce.config.getStrings(key));
            else
                value = (String) ce.config.getString(key);
        } catch (ConfigException e) {}
        return value;
    }


    /**
     *  Get the class name for the given servlet. This value is initally
     *  fetched from the ServletManager. Changes made via
     *  <CODE>setClassName()</CODE> are stored internally in memory,
     *  until <CODE>save()</CODE> is called.
     *
     * @param servletID Which servlet to work on.
     * @return The servlet's class name, or null if there was an error.
     */
    synchronized static public String getClassName(String servletID) {
        ConfEditorEntry ce = getEntry(servletID);
        if (ce == null)
            return null;
        return ce.status.className;
    }

    /**
     *  Set the class name for the given servlet. If the new value is the
     *  same as the old value, no action is taken. The new value is saved
     *  in memory; there will be no effect on the MultiServer until
     *  <CODE>save()</CODE> is called. At that time, if the values of
     *  class name, class path, doc root, init args or description have
     *  been changed, then the servlet is deleted from the MultiServer's
     *  ServletManager and a new servlet is added with the same id and
     *  the new data.
     *
     * @param servletID Which servlet to work on.
     * @param className The new class name for the servlet.
     */
    synchronized static public void setClassName(
                                       String servletID, String className) {
        ConfEditorEntry ce = getEntry(servletID, false);
        if (ce == null)
            return;
        if (!className.equals(ce.status.className)) {
            ce.status.className = className;
            ce.statusDirty = true;
        }
    }



    /**
     *  Get the classpath  for the given servlet. This value is initally
     *  fetched from the ServletManager. Changes made via
     *  <CODE>setClassPath()</CODE> are stored internally in memory,
     *  until <CODE>save()</CODE> is called. A single string is returned,
     *  all the classpath elements, newline separated. This value should
     *  be presented to the user for editing as an array.
     *
     * @param servletID Which servlet to work on.
     * @return The servlet's classpath, or null if there was an error.
     */
    synchronized static public String getClassPath(String servletID) {
        ConfEditorEntry ce = getEntry(servletID);
        if (ce == null)
            return null;
        return Utils.joinArray(ce.status.classPath);
    }

    /**
     *  Set the class path for the given servlet. If the new value is the
     *  same as the old value, no action is taken. The new value is saved
     *  in memory; there will be no effect on the MultiServer until
     *  <CODE>save()</CODE> is called. At that time, if the values of
     *  class name, class path, doc root, init args or description have
     *  been changed, then the servlet is deleted from the MultiServer's
     *  ServletManager and a new servlet is added with the same id and
     *  the new data. A single string should be passed in, it will be
     *  split at the newlines to make an array.
     *
     * @param servletID Which servlet to work on.
     * @param classPath The new class path for the servlet.
     */
    synchronized static public void setClassPath(
                                     String servletID, String classPath) {
        ConfEditorEntry ce = getEntry(servletID, false);
        if (ce == null)
            return;
        String cp = Utils.joinArray(ce.status.classPath);
        if (!cp.equals(classPath)) {
            ce.status.classPath = Utils.splitString(classPath);
            ce.statusDirty = true;
        }
    }



    /**
     *  Get the doc root for the given servlet. This value is initally
     *  fetched from the ServletManager. Changes made via
     *  <CODE>setDocRoot()</CODE> are stored internally in memory,
     *  until <CODE>save()</CODE> is called.
     *
     * @param servletID Which servlet to work on.
     * @return The servlet's doc root, or null if there was an error.
     */
    synchronized static public String getDocRoot(String servletID) {
        ConfEditorEntry ce = getEntry(servletID);
        if (ce == null)
            return null;
        return ce.status.docRoot;
    }

    /**
     *  Set the doc root for the given servlet. If the new value is the
     *  same as the old value, no action is taken. The new value is saved
     *  in memory; there will be no effect on the MultiServer until
     *  <CODE>save()</CODE> is called. At that time, if the values of
     *  class name, class path, doc root, init args or description have
     *  been changed, then the servlet is deleted from the MultiServer's
     *  ServletManager and a new servlet is added with the same id and
     *  the new data.
     *
     * @param servletID Which servlet to work on.
     * @param docRoot The new doc root for the servlet.
     */
    synchronized static public void setDocRoot(
                                       String servletID, String docRoot) {
        ConfEditorEntry ce = getEntry(servletID, false);
        if (ce == null)
            return;
       if (!docRoot.equals(ce.status.docRoot)) {
            ce.status.docRoot = docRoot;
            ce.statusDirty = true;
        }
    }



    /**
     *  Get the initial arguments for the given servlet. This value is initally
     *  fetched from the ServletManager. Changes made via
     *  <CODE>setClassName()</CODE> are stored internally in memory,
     *  until <CODE>save()</CODE> is called. The initial arguments are
     *  returned as a single string, newline separated. They should be
     *  presented to the user as an array for editing.
     *
     * @param servletID Which servlet to work on.
     * @return The servlet's initial arguments, or null if there was an error.
     */
    synchronized static public String getInitArgs(String servletID) {
        ConfEditorEntry ce = getEntry(servletID);
        if (ce == null)
            return null;
        return Utils.propertiesToString(ce.status.initArgs);
    }

    /**
     *  Set the initial arguments for the given servlet. If the new value
     *  is the same as the old value, no action is taken. The new value is
     *  saved in memory; there will be no effect on the MultiServer until
     *  <CODE>save()</CODE> is called. At that time, if the values of
     *  class name, class path, doc root, init args or description have
     *  been changed, then the servlet is deleted from the MultiServer's
     *  ServletManager and a new servlet is added with the same id and
     *  the new data. Pass the data in as a single string. It will be
     *  split at the newlines and converted into a Properties object.
     *
     * @param servletID Which servlet to work on.
     * @param initArgs The new initial arguments for the servlet.
     */
    synchronized static public void setInitArgs(
                               String servletID, String initArgs) {
        ConfEditorEntry ce = getEntry(servletID);
        if (ce == null)
            return;
        Properties newProps = Utils.stringToProperties(initArgs);
        if (!Utils.compareProperties(newProps, ce.status.initArgs)) {
            ce.status.initArgs = newProps;
            ce.statusDirty = true;
        }
    }



    /**
     *  Get the description for the given servlet. This value is initally
     *  fetched from the ServletManager. Changes made via
     *  <CODE>setClassName()</CODE> are stored internally in memory,
     *  until <CODE>save()</CODE> is called.
     *
     * @param servletID Which servlet to work on.
     * @return The servlet's description, or null if there was an error.
     */
    synchronized static public String getDescription(String servletID) {
        ConfEditorEntry ce = getEntry(servletID);
        if (ce == null)
            return null;
        return ce.status.description;
    }

    /**
     *  Set the description for the given servlet. If the new value is the
     *  same as the old value, no action is taken. The new value is saved
     *  in memory; there will be no effect on the MultiServer until
     *  <CODE>save()</CODE> is called. At that time, if the values of
     *  class name, class path, doc root, init args or description have
     *  been changed, then the servlet is deleted from the MultiServer's
     *  ServletManager and a new servlet is added with the same id and
     *  the new data.
     *
     * @param servletID Which servlet to work on.
     * @param description The new description for the servlet.
     */
    synchronized static public void setDescription(
                                       String servletID, String description) {
        ConfEditorEntry ce = getEntry(servletID, false);
        if (ce == null)
            return;
       if (!description.equals(ce.status.description)) {
            ce.status.description = description;
            ce.statusDirty = true;
        }
    }

    /**
     *  Get the defaultSessionTimeout for the WAR
     *
     * @param servletID Which servlet to work on.
     * @return The WAR session Timeout
     */
    synchronized static public String getDefaultSessionTimeout(String servletID) {
        ConfEditorEntry ce = getEntry(servletID);
        if (ce == null) {
            return null;
         }
        return ce.status.defaultSessionTimeOut;
    }

    /**
     *
     * @param servletID Which servlet to work on.
     * @param sessionTimeout
     */
    synchronized static public void setDefaultSessionTimeout(
                                       String servletID,
                                       String defaultSessionTimeOut) {
        ConfEditorEntry ce = getEntry(servletID, false);
        if (ce == null)
            return;
       if (!defaultSessionTimeOut.equals(ce.status.defaultSessionTimeOut)) {
            ce.status.defaultSessionTimeOut = defaultSessionTimeOut;
            ce.statusDirty = true;
        }
    }


    /**
     *  Get the IsWarExpanded for the WAR
     *
     * @param servletID Which servlet to work on.
     * @return Is the war expanded
     */
    synchronized static public String getIsWarExpanded(String servletID) {
        ConfEditorEntry ce = getEntry(servletID);
        if (ce == null)
            return null;
        return ce.status.isWARExpanded;
    }

    /**
     *
     * @param servletID Which servlet to work on.
     * @param isWarExpanded
     */
    synchronized static public void setIsWarExpanded(
                                       String servletID,
                                       String isWarExpanded) {
        ConfEditorEntry ce = getEntry(servletID, false);
        if (ce == null)
            return;
       if (!isWarExpanded.equals(ce.status.isWARExpanded)) {
            ce.status.isWARExpanded = isWarExpanded;
            ce.statusDirty = true;
        }
    }


    /**
     *  Get the IsWarExpanded for the WAR
     *
     * @param servletID Which servlet to work on.
     * @return Is the war expanded
     */
    synchronized static public String getIsWarValidated(String servletID) {
        ConfEditorEntry ce = getEntry(servletID);
        if (ce == null)
            return null;
        return ce.status.isWARValidated;
    }

    /**
     *
     * @param servletID Which servlet to work on.
     * @param isWarValidated
     */
    synchronized static public void setIsWarValidated(
                                       String servletID,
                                       String isWarValidated) {
        ConfEditorEntry ce = getEntry(servletID, false);
        if (ce == null)
            return;
       if (!isWarValidated.equals(ce.status.isWARValidated)) {
            ce.status.isWARValidated = isWarValidated;
            ce.statusDirty = true;
        }
    }


    /**
     *  Get the IsWarExpanded for the WAR
     *
     * @param servletID Which servlet to work on.
     * @return Is the War Directory expanded
     */
    synchronized static public String getIsInvokerEnabled(String servletID) {
        ConfEditorEntry ce = getEntry(servletID);
        if (ce == null)
            return null;
        return ce.status.isInvokerEnabled;
    }

    /**
     *
     * @param servletID Which servlet to work on.
     * @param isInvokerEnabled
     */
    synchronized static public void setIsInvokerEnabled(
                                       String servletID,
                                       String isInvokerEnabled) {
        ConfEditorEntry ce = getEntry(servletID, false);
        if (ce == null)
            return;
       if (!isInvokerEnabled.equals(ce.status.isInvokerEnabled)) {
            ce.status.isInvokerEnabled = isInvokerEnabled;
            ce.statusDirty = true;
        }
    }


    /**
     *  Get the IsWarExpanded for the WAR
     *
     * @param servletID Which servlet to work on.
     * @return Is the War Directory expanded
     */
    synchronized static public String getIsWorkDirPersistent(String servletID) {
        ConfEditorEntry ce = getEntry(servletID);
        if (ce == null)
            return null;
        return ce.status.isWorkDirPersistent;
    }

    /**
     *
     * @param servletID Which servlet to work on.
     * @param isWorkDirPersistent
     */
    synchronized static public void setIsWorkDirPersistent(
                                       String servletID,
                                       String isWorkDirPersistent) {

        ConfEditorEntry ce = getEntry(servletID, false);
        if (ce == null)
            return;
       if (!isWorkDirPersistent.equals(ce.status.isWorkDirPersistent)) {
            ce.status.isWorkDirPersistent = isWorkDirPersistent;
            ce.statusDirty = true;
        }
    }






    /**
     * Have the MultiServer/ServletStatus values been modified?
     * This ignores changes to the application config file.
     *
     * @param servletID Which servlet to work on.
     * @return True if the values of class name, class path,
     * doc root, init args or description have been changed.
     */
    synchronized static public boolean isDirty(String servletID) {
        ConfEditorEntry ce = getEntry(servletID, false);
        if (ce == null)
            return false;
        return ce.statusDirty;
    }




    /*
     * Open the file, and get the config objects ready to use.
     * Get the i/o streams, and leave them around. This will force the
     * file to stay open. We will need the output stream to write the
     * file later.
     * This adds the file to the hashtable, and returns it for use now.
     */
    synchronized static private ConfEditorEntry getEntry(String servletID) {
        return getEntry(servletID, true);
    }

    synchronized static private ConfEditorEntry getEntry(
                                 String servletID, boolean create) {
        ConfEditorEntry ce = (ConfEditorEntry) allFiles.get(servletID);
        if (ce != null) {
            return ce;
        } else {
            ce = new ConfEditorEntry();
	}

        /*
         *   Be sure the multiserver is ok.
         */
	ServletContainer sc = (ServletContainer)
		MultiServer.getService("ServletContainer");
	ServletManager sm = sc.getServletManager();
        if ((sm == null) || (sc.getConfigDirectory() == null)) {
            return null;
	}

        /*
         * Save a copy of the servlet status.
         */
        ServletStatus ss = sm.getStatus(servletID);
        if (ss == null) {
            return null;
	}
        ce.status = ss;

        /*
         *    Find the config file to read. This only applies to
         *    Enhydra applications.
         */
        if (Utils.isApplication(ss)) {
            /*
             * This servlet is an Enhydra application.
             */
            String confFilename = ss.initArgs.getProperty("ConfFile");
            if (confFilename == null)
                return null;
            /*
             *    Read the config file.
             */
            ce.filename = confFilename;
            ce.file = new File(confFilename);
            if (!ce.file.exists() ||
                !ce.file.isFile() ||
                !ce.file.canRead())
                return null;
            FileInputStream in = null;
            try {
                in = new FileInputStream(ce.file);
            } catch (FileNotFoundException e) {
                return null;
            }
            try {
                ce.configFile = new ConfigFile(in);
            } catch (ConfigException e) {
                return null;
            } finally {
                try {
                    in.close();
                } catch (IOException e) {
                    // Nothing we can do about it.
                }
            }
            ce.config = ce.configFile.getConfig();
        } else {
            /*
             * This servlet is not an Enhydra application.
             * Leave all the config file related variables in ce set to null.
             */
        }

        /*
         *   Save the file contents and the servlet status in the hashtable.
         */
        allFiles.put(servletID, ce);
        return ce;
    }


    /**
     *   Set the value of a key in the application's config file.
     *   The change is made to the copy in memory; no changes occur to the
     *   application's config file until <CODE>save()</CODE> is called.
     *   If the new value is the same as the old value, no action is
     *   taken. <P>
     *
     *   Assumes that the servletID had at least one getXXXTable()
     *   called on it. Assumes the key is already in the table.
     *   Returns true if sucessfull, false if an error. <P>
     *
     *   If the key (which already exists in the table) is an array, then
     *   the incomming string will be split at the newlines into an array.
     *
     *   @param servletID Which servlet to work on.
     *   @param key The key to set.
     *   @param value The value for the key.
     *   @return True if successfull.
     */
    synchronized static public boolean set(String servletID, String key, String value) {
        /*
         * Error checking.
         */
        ConfEditorEntry ce = getEntry(servletID, false);
        if (ce == null)
            // Never heard of that servlet ID before.
            return false;
        if (ce.config == null)
            // Not an application. So no config file.
            return false;
        /*
         * Get the current value (single string version).
         */
        String currentVal = null;
        try {
            if (ce.config.isArray(key))
                currentVal = Utils.joinArray(ce.config.getStrings(key));
            else
                currentVal = (String) ce.config.getString(key);
        } catch (ConfigException e) {}
        if (currentVal == null)
            // Don't add new keys.
            return false;
        /*
         * Has there been a change?
         */
        if (currentVal.equals(value))
            // No change, nothing to do.
            return true;
        /*
         * OK, save the new value.
         */
        try {
            if (ce.config.isArray(key))
                ce.config.set(key, Utils.splitString(value));
            else
                ce.config.set(key, value);
        } catch (KeywordValueException e) {
            // Uh oh...?
            return false;
        }
        ce.configDirty = true;
        return true;
    }


    /**
     *  Saves all the changes to the key/value pairs back to the
     *  config file where they came from.
     *  Returns null on sucess, else a human readable error string.
     *
     *  @param servletID Which servlet to work on.
     */
    synchronized static public String save(String servletID) {
        ConfEditorEntry ce = getEntry(servletID, false);
        if (ce == null)
            // Never heard of that servlet ID before.
            return "Illegal application id: " + servletID;
        /*
         * First re-rwite the config file, if any changes were made.
         */
        if (ce.configDirty) {
            // Re-write using the new keys in the config object.
            String tempName = ce.filename + ".tmp";
            File tempFile = null;
            FileOutputStream tempOut = null;
            try {
                tempFile = new File(tempName);
                tempOut = new FileOutputStream(tempFile);
            } catch (IOException e) {
                return "I/O Error writing to file " + tempName + ".";
            }
            if (tempOut == null)
                return "I/O Error writing to file " + tempName + ".";
            try {
                try {
                    ce.configFile.write(tempOut);
                } catch (Error e) {
                    return "Error writing to file " + tempName +
                           ": " + e.getMessage();
                }
            } finally {
                /*
                 * Be sure we close the open output stream.
                 */
                try {
                    tempOut.close();
                } catch (IOException e) {
                    return "Unable to close output stream after writing " +
                           tempName;
                }
            }
            /*
             * When running under windows, we need to close the output stream,
             * and delete the target file, or the rename will fail.
             */
            if (!ce.file.delete())
                return "Unable to remove old version of config file " +
                       ce.filename;
            if (!tempFile.renameTo(ce.file))
                return "Unable to rename " + tempName + " to " +
                       ce.filename + ".";
        }
        // Attempt to close the file. Will get GC'd sooner.
        ce.file = null;

	ServletContainer sc = (ServletContainer)
		MultiServer.getService("ServletContainer");
	ServletManager sm = sc.getServletManager();

        /*
         * Now re-register the servlet with the servlet manager,
         * if it has changed.
         */
        if (ce.statusDirty) {
            if (sm == null) {
                return "Error connecting to MultiServer.";
	    }
            ServletStatus ss = sm.getStatus(servletID);
            if (ss.running) {
                return "Unable to modify running servlet " + servletID;
	    }

           boolean isWar = ss.isWar;

            try {
                sm.delete(servletID);
            } catch (ServletException se) {
                return "Unable to delete old version of the servlet.";
            }

            try {
                sm.add(servletID, ce.status.className, ce.status.classPath,
                        ce.status.docRoot, ce.status.initArgs,
                	false,ce.status.description,isWar);

                sm.recordWarOnlySettings(servletID,
                        ce.status.defaultSessionTimeOut,ce.status.isWARExpanded,
                        ce.status.isWARValidated,ce.status.isInvokerEnabled,
                        ce.status.isWorkDirPersistent);

                int i = 0;
                while (i != ss.paths.size()) {
                    String path = (String) ss.paths.elementAt(i);
                    sm.recordContextPath(servletID,path);
                    i++;
                }

            } catch (ServletException se) {
                return "Unable to add new version of the servlet.";
            } catch (ConnectionMethodException cme) {
                // the ConnectionMethodException should never be thrown as
                // the Context/servlet will be fresh.
                cme.printStackTrace();
            }
        }
        // Forget about this config file.
        allFiles.remove(servletID);
        return null;
    }

}

