/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: DisplayRequests.java,v 1.21.10.1 2000/10/19 17:59:08 jasona Exp $
 */




package com.lutris.appserver.admin.presentation;

import java.net.*;
import java.io.*;
import java.util.Vector;
import com.lutris.appserver.admin.Admin;
import com.lutris.appserver.admin.business.*;
import com.lutris.http.*;
import com.lutris.appserver.server.httpPresentation.*;
import com.lutris.appserver.server.session.*;
import org.enhydra.servlet.ServletIOException;
import org.enhydra.servlet.debug.*;

/**
 * This is an http presentation object, which also implements the 
 * ServletRecordCallback interface, so it can register its self with the
 * DebugManager. <P>
 * This presentation tracks changes to the queue. In the future it will
 * talk to an applet, telling it to add and remove lines of text (each of
 * which is a URL). For now it just creates a neverending list of
 * anchors in normal HTML. Every time a new transaction is added to the
 * queue in the DebugManager, one more anchor appears, on it's own line.
 *
 * Expected query string inputs:
 *     id           The name of the servlet being debugged.
 *     targetUrl    The presentation object the anchors will point to.
 *     targetFrame  The target (frame name) for the generated anchors.
 *     history      true or false on whether history in queue is displayed.
 * 
 * @author Andrew John
 * @version $Revision: 1.21.10.1 $
 */
public class DisplayRequests implements HttpPresentation, 
       ServletRecordCallback {

    private Vector pendingRecords;

    public void run(HttpPresentationComms comms)
            throws IOException, PageRedirectException, Exception {
    //
    // Get the required fields from the query string.
    //
    String targetURL = comms.request.getParameter("targetUrl");
    if (targetURL == null)
        throw new HttpPresentationException("targetUrl not set.");
    String servletID = comms.request.getParameter("id");
    if (servletID == null)
        throw new HttpPresentationException("id (of servlet) not set.");
    String targetFrame = comms.request.getParameter("targetFrame");
    if (targetFrame == null)
        throw new HttpPresentationException("targetFrame not set.");

    //
    // Look up the DebugManager for this servlet.
    //
    if (comms.session == null)
        throw new HttpPresentationException("Null session manager.");
    Admin myApp = (Admin) comms.application;
    DebugManager dbm = (DebugManager) myApp.debuggers.get(servletID);
    if (dbm == null)
        throw new HttpPresentationException("Not debugging the servlet " +
                servletID);
    
    //
    //  Get ready to recieve records.
    //
    pendingRecords = new Vector();

    //
    // Register for callbacks.
    //
    String history = comms.request.getParameter("history");
    HttpPresentationOutputStream out = null;
    dbm.addTransactionCallback(this, history.equals("true"));
    try {
        /*
         * Start the output! This will not stop until the user clicks on
         * the stop button, or they click on one of the anchors.
         */ 
	comms.response.setContentType("text/html");
        out = comms.response.getOutputStream();
        out.println("<HTML>\n<HEAD>\n</HEAD>\n" + 
            "<BODY BACKGROUND=media/BackgroundWide.jpg>\n");
        out.println("<CODE><FONT SIZE=-1><SCRIPT></SCRIPT>\n");
        out.flush();
        synchronized(pendingRecords) {
            while (true) {
                while (!pendingRecords.isEmpty()) {
                    ServletTransactionRecord tr =
                        (ServletTransactionRecord) pendingRecords.firstElement();
                    pendingRecords.removeElement(tr);
                    RequestRecord rr = tr.getRequestRecord();
                    if (rr == null)
                        continue;
                    String status = ""; 
                    ResponseRecord resp = tr.getResponseRecord();
                    if (resp != null) {
                        int sc = resp.getStatusCode();
                        if (sc == -1)
                            sc = 200;
                        if (sc == 200)
                            status = sc + " ";
                        else if ((sc > 200) && (sc <=299))
                            status = "<FONT COLOR=green>" + sc + "</FONT> ";
                        else if ((sc >= 300) && (sc <=399))
                            status = "<FONT COLOR=yellow>" + sc + "</FONT> ";
                        else if ((sc >= 400) && (sc <=499))
                            status = "<FONT COLOR=red>" + sc + "</FONT> ";
                        else 
                             status = "<FONT COLOR=orange>" + sc + "</FONT> ";
                    }
                    int id = tr.getId();
                    String host = rr.getRemoteHost();
                    String url = rr.getPathInfo();
                    String method = rr.getMethod();
                    if (host == null)
                        host = "<I>N/A</I>";
                    else if (host.indexOf(".") > 0)
                        host = host.substring(0, host.indexOf("."));
                    if (host.length() > 9)
                        host = host.substring(0, 6) + "...";
                    if (url == null)
                        url = "&quot;&quot;";
                    else if (url.length() > 16)
                        url = "..." + url.substring(url.length() - 13);
                    if (method == null)
                        method = "<I>N/A</I>";
                    out.println(
                        status + host + " <A HREF=" +
                        targetURL +
                        "?id=" + servletID +
                        "&transaction=" + id + 
                        " TARGET=" + targetFrame + ">" +
                        method + "</A> "  + 
                        url + "<BR>\n");
                    out.println("<SCRIPT>window.scroll(0,32000)</SCRIPT>\n");
                    out.flush();
                }

                /*
                 * If the browser broke off the connection, this should
                 * force a network io exception.
                 */
                comms.request.getInputStream().available();

                /*
                 *  We processed all the pending records. Now wait for more.
                 */ 
                pendingRecords.wait(2000);
            }
        }
    } catch (ServletIOException e) {
        /*
         * This is expected, when the user hits "stop" in the browser.
         * Ignore this exception. All others will pass though and get 
         * caught by the standard mechanisms.
         */
    } finally {
        /*
         * No matter what, do this clean up code.
         */
        dbm.removeTranactionCallback(this);
        out.println("</CODE></FONT></BODY>\n</HTML>\n");
        out.flush();
    }
}


    /*
        For now, just add the new one. Ignore deleting elements until
        we are talking to an applet.
    */
    public void transactionRecordAddEvent(
        ServletTransactionRecord newRecord,
        ServletTransactionRecord deletedRecord) {
        synchronized(pendingRecords) {
            if (newRecord == null)
                return;
            pendingRecords.addElement(newRecord);
            pendingRecords.notifyAll();
        }
    }


    public void clearRecordsEvent() {
    }

}


