/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: ServletContainer.java,v 1.7.2.4.2.2 2000/10/19 17:59:09 jasona Exp $
 */

package org.enhydra.servlet;

import java.io.*;
import java.net.*;
import java.rmi.RemoteException;
import java.util.*;
import java.text.DateFormat;
import javax.servlet.ServletException;
//import java.text.MessageFormat;

import org.enhydra.servlet.connectionMethods.*;
import org.enhydra.servlet.filter.*;
import org.enhydra.servlet.servletManager.*;
import com.lutris.util.*;
import com.lutris.appserver.server.session.*;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

import com.lutris.logging.*;
import com.lutris.multiServer.Service;

import org.apache.tomcat.core.*;
import org.apache.tomcat.util.URLUtil;
import org.enhydra.i18n.ResManager;


/**
 * Handles requests for Servlets. Multiple connection methods may be
 * routed to one Servlet, or one connection method may direct requests
 * to multiple Servlets (based on URL). <P>
 *
 * Servlets or Enhydra applications (which are a special kind of servlet)
 * or Servlet 2.2 Web applications may be served. <P>
 *
 * This class uses a ServletManager to hold a set of Servlets,
 * a ConnectionMethodManager to hold a set of ConnectionMethods,
 * and a FilterManager to hold a set of Filters. These
 * are available to other classes (for example the
 * admin application). <P>
 *
 * At startup time a configuration file is read, and it describes what
 * ConnectionMethods to create, what Servlets to register, and the connections
 * to make from the ConnectionMethods to the Servlets. It also describes
 * the filters to create and what channels to install them in. These filters
 * are all descendants of the MultServerFilter class. <P>
 *
 * After this config file is read, and the servlet container is set up, no
 * further action is taken. The thread that ran through start() exits.
 * The ConnectionMethods must create threads, and they are responsible for
 * receiving requests from the outside world and passing them on to the
 * handleRequest() methods of the ServletContainer. <P>
 *
 * Normally one of the Servlets will be an administration application. This
 * admin app knows about this class, and so it knows how to talk to the
 * current ConnectionMethods and Servlets. It may change the server setup
 * on the fly. For example, it can create and add a new ConnectionMethod.
 * Or it could add a new channel to an already existing ConnectionMethod.
 * Or register a new Servlet. Or add a logging filter to a channel.
 * Or monitor performance. <P>
 *
 * @see org.enhydra.servlet.connectionMethod.ConnectionMethod
 * @see org.enhydra.servlet.connectionMethod.ConnectionMethodManager
 * @see org.enhydra.servlet.filter.FilterManager
 * @see org.enhydra.servlet.servletManager.ServletManager
 * @see org.enhydra.servlet.filter.Filter
 * @see javax.servlet.Servlet
 * @author Shawn McMurdo
 * @author Andy John
 */
public class ServletContainer implements Service {

    /**
     * The manager for all the ConnectionMethods in use by the server.
     */
    private ConnectionMethodManager connectionMethodManager;

    /**
     * The manager for all the Servlets in use by the server.
     */
    private ServletManager servletManager;

    /**
     * The manager for all the Filters (which will all be
     * descendants of MultiServerFilter) in use by the server.
     */
    private FilterManager filterManager;

    /**
     * Has the servlet container been initialized and started?
     */
    private boolean started = false;

    /**
     * The directory where Enhydra application config files
     * are expected to reside. Only the admin application is expected to need
     * to use this.
     */
    private String configDirectory;

    /*
     * The default config file to use if none is specified.
     */
    private static final String defaultConfigFileName = "./multiserver.conf";

    /*
     * The config file that was actually read.
     */
    private String configFileName;

    /*
     *   This is the file we read to start up the server. We need to keep
     *   it around so that we can save the current settings back to the same
     *   file. The Config object will preserve comments.
     */
    private ConfigFile configFile;

    /*
     *   This is the keyword value table that was generated by the Config
     *   object above. When writing the config file, this is used as the
     *   starting point.
     */
    public Config config;

    private LogChannel logChannel;

    // Strings used in config file names
    public static final String CONF_DIR = "ConfDir";
    public static final String SERVER = "Server";

    public static final String APPLICATION = "Application";
    public static final String CONF_FILE = "ConfFile";
    public static final String SERVLET = "Servlet";
    public static final String CLASS_NAME = "ClassName";
    public static final String DOC_ROOT = "DocRoot";
    public static final String DESCRIPTION = "Description";
    public static final String RUNNING = "Running";
    public static final String CLASS_PATH = "ClassPath";
    public static final String INIT_ARGS = "InitArgs";
    public static final String URL = "Url";
    public static final String CONNECTION = "Connection";
    public static final String FILTERS = "Filters";
    public static final String CHANNEL = "Channel";
    public static final String FILTER = "Filter";
    public static final String ENABLED = "Enabled";
    public static final String APP_CLASS = "AppClass";
    public static final String PRESENTATION_PREFIX = "PresentationPrefix";
    public static final String AUTO_RELOAD = "AutoReload";
    public static final String SESSION_MANAGER = "SessionManager";

    // config info for a wab application
    public static final String DEFAULT_SESSION_TIMEOUT = "defaultSessionTimeOut";
    public static final String IS_WAR_EXPANDED         = "isWARExpanded";
    public static final String IS_WAR_VALIDATED        = "isWARValidated";
    public static final String IS_INVOKER_ENABLED      = "isInvokerEnabled";
    public static final String IS_WORKDIR_PERSISTENT   = "isWorkDirPersistent";

    //https stuff
    public static final String SSL = "SSL";
    public static final String SERVER_CERTIFICATE = "ServerCertificate";
    public static final String SECURE_RANDOM_ALGORITHM ="SecureRandomAlgorithm";
    public static final String SECURE_RANDOM_PROVIDER ="SecureRandomProvider";
    public static final String SSL_CONTEXT_PROVIDER = "SSLContextProvider";
    public static final String SSL_CONTEXT_PROTOCOL = "SSLContextProtocol";
    public static final String KEY_STORE_LOCATION ="KeyStoreLocation";
    public static final String KEY_STORE_PROVIDER = "KeyStoreProvider";
    public static final String KEY_MANAGER_ALGORITHM ="KeyManagerAlgorithm";
    public static final String KEY_MANAGER_PROVIDER = "KeyManagerProvider";
    public static final String TRUST_MANAGER = "TrustManager";
    public static final String PASSWORD = "Password";
    public static final String CLIENT_AUTHENTICATION = "clientAuthentication";

    // Strings used in config file values
    public static final String HTTP = "http";
    public static final String HTTPS= "https";
    public static final String CGI = "cgi";
    public static final String RMI = "rmi";
    public static final String WAI = "wai";
    public static final String APACHE = "apache";
    public static final String FULL = "full";
    public static final String NONE = "none";
    public static final String YES = "yes";
    public static final String NO = "no";

    // Class name of an enhydra app
    public static final String ENYHDRA_APP_CLASSNAME =
      "com.lutris.appserver.server.httpPresentation.servlet.HttpPresentationServlet";


    private StandardSessionManager  standardSessionManager = null;
    private String yesOrNo = YES+","+NO;

    private ResManager rez;


    /**
     * Default constructor.
     */
    public ServletContainer() {
        // Gets a string  manager wrapped around "org.enhydra.servlet.Res"
        rez = ResManager.getResManager(this.getClass());
    }

    /**
     * Default initialization.  Uses the existing config file if previously
     * set, otherwise uses the default config file.
     */
    public void init() {
	if (configFileName == null) {
	    init(defaultConfigFileName);
	} else {
	    init(configFileName);
	}
    }

    /**
     * Initialization uses the passed in config file.
     *
     * @param configFileName
     *		a String filename to be read for configuration.
     */
    public void init(String configFileName) {
	this.configFileName = configFileName;
	logChannel = Logger.getCentralLogger().getChannel("Multiserver");
    }

    /**
     * Start the servlet container.
     */
    public void start() {
        // Get the server ready to read it's config file.
        initialize();

	// X22
	started = true;

        // Read the config file.
        readConfigFile(configFileName);

	// X22 Read the server.xml file
	// readServerConfig(new String[0]);
    }

    /**
     *   Get the server ready to read it's config file. All the
     *   manager objects start off empty. There are no "built in"
     *   servlets, connection methods, channels, or filters.
     *   They must all be specified in the config file. This
     *   gives more control to the administrator.
     */
    private void initialize() {
        connectionMethodManager = new ConnectionMethodManager();
        servletManager = new ServletManager();
        filterManager = new FilterManager();

    }

    /**
     *  Stop the servlet container.
     *
     *  This fucnction disables all the channels in all the
     *  connection methods, then destroys all the connection
     *  methods. This causes them to kill all their threads.
     *  When the last thread exits, the java application will
     *  exit.
     */
    public void stop() {
	started = false;

        String[] cmIDs = connectionMethodManager.getIDs();

        // Disable all channels first.
        for (int i=0; i<cmIDs.length; i++) {
            ConnectionMethod cm = connectionMethodManager.get(cmIDs[i]);
            if (cm == null)
                continue;
            String[] chIDs = cm.getChannelIDs();
            for (int j=0; j<chIDs.length; j++) {
                try {
                    cm.disableChannel(chIDs[j]);
                } catch (ConnectionMethodException cme) {
                }
            }
        }

        // Now destroy all ConnectionMethods.
        for (int i=0; i<cmIDs.length; i++) {
            ConnectionMethod cm = connectionMethodManager.get(cmIDs[i]);
            if (cm == null)
                continue;
            try {
                cm.destroy();
            } catch (ConnectionMethodException cme) {
            }
        }

        /*
         *   Shut down the servlet manager. It will shut down all the
         *   servlets, and stop it's trans/min monitoring thread.
         */
         if (servletManager != null) {
            servletManager.destroy();
            servletManager = null;
        }
    }

    /**
     * Handle a servlet request.
     *
     * @param request
     *		a Tomcat Request.
     * @param response
     *		a Tomcat Response.
     */
     /*
    public void handleRequest(Request request, Response response) {
	if (!started) {
	    init();
	    start();
	}
	contextManager.service(request, response);
    }
    */

    /**
     * Get the connection manager for the ServletContainer.
     *
     * @return the ConnectionMethodManager
     */
    public ConnectionMethodManager getConnectionMethodManager() {
	if (!started) {
	    init();
	    start();
	}
	return connectionMethodManager;
    }

    /**
     * Get the servlet manager for the ServletContainer.
     *
     * @return the ServletManager
     */
    public ServletManager getServletManager() {
	return servletManager;
    }


    /**
     * Get the servlet manager for the ServletContainer.
     *
     * @return the ServletManager
     */
    public SessionManager getSessionManager() {
	return standardSessionManager;
    }



    /**
     * Get the filter manager for the ServletContainer.
     *
     * @return the FilterManager
     */
    public FilterManager getFilterManager() {
    // This can cause recursive calls to init() and start()
    // after an error has shut down the ServletContainer
    //
    /*
	if (!started) {
	    init();
	    start();
	}
     */
	return filterManager;
    }


    /**
     * Get the config directory used by this ServletContainer to find apps.
     * X22 This seems like it should be in the MultiServer Admin rather
     * X22 than here.
     *
     * @return a String config directory.
     */
    public String getConfigDirectory() {
	return configDirectory;
    }

    /**
     * Read the given config file.
     *
     * @param configFileName
     *		a String config file name.
     */
    private void readConfigFile(String configFileName) {

        try {
	    FileInputStream configFIS = new FileInputStream(configFileName);
	    configFile = new ConfigFile(configFIS);
            config = configFile.getConfig();
            parseServerSection(myGetSection(config, SERVER, false));
            configFIS.close();
        } catch (Throwable t) {
            // the stack trace (if any) will already be reported.
            String msg = rez.format("Error encountered while reading config file {0}.",configFileName);
            logChannel.write(Logger.ERROR, msg);
            return;
        }

	KeywordValueTable applications =
	    myGetSection(config, APPLICATION, false);
	KeywordValueTable servlets =
	    myGetSection(config, SERVLET, false);
	KeywordValueTable channels =
	    myGetSection(config, CHANNEL, false);
	KeywordValueTable filters =
	    myGetSection(config, FILTER, false);


	// FIX error handling and reporting.
        try {
            if (filters != null) {
                String[] keys = filters.keys();
                for (int i=0; i<keys.length && started; i++)
                    addFilter((Config)filters.getSection(keys[i]), keys[i]);
            }
            if (applications != null && started) {
                String[] keys = applications.keys();
                for (int i=0; i<keys.length && started; i++)
                    addApplication((Config)applications.getSection(keys[i]),
                                   keys[i]);
            }
            if (servlets != null && started) {
                String[] keys = servlets.keys();
                for (int i=0; i<keys.length && started; i++)
                    addServlet((Config)servlets.getSection(keys[i]),
                               keys[i]);
            }

	    try {
                if (started)
                   connectionMethodManager.initConnectionMethods(config);
	    } catch (ConnectionMethodException e) {
                // stack trace is already reported.
                String msg = rez.format("Exiting the server.");
		logChannel.write(Logger.ERROR, msg);
		System.err.println(msg);
		Runtime vm = Runtime.getRuntime();
                // This is deprecated in JDK 1.2
		Runtime.runFinalizersOnExit(true);
		vm.exit(1);
	    }

            if (channels != null && started) {
                String[] keys = channels.keys();
                for (int i=0; i<keys.length && started; i++)
                    addChannels((Config)channels.getSection(keys[i]), keys[i]);
             }
        } catch (Throwable e) {
            String msg = rez.format("Error encountered while reading config file {0}.",configFileName);
            logChannel.write(Logger.ERROR, msg, e);
            return;
        }

        String workDir = "work";
        int i =  configDirectory.lastIndexOf('/');
        if (i >0) {
            workDir = configDirectory.substring(0,i) +
                        File.separator + "work";
        }
        getServletManager().setWorkDir(workDir);

        if (started) {
            try {
                getServletManager().startAllContexts();
            } catch (ServletException se) {
                throw new ServletError(se);
            }
        }
    }

    /**
     * Retrieve the server parameters from the config file and
     * establish the logging facility.
     *
     * @param server
     *   the config object for the server section.
     */
    private void parseServerSection(Config server) {
        configDirectory = myGetString(server, CONF_DIR, SERVER);
    }


    private void addApplication(Config info, String id) {

        if (info == null)
            return;

        String running = myGetString(info, RUNNING, APPLICATION + "." + id);

        if (!running.equalsIgnoreCase(YES) && !running.equalsIgnoreCase(NO)) {
            String msg = rez.format("Illegal running entry {0} for application {1}. Expected one of {2}.",
                    running,id,yesOrNo);
            fatalErr(msg);
            return;
        }

        String conf = myGetString(info, CONF_FILE, APPLICATION + "." + id,false);
        boolean bRunning = running.equalsIgnoreCase(YES);
        if (conf != null)  {
            addEnhydraApplication(info,id,bRunning,conf);
        }
        else {
            addWarApplication(info,id,bRunning);
        }

    }

    private void addEnhydraApplication(Config info, String id,boolean runOnInit,
                                       String conf) {

        String description = myGetString(info, DESCRIPTION, APPLICATION + "." + id);
        /*
         *    Read the application's config file.
         */
	String confFilename = conf;
	if (!confFilename.startsWith(File.separator)) {
	    confFilename = configDirectory + File.separator + conf;
	}

        File confFile = new File(confFilename);

        if (!confFile.exists()) {
            fatalErr(rez.format("No conf file {0}" ,confFilename));
            return;
        }

        if (!confFile.isFile()) {
            fatalErr(rez.format("Not a file: {0}",confFilename));
            return;
        }

        Config cfg = null;
        try {
	    FileInputStream cfgFIS = new FileInputStream(confFilename);
	    ConfigFile cfgF = new ConfigFile(cfgFIS);
            cfgFIS.close();
            cfg = cfgF.getConfig();
        } catch (IOException ioe) {
            fatalErr("IO error reading " + confFilename +
                     ":\n" + ioe.getMessage(), ioe);
            return;
        } catch (ConfigException ce) {
            fatalErr("Bad file format reading " + confFilename +
                     ":\n" + ce.getMessage(),ce);
            return;
        }

        /*
         *   Get the classpath from the config file. This is an optional
         *   field.
         */
        String[] classPath = new String[0];
        if (cfg.containsKey(SERVER + "." + CLASS_PATH)) {
            try {
                classPath = cfg.getStrings(SERVER + "." + CLASS_PATH);
            } catch (ConfigException ce) {
                fatalErr("Invalid config file " +
                      confFilename + ". No key " + SERVER + "." + CLASS_PATH);
                return;
            }
        }

        /*
         *    Get the docRoot from the config file.
         *    FOR NOW ENHYDRA APPS DO NOT NEED THIS.
         */
        String docRoot = "/";   // docRoot will be fed to the context it should not be null
	// X22 ?
        if (cfg.containsKey(SERVER + "." + DOC_ROOT)) {
            try {
                docRoot = cfg.getString(SERVER + "." + DOC_ROOT);
            } catch (ConfigException ce) {
                fatalErr("Error getting " + SERVER + "." + DOC_ROOT, ce);
                return;
            }
        }

        /*
         *    Create the initial arguments to the servlet.
         */
        Properties initArgs = new Properties();
        initArgs.put(CONF_FILE, confFilename);

        /*
         *    Go ahead and register the new servlet.
         */
	logChannel.write(Logger.DEBUG, "ADDING APPLICATION:");
	logChannel.write(Logger.DEBUG, "ID = " + id);
	for (int z=0; z<classPath.length; z++)
	    logChannel.write(Logger.DEBUG, "CP [" + z + "] = " + classPath[z]);
	logChannel.write(Logger.DEBUG, "initArgs = " + initArgs);
	logChannel.write(Logger.DEBUG, "desc = " + description);

        try {
            servletManager.add(id,
                "com.lutris.appserver.server.httpPresentation." +
                "servlet.HttpPresentationServlet",
                classPath, docRoot,
                initArgs, runOnInit,description,false);
        } catch (ServletException sme) {
            fatalErr("Error adding application" + id + " to server:\n" +
                     sme.getMessage(),sme);
        }

    }

    // The application did not define a confFile so we assume it is a WAR
    // There must be a docRoot
    private void addWarApplication(Config info, String id,boolean runOnInit)
            {



        Config sessionConfig = null;
        try {
            sessionConfig = (Config)info.getSection(SESSION_MANAGER);
        }
        catch (KeywordValueException e) {
            fatalErr(rez.format("Unable to read config file."),e);
        }

        String app = APPLICATION + "." + id;
        String description = myGetString(info, DESCRIPTION,app);
        String docRoot = myGetString(info, DOC_ROOT, APPLICATION + "." + id,false);

        if (docRoot == null) {
            String msg = rez.format("DocRoot required for Web Application {0}.",id);
            logChannel.write(Logger.ERROR, msg);
            return;
        }

        File docRootFile = new File(docRoot);
        if (!docRootFile.exists()) {
            String msg;
            msg = rez.format("DocRoot \"{0}\" not found for Web Application {1}.",docRoot,id);
            logChannel.write(Logger.ERROR, msg);
            if (runOnInit) {
                msg = rez.format("Web Application {0} will not be started.",id);
                logChannel.write(Logger.INFO, msg);
                runOnInit = false;
            }
        }
        else if (!docRootFile.isDirectory()) {
            // untill tomcat expands the wars.
            String rezKey = "DocRoot \"{0}\" for apllication {1} is not a directory .";
            String msg = rez.format(rezKey,docRoot,id);
            logChannel.write(Logger.WARNING, msg);
        }


        // all of these are optional.
        String defaultSessionTimeout = myGetString(info, DEFAULT_SESSION_TIMEOUT, app, false);
        String isWARExpanded         = myGetString(info, IS_WAR_EXPANDED, app, false);
        String isWARValidated        = myGetString(info, IS_WAR_VALIDATED, app, false);
        String isInvokerEnabled      = myGetString(info, IS_INVOKER_ENABLED, app, false);
        String isWorkDirPersistent   = myGetString(info, IS_WORKDIR_PERSISTENT, app, false);

        Properties initArgs = new Properties();
        String[] classPath = new String[0];

        try {
            servletManager.add(id,
                "", // className
                classPath,
                docRoot, initArgs, runOnInit,
                description,true);
        } catch (ServletException sme) {
            String rezKey = "Error adding application {0} to server.";
            String msg = rez.format(rezKey,id);
            logChannel.write(Logger.ERROR, msg,sme);
            return;
        }

        // now add in the optional settings
        servletManager.recordWarOnlySettings(id, defaultSessionTimeout,
                                                 isWARExpanded,
                                                 isWARValidated,
                                                 isInvokerEnabled,
                                                 isWorkDirPersistent );

        servletManager.recordSessionConfig(id,sessionConfig);

    }

    private void addServlet(Config info, String id) {

        if (info == null)
            return;

        String className   = myGetString(info, CLASS_NAME, SERVLET + "." + id);
        String root        = myGetString(info, DOC_ROOT,  SERVLET + "." + id);
        String description = myGetString(info, DESCRIPTION, SERVLET, false);

        if (description == null)
            description = "";

        String running = myGetString(info, RUNNING, SERVLET + "." + id);

        if (!running.equalsIgnoreCase(YES) && !running.equalsIgnoreCase(NO)) {

            String rezKey = "Illegal running entry {0} for servlet {1}. Expected one of {2}.";
            String msg = rez.format(rezKey,running,id,yesOrNo);
            fatalErr(msg);
            return;
        }


        /*
         *   ClassPath (optional) is a comma-separated list.
         */
        String[] classPath = new String[0];
        if (info.containsKey(CLASS_PATH)) {
            try {
                classPath = info.getStrings(CLASS_PATH);
            } catch (ConfigException ce) {
                // resource bundle this string
                String rezKey = "Illegal class path specified for servlet {0}.";
                String msg = rez.format(rezKey,id);
                fatalErr(msg,ce);
                return;
            }
        }
        /*
         *  Init (optional) is a section.
         */
        Config init = null;
        if (info.containsKey(INIT_ARGS)) {
            try {
                init = (Config) info.getSection(INIT_ARGS);
            } catch (KeywordValueException e) {
                fatalErr("Error getting " + INIT_ARGS, e);
                return;
            }
        } else {
            init = new Config();
        }
        Properties initArgs = new Properties();
        String[] keys = init.keys();
        for (int i=0; i<keys.length; i++) {
            String val = null;
            try {
                val = init.getString(keys[i]);
            } catch (ConfigException e) {}
            if (val == null)
                val = "";
            initArgs.put(keys[i], val);
        }
        logChannel.write(Logger.DEBUG, "ADDING SERVLET:");
        logChannel.write(Logger.DEBUG, "ID = " + id);
        logChannel.write(Logger.DEBUG, "CN = " + className);
        for (int z=0; z<classPath.length; z++)
            logChannel.write(Logger.DEBUG, "CP [" + z + "] = " + classPath[z]);
        logChannel.write(Logger.DEBUG, "root = " + root);
        logChannel.write(Logger.DEBUG, "initArgs = " + initArgs);
        logChannel.write(Logger.DEBUG, "desc = " + description);
        try {
            servletManager.add(id, className,
                               classPath, root, initArgs,
                               running.equalsIgnoreCase(YES),
                               description,false);
        } catch (ServletException e) {
            String rezKey = "Unable to add servlet {0}.";
            String msg = rez.format(rezKey,id);
            fatalErr(msg,e);
            return;

        }
    }



    private void addChannels(Config info, String cmID) {
        logChannel.write(Logger.DEBUG, "ADD CHANNEL : " + cmID);
        if (info == null)
            return;
        ConnectionMethod cm = connectionMethodManager.get(cmID);
        if (cm == null) {
            String rezKey = "Unable to add channel: illegal connection method id {0}.";
            fatalErr(rez.format(rezKey,cmID));
        }
        String[] chanIDs = info.keys();
        for (int i=0; i<chanIDs.length; i++) {
            Config props = null;
            try {
                props = (Config) info.getSection(chanIDs[i]);
            } catch (ConfigException e) {
            } catch (KeywordValueException e) {
            }
            if (props == null) {
                String rezKey = "Error reading channel id {0} for connection method {1}.";
                String msg = rez.format(rezKey,chanIDs[i],cmID);
                fatalErr(msg);
            }

            String servletID = myGetString(props, SERVLET,
                CHANNEL + "." + cmID + "." + chanIDs[i]);
            String url = myGetString(props, URL,
                CHANNEL + "." + cmID + "." + chanIDs[i]);
            String enabled = myGetString(props, ENABLED,
                CHANNEL + "." + cmID + "." + chanIDs[i]);
            if (!enabled.equalsIgnoreCase(YES) &&
                !enabled.equalsIgnoreCase(NO)) {
                String rezKey = "Illegal enabled value {0} for channel {1} in connection method {2}. Expected one of {3}.";
                String msg = rez.format(rezKey,enabled,chanIDs[i],cmID,yesOrNo);
                fatalErr(msg);
            }
            String filters = myGetString(props, FILTERS, CONNECTION, false);
            if (filters == null)
                filters = "";
            try {
            //	contextFromConfig(props,cmID,cm);
                cm.addChannel(chanIDs[i], url, servletID);
            } catch (ConnectionMethodException e) {
                String rezKey = "Error adding channel {0} to connection method {1}.";
                String msg = rez.format(rezKey,chanIDs[i],cmID);
                fatalErr(msg,e);
            }
            /*
             *    Now try to add all the requested filters (if any).
             */
            StringTokenizer stok = new StringTokenizer(filters, ",");
            while (stok.hasMoreTokens()) {
                String id = trim(stok.nextToken());
                Filter tf = filterManager.get(id);
                if (tf == null) {
                    String rezKey = "Channel {0} in connection method {1} has illegal filter id {2}.";
                    String msg = rez.format(rezKey,chanIDs[i],cmID,id);
                    fatalErr(msg);
                }
                try {
                    cm.addTransactionFilter(chanIDs[i], id);
                } catch (ConnectionMethodException cme) {
                    String rezKey = "Error adding filter {0} to channel {1} in connection method {2}.";
                    String msg = rez.format(rezKey,id,chanIDs[i],cmID);
                    fatalErr(msg, cme);
                }
            }

            /*
             *    Maybe enable the channel.
             */
            if (enabled.equalsIgnoreCase(YES)) {
                try {
                    cm.enableChannel(chanIDs[i]);
                } catch (ConnectionMethodException e) {
                    String rezKey = "Unable to enable channel {0} in connection method {1}.";
                    String msg = rez.format(rezKey,chanIDs[i],cmID);
                    fatalErr(msg, e);
                }
            }
        }
    }

    private void addFilter(Config info, String id) {
        if (info == null)
            return;
        String classname = myGetString(info, CLASS_NAME, FILTER);
        String description = myGetString(info, DESCRIPTION, FILTER);
        /*
         *   Init (optional) is a section.
         */
        KeywordValueTable init = null;
        try {
            init = info.getSection(INIT_ARGS);
        } catch (KeywordValueException e) {
        }
        /*
         *   Try to create the filter. It must
         *   have a constructor that takes a String and a
         *   KeywordValueTable.
         */
        Class c = null;
        try {
            c = Class.forName(classname);
        } catch (ClassNotFoundException cnfe) {
            String rezKey = "Unable to find filter class {0}.";
            String msg = rez.format(rezKey,classname);
            fatalErr(msg, cnfe);
        }
        try {
            Class tf = Class.forName("org.enhydra.servlet.filter.Filter");
            if (!(tf.isAssignableFrom(c))) {
                String rezKey = "Filter class {0} does not extend Filter.";
                String msg = rez.format(rezKey,classname);
                fatalErr(msg);
            }
        } catch (ClassNotFoundException cnfe) {
            fatalErr(rez.format("Error loading filter {0}.",id),cnfe);
        }
        Constructor cnst = null;
        try {
            Class[] types = new Class[2];
            types[0] = null;
            types[1] = null;
            try {
                types[0] = Class.forName("java.lang.String");
                types[1] = Class.forName("com.lutris.util.KeywordValueTable");
            } catch (ClassNotFoundException cnfe) {
                fatalErr(rez.format("Error loading filter {0}.",id),cnfe);
            }
            cnst = c.getConstructor(types);
        } catch (NoSuchMethodException nsme) {
            String rezKey = "Filter class {0} does not support the correct constructor. See class org.enhydra.servlet.filter.MultiServerFilter.";
            String msg = rez.format(rezKey,classname)            ;
            fatalErr(msg, nsme);
        } catch (SecurityException se) {
            String rezKey = "Filter class {0} caused a security exception.";
            String msg = rez.format(rezKey,classname)            ;
            fatalErr(msg,se);
        }
        Object o = null;
        try {
            Object[] args = new Object[2];
            args[0] = description;
            args[1] = init;
            o = cnst.newInstance(args);
        } catch (InstantiationException ie) {
            String rezKey = "Filter {0} could not be instantiated.";
            String msg = rez.format(rezKey,id);
            fatalErr(msg, ie);
        } catch (IllegalAccessException iae) {
            String rezKey = "Filter {0} could not be accessed. (is it a public class?).";
            String msg = rez.format(rezKey,id);
            fatalErr(msg, iae);
        } catch (IllegalArgumentException iae2) {
            fatalErr("Filter " + id + " could not be instantiated", iae2);
        } catch (InvocationTargetException ite) {
            String rezKey = "Filter {0} threw an expection";
            String msg = rez.format(rezKey,id);
            fatalErr(msg,ite.getTargetException());
        }
        if (o == null) {
            String rezKey = "Filter {0} could not be instantiated.";
            String msg = rez.format(rezKey,id);
            fatalErr(msg);
        }
        Filter tf = null;
        try {
            tf = (Filter) o;
        } catch (Throwable t) {
            String rezKey = "Filter class {0} is not castable to Filter.";
            fatalErr("Filter class " + classname +
                     " is not castable to Filter.", t);
        }
        logChannel.write(Logger.DEBUG, "ADDING FILTER: " + id);
        filterManager.add(id, tf, description);
    }

    private String myGetString(Config cfg, String key) {
        return myGetString(cfg, key, null, true);
    }

    private String myGetString(Config cfg, String key, String prefix) {
        return myGetString(cfg, key, prefix, true);
    }

    private String[] myGetStrings(Config t,
                   String key, String prefix) {
        return myGetStrings(t, key, prefix, true);
    }

    private String myGetString(Config cfg, String key,
                                      String prefix, boolean required) {
        String s = null;
        try {
            s = cfg.getString(key);
        } catch (ConfigException e) {
        }
        if ((s == null) && required) {
            if (prefix != null)
                prefix += ".";
            String msg = rez.format("Error reading config file. Key {0} not found",(prefix+key));
            fatalErr(msg);
        }
        return s;
    }

    private String[] myGetStrings(Config t,
                             String key, String prefix, boolean required) {
        String[] s = null;
        try {
            s = t.getStrings(key);
        } catch (ConfigException e) {
            // Handled below..
        }
        if ((s == null) && required) {
            if (prefix != null)
                prefix += ".";
            String msg = rez.format("Error reading config file. Key {0} not found",(prefix+key));
            fatalErr(msg);
        }
        return s;
    }

    private Config myGetSection(Config cfg, String key) {
        return myGetSection(cfg, key, true);
    }

    private Config myGetSection(Config cfg,
                                       String key, boolean required) {
        Config cfg2 = null;
        try {
            cfg2 = (Config) cfg.getSection(key);
        } catch (ConfigException e) {
            // Do nothing.
        } catch (Exception e) {
            String rezKey = "Internal error parsing config file (section {0}).";
            String msg = rez.format(rezKey,key);
            fatalErr(msg, e);
        }
        if ((cfg2 == null) && required) {
            String rezKey = "Illegal config file. Section {0} not found.";
            String msg = rez.format(rezKey,key);
            fatalErr(msg);
        }
        return cfg2;
    }

    private String trim(String s) {
        if (s == null)
            return null;
        int len = s.length();
        while(len > 0) {
            char c = s.charAt(0);
            if ((c == ' ') || (c == '\t')) {
                s = s.substring(1);
                len--;
            } else
                break;
        }
        while(len > 0) {
            char c = s.charAt(len - 1);
            if ((c == ' ') || (c == '\t')) {
                s = s.substring(0, len - 1);
                len--;
            } else
                break;
        }
        return s;
    }

    /**
     *  This function examines the current state of the server and
     *  writes it out to the given config file.
     *
     * @param filename The filename of the config file to write out.
     * If null, the same file that was read at startup time will be
     * rewritten.
     */
    public String writeConfigFile(String filename) {
        if (filename == null)
            filename = configFileName;
	File realFile = null;
        String tempName = filename + ".tmp";
        File tempFile = null;
        FileOutputStream tempOut = null;
        try {
            realFile = new File(filename);
	    tempFile = new File(tempName);
            tempOut = new FileOutputStream(tempFile);
        } catch (IOException e) {
            String err = rez.format("IO error writing to {0}.",tempName);
            logChannel.write(Logger.ERROR, err, e);
            return err;
        }
        try {
            try {
                createApplicationSection();
                createServletSection();
		config.remove(CONNECTION);
                connectionMethodManager.createConnectionSection(config);
                createChannelSection();
            } catch (ConfigException e) {
                String err = rez.format("Internal error recreating config file.");
                logChannel.write(Logger.ERROR, err, e);
                return err;
            } catch (KeywordValueException e) {
                String err = rez.format("Internal error recreating config file.");
                logChannel.write(Logger.ERROR, err, e);
                return err;
            }
            try {
                configFile.write(tempOut);
            } catch (Error e) {
                String err = rez.format("Unable to write config file to: {0}.",tempOut);
                logChannel.write(Logger.ERROR, err, e);
                return err;
            }
        } finally {
            /*
             * Be sure we close the output stream no matter what.
             */
            try {
                tempOut.close();
            } catch (IOException e) {
                // Nothing we can do about it.
            }
        }
        /*
         * When running under windows, we need to close the output streams
         * and delete the target file or the rename will fail.
         */
        if (!realFile.delete()) {
            String err = rez.format("Unable to delete {0}.",filename);
            logChannel.write(Logger.ERROR, err);
            return err;
        }
        if (!tempFile.renameTo(realFile)) {
            String rezKey = "Unable to rename {0} to {1}.";
            String err = rez.format(rezKey,tempName,filename);
            logChannel.write(Logger.ERROR, err);
            return err;
        }
        return null;
    }

    private void createApplicationSection()
                            throws ConfigException, KeywordValueException {
        // Delete the old application section.
        config.remove(APPLICATION);
        // Recreate it from the current state of server.
        String[] ids = servletManager.getServletIDs();
        for (int i=0; i<ids.length; i++) {
            ServletStatus status = servletManager.getStatus(ids[i]);
            if (status == null)
                continue;
            // sjip if it's not an Enhydra App or a War.
            if (!status.isWar && !status.className.equals(ENYHDRA_APP_CLASSNAME))
                continue;
            String base = APPLICATION + "." + ids[i] + ".";
            if (!status.isWar) {
              // It's an Enhydra app
              String confName = status.initArgs.getProperty(CONF_FILE);
              if (confName == null) {
                  String msg = rez.format(
                        "Unable to determine config file for application {0}",
                        ids[i]);
                  throw new ConfigException(msg);
              }
              StringTokenizer stok = new StringTokenizer(confName,
                                                         File.separator);
              while (stok.hasMoreTokens())
                  confName = stok.nextToken();
              config.set(base + CONF_FILE, confName);
            }
            else {
              // It's a Web app
              config.set(base + DOC_ROOT, status.docRoot);
              config.set(base + DEFAULT_SESSION_TIMEOUT, status.defaultSessionTimeOut);
              config.set(base + IS_WAR_EXPANDED, status.isWARExpanded);
              config.set(base + IS_WAR_VALIDATED, status.isWARValidated);
              config.set(base + IS_INVOKER_ENABLED, status.isInvokerEnabled);
              config.set(base + IS_WORKDIR_PERSISTENT, status.isWorkDirPersistent);
            }

            if (status.running)
                config.set(base + RUNNING, YES);
            else
                config.set(base + RUNNING, NO);
            config.set(base + DESCRIPTION, status.description);
        }
    }

    private void createServletSection()
                       throws ConfigException, KeywordValueException {
        // Delete the old servlet section.
        config.remove(SERVLET);
        // Recreate it from the current state of server.
        String[] ids = servletManager.getServletIDs();
        for (int i=0; i<ids.length; i++) {
            ServletStatus status = servletManager.getStatus(ids[i]);
            if (status == null)
                continue;
            if (status.isWar || status.className.equals(ENYHDRA_APP_CLASSNAME))
                continue;
            String base = SERVLET + "." + ids[i] + ".";
            config.set(base + CLASS_NAME, status.className);
            if (status.running)
                config.set(base + RUNNING, YES);
            else
                config.set(base + RUNNING, NO);
            config.set(base + DESCRIPTION, status.description);

            if (status.classPath.length > 0) {
                config.set(base + CLASS_PATH, status.classPath);
	    }
            config.set(base + DOC_ROOT, status.docRoot);

            Enumeration e = status.initArgs.propertyNames();
            while (e.hasMoreElements()) {
                String key = (String) e.nextElement();
                String value = status.initArgs.getProperty(key, "");
                config.set(base + INIT_ARGS + "." + key, value);
            }
        }
    }

    private void createChannelSection()
                         throws ConfigException, KeywordValueException {
        // Delete old channel section.
        config.remove(CHANNEL);
        // Recreate it from the current channels.
        String[] cmIds = connectionMethodManager.getIDs();
        for (int i=0; i<cmIds.length; i++) {
            ConnectionMethod cm = connectionMethodManager.get(cmIds[i]);
            if (cm == null)
                continue;
            String[] chIds = cm.getChannelIDs();
            for (int j=0; j<chIds.length; j++) {
                String base = CHANNEL + "." + cmIds[i] + "." + chIds[j] + ".";
                ChannelStatus cs = cm.getChannelStatus(chIds[j]);
                config.set(base + SERVLET, cs.servletID);
                config.set(base + URL, cs.URLPrefix);
                if (cs.enabled)
                    config.set(base + ENABLED, YES);
                else
                    config.set(base + ENABLED, NO);
                if (cs.filterIDs.length > 0) {
                    String filters = "";
                    boolean first = true;
                    for (int k=0; k<cs.filterIDs.length; k++) {
                        Filter filter =
                            filterManager.get(cs.filterIDs[k]);
                        if (filter == null)
                            continue;
                        if (filter instanceof
                            org.enhydra.servlet.debug.ServletRecorderFilter)
                            // Do not save the debugging filters.
                            continue;
                        if (!first)
                            filters += ", ";
                        filters += cs.filterIDs[k];
                        first = false;
                    }
                    config.set(base + FILTERS, filters);
                }
            }
        }
    }

    /*
     * Log the message, then stop the ServletContainer.
     */
    private void fatalErr(String msg, Throwable t) {
        fatalErr(msg, true,t);
    }

    /*
     * Log the message, then stop the ServletContainer.
     */
    private void fatalErr(String msg) {
        fatalErr(msg, true,null);
    }

    /*
     * Either log the message or write it to stderr, then stop the
     * ServletContainer. We should always log the messages, unless that
     * is not possible.
     */
    private void fatalErr(String msg, boolean useLogger,Throwable t) {
        if (useLogger) {
            String exitMsg = rez.format("Exiting the server.");
            if (t == null) {
                logChannel.write(Logger.ERROR, msg);
                logChannel.write(Logger.ERROR, exitMsg);
            }
            else {
                logChannel.write(Logger.ERROR, msg,t);
                logChannel.write(Logger.ERROR, exitMsg,t);
            }
        }
        if (t != null) {
            t.printStackTrace(System.err);
        }
        System.err.println(msg);
        System.err.println(rez.format("Stopping the ServletContainer."));
	stop();
        if (t != null) {
            throw new ServletError(t);
        } else {
            throw new ServletError(msg);
        }
    }

/*
    public void readServerConfig(String[] args) {
        Config config = getConfig(args);

	if (config == null) {
            return;
	}

	checkClassDependencies();

	ServerConfig serverConfig = config.getServerConfig();
	//Registry registry =  createRegistry(serverConfig.getAdminPort());

	Enumeration contextManagers = serverConfig.getContextManagers();

	while (contextManagers.hasMoreElements()) {
	    ContextManagerConfig contextManagerConfig =
	        (ContextManagerConfig)contextManagers.nextElement();
	    ContextManager contextManager = new ContextManager();
	    Enumeration contexts = contextManagerConfig.getContextConfigs();

	    while (contexts.hasMoreElements()) {
	        ContextConfig contextConfig =
		    (ContextConfig)contexts.nextElement();
                Context context = contextManager.addContext(
		    contextConfig.getPath(),
		    contextConfig.getDocumentBase());
		String contextWorkDirPath =
		    getContextWorkDirPath(serverConfig,
			contextManagerConfig,
		        contextConfig.getPath());

		context.setSessionTimeOut(
		    contextConfig.getDefaultSessionTimeOut());
		context.setInvokerEnabled(contextConfig.isInvokerEnabled());
		context.setIsWARExpanded(contextConfig.isWARExpanded());
		context.setIsWARValidated(contextConfig.isWARValidated());
		context.setWorkDir(contextWorkDirPath,
		    contextConfig.isWorkDirPersistent());
		context.setIsWARValidated(contextConfig.isWARValidated());

		if (contextConfig.getPath().equals(
                    org.apache.tomcat.core.Constants.Context.Default.Path)) {
		    contextManager.setDefaultContext(context);
		    contextManager.setServerInfo(
		        contextManager.getDefaultContext().getEngineHeader());
		}
	    }
	}
    }

    private Config getConfig(String args[]) {
	Config config = new Config(args);

	System.out.println(sm.getString("startup.banner"));

	if (config.isArg(Constants.Arg.Help)) {

	    System.out.println(sm.getString("startup.help"));
	    System.out.println();

	    return null;
	}

	String configFile = Constants.Server.ConfigFile;

	if (config.isArg(Constants.Arg.Config)) {
	    configFile = config.getArg(Constants.Arg.Config);
	}

	config.loadConfig(configFile);

	return config;
    }

    private String getContextWorkDirPath(ServerConfig serverConfig,
	ContextManagerConfig contextManagerConfig, String path) {
        String s = "";
	String baseDir = serverConfig.getWorkDir();
	String hostName = contextManagerConfig.getHostName();
	String iNet = contextManagerConfig.getINet();
	int port = contextManagerConfig.getPort();

	if (hostName.trim().length() > 0) {
	    s += hostName;
	} else if (iNet.trim().length() > 0) {
	    s += iNet;
	}

	if (s.length() != 0) {
	    s += ":";
	}

	if (port > -1) {
	    s += Integer.toString(port);
	}

	return baseDir + File.separator + URLEncoder.encode(s + path);
    }

    public void checkClassDependencies() {
        for (int i = 0; i < Constants.REQUIRED_CLASSES.length; i++) {
	    String clazz = (String)Constants.REQUIRED_CLASSES[i];

	    try {
	        Class.forName(clazz);
	    } catch (ClassNotFoundException cnfe) {
	        String msg = sm.getString("startup.classes.cnfe", clazz);

		System.out.println("warning: " + msg);
	    }
	}
    }
*/



}

