/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: Channel.java,v 1.4.12.1 2000/10/19 17:59:09 jasona Exp $
 */

package org.enhydra.servlet.connectionMethods;

import java.net.*;
import java.io.File;
import java.util.Vector;
import java.util.Date;
//import org.enhydra.servlet.filter.TransactionFilter;

import org.apache.tomcat.util.*;

/**
 * Everything there is to store about a channel.
 * @see org.enhydra.servlet.connectionMethods.ChannelTable
 * @see org.enhydra.servlet.connectionMethods.ConnectionMethod
 * @see org.enhydra.servlet.connectionMethods.StandardConnectionMethod
 * @author Shawn McMurdo
 */
public class Channel
{
    private String   channelID;
    private String   urlPrefix;
    private String   servletID;
    private boolean  enabled;
    protected Vector filterIDs;
    private long     requestCount;
    private Date     createTime;
    private Date     enabledDisabledTime;

    public Channel() {
        filterIDs = new Vector();
        enabled = false;
        requestCount = 0L;
        createTime = new Date();
        enabledDisabledTime = createTime;
    }

    /**
     * Set the channel ID.
     * @param channelID The channel ID.
     */
    public void setChannelID(String channelID) {
        this.channelID = channelID;
    }

    /**
     * Get the channel ID.
     * @return The channel ID.
     */
    public String getChannelID() {
        return channelID;
    }

    /**
     * Set the URL prefix.
     * @param urlPrefix The URL prefix.
     */
    public void setURLPrefix(String urlPrefix) {
        this.urlPrefix = urlPrefix;
    }

    /**
     * Get the URL prefix.
     * @return The URL prefix.
     */
    public String getURLPrefix() {
        return urlPrefix;
    }

    /**
     * Set the servlet ID.
     * @param servletID The servlet ID.
     */
    public void setServletID(String servletID) {
        this.servletID = servletID;
    }

    /**
     * Get the servlet ID.
     * @return The servlet ID.
     */
    public String getServletID() {
        return servletID;
    }

    /**
     * Set the channel enabled state.
     * @param enabled True to enable, false to disable.
     */
    public void setEnabled(boolean enabled) {
        this.enabled = enabled;
        enabledDisabledTime = new Date();
    }

    /**
     * Set the channel enabled state to true.
     */
    public void enable() {
        this.enabled = true;
        enabledDisabledTime = new Date();
    }

    /**
     * Set the channel enabled state to false.
     */
    public void disable() {
        this.enabled = false;
        enabledDisabledTime = new Date();
    }

    /**
     * Return the channel enabled state.
     * @return True if the channel is enabled.
     */
    public boolean getEnabled() {
        return enabled;
    }


    /**
     * Return the time when this channel was created.
     * @return The time when the constructor was called.
     */
    public Date getCreateTime() {
        return createTime;
    }


    /**
     * Return the time when <CODE>enable()</CODE>, <CODE>disable()</CODE>,
     * or <CODE>setEnabled()</CODE> was last called. Channels start off
     * disabled, and this value starts off
     * with the same value as <CODE>getCreateTime()</CODE>
     * @return The time when the the channel was last enabled or disabled.
     */
    public Date getEnabledDisabledTime() {
        return enabledDisabledTime;
    }



    /**
     * Add a new filter to the stack of filters. The first filter
     * added should be the first filter wrapped around the Servlet,
     * request and response objects.
     * The filter is refered to in the future by it's filterID.
     * The filters are fetched from the filter manager at request time.
     *
     * @param filterID The symbolic name to use to refer to this filter.
     * @exception ConnectionMethodException If the name is already in use.
     */
    public void addFilter(String filterID)
        throws ConnectionMethodException {
        synchronized (filterIDs) {
            int len = filterIDs.size();
            for (int i=0; i<len; i++) {
                if (filterID.equals(filterIDs.elementAt(i)))
                    throw new ConnectionMethodException("Unable to add " +
                        "filter. ID " + filterID + " already in use.");
            }
            filterIDs.addElement(filterID);
        }
    }


    /**
     * Remove a filter to the stack of filters. Use the filter ID passed
     * in to <CODE>addFilter()</CODE>.
     *
     * @param filterID The symbolic name of the filter to remove.
     * @exception ConnectionMethodException If the name is not found.
     */
    public void removeFilter(String filterID)
        throws ConnectionMethodException {
        synchronized (filterIDs) {
            if (filterIDs.removeElement(filterID))
                    return;
            else
                throw new ConnectionMethodException("Unable to remove " +
                    "filter. ID " + filterID + " not found.");
        }
    }




    /**
     * Returns an array of Strings, the filter names.
     * @return An array of the current TransactionFilter names.
     */
    public String[] getFilterIDs()
    {
        synchronized (filterIDs) {
            int len = filterIDs.size();
            String[] result = new String[len];
            for (int i=0; i<len; i++)
                result[i] = (String) filterIDs.elementAt(i);
            return result;
        }
    }




    /**
     * Set the request count.
     * @param count The request count.
     */
    public void setRequestCount(long count) {
        this.requestCount = count;
    }

    /**
     * Reset the request count.
     */
    public void resetRequestCount() {
        this.requestCount = 0L;
    }

    /**
     * Increment the request count.
     */
    public void incrRequestCount() {
        this.requestCount++;
    }

    /**
     * Return the request count.
     * @return The request count.
     */
    public long getRequestCount() {
        return requestCount;
    }



    /**
     * Compare two URL prefixes in terms of specificity. If a is more
     * specific than b, 1 is returned. If b is more specific than a,
     * -1 is returned. If a and b are equally specific, 0 is returned.
     * <P>
     * This comparison can be used to sort channels into the order that
     * should be used for matching request URLs. When a request comes in,
     * it's URL is extracted. Then the list of URL prefixes is searched.
     * The first one that matches the beginning of the request URL is
     * found. This channel then recieves the request. The list of URL
     * prefixes should be sorted via this comparison operator. This will
     * ensure, for example, that <CODE>/admin/</CODE> is examined before
     * <CODE>/</CODE> is. If they were compared in the other order, the
     * <CODE>/</CODE> would always match, and the channel associated with
     * <CODE>/admin/</CODE> would never get any requests.
     * <P>
     * The URL prefixes are sorted based on the number of directory
     * separator characters they have.  More separators means the prefix
     * is more specific. If two prefixes have the same number of
     * directory components, then the one with the longer final path
     * element is considered more specific. If the final path components
     * are the same length, the strings are considered equal.
     * <P>
     * Here is an example of a resulting search order using this method:
     *
     *<PRE>
     *  /app1/a/b/c/d/e/
     *  /app2/longer-name/but-less-path/elements/
     *  /demo/free/supersavers/
     *  /demo/free/supersavers
     *  /demo/free/super
     *  /index.html
     *  /index
     *  /
     *  index.html
     *  index
     *  "" (the empty string)
     *</PRE>
     * Note that the empty string always comes last, and it matches all
     * incoming URLs.
     */
    public static int compareURLPrefixes(String a, String b) {
        byte[] aa = a.getBytes();
        byte[] bb = b.getBytes();
        int aaElements = 0;
        int bbElements = 0;
        int lastAIndex = 0;
        int lastBIndex = 0;
        for(int i=0; i<aa.length; i++)
            if (aa[i] == File.pathSeparatorChar) {
                aaElements++;
                lastAIndex = i;
            }
        for(int i=0; i<bb.length; i++)
            if (bb[i] == File.pathSeparatorChar) {
                bbElements++;
                lastBIndex = i;
            }
        if (aaElements < bbElements)
            return -1;
        else if (aaElements > bbElements)
            return 1;
        int aLen = aa.length - lastAIndex;
        int bLen = bb.length - lastBIndex;
        if (aLen < bLen)
            return -1;
        else if (aLen > bLen)
            return 1;
        else
            return 0;
    }
}


