/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: ChannelTable.java,v 1.4.12.1 2000/10/19 17:59:09 jasona Exp $
 */

package org.enhydra.servlet.connectionMethods;

import java.util.Vector;

/**
 * A table of Channel elements.
 * @see org.enhydra.servlet.connectionMethods.Channel
 * @see org.enhydra.servlet.connectionMethods.ConnectionMethod
 * @see org.enhydra.servlet.connectionMethods.StandardConnectionMethod
 * @author Andy John
 * @author Shawn McMurdo
 */
public class ChannelTable
{
    /**
     * This is the actual table. If you need to synchronize across multiple
     * calls to this class, you may synchronize on this object.
     * Internally, all synchronization is done on this object.
     * <P><B> Maybe this should be replaced with a public Object, and it
     * should be used for all locking? <B>
     */
    protected Vector channelTable = null;

    /**
     * Constructor
     */
    public ChannelTable()
    {
	channelTable = new Vector();
    }

    /**
     * Add a new channel.  The new channel is not enabled and has
     * no TransactionFilters.
     * @param channelID The symbolic name to use for this channel.
     * @param URLPrefix If a request's URL starts with this string,
     *	it will be sent to the servlet identified by servletID.
     * @param servletID The servlet to send requests to. Use the
     *	ServletManager's <CODE>get()</CODE> method to get the servlet.
     *	Call <CODE>get()</CODE> for each request, do not save a reference
     *	to the servlet.
     * @see org.enhydra.servlet.servletManager.ServletManager
     */
    public void add(String channelID, String URLPrefix,
            String servletID)
    {
        Channel c = new Channel();
        c.setChannelID(channelID);
        c.setURLPrefix(URLPrefix);
        c.setServletID(servletID);
        c.disable();
        c.resetRequestCount();
        add(c);
    }

    /**
     * Add a new channel.  The new channel is not enabled and has
     * no TransactionFilters.
     * @param channel the channel to add.
     */
    public void add(Channel channel)
    {
        synchronized (channelTable) {
            /*
              Insert in order, based on how specific the URL prefix is.
              We want the most specific prefixes first.
            */
            int i = 0;
            while(i < channelTable.size()) {
                Channel ch = (Channel) channelTable.elementAt(i);
                if (Channel.compareURLPrefixes(channel.getURLPrefix(),
                                               ch.getURLPrefix()) == 1)
                    break;
                i++;
            }
            channelTable.insertElementAt(channel, i);
        }
    }

    /**
     * Delete a channel.
     * @param channelID Which channel to remove.
     */
    public void delete(String channelID)
    {
        synchronized (channelTable) {
            Channel c = findID(channelID);
            if (c != null) {
                channelTable.removeElement(c);
	    }
        }
    }

    /**
     * Returns the identifier names of all the channels currently in
     * the ChannelTable.
     * @return An array of Strings, the names of all the channels
     */
    public String[] getChannelIDs()
    {
        synchronized (channelTable) {
            int n = channelTable.size();
            String[] results = new String[n];
            for (int i = 0; i < n; i++) {
		Channel c = (Channel) channelTable.elementAt(i);
           	results[i] = c.getChannelID();
            }
            return results;
	}
    }

    /**
     * Return the Channel for the given identification
     * name. Returns null if not found.
     * @param channelID Which channel to search for.
     * @return The Channel (see above) for the given channel, or null
     *	if not found.
     */
    public Channel getChannelByID(String channelID)
    {
	Channel c = null;
        synchronized (channelTable) {
            c = findID(channelID);
        }
        return c;
    }

    /**
     * Return the Channel for the given identification
     * name. Returns null if not found.
     *
     * <I>Warning: this method does not synchronize on channelTable.
     * The caller must synchronize on it.
     * Normally you will want to synchronize on it,
     * call this method to look up a Channel,
     * modify the Channel or the
     * table itself, then release the lock on channelTable. That way you
     * know that the table will not change between the time you fetch an
     * entry to the time you modify the entry or the table. </I>
     *
     * @param channelID Which channel to search for.
     * @return The Channel (see above) for the given channel, or null
     *	if not found.
     */
    public Channel findID(String channelID)
    {
	for (int i = 0; i < channelTable.size(); i++) {
            Channel c = (Channel) channelTable.elementAt(i);
            if (c.getChannelID().equals(channelID)) {
		return c;
	    }
	}
	return null;
    }

    /**
     * Return the Channel for the given URL prefix.
     * Returns null if not found.
     * @param urlPrefix The URL prefix to search for.
     * @return The Channel for the given URL prefix, or null if not found.
     */
    public Channel getChannelByURLPrefix(String urlPrefix)
    {
	Channel c = null;
        synchronized (channelTable) {
            c = findURLPrefix(urlPrefix);
        }
        return c;
    }

    /**
     * Search the set of channels for one whose URLPrefix matches the
     * beginning of the given URL. Returns null if not found.
     * If you are extending this class, you should use this method to
     * determine which channel applies to a request URL.
     * The Channel returned is a reference to the object stored in
     * the table, so any changes made to it are also made to the working
     * instance in the table.<P>
     *
     * <I>Warning: this method does not synchronize on channelTable.
     * The caller must synchronize on it.
     * Normally you will want to synchronize on it,
     * call this method to look up a Channel,
     * modify the Channel or the
     * table its self, then release the lock on channelTable. That way you
     * know that the table will not change between the time you fetch an
     * entry to the time you modify the entry or the table. </I>
     *
     * @param URL The URL to try to find a match for.
     * @return The Channel (see above) that matches the given URL,
     * or null if not found.
     */
    public Channel findURLPrefix(String URL)
    {
	for (int i = 0; i<channelTable.size(); i++) {
            Channel c = (Channel) channelTable.elementAt(i);
            String preFix = c.getURLPrefix();
            if (URL.startsWith(preFix)) {
                return c;
	    }
        }
        return null;
    }

    /**
     *  Return a multi-line string listing the channel ids.
     *
     * @return The current channel ids.
     */
    public String toString() {
        String result = "Channel table contents:\n";
        synchronized (channelTable) {
            int n = channelTable.size();
            for (int i = 0; i < n; i++) {
		Channel c = (Channel) channelTable.elementAt(i);
           	result += c.getChannelID() + "\n";
            }
	}
        return result;
    }
}

