/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: ConnectionMethod.java,v 1.4.12.1 2000/10/19 17:59:09 jasona Exp $
 */

package org.enhydra.servlet.connectionMethods;

import com.lutris.util.*;
import java.io.Serializable;
import org.enhydra.servlet.servletManager.ServletManager;
import org.enhydra.servlet.filter.FilterManager;

import org.apache.tomcat.core.*;

/**
 *
 *  A ConnectionMethod is how requests get into the server from the
 *  outside world. This interface defines the API an object must
 *  implement in order to be administered by the server. <P>
 *
 *  Each ConnectionMethod must maintain an ordered list of "channels".
 *  Each channel has an ID (a String), used to identify and refer to
 *  the channel. See <CODE>addChannel()</CODE>, below. <P>
 *
 *  Each channel can be enabled or disabled. When disabled, connections
 *  must be refused, and not passed on to any servlets. When enabled,
 *  connections are allowed through to the destination servlet. <P>
 *
 *  Each channel also contains an ordered list of TransactionFilters,
 *  which can be used for debugging and logging. <P>
 *
 *  When a ConnectionMethod is being used as part of a server, the
 *  <CODE>initialize()</CODE> method will be called before any other methods.
 *  The servlet IDs in the channels are only valid in the ServletManager
 *  object provided to <CODE>initialize()</CODE>. Filter IDs are only
 *  valid in the FilterManager passed in to <CODE>initialize()</CODE>.<P>
 *
 *  The <CODE>destroy()</CODE> method will be called when the
 *  ConnectionMethod should shut down. Any external network exposure should
 *  be eliminated, for example any ServerSockets should be closed, or any
 *  WAI registrations should be unregistered.
 *  After <CODE>destroy()</CODE> is called, no further calls will be
 *  made to the object. <P>
 *
 *  HTTP, WAI and RMI are implementations of this interface. See
 *  StandardConnectionMethod for a sample implementation which mananges
 *  the channels, but does not actually recieve or generate requests.
 *  If you are implementing a ConnectionMethod, it is recommended that
 *  you extend StandardConnectionMethod.
 *
 *  Connection methods must contain a running non-daemon thread. The
 *  thread must be started either in the constructor or (perfered)
 *  the <CODE>initialize()</CODE> method, and stopped in the
 *  <CODE>destroy()</CODE> method.
 *
 *
 *  @see org.enhydra.servlet.connectionMethods.ChannelStatus
 *  @see org.enhydra.servlet.servletManager.ServletManager
 *  @see org.enhydra.servlet.filter.TransactionFilter
 *  @see org.enhydra.servlet.connectionMethods.StandardConnectionMethod
 *  @author Andy John
 */
public interface ConnectionMethod
    extends Serializable {
    /**
     *  Initialize the ConnectionMethod. The ServletManager to use is
     *  passed in. This ServletManager is used to look up Servlets
     *  based on their String servlet IDs. <P>
     *  <B>This will be called before any other method in this interface.</B>
     *
     *  @param connectionConfig The portion of the config file relevant to
     *  this connection method
     *  @param id The id associated with this connection method
     *  @param servletManager The ServletManager to use to obtain servlets.
     *  Servlet ID strings given in <CODE>addChannel()</CODE> are valid only
     *  in this ServletManager.
     *  @param filterManager The FilterManager to use to get filters.
     *  Filter ID strings given in <CODE>addTransactionFilter()</CODE>
     *  are valid only in this FilterManager.
     *  @exception ConnectionMethodException If an error occurs.
     */
    public void initialize(Config connectionConfig,
			   String id,
			   ServletManager servletManager,
			   FilterManager filterManager)
        throws ConnectionMethodException;

    /**
     *  Initialize the ConnectionMethod. The ServletManager to use is
     *  passed in. This ServletManager is used to look up Servlets
     *  based on their String servlet IDs. <P>
     *
     *  @param id The id associated with this connection method
     *  @param servletManager The ServletManager to use to obtain servlets.
     *  Servlet ID strings given in <CODE>addChannel()</CODE> are valid only
     *  in this ServletManager.
     *  @param filterManager The FilterManager to use to get filters.
     *  Filter ID strings given in <CODE>addTransactionFilter()</CODE>
     *  are valid only in this FilterManager.
     *  @exception ConnectionMethodException If an error occurs.
     */
    public void initialize(String id,
			   ServletManager servletManager,
			   FilterManager filterManager)
        throws ConnectionMethodException;

    /**
     * Write out self into config fie
     *
     * @param config The config file
     * @param base The base key string
     **/
    public void writeToConfig(Config connectionConfig,
			      String base)
	throws ConfigException, KeywordValueException;


    /**
     *  Add a new channel. The new channel is not enabled,
     *  and has no TransactionFilters.
     *  The Servlet ID is only valid in the
     *  ServletManager passed in to <CODE>initialize()</CODE>.
     *  When the Servlet ID
     *  is used to fetch the actual Servlet, the reference should only
     *  be kept in the ConnectionMethod for the duration of the request.
     *  After the request is finished, the ConnectionMehtod should eliminate
     *  all pointers to the Servlet. This will allow the ServletManager
     *  to destroy the Servlet and cause it to get garbage collected.
     *  The ServletManager's <CODE>get()</CODE> method should be called for
     *  every request.
     *
     *  @param channelID The symbolic name to use for this channel.
     *  @param URLPrefix If a request's URL starts with this string,
     *      it should be sent to the servlet identified by servletID.
     *      To obtain the servlet, use the ServletManager's <CODE>get()</CODE>
     *      method.
     *  @param servletID The servlet to send requests to. Use the
     *      ServletManager's <CODE>get()</CODE> method to get the servlet.
     *      Call <CODE>get()</CODE> for each request, do not save a reference
     *      to the servlet.
     *  @exception ConnectionMethodException If an error occurs.
     */
    void addChannel(String channelID, String URLPrefix, String servletID)
         throws ConnectionMethodException;


    /**
     *  Delete a channel.
     *
     *  @param channelID Which channel to remove.
     *  @exception ConnectionMethodException If an error occurs.
     */
    void deleteChannel(String channelID) throws ConnectionMethodException;


    /**
     *  Get the current status of a channel.
     *  @param channelID Which channel.
     *  @return The status of the channel, stored in a ChannelStatus object.
     *  Null is returned if the given channelID cannot be found.
     *  @see org.enhydra.servlet.connectionMethods.ChannelStatus
     */
    ChannelStatus getChannelStatus(String channelID);


    /**
     *  Returns the identifier names of all the channels currently in
     *  the ConnectionMethod.
     *
     *  @return An array of Strings, the names of all the channels
     *  in the ConnectionMethod.
     */
    String[] getChannelIDs();


    /**
     *  Turn on the channel. Allows requests to be sent to the servlet.
     *
     *  @param channelID Which channel to enable.
     *  @exception ConnectionMethodException If an error occurs.
     */
    void enableChannel(String channelID) throws ConnectionMethodException;


    /**
     *  Turn off the channel. Incomming requests should be refused.
     *  No requests should be allowd through to the servlet.
     *
     *  @param channelID Which channel to disable.
     *  @exception ConnectionMethodException If an error occurs.
     */
    void disableChannel(String channelID) throws ConnectionMethodException;


    /**
     *  Reset the specified channel's request count to zero.
     *  The counter measures the number of requests processed by this channel.
     *
     *  @param channelID Which channel.
     *  @exception ConnectionMethodException If an error occurs.
     */
    void resetRequestCount(String channelID) throws ConnectionMethodException;


    /**
     *  Add a new TransactionFilter to the specified channel.
     *  Use this ID to refer to the filter in
     *  <CODE>removeTransactionFilter()</CODE>. The filter will be
     *  fetched from the filter manager when a request comes in.
     *
     *  @param channelID Which channel to add a TransactionFilter to.
     *  @param filterID The TransactionFilter to add.
     *  The filter is fetched from the FilterManager passed in to
     *  <CODE>initialize()</CODE> each time a request comes in.
     *  @exception ConnectionMethodException If an error occurs.
     *  @see org.enhydra.servlet.filter.TransactionFilter
     */
    void addTransactionFilter(String channelID, String filterID)
        throws ConnectionMethodException;


    /**
     *  Remove the given TransactionFilter from the specified channel.
     *
     *  @param channelID Which channel to remove a TransactionFilter from.
     *  @param filter The TransactionFilter to remove.
     *  @exception ConnectionMethodException If an error occurs.
     *  @see org.enhydra.servlet.filter.TransactionFilter
     */
    void removeTransactionFilter(String channelID, String filterID)
        throws ConnectionMethodException;


    /**
     *  Get an array of all the current filter names for a given channel.
     *  If there are no filters, this will return a zero-length array.
     *
     *  @return An array of filter names. You can pass these to
     *  <CODE>removeTransactionFilter()</CODE>.
     *  @exception ConnectionMethodException If an error occurs, for example
     *  if the channel does not exist.
     */
    public String[] getTransactionFilterIDs(String channelID)
        throws ConnectionMethodException;


    /**
     *  This returns a channel name that is not currently in use. Use this
     *  when you want to programatically create a channel, and you don't
     *  care what it's name is. The name returned will not be in the list
     *  of current channels, and will not be returned on any other calls
     *  to this function.
     *  @return A name that can be used with <CODE>addChannel()</CODE>.
     */
    public String getUniqueChannelName();


    /**
     *  Returns the URL that a user would use to connect to this
     *  channel. For example, with an HTTP method on port 2000 on
     *  a machine called <CODE>enhydra.lutris.com</CODE>,
     *  and a url prefix of <CODE>/abalone/</CODE>, the
     *  returned string would be
     *  <CODE>"http://enhydra.lutris.com:2000/abalone/"</CODE>. <P>
     *
     *  Note: it is not always possible for a connection method to
     *  know the exact URL a user should use, so the returned
     *  string may be an approximation. For example, the RMI
     *  connection method may be accessed from any host that has
     *  access to it.
     *
     *  @return A URL string.
     *  @exception ConnectionMethodException If an error occurs,
     *  for example if the channel does not exist.
     */
    public String getChannelURL(String channelID)
        throws ConnectionMethodException;


    /**
     *  Is the string returned from <CODE>getChannelURL()</CODE> a valid
     *  url?
     *
     *  @return True if the url returned from
     *  <CODE>getChannelURL()</CODE> is a valid url that users could
     *  sucessfully enter into their browsers.
     *  @exception ConnectionMethodException If an error occurs,
     *  for example if the channel does not exist.
     */
    public boolean channelURLIsValid(String channelID)
        throws ConnectionMethodException;


    /**
     *  Shut down the connection method. Any threads created in
     *  <CODE>initialize()</CODE> or any other mehtod must be terminated.<P>
     *  <B>After this call, none of these methods will be called again.</B>
     *  @exception ConnectionMethodException If an error occurs.
     */
    public void destroy() throws ConnectionMethodException;

    /**
     * Compares the connection specific attributes for equivalency
     *
     *  @param compareObject The object to compare this to
     *  @exception ConnectionMethodException If an error occurs.
     **/
    public boolean equivalent(ConnectionMethod compareObject);

    public String getType();

    /*********************************************************************
     *  Error strings for common errors shared by all ConnectionMethods.
     *********************************************************************/

    /**
     *  Return this HTML when an incomming request does not match any
     *  of the channels. Suggested use:
     *  <CODE>response.sendError(HttpServletResponse.SC_NOT_FOUND,
     *  ConnectionMethod.noChannelHtml);</CODE>
     */
    public final static String noChannelHtml =
        "<TITLE>Not Found</TITLE><H1>Not Found</H1>\n" +
        "The requested object does not exist on this " +
        "Enhydra Multiserver.<P>\n" +
        "The link you followed is either outdated, inaccurate, " +
        "or the server has been instructed not to let you have it.";

    /**
     *  Return this HTML when an incomming request matches a channel,
     *  but that channel is disabled. Suggested use:
     *  <CODE>response.sendError(HttpServletResponse.SC_MOVED_TEMPORARILY,
     *  ConnectionMethod.disabledChannelHtml);</CODE>
     */
    public final static String disabledChannelHtml =
        "<TITLE>Temporarily Unavailable</TITLE>\n" +
        "<H1>Temporarily Unavailable</H1>\n" +
        "The requested object is temporarily unavailable. " +
        "Please try again later.\n";

    /**
     *  Return this HTML if an unexpected exception is thrown.
     *  Suggested use:
     *  <CODE>response.sendError(HttpServletResponse.SC_SERVER_ERROR,
     *  ConnectionMethod.errorHtml);</CODE>
     */
    public final static String errorHtml =
        "<TITLE>Server Error</TITLE><H1>Server Error</H1> " +
        "This Enhydra Multiserver has encountered an internal " +
        "error which prevents it from fulfilling your request. " +
        "The most likely cause is a misconfiguration. Please ask " +
        "the administrator to look for messages in the server's error log.";


}




