/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: EnhydraDirectorConnection.java,v 1.9.2.11.2.1 2000/10/19 17:59:09 jasona Exp $
 */

package org.enhydra.servlet.connectionMethods.EnhydraDirector;

import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.ServletInputStream;
import javax.servlet.ServletOutputStream;
import javax.servlet.ServletContext;
import javax.servlet.RequestDispatcher;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpSession;
import javax.servlet.http.HttpUtils;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.text.ParseException;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.NoSuchElementException;
import java.util.TimeZone;
import java.util.Date;
import java.util.Locale;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.IOException;
import java.security.Principal;
import org.enhydra.servlet.ServletError;
import org.enhydra.servlet.ServletIOException;
import org.enhydra.servlet.tomcat.HeaderConverter;
import com.lutris.http.HttpStatus;
import com.lutris.http.HttpCookieParser;
import com.lutris.logging.LogChannel;
import com.lutris.logging.Logger;
import com.lutris.Lutris;
import org.apache.tomcat.core.RequestAdapter;
import org.apache.tomcat.core.ResponseAdapter;
import org.apache.tomcat.util.MimeHeaders;
import org.apache.tomcat.util.MimeHeaderField;


public class EnhydraDirectorConnection extends EnhydraDirectorProtocol
        implements HttpServletRequest, HttpServletResponse, Enumeration,
        RequestAdapter, ResponseAdapter {
    //=========================================================
    // Class static variables.
    //=========================================================

    /**
     * Standard Mime type for POST-ed forms.
     */
    public static final String FORM_POST_MIME_TYPE =
	    "application/x-www-form-urlencoded";

    /**
     * Parser for dates.
     */
    private static final DateFormat dateParser = DateFormat.getDateInstance();

    /**
     * Parser for dates.
     */
    private static DateFormat dateFormatter = null;

    //=========================================================
    // Class per-connection variables.
    //=========================================================

    /**
     * Authentication key.
     */
    private byte[] authData = null;

    /**
     * Table of incoming request headers.
     */
    private Hashtable requestHeaders;

    /**
     * Table of incoming request headers.
     */
    private Hashtable cgiVars;

    /**
     * Table of incoming request headers.
     */
    private Hashtable privateHeaders;

    /**
     * Cookies received from the client.
     */
    private Vector requestCookies;

    /**
     * Query String parameters.
     */
    private Hashtable queryParameters;

    /**
     * Whether we have query parameters yet.
     */
    private boolean haveQueryParameters;

    /**
     * Whether <code>getInputStream() has been called on this instance.
     */
    private boolean usingInputStream;

    /**
     * Internal <code>ServletInputStream</code>
     */
    private EnhydraDirectorInputStream inStream;

    /**
     * If <code>getReader()</code> has been called, this keeps a reference
     * to the created <code>BufferedReader</code>.
     */
    private BufferedReader reader;

    /**
     * Whether <code>getOutputStream() has been called on this instance.
     */
    private boolean usingOutputStream;

    /**
     * Internal <code>ServletOutputStream</code>
     */
    private EnhydraDirectorOutputStream outStream;

    /**
     * Writer reference if the servlet requested a Writer.
     */
    private PrintWriter writer;

    /**
     * Servlet Context passed to us by the handler.
     */
    private ServletContext servletContext;

    /**
     * Servlet Path passed to us by the handler.
     */
    private String servletPath;

    /**
     * Our locale.
     */
    private Locale locale;

    /**
     * Cookie string to stick in request headers. XXX HACK
     * <P>
     * This is lame.  When/if we do persistent appserver connections,
     * this will need to be set to null.
     */
    private String cookieMangle = null;

    //=========================================================
    // Per-instance variables related to response calls.
    //=========================================================

    /**
     * Boolean to indicate whether data bytes have been sent over
     * the wire, and thus that headers and status can no longer
     * be set.
     */
    private boolean committed = false;

    /**
     * Current table of response headers.
     */
    private Hashtable responseHeaders = new Hashtable();;

    /**
     * Current table of outgoing cookies.  This is a <code>Vector</code>
     * of <code>Cookie</code> objects.
     */
    private Vector responseCookies = new Vector();

    /**
     * Status code if no error.
     */
    private int statusCode = 200;

    /**
     * Status code if no error.
     */
    private String statusMsg = "OK";

    /**
     * Debug flag.
     */
    private static final boolean debug = false;

    /**
     * The IP address and Port that the connection is listening on
     **/
    private String ipPortToken = null;

    /**
     * The session affinity flag
     **/
    private boolean sessionAffinity = true;

    /**
     * Request attributes
     */
    private Hashtable attributes = new Hashtable();


    //=========================================================
    // Methods defined by EnhydraDirectorConnection.
    //=========================================================

    public
    EnhydraDirectorConnection(LogChannel chan)
    throws IOException
    {
	super(chan);
	if (dateFormatter == null) setDateFormatter();
	inStream = new EnhydraDirectorInputStream(this);
	outStream = new EnhydraDirectorOutputStream(this);
    }

    /**
     * Sets information that can not be obtained from the request.
     * This should be called after next() and before servlet.service() is
     * called.
     *
     * @param reqServletPath The servlet path portion of the URL.
     * @param content The ServletContext object for the servlet.
     */
    public
    void setServletInfo(String reqServletPath, ServletContext context)
    {
        servletPath = reqServletPath;
        servletContext = context;
    }

    private Hashtable postParams = null;
    private static final Hashtable emptyHashTable = new Hashtable();

    /**
     * Parse query string or POST form parameters.
     * <P>
     * If query parameters exist, obtain them from the QUERY_STRING or
     * POST data and put them in the <code>queryParameters</code> hash
     * table.  This method does not handle multipart file upload.
     *
     * @exception IOException if an error occurs while reading POST data, or
     * 		invalid Content-Type or Content-Length headers are
     * 		encountered.
     */
    private synchronized
    void parseQueryParameters()
    throws IOException
    {
	if (!haveQueryParameters) {
	    Hashtable qp = null;
	    String requestMethod = getMethod().toUpperCase();
	    String contentType = getContentType();
	    int contentLength = getContentLength();
	    if (debug) {
		logChannel.write(Logger.DEBUG,
				"contentType = " + contentType);
		logChannel.write(Logger.DEBUG,
				"contentLength = " + contentLength);
		logChannel.write(Logger.DEBUG,
				"requestMethod = " + requestMethod);
	    }
	    String queryString = getQueryString();
	    if (requestMethod.equals("POST")) {
		if (contentType == null) {
		    throw new ServletIOException("Invalid HTTP post request:" +
			    " Content-Type not specified");
		}
		if (contentType.startsWith(FORM_POST_MIME_TYPE)) {
		    if (debug) {
			logChannel.write(Logger.DEBUG, "reading POST data");
		    }
		    if (contentLength < 0) {
			throw new ServletIOException("Invalid HTTP post " +
			    "request: Content-Length invalid.");
		    }
	            if (postParams == null) {
		        postParams = HttpUtils.parsePostData(contentLength, inStream);
	            }
		} else {
		    if (debug) {
			logChannel.write(Logger.DEBUG,
					"unknown POST data format");
		    }
		}
	    }
            if (postParams == null) {
                postParams = emptyHashTable;
            }
            // Add in the query even if the method was a POST.
	    if (queryString != null) {
		qp = HttpUtils.parseQueryString(queryString);
                // Do not call mergeParameters unless the there is a querystring.
                queryParameters = org.apache.tomcat.core.RequestUtil.mergeParameters(qp,
		    postParams);
	    }
            else {
	        queryParameters = postParams;
            }
	    haveQueryParameters = true;
	}
    }

    //=============================================================
    // EnhydraDirectorConnection methods for handling response calls.
    //=============================================================

    private static synchronized
    void setDateFormatter()
    {
	if (dateFormatter == null) {
	    dateFormatter =
		new SimpleDateFormat("EEE, dd MMM yyyy kk:mm:ss zzz");
	    dateFormatter.setTimeZone(TimeZone.getTimeZone("GMT"));
	}
    }

    /**
     * Called by <code>EnhydraDirectorOutputStream</code> to indicate that
     * actual data bytes have been written.
     * <P>
     * Once this method has
     * been called, no more changes can be made to the response
     * status or headers.
     */
    synchronized
    void commit()
    throws IOException
    {
	if (committed) return;
	String name, value, s;
	InternalHeader hdr;

	// Throw away "Status" in headers.  Use statusCode, statusMsg.
	responseHeaders.remove("status");
	if ((statusMsg == null) || (statusMsg.length() < 1))
	    value = "" + statusCode + " " + HttpStatus.getStatusMsg(statusCode);
	else
	    value = "" + statusCode + " " + statusMsg;
        sendHeader("Status", value);

	// Send Content-Type header.
	hdr = (InternalHeader) responseHeaders.get("content-type");
	if ((hdr == null) || (hdr.size() < 1)) {
	    value = "text/html";
	} else {
	    value = (String) hdr.lastElement();
	}
	responseHeaders.remove("content-type");
	sendHeader("Content-Type", value);

	// Send Content-Length header.
	boolean haveContentLength = false;
	hdr = (InternalHeader) responseHeaders.get("content-length");
	if ((hdr == null) || (hdr.size() < 1)) {
	    value = null;
	} else {
	    value = (String) hdr.lastElement();
	}
	responseHeaders.remove("content-length");
	if (value != null) {
	    sendHeader("Content-Length", value);
	    haveContentLength = true;
	}

	// Send Connection: Keep-Alive or Connection: close if necessary.
	hdr = (InternalHeader) requestHeaders.get("connection");
	if (hdr != null) {
	    if (haveContentLength) {
		value = (String) hdr.lastElement();
		if ((value != null) &&
		     value.trim().equalsIgnoreCase("keep-alive")) {
		    sendHeader("Connection", "Keep-Alive");
		} else {
		    sendHeader("Connection", "close");
		}
	    } else {
		sendHeader("Connection", "close");
	    }
	}

	// Don't Send the "Server" header.  Client HTTP daemons do that
	// for us.
	//sendHeader("Server",
	//		"Enhydra-Multiserver/" + Lutris.getEnhydraVersion());
	responseHeaders.remove("server");

	// Don't Send the "Date" header.  Client HTTP daemons do that
	// for us.
	//sendHeader("Date", dateFormatter.format(new Date()));
	responseHeaders.remove("date");

	// Send rest of headers.
	Enumeration enum = responseHeaders.elements();
	if (enum != null) {
	    while (enum.hasMoreElements()) {
		hdr = (InternalHeader) enum.nextElement();
		Enumeration valEnum = hdr.elements();
		if (valEnum != null) {
		    while (valEnum.hasMoreElements()) {
			value = (String) valEnum.nextElement();
			sendHeader(hdr.name, value);
		    }
		}
	    }
	}

	// Send the cookies.  The client-side module will mangle
	// them together if necessary.
	Cookie c;
	String cval, cname;
	int cnl;
	enum = responseCookies.elements();
	if (debug) {
	    logChannel.write(Logger.DEBUG, "About to do cookies...");
	}
	while (enum.hasMoreElements()) {
	    c = (Cookie) enum.nextElement();
	    cval = HttpCookieParser.formatResponseCookie(c);
	    cname = c.getName();
	    if (cname == null) cname = "";
	    cnl = cname.length();
	    cval = cval.substring(cnl + 1);
	    if (debug) {
		logChannel.write(Logger.DEBUG,
				"Cookie name='"+cname+"' value='"+cval);
	    }
	    sendCookie(cname, cval);
	}

	// Tell the connection object to finish sending headers.
	finishHeaders();

	committed = true;
    }

    /**
     * These values are used by the session manager to determine what
     * IP and Port number to include with the user's session cookie.
     * This value default to null and is set by the EnhydraDirectorHandler.java
     **/
    public void setIPPortToken(String ipPortToken) {
	this.ipPortToken = ipPortToken;
    }

    public String getIPPortToken() {
	return ipPortToken;
    }

    /**
     * This flag is used by the session manager to determine if the IP and
     * Port string should be included in the session cookie.  This value
     * defaults to true and is set by EnhydraDirectorHandler.java
     **/
    public void setSessionAffinity(boolean sessionAffinity) {
	this.sessionAffinity = sessionAffinity;
	if (sessionAffinity && (ipPortToken != null)) {
	    stuffInternalHeader(":aff:", ipPortToken);
	}
    }

    public boolean getSessionAffinity() {
	return sessionAffinity;
    }

    /**
     * Override the activate() method to allow per-connection variables
     * to be reset for each new connection.
     *
     * @exception IOException If an error occurs.
     */
    public
    void activate()
    throws IOException
    {
	committed = false;
	statusCode = 200;
	statusMsg = "OK";
	requestHeaders = new Hashtable();
	cgiVars = new Hashtable();
	responseCookies = new Vector();
	privateHeaders = new Hashtable();
	responseHeaders = new Hashtable();
      	requestCookies = new Vector();
	queryParameters = null;
	haveQueryParameters = false;
	usingInputStream = false;
	inStream.reset();
	reader = null;
	usingOutputStream = false;
	outStream.reset();
	writer = null;
	servletContext = null;
	servletPath = null;
	locale = Locale.getDefault();
	cookieMangle = null;
	super.activate();

	/*
	 * Stuff some pseudo-headers to notify the app manager
	 * about this connection's session affinity status.
	 * XXX This may be a hack but it's portable.
	 */

    }

    //=========================================================
    // Abstract method overrides for EnhydraDirectorProtocol
    //=========================================================

    /**
     * Internal dispatch table for server side function handlers.
     * Supported functions are permanently registered at compile time
     * by placing them in this method's switch block.
     *
     * @param type Protocol-defined function type number.
     * @return Reference to the appropriate function handler instance.
     */
    protected
    EnhydraDirectorFunction getFunction(int type)
    {
	/* XXX No functions supported yet */
	return null;
    }

    /**
     * Implementation class must provide this method.  It should check
     * the configuration data to determine if authentication is required
     * for connections.
     *
     * @return <code>true</code> if authentication is required.
     */
    protected
    boolean authRequired()
    {
	if (this.authData == null)
	    return false;
	else
	    return true;
    }

    /**
     * Abstract method to get authentication secret data.
     * <P>
     * This method must be provided by the implementation class.
     * It checks the configuration and finds the secret authentication
     * data used to verify authentication responses from clients.
     *
     * @return A byte array of authentication data, or <code>null</code>
     * 		if authentication data does not exist.
     */
    protected
    byte[] getAuthData()
    {
	return this.authData;
    }

    /**
     * Sets authentication data.
     * <P>
     * This method allows the connection method handler to pass the
     * configured authentication string to this instance for use
     * in establishing connections.
     *
     * @param A string containing the authentication key.  If empty or
     * 		<code>null</code>, then authentication is not required.
     */
    public
    void setAuthData(String authData)
    {
	if ((authData == null) || (authData.length() < 1))
	    this.authData = null;
	else
	    this.authData = authData.getBytes();
    }

    /**
     * Abstract method to get authentication secret data.
     * <P>
     *
     * @return A byte array of authentication data, or <code>null</code>
     * 		if authentication data does not exist.
     */
    protected
    void processCgiHeader(byte[] nameBytes, byte[] valueBytes)
    {
	int i;
	char[] chars = new char[nameBytes.length];
	char[] lcchars = new char[nameBytes.length];
	for (i=0; i<chars.length; i++) {
	    chars[i] = (char)(nameBytes[i] & 0xff);
	    lcchars[i] = Character.toLowerCase(chars[i]);
	}
	String name = new String(chars);
	String lcname = new String(lcchars);
	chars = new char[valueBytes.length];
	for (i=0; i<chars.length; i++) chars[i] = (char)(valueBytes[i] & 0xff);
	String value = new String(chars);
	if (debug) {
	    logChannel.write(Logger.DEBUG,
			"GOT CGI HEADER " + name + "='" + value + "'");
	}

	InternalHeader hdr = (InternalHeader) cgiVars.get(lcname);
	if (hdr == null) {
	    hdr = new InternalHeader();
	    hdr.name = name;
	    cgiVars.put(lcname, hdr);
	}
	hdr.addElement(value);
    }

    /**
     * Private helper method to stuff a header into the request headers list.
     * <P>
     * If this is used to stuff 'pseudo' headers for internal code,
     * then the headers should have illegal names that can't conflict
     * with valid HTTP headers.
     *
     * @param name The header name.
     * @param name The header value.
     * @return A byte array of authentication data, or <code>null</code>
     * 		if authentication data does not exist.
     */
    protected
    void stuffInternalHeader(String name, String value)
    {
	String lcname = name.toLowerCase();
	if (debug) {
	    logChannel.write(Logger.DEBUG,
			"PRIVATE HEADER " + name + "='" + value + "'");
	}
	InternalHeader hdr = (InternalHeader) privateHeaders.get(lcname);
	if (hdr == null) {
	    hdr = new InternalHeader();
	    hdr.name = name;
	    privateHeaders.put(lcname, hdr);
	}
	hdr.addElement(value);
    }

    /**
     * Abstract method to get authentication secret data.
     * <P>
     *
     * @return A byte array of authentication data, or <code>null</code>
     * 		if authentication data does not exist.
     */
    protected
    void processHttpHeader(byte[] nameBytes, byte[] valueBytes)
    {
	int i;
	char[] chars = new char[nameBytes.length];
	char[] lcchars = new char[nameBytes.length];
	for (i=0; i<chars.length; i++) {
	    chars[i] = (char)(nameBytes[i] & 0xff);
	    lcchars[i] = Character.toLowerCase(chars[i]);
	}
	String name = new String(chars);
	String lcname = new String(lcchars);
	chars = new char[valueBytes.length];
	for (i=0; i<chars.length; i++) chars[i] = (char)(valueBytes[i] & 0xff);
	String value = new String(chars);
	if (debug) {
	    logChannel.write(Logger.DEBUG,
			"GOT HTTP HEADER " + name + "='" + value + "'");
	}
	InternalHeader hdr = (InternalHeader) requestHeaders.get(lcname);
	if (hdr == null) {
	    hdr = new InternalHeader();
	    hdr.name = name;
	    requestHeaders.put(lcname, hdr);
	}
	hdr.addElement(value);
    }

    /**
     * Abstract method to get authentication secret data.
     * <P>
     *
     * @return A byte array of authentication data, or <code>null</code>
     * 		if authentication data does not exist.
     */
    protected
    void processCookie(byte[] nameBytes, byte[] valueBytes)
    {
	int i;
	char[] chars = new char[nameBytes.length];
	char[] lcchars = new char[nameBytes.length];
	for (i=0; i<chars.length; i++) {
	    chars[i] = (char)(nameBytes[i] & 0xff);
	    lcchars[i] = Character.toLowerCase(chars[i]);
	}
	String name = new String(chars);
	String lcname = new String(lcchars);
	chars = new char[valueBytes.length];
	for (i=0; i<chars.length; i++) chars[i] = (char)(valueBytes[i] & 0xff);
	String value = new String(chars);
	if (debug) {
	    logChannel.write(Logger.DEBUG,
			"GOT COOKIE " + name + "='" + value + "'");
	}
	Cookie c = new Cookie(name, value);
	requestCookies.addElement(c);
	/* This is lame but necessary - XXX */
	if (cookieMangle != null) {
	    cookieMangle += ";" + name + "=" + value;
	} else {
	    cookieMangle = name + "=" + value;
	}
	InternalHeader hdr = (InternalHeader) requestHeaders.get("cookie");
	if (hdr == null) {
	    hdr = new InternalHeader();
	    hdr.name = "cookie";
	    requestHeaders.put("cookie", hdr);
	}
	hdr.removeAllElements();
	hdr.addElement(cookieMangle);
    }

    //=============================================================
    // Methods implementing java.util.Enumeration.
    //=============================================================

    /**
     * Stub <code>hasMoreElements</code> implementation so this instance
     * can be passed as an empty enumeration.
     * <P>
     * This is necessary because <code>Enumeration</code> cannot
     * be instantiated by itself, and some methods in this class
     * must return an empty enumeration if certain headers or
     * parameters are not found.
     *
     * @return Always returns <code>false</code>, indicating no more
     * 		elements are in this enumeration.
     */
    public
    boolean hasMoreElements()
    {
	return false;
    }

    /**
     * Stub <code>nextElement</code> implementation so this instance
     * can be passed as an empty enumeration.
     * <P>
     * This is necessary because <code>Enumeration</code> cannot
     * be instantiated by itself, and some methods in this class
     * must return an empty enumeration if certain headers or
     * parameters are not found.
     *
     * @return Normally an <code>Object</code>, but in this case it
     * 		only throws exceptions.
     * @exception NoSuchElementException Always.
     */
    public
    Object nextElement()
    throws NoSuchElementException
    {
	throw new NoSuchElementException("Empty enumeration");
    }

    //=============================================================
    // Methods implementing javax.servlet.http.HttpServletRequest
    //=============================================================

    /**
     * Gets the authentication scheme of this request.
     *
     * @return This request's authentication scheme, or null if none.
     */
    public
    String getAuthType()
    {
	String ret;
	InternalHeader hdr = (InternalHeader) cgiVars.get("auth_type");
	if (hdr == null) ret = null;
	else ret = (String) hdr.lastElement();
	if (debug) {
	    logChannel.write(Logger.DEBUG, "getContextPath() returns \"" +
			     ((ret != null) ? ret : "(null)") +  "\"");
	}
	return ret;
    }

    /**
     * Returns the portion of the request URI that indicates the context
     * of the request.
     * <P>
     * The context path always comes first in a request URI.  The path
     * starts with a "/" character but does not end with a "/" character.
     * For servlets in the default (root) context, this method returns "".
     *
     * @return A <code>String</code> specifying the portion of the request
     * 		URI that indicates the context of the request.
     */
    public
    String getContextPath()
    {
	String ret;
	// XXX Need the handler to set this.
	InternalHeader hdr =
		(InternalHeader) privateHeaders.get("context_path");
	if (hdr == null) ret = null;
	else ret = (String) hdr.lastElement();
	if (debug) {
	    logChannel.write(Logger.DEBUG, "getContextPath() returns \"" +
			     ((ret != null) ? ret : "(null)") +  "\"");
	}
	return ret;
    }

    /**
     * Gets the array of cookies found in this request.
     *
     * @return The array of cookies found in this request
     */
    public
    Cookie[] getCookies()
    {
	Cookie[] cookies = new Cookie[requestCookies.size()];
	requestCookies.copyInto(cookies);
	return cookies;
    }

    /**
     * Gets the value of the requested date header field of this request.
     *
     * @param name the String containing the name of the requested
     *		header field.   The comparison is case insensitive.
     * @return the value the requested date header field, or -1 if not
     *		found.
     * @exception IllegalArgumentException If value of header can't be
     *		converted to an int.
     */
    public
    long getDateHeader(String name)
    {
	InternalHeader hdr =
	    (InternalHeader) requestHeaders.get(name.toLowerCase());
	if (hdr == null) return -1;
        String value = (String) hdr.lastElement();
        if (value == null) {
            return -1;
        } else {
            try {
                return dateParser.parse(value).getTime();
            } catch(ParseException except) {
                throw new IllegalArgumentException(except.getMessage());
            }
        }
    }

    /**
     * Gets the value of the requested header field of this request.
     *
     * @param name The String containing the name of the requested
     * header field.  The comparison is case insensitive.
     * @return The value of the requested header field, or null if not
     * known.
     */
    public
    String getHeader(String name)
    {
	String ret;
	InternalHeader hdr;
	String lcname = name.toLowerCase();
	if (lcname.startsWith(":")) {
	    hdr = (InternalHeader) privateHeaders.get(lcname);
	} else {
	    hdr = (InternalHeader) requestHeaders.get(lcname);
	}
	if (hdr == null) ret = null;
	else ret = (String) hdr.lastElement();
	if (debug) {
	    logChannel.write(Logger.DEBUG, "getHeader(\"" + name +
			     "\") returns \"" +
			     ((ret != null) ? ret : "(null)") +  "\"");
	}
	return ret;
    }

    /**
     * Gets the header names for this request.
     *
     * @return An enumeration of strings representing the header names
     * for this request.
     */
    public
    Enumeration getHeaderNames()
    {
	return new InternalHeaderKeyEnumeration(requestHeaders);
    }

    /**
     * Returns all the values of the specified requst header as an
     * <code>Enumeration</code> of <code>String</code> objects.
     * <P>
     * Some headers, such as <code>Accept-Language</code> can be sent
     * by clients as several headers each with a different value rather
     * than sending the header as a comma separated list.
     * <P>
     * If the request did not include any headers of the specified name,
     * this method returns an empty <code>Enumeration</code>.  The
     * header name is case insensitive.  You can use this method with
     * any request header.
     *
     * @param name A <code>String</code> specifying the header name.
     * @return An <code>Enumeration</code> containing the values of the
     * 		requested header, or an empty <code>Enumeration</code>
     * 		if the request does not have any headers of that name.
     */
    public Enumeration getHeaders(String name)
    {
	String lcname = name.toLowerCase();
	InternalHeader hdr;
	if (lcname.startsWith(":")) {
	    hdr = (InternalHeader) privateHeaders.get(name.toLowerCase());
	} else {
	    hdr = (InternalHeader) requestHeaders.get(name.toLowerCase());
	}
	if ((hdr == null) || (hdr.size() < 1)) return (Enumeration) this;
	return hdr.elements();
    }

    /**
     * Gets the value of the specified integer header field of this
     * request.
     *
     * @param name The String containing the name of the requested
     * header field.   The comparison is case insensitive.
     * @return the value of the requested header field, or -1 if not
     * found.
     * @exception NumberFormatException If value of header can't be
     *  converted to an int.
     */
    public
    int getIntHeader(String name)
    throws NumberFormatException
    {
	InternalHeader hdr =
	    (InternalHeader) requestHeaders.get(name.toLowerCase());
	if ((hdr == null) || (hdr.size() < 1)) return -1;
	String value = (String) hdr.lastElement();
	return Integer.parseInt(value);
    }

    /**
     * Gets the HTTP method (for example, GET, POST, PUT) with which
     * this request was made.
     *
     * @return the HTTP method with which this request was made
     */
    public
    String getMethod()
    {
	InternalHeader hdr = (InternalHeader) cgiVars.get("request_method");
	if ((hdr == null) || (hdr.size() < 1)) return "GET";
	return (String) hdr.lastElement();
    }

    /**
     * Gets any optional extra path information following the servlet
     * path of this request's URI, but immediately preceding its query
     * string.
     *
     * @return The optional path information following the servlet
     * path, but before the query string, in this request's URI; null
     * if this request's URI contains no extra path information.
     */
    public
    String getPathInfo()
    {
	String ret;
	InternalHeader hdr = (InternalHeader) cgiVars.get("path_info");
	if ((hdr == null) || (hdr.size() < 1)) ret = "";
	else ret =  (String) hdr.lastElement();
	if (debug) {
	    logChannel.write(Logger.DEBUG, "getPathInfo() returns \"" +
			     ((ret != null) ? ret : "(null)") +  "\"");
	}
	return ret;
    }

    /**
     * Gets any optional extra path information following the servlet
     * path of this request's URI, but immediately preceding its query
     * string, and translates it to a real path.
     *
     * @return Extra path information translated to a real path or null
     * if no extra path information is in the request's URI.
     */
    public
    String getPathTranslated()
    {
	String ret;
	InternalHeader hdr = (InternalHeader) cgiVars.get("path_translated");
	if ((hdr == null) || (hdr.size() < 1)) ret = null;
	else ret = (String) hdr.lastElement();
	if (debug) {
	    logChannel.write(Logger.DEBUG, "getPathTranslated() returns \"" +
			     ((ret != null) ? ret : "(null)") +  "\"");
	}
	return ret;
    }



    private String queryString = null;
    /**
     * Gets any query string that is part of the HTTP request URI.
     *
     * @return query string that is part of this request's URI, or null
     * if it contains no query string
     */
    public
    String getQueryString()
    {
        if (queryString == null) {
	    InternalHeader hdr = (InternalHeader) cgiVars.get("query_string");
	    if ((hdr == null) || (hdr.size() < 1)) queryString = null;
	    else queryString =  (String) hdr.lastElement();
	    if (debug) {
	        logChannel.write(Logger.DEBUG, "getQueryString() returns \"" +
		    ((queryString != null) ? queryString : "(null)") +  "\"");
	    }
        }
	return queryString;
    }

    /**
     * Sets the query string that is part of the HTTP request URI.
     * The querystring can be set by a RequestDispatcher forward.
     */
    public void setQueryString(String queryString) {
        this.queryString = queryString;
        // force the parameters to be recalculated
        haveQueryParameters = false;
    }


    /**
     * Gets the name of the user making this request.  The user name is
     * set with HTTP authentication.
     *
     * @return The name of the user making this request, or null if not
     * known.
     */
    public
    String getRemoteUser()
    {
	String ret;
	InternalHeader hdr = (InternalHeader) cgiVars.get("remote_user");
	if ((hdr == null) || (hdr.size() < 1)) ret = null;
	else ret = (String) hdr.lastElement();
	if (debug) {
	    logChannel.write(Logger.DEBUG, "getRemoteUser() returns \"" +
			     ((ret != null) ? ret : "(null)") +  "\"");
	}
	return ret;
    }

    /**
     * Gets the session id specified with this request.
     *
     * @return The session id specified by this request, or null if the
     * request did not specify a session id.
     *
     * @see #isRequestedSessionIdValid
     */
    public
    String getRequestedSessionId ()
    {
        throw new ServletError("Servlet sessions are not yet implemented");
    }

    /**
     * Get the part of this
     * request's URI that is to the left of any query string.
     *
     * @return This request's URI
     * @see javax.servlet.http.HttpUtils#getRequestURL
     */
    public
    String getRequestURI()
    {
	String ret;
	InternalHeader hdr = (InternalHeader) cgiVars.get("request_uri");
	if ((hdr == null) || (hdr.size() < 1)) ret =  "";
	else ret = (String) hdr.lastElement();
	if (debug) {
	    logChannel.write(Logger.DEBUG, "getRequestURI() returns \"" +
			     ((ret != null) ? ret : "(null)") +  "\"");
	}
	return ret;
    }

    /**
     * Gets the part of this request's URI that refers to the servlet
     * being invoked.
     *
     * @return The servlet being invoked, as contained in this
     * request's URI.
     */
    public
    String getServletPath()
    {
	String ret;
	InternalHeader hdr = (InternalHeader) cgiVars.get("script_name");
	if ((hdr == null) || (hdr.size() < 1)) {
	    ret =  "/";
	} else {
	    String sn = (String) hdr.lastElement();
	    if (!sn.endsWith("/")) ret = sn + "/";
	    else ret = sn;
	}
	if (debug) {
	    logChannel.write(Logger.DEBUG, "getServletPath() returns \"" +
			     ((ret != null) ? ret : "(null)") +  "\"");
	}
	return ret;
    }

    /**
     * Gets the current valid session associated with this request, if
     * create is false or, if necessary, creates a new session for the
     * request, if create is true.
     *
     * @return The session associated with this request or null if
     * create was false and no valid session is associated
     * with this request.
     */
    public
    HttpSession getSession()
    {
	return getSession(true);
    }

    /**
     * Gets the current valid session associated with this request, if
     * create is false or, if necessary, creates a new session for the
     * request, if create is true.
     *
     * @return The session associated with this request or null if
     * create was false and no valid session is associated
     * with this request.
     */
    public
    HttpSession getSession (boolean create)
    {
        throw new ServletError("Servlet sessions are not yet implemented");
    }

    /**
     * NEW
     */
    public Principal getUserPrincipal()
    {
        throw new ServletError("Servlet 2.1 Security not yet implemented");
    }

    /**
     * Checks whether the session id specified by this request came in
     * as a cookie.
     *
     * @return True if the session id specified by this request came in
     * as a cookie; false otherwise.
     *
     * @see #getSession
     */
    public
    boolean isRequestedSessionIdFromCookie ()
    {
        throw new ServletError("Servlet sessions are not yet implemented");
    }

    /**
     * Checks whether the session id specified by this request came in
     * as part of the URL.
     *
     * @return True if the session id specified by the request for this
     * session came in as part of the URL; false otherwise.
     *
     * @see #getSession
     */
    public
    boolean isRequestedSessionIdFromUrl ()
    {
        throw new ServletError("Servlet sessions are not yet implemented");
    }

    /**
     * Checks whether the session id specified by this request came in
     * as part of the URL.
     *
     * @return True if the session id specified by the request for this
     * session came in as part of the URL; false otherwise.
     *
     * @see #getSession
     */
    public
    boolean isRequestedSessionIdFromURL()
    {
        throw new ServletError("Servlet sessions are not yet implemented");
    }

    /**
     * Checks whether this request is associated with a session that
     * is valid in the current session context.
     *
     * @return True if this request is assocated with a session that is
     * valid in the current session context.
     *
     * @see #getRequestedSessionId
     * @see javax.servlet.http.HttpSessionContext
     * @see #getSession
     */
    public
    boolean isRequestedSessionIdValid()
    {
        throw new ServletError("Servlet sessions are not yet implemented");
    }

    /**
     * NEW
     */
    public
    boolean isUserInRole(String role)
    {
        throw new ServletError("Servlet 2.1 Security not yet implemented");
    }

    //=============================================================
    // Methods implementing javax.servlet.ServletRequest
    //=============================================================

    /**
     * Returns the value of an attribute of the request.
     *
     * @param name The name of the attribute whose value is required.
     * @return The value of the named attribute of the request, or
     * null if the attribute does not exist.
     */
    public
    Object getAttribute(String name)
    {
        return attributes.get(name);
    }

    /**
     * NEW
     */
    public
    Enumeration getAttributeNames()
    {
        return attributes.keys();
    }

    /**
     * Get the character set encoding for the input of this request.
     * @return The character set encoding.
     */
/* Already part of Response implementation.
    public
    String getCharacterEncoding()
    {
        return "ISO8859_1"; //FIXME
    }
*/

    /**
     * Get the size of the request data.
     * @returns The size of the request data, or -1 if not known.
     */
    public
    int getContentLength()
    {
	int ret;
	try {
	    ret = getIntHeader("content-length");
	} catch (NumberFormatException e) {
	    ret = -1;
	}
	if (debug) {
	    logChannel.write(Logger.DEBUG, "getContentLength() returns " +ret);
	}
	return ret;
    }

    /**
     * Get the content type of the request.
     *
     * @return The Internet Media Type of the request entity data, or
     * null if not known.
     */
    public
    String getContentType()
    {
	String ret;
	String cType = getHeader("content-type");
	if (cType == null) {
	    ret = null;
	} else {
	    ret = cType.trim().toLowerCase();
	}
	if (debug) {
	    logChannel.write(Logger.DEBUG, "getContentType() returns \"" +
			     ((ret != null) ? ret : "(null)") +  "\"");
	}
	return ret;
    }

    /**
     * Get an input stream for reading binary data in the request body.
     *
     * @see getReader
     * @exception IllegalStateException if getReader has been
     *	called on this same request.
     * @exception IOException on other I/O related errors.
     * @returns An input stream.
     */
    public
    ServletInputStream getInputStream()
    throws IOException
    {
        if (reader != null) {
            throw new IllegalStateException(
			"getReader() has been called on this " +
			"request; can't use getInputStream()");
        }
	usingInputStream = true;
        return inStream;
    }

    /**
     * NEW
     */
    public Enumeration getLocales()
    {
	// XXX FIXME.
	return (Enumeration) this;
    }

    /**
     * Get a string containing the lone value of the specified parameter.
     * <P>
     * This call is case sensitive.
     *
     * @see #getParameterValues
     *
     * @param name The name of the parameter whose value is required.
     * @return The parameter value or null if the parameter does not exist.
     */
    public
    String getParameter(String name)
    {
	try {
	    if (!haveQueryParameters) parseQueryParameters();
	    if (queryParameters == null) return null;
	    Object values = queryParameters.get(name);
	    if (values == null) return null;
	    if (values instanceof String) {
		return (String) values;
	    }
	    if ((values instanceof String[]) &&
		(((String [])values).length > 0)) {
		    return ((String [])values)[0];
	    }
	    return null;
	} catch (IOException e) {
	    // Not much we can do here, since servlet I/F does not
	    // allow this method to throw IOException.
	    logChannel.write(Logger.ERROR,
		"EnhydraDirector Connection " + toString() + ": " +
		"Error processing query parameters: " + e);
	}
	return null;
    }

    /**
     * Get the parameter names for this request.
     * @return The parameter names for this request as an enumeration.
     */
    public
    Enumeration getParameterNames()
    {
	try {
	    if (!haveQueryParameters) parseQueryParameters();
	    if (queryParameters == null) return (Enumeration) this;
	    return queryParameters.keys();
	} catch (IOException e) {
	    // Not much we can do here, since servlet I/F does not
	    // allow this method to throw IOException.
	    logChannel.write(Logger.ERROR,
		"EnhydraDirector Connection " + toString() + ": " +
		"Error processing query parameters: " + e);
	}
	return (Enumeration) this;
    }

    /**
     * Get the values of the specified parameter for the request.
     * <P>
     * The name is case sensitive.
     *
     * @param name the name of the parameter whose value is required.
     * @return The values of the specified parameter for the request as
     *		an array of strings, or null if the named parameter does not
     *		exist.
     * @see javax.servlet.ServletRequest#getParameter
     */
    public
    String[] getParameterValues(String name)
    {
	try {
	    if (!haveQueryParameters) parseQueryParameters();
	    if (queryParameters == null) return null;
	    Object values = queryParameters.get(name);
	    String[] ret = null;
	    if (values == null) return null;
	    if (values instanceof String) {
		ret = new String[1];
		ret[0] = (String) values;
	    } else if (values instanceof String[]) {
		ret = new String[((String [])values).length];
		System.arraycopy((String []) values, 0, ret, 0, ret.length);
	    }
	    return ret;
	} catch (IOException e) {
	    // Not much we can do here, since servlet I/F does not
	    // allow this method to throw IOException.
	    logChannel.write(Logger.ERROR,
		"EnhydraDirector Connection " + toString() + ": " +
		"Error processing query parameters: " + e);
	}
	return null;
    }

    /**
     * Get the protocol and version of the request.
     *
     * @return The protocol and version of the request as a string of
     * the form <code>&lt;protocol&gt;/&lt;major version&gt;.&lt;minor
     * version&gt</code>.
     */
    public
    String getProtocol()
    {
	String ret;
	InternalHeader hdr = (InternalHeader) cgiVars.get("server_protocol");
	if ((hdr == null) || (hdr.size() < 1)) ret = "HTTP/1.0";
	else ret = (String) hdr.lastElement();
	if (debug) {
	    logChannel.write(Logger.DEBUG, "getRemoteAddr() returns \"" +
			     ((ret != null) ? ret : "(null)") +  "\"");
	}
	return ret;
    }

    /**
     * Returns a buffered reader for reading text in the request body.
     * This translates character set encodings as appropriate.
     *
     * @see getInputStream
     *
     * @exception UnsupportedEncodingException if the character set encoding
     *  is unsupported, so the text can't be correctly decoded.
     * @exception IllegalStateException if getInputStream has been
     *	called on this same request.
     * @exception IOException on other I/O related errors.
     * @return A buffered reader object.
     */
    public synchronized
    BufferedReader getReader()
    throws IOException
    {
	if (usingInputStream) {
            throw new IllegalStateException("getInputStream() has been " +
		"called on this request; can't use getReader()");
	}
	usingInputStream = true;
        if (reader == null) {
            reader = new BufferedReader(new InputStreamReader(inStream));
        }
        return reader;
    }

    /**
     * Applies alias rules to the specified virtual path and returns
     * the corresponding real path, or null if the translation can not
     * be performed for any reason.
     *
     * @param path The virtual path to be translated to a real path.
     * @return The real path.
     */
    public
    String getRealPath(String path)
    {
	/* XXX Should call back to web server to get translated path. */
        return servletContext.getRealPath(path);
    }

    /**
     * Get the IP address of the agent that sent the request.
     * @returns The IP address.
     */
    public String getRemoteAddr()
    {
	String ret;
	InternalHeader hdr = (InternalHeader) cgiVars.get("remote_addr");
	if ((hdr == null) || (hdr.size() < 1)) ret = null;
	else ret = (String) hdr.lastElement();
	if (debug) {
	    logChannel.write(Logger.DEBUG, "getRemoteAddr() returns \"" +
			     ((ret != null) ? ret : "(null)") +  "\"");
	}
	return ret;
    }

    /**
     * Get the fully qualified host name of the agent that sent the
     * request.
     *
     * @return The host name.
     */
    public
    String getRemoteHost()
    {
	String ret;
	InternalHeader hdr = (InternalHeader) cgiVars.get("remote_host");
	if ((hdr == null) || (hdr.size() < 1)) ret = null;
	else ret = (String) hdr.lastElement();
	if (debug) {
	    logChannel.write(Logger.DEBUG, "getRemoteHost() returns \"" +
			     ((ret != null) ? ret : "(null)") +  "\"");
	}
	return ret;
    }

    /*
     * NEW
     */
/* XXX 2.1
    public
    RequestDispatcher getRequestDispatcher(String path)
    {
	return servletContext.getRequestDispatcher(path);
    }
 */

    /**
     * Get the scheme of the URL used in this request, for example
     * "http", "https", or "ftp".
     *
     * @return The scheme of the URL.
     */
    public
    String getScheme()
    {
	String ret;
	InternalHeader hdr = (InternalHeader) cgiVars.get("server_scheme");
	if ((hdr == null) || (hdr.size() < 1)) ret = "http";
	else ret = (String) hdr.lastElement();
	if (debug) {
	    logChannel.write(Logger.DEBUG, "getScheme() returns \"" +
			     ((ret != null) ? ret : "(null)") +  "\"");
	}
	return ret;
    }

    /**
     * Get the host name of the server that received the request.
     * @return The host name.
     */
    public String getServerName()
    {
	String ret;
	InternalHeader hdr = (InternalHeader) cgiVars.get("server_name");
	if ((hdr == null) || (hdr.size() < 1)) ret = "http";
	else ret = (String) hdr.lastElement();
	if (debug) {
	    logChannel.write(Logger.DEBUG, "getServerName() returns \"" +
			     ((ret != null) ? ret : "(null)") +  "\"");
	}
	return ret;
    }

    /**
     * Get the port number on which this request was received.
     * @returns The port number.
     */
    public
    int getServerPort()
    {
	String ret;
	int iret;
	InternalHeader hdr = (InternalHeader) cgiVars.get("server_port");
	if ((hdr == null) || (hdr.size() < 1)) ret = "-1";
	else ret = (String) hdr.lastElement();
	try {
	    iret = Integer.parseInt(ret);
	} catch (Exception e) { iret = -1; }
	if (debug) {
	    logChannel.write(Logger.DEBUG, "getServerPort() returns " + iret);
	}
	return iret;
    }

    /**
     * NEW
     */
    public
    boolean isSecure()
    {
        // XXX FIXME: Not implemented.
	return false;
    }

    /**
     * NEW
     */
    public
    void removeAttribute(String name)
    {
	attributes.remove(name);
    }

    /**
     * NEW
     */
    public
    void setAttribute(String name)
    {
        // XXX FIXME: Attributes are not yet supported by the Multiserver.
    }


    //=============================================================
    // Methods to implement the javax.servlet.ServletResponse
    //=============================================================

    /**
     * Sets a response header with the given name and integer value.
     * <P>
     * If the header has already been set, the new value overwrites the
     * previous one.  The <code>containsHeader</code> method can be used
     * to test for the presence of a header before setting its value.
     *
     * @param name The name of the header.
     * @param value The value to assign.
     */
    public
    void setHeader(String name, String value)
    {
	if (debug) {
	    logChannel.write(Logger.DEBUG,
		    "setHeader(\""+name+"\", \""+value+"\")");
	}
	String lname = name.toLowerCase();
	responseHeaders.remove(lname);
	InternalHeader hdr = new InternalHeader();
	hdr.name = name;
	hdr.addElement(value);
	responseHeaders.put(lname, hdr);
    }

    /**
     * Adds a response header with the given name and integer value.
     * <P>
     * This method allows response headers to have multiple values.
     *
     * @param name The name of the header.
     * @param value The value to assign.
     */
    public
    void addHeader(String name, String value)
    {
	if (debug) {
	    logChannel.write(Logger.DEBUG,
		    "addHeader(\""+name+"\", \""+value+"\")");
	}
	String lname = name.toLowerCase();
	InternalHeader hdr = (InternalHeader) responseHeaders.get(lname);
	if (hdr == null) {
	    hdr = new InternalHeader();
	    hdr.name = name;
	    responseHeaders.put(lname, hdr);
	}
	hdr.addElement(value);
    }

    /**
     * Add a cookie to the list of cookies to send to the browser.
     *
     * @param cookie The cookie to add.
     */
    public
    void addCookie(Cookie cookie)
    {
	responseCookies.addElement(cookie);
    }

    /**
     * Returns <code>true</code> if the current response contains the
     * named header.
     *
     * @param name The name of the header.
     * @return whether the header is contained in the response.
     */
    public
    boolean containsHeader(String name)
    {
	String lname = name.toLowerCase();
	return (responseHeaders.get(lname) != null);
    }

    /**
     * Encodes the specified URL by including the session ID in it,
     * or if encoding is not needed, returns the URL unchanged.
     *
     * @param the url to be encoded.
     * @return The encoded URL if encoding is needed; the unchanged URL
     * 		otherwise.
     */
    public
    String encodeUrl(String url)
    {
	return url; // Sessions not yet implemented. XXX FIXME
    }

    /**
     * Encodes the specified URL for use in the
     * <code>sendRedirect</code> method.
     *
     * @param the url to be encoded.
     * @return The encoded URL if encoding is needed; the unchanged URL
     * 		otherwise.
     */
    public
    String encodeRedirectUrl(String url)
    {
	return url; // Sessions not yet implemented. XXX FIXME
    }

    /**
     * Generate an HTML error page for sendError.
     */
    private
    String getErrorPage(int sc, String msg)
    {
        StringBuffer buf = new StringBuffer();

        //FIXME: Need to figure who generates <TITLE>
        //Should it be in msg or here.
        buf.append("<HTML><HEAD><TITLE>");
        buf.append(HttpStatus.getStatusMsg(sc));
        buf.append("</TITLE><CENTER><H1>");
        buf.append(HttpStatus.getStatusMsg(sc));
        buf.append("</H1></CENTER>");
        if (msg != null) {
            buf.append(msg);
            buf.append("<P>");
        }
        buf.append("</BODY></HTML>");
        return buf.toString();
    }

    /**
     * Sends an error response to the client using the specified status
     * code and descriptive message.
     *
     * @param sc The status code.
     * @param msg The descriptive message.
     * @exception IOException If an I/O error has occured.
     */
    public
    void sendError(int sc, String msg)
    throws IOException
    {
	String page = getErrorPage(sc, msg);
	if (!committed) {
	    reset();
	    setStatus(sc);
	    setContentType("text/html");
	    setContentLength(page.length());
	    commit();
	}
	outStream.print(page);
	outStream.flush();
    }

    /**
     * Sends an error response to the client using the specified status
     * code and a default message.
     *
     * @param sc The status code.
     * @exception IOException If an I/O error has occured.
     */
    public
    void sendError(int sc)
    throws IOException
    {
	sendError(sc, HttpStatus.getStatusMsg(sc));
    }

    /**
     * Generate an HTML page for sendRedirect.
     *
     * @param url The URL to redirect to.
     * @return the HTML page.
     */
    private
    String getRedirectPage(String url)
    {
        StringBuffer buf = new StringBuffer();
        buf.append("<HTML><HEAD><TITLE>Document Has Moved</TITLE>");
        buf.append("<BODY><CENTER><H1>Document Has Moved</H1></CENTER>");
        buf.append("This document has moved to <A HREF=\"");
        buf.append(url);
        buf.append("\">");
        buf.append(url);
        buf.append("</A><P>");
        buf.append("</BODY></HTML>\r\n");
        return buf.toString();
    }

    /**
     * Sends a temporary redirect response to the client using the
     * specified redirect location URL.
     *
     * @param location The redirect location URL.
     * @exception IOException If an I/O error has occured.
     */
    public
    void sendRedirect(String location)
    throws IOException
    {
	if (debug) {
	    logChannel.write(Logger.DEBUG, "REDIRECT: " + location);
	}
	String page = getRedirectPage(location);
	Hashtable saveHeaders = responseHeaders;
	Vector saveCookies = responseCookies;
	reset();
	responseHeaders = saveHeaders;
	responseCookies = saveCookies;
	setStatus(HttpStatus.movedTemporarily);
	setContentType("text/html");
	setHeader("Location", location);
	setContentLength(page.length());
	commit();
	outStream.print(page);
	outStream.flush();
    }

    /**
     * Sets a response header with the given name and date value.
     * <P>
     * If the header has already been set, the new value overwrites the
     * previous one.  The <code>containsHeader</code> method can be used
     * to test for the presence of a header before setting its value.
     *
     * @param name The name of the header.
     * @param value The value to assign.
     */
    public
    void setDateHeader(String name, long date)
    {
        setHeader(name, dateFormatter.format(new Date(date)));
    }

    /**
     * Adds a response header with the given name and date value.
     * <P>
     * This method allows response headers to have multiple values.
     *
     * @param name The name of the header.
     * @param value The value to assign.
     */
    public
    void addDateHeader(String name, long date)
    {
        addHeader(name, dateFormatter.format(new Date(date)));
    }

    /**
     * Sets a response header with the given name and integer value.
     * <P>
     * If the header has already been set, the new value overwrites the
     * previous one.  The <code>containsHeader</code> method can be used
     * to test for the presence of a header before setting its value.
     *
     * @param name The name of the header.
     * @param value The value to assign.
     */
    public
    void setIntHeader(String name, int value)
    {
	setHeader(name, "" + value);
    }

    /**
     * Adds a response header with the given name and integer value.
     * <P>
     * This method allows response headers to have multiple values.
     *
     * @param name The name of the header.
     * @param value The value to assign.
     */
    public
    void addIntHeader(String name, int value)
    {
	addHeader(name, "" + value);
    }

    /**
     * Sets the status code for this response.
     *
     * @param sc The HTTP status code.
     * @param sm Message to send with status code.
     */
    public
    void setStatus(int sc, String sm)
    {
	if (debug) {
	    logChannel.write(Logger.DEBUG, "setStatus("+sc+", \""+sm+"\")");
	}
	if (committed) throw new IllegalStateException("Already committed.");
	statusCode = sc;
	if (sm == null) statusMsg = HttpStatus.getStatusMsg(sc);
	else statusMsg = sm;
    }

    /**
     * Sets the status code for this response.
     * <P>
     * This method is used to set the return status code when there
     * is no error.  (For example, SC_OK or SC_MOVED_TEMPORARILY).  If there
     * is an error, the <code>sendError</code> method should be used instead.
     *
     * @param sc The status code.
     */
    public
    void setStatus(int sc)
    {
	setStatus(sc, HttpStatus.getStatusMsg(sc));
    }

    //=============================================================
    // Methods to implement the javax.servlet.ServletResponse
    //=============================================================

    /**
     * Returns the charset used for the body.
     * <P>
     * This is not fully implemented.  It just returns "ISO-8859-1".
     * FIXME XXX This should use correct encodings.
     *
     * @return "ISO-8859-1".
     */
    public
    String getCharacterEncoding()
    {
	return "ISO-8859-1";
    }

    /**
     * Returns the byte output stream object for this response.
     * <P>
     * This method may not be called if it or <code>getWriter()</code>
     * have already been called.
     *
     * @return The output stream.
     * @exception IllegalStateException If this method is called after the
     * 		output stream has already been gotten.
     */
    public
    ServletOutputStream getOutputStream()
    throws IOException
    {
	if (writer != null) {
	    throw new IllegalStateException("getWriter() alrady called.");
	}
	usingOutputStream = true;
	return outStream;
    }

    /**
     * Gets a character stream printing writer for the response.
     * <P>
     * This implementation is not fully functional, since it uses
     * the default character encoding instead of mapping the charset
     * portion of the Content-Type to the correct encoding.
     *
     * @return A <code>PrintWriter</code> stream for output.
     * @exception IOException If an I/O error occurs.
     * @exception IllegalStateException If getOutputStream() already called.
     * @exception UnsupportedEncodingException If an unknown character
     * 		encoding is specified.
     */
    public
    PrintWriter getWriter()
    throws IOException
    {
	if (usingOutputStream) {
	    throw new IllegalStateException(
			"getOutputStream() already called.");
	}
	// XXX FIX ME - Should use the correct character encoding.
	if (writer == null) {
	    OutputStreamWriter osw = new OutputStreamWriter(outStream);
	    writer = new PrintWriter(osw);
	}
	return writer;
    }

    /**
     * Sets the value of the outgoing <code>Content-Length</code> header.
     * <P>
     * Sets the length of the outgoing content.  Any data sent beyond
     * this length will be discarded, and the output will be padded
     * to this length after the output stream is closed.  Padding is
     * with octets of value 0 (zero).
     *
     * @param len The Content-Length value.
     */
    public
    void setContentLength(int len)
    {
	setIntHeader("Content-Length", len);
    }

    /**
     * Sets the MIME content type of the outgoing response.
     * <P>
     * FIXME XXX - This is not fully implemented.  It simply copies
     * the type verbatim to the MIME headers sent with the response.
     * Future versions of this class may attempt to read the
     * "charset" portion of the content type to determine which
     * encoding, if any, should be used for streams returned by
     * <code>getWriter()</code>.  This will require a configuration
     * table that maps locales, MIME charsets, and Java encodings.
     *
     * @param type The MIME type of this response.
     */
    public
    void setContentType(String type)
    {
	setHeader("Content-Type", type);
    }

    /**
     * Sets the preferred buffer size for the body of the response.
     * <P>
     * The servlet container will use a buffer at least as large as
     * the size requested.  The actual buffer size used can be found
     * using <code>getBufferSize</code>.
     * <P>
     * A larger buffer allows more content to be written before anything
     * is actually sent, this providing the servlet with more time to set
     * appropriate status codes and headers.  A smaller buffer decreases
     * server memory load and allows the client to start receiving data
     * more quickly.
     * <P>
     * This method must be called before any response body content is
     * written; if content has been written, this method throws
     * an <code>IllegalStateException</code>.
     *
     * @param size The preferred buffer size.
     * @exception IllegalStateException If called after content written.
     */
    public
    void setBufferSize(int size)
    {
	outStream.setBufferSize(size);
    }

    /**
     * Returns the actual buffer size used for the response.
     * <P>
     * If no buffering is used, this method returns <code>0</code>.
     *
     * @return The size of the send buffer.
     */
    public
    int getBufferSize()
    {
	return outStream.getBufferSize();
    }

    /**
     * Forces any content in the buffer to be written to the client.
     * <P>
     * A call to this method automatically commits the response, meaning the
     * status code and headers will be written.
     *
     * @exception IOException If an I/O error occurs.
     */
    public
    void flushBuffer()
    throws IOException
    {
	if (!committed) commit();
	outStream.flush();
    }

    /**
     * Returns a boolean indicating whether the response has been committed.
     * <P>
     * A committed response has already had its status code and headers
     * written.
     *
     * @return A boolean indicating if the response has been committed.
     */
    public
    boolean isCommitted()
    {
	return committed;
    }

    /**
     * Clears any data that exists in the buffer as well as the status code
     * and headers.  If the response has been commited, this method throws
     * an <code>IllegalStateException</code>.
     *
     * @exception IllegalStateException If the response has already
     * 		been committed.
     */
    public
    void reset()
    throws IllegalStateException
    {
	if (committed)
	    throw new IllegalStateException("Response already committed.");
	outStream.reset();
	responseHeaders = new Hashtable();
	responseCookies = new Vector();
	statusCode = 200;
	statusMsg = HttpStatus.getStatusMsg(statusCode);
    }

    /**
     * Sets the locale of the response.
     * <P>
     * This is supposed to set the headers, including <code>Content-Type</code>
     * and <code>Content-Encoding</code>.  However, it is not * clear how to
     * map between character encodings and locales.
     * <P>
     * Probably, future revisions of this class will use some sort of mapping
     * table to set the encodings.  For now, it just keeps track of which
     * locale is set and uses Latin1 encoding in the underlying output
     * stream.  Servlets that want to use other encodings should create an
     * appropriate <code>PrintWriter</code> on the binary stream given by
     * <code>getOutputStream</code>.
     * XXX - FIXME.
     *
     * @param loc The local set for output on this response.
     */
    public
    void setLocale(Locale loc)
    {
	if (loc != null) {
	    locale = loc;
	} else {
	    locale = Locale.getDefault();
	}
    }

    /**
     * Gets the locale of the response.
     * <P>
     * Locale is supposed to determine some attributes of headers, including
     * <code>Content-Type </code> and <code>Content-Encoding</code>.
     * Unfortunately, it is not clear how to map between character encodings
     * and locales.
     * <P>
     * Probably, future revisions of this class will use some sort
     * of mapping table to set the encodings.  For now, it just keeps
     * track of which locale is set and uses Latin1 encoding in the
     * underlying output stream.  Servlets that want to use other
     * encodings should create an appropriate <code>PrintWriter</code>
     * on the binary stream given by <code>getOutputStream</code>.
     * XXX - FIXME.
     *
     * @return The local set for output on this response.
     */
    public
    Locale getLocale()
    {
	return locale;
    }

    public RequestDispatcher getRequestDispatcher(String name) {
	// X22
	throw new Error("EnhydraDirectorConnection.getRequestDispatcher() not implemented!");
    }

    public String encodeURL(String url) {
	// X22
	throw new Error("EnhydraDirectorConnection.encodeURL() not implemented!");
    }

    public void setAttribute(String name, Object value) {
        attributes.put(name, value);
    }

    public String encodeRedirectURL(String url) {
	// X22
	throw new Error("EnhydraDirectorConnection.encodeRedirectURL() not implemented!");
    }

    // X22 for RequestAdapter
    public void recycle() {
        queryString = null;
        attributes.clear();
	// XXX call close()?
	// servletPath = null;
	// servletContext = null;
	// reader = null;
    }

    // X22 for RequestAdapter
    public MimeHeaders getMimeHeaders() {
	/* InternalHeader extends Vector, so this should be OK */
	return HeaderConverter.toMimeHeaders(requestHeaders);
    }

    // X22 for RequestAdapter
    public String[] getCookieHeaders() {
	// X22 doesn't seem to be used
	throw new Error("not implemented");
    }

    // X22 for RequestAdapter
    public String getServletPrefix() {
	// X22 doesn't seem to be used
	throw new Error("not implemented");
    }

    // X22 for RequestAdapter
    public String getServletName() {
	// X22 doesn't seem to be used
	throw new Error("not implemented");
    }

    // X22 for RequestAdapter
    public int doRead() throws IOException {
	// X22 might need some checking here first
	return inStream.read();
    }

    // X22 for RequestAdapter
    public int doRead(byte[] b, int off, int len) throws IOException {
	// X22 might need some checking here first
	return inStream.read(b, off, len);
    }

    // X22 ResponseAdapter
    // XXX This one or multiple addHeader?
    // Probably not a big deal - but an adapter may have
    // an optimized version for this one ( one round-trip only )
    public void addMimeHeaders(MimeHeaders headers) throws IOException {
	int size = headers.size();
        for (int i = 0; i < size; i++) {
            MimeHeaderField h = headers.getField(i);
            addHeader(h.getName(), h.getValue());
        }
    }

    // X22 ResponseAdapter
    /** Signal that we're done with a particular request, the
	server can go on and read more requests or close the socket
    */
    public void endResponse() throws IOException {
	// X22 should we call finish here?
    }

    private static final byte[] CRLF = { (byte)'\r', (byte)'\n' };

    // X22 ResponseAdapter
    /** Signal that we're done with the headers, and body will follow.
	The adapter doesn't have to maintain state, it's done inside the engine
    */
    public void endHeaders() throws IOException {
	// X22 need to send CRLF here?
	// outStream.write(CRLF, 0, 2);
    }

    // X22 ResponseAdapter
    public ServletOutputStream getServletOutputStream() throws IOException {
	return outStream;
    }

    // X22 ResponseAdapter
    // X22 recycle is in both RequestAdapter and ResponseAdapter

    /** Write a chunk of bytes. Should be called only from ServletOutputStream implementations,
     *	No need to implement it if your adapter implements ServletOutputStream.
     *  Headers and status will be written before this method is exceuted.
     */
    public void doWrite(byte buffer[], int pos, int count) throws IOException {
	// X22 is it right to flush() here?
	outStream.write(buffer, pos, count);
	outStream.flush();
    }

    /**
     * Method to close output stream.
     * Overrides protocol engine's  closeOutput to add a commit().
     */
    public
    void closeOutput()
    throws IOException
    {
	if (!committed) commit();
	super.closeOutput();
    }

}

class InternalHeader extends Vector {
    public String	name;
}

class InternalHeaderKeyEnumeration implements Enumeration {
    private Hashtable headers = null;
    private Enumeration keyEnum = null;

    public InternalHeaderKeyEnumeration(Hashtable headers)
    {
	this.headers = headers;
	keyEnum = headers.keys();
    }

    public boolean hasMoreElements() {
	return keyEnum.hasMoreElements();
    }

    public Object nextElement() throws NoSuchElementException {
	String key = (String) keyEnum.nextElement();
	InternalHeader hdr = (InternalHeader) headers.get(key);
	if (hdr == null) return key;
	return hdr.name;
    }
}

