/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: EnhydraDirectorOutputStream.java,v 1.4.2.1.2.1 2000/10/19 17:59:09 jasona Exp $
 */

package org.enhydra.servlet.connectionMethods.EnhydraDirector;

import javax.servlet.ServletOutputStream;
import java.util.Enumeration;
import java.io.OutputStream;
import java.io.IOException;
import java.net.Socket;
import org.enhydra.servlet.ServletIOException;
import com.lutris.logging.LogChannel;
import com.lutris.logging.Logger;

/**
 * <code>OutputStream</code> interface for the Enhydra connection method.
 * <P>
 * Implemention of <code>ServletOutputStream</code> to be passed to 
 * servlets to allow them to send response data to the client.
 * <P>
 * This class automatically wrappers a lower level output stream in
 * a <code>BufferedOutputStream</code> if buffering is requested by
 * the servlet.
 *
 * @author John Marco
 * @version $Revision: 1.4.2.1.2.1 $
 * @see EnhydraResponse
 * @see javax.servlet.ServletOutputStream
 */
class EnhydraDirectorOutputStream extends ServletOutputStream
{
    /**
     * Internal EnhydraDirector protocol driver instance.
     */
    private EnhydraDirectorConnection conn;

    /**
     * The actual OutputStream.  Not set until buffering is finalized.
     */
    private OutputStream out;

    /**
     * The underlying <code>EnhydraDirectorProtoOutputStream</code>.
     */
    private OutputStream os;

    /**
     * Whether headers have been written.  Saves a method call.
     */
    private boolean isFirstWrite;

    /**
     * Buffer size.  If less than or equal to zero, no buffering.
     */
    private int bufSize;

    /**
     * Debug flag
     */
    private static final boolean debug = false;

    /**
     * Construct an output stream associate with a response object.
     * 
     * @param conn EnhydraDirector protocol driver instance.
     * @param resp EnhydraResponse instance to allow headers to be sent.
     * @param bufsize Buffer size to use if greater than zero.
     */
    public
    EnhydraDirectorOutputStream(EnhydraDirectorConnection conn)
    throws IOException
    {
	if (debug) conn.logChannel.write(Logger.DEBUG,
			"EnhydraDirectorOutputStream(" + conn + ")");
        this.conn = conn;
	this.out = null;
	this.isFirstWrite = true;
	this.bufSize = 0;
	this.os = new EnhydraDirectorProtoOutputStream(conn);
    }

    /**
     * Set the buffer size.
     * <P>
     * Sets the buffer size for this output stream.  If less than or equal 
     * to zero, no buffering is done and output is sent directly to
     * the protocol engine's output methods.
     * <P>
     * This method throws an <code>IllegalStateException</code> if it
     * is called after the first write to response output.
     * 
     * @param size The size of the buffer, or <code>0</code> for no buffering.
     * @exception IllegalStateException If data has already been written.
     */
    public
    void setBufferSize(int size)
    {
	if (!isFirstWrite)
	    throw new IllegalStateException("Data already written.");
	bufSize = size;
    }

    /**
     * Get the buffer size.
     * <P>
     * Gets the buffer size for this output stream.  If less than or equal 
     * to zero, no buffering is done and output is sent directly to
     * the protocol engine's output methods.
     * 
     * @return The size of the buffer, or <code>0</code> for no buffering.
     */
    public
    int getBufferSize()
    {
	return bufSize;
    }

    /**
     * Internal method to be called upon the first write of data.
     * <P>
     * Sets up the output stream as either a buffered or unbuffered
     * stream depending on the buffer size.
     * <P>
     * Note.  Does NOT commit() the connection, since we only send
     * headers when data bytes are about to be actually sent over the
     * wire.
     * 
     * @exception IOException If there is an error setting up the stream.
     */
    private synchronized
    void firstWrite()
    throws IOException
    {
	if (isFirstWrite) {
	    try {
		if (bufSize > 0) {
		    out = new EnhydraDirectorBufferedOutputStream(os,
							    bufSize, conn);
		} else {
		    out = os;
		}
		isFirstWrite = false;
	    } catch (IllegalArgumentException iae) {
		// We don't ever expect illegal argument because
		// we check.
		throw new IOException("Unexpected Illegal argument");
	    }
	}
    }

    /**
     * Write a single byte of response data.
     * <P>
     * This method is extremely inefficient, since it sends a separate
     * packet for each outgoing response byte.  Servlets should pass this
     * output stream to the constructor of a new <code>BufferedOutputStream
     * </code> if the single byte <code>write()</code> call is going 
     * to be used frequently.
     *
     * @param b Integer with a value between 0 and 255, inclusive.
     * @exception IOException If an I/O error occurs.
     */
    public
    void write(int b)
    throws IOException
    {
	if (debug) conn.logChannel.write(Logger.DEBUG, "write(" + b + ")");
	if (isFirstWrite) firstWrite();
	if (out == null) throw new IOException("Stream closed");
	out.write(b);
    }

    /**
     * Write an array of bytes to this output stream. 
     *
     * @param bytes Array of byte data to write to the stream.
     * @param off Starting offset in the array.
     * @param len Number of bytes to write.
     * @exception  IOException If an I/O error occurs.
     */
    public
    void write(byte[] bytes, int off, int len)
    throws IOException
    {
	if (debug) conn.logChannel.write(Logger.DEBUG, "write(" + bytes + 
			    ", " + off + ", " + len + ")");
	if (isFirstWrite) firstWrite();
	if (out == null) throw new IOException("Stream closed");
	out.write(bytes, off, len);
    }

    /**
     * Write an array of bytes to this output stream. 
     *
     * @param bytes Array of byte data to write to the stream.
     * @exception  IOException If an I/O error occurs.
     */
    public
    void write(byte[] bytes)
    throws IOException
    {
	if (debug) conn.logChannel.write(Logger.DEBUG, "write(" + bytes + ")");
	if (isFirstWrite) firstWrite();
	if (out == null) throw new IOException("Stream closed");
	out.write(bytes, 0, bytes.length);
    }

    /**
     * Writes a string to response data using simple ASCII-8 encoding.
     * <P>
     * <B>Sigh</B>
     * <P>
     * This <B>really</B> does not belong at this level.  The correct
     * solution would have been to have servlets pass a plain, simple
     * well defined binary <code>OutputStream</code> to a <code>
     * PrintStream</code> or <code>PrintWriter</code> if they wanted
     * to have fully encoded string writing.
     * <P>
     * This implementation supports <I>only</I> simple 8 bit ASCII
     * encoding.  If non-ASCII unicode characters are used, they WILL
     * be truncated.  The correct solution in this case is to use
     * the appropriate <code>PrintStream</code> or <code>PrintWriter</code>
     * classes to do the string-to-binary encoding at a higher level.
     * <P>
     * String-to-binary encoding is extremely slow.
     *
     * @param s The string to write.
     * @exception IOException if an I/O error has occurred
     * @see javax.servlet.ServletOutputStream#print
     */
    public
    void print(String s)
    throws IOException
    {
	if (debug) conn.logChannel.write(Logger.DEBUG, "print(\"" + s + "\")");
	if (isFirstWrite) firstWrite();
	if (out == null) throw new IOException("Stream closed");
	/* This is what I HATE about java.  There are two full copies
	 * of the string in the code below.
	 */
	char[] cbuf = s.toCharArray();
	byte[] bbuf = new byte[cbuf.length];
	for (int i=0; i<bbuf.length; i++) {
	    bbuf[i] = (byte) cbuf[i];
	}
	out.write(bbuf, 0, bbuf.length);
    }

    /**
     * Writes a character to response data using simple Latin-1 encoding.
     * <P>
     * <B>Sigh</B>
     * <P>
     * This <B>really</B> does not belong at this level.  The correct
     * solution would have been to have servlets pass a plain, simple
     * well defined binary <code>OutputStream</code> to a <code>
     * PrintStream</code> or <code>PrintWriter</code> if they wanted
     * to have fully encoded string writing.
     * <P>
     * This implementation supports <I>only</I> simple 8 bit ASCII
     * encoding.  If non-ASCII unicode characters are used, they WILL
     * be truncated.  The correct solution in this case is to use
     * the appropriate <code>PrintStream</code> or <code>PrintWriter</code>
     * classes to do the string-to-binary encoding at a higher level.
     * <P>
     * String-to-binary encoding is extremely slow.
     *
     * @param c The character to write.
     * @exception IOException if an I/O error has occurred
     * @see javax.servlet.ServletOutputStream#print
     */
    public
    void print(char c)
    throws IOException
    {
	if (debug) conn.logChannel.write(Logger.DEBUG, "print('" + c + "')");
	if (isFirstWrite) firstWrite();
	if (out == null) throw new IOException("Stream closed");
	out.write((byte) c);
    }

    /**
     * Commits headers and immediately sends any outgoing
     * data over the connection.
     *
     * @exception  IOException If an I/O error occurs.
     */
    public
    void close()
    throws IOException
    {
	if (debug) conn.logChannel.write(Logger.DEBUG, "close()");
	if (isFirstWrite) firstWrite();
	if (out == null) throw new IOException("Stream closed");
	out.flush();
	out.close();
	out = null;
    }

    /**
     * Closes this output stream and releases any system resources 
     * associated with the stream. 
     *
     * @exception  IOException If an I/O error occurs.
     */
    public
    void flush()
    throws IOException
    {
	if (debug) conn.logChannel.write(Logger.DEBUG, "flush()");
	if (isFirstWrite) firstWrite();
	if (out == null) throw new IOException("Stream closed");
	out.flush();
    }

    /**
     * Resets the underlying output stream, discarding any buffers.
     * 
     * @exception IllegalStateException
     * 		If the stream is not buffered, or if buffer data has already
     * 		been flushed/written to the connection.
     */
    public
    void reset()
    throws IllegalStateException
    {
	if (debug) conn.logChannel.write(Logger.DEBUG, "reset()");
	if (isFirstWrite || (out == null)) return; // Haven't written yet.
	if (bufSize < 1) 
	    throw new IllegalStateException("No buffering: cannot reset");
	if (os == null)
	    throw new IllegalStateException("No stream: cannot reset");
	((EnhydraDirectorBufferedOutputStream)out).reset();
	// Start again with a fresh buffer stream.
	out = new EnhydraDirectorBufferedOutputStream(os, bufSize, conn);
    }
}

