/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: EnhydraDirectorProtocol.java,v 1.5.2.4.2.1 2000/10/19 17:59:09 jasona Exp $
 */

package org.enhydra.servlet.connectionMethods.EnhydraDirector;

import java.net.ServerSocket;
import java.net.Socket;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.security.SecureRandom;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import com.lutris.logging.LogChannel;
import com.lutris.logging.Logger;

public abstract class EnhydraDirectorProtocol
{
    protected static final int		MAJOR_VERSION = 1;
    protected static final int		MINOR_VERSION = 0;

    protected static final int		STATE_CLOSED = 1;
    protected static final int		STATE_NEWCONNECT = 2;
    protected static final int		STATE_WAIT_AUTH = 3;
    protected static final int		STATE_WAIT_VERSION = 4;
    protected static final int		STATE_CHECK_AUTH = 5;
    protected static final int		STATE_OPENING = 6;
    protected static final int		STATE_WAIT_OPEN_CALLBACK = 7;
    protected static final int		STATE_OPEN = 8;
    protected static final int		STATE_PROCESSING_FUNC = 9;
    protected static final int		STATE_PROCESSING_HEADER = 10;
    protected static final int		STATE_RECEIVING_HEADERS = 11;
    protected static final int		STATE_HAVE_HEADERS = 12;
    protected static final int		STATE_WAIT_HEADER_CALLBACK = 13;
    protected static final int		STATE_RECEIVING_ENTITY = 14;
    protected static final int		STATE_WAIT_ENTITY_CALLBACK = 15;
    protected static final int		STATE_PROCESSING_ENTITY_FUNC = 16;
    protected static final int		STATE_PROCESSING_REQUEST = 17;

    protected static final int		PKT_TYPE_AUTH = 0;
    protected static final int		PKT_SUBTYPE_AUTH_SVR = 0;
    protected static final int		PKT_SUBTYPE_AUTH_CLI = 1;
    protected static final int		PKT_SUBTYPE_AUTH_OK = 2;
    protected static final int		PKT_SUBTYPE_AUTH_FAIL = 3;

    protected static final int		PKT_TYPE_REQ = 1;
    protected static final int		PKT_SUBTYPE_REQ_DATA = 0;
    protected static final int		PKT_SUBTYPE_REQ_LASTDATA = 1;

    protected static final int		PKT_TYPE_FUNC = 2;
    protected static final int		PKT_SUBTYPE_FUNC_FAIL = 0;
    protected static final int		PKT_SUBTYPE_FUNC_OK = 1;
    protected static final int		PKT_SUBTYPE_FUNC_CALL = 2;
    protected static final int		PKT_SUBTYPE_FUNC_ONEWAY = 3;
    protected static final int		PKT_SUBTYPE_FUNC_END = 4;

    protected static final int		PKT_TYPE_RESP = 3;
    protected static final int		PKT_SUBTYPE_RESP_DATA = 0;
    protected static final int		PKT_SUBTYPE_RESP_LOGS = 1;
    protected static final int		PKT_SUBTYPE_RESP_WARNING = 2;
    protected static final int		PKT_SUBTYPE_RESP_ERROR = 3;
    protected static final int		PKT_SUBTYPE_RESP_END = 4;

    protected static final int		PKT_TYPE_HEADER = 4;
    protected static final int		PKT_SUBTYPE_HEADER_HEADER = 0;
    protected static final int		PKT_SUBTYPE_HEADER_HTTP = 1;
    protected static final int		PKT_SUBTYPE_HEADER_COOKIE = 2;
    protected static final int		PKT_SUBTYPE_HEADER_LAST = 15;

    protected static final int		PKT_TYPE_CTRL_CLI = 5;
    protected static final int		PKT_SUBTYPE_CTRL_CLI_ACTIVATE = 0;
    protected static final int		PKT_SUBTYPE_CTRL_CLI_DEACTIVATE = 1;
    protected static final int		PKT_SUBTYPE_CTRL_CLI_CLOSE = 2;

    protected static final int		PKT_TYPE_CTRL_SVR = 6;
    protected static final int		PKT_SUBTYPE_CTRL_SVR_REFRESH = 0;

    protected static final int		PKT_TYPE_PROTO = 15;
    protected static final int		PKT_SUBTYPE_PROTO_CLOSE = 0;
    protected static final int		PKT_SUBTYPE_PROTO_FATAL = 1;
    protected static final int		PKT_SUBTYPE_PROTO_VERSION = 15;

    protected static final int		PKT_OPENFLAGS_OPENCB = 0x01;
    protected static final int		PKT_OPENFLAGS_HEADERCB = 0x02;

    protected static final int		PKT_HEADER_SIZE = 4;

    protected static final int		MAX_AUTHDATA_LENGTH = 128;

    protected static final int		MAX_FUNCDATA_LENGTH = 10000;

    protected static final int		MAX_SMALL_PACKET_LENGTH = 10000;

    /**
     * The current state of this connection's state machine.
     */
    protected int state = STATE_CLOSED;

    /**
     * Client side protocol major version.
     */
    protected int clientMajor = 0;

    /**
     * Client side protocol minor version.
     */
    protected int clientMinor = 0;

    /**
     * Whether or not this connection is active.
     */
    protected boolean active = false;

    /**
     * The TCP socket connection to the client.
     */
    private Socket 	 sock = null;

    /**
     * The input stream for <code>sock</code>.
     */
    private InputStream in;

    /**
     * The output stream for <code>sock</code>.
     */
    private OutputStream out;

    /**
     * Random number generator used to generate random authentication
     * challenges to send to the client.
     */
    private static SecureRandom randomizer = new SecureRandom();

    /**
     * Whether the random number generator has already been seeded.
     */
    private boolean randomizerSeeded = false;

    /**
     * Whether post-open callbacks have been requested.
     */
    private boolean openCBRequested = false;

    /**
     * Whether post-header callbacks have been requested.
     */
    private boolean headerCBRequested = false;

    /**
     * Packet type of the currently processed input packet.
     */
    private int packetType;

    /**
     * Packet subtype of the currently processed input packet.
     */
    private int packetSubType;

    /**
     * Packet data length of the currently processed input packet.
     */
    private int packetLength;

    /**
     * Whether the incoming request entity has been completely received.
     */
    private boolean entityReceived = false;

    /**
     * Whether someone has called closeInput on this request.
     */
    private boolean inputClosed = false;

    /**
     * Temporary buffer for incoming packet data.
     */
    private byte[] packetData = null;

    /**
     * Temporary buffer for outgoing function data.
     */
    private byte[] funcData = null;

    /**
     * Temporary between-state holder for incoming HTTP header name
     */
    private byte[] headerName = null;

    /**
     * Temporary between-state holder for incoming HTTP header value
     */
    private byte[] headerValue = null;

    /**
     * Wheter the last function call succeeded.
     */
    private boolean funcSuccess = false;

    /**
     * The number of remaining data bytes in the current entity fragment
     * packet.
     */
    private int entityRemainder = 0;

    /**
     * The number of remaining data bytes in the current entity fragment
     * packet.
     */
    private int packetRemainder = 0;

    /**
     * Local scratch buffer used by the sendPacket() methods to
     * build packet headers.
     */
    private byte[] pktHdrBuf = new byte[4];

    /**
     * Local scratch buffer used by the receiveSinglePacket() method to
     * receive packet headers.
     */
    private byte[] pktHdrInBuf = new byte[4];

    /**
     * Local scratch buffer for reading a single byte.
     */
    private byte[] singleByteBuf = new byte[1];

    /**
     * Expected authentication response from client.
     */
    private byte[] authExpectedResponse = null;

    /**
     * Logging channel
     */
    protected LogChannel logChannel = null;

    /**
     * Lingering socket timeout.
     */
    private static final int LINGER_TIMEOUT = 30 * 1000; // Milliseconds.

    /**
     * Lingering socket max read total.
     */
    private static final int LINGER_MAXBYTES = 10000;

    /**
     * Size of the trash buffer.
     */
    private static final int TRASHBUFSIZE = 1000;

    /**
     * Trash buffer for discarding bytes.
     * We don't synchronize this because we don't care if its contents
     * get garbled.  This is a write-only discard buffer.
     */
    private static byte[] trashBuf = new byte[TRASHBUFSIZE];

    /**
     * debug flag
     */
    private static final boolean debug = false;

    /**
     * Create a new EnhydraDirector connection object.
     * <P>
     * Sets the initial state to CLOSED.  Before this instance may be
     * used, it must be assigned a connection using the <code>connect()</code>
     * method.
     */
    public
    EnhydraDirectorProtocol(LogChannel chan)
    {
	logChannel = chan;
	state = STATE_CLOSED;
    }

    /**
     * Internal dispatch table for server side function handlers.
     * Supported functions are permanently registered at compile time
     * by placing them in this method's switch block.
     * 
     * @param type Protocol-defined function type number.
     * @return Reference to the appropriate function handler instance.
     */
    protected abstract
    EnhydraDirectorFunction getFunction(int type);

    /**
     * Implementation class must provide this method.  It should check
     * the configuration data to determine if authentication is required
     * for connections.
     * 
     * @return <code>true</code> if authentication is required.
     */
    protected abstract
    boolean authRequired();

    /**
     * Abstract method to get authentication secret data.
     * <P>
     * This method must be provided by the implementation class.
     * It checks the configuration and finds the secret authentication
     * data used to verify authentication responses from clients.
     * 
     * @return A byte array of authentication data, or <code>null</code>
     * 		if authentication data does not exist.
     */
    protected abstract
    byte[] getAuthData();

    /**
     * Abstract method to process a CGI-BIN header.
     * <P>
     * 
     * @return A byte array of authentication data, or <code>null</code>
     * 		if authentication data does not exist.
     */
    protected abstract
    void processCgiHeader(byte[] name, byte[] value);

    /**
     * Abstract method to process an HTTP request header.
     * <P>
     * 
     * @return A byte array of authentication data, or <code>null</code>
     * 		if authentication data does not exist.
     */
    protected abstract
    void processHttpHeader(byte[] name, byte[] value);

    /**
     * Abstract method to process a cookie.
     * <P>
     * 
     * @return A byte array of authentication data, or <code>null</code>
     * 		if authentication data does not exist.
     */
    protected abstract
    void processCookie(byte[] name, byte[] value);

    /**
     * This method causes a new <code>EnhydraDirectorProtocol</code> instance
     * to accept the next socket connection from a server socket. 
     * This method may only be called if <code>isConnected()</code>
     * returns <code>false</code>.  Otherwise an exception is thrown.
     * 
     * @param socket Server socket from which to obtain a new connection.
     * 
     * @exception IOException
     * 		If a network failure occurs.
     * @exception EnhydraDirectorBadStateException
     * 		If this method is called while the state is not CLOSED.
     */
    public
    void connect(Socket socket)
    throws EnhydraDirectorBadStateException,
	   IOException
    {
	switch (state) {
	    case STATE_CLOSED:
		sock = socket;
		break;
	    default:
		throw new EnhydraDirectorBadStateException(
			"connect: expected state CLOSED " +
			"(was " + getStateName() + ")");
	}
	stateMachine();
    }

    /**
     * Returns <code>false</code> if this connection is in the CLOSED
     * state, otherwise returns <code>true</code>.
     * <P>
     * After an exception is raised by this connection, this method
     * should be checked to verify the state of the connection.
     * 
     * @return True if not CLOSED, otherwise false.
     */
    public
    boolean isConnected()
    {
	switch (state) {
	    case STATE_CLOSED:
		return false;
	    default:
		break;
	}
	return true;
    }

    /**
     * Causes the authentication challenge to be sent to the client.
     * <P>
     * Authentication data is retrieved in an implementation
     * specific manner by the abstract <code>getAuthData</code> method.
     * <P>
     * This method must only be called after a <code>connect()</code>
     * call and before any other calls.  However, authentication can
     * be delayed until a thread becomes available to process the
     * authentication.  Once authentication returns, this instance
     * may be placed back in inactive status until it is re-activated
     * to handle an incoming request.  Upon authentication failure,
     * this instance reverts to the CLOSED state, and the connection
     * socket is closed.  In such a case, this object must be deleted
     * and a new one created in order to create a new connection.
     * <P>
     * Only call this method if <code>isAuthenticated()</code> returns
     * <code>false</code>.  Otherwise an exception will be thrown.
     * 
     * @exception IOException
     * 		If a network failure occurs.
     * @exception EnhydraDirectorBadStateException
     * 		If this method is called while in a state other than
     * 		NEWCONNECT.
     */
    public 
    void authenticate()
    throws EnhydraDirectorBadStateException, IOException
    {
	/*
	 * Check state sanity.
	 */
	switch (state) {
	    case STATE_NEWCONNECT:
		break;
	    default:
		throw new EnhydraDirectorBadStateException(
			"connect: expected state " +
			"NEWCONNECT (was"+ getStateName() +")");
	}
	openCBRequested = false;
	headerCBRequested = false;
	stateMachine();
    }

    /**
     * Second version of <code>authenticate</code> that allows a request
     * for post-open and post-header callbacks. 
     * <P>
     * If either of these
     * options are set, performance will be degraded because the client
     * must go into a wait state until it receives all of the callbacks
     * and the final callback termination packet.  Use these options
     * sparingly.
     * <P>
     * Only call this method if <code>isAuthenticated()</code> returns
     * <code>false</code>.  Otherwise an exception will be thrown.
     * <P>
     * <B>Note:</B>  If either callback option is set to <code>true</code>
     * but callbacks are not made, the state machine automatically detects
     * this and sends callback termination to prevent a mutual-receive 
     * deadlock condition from arising.
     * 
     * @param openCB	If <code>true</code> callbacks may be made to the
     * 			client immediately prior to finishing the connection
     * 		 	establishment.
     * @param headerCB	If <code>true</code> callbacks may be made to the
     * 			client immediately after all HTTP request headers
     * 			have been received.
     * @exception IOException
     * 		If a network failure occurs.
     * @exception EnhydraDirectorBadStateException
     * 		If this method is called while in a state other than
     * 		NEWCONNECT.
     * @see #authenticate()
     * @see #activate()
     */
    public 
    void authenticate(boolean openCB, boolean headerCB)
    throws EnhydraDirectorBadStateException, IOException
    {
	/*
	 * Check state sanity.
	 */
	switch (state) {
	    case STATE_NEWCONNECT:
		break;
	    default:
		throw new EnhydraDirectorBadStateException(
			"connect: expected state " +
			"NEWCONNECT (was"+getStateName()+")");
	}
	openCBRequested = openCB;
	headerCBRequested = headerCB;
	stateMachine();
    }

    /**
     * Returns <code>true</code> if the client has successfully authenticated
     * itself for this connection.
     * <P>
     * Use this method to determine whether <code>authenticate()</code>
     * must be called.
     * 
     * @return True if the client has successfuly authenticted itself
     */
    public
    boolean isAuthenticated()
    {
	switch (state) {
	    case STATE_CLOSED:
	    case STATE_NEWCONNECT:
	    case STATE_WAIT_AUTH:
	    case STATE_CHECK_AUTH:
		return false;
	    default:
		break;
	}
	return true;
    }

    /**
     * Complete connection establishment after authentication.
     * <P>
     * After <code>authenticate()</code> returns, the connection is not
     * fully open and available for requests until <code>activate</code>.
     * has been called.  This allows callbacks to be sent prior to 
     * full connection establishment.  These callbacks happen only once
     * in the lifetime of a connection, and not on a per-request basis, 
     * as would be the case with post-header or post-entity callbacks.
     * <P>
     * If post-open callbacks have not been requested, this method goes
     * straight to the OPEN state.  If callbacks were requested, the
     * callback termination packet is sent to tell the client that
     * there are no more callbacks.
     */
    public
    void activate()
    throws IOException
    {
	/*
	 * Check state sanity.
	 */
	switch (state) {
	    case STATE_OPENING:
		break;
	    case STATE_OPEN:
		break;
	    default:
		throw new EnhydraDirectorBadStateException(
			"connect: expected state " + 
			"OPENING (was"+getStateName()+")");
	}
	active = true;
	stateMachine();
    }

    /**
     * Cause this connection to go into an inactive state.
     * <P>
     * All buffers are freed and the state is reset to OPEN.
     * This method is called by a higher level connection/thread
     * manager to disassociate this instance with any thread.  Usually
     * this means the instance is being placed on an inactive list,
     * awaiting future activation by a running client load manager.
     * 
     * @exception EnhydraDirectorBadStateException
     * 		If this instance is currently processing.  Only instances
     * 		in the OPEN state may be activated or deactivated.
     */
    public
    void deactivate()
    throws IOException
    {
	if (state > STATE_OPEN) {
	    throw new EnhydraDirectorBadStateException(
		    "Connection busy (state=" + getStateName() + ")");
	}
	active = false;
	if (state == STATE_OPEN) {
	    // Make sure all resources have been freed.
	    packetData = null;
	    headerName = null;
	    headerValue = null;
	}
    }

    /**
     * Returns <true> if the connection is fully established.
     * 
     * @return True if the connection is authenticated and established.
     */
    public
    boolean isOpen()
    {
	switch (state) {
	    case STATE_CLOSED:
	    case STATE_NEWCONNECT:
	    case STATE_WAIT_AUTH:
	    case STATE_CHECK_AUTH:
	    case STATE_OPENING:
	    case STATE_WAIT_OPEN_CALLBACK:
		return false;
	    default:
		break;
	}
	return true;
    }

    /**
     * Send a callback to the client.
     * <P>
     * This method may only be send in the following situations:<BR>
     *   1.  If post-open callbacks are requested, just before 
     *       <code>activate()</code> is called.
     *   2.  If post-header callbacks are requested, just before 
     *       calls to <code>read()</code> occur.
     *   3.  Once the request entity has been fully read.
     *   
     * @param number The ID number (0..255) that identifies the callback
     * 		to run on the client.
     * @param argsIn Arguments to pass to the remote callback.  Format
     * 		depends on the function being called.
     * @param argsOut Array of byte arrays of length 1.  Element zero
     * 		gets filled in with the return data byte array upon
     * 		return from the callback.  If no data bytes were returned
     * 		by the callback, then element zero will be set to 
     * 		<code>null</code>.
     * @return True if the callback returned successful status, false
     * 		if the callback failed.
     */
    public
    boolean callback(int number, byte[] argsIn, byte[][] dataOut)
    throws IllegalArgumentException,
	   EnhydraDirectorBadStateException,
	   IOException
    {
	int newState = state;
	if (dataOut.length != 1) {
	    throw new IllegalArgumentException("dataOut not 1xN array");
	}
	if ((number < 0) || (number > 255)) {
	    throw new IllegalArgumentException("number not in 0..255");
	}
	/*
	 * Check state sanity.
	 */
	switch (state) {
	    case STATE_OPENING:
		if (!openCBRequested) {
		    throw new EnhydraDirectorBadStateException(
			"callback: post-open callbacks not requested");
		}
		newState = STATE_WAIT_OPEN_CALLBACK;
		break;
	    case STATE_HAVE_HEADERS:
		if (!headerCBRequested) {
		    throw new EnhydraDirectorBadStateException(
			"callback: post-header callbacks not requested");
		}
		newState = STATE_WAIT_HEADER_CALLBACK;
		break;
	    case STATE_PROCESSING_REQUEST:
		newState = STATE_WAIT_ENTITY_CALLBACK;
		break;
	    default:
		throw new EnhydraDirectorBadStateException(
			"callback: Bad state for callback: " + getStateName());
	}
	sendPacket(PKT_TYPE_FUNC,PKT_SUBTYPE_FUNC_CALL, (byte) number, argsIn);
	state = newState;
	stateMachine();
	/* At this point, client reply has been received. */
	dataOut[0] = packetData;
	return funcSuccess;
    }

    /**
     * Read at least one and at most <code>len</code> request entity bytes.
     * <P>
     * If no EOF or error condition occurs, at least one byte will 
     * be read into the buffer, up to the length of the buffer.
     * This method returns immediately when bytes become available and
     * may read fewer bytes than the allocated buffer length.
     * 
     * @param inBuf Buffer to read into.
     * @param off Offset within the buffer.
     * @param len Number of bytes to read.  The read will stop before reaching
     * 		the end of the buffer, so fewer than <code>len</code> bytes
     * 		may be read if <code>inBuf.length - off</code> is less than
     * 		<code>len</code>.
     * @return The number of bytes actually read.
     */
    public synchronized
    int	read(byte[] inBuf, int off, int len)
    throws IOException
    {
	/*
	 * Check state sanity.
	 */
	switch (state) {
	    case STATE_HAVE_HEADERS:
	    case STATE_RECEIVING_ENTITY:
		break;
	    case STATE_PROCESSING_REQUEST:
		if (inputClosed) throw new IOException("Closed.");
		// If called in this state, ALWAYS return end-of-data.
		//return 0;
		break;
	    default:
		throw new IOException(
			"EnhydraDirector: Illegal state for read: " + getStateName());
	}
	if (inputClosed) throw new IOException("Closed.");
	int actual;
	int bufBytes = inBuf.length - off;
	if (bufBytes < 1) return 0;
	int bytesToRead;
	if (len <= bufBytes) bytesToRead = len;
	else bytesToRead = bufBytes;
	if (entityRemainder > 0) {
	    /* Just read what's available. */
	    if (bytesToRead > entityRemainder) bytesToRead = entityRemainder;
	    actual = in.read(inBuf, off, bytesToRead);
	    entityRemainder -= actual;
	    return actual;
	}
	/* Ok, current entity fragment packet is done.  Get next. */
	stateMachine();
	if (entityRemainder <= 0) {
	    /* No more entity fragments from client.  we're done. */
	    return 0; // end-of-data.
	}
	/* NOW, read from new packet. */
	if (bytesToRead > entityRemainder) bytesToRead = entityRemainder;
	actual = in.read(inBuf, off, bytesToRead);
	entityRemainder -= actual;
	return actual;
    }

    /**
     * Read at least one and at most <code>inBuf.length</code> bytes.
     * <P>
     * If no EOF or error condition occurs, at least one byte will 
     * be read into the buffer, up to the length of the buffer.
     * This method returns immediately when bytes become available and
     * may read fewer bytes than the allocated buffer length.
     * 
     * @param inBuf Buffer to read into.
     * @return The number of bytes actually read.
     */
    public synchronized
    int	read(byte[] inBuf)
    throws IOException
    {
	/*
	 * Check state sanity.
	 */
	switch (state) {
	    case STATE_HAVE_HEADERS:
	    case STATE_RECEIVING_ENTITY:
		break;
	    case STATE_PROCESSING_REQUEST:
		// If called in this state, ALWAYS return end-of-data.
		if (inputClosed) throw new IOException("Closed.");
		//return 0;
		break;
	    default:
		throw new IOException(
			"EnhydraDirector: Illegal state for read: " + getStateName());
	}
	if (inputClosed) throw new IOException("Closed.");
	return read(inBuf, 0, inBuf.length);
    }

    /**
     * Read a single byte of entity input and return it as an integer from
     * 0 to 255.
     * <P>
     * Returns -1 if the end-of-data is reached.
     * 
     * @return The value of the next input byte from 0 to 255,
     * 		as a signed integer.  Returns -1 on end-of-data.
     * @exception IOException If an I/O error occurs.
     */
    public synchronized
    int	read()
    throws IOException
    {
	/*
	 * Check state sanity.
	 */
	switch (state) {
	    case STATE_HAVE_HEADERS:
	    case STATE_RECEIVING_ENTITY:
		break;
	    case STATE_PROCESSING_REQUEST:
		// If called in this state, ALWAYS return end-of-data.
		if (inputClosed) throw new IOException("Closed.");
		//return 0;
		break;
	    default:
		throw new IOException(
			"EnhydraDirector: Illegal state for read: " + getStateName());
	}
	if (inputClosed) throw new IOException("Closed.");
	int n = read(singleByteBuf, 0, 1);
	if (n < 1) return -1;
	return (int)(singleByteBuf[0]) & 0xff;
    }

    /**
     * Read exactly <code>len</code> request entity bytes.
     * <P>
     * If no EOF or error condition occurs, at least one byte will 
     * be read into the buffer, up to the length of the buffer.
     * This method returns only after the buffer is full, end-of-data
     * is reached, or an error occurs.  This is convenient when an
     * exact number of bytes are expected.
     * <P>
     * The actual number of bytes read is returned.  This may be less
     * than the buffer length if an error occurs, or if end-of-data
     * is reached.
     * 	
     * @param inBuf Buffer to read into.
     * @return The actual number of bytes read.  On end-of-data,
     * 		returns 0. Receiving more than one byte, but fewer
     * 		bytes than the buffer size also indicates an end-of-data
     * 		condition.  Once end-of-data occurs, all further calls
     * 		return zero.
     * @exception IOException
     * 		  If a network failure occurs.
     */
    public synchronized
    int	readExactly(byte[] inBuf, int off, int len)
    throws IOException
    {
	/*
	 * Check state sanity.
	 */
	switch (state) {
	    case STATE_HAVE_HEADERS:
	    case STATE_RECEIVING_ENTITY:
		break;
	    case STATE_PROCESSING_REQUEST:
		// If called in this state, ALWAYS return end-of-data.
		if (inputClosed) throw new IOException("Closed.");
		//return 0;
		break;
	    default:
		throw new IOException(
			"EnhydraDirector: Illegal state for read: " + getStateName());
	}
	if (inputClosed) throw new IOException("Closed.");
	int actual = 0;
	int bufBytes = inBuf.length - off;
	if (bufBytes < 1) return 0;
	int bytesToRead;
	if (len <= bufBytes) bytesToRead = len;
	else bytesToRead = bufBytes;
	while (actual < bytesToRead) {
	    actual += read(inBuf, off + actual, bytesToRead - actual);
	    if (actual <= 0) break;
	}
	return actual;
    }

    /**
     * Read request entity bytes with Java read semantics.
     * <P>
     * If no EOF or error condition occurs, at least one byte will 
     * be read into the buffer, up to the length of the buffer.
     * This method returns only after the buffer is full, end-of-data
     * is reached, or an error occurs.  This is convenient when an
     * exact number of bytes are expected.
     * <P>
     * The actual number of bytes read is returned.  This may be less
     * than the buffer length if an error occurs, or if end-of-data
     * is reached.
     * 	
     * @param inBuf Buffer to read into.
     * @return The actual number of bytes read.  On end-of-data,
     * 		returns 0. Receiving more than one byte, but fewer
     * 		bytes than the buffer size also indicates an end-of-data
     * 		condition.  Once end-of-data occurs, all further calls
     * 		return zero.
     * @exception IOException
     * 		  If a network failure occurs.
     */
    public synchronized
    int	readExactly(byte[] inBuf)
    throws IOException
    {
	/*
	 * Check state sanity.
	 */
	switch (state) {
	    case STATE_HAVE_HEADERS:
	    case STATE_RECEIVING_ENTITY:
		break;
	    case STATE_PROCESSING_REQUEST:
		// If called in this state, ALWAYS return end-of-data.
		if (inputClosed) throw new IOException("Closed.");
		//return 0;
		break;
	    default:
		throw new IOException(
			"EnhydraDirector: Illegal state for read: " + getStateName());
	}
	if (inputClosed) throw new IOException("Closed.");
	int actual = 0;
	while (actual < inBuf.length) {
	    actual += read(inBuf, actual, inBuf.length - actual);
	    if (actual <= 0) break;
	}
	return actual;
    }

    /**
     * Skip over <code>nbytes</code> bytes of entity input data.
     * <P>
     * Receives and discards <code>n</code> bytes of entity data and 
     * returns the number of bytes discarded.  Returns 0 when end-of-data
     * is reached, or if <code>n</code> is 0.
     * 
     * @param n The number of bytes to discard.
     * @return The number of bytes skipped, or 0 when end-of-data reached.
     * @exception IOException If a network failure occurs.
     */
    public synchronized
    long skip(long n)
    throws IOException
    {
	/*
	 * Check state sanity.
	 */
	switch (state) {
	    case STATE_HAVE_HEADERS:
	    case STATE_RECEIVING_ENTITY:
		break;
	    case STATE_PROCESSING_REQUEST:
		// If called in this state, ALWAYS return end-of-data.
		if (inputClosed) throw new IOException("Closed.");
		return 0;
	    default:
		throw new IOException(
			"EnhydraDirector: Illegal state for skip: " + getStateName());
	}
	if (inputClosed) throw new IOException("Closed.");
	if (n <= trashBuf.length) {
	    return read(trashBuf, 0, (int)n);
	}
	long left = n;
	long skipped = 0;
	int act = 0;
	while (left > trashBuf.length) {
	    if ((act = read(trashBuf)) < trashBuf.length) {
		return (skipped + ((act >= 0) ? act : 0));
	    }
	    skipped += act;
	    left -= act;
	}
	act += read(trashBuf, 0, (int)left);
	skipped += ((act >= 0) ? act : 0);
	return skipped;
    }

    /**
     * Discards any further incoming request entity data and marks the
     * input stream as "closed".
     * 
     * @exception IOException if an error occurs while discarding 
     * 			entity data.  This usually will cause the
     * 			protocol to go out of sync.
     */
    public
    void closeInput()
    throws IOException
    {
	/*
	 * Check state sanity.
	 */
	if (inputClosed) return; /* XXX */
	discardEntity();
	inputClosed = true;
    }

    /**
     * Returns a maximum number of bytes which can be read without blocking.
     * <P>
     * No internal byte[] buffer is kept around.  The read() methods
     * read directly from the socket to the passed-in buffer.
     * This means that the number of bytes guaranteed to be available
     * in a non-blocking readExactly() is the minimum of the available
     * bytes on the input socket, and the number of bytes remaining 
     * in the current incoming data packet.  We can't just read the
     * next packet header, because it may contain a callback, and
     * thus require a block.  We don't have a way to make the state
     * machine go backwards in a "putback" fashion.  (which would be ugly)
     * 
     * @return The number of bytes which can be read with a guarantee of
     * 		not blocking on the input stream.
     * @exception IOException If an I/O error occurs.
     */
    public
    int inputAvailable()
    throws IOException
    {
	/*
	 * Check state sanity.
	 */
	switch (state) {
	    case STATE_HAVE_HEADERS:
	    case STATE_RECEIVING_ENTITY:
		break;
	    case STATE_PROCESSING_REQUEST:
		// If called in this state, ALWAYS return end-of-data.
		if (inputClosed) throw new IOException("Closed.");
		return 0;
	    default:
		throw new IOException(
			"EnhydraDirector: Illegal state for available: " + getStateName());
	}
	if (inputClosed) throw new IOException("Closed.");
	int nbytes = in.available();
	/*
	 * No guarantee that the next packet won't block due
	 * to callback, etc.
	 * The spec for available() only requires a number of bytes
	 * GUARANTEED to be readable without blocking.  It does NOT
	 * require ALL bytes readable without blocking. :)
	 */
	if ((nbytes) > entityRemainder) nbytes = entityRemainder;
	return nbytes;
    }

    /**
     * Discard the entire entity body We don't care about it.
     * 
     * @return The number of bytes skipped.
     * @exception IOException If a network failure occurs.
     */
    protected
    long discardEntity()
    throws IOException
    {
	/*
	 * Check state sanity.
	 */
	switch (state) {
	    case STATE_HAVE_HEADERS:
	    case STATE_RECEIVING_ENTITY:
		break;
	    case STATE_PROCESSING_REQUEST:
		// If called in this state, ALWAYS return end-of-data.
		return 0;
	    default:
		throw new IOException(
			"EnhydraDirector: Illegal state for discard: " +
			getStateName());
	}
	long skipped = 0;
	long act;
	while ((act = skip(4096)) > 0) skipped += act;
	if (debug) {
	    logChannel.write(Logger.DEBUG, "DISCARDED " + skipped + " bytes.");
	}
	state = STATE_PROCESSING_REQUEST;
	return skipped;
    }

    /**
     * Method to write a single byte of response data.
     */
    public
    void write(int b)
    throws IOException
    {
	if (debug) {
	    StringBuffer sbuf = new StringBuffer();
	    sbuf.append((char)(b & 0xff));
	    logChannel.write(Logger.DEBUG, "write byte: " + sbuf.toString());
	}
	/*
	 * Check state sanity.
	 */
	switch (state) {
	    case STATE_HAVE_HEADERS:
	    case STATE_RECEIVING_ENTITY:
		// live dangerously and dont: discardEntity();
	    case STATE_PROCESSING_REQUEST:
		break;
	    default:
		throw new IOException(
			"EnhydraDirector: Illegal state for write: " +
			getStateName());
	}
	synchronized (singleByteBuf) {
	    singleByteBuf[0] = (byte) b;
	    sendPacket(PKT_TYPE_RESP, PKT_SUBTYPE_RESP_DATA, singleByteBuf);
	}
    }

    /**
     * Method to write a single byte of response data.
     */
    public
    void write(byte[] buf)
    throws IOException
    {
	if (debug) {
	    String s = new String(buf);
	    //logChannel.write(Logger.DEBUG, "write  buf: " + s);
	    logChannel.write(Logger.DEBUG,
			    "write pbuf: " + s.length() + " bytes");
	}
	/*
	 * Check state sanity.
	 */
	switch (state) {
	    case STATE_HAVE_HEADERS:
	    case STATE_RECEIVING_ENTITY:
		// live dangerously and dont: discardEntity();
	    case STATE_PROCESSING_REQUEST:
		break;
	    default:
		throw new IOException(
			"EnhydraDirector: Illegal state for write: " +
			getStateName());
	}
	sendPacket(PKT_TYPE_RESP, PKT_SUBTYPE_RESP_DATA, buf);
    }

    /**
     * Method to write a single byte of response data.
     */
    public
    void write(byte[] buf, int off, int len)
    throws IOException
    {
	if (debug) {
	    char[] cbuf = new char[len];
	    for (int i=off,j=0; j<len; i++,j++)
		    cbuf[j] = (char)(buf[i] & 0xff);
	    String s = new String(cbuf);
	    logChannel.write(Logger.DEBUG,
		    "write pbuf: " + s.length() + " bytes " + len);
	}
	/*
	 * Check state sanity.
	 */
	switch (state) {
	    case STATE_HAVE_HEADERS:
	    case STATE_RECEIVING_ENTITY:
		// live dangerously and dont: discardEntity();
	    case STATE_PROCESSING_REQUEST:
		break;
	    default:
		throw new IOException(
			"EnhydraDirector: Illegal state for write: " +
			getStateName());
	}
	sendPacket(PKT_TYPE_RESP, PKT_SUBTYPE_RESP_DATA, buf, off, len);
    }

    /**
     * Method to close output stream.
     */
    public
    void closeOutput()
    throws IOException
    {
	if (debug) {
	    logChannel.write(Logger.DEBUG,"closeOutput()");
	}
	if (state == STATE_OPEN) return; // Allow multiple close.
	switch (state) {

	    // NO-OP states.
	    case STATE_CLOSED:
	    case STATE_OPEN:
		break;

	    // The "normal" state in which to do this.
	    case STATE_PROCESSING_REQUEST:
		// This ends the request and resets to STATE_OPEN
		// Sends "last data" header to client.
		stateMachine();
		break;

	    // Try our best here.
	    case STATE_HAVE_HEADERS:
		state = STATE_RECEIVING_ENTITY; // Short-circuit.
	    case STATE_RECEIVING_ENTITY:
		try {
		    // Try to discard entity and end request normally.
		    discardEntity();
		    stateMachine();
		} catch (Exception e) {}
		if ((state != STATE_OPEN) && (state != STATE_CLOSED)) {
		    try {
			sendPacket(PKT_TYPE_PROTO,PKT_SUBTYPE_PROTO_CLOSE,
				    "Connection closed by server.");
		    } catch (Exception e) {}
		    errorShutdown();
		    throw new IOException(this.toString() +
			    " Closed in recv state");
		}
		break;

	    // PROTO CLOSE states (waiting for us to send).
	    // Closing here results in dropped socket connection!
	    case STATE_NEWCONNECT:
	    case STATE_CHECK_AUTH:
	    case STATE_OPENING:
	    case STATE_PROCESSING_FUNC:
	    case STATE_PROCESSING_ENTITY_FUNC:
		sendPacket(PKT_TYPE_PROTO,PKT_SUBTYPE_PROTO_CLOSE,
				"Connection closed by server.");
		errorShutdown();
		throw new IOException(this.toString() +
			" Closed in send state");

	    // Iffy PROTO CLOSE states (we're receiving - a bit dicey here)
	    // Closing here results in dropped socket connection!
	    case STATE_WAIT_AUTH:
	    case STATE_WAIT_OPEN_CALLBACK:
	    case STATE_PROCESSING_HEADER:
	    case STATE_RECEIVING_HEADERS:
	    case STATE_WAIT_HEADER_CALLBACK:
	    case STATE_WAIT_ENTITY_CALLBACK:
	    default:
		try {
		    sendPacket(PKT_TYPE_PROTO,PKT_SUBTYPE_PROTO_CLOSE,
				"Connection closed by server.");
		} catch (Exception e) {}
		errorShutdown();
		throw new IOException(this.toString() +
			" Closed in recv state");
	}
    }

    /**
     * Method to close output stream.
     */
    public
    void closeConnection()
    throws IOException
    {
	if (debug) {
	    logChannel.write(Logger.DEBUG,"closeConnection()");
	}

	// XXX Need to think this out more.
	if (sock != null) {
	    if (out != null) {
		try { out.flush(); } catch (IOException e) {}
	    }
	    try { lingeringClose(sock); } catch (IOException e) {}
	    sock = null;
	}
	state = STATE_CLOSED;
	packetData = null;
	headerName = null;
	headerValue = null;
	authExpectedResponse = null;
	
    }

    /**
     * Method to send an HTTP header.
     */
    public
    void sendHeader(String name, String value)
    throws IOException
    {
	if (debug) {
	    logChannel.write(Logger.DEBUG,
			    "sendHeader(\""+name+"\", \""+value+"\")");
	}
	switch (state) {
	    case STATE_HAVE_HEADERS:
	    case STATE_RECEIVING_ENTITY:
		// live dangerously and don't: discardEntity();
	    case STATE_PROCESSING_REQUEST:
		break;
	    default:
		throw new IOException(
			"EnhydraDirector: Illegal state for sendHeader: " +
			getStateName());
	}
	sendHeaderPacket(PKT_SUBTYPE_HEADER_HEADER, name, value);
    }

    /**
     * Method to send an HTTP cookie.
     */
    public
    void sendCookie(String name, String value)
    throws IOException
    {
	if (debug) {
	    logChannel.write(Logger.DEBUG,
			"sendCookie(\""+name+"\", \""+value+"\")");
	}
	switch (state) {
	    case STATE_HAVE_HEADERS:
	    case STATE_RECEIVING_ENTITY:
		// live dangerously and don't: discardEntity();
	    case STATE_PROCESSING_REQUEST:
		break;
	    default:
		throw new IOException(
			"EnhydraDirector: Illegal state for sendCookie: " +
			getStateName());
	}
	sendHeaderPacket(PKT_SUBTYPE_HEADER_COOKIE, name, value);
    }

    /**
     * Method to send an HTTP cookie.
     */
    public
    void finishHeaders()
    throws IOException
    {
	if (debug) {
	    logChannel.write(Logger.DEBUG,"finishHeaders()");
	}
	switch (state) {
	    case STATE_HAVE_HEADERS:
	    case STATE_RECEIVING_ENTITY:
		// live dangerously and don't: discardEntity();
	    case STATE_PROCESSING_REQUEST:
		break;
	    default:
		throw new IOException(
			"EnhydraDirector: Illegal state for sendHeader: " +
			getStateName());
	}
	sendPacket(PKT_TYPE_HEADER, PKT_SUBTYPE_HEADER_LAST);
    }
    
    /**
     * The main loop of the internal state machine.
     * <P>
     * Continuously loops, processing events and running state
     * logic until one of its states transitions back to the
     * calling public method because of an error or to allow
     * the Connection Method logic to do processing at a safe point.
     * 
     * @exception IOException
     * 		  If a network failure occurs while processing.  If this
     * 		  exception is caught, the connection will have been 
     * 		  closed due to the error.
     */
    private
    void stateMachine()
    throws EnhydraDirectorBadStateException, IOException
    {
	try {
	    while (true) {
		if (debug) {
		    logChannel.write(Logger.DEBUG, "IN:  " + this.toString());
		}
		switch (state) {
		    case STATE_CLOSED:
			if (!state_closed()) return;
			break;
		    case STATE_NEWCONNECT:
			if (!state_newconnect()) return;
			break;
		    case STATE_WAIT_VERSION:
			if (!state_wait_version()) return;
			break;
		    case STATE_WAIT_AUTH:
			if (!state_wait_auth()) return;
			break;
		    case STATE_CHECK_AUTH:
			if (!state_check_auth()) return;
			break;
		    case STATE_OPENING:
			if (!state_opening()) return;
			break;
		    case STATE_WAIT_OPEN_CALLBACK:
			if (!state_wait_open_callback()) return;
			break;
		    case STATE_OPEN:
			if (!state_open()) return;
			break;
		    case STATE_PROCESSING_FUNC:
			if (!state_processing_func()) return;
			break;
		    case STATE_PROCESSING_HEADER:
			if (!state_processing_header()) return;
			break;
		    case STATE_RECEIVING_HEADERS:
			if (!state_receiving_headers()) return;
			break;
		    case STATE_HAVE_HEADERS:
			if (!state_have_headers()) return;
			break;
		    case STATE_WAIT_HEADER_CALLBACK:
			if (!state_wait_header_callback()) return;
			break;
		    case STATE_RECEIVING_ENTITY:
			if (!state_receiving_entity()) return;
			break;
		    case STATE_PROCESSING_ENTITY_FUNC:
			if (!state_processing_entity_func()) return;
			break;
		    case STATE_PROCESSING_REQUEST:
			if (!state_processing_request()) return;
			break;
		    case STATE_WAIT_ENTITY_CALLBACK:
			if (!state_wait_entity_callback()) return;
			break;
		    default:
			throw new EnhydraDirectorBadStateException(
			    "Illegal state: " + state);
		}
		if (debug) {
		    logChannel.write(Logger.DEBUG, "OUT: " + this.toString());
		}
	    }
	} catch (IOException ioe) {
	    // Close the connection, try to send PROTO FATAL error.
	    try {
		sendPacket(PKT_TYPE_PROTO, PKT_SUBTYPE_PROTO_FATAL,
		       "I/O Error: " + ioe);
	    } catch (IOException ioe2) {}
	    try { if (sock != null) lingeringClose(sock); } catch (IOException ioe3) {}
	    sock = null;
	    state = STATE_CLOSED;
	    deactivate();
	    throw ioe; // Rethrow.
	} finally {
	    if (debug) {
		logChannel.write(Logger.DEBUG, "OUT: " + this.toString());
	    }
	}
    }

    /**
     * Logic for state <code>CLOSED</code>.
     * <P>
     * This is the state of a newly created EnhydraDirector connection,
     * or of one that has been closed.  To receive another connection,
     * the <code>connect</code> method should be called with a newly
     * <code>accept()-ed</code> socket.
     *
     * <P>Blocks for events: No
     * <BR>Exits state machine on success: Yes
     * <P>State Transitions: NEWCONNECT
     * 
     * @return True to remain in state machine, false to exit state machine.
     * @exception IOException If a network or protocol failure occurs.
     */
    private
    boolean state_closed()
    throws IOException
    {
	if (sock == null) {
	    throw new IOException("Missing socket.");
	}
	in = sock.getInputStream();
	out = sock.getOutputStream();
	state = STATE_NEWCONNECT;
	return false; // Break out of state machine.
    }

    /**
     * Logic for state <code>NEWCONNECT</code>.
     * <P>
     * Protocol has been revised slightly to send a PROTO,VERSION 
     * packet as the first packet either side receives.  We
     * Send off a VERSION packet, and if auth is required, we
     * send an auth challenge packet.  Then we go to the WAIT_VERSION
     * state and wait for the version packet to come back.  At
     * that point, if auth is required, we go to WAIT_AUTH, else we
     * go to OPENING.
     * <P>
     * This method is entered via a call to the <code>authenticate()</code>
     * method.
     *
     * <P>Blocks for events: No
     * <BR>Exits state machine on success: No
     * <P>State Transitions: OPENING, WAIT_AUTH
     * 
     * @return True to remain in state machine, false to exit state machine.
     * @exception IOException
     * 		If a network or protocol failure occurs.
     * @exception EnhydraDirectorException
     * 		If an internal error or misconfiguration is detected.
     */
    private
    boolean state_newconnect()
    throws IOException, EnhydraDirectorException
    {
	/*
	 * Always send a [PROTO,VERSION] packet with our major,minor
	 * version to give the client a chance to select its protocol.
	 */
	sendVersionPacket();

	authExpectedResponse = null;
	clientMajor = clientMinor = 0;


	/*
	 * If authentication is required, we send an [AUTH,SVR] packet
	 * to request digest authentication.  This immediately follows
	 * the VERSION packet and does not wait for any response 
	 * from the client.
	 */
	if (authRequired()) {
	    MessageDigest md5 = null;
	    try {
	        md5 = MessageDigest.getInstance("MD5");
	    } catch (NoSuchAlgorithmException nsae) {
		/* No MD5 algorithm.  Fail the connection. */
		sendPacket(PKT_TYPE_AUTH, PKT_SUBTYPE_AUTH_FAIL,
			   "server: cannot generate auth challenge");
		try { lingeringClose(sock); } catch (IOException ioe) {}
		sock = null;
		state = STATE_CLOSED;
		throw new EnhydraDirectorException("No MD5 digester!");
	    }
	    byte[] authDataBytes = null;
	    authDataBytes = getAuthData();
	    if (authDataBytes != null) {
		/* Supplement seed with auth data itself. */
		if (!randomizerSeeded) {
		    randomizer.setSeed(authDataBytes);
		    randomizerSeeded = true;
		}
		byte[] authChallengeBytes = new byte[32];
		randomizer.nextBytes(authChallengeBytes);
		sendPacket(PKT_TYPE_AUTH, PKT_SUBTYPE_AUTH_SVR,
			   authChallengeBytes);
		md5.update(authChallengeBytes);
		md5.update(authDataBytes);
		authExpectedResponse = md5.digest();
		state = STATE_WAIT_VERSION;
	    } else {
		/* No auth data.  Fail the connection. */
		sendPacket(PKT_TYPE_AUTH, PKT_SUBTYPE_AUTH_FAIL,
			   "server: cannot generate auth challenge");
		try { lingeringClose(sock); } catch (IOException ioe) {}
		sock = null;
		state = STATE_CLOSED;
		throw new EnhydraDirectorException("Auth data not found");
	    }
	} else {
	    sendPacket(PKT_TYPE_AUTH, PKT_SUBTYPE_AUTH_OK);
	    state = STATE_WAIT_VERSION;
	}
	return true; // Continue in state machine.
    }

    /**
     * Logic for state <code>WAIT_VERSION</code>.
     * <P>
     * In this state, we are waiting for the client to send back 
     * a [PROTO,VERSION] packet indicating its supported protocol
     * version.  After this we go to OPENING or WAIT_AUTH, depending
     * upon whether we're expecting an authentication response from
     * the client.
     *
     * <P>Blocks for events: Yes (Receipt of [PROTO,VERSION])
     * <BR>Exits state machine on success: No
     * <P>State Transitions: WAIT_AUTH, OPENING
     * 
     * @return True to remain in state machine, false to exit state machine.
     * @exception IOException
     *		  If a network failure occurs.
     */
    private
    boolean state_wait_version()
    throws  IOException
    {
	receiveSinglePacket();
	switch (packetType) {
	    case PKT_TYPE_PROTO:
		switch (packetSubType) {
		    case PKT_SUBTYPE_PROTO_VERSION:
			clientMajor = (packetLength >> 12) & 0xfff;
			clientMinor = packetLength & 0xfff;
			if (clientMajor != MAJOR_VERSION) {
			    /* Complain to client */
			    try {
				sendPacket(	
				    PKT_TYPE_PROTO,PKT_SUBTYPE_PROTO_FATAL,
				    "Incompatible protocol version " +
				    clientMajor + "." +  clientMinor +
				    ": I handle version " + MAJOR_VERSION +
				    "." + MINOR_VERSION);
			    } catch (Exception e) {}
			    errorShutdown();
			    throw new EnhydraDirectorException(
				    "Incompatible protocol version " +
				    clientMajor + "." +  clientMinor +
				    ": I handle version " + MAJOR_VERSION +
				    "." + MINOR_VERSION);
			}
			if (authExpectedResponse != null) {
			    state = STATE_WAIT_AUTH;
			} else {
			    state = STATE_OPENING;
			    return false; /* handler must call activate() */
			}
			break;
		    default:
			throw protoERROR();
		}
		break;
	    default:
		throw protoBADSYNC();
	}
	return true; // Continue in state machine.
    }

    /**
     * Logic for state <code>WAIT_AUTH</code>.
     * <P>
     * This is the state of a connection that is waiting for an
     * authentication response to a previously sent challenge.
     * <P>
     * Upon receipt of an auth response packet, this method simply
     * changes state to CHECK_AUTH, and returns <code>true</code> to
     * continue processing by the state machine.
     *
     * <P>Blocks for events: Yes (Receipt of AUTH REPLY)
     * <BR>Exits state machine on success: No
     * <P>State Transitions: CHECK_AUTH, CLOSED
     * 
     * @return True to remain in state machine, false to exit state machine.
     * @exception IOException
     *		  If a network failure occurs.
     */
    private
    boolean state_wait_auth()
    throws  IOException
    {
	receiveSinglePacket();
	switch (packetType) {
	    case PKT_TYPE_AUTH:
		switch (packetSubType) {
		    case PKT_SUBTYPE_AUTH_CLI:
			if (packetLength > MAX_AUTHDATA_LENGTH) {
			    discardPacketData();
			    sendPacket(PKT_TYPE_AUTH, PKT_SUBTYPE_AUTH_FAIL);
			    state = STATE_CLOSED;
			    try { lingeringClose(sock); } catch (IOException e) {}
			    sock = null;
			    return false;
			}
			receivePacketData();
			state = STATE_CHECK_AUTH;
			break;
		    default:
			throw protoBADSYNC(); // Throws ProtoException
		}
		break;
	    case PKT_TYPE_PROTO:
		throw protoERROR();
	    default:
		throw protoBADSYNC();
	}
	return true; // Continue in state machine.
    }

    /**
     * Logic for state <code>CHECK_AUTH</code>.
     * <P>
     * This is the state in which the authentication data sent by 
     * the client is verified against local authentication data.
     * <P>
     * If authentication succeeds, send "AUTH OK" to client and go
     * to state OPENING.  If authentication fails, send "AUTH FAIL",
     * close the socket, and go to state CLOSED.
     * <P>
     * <code>packetData</code> contains auth reply packet payload.
     * 
     * <P>Blocks for events: No
     * <BR>Exits state machine on success: Yes
     * <P>State Transitions: OPENING, CLOSED
     * 
     * @return True to remain in state machine, false to exit state machine.
     * @exception IOException If a network failure occurs.
     */
    private
    boolean state_check_auth()
    throws IOException
    {
	boolean match = true;
	if (packetData.length == authExpectedResponse.length) {
	    for (int i=0; i<packetData.length; i++) {
		if (packetData[i] != authExpectedResponse[i]) {
		    match = false;
		    break;
		}
	    }
	} else {
	    match = false;
	}
	authExpectedResponse = null; // No longer needed.
	if (match) {
	    sendPacket(PKT_TYPE_AUTH, PKT_SUBTYPE_AUTH_OK);
	    state = STATE_OPENING;
	} else {
	    sendPacket(PKT_TYPE_AUTH, PKT_SUBTYPE_AUTH_FAIL);
	    errorShutdown();
	    state = STATE_CLOSED;
	}
	return false;
    }

    /**
     * Logic for state <code>OPENING</code>.
     * <P>
     * This state handler is reached _after_ the Connection Method
     * code has finished processing any callbacks.  The only thing
     * to do here is send a "Last callback" if open callbacks were
     * requested in the AUTH_OK header.
     * 
     * <P>Blocks for events: No
     * <BR>Exits state machine on success: No
     * <P>State Transitions: OPEN
     * 
     * @return True to remain in state machine, false to exit state machine.
     * @exception IOException If a network failure occurs.
     */
    private
    boolean state_opening()
    throws IOException
    {
	if (openCBRequested) {
	    /* Send "no more callbacks" */
	    sendPacket(PKT_TYPE_FUNC, PKT_SUBTYPE_FUNC_END);
	}
	state = STATE_OPEN;
	return true; // Continue state machine.
    }

    /**
     * Logic for state <code>WAIT_OPEN_CALLBACK</code>.
     * <P>
     * If the Connection Method code has requested opening callbacks,
     * then the client (the web server module) will block, waiting for
     * callback requests until a PKT_SUBTYPE_FUNC_END packet is sent
     * to signal the end of callback requests.
     * <P>
     * This method is only entered from the openCallback() method.
     * It blocks and waits for the client to reply with a packet
     * of type PKT_SUBTYPE_FUNC_OK or PKT_SUBTYPE_FUNC_FAIL.
     * 
     * <P>Blocks for events: Yes (PKT_SUBTYPE_FUNC_{OK,FAIL})
     * <BR>Exits state machine on success: Yes
     * <P>State Transitions: OPENING
     * 
     * @return True to remain in state machine, false to exit state machine.
     * @exception IOException
     *		  If a network failure occurs.
     */
    private
    boolean state_wait_open_callback()
    throws 	IOException
    {
	receiveSinglePacket();
	switch (packetType) {
	    case PKT_TYPE_FUNC:
		switch (packetSubType) {
		    case PKT_SUBTYPE_FUNC_OK:
			funcSuccess = true;
			break;
		    case PKT_SUBTYPE_FUNC_FAIL:
			funcSuccess = false;
			break;
		    default:
			throw protoBADSYNC();
		}
		break;
	    case PKT_TYPE_PROTO:
		throw protoERROR();
	    default:
		throw protoBADSYNC();
	}
	/* Receive function return data (if any) */
	receivePacketData(MAX_FUNCDATA_LENGTH);
	state = STATE_OPENING;
	return false; // Return to Connection Method.
    }

    /**
     * Logic for state <code>OPEN</code>.
     * <P>
     * This is the main "waiting" state for a fully established connection.
     * When processing of a request is finished, the connection will 
     * normally return to this state, awaiting another request over the
     * (still open) socket.
     * <P>
     * In the open state, the connection can handle client function
     * calls, or it can handle client requests.
     * 
     * <P>Blocks for events: Yes (..FUNC_CALL,..REQUEST_[LAST]HEADER)
     * <BR>Exits state machine on success: No
     * <P>State Transitions: PROCESSING_HEADER, PROCESSING_FUNC
     * 
     * @return True to remain in state machine, false to exit state machine.
     * @exception IOException
     *		  If a network failure occurs.
     */
    private
    boolean state_open()
    throws IOException
    {
	inputClosed = true; /* Don't allow read quite yet. */
	entityReceived = false;
	receiveSinglePacket();
	switch (packetType) {
	    case PKT_TYPE_HEADER:
		switch (packetSubType) {
		    case PKT_SUBTYPE_HEADER_HEADER:
		    case PKT_SUBTYPE_HEADER_HTTP:
		    case PKT_SUBTYPE_HEADER_COOKIE:
		    case PKT_SUBTYPE_HEADER_LAST:
			state = STATE_PROCESSING_HEADER;
			break;
		    default:
			throw protoBADSYNC();
		}
		break;
	    case PKT_TYPE_FUNC:
		switch (packetSubType) {
		    case PKT_SUBTYPE_FUNC_CALL:
			state = STATE_PROCESSING_FUNC;
			break;
		    default:
			throw protoBADSYNC();
		}
		break;
	    case PKT_TYPE_PROTO:
		throw protoERROR();
	    default:
		throw protoBADSYNC();
	}
	return true; // Continue state machine.
    }

    /**
     * Logic for state <code>PROCESSING_FUNC</code>.
     * <P>
     * 
     * <P>Blocks for events: Yes (..FUNC_CALL,..REQUEST_[LAST]HEADER)
     * <BR>Exits state machine on success: No
     * <P>State Transitions: PROCESSING_HEADER, PROCESSING_FUNC
     * 
     * @return True to remain in state machine, false to exit state machine.
     * @exception IOException
     *		  If a network failure occurs.
     */
    private
    boolean state_processing_func()
    throws IOException
    {
	receivePacketData();
	if (packetData.length < 1) {
	    sendPacket(PKT_TYPE_FUNC,PKT_SUBTYPE_FUNC_FAIL, 
			"Missing function type");
	    state = STATE_OPEN;
	    return false; // Continue in OPEN state.
	}
	int type = packetData[0] & 0xff;
	EnhydraDirectorFunction func = getFunction(type);
	if (func == null) {
	    sendPacket(PKT_TYPE_FUNC,PKT_SUBTYPE_FUNC_FAIL, 
			"Function not defined");
	    state = STATE_OPEN;
	    return false; // Continue in OPEN state.
	}
	func.run(this);
	if (funcSuccess) {
	    sendPacket(PKT_TYPE_FUNC, PKT_SUBTYPE_FUNC_OK, funcData);
	} else {
	    sendPacket(PKT_TYPE_FUNC, PKT_SUBTYPE_FUNC_FAIL, funcData);
	}
	state = STATE_OPEN;
	return false; // Continue in OPEN state.
    }

    /**
     * Logic for state <code>PROCESSING_HEADER</code>.
     * <P>
     * Processes a previously received header.  If the header is
     * PKT_SUBTYPE_REQ_LASTHEADER, then we go into the HAVE_HEADERS state,
     * otherwise we continue in the RECEIVING_HEADERS state until
     * all headers have been received or an error occurs.
     * 
     * <P>Blocks for events: No (Reads previous header data.)
     * <BR>Exits state machine on success: Yes - on PKT_SUBTYPE_REQ_LASTHEADER
     * <P>State Transitions: RECEIVING_HEADERS, HAVE_HEADERS
     * 
     * @return True to remain in state machine, false to exit state machine.
     * @exception IOException
     *		  If a network failure occurs.
     */
    private
    boolean state_processing_header()
    throws  IOException
    {
	// Packet length for headers is encoded differently.
	// First 10 bits of 24 bit length is header name length.
	// Last 14 bits is header value length.  This allows
	// 1024 bit header names, and 16K header values.  This
	// is ridiculously large for headers, so should not be
	// exceeded in any normal app.
	int nameLen;
	int valueLen;
	nameLen = (packetLength >> 14) & 0x3ff;
	valueLen = packetLength & 0x3fff;

	// Get name in first byte array.
	packetRemainder = nameLen;
	receivePacketData();
	headerName = packetData;
	// Get value in separate byte array.  This will save
	//   the Connection Method code some trouble separating them.
	packetRemainder = valueLen;
	receivePacketData();
	headerValue = packetData;
	packetData = null;
	switch (packetSubType) {
	    case PKT_SUBTYPE_HEADER_HEADER:
		// Ordinary CGI header.
		processCgiHeader(headerName, headerValue);
		break;
	    case PKT_SUBTYPE_HEADER_HTTP:
		// HTTP_* header
		processHttpHeader(headerName, headerValue);
		break;
	    case PKT_SUBTYPE_HEADER_COOKIE:
		// Cookie header (parsed)
		processCookie(headerName, headerValue);
		break;
	    default:
		// Ordinary CGI header.
		processCgiHeader(headerName, headerValue);
		break;
	}
	headerName = null;
	headerValue = null;
	if (packetSubType == PKT_SUBTYPE_HEADER_LAST) {
	    state = STATE_HAVE_HEADERS;
	    inputClosed = false; /* Can now read input */
	    return false; // Return to Connection Method code.
	}
	state = STATE_RECEIVING_HEADERS;
	return true; // Continue in state RECEIVING_HEADERS.
    }

    /**
     * Logic for state <code>RECEIVING_HEADERS</code>.
     * <P>
     * This method waits for the next PKT_SUBTYPE_REQ_[LAST]HEADER packet
     * to arrive on the socket.  Upon receipt of the next header, we
     * transition to the PROCESSING_HEADER state to decode and store it.
     * The PROCESSING_HEADER state transitions back to this state when
     * done, unless it sees that the packet subtype is
     * PKT_SUBTYPE_REQ_LASTHEADER, in which case it transitions to the
     * HAVE_HEADERS state.
     * 
     * <P>Blocks for events: Yes (PKT_TYPE_HEADER)
     * <BR>Exits state machine on success: Yes - on PKT_SUBTYPE_REQ_LASTHEADER
     * <P>State Transitions: RECEIVING_HEADERS, HAVE_HEADERS
     * 
     * @return True to remain in state machine, false to exit state machine.
     * @exception IOException
     *		  If a network failure occurs.
     */
    private
    boolean state_receiving_headers()
    throws IOException
    {
	receiveSinglePacket();
	switch (packetType) {
	    case PKT_TYPE_HEADER:
		switch (packetSubType) {
		    case PKT_SUBTYPE_HEADER_HEADER:
		    case PKT_SUBTYPE_HEADER_HTTP:
		    case PKT_SUBTYPE_HEADER_COOKIE:
		    case PKT_SUBTYPE_HEADER_LAST:
			state = STATE_PROCESSING_HEADER;
			break;
		    default:
			throw protoBADSYNC();
		}
		break;
	    case PKT_TYPE_PROTO:
		throw protoERROR();
	    default:
		throw protoBADSYNC();
	}
	return true; // Continue in state PROCESSING_HEADER.
    }

    /**
     * Logic for state <code>HAVE_HEADERS</code>.
     * <P>
     * In this state, the Connection Method code can process the request
     * headers as necessary before beginning to receive any request
     * entity data.  Also, callbacks can be made at this point to the
     * client for control purposes, assuming post-header callbacks were
     * requested by the AUTH_OK reply.
     * <P>
     * When the Connection method is finished in the HAVE_HEADERS state,
     * it begins to call readEntityData().  This method simply jumps
     * to the RECEIVING_ENTITY state.
     * 
     * <P>Blocks for events: No
     * <BR>Exits state machine on success: No
     * <P>State Transitions: RECEIVING_ENTITY
     * 
     * @return True to remain in state machine, false to exit state machine.
     */
    private
    boolean state_have_headers()
    throws IOException
    {
	// Free now-unused header value strings.
	state = STATE_RECEIVING_ENTITY;
	return true; // Continue in state RECEIVING_ENTITY.
    }

    /**
     * Logic for state <code>WAIT_HEADER_CALLBACK</code>.
     * <P>
     * If the Connection Method code has requested post-header callbacks,
     * then the client (the web server module) will block, waiting for
     * callback requests until a PKT_SUBTYPE_FUNC_END packet is sent
     * to signal the end of callback requests.
     * <P>
     * This method is only entered from the callback() method.
     * It blocks and waits for the client to reply with a packet
     * of type PKT_SUBTYPE_FUNC_OK or PKT_SUBTYPE_FUNC_FAIL.
     * 
     * <P>Blocks for events: Yes (PKT_SUBTYPE_FUNC_{OK,FAIL})
     * <BR>Exits state machine on success: Yes
     * <P>State Transitions: HAVE_HEADERS
     * 
     * @return True to remain in state machine, false to exit state machine.
     * @exception IOException
     *		  If a network failure occurs.
     */
    private
    boolean state_wait_header_callback()
    throws IOException
    {
	receiveSinglePacket();
	switch (packetType) {
	    case PKT_TYPE_FUNC:
		switch (packetSubType) {
		    case PKT_SUBTYPE_FUNC_OK:
			funcSuccess = true;
			break;
		    case PKT_SUBTYPE_FUNC_FAIL:
			funcSuccess = false;
			break;
		    default:
			throw protoBADSYNC();
		}
		break;
	    case PKT_TYPE_PROTO:
		throw protoERROR();
	    default:
		throw protoBADSYNC();
	}
	/* Receive function return data (if any) */
	receivePacketData(MAX_FUNCDATA_LENGTH);
	state = STATE_HAVE_HEADERS;
	return false; // Return to Connection Method.
    }

    /**
     * Logic for state <code>RECEIVING_ENTITY</code>.
     * <P>
     * Now that headers have been received and processed, and any
     * post-header callbacks have been completed, we can begin receiving
     * the request entity, if any.  This state is entered via the
     * <code>read()</code> method and this method is only called if 
     * this connection context's internal read buffer is completely read.
     * 
     * <P>Blocks for events: Yes (PKT_SUBTYPE_REQ_[LAST]DATA)
     * <BR>Exits state machine on success: Yes
     * <P>State Transitions: RECEIVING_ENTITY,PROCESSING_REQUEST
     * 
     * @return True to remain in state machine, false to exit state machine.
     * @exception IOException
     *		  If a network failure occurs.
     */
    private
    boolean state_receiving_entity()
    throws IOException
    {
	if (entityReceived) {
	    /* entity receive buffer empty, and no more incoming.
	     * we're done receiving entity now.
	     */
	    state = STATE_PROCESSING_REQUEST;
	    return false; // Break back to read() method.
	}
	receiveSinglePacket();
	switch (packetType) {
	    case PKT_TYPE_REQ:
		switch (packetSubType) {
		    case PKT_SUBTYPE_REQ_LASTDATA:
			entityReceived = true;
			break;
		    case PKT_SUBTYPE_REQ_DATA:
			break;
		    default:
			throw protoBADSYNC();
		}
		break;
	    case PKT_TYPE_FUNC:
		switch (packetSubType) {
		    case PKT_SUBTYPE_FUNC_CALL:
			state = STATE_PROCESSING_ENTITY_FUNC;
			entityReceived = true;
			break;
		    default:
			throw protoBADSYNC();
		}
		break;
	    case PKT_TYPE_PROTO:
		throw protoERROR();
	    default:
		throw protoBADSYNC();
	}
	/* 
	 * Set up the counters that tell how many bytes remain in
	 * this packet.  A packet can hold 2^24 (16M) bytes!  We 
	 * DON'T want a buffer that big, so we read in chunks.
	 */
	entityRemainder = packetLength;
	state = STATE_RECEIVING_ENTITY;
	/* In case goofy client sends empty entity fragment. */
	if ((entityRemainder <= 0) && (!entityReceived)) {
	    return true; // Loop back around and get next packet.
	}
	return false; // Break back to read() method.
    }

    /**
     * Logic for state <code>PROCESSING_ENTITY_FUNC</code>.
     * <P>
     * If, during receipt of entity data, the client sends a function
     * request, this state is entered, and the function is passed to
     * the correct EnhydraDirector protocol function handler.
     * 
     * <P>Blocks for events: Yes (..FUNC_CALL,..REQUEST_[LAST]HEADER)
     * <BR>Exits state machine on success: No
     * <P>State Transitions: RECEIVING_ENTITY
     * 
     * @return True to remain in state machine, false to exit state machine.
     * @exception IOException
     *		  If a network failure occurs.
     */
    private
    boolean state_processing_entity_func()
    throws IOException
    {
	receivePacketData(MAX_FUNCDATA_LENGTH);
	if (packetData.length < 1) {
	    sendPacket(PKT_TYPE_FUNC,PKT_SUBTYPE_FUNC_FAIL, 
			"Missing function type");
	    state = STATE_RECEIVING_ENTITY;
	    return false; // Continue in OPEN state.
	}
	int type = packetData[0] & 0xff;
	EnhydraDirectorFunction func = getFunction(type);
	if (func == null) {
	    sendPacket(PKT_TYPE_FUNC,PKT_SUBTYPE_FUNC_FAIL, 
			"Function not defined");
	    state = STATE_RECEIVING_ENTITY;
	    return false; // Continue in OPEN state.
	}
	func.run(this);
	if (funcSuccess) {
	    sendPacket(PKT_TYPE_FUNC, PKT_SUBTYPE_FUNC_OK, funcData);
	} else {
	    sendPacket(PKT_TYPE_FUNC, PKT_SUBTYPE_FUNC_FAIL, funcData);
	}
	state = STATE_RECEIVING_ENTITY;
	return false; // Continue in OPEN state.
    }

    /**
     * Logic for state <code>PROCESSING_REQUEST</code>.
     * <P>
     * In this state, we expect no more client-initiated data for
     * this request. (Unless we do a post-entity callback).  When the
     * Connection Method code calls the <code>finish()</code> method,
     * this state method is run.  It simply writes out any last data
     * to the output socket, resets the state to OPEN, and returns.
     * At that point, the Connection manager must release this 
     * connection back to the EnhydraDirector listener.
     * 
     * <P>Blocks for events: No
     * <BR>Exits state machine on success: Yes
     * <P>State Transitions: OPEN  (The circle is complete)
     * 
     * @return True to remain in state machine, false to exit state machine.
     * @exception IOException
     *		  If a network failure occurs.
     */
    private
    boolean state_processing_request()
    throws IOException
    {
	sendPacket(PKT_TYPE_RESP,PKT_SUBTYPE_RESP_END);
	state = STATE_OPEN;
	// Release now-unused resources.
	packetData = null;
	headerName = null;
	headerValue = null;
	return false;
    }

    /**
     * Logic for state <code>WAIT_ENTITY_CALLBACK</code>.
     * <P>
     * Once the request entity has been entirely received, the 
     * Connection Method code may make any number of callbacks,
     * at any time until the connection is finished, since the
     * packet-oriented nature of this protocol allows the client
     * to receive function requests while it's receiving the response.
     * <P>
     * This method is only entered from the callback() method.
     * It blocks and waits for the client to reply with a packet
     * of type PKT_SUBTYPE_FUNC_OK or PKT_SUBTYPE_FUNC_FAIL.
     * 
     * <P>Blocks for events: Yes (PKT_SUBTYPE_FUNC_{OK,FAIL})
     * <BR>Exits state machine on success: Yes
     * <P>State Transitions: PROCESSING_REQUEST
     * 
     * @return True to remain in state machine, false to exit state machine.
     * @exception IOException
     *		  If a network failure occurs.
     */
    private
    boolean state_wait_entity_callback()
    throws IOException
    {
	receiveSinglePacket();
	switch (packetType) {
	    case PKT_TYPE_FUNC:
		switch (packetSubType) {
		    case PKT_SUBTYPE_FUNC_OK:
			funcSuccess = true;
			break;
		    case PKT_SUBTYPE_FUNC_FAIL:
			funcSuccess = false;
			break;
		    default:
			throw protoBADSYNC();
		}
		break;
	    case PKT_TYPE_PROTO:
		throw protoERROR();
	    default:
		throw protoBADSYNC();
	}
	/* Receive function return data (if any) */
	receivePacketData(MAX_FUNCDATA_LENGTH);
	state = STATE_PROCESSING_REQUEST;
	return false; // Return to Connection Method.
    }

    private
    void sendPacket(int type, int subtype, byte[] data)
    throws IOException
    {
	int length = 0;
	if (data != null) length = data.length;
	pktHdrBuf[0] = (byte) ((type << 4) | subtype);
	pktHdrBuf[1] = (byte) (length >> 16);
	pktHdrBuf[2] = (byte) (length >> 8);
	pktHdrBuf[3] = (byte) length;
	out.write(pktHdrBuf);
	if (length > 0) out.write(data);
    }

    private
    void sendPacket(int type, int subtype, byte[] data, int off, int len)
    throws IOException
    {
	int length = 0;
	if (data != null) length = len;
	pktHdrBuf[0] = (byte) ((type << 4) | subtype);
	pktHdrBuf[1] = (byte) (length >> 16);
	pktHdrBuf[2] = (byte) (length >> 8);
	pktHdrBuf[3] = (byte) length;
	out.write(pktHdrBuf);
	if (length > 0) out.write(data, off, len);
    }

    private
    void sendPacket(int type, int subtype, byte funcNum, byte[] data)
    throws IOException
    {
	int length = 1;
	if (data != null) length += data.length;
	pktHdrBuf[0] = (byte) ((type << 4) | subtype);
	pktHdrBuf[1] = (byte) (length >> 16);
	pktHdrBuf[2] = (byte) (length >> 8);
	pktHdrBuf[3] = (byte) length;
	out.write(pktHdrBuf);
	out.write(funcNum);
	if (length > 1) out.write(data);
    }

    private
    void sendPacket(int type, int subtype)
    throws IOException
    {
	pktHdrBuf[0] = (byte) ((type << 4) | subtype);
	pktHdrBuf[1] = pktHdrBuf[2] = pktHdrBuf[3] = 0;
	out.write(pktHdrBuf);
    }

    private
    void sendVersionPacket()
    throws IOException
    {
	long val;
	pktHdrBuf[0] = (byte) 0xff; /* PROTO VERSION is 15,15, which is 0xff */
	val = MAJOR_VERSION << 12;
	val |= (MINOR_VERSION & 0xfff);
	pktHdrBuf[3] = (byte) (val & 0xff); val >>= 8;
	pktHdrBuf[2] = (byte) (val & 0xff); val >>= 8;
	pktHdrBuf[1] = (byte) (val & 0xff);
	out.write(pktHdrBuf);
    }

    private
    void sendPacket(int type, int subtype, String message)
    throws IOException
    {
	int length = 0;
	byte[] data = null;
	if (message != null) {
	    try {
		data = message.getBytes("UTF8");
	    } catch (UnsupportedEncodingException e) {
		data = message.getBytes();
	    }
	}
	if (data != null) length = data.length;
	pktHdrBuf[0] = (byte) ((type << 4) | subtype);
	pktHdrBuf[1] = (byte) (length >> 16);
	pktHdrBuf[2] = (byte) (length >> 8);
	pktHdrBuf[3] = (byte) length;
	out.write(pktHdrBuf);
	if (length > 0) out.write(data);
    }

    private
    void sendHeaderPacket(int subtype, String name, String value)
    throws IOException
    {
	if (value == null) value = "";
	int i;
	int nameLen = name.length();
	int valueLen = value.length();
	int length = ((nameLen & 0x3ff) << 14) | (valueLen & 0x3fff);
	byte[] data = new byte[nameLen + valueLen];
	char[] nameChars = name.toCharArray();
	char[] valChars = value.toCharArray();
	for (i=0; i<nameLen; i++) data[i] = (byte) nameChars[i];
	for (i=0; i<valueLen; i++) data[i + nameLen] = (byte) valChars[i];
	pktHdrBuf[0] = (byte) ((PKT_TYPE_HEADER << 4) | subtype);
	pktHdrBuf[1] = (byte) (length >> 16);
	pktHdrBuf[2] = (byte) (length >> 8);
	pktHdrBuf[3] = (byte) length;
	out.write(pktHdrBuf);
	if (length > 0) out.write(data);
    }

    /**
     * Close the connection and abort.
     * <P>
     * Closes the socket and sets the state to CLOSED.  Drops any
     * referenced data buffers.
     */
    private
    void errorShutdown()
    {
	if (sock != null) {
	    try { lingeringClose(sock); } catch (IOException e) {}
	    sock = null;
	}
	state = STATE_CLOSED;
	packetData = null;
	headerName = null;
	headerValue = null;
	authExpectedResponse = null;
    }

    /**
     * Abort the connection and generate an exception object signalling
     * a severe remote protocol failure or premature closure by the peer.
     * <P>
     * If this method is called it is because the protocol has gotten
     * out of sync and this connection is effectively dead and
     * unrecoverable.  Try to send a FATAL protocol error to the client,
     * then close the connection and return an exception object to
     * be thrown by the caller.
     * 
     * @return An EnhydraDirectorException exception object to be thrown by
     * 		the caller.
     */
    private
    EnhydraDirectorException protoERROR()
    {
	EnhydraDirectorException exc = null;
	switch (packetSubType) {
	    case PKT_SUBTYPE_PROTO_CLOSE:
		exc = new EnhydraDirectorException(
		    "Connection prematurely closed by peer. " +
		    "(state="+ getStateName() + ")");
		break;
	    case PKT_SUBTYPE_PROTO_FATAL:
		exc = new EnhydraDirectorException(
		    "Connection aborted by peer. " +
		    "(state="+ getStateName() + ")");
		break;
	    default:
		exc = new EnhydraDirectorException(
		    "Unknown protocol failure. " +
		    "(state="+ getStateName() + ")");
		break;
	}
	errorShutdown();
	return exc;
    }

    /**
     * Abort the connection and generate an exception object signalling
     * a severe protocol sync failure.
     * <P>
     * If this method is called it is because the protocol has gotten
     * out of sync and this connection is effectively dead and
     * unrecoverable.  Try to send a FATAL protocol error to the client,
     * then close the connection and return an exception object to
     * be thrown by the caller.
     * 
     * @return An EnhydraDirectorException exception object to be thrown by
     * 		the caller.
     */
    private
    EnhydraDirectorException protoBADSYNC()
    {
	EnhydraDirectorException exc = null;
	/* Make a desperate attempt to send some kind of error to peer. */
	try {
	    sendPacket(PKT_TYPE_PROTO,PKT_SUBTYPE_PROTO_FATAL,
			"Connection out of sync.  Aborting.");
	} catch (Exception e) {}
	exc = new EnhydraDirectorException(
		"Protocol out of sync: state=" + getStateName() + " packet=[" +
		packetType + "," + packetSubType + "]");
	errorShutdown();
	return exc;
    }

    /**
     * Read the next packet header from the input stream and decode its
     * fields.
     * 
     * @exception EnhydraDirectorException
     * 		If the header was too short, or if it could not be decoded.
     * @exception IOException
     * 		If an I/O failure occured while reading data.
     */
    private
    void receiveSinglePacket()
    throws EnhydraDirectorException, IOException
    {
	int act, total = 0, len = pktHdrInBuf.length;
	while (total < len) {
	    act = in.read(pktHdrInBuf, total, len - total);
	    if (act <= 0) break;
	    total += act;
	}
	if (total < pktHdrInBuf.length) {
	    /* Send fatal proto error and die. */
	    try {
		sendPacket(PKT_TYPE_PROTO,PKT_SUBTYPE_PROTO_FATAL,
			    "Header too short: " + total + " octets. (EOF)");
	    } catch (Exception e) {}
	    EnhydraDirectorException exc = new EnhydraDirectorException(
		    "Header too short: " + total + " octets, state="+
		    getStateName() + " (EOF)");
	    throw exc;
	}
	packetType = packetSubType = pktHdrInBuf[0] & 0xff;
	packetType >>= 4;
	packetSubType &= 0xf;
	packetLength = (pktHdrInBuf[1] & 0xff) << 16;
	packetLength |= (pktHdrInBuf[2] & 0xff) << 8;
	packetLength |= (pktHdrInBuf[3] & 0xff);
	packetRemainder = packetLength;
    }

    /**
     * Reads and discards all remaining data in the current packet.
     * 
     * @exception IOException
     * 		If an I/O error occurs reading the socket.
     * @exception EnhydraDirectorException
     * 		If unexpected end-of-data is encountered.
     */
    private
    void discardPacketData()
    throws IOException, EnhydraDirectorException
    {
	if (packetRemainder <= 0) return;
	byte[] buf = new byte[4096];
	int act, toRead;
	while (packetRemainder > 0) {
	    if (packetRemainder < 4096) toRead = packetRemainder;
	    else toRead = 4096;
	    act = in.read(buf, 0, toRead);
	    if (act <= 0) break;
	    packetRemainder -= act;
	}
	if (packetRemainder > 0) {
	    /* Send fatal proto error and die. */
	    try {
		sendPacket(PKT_TYPE_PROTO,PKT_SUBTYPE_PROTO_FATAL,
			    "Packet body too short - missing: " + 
			       packetRemainder + " octets.");
	    } catch (Exception e) {}
	    EnhydraDirectorException exc = new EnhydraDirectorException(
			    "Packet body too short - missing: " + 
			       packetRemainder + " octets, state="+
				getStateName());
	    throw exc;
	}
	packetData = null;
    }

    /**
     * Reads up to <code>maxLen</code> octets of the current packet's
     * data body.
     * <P>
     * Any remaining bytes are discarded after <code>maxLen</code>
     * octets have been read.  This method is called in cases where
     * large streaming packet bodies are not expected.
     * <P>
     * After reading, the received and kept octets can be found in
     * <code>packetData</code>
     * 
     * @param maxLen The maximum number of data bytes to keep.  The rest
     * 		are discarded.
     * 
     * @exception IOException
     * 		If an I/O error occurs reading the socket.
     * @exception EnhydraDirectorException
     * 		If unexpected end-of-data is encountered.
     */
    private
    void receivePacketData(int maxLen)
    throws IOException, EnhydraDirectorException
    {
	int len = packetRemainder;
	int act, total = 0;
	if (len > maxLen) len = maxLen;
	packetData = new byte[len];
	while (total < len) {
	    if ((act = in.read(packetData, total, len - total)) <= 0) break;
	    total += act;
	}
	packetRemainder -= total;
	if (total < len) {
	    /* Send fatal proto error and die. */
	    try {
		sendPacket(PKT_TYPE_PROTO,PKT_SUBTYPE_PROTO_FATAL,
			    "Packet body too short - missing: " + 
			       packetRemainder + " octets.");
	    } catch (Exception e) {}
	    EnhydraDirectorException exc = new EnhydraDirectorException(
			    "Packet body too short - missing: " + 
			       packetRemainder + " octets, state="+
				getStateName());
	    throw exc;
	}
	if (packetRemainder > 0) discardPacketData();
    }

    /**
     * Reads up to <code>MAX_SMALL_PACKET_LENGTH</code> octets of the
     * current packet's data body.
     * <P>
     * Any remaining bytes are discarded after the desired number of
     * octets have been read.  This method is called in cases where
     * large streaming packet bodies are not expected.
     * <P>
     * After reading, the received and kept octets can be found in
     * <code>packetData</code>
     * 
     * @exception IOException
     * 		If an I/O error occurs reading the socket.
     * @exception EnhydraDirectorException
     * 		If unexpected end-of-data is encountered.
     */
    private
    void receivePacketData()
    throws IOException, EnhydraDirectorException
    {
	receivePacketData(MAX_SMALL_PACKET_LENGTH);
    }

    /**
     * Returns a text name for the current state of this instance.
     * 
     * @return The text state name.
     */
    protected
    String getStateName()
    {
	return getStateName(state);
    }

    /**
     * Returns a text name for the specified state number.
     * 
     * @param The state number.
     * @return The text state name.
     */
    protected
    String getStateName(int stateNum)
    {
	String name;
	switch (stateNum) {
	    case STATE_CLOSED:
		name = "CLOSED";
		break;
	    case STATE_NEWCONNECT:
		name = "NEWCONNECT";
		break;
	    case STATE_WAIT_AUTH:
		name = "WAIT_AUTH";
		break;
	    case STATE_CHECK_AUTH:
		name = "CHECK_AUTH";
		break;
	    case STATE_OPENING:
		name = "OPENING";
		break;
	    case STATE_WAIT_OPEN_CALLBACK:
		name = "WAIT_OPEN_CALLBACK";
		break;
	    case STATE_OPEN:
		name = "OPEN";
		break;
	    case STATE_PROCESSING_FUNC:
		name = "PROCESSING_FUNC";
		break;
	    case STATE_PROCESSING_HEADER:
		name = "PROCESSING_HEADER";
		break;
	    case STATE_RECEIVING_HEADERS:
		name = "RECEIVING_HEADERS";
		break;
	    case STATE_HAVE_HEADERS:
		name = "HAVE_HEADERS";
		break;
	    case STATE_WAIT_HEADER_CALLBACK:
		name = "WAIT_HEADER_CALLBACK";
		break;
	    case STATE_RECEIVING_ENTITY:
		name = "RECEIVING_ENTITY";
		break;
	    case STATE_WAIT_ENTITY_CALLBACK:
		name = "WAIT_ENTITY_CALLBACK";
		break;
	    case STATE_PROCESSING_ENTITY_FUNC:
		name = "PROCESSING_ENTITY_FUNC";
		break;
	    case STATE_PROCESSING_REQUEST:
		name = "PROCESSING_REQUEST";
		break;
	    default:
		name = "" + state;
		break;
	}
	return name;
    }

    /**
     * Implements a "lingering close" to ensure final data is sent to
     * the client.
     * <P>
     * An annoying "feature" of the TCP protocol is that if a server does a
     * close() on a socket while the client is still trying to write to the
     * server, then there is a good possibility that the "reset" packet will
     * get to the client and be detected by a write operation.  This will
     * cause BOTH the send AND receive buffers to be flushed, including the
     * last data sent by the server.  Grrr.  Apache has this problem as
     * well.  The normal kernel solution "should" be to use SO_LINGER, except
     * that SO_LINGER is nonfunctional or outright broken on many platforms.
     * The solution we do here is to repeatedly read from the socket until:
     * <PRE>
     * 	- End-of-data is detected (read returns zero bytes) OR..
     * 	- We time out (Using SO_TIMEOUT).
     * 	- We read more than a finite number of bytes (to prevent DoS attacks
     * 	  by spamming us with gigabytes of data).
     * </PRE>
     * 
     * @param sock The socket to close.
     * @exception IOException If an I/O error is detected.
     */
    private
    void lingeringClose(Socket sock)
    throws IOException
    {
	long		currentTime = System.currentTimeMillis();
	long		startTime = currentTime;
	long		endTime = currentTime + LINGER_TIMEOUT;
	int		nbytes, bytesLeft = LINGER_MAXBYTES;
	InputStream 	in = sock.getInputStream();
	if (debug) {
	    logChannel.write(Logger.DEBUG, "Lingering close on socket: "
				+ sock.toString());
	}
	try {
	    while ((currentTime < endTime) && (bytesLeft > 0)) {
		sock.setSoTimeout((int)(endTime - currentTime));
		try {
		    nbytes = in.read(trashBuf);
		} catch (java.io.InterruptedIOException ioe) {
		    nbytes = 0;
		}
		if (nbytes < 0) break;
		currentTime = System.currentTimeMillis();
		bytesLeft -= nbytes;
	    }
	} finally {
	    sock.close();
	}
	if (debug) {
	    logChannel.write(Logger.DEBUG, "Socket close on "
			     + sock.toString() + " lingered for "
			     + (currentTime - startTime)
			     + " ms/" + (LINGER_MAXBYTES - bytesLeft)
			     + " bytes");
	}
    } 

    /**
     * Generates a string that displays the current state of this instance.
     * <P>
     * Useful for debugging and logging.
     * 
     * @return String that shows the current state of this instance.
     */
    public
    String toString()
    {
	String myIP=null, hisIP=null;
	int myPort=0, hisPort=0;
	if (sock != null) {
	    myIP = sock.getLocalAddress().getHostAddress();
	    myPort = sock.getLocalPort();
	    hisIP = sock.getInetAddress().getHostAddress();
	    hisPort = sock.getPort();
	}
	return	"[ " + ((myIP==null)  ? "0.0.0.0" : myIP) + ":" + myPort + " , "
		     + ((hisIP==null) ? "0.0.0.0" : hisIP) + ":" + hisPort
		     + " , " + getStateName() + " ]";
    }
}
