/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: StandardConnectionMethod.java,v 1.4.2.1.2.1 2000/10/19 17:59:09 jasona Exp $
 */

package org.enhydra.servlet.connectionMethods;

import java.net.*;
import java.io.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
import org.enhydra.servlet.servletManager.*;
import org.enhydra.servlet.filter.FilterManager;
import org.enhydra.servlet.filter.Filter;

import org.enhydra.servlet.*;
import com.lutris.util.*;
import com.lutris.multiServer.MultiServer;


/**
 *  This is a sample implementation of a ConnectionMethod. It is
 *  an abstract class because this class does not actually recieve or
 *  generate requests. It does do all the channel managment.
 *  Implementations of ConnectionMethod should extend this object,
 *  unless there is a specific reason not to. <P>
 *
 *  Classes that extend this class will need to implement the
 *  <CODE>getChannelURL()</CODE> and
 *  <CODE>channelURLIsValid()</CODE> functions. They will also need to
 *  ensure that they contain a thread that is started in the constructor
 *  or (prefered) <CODE>initialize()</CODE>, and stopped in
 *  <CODE>destroy()</CODE>.
 *
 *  @see org.enhydra.servlet.connectionMethods.ConnectionMethod
 *  @author Andy John
 */
public abstract class StandardConnectionMethod
    implements ConnectionMethod {

    /**
     *  The ServletManager passed in to the <CODE>initialize()</CODE> method.
     *  If you are extending this class, you should use this ServletManager
     *  to convert the servletID string to Servlets. The servletID strings
     *  are only valid in this ServletManager.
     *
     *  @see org.enhydra.servlet.servletManager.ServletManager
     */
    public ServletManager servletManager;

    /**
     *  The FilterManager passed in to the <CODE>initialize()</CODE> method.
     *  If you are extending this class, you should use this FilterManager
     *  to convert the filterID string to Filters.
     *  The filterID strings are only valid in this FilterManager.
     *
     *  @see org.enhydra.servlet.filter.FilterManager
     */
    public FilterManager filterManager;

    /**
     *  Stores a set of ChannelTableEntries. This is where all the information
     *  about the channels is stored. All the methods that modify the
     *  status of channels synchronize on this object. <P>
     */
    protected ChannelTable channelTable;


    /**
     *  Start up. Initially there are no channels.
     *  Saves the ServletManager for later use.<P>
     *
     *  <B>This will be called before any other method in this class.</B>
     *
     *  @param connectionConfig The portion of the config file relevant to
     *  this connection method
     *  @param id The id associated with this connection method
     *  @param servletManager The ServletManager to use to convert servletIDs
     *  into Servlets.
     *  @param filterManager The FilterManager to use to convert filterIDs
     *  into Filters.
     *  @exception ConnectionMethodException If unable to start up. This
     *  class will not throw this exception, however classes which extend
     *  this class may.
     *  @see org.enhydra.servlet.filter.Filter
     */
    public void initialize(Config connectionConfig,
			   String id,
			   ServletManager servletManager,
			   FilterManager filterManager)
        throws ConnectionMethodException {
        this.servletManager = servletManager;
        this.filterManager = filterManager;
        channelTable = new ChannelTable();
    }

    /**
     *  Start up. Initially there are no channels.
     *  Saves the ServletManager for later use.<P>
     *
     *  <B>This will be called before any other method in this class.</B>
     *
     *  @param id The id associated with this connection method
     *  @param servletManager The ServletManager to use to convert servletIDs
     *  into Servlets.
     *  @param filterManager The FilterManager to use to convert filterIDs
     *  into Filters.
     *  @exception ConnectionMethodException If unable to start up. This
     *  class will not throw this exception, however classes which extend
     *  this class may.
     *  @see org.enhydra.servlet.filter.Filter
     */
    public void initialize(String id,
			   ServletManager servletManager,
			   FilterManager filterManager)
        throws ConnectionMethodException {
        this.servletManager = servletManager;
        this.filterManager = filterManager;
        channelTable = new ChannelTable();
    }

    /**
     * Write out self into config fie
     *
     * @param config The config file
     * @param base The base key string
     **/
    public void writeToConfig(Config connectionConfig,
			      String base)
	throws ConfigException, KeywordValueException {
	// Do nothing here
    }


/*
    private void buildContext(String path, String servletID)
                                     throws ConnectionMethodException {

        ServletContainer sc = (ServletContainer)
	    MultiServer.getService("ServletContainer");

        ServletStatus servletStatus = sc.getServletManager().getStatus(servletID);
        String docRoot;
        if (servletStatus.docRoot != null) {
           docRoot = servletStatus.docRoot;
        }
        else {
           docRoot = path;
        }
  	URL url = null;
        try {
            url = URLUtil.resolve(docRoot);
        } catch (MalformedURLException mue) { }



        if (servletStatus.isWar) {
           buildWarContext(path, servletID,url,servletStatus);
        }
        else {
           buildEnhydraContext(path, servletID,url,servletStatus);
        }

        // now add the contextManager to the serveltManager
//        sc.getServletManager().updateContextManager(servletID, getContextManager(),path);
    }


    private void buildEnhydraContext(String path, String servletID,URL url,ServletStatus servletStatus)
                                     throws ConnectionMethodException {

        if (path.length() == 0) {
          path = "/";  /// ??????
        }

        Context context = null;
        try {
          context = getContextManager().addContext(path,url);
        }
        catch (Exception ex) {

          String[] args = new String[] { path };
          // This string needs to be in a resource
          String errorMsg = java.text.MessageFormat.format("The URL {0} is already in use.",args);

          ConnectionMethodException cme =  new ConnectionMethodException(errorMsg);
          throw cme;
        }

        context.setPath(""); // PAJ
        // context.setPath(path);
        context.setDocumentBase(url);

 //       ServletContainer sc = (ServletContainer)
 //	    MultiServer.getService("ServletContainer");

//        ServletStatus servletStatus = sc.getServletManager().getStatus(servletID);

        context.getContainer().addServlet(servletID,  servletStatus.className, null);
        context.getContainer().addMapping(servletID,"/");
        context.getContainer().addMapping(servletID,path);

        // Now we have to set the initParameters of the context
        // copy these out of the property names of the property file  we have
        // now

        Hashtable initParams = new Hashtable();
        Enumeration argNames = servletStatus.initArgs.propertyNames();
        while (argNames.hasMoreElements()) {
           String name = (String) argNames.nextElement();
           String value = servletStatus.initArgs.getProperty(name);
//logChannel.write(Logger.DEBUG, "CONTEXT INIT PARAMS : " + name + "=" + value);
             initParams.put(name,value);
        }
        context.getContainer().setServletInitParams(servletID, initParams) ;

        // a war might have a diffirent base and doc url
        context.getContainer().setServletBase(url);

        // If it's a war we need to set up the classpath
        // and call init(); ?

        // context.setIsWARExpanded(boolean);
        // context.setWorkDir();

    }


    private void buildWarContext(String path, String servletID,URL url,ServletStatus servletStatus)
                                     throws ConnectionMethodException {
        Context context = null;
        try {
          context = getContextManager().addContext(path,url);
        }
        catch (Exception ex) {
         ex.printStackTrace();
        }

    }

*/
    /**
     *  Add a new channel. The new channel is not enabled, and has
     *  no Filters.
     *
     *  @param channelID The symbolic name to use for this channel.
     *  @param URLPrefix If a request's URL starts with this string,
     *      it should be sent to the servlet identified by servletID.
     *      To obtain the servlet, use the ServletManager's <CODE>get()</CODE>
     *      method.
     *  @param servletID The servlet to send requests to. Use the
     *      ServletManager's <CODE>get()</CODE> method to get the servlet.
     *      Call <CODE>get()</CODE> for each request, do not save a reference
     *      to the servlet.
     *  @exception ConnectionMethodException If an error occurs, for example
     *      using an already existing channelID.
     *  @see org.enhydra.servlet.servletManager.ServletManager
     */
    public void addChannel(String channelID, String URLPrefix,
                String servletID) throws ConnectionMethodException {


        ServletContainer sc = (ServletContainer)
	     MultiServer.getService("ServletContainer");

        sc.getServletManager().recordContextPath(servletID,URLPrefix);


        synchronized (channelTable.channelTable) {
            if (channelTable.getChannelByID(channelID) != null)
                throw new ConnectionMethodException("Unable to add channel. " +
                    "The id " + channelID + " is already in use.");
            channelTable.add(channelID, URLPrefix, servletID);
        }
    }

    /**
     *  Get the current status of a channel.
     *  @param channelID Which channel.
     *  @return The status of the channel, stored in a ChannelStatus object.
     *  @see org.enhydra.servlet.connectionMethods.ChannelStatus
     */
    public ChannelStatus getChannelStatus(String channelID) {
        Channel c = channelTable.getChannelByID(channelID);
        if (c != null) {
            ChannelStatus s = new ChannelStatus(c);
            return s;
        } else
            return null;
    }


    /**
     *  Delete a channel.
     *
     *  @param channelID Which channel to remove.
     *  @exception ConnectionMethodException If an error occurs, for example
     *  if the channelID is not found.
     */
    public void deleteChannel(String channelID)
        throws ConnectionMethodException {

        // remove the tomcat context
        try {
          String path = channelTable.getChannelByID(channelID).getURLPrefix();
        }
        catch (Exception e) {
            e.printStackTrace();
        }


        ServletContainer sc = (ServletContainer)
	     MultiServer.getService("ServletContainer");

        Channel channel;
        synchronized (channelTable.channelTable) {
            channel = channelTable.getChannelByID(channelID);
            if (channel == null) {
                throw new ConnectionMethodException("Unable to delete " +
                    "channel. ID " + channelID + " not found.");
            }
            channelTable.delete(channelID);
        }
        sc.getServletManager().removeContextPath(channel.getServletID(),
                    channel.getURLPrefix());


    }

    /**
     *  Returns the identifier names of all the channels currently in
     *  the ConnectionMethod.
     *
     *  @return An array of Strings, the names of all the channels
     *  in the ConnectionMethod.
     */
    public String[] getChannelIDs() {
        return channelTable.getChannelIDs();
    }


    /**
     *  Turn on the channel. Allows requests to be sent to the servlet.
     *
     *  @param channelID Which channel to enable.
     *  @exception ConnectionMethodException If an error occurs, for example
     *  if the channelID is not found.
     */
    public void enableChannel(String channelID)
        throws ConnectionMethodException {
        synchronized (channelTable.channelTable) {
            Channel c = channelTable.findID(channelID);
            if (c == null)
                throw new ConnectionMethodException("Unable to enable " +
                    "channel. ID " + channelID + " not found.");
            c.enable();
        }
    }

    /**
     *  Turn off the channel. Incomming requests should be refused.
     *  No requests should be allowed through to the servlet.
     *
     *  @param channelID Which channel to disable.
     *  @exception ConnectionMethodException If an error occurs, for example
     *  if the channelID is not found.
     */
    public void disableChannel(String channelID)
        throws ConnectionMethodException {
        synchronized (channelTable.channelTable) {
            Channel c = channelTable.findID(channelID);
            if (c == null)
                throw new ConnectionMethodException("Unable to disable " +
                    "channel. ID " + channelID + " not found.");
            c.disable();
        }
    }



    /**
     *  Reset the specified channel's request count to zero.
     *  The counter measures the number of requests processed by this channel.
     *
     *  @param channelID Which channel.
     *  @exception ConnectionMethodException If an error occurs, for example
     *  if the channelID is not found.
     */
    public void resetRequestCount(String channelID)
        throws ConnectionMethodException {
        synchronized (channelTable.channelTable) {
            Channel c = channelTable.findID(channelID);
            if (c == null)
                throw new ConnectionMethodException("Unable to reset " +
                    "reference count. ID " + channelID + " not found.");
            c.resetRequestCount();
        }
    }


    /**
     *  Add a new Filter to the specified channel.
     *
     *  @param channelID Which channel to add a Filter to.
     *  @exception ConnectionMethodException If the channel does not exist, or
     *  the filter name is already in use.
     *  @see org.enhydra.servlet.filter.Filter
     */
    public void addTransactionFilter(String channelID, String filterID)
        throws ConnectionMethodException {
        Filter filter = filterManager.get(filterID);
        if (filter == null)
            throw new ConnectionMethodException("Illegal filter ID " +
                filterID);
        synchronized (channelTable.channelTable) {
            Channel c = channelTable.findID(channelID);
            if (c == null)
                throw new ConnectionMethodException("Unabe to add filter. " +
                    "Channel ID " + channelID + " not found.");
            c.addFilter(filterID);
	    // Interceptors get registered in ServletManager.buildContext()
	    // which is called by ServletManager.start() since the Context
	    // does not exist until then.
	    ServletContainer sc = (ServletContainer)
		MultiServer.getService("ServletContainer");
	    ServletManager sm = sc.getServletManager();
	    String servletID = c.getServletID();
	    // This assumes that the servlet id exists in the channel before
	    // the filter gets added.
	    sm.addFilter(servletID, filterID);
        }
    }


    /**
     *  Remove the given Filter from the specified channel.
     *
     *  @param channelID Which channel to remove a Filter from.
     *  @param filter The Filter to remove.
     *  @exception ConnectionMethodException If the channel or the filter
     *  do not exist.
     *  @see org.enhydra.servlet.filter.Filter
     */
    public void removeTransactionFilter(String channelID, String filterID)
        throws ConnectionMethodException {
        synchronized (channelTable) {
            Channel c = channelTable.findID(channelID);
            if (c == null)
                throw new ConnectionMethodException("Unabe to remove filter." +
                    " Channel ID " + channelID + " not found.");
            c.removeFilter(filterID);
	    ServletContainer sc = (ServletContainer)
		MultiServer.getService("ServletContainer");
	    ServletManager sm = sc.getServletManager();
	    String servletID = c.getServletID();
	    // This assumes that the servlet id exists in the channel before
	    // the filter gets added.
	    sm.removeFilter(servletID, filterID);
        }
     }


    /**
     *  Get a list of all the filter names.
     *
     *  @return An array of filter names.
     *  @exception ConnectionMethodException If the channel ID does not exist.
     */
    public String[] getTransactionFilterIDs(String channelID)
        throws ConnectionMethodException {
        synchronized (channelTable) {
            Channel c = channelTable.findID(channelID);
            if (c == null)
                throw new ConnectionMethodException("Unabe to get filter" +
                    " IDs. Channel ID " + channelID + " not found.");
            return c.getFilterIDs();
        }
    }



    // Interal counter, used by getUniqueChannelName().
    static private int channelNameCounter = 0;

    /**
     *  This returns a channel name that is not currently in use. Use this
     *  when you want to programatically create a channel, and you don't
     *  care what it's name is. The name returned will not be in the list
     *  of current channels, and will not be returned on any other calls
     *  to this function.<BR>
     *  The names will be of the form "channel<N>".
     *  @return A name that can be used with <CODE>addChannel()</CODE>.
     */
    public String getUniqueChannelName() {
        synchronized (channelTable) {
            while (true) {
                String name = "channel_" + channelNameCounter++;
                if (channelTable.findID(name) == null)
                    return name;
            }
        }
    }



    // Internal counter, used by getUniqueFilterName().
    static private int filterNameCounter = 0;

    /*
     *  This returns a filter name that is not currently in use. Use this
     *  when you want to programatically create a filter, and you don't
     *  care what it's name is. The name returned will not be in the list
     *  of current filters, and will not be returned on any other calls
     *  to this function.<BR>
     *  The names will be of the form "filter<N>".
     *  @return A name that can be used with
     *  <CODE>addTransactionFilter()</CODE>.
    public String getUniqueFilterName(String channelID)
        throws ConnectionMethodException {
        synchronized (channelTable) {
            Channel c = channelTable.findID(channelID);
            if (c == null)
                throw new ConnectionMethodException("Unabe to generate " +
                    "unique filter name. Channel ID " +
                    channelID + " not found.");
            while (true) {
                String name = "filter" + filterNameCounter++;
                if (c.getFilter(name) == null)
                    return name;
            }
        }
    }
*/



    /**
     *  Shut down the connection method. This class simply throws out all
     *  the channels. If you extend this class, you will probably want to
     *  perform some actions to close down the external part of the
     *  connection and kill any threads you started.<P>
     *
     *  <B>After this call, none of these methods will be called again.</B>
     */
    public void destroy() throws ConnectionMethodException {
        synchronized (channelTable) {
            channelTable = new ChannelTable();
        }
        this.servletManager = null;
        this.filterManager = null;
    }

    /**
     * Compares the connection specific attributes for equivalency
     *
     *  @param compareObject The object to compare this to
     *  @exception ConnectionMethodException If an error occurs.
     **/
    public boolean equivalent(ConnectionMethod compareObject) {
	// This should never be invoked at this level
	return false;
    }

    public String getType() {
	// This should never be invoked at this level
	return null;
    }
}
