/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HttpRequest.java,v 1.6.2.2.2.1 2000/10/19 17:59:09 jasona Exp $
 */

package org.enhydra.servlet.connectionMethods.http;

import javax.servlet.ServletRequest;
import javax.servlet.ServletInputStream;
import javax.servlet.RequestDispatcher;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpSession;
import java.text.DateFormat;
import java.text.ParseException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Locale;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.File;
import org.enhydra.servlet.ServletError;
import org.enhydra.servlet.tomcat.HeaderConverter;
import org.apache.tomcat.core.RequestAdapter;
import org.apache.tomcat.util.MimeHeaders;

/**
 * Implementation of a HTTP servlet request for the HTTP connection
 * method.
 *
 * @author Mark Diekhans
 * @version $Revision: 1.6.2.2.2.1 $
 * @see javax.servlet.http.HttpServletRequest
 * @see javax.servlet.ServletRequest
 */
public class HttpRequest extends HttpRequestProcessor
    implements HttpServletRequest, RequestAdapter {

    /**
     * Parser for dates.
     */
    private static final DateFormat dateParser = DateFormat.getDateInstance();

    /**
     * Request attributes.
     */
    protected Hashtable attributes = new Hashtable();

    /**
     * Possible file path separators.  This is done because there are
     * two legal file path separators on Win32, but File only recognizes
     * `\'.
     */
    private static final String fileSeparators;

    /**
     * Class initializer.
     */
    static {
        if (File.separatorChar == '\\') {
            fileSeparators = File.separator + "/";
        } else {
            fileSeparators = File.separator;
        }
    }

    /**
     * Merge two file path names, avoiding double path seperators.
     */
    protected static String mergeFilePathNames(String path1, String path2) {
        StringBuffer path = new StringBuffer();
        path1 = trimTrailingSeparators(path1, fileSeparators);
        path2 = trimLeadingSeparators(path2, fileSeparators);
        path2 = trimTrailingSeparators(path2, fileSeparators);
        if (path1 != null) {
            path.append(path1);
            if (path2 != null) {
                path.append(File.separator);
                path.append(path2);
            }
        } else {
            if (path2 != null) {
                path.append(path2);
            } else {
                path.append(File.separator);
            }
        }
        return path.toString();
    }

    /**
     * Convert the a platform-specific file path to a URL file path.
     */
    protected static String filePathToUrlPath(String filePath) {
        StringBuffer urlPath = new StringBuffer(filePath);
        for (int sepIdx = 0; sepIdx < fileSeparators.length(); sepIdx++) {
            char sep = fileSeparators.charAt(sepIdx);
            int idx = 0;
            while ((idx = filePath.indexOf(sep, idx)) >= 0) {
                urlPath.setCharAt(idx, '/');
                idx++;
            }
        }
        return urlPath.toString();
    }

    /**
     * Convert the file portion of a URL path to a platform-specific file
     * path.
     */
    protected static String urlPathToFilePath(String urlPath) {
        StringBuffer filePath = new StringBuffer(urlPath);
        int idx = 0;
        while ((idx = urlPath.indexOf('/', idx)) >= 0) {
            filePath.setCharAt(idx, File.separatorChar);
            idx++;
        }
        return filePath.toString();
    }

    /**
     * Construct a new request object.  This object is designed
     * to be reused for multiple requests, with the
     * <CODE>setConnection</CODE> function being called for each
     * connection.
     *
     * @see HttpRequestProcessor#setConnection
     * @see HttpRequestProcessor#next
     * @see HttpRequestProcessor#setServletInfo
     */
    protected HttpRequest() {
    }

    /**
     * Trim file separators from the start of a path.
     */
    private static String trimLeadingSeparators(String path, String separators) {
        if (path == null) {
            return null;
        }
        int idx = 0;
        while ((idx < path.length()) &&
               (separators.indexOf(path.charAt(idx)) >= 0)) {
            idx++;
        }
        if (idx >= path.length()) {
            return null;
        } else {
            return path.substring(idx);
        }
    }

    /**
     * Trim file separators from the end of a path.
     */
    private static String trimTrailingSeparators(String path, String separators) {
        if (path == null) {
            return null;
        }
        int idx = path.length();
        do {
            idx--;
        } while ((idx >= 0) &&
                 (separators.indexOf(path.charAt(idx)) >= 0));
        if (idx < 0) {
            return null;
        } else {
            return path.substring(0, idx+1);
        }
    }

    /**
     * Get the size of the request data.
     * @returns The size of the request data, or -1 if not known.
     */
    public int getContentLength() {
        return headers.getInt("content-length");
    }

    /**
     * Get the content type of the request.
     *
     * @return The Internet Media Type of the request entity data, or
     * null if not known.
     */
    public String getContentType() {
        return headers.getValue("content-type");
    }

    /**
     * Get the protocol and version of the request.
     *.
     * @return The protocol and version of the request as a string of
     * the form <code>&lt;protocol&gt;/&lt;major version&gt;.&lt;minor
     * version&gt</code>.
     */
    public String getProtocol() {
        return httpVersion;
    }

    /**
     * Get the scheme of the URL used in this request, for example
     * "http", "https", or "ftp".
     *
     * @return The scheme of the URL.
     */
    public String getScheme() {
        return "http"; //FIXME: Support https..
    }

    /**
     * Get the host name of the server that received the request.
     * @return The host name.
     */
    public String getServerName() {
	if (serverName == null) {
	    String host = getHeader("host");
	    if (host != null) {
	        int i = host.indexOf(':');
	        if (i > -1) {
		    host = host.substring(0, i);
	        }
	        serverName = host;
	    } else {
		serverName = "localhost";
	    }
	}
	return serverName;
    }

    /**
     * Get the port number on which this request was received.
     * @returns The port number.
     */
    public int getServerPort() {
        return serverPort;
    }

    /**
     * Get the IP address of the agent that sent the request.
     * @returns The IP address.
     */
    public String getRemoteAddr() {
        return socket.getInetAddress().getHostAddress();
    }

    /**
     * Get the fully qualified host name of the agent that sent the
     * request.
     * @return The host name.
     */
    public String getRemoteHost() {
        return socket.getInetAddress().getHostName();
    }

    /**
     * Applies alias rules to the specified virtual path and returns
     * the corresponding real path, or null if the translation can not
     * be performed for any reason.
     *
     * @param path The virtual path to be translated to a real path.
     * @return The real path.
     */
    public String getRealPath(String path) {
        return servletContext.getRealPath(path);
    }

    /**
     * Get an input stream for reading binary data in the request body.
     *
     * @see getReader
     * @exception IllegalStateException if getReader has been
     *	called on this same request.
     * @exception IOException on other I/O related errors.
     * @returns An input stream.
     */
    public ServletInputStream getInputStream() throws IOException {
        if (reader != null) {
            throw new IllegalStateException("getReader() has been called on this request; can't use getInputStream()");
        }
        getInputStreamCalled = true;
        return inputStream;
    }

    /**
     * Get a string containing the lone value of the specified
     * parameter.
     *
     * @see #getParameterValues
     *
     * @param name The name of the parameter whose value is required.
     * @return The parameter value or null if the parameter does not exist.
     */
    public String getParameter(String name) {
        if (parameters == null) {
            getParsedParameters();
        }
        if (parameters.containsKey(name)) {
            return ((String[])parameters.get(name))[0];
        } else {
            return null;
        }
    }

    /**
     * Get the values of the specified parameter for the request.
     *
     * @param name the name of the parameter whose value is required.
     * @see javax.servlet.ServletRequest#getParameter
     * @return The values of the specified parameter for the request as
     * an array of strings, or null if the named parameter does not
     * exist.
     */
    public String[] getParameterValues(String name) {
        if (parameters == null) {
            getParsedParameters();
        }
        if (parameters.containsKey(name)) {
            return (String[])parameters.get(name);
        } else {
            return null;
        }
    }

    /**
     * Get the parameter names for this request.
     * @return The parameter names for this request as an enumeration.
     */
    public Enumeration getParameterNames() {
        if (parameters == null) {
            getParsedParameters();
        }
        return parameters.keys();
    }

    /**
     * Returns the value of an attribute of the request.
     *
     * @param name The name of the attribute whose value is required.
     * @return The value of the named attribute of the request, or
     * null if the attribute does not exist.
     */
    public Object getAttribute(String name) {
        return attributes.get(name);
    }

    /**
     * Returns a buffered reader for reading text in the request body.
     * This translates character set encodings as appropriate.
     *
     * @see getInputStream
     *
     * @exception UnsupportedEncodingException if the character set encoding
     *  is unsupported, so the text can't be correctly decoded.
     * @exception IllegalStateException if getInputStream has been
     *	called on this same request.
     * @exception IOException on other I/O related errors.
     * @return A buffered reader object.
     */
    public synchronized BufferedReader getReader() throws IOException {
        if (getInputStreamCalled) {
            throw new IllegalStateException("getInputStream() has been called on this request; can't use getReader()");
        }
        if (reader == null) {
            reader = new BufferedReader(new InputStreamReader(inputStream));
        }
        return reader;
    }

    /**
     * Get the character set encoding for the input of this request.
     * @return The character set encoding.
     */
    public String getCharacterEncoding() {
        return null; //FIXME
    }

    /**
     * Gets the array of cookies found in this request.
     *
     * @return The array of cookies found in this request
     */
    public Cookie[] getCookies() {
        return (Cookie[])cookies.clone();
    }

    /**
     * Gets the HTTP method (for example, GET, POST, PUT) with which
     * this request was made.
     *
     * @return the HTTP method with which this request was made
     */
    public String getMethod() {
        return method;
    }

    /**
     * Get the part of this
     * request's URI that is to the left of any query string.
     *
     * @return This request's URI
     * @see javax.servlet.http.HttpUtils#getRequestURL
     */
    public String getRequestURI() {
        return requestUri;
    }

    /**
     * Gets the part of this request's URI that refers to the servlet
     * being invoked.
     *
     * @return The servlet being invoked, as contained in this
     * request's URI.
     */
    public String getServletPath() {
        return servletPath;
    }

    /**
     * Gets any optional extra path information following the servlet
     * path of this request's URI, but immediately preceding its query
     * string.
     *
     * @return The optional path information following the servlet
     * path, but before the query string, in this request's URI; null
     * if this request's URI contains no extra path information.
     */
    public String getPathInfo() {
        if (servletPath.length() >= requestUri.length()) {
            return null;
        }
        return requestUri.substring(servletPath.length());
    }


    /**
     * Gets any optional extra path information following the servlet
     * path of this request's URI, but immediately preceding its query
     * string, and translates it to a real path.
     *
     * @return Extra path information translated to a real path or null
     * if no extra path information is in the request's URI.
     */
    public String getPathTranslated() {
        String pathInfo = getPathInfo();
        if (pathInfo == null) {
            return null;
        }
        return getRealPath(pathInfo);
    }

    /**
     * Gets any query string that is part of the HTTP request URI.
     *
     * @return query string that is part of this request's URI, or null
     * if it contains no query string
     */
    public String getQueryString() {
        return queryString;
    }

    /**
     * Sets any query string that is part of the HTTP request URI.
     * The querystring can be set by a RequestDispatcher forward.
     */
    public void setQueryString(String queryString) {
        this.queryString = queryString;
        // force the parameters to be recalculated
        parameters = null;
    }


    /**
     * Gets the name of the user making this request.  The user name is
     * set with HTTP authentication.
     *
     * @return The name of the user making this request, or null if not
     * known.
     */
    public String getRemoteUser() {
        return remoteUser;
    }

    /**
     * Gets the authentication scheme of this request.
     *
     * @return This request's authentication scheme, or null if none.
     */
    public String getAuthType() {
        return authType;
    }

    /**
     * Gets the value of the requested header field of this request.
     *
     * @param name The String containing the name of the requested
     * header field.  The comparison is case insensitive.
     * @return The value of the requested header field, or null if not
     * known.
     */
    public String getHeader(String name) {
        return headers.getValue(name);
    }

    /**
     * Gets the value of the specified integer header field of this
     * request.
     *
     * @param name The String containing the name of the requested
     * header field.   The comparison is case insensitive.
     * @return the value of the requested header field, or -1 if not
     * found.
     * @exception NumberFormatException If value of header can't be
     *  converted to an int.
     */
    public int getIntHeader(String name) {
        return headers.getInt(name);
    }

    /**
     * Gets the value of the requested date header field of this
     * request.
     *
     * @param name the String containing the name of the requested
     * header field.   The comparison is case insensitive.
     * @return the value the requested date header field, or -1 if not
     * found.
     * @exception IllegalArgumentException If value of header can't be
     *  converted to an int.
     */
    public long getDateHeader(String name) {
        String value = headers.getValue(name);
        if (value == null) {
            return -1;
        } else {
            try {
                return dateParser.parse(value).getTime();
            } catch(ParseException except) {
                throw new IllegalArgumentException(except.getMessage());
            }
        }
    }

    /**
     * Gets the header names for this request.
     *
     * @return An enumeration of strings representing the header names
     * for this request.
     */
    public Enumeration getHeaderNames() {
        return headers.getNames();
    }

    /**
     * Gets the current valid session associated with this request, if
     * create is false or, if necessary, creates a new session for the
     * request, if create is true.
     *
     * @return The session associated with this request or null if
     * create was false and no valid session is associated
     * with this request.
     */
    public HttpSession getSession (boolean create) {
	// X22
        throw new ServletError("Servlet sessions are not yet implemented");
    }

    /**
     * Gets the session id specified with this request.
     *
     * @return The session id specified by this request, or null if the
     * request did not specify a session id.
     *
     * @see #isRequestedSessionIdValid
     */
    public String getRequestedSessionId () {
	// X22
        throw new ServletError("Servlet sessions are not yet implemented");
    }

    /**
     * Checks whether this request is associated with a session that
     * is valid in the current session context.
     *
     * @return True if this request is assocated with a session that is
     * valid in the current session context.
     *
     * @see #getRequestedSessionId
     * @see javax.servlet.http.HttpSessionContext
     * @see #getSession
     */
    public boolean isRequestedSessionIdValid () {
	// X22
        throw new ServletError("Servlet sessions are not yet implemented");
    }

    /**
     * Checks whether the session id specified by this request came in
     * as a cookie.
     *
     * @return True if the session id specified by this request came in
     * as a cookie; false otherwise.
     *
     * @see #getSession
     */
    public boolean isRequestedSessionIdFromCookie () {
	// X22
        throw new ServletError("Servlet sessions are not yet implemented");
    }

    /**
     * Checks whether the session id specified by this request came in
     * as part of the URL.
     *
     * @return True if the session id specified by the request for this
     * session came in as part of the URL; false otherwise.
     *
     * @see #getSession
     */
    public boolean isRequestedSessionIdFromUrl () {
	// X22
        throw new ServletError("Servlet sessions are not yet implemented");
    }


    public boolean isUserInRole(String role) {
	// X22
	throw new Error("not implemented");
    }

    public Locale getLocale() {
	// X22
	throw new Error("not implemented");
    }

    public java.security.Principal getUserPrincipal() {
	// X22
	throw new Error("not implemented");
    }

    public Enumeration getLocales() {
	// X22
	throw new Error("not implemented");
    }

    public String getContextPath() {
	// X22
	throw new Error("not implemented");
    }

    public Enumeration getHeaders(String name) {
	// X22
	throw new Error("not implemented");
    }

    public void removeAttribute(String name) {
	// X22
	attributes.remove(name);
    }

    public boolean isSecure() {
	// X22
	throw new Error("not implemented");
    }

    public boolean isRequestedSessionIdFromURL() {
	// X22
	throw new Error("not implemented");
    }

    public HttpSession getSession() {
	// X22
	throw new Error("not implemented");
    }

    public Enumeration getAttributeNames() {
	// X22
	return attributes.keys();
    }

    public void setAttribute(String name, Object value) {
	// X22
	attributes.put(name, value);
    }

    public RequestDispatcher getRequestDispatcher(String name) {
	// X22
	throw new Error("not implemented");
    }


    /**
     * Indicate the end of handing a request.
     */
    protected void finish() {
        attributes.clear();
    }


    // X22 for RequestAdapter
    public void recycle() {
	// XXX call close()?
	try {
	    close();
	} catch (IOException e) {
	    // don't care
	}
	socket = null;
	inputStream = new HttpInputStream();
	method = null;
	requestUri = null;
	httpVersion = null;
	queryString = null;
	servletPath = null;
	serverName = null;
	serverPort = 0;
	remoteUser = null;
	authType = null;
	headers = new HttpHeaders();
	cookies = null;
	parameters = null;
	servletContext = null;
	getInputStreamCalled = false;
	reader = null;
	keepAlive = false;
    }

    // X22 for RequestAdapter
    public MimeHeaders getMimeHeaders() {
	return HeaderConverter.toMimeHeaders(headers);
    }

    // X22 for RequestAdapter
    public String[] getCookieHeaders() {
	// X22 doesn't seem to be used
	throw new Error("not implemented");
    }

    // X22 for RequestAdapter
    public String getServletPrefix() {
	// X22 doesn't seem to be used
	throw new Error("not implemented");
    }

    // X22 for RequestAdapter
    public String getServletName() {
	// X22 doesn't seem to be used
	throw new Error("not implemented");
    }

    // X22 for RequestAdapter
    public int doRead() throws IOException {
	// X22 might need some checking here first
	return inputStream.read();
    }

    // X22 for RequestAdapter
    public int doRead(byte[] b, int off, int len) throws IOException {
	// X22 might need some checking here first
	return inputStream.read(b, off, len);
    }

}

