/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HttpRequestProcessor.java,v 1.5.2.4.2.1 2000/10/19 17:59:09 jasona Exp $
 */

package org.enhydra.servlet.connectionMethods.http;

import javax.servlet.*;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpUtils;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;
import java.util.Vector;
import java.io.IOException;
import java.io.BufferedReader;
import java.net.Socket;
import com.lutris.mime.MimeHeader;
import com.lutris.http.HttpCookieParser;
import org.enhydra.servlet.ServletIOException;
import com.lutris.util.FatalExceptionError;

/**
 * Class that implements the request setup and parsing for
 * <CODE>HttpRequest</CODE>.  This functionality is seperated
 * into this class for organization purposes, with the internal
 * functions being implemented here and the servlet interface
 * being implemented in <CODE>HttpRequest</CODE>.
 *
 * @author Mark Diekhans
 * @version $Revision: 1.5.2.4.2.1 $
 * @see HttpRequest
 */
class HttpRequestProcessor {
    /**
     * The socket associated with the request.
     */
    protected Socket socket;

    /**
     * Input stream associated with the socket.
     */
    protected HttpInputStream inputStream = new HttpInputStream();

    /**
     * Request method.
     */
    protected String method;

    /**
     * Request URI (less query string).
     */
    protected String requestUri;

    /**
     * Request HTTP Version.
     */
    protected String httpVersion;

    /**
     * Query string from URI.
     */
    protected String queryString;

    /**
     * The servlet path.  Must be set externally before the servlet is called.
     * We can't set it here, since its not known until the connection manager
     * has looked at the path.
     */
    protected String servletPath;

    /**
     * The server name.
     */
    protected String serverName;

    /**
     * The server port.
     */
    protected int serverPort;

    /**
     * The remote user, specified externally.
     */
    protected String remoteUser;

    /**
     * The auth type, specified externally.
     */
    protected String authType;

    /**
     * Parsed MIME headers.  Values are either a string or an array of
     * strings if multiple copies of the same header where supplied.
     */
    protected HttpHeaders headers = new HttpHeaders();

    /**
     * Parser Cookies.
     */
    protected Cookie[] cookies;

    /**
     * Parsed parameters from query string or POST.
     */
    protected Hashtable parameters;

    /**
     * Hash table that is always empty; user when no parameters.
     */
    private static final Hashtable emptyHashTable = new Hashtable();

    /**
     * ServletContext object for the current servlet.
     */
    protected ServletContext servletContext;

    /**
,     * Has getInputStream() been called on the current request?
     */
    protected boolean getInputStreamCalled;

    /**
     * Reader for stream.
     */
    protected BufferedReader reader;

    /**
     * Was a keep-alive connection requested?
     */
    protected boolean keepAlive;

    /**
     * Table of post parameters.  This is kept to work around the servlet
     * debugger asking for parameters multiple times.
     * FIXME: this should be change into a error check once the
     * request object handling in the servlet debugger is cleaned up.
     */
    private Hashtable postParams = null;

    /**
     * Associate a connected socket with this object.  The
     * <CODE>next</CODE> function must be called before a
     * request can be processed.
     * @param requestSocket The HTTP connected socket.
     * @exception IOException If an error occured setting up the socket
     *  input stream.
     */
    protected void setConnection(Socket requestSocket) throws IOException {
        socket = requestSocket;
        inputStream.setConnection(requestSocket);
	// XXX Don't do this, since it may return only the short name.
	// XXX Use the host header instead.
        //serverName = socket.getLocalAddress().getHostName();
        serverPort = socket.getLocalPort();
    }

    /**
     * Close the socket associate with this object.
     * @exception IOException If an error occured setting up the socket
     *  input stream.
     */
    protected void close() throws IOException {
        try {
            if (reader != null) {
                reader.close();
            }
            inputStream.close();
        } catch (IOException except) {
            throw ServletIOException.rethrow(except);
        }
    }

    /**
     * Read and parse the HTTP request line.
     * @return True if headers where read, false if an eof was encountered.
     * @exception IOException If a read or parse error or occurs.
     */
    private boolean readRequest() throws IOException {
        String request = inputStream.readHeader();
        if (request == null) {
            return false;  //EOF
        }

        // Request-Line  = Method SP Request-URI SP HTTP-Version CRLF

        StringTokenizer tokens = new StringTokenizer(request);
        method = tokens.nextToken().toUpperCase();
        String uri = tokens.nextToken();
        httpVersion = tokens.nextToken();

        // Parse URL
        int pos = uri.indexOf('?');
        if (pos >= 0) {
            requestUri = uri.substring(0, pos);
            queryString = uri.substring(pos+1);
        } else {
            requestUri = uri;
            queryString = null;
        }
        return true;
    }

    /**
     * Parse cookies from header.
     */
    private void parseCookies(String header) {
        Vector newCookies = HttpCookieParser.parseRequestHeader(header);
        Cookie[] newTable;
        if (cookies == null) {
            newTable = new Cookie[newCookies.size()];
            newCookies.copyInto(newTable);
        } else {
            newTable = new Cookie[cookies.length + newCookies.size()];
            System.arraycopy(cookies, 0, newTable, 0, cookies.length);
            int idx = cookies.length;

            Enumeration elems = newCookies.elements();
            while (elems.hasMoreElements()) {
                newTable[idx++] = (Cookie)elems.nextElement();
            }
        }
        cookies = newTable;
    }

    /**
     * Parse a header line and store in the table.
     */
    private void parseHeaderLine(String headerLine) throws IOException {
        int pos = headerLine.indexOf(':');
        if (pos < 0) {
            throw new ServletIOException("Invalid HTTP request header: \""
                                         + headerLine + "\"");
        }
        String name = headerLine.substring(0, pos).trim();
        String value = headerLine.substring(pos+1).trim();

        headers.add(name, value);
        if (name.equalsIgnoreCase("cookie")) {
            parseCookies(value);
        }
    }

    /**
     * Read and parse the mime headers into this object.
     * @exception IOException If a read or parse error or occurs.
     */
    private void readHeaders() throws IOException {
        String headerLine;
        while ((headerLine = inputStream.readHeader()) != null) {
            parseHeaderLine(headerLine);
        }
        if (cookies == null) {
            cookies = new Cookie[0];
        }
    }

    /**
     * Parse a post request.
     * @return The parameters, or null if they where not obtained.
     * @exception IOException If a read or parse error or occurs.
     */
    private Hashtable parsePostParameters() throws IOException {
        String contentType = headers.getValue("Content-Type");
        if (contentType == null) {
            throw new ServletIOException("Invalid HTTP post request: Content-Type not specified");
        }
        if (contentType.equals("application/x-www-form-urlencoded")) {
            int contentLength;
            try {
                contentLength = headers.getInt("content-length");
            } catch (NumberFormatException except) {
                throw new ServletIOException("Invalid HTTP post request: Content-Length invalid");
            }
            if (contentLength < 0) {
                throw new ServletIOException("Invalid HTTP post request: Content-Length not specified");
            }
            // Only parse parameters from stream once.
            // FIXME: This is a hack to work around a problem with the servlet
            // debugger requesting parameters twice.  Particularlly if
            // 
	    if (postParams == null) {
		postParams = HttpUtils.parsePostData(contentLength, inputStream);
	    }
            return postParams;
        } else {
            if (queryString != null) {
                return HttpUtils.parseQueryString(queryString);
            }
        }
        return null;
    }

    /**
     * Parse either the query string or POST input parameters. Or both
     * A form may have a POST method and an action that includes a query string
     * Tomcat provides a utility that can merge the two tables. so a variable
     * can appear in both. both values will be saved.
     *
     * @exception IOException If a read or parse error or occurs.
     */
    private void parseParameters() throws IOException {
        if (method.equals("POST")) {
            parameters = parsePostParameters();
        }
        if (parameters == null) {
            parameters = emptyHashTable;
        }
        // Add in the query even if the method was a POST.
        if (queryString != null) {
            Hashtable qParams =  HttpUtils.parseQueryString(queryString);
            parameters = org.apache.tomcat.core.RequestUtil.mergeParameters(qParams, parameters);
        }
    }

    /**
     * Parse either the query string or POST input parameters.  To conform
     * with the behavior of the Java WWW server, this should be call only
     * when parameters are first requested.  Otherwise, on a POST, the
     * input stream will be read in an unexpected way.  This turns all
     * IOExceptions into FatalExceptionErrors as the servlet methods that
     * call this don't return exceptions.
     */
    protected void getParsedParameters() {
        try {
            parseParameters();
        } catch (IOException except) {
            throw new FatalExceptionError(except);
        }
    }



    /**
     * Initialize for a request on a socket. With keep-alive, multiple
     * requests can be made on a single connection.  This parses the
     * headers for the request and saves them in this object for access.
     *
     * @exception IOException If a parse error or IOException occurs
     *  reading the request.
     * @return True if a request has been processed; false if the socket
     *  was closed and there are no more requests.
     */
    protected boolean next() throws IOException {
        inputStream.next();
        if (!readRequest()) {
            return false;
        }
        parameters = null;
        postParams = null;
        cookies = null;
        headers.clear();
        readHeaders();

        // Determine if Keep-Alive is requested.
        String connValue = headers.getValue("Connection");
        if ((connValue == null) || !connValue.equalsIgnoreCase("Keep-Alive")) {
            keepAlive = false;
        } else {
            keepAlive = true;
        }

        // Clear values that should be initialized each request.
        remoteUser = null;
        authType = null;
        getInputStreamCalled = false;
        reader = null;
        return true;
    }


    /**
     * Sets information that can not be obtained from the request.
     * This should be called after next() and before servlet.service() is
     * called.
     *
     * @param reqServletPath The servlet path portion of the URL.
     * @param content The ServletContext object for the servlet.
     */
    protected void setServletInfo(String reqServletPath,
                                  ServletContext context) {
        servletPath = reqServletPath;
        servletContext = context;
    }

    /**
     * Sets the name of the user making this request.  The user name is
     * set with HTTP authentication.
     *
     * @param The name of the user making this request.
     */
    public void setRemoteUser(String reqRemoteUser) {
        remoteUser = reqRemoteUser;
    }

    /**
     * Sets the authentication scheme of this request.
     *
     * @param This request's authentication scheme.
     */
    public void setAuthType(String reqAuthType) {
        authType = reqAuthType;
    }

    /**
     * Check if keep-alive requested?.
     *
     * @return True if a keep-alive connection was requested.
     */
    protected boolean keepAliveRequested() {
        return keepAlive;
    }

}

