/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: OutputStreamRecord.java,v 1.4.12.1 2000/10/19 17:59:10 jasona Exp $
 */

package org.enhydra.servlet.debug;

import java.io.PrintWriter;
import java.io.IOException;
import javax.servlet.ServletOutputStream;

/**
 *  This class is used to trace the calls to <CODE>print()</CODE> and
 *  <CODE>println()</CODE> and, optionally, to save a copy of the data
 *  written to the output stream. The tracing and data-copying are 
 *  accomplished via an associated ResponseRecord object.
 */
class OutputStreamRecord extends ServletOutputStream {

    private ResponseRecord          responseRecord;
    private ServletOutputStream     realStream;
    private StringBuffer            responseData; // Use if not null.

    /**
     *  All calls will also be logged to the trace string via the
     *  <CODE>ResponseRecord</CODE> passed in. If saveResponseData
     *  is true, the data written will also be saved to the
     *  ResponseData's StringBuffer called responseData. <P>
     *
     *  This class passes all data through unmodified. <P>
     *
     * @param responseRecord The response associated with this output
     * stream. It's <CODE>addTrace()</CODE> method will be called to log the 
     * calls to the <CODE>print()</CODE> and <CODE>println()</CODE>
     * methods. If saveResponseData is true, then it's StringBuffer
     * responseData will be appened to, saving a copy of the data written
     * to the output stream.
     *
     * @param realStream After saving a copy, all the data is passed
     * through unmodified to this output stream.
     */
    OutputStreamRecord (ResponseRecord responseRecord,
                        boolean saveResponseData,
                        ServletOutputStream realStream) {
        this.responseRecord = responseRecord;
        this.realStream = realStream;
        if (saveResponseData)
            responseData = responseRecord.responseData;
        else
            responseData = null;
     }

    public void print(String s) throws IOException {
        try {
            realStream.print(s);
        } catch (IOException e) {
            addTrace("print(String, " + s.length() + " chars) threw " + e);
            throw e;
        }
        if (responseData != null)
            responseData.append(s);
        addTrace("print(String, " + s.length() + " chars)");
    }

    public void print(boolean b) throws IOException {
        try {
            realStream.print(b);
        } catch (IOException e) {
            addTrace("print(" + b + ") threw " + e);
            throw e;
        }
        if (responseData != null)
            responseData.append(b);
        addTrace("print(" + b + ")");
    }

    public void print(char c) throws IOException {
        try {
            realStream.print(c);
        } catch (IOException e) {
            addTrace("print(char " + (int)c + ") threw " + e);
            throw e;
        }
        if (responseData != null)
            responseData.append(c);
        addTrace("print(char " + (int)c + ")");
    }

    public void print(int i) throws IOException {
        try {
            realStream.print(i);
        } catch (IOException e) {
            addTrace("print(int " + i + ") threw " + e);
            throw e;
        }
       if (responseData != null)
            responseData.append(i);
        addTrace("print(int " + i + ")");
    }

    public void print(long l) throws IOException {
        try {
            realStream.print(l);
        } catch (IOException e) {
            addTrace("print(long " + l + ") threw " + e);
            throw e;
        }
        if (responseData != null)
            responseData.append(l);
        addTrace("print(long " + l + ")");
    }

    public void print(float f) throws IOException {
        try {
            realStream.print(f);
        } catch (IOException e) {
            addTrace("print(float " + f + ") threw " + e);
            throw e;
        }
        if (responseData != null)
            responseData.append(f);
        addTrace("print(float " + f + ")");
    }

    public void print(double d) throws IOException {
        try {
            realStream.print(d);
        } catch (IOException e) {
            addTrace("print(double " + d + ") threw " + e);
            throw e;
        }
        if (responseData != null)
            responseData.append(d);
        addTrace("print(double " + d + ")");
    }


    public void println() throws IOException {
        try {
            realStream.println();
        } catch (IOException e) {
            addTrace("println() threw " + e);
            throw e;
        }
        if (responseData != null)
            responseData.append('\n');
        addTrace("println()");
    }

    public void println(String s) throws IOException {
        try {
            realStream.println(s);
        } catch (IOException e) {
            addTrace("println(String, " + s.length() + " chars) threw " + e);
            throw e;
        }
        if (responseData != null) {
            responseData.append(s);
            responseData.append('\n');
        }
        addTrace("println(String, " + s.length() + " chars)");
    }

    public void println(boolean b) throws IOException {
        try {
            realStream.println(b);
        } catch (IOException e) {
            addTrace("println(" + b + ") threw " + e);
            throw e;
        }
        if (responseData != null) {
            responseData.append(b);
            responseData.append('\n');
        }
        addTrace("println(" + b + ")");
    }

    public void println(char c) throws IOException {
        try {
            realStream.println(c);
        } catch (IOException e) {
            addTrace("println(char " + (int)c + ") threw " + e);
            throw e;
        }
        if (responseData != null) {
            responseData.append(c);
            responseData.append('\n');
        }
        addTrace("println(char " + (int)c + ")");
    }

    public void println(long l) throws IOException {
        try {
            realStream.println(l);
        } catch (IOException e) {
            addTrace("println(long " + l + ") threw " + e);
            throw e;
        }
        if (responseData != null) {
            responseData.append(l);
            responseData.append('\n');
        }
        addTrace("println(long " + l + ")");
    }

    public void println(float f) throws IOException {
        try {
            realStream.println(f);
        } catch (IOException e) {
            addTrace("println(float " + f + ") threw " + e);
            throw e;
        }
        if (responseData != null) {
            responseData.append(f);
            responseData.append('\n');
        }
        addTrace("println(float " + f + ")");
    }

    public void println(double d) throws IOException {
        try {
            realStream.println(d);
        } catch (IOException e) {
            addTrace("println(double " + d + ") threw " + e);
            throw e;
        }
        if (responseData != null) {
            responseData.append(d);
            responseData.append('\n');
        }
        addTrace("println(double " + d + ")");
    }

    public void write(int n) throws IOException {
        try {
            realStream.write(n);
        } catch (IOException e) {
            addTrace("write(int " + n + ") threw " + e);
            throw e;
        }
        if (responseData != null)
            responseData.append(n);
        addTrace("write(int " + n + ")");
    }

    public void write(byte b[]) throws IOException {
        try {
            realStream.write(b);
        } catch (IOException e) {
            addTrace("write(byte[" + b.length + "]) threw " + e); 
            throw e;
        }
        if (responseData != null) {
            String temp = new String(b);
            responseData.append(temp.toCharArray());
        }
        addTrace("write(byte[" + b.length + "])"); 
    }


    public void write(byte b[], int off, int len) throws IOException {
        try {
            realStream.write(b, off, len);
        } catch (IOException e) {
            addTrace("write(byte[" + b.length + "], " + off + ", " + 
                     len + ") threw " + e); 
            throw e;
        }
        if (responseData != null) {
            String temp = new String(b, off, len);
            responseData.append(temp.toCharArray());
        }
        addTrace("write(byte[" + b.length + "], " + off + ", " + len + ")"); 
    }


    public void flush() throws IOException {
        try {
            realStream.flush();
        } catch (IOException e) {
            addTrace("flush() threw " + e);
            throw e;
        }
        addTrace("flush()");
    }

    public void close() throws IOException {
        try {
            realStream.close();
        } catch (IOException e) {
            addTrace("close() threw " + e);
            throw e;
        }
        addTrace("close()");
    }

    /*
     * How many times have we added to the trace log?
     */
    private int counter = 0;

    /*
     * What is the maxmimum number of times to write to the trace log?
     * There could be hundreds of calls to write()...
     */
    private static final int maxTraces = 64;

    /*
     * If we havn't logged too much, log a message to the trace string.
     */
    private void addTrace(String msg) {
        counter++;
        if (counter > maxTraces)
            return;
        if (responseRecord == null)
            return;
        msg = "ServletOutputStream: " + msg;
        if (counter == maxTraces) {
            responseRecord.addTrace(msg); 
            responseRecord.addTrace("(No more ServletOutputStream calls " +
                                    "will be shown)");
            return;
        }
        responseRecord.addTrace(msg); 
    }
}


