/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: RequestRecord.java,v 1.4.2.1.2.2 2000/12/02 01:45:16 peterj Exp $
 */

package org.enhydra.servlet.debug;

import java.io.IOException;
import java.io.BufferedReader;
import java.util.Enumeration;
import java.util.Hashtable;
import java.text.SimpleDateFormat;
import java.text.ParseException;
import java.util.Date;
import java.util.Vector;
import java.util.Locale;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpSession;
import javax.servlet.ServletInputStream;
import javax.servlet.RequestDispatcher;
import org.enhydra.servlet.filter.*;
import com.lutris.http.*;
import com.lutris.http.debug.*;


/**
 * HttpServletRequest filter that generate a record of the request that was
 * made.  Once the request is complete, this object contains the information
 * and is used as the record of the request's content.
 *
 * @version  $Revision: 1.4.2.1.2.2 $
 * @author   Mark Diekhans
 * @see com.lutris.filter.TransactionFilter
 * @see javax.servlet.ServletRequest
 * @see javax.servlet.http.HttpServletRequest
 */
public class RequestRecord extends RequestFilter implements HttpServletRequest, DebugRequest {
    /**
     * If not null, the actual request object that will handle the request.
     * Set to null when complete() is called.
     */
    private HttpServletRequest request;

    /**
     * Transaction record that this object is associated with.
     */
    protected ServletTransactionRecord transactionRecord;

    /**
     * The wrapper around the input stream, logs calls.
     */
    private InputStreamRecord inputStreamRecord;

    /**
     * Parsers for legal formats for data headers.
     */
    private static SimpleDateFormat[] dateParsers = {
            new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss zzz"),
            new SimpleDateFormat("EEEEEE, dd-MMM-yy HH:mm:ss zzz"),
            new SimpleDateFormat("EEE MMMM d HH:mm:ss yyyy")
    };

    /*
     * Copies of all of the data field from the request.
     */
    private int contentLength;
    private String contentType;
    private String protocol;
    private String scheme;
    private String serverName;
    private int serverPort;
    private String remoteAddr;
    private String remoteHost;
    private Hashtable parameters;
    private String method;
    private String requestURI;
    private String servletPath;
    private String pathInfo;
    private String pathTranslated;
    private String queryString;
    private String remoteUser;
    private String authType;
    private Hashtable headers;
    private Cookie[] cookies;
    private String characterEncoding;

    /**
     * Constructor.  Wrap the object around a single request and record
     * infomation about it.  The information is stored in this object
     * for future access.
     *
     * @param originalRequest The real HttpServletRequest, which all calls
     *  will be passed on to.
     * @see javax.servlet.ServletRequest
     * @see javax.servlet.http.HttpServletRequest
     */
    protected RequestRecord(HttpServletRequest originalRequest) {
	super(originalRequest);
        request = originalRequest;

        /*
         * Save the content of the request.
         */
        contentLength = request.getContentLength();
        contentType = request.getContentType();
        protocol = request.getProtocol();
        scheme = request.getScheme();
        serverName = request.getServerName();
        serverPort = request.getServerPort();
        remoteAddr = request.getRemoteAddr();
        remoteHost = request.getRemoteHost();

        parameters = new Hashtable();
        Enumeration pnames = request.getParameterNames();
        while (pnames.hasMoreElements()) {
            String pname = (String)pnames.nextElement();
            parameters.put(pname, request.getParameterValues(pname));
        }
        method = request.getMethod();
        requestURI = request.getRequestURI();
        servletPath = request.getServletPath();
        pathInfo = request.getPathInfo();
// X22 this is a problem
try {
        pathTranslated = request.getPathTranslated();
} catch (Exception e) {
pathTranslated = null;
//System.err.println("pathTranslated threw exception " + e);
}
        queryString = request.getQueryString();
        remoteUser = request.getRemoteUser();
        authType = request.getAuthType();

        headers = new Hashtable();
        Enumeration hnames = request.getHeaderNames();
        while (hnames.hasMoreElements()) {
            String hname = (String)hnames.nextElement();
            if (hname == null)
                continue;
            String value = request.getHeader(hname);
            if (value == null)
                continue;  // Don't put nulls into hash table. Maybe we should?
            headers.put(hname.toLowerCase(), value);
        }
        cookies = request.getCookies();
        if (cookies == null)
            cookies = new Cookie[0];
        characterEncoding = getCharacterEncoding();
    }

    /**
     * Flag the request as completed.  This any summarization to occur
     * and the object to disassociated with the original request object
     * to help catch errors.
     */
    protected void completed() {
        request = null;
    }

    /**
     * Add a string to the trace buffer. InputStreamRecords will also call this.
     */
    public void addTrace(String str) {
        if (transactionRecord != null) {
            transactionRecord.addTrace(str);
        }
    }

    /**
     * Record call and return cached value.
     *
     * @see javax.servlet.ServletRequest#getContentLength
     */
    public int getContentLength() {
        addTrace("getContentLength() = " + contentLength);
        return contentLength;
    }

    /**
     * Record call and return cached value.
     *
     * @see javax.servlet.ServletRequest#getContentType
     */
    public String getContentType() {
        addTrace("getContentType() = " + contentType);
        return contentType;
    }

    /**
     * Record call and return cached value.
     *
     * @see javax.servlet.ServletRequest#getProtocol
     */
    public String getProtocol() {
        addTrace("getProtocol() = " + protocol);
        return protocol;
    }

    /**
     * Record call and return cached value.
     *
     * @see javax.servlet.ServletRequest#getScheme
     */
    public String getScheme() {
        addTrace("getScheme() = " + scheme);
        return scheme;
    }

    /**
     * Record call and return cached value.
     *
     * @see javax.servlet.ServletRequest#getServerName
     */
    public String getServerName() {
        addTrace("getServerName() = " + serverName);
        return serverName;
    }

    /**
     * Record call and return cached value.
     *
     * @see javax.servlet.ServletRequest#getServerPort
     */
    public int getServerPort() {
        addTrace("getServerPort() = " + serverPort);
        return serverPort;
    }

    /**
     * Record call and return cached value.
     *
     * @see javax.servlet.ServletRequest#getRemoteAddr
     */
    public String getRemoteAddr() {
        addTrace("getRemoteAddr() = " + remoteAddr);
        return remoteAddr;
    }

    /**
     * Record call and return cached value.
     *
     * @see javax.servlet.ServletRequest#getRemoteHost
     */
    public String getRemoteHost() {
        addTrace("getRemoteHost() = " + remoteHost);
        return remoteHost;
    }

    /**
     * Pass on to actual request object.  Will generate an error if used
     * after request has completed.
     *
     * @see javax.servlet.ServletRequest#getRealPath
     */
    public String getRealPath(String path) {
        String realPath = request.getRealPath(path);
        addTrace("getRealPath(" + path + ") = " + realPath);
        return realPath;
    }

    /**
     * Returns a wrapper around the real input stream. All read calls will
     * be logged via this object's addTrace() method.
     *
     * @see javax.servlet.ServletRequest#getInputStream
     */
    public ServletInputStream getInputStream() throws IOException {
        ServletInputStream is = request.getInputStream();
        addTrace("getInputStream() = " + is);
        if (inputStreamRecord == null)
            inputStreamRecord = new InputStreamRecord(this, is);
        return inputStreamRecord;
    }

    /**
     * Record call and return cached value.
     *
     * @see javax.servlet.ServletRequest#getParameter
     */
    public String getParameter(String name) {
        String [] value = (String[])parameters.get(name);
        if (value == null) {
            addTrace("getParameter(" + name + ") = null");
            return null;
        } else {
            addTrace("getParameter(" + name + ") = " + value[0]);
            return value[0];
        }
    }

    /**
     * Record call and return cached value.
     *
     * @see javax.servlet.ServletRequest#getParameterValues
     */
    public String[] getParameterValues(String name) {
        String[] results = (String[])parameters.get(name);
        if (results == null || results.length == 0)
            addTrace("getParameterValues(" + name + ") = []");
        else {
            addTrace("getParameterValues(" + name + ") = [");
            for (int i=0; i<results.length; i++)
                addTrace("&nbsp;&nbsp;&nbsp;&nbsp;" +
                         results[i]);
            addTrace("]");
        }
        return results;
    }

    /**
     * Record call and return cached value.
     *
     * @see javax.servlet.ServletRequest#getParameterNames
     */
    public Enumeration getParameterNames() {
        Vector v = new Vector();
        Enumeration e = parameters.keys();
        while (e.hasMoreElements())
            v.addElement(e.nextElement());
        if (v.size() == 0)
            addTrace("getParameterNames() = {}");
        else {
            addTrace("getParameterNames() = {");
            for (int i=0; i<v.size(); i++)
                addTrace("&nbsp;&nbsp;&nbsp;&nbsp;" +
                         (String) v.elementAt(i));
            addTrace("}");
        }
        return parameters.keys();
    }

    /**
     * Record call and pass on to actual request object.  Will return null
     * if used after request is complete.
     *
     * @see javax.servlet.ServletRequest#getAttribute
     */
    public Object getAttribute(String name) {
        Object result = null;
        if (request != null)
            result = request.getAttribute(name);
        if (!name.startsWith("Enhydra.Recorder"))
            addTrace("getAttribute(" + name + ") = " + result);
        return result;
    }

    /**
     * Record call and return cached value.
     *
     * @see javax.servlet.http.HttpServletRequest#getMethod
    */
    public String getMethod() {
        addTrace("getMethod() = " + method);
        return method;
    }

    /**
     * Record call and return cached value.
     *
     * @see javax.servlet.http.HttpServletRequest#getRequestURI
    */
    public String getRequestURI() {
        addTrace("getRequestURI() = " + requestURI);
        return requestURI;
    }

    /**
     * Record call and return cached value.
     *
     * @see javax.servlet.http.HttpServletRequest#getServletPath
     */
    public String getServletPath() {
        addTrace("getServletPath() = " + servletPath);
        return servletPath;
    }

    /**
     * Record call and return cached value.
     *
     * @see javax.servlet.http.HttpServletRequest#getPathInfo
     */
    public String getPathInfo() {
        addTrace("getPathInfo() = " + pathInfo);
        return pathInfo;
    }

    /**
     * Record call and return cached value.
     *
     * @see javax.servlet.http.HttpServletRequest#getPathTranslated
     */
    public String getPathTranslated() {
        addTrace("getPathTranslated() = " + pathTranslated);
        return pathTranslated;
    }

    /**
     * Record call and return cached value.
     *
     * @see javax.servlet.http.HttpServletRequest#getQueryString
     */
    public String getQueryString() {
        addTrace("getQueryString() = " + queryString);
        return queryString;
    }

    /**
     * Record call and return cached value.
     *
     * @see javax.servlet.http.HttpServletRequest#getRemoteUser
     */
    public String getRemoteUser() {
        addTrace("getRemoteUser() = " + remoteUser);
        return remoteUser;
    }

    /**
     * Record call and return cached value.
     *
     * @see javax.servlet.http.HttpServletRequest#getAuthType
     */
    public String getAuthType() {
        addTrace("getAuthType() = " + authType);
        return authType;
    }

    /**
     * Record call and return cached value.
     *
     * @see javax.servlet.http.HttpServletRequest#getHeader
     */
    public String getHeader(String name) {
        String value = (String)headers.get(name.toLowerCase());
        addTrace("getHeader(" + name + ") = " + value);
        return value;
    }

    /**
     * Record call and return cached value.
     *
     * @see javax.servlet.http.HttpServletRequest#getIntHeader
     */
    public int getIntHeader(String name) {
        String svalue = (String)headers.get(name.toLowerCase());
        int ivalue;
        if (svalue == null) {
            ivalue = -1;
        } else {
            ivalue = Integer.parseInt(svalue);
        }
        addTrace("getIntHeader(" + name + ") = " + ivalue);
        return ivalue;
    }

    /**
     * Record call and return cached value.
     *
     * @see javax.servlet.http.HttpServletRequest#getDateHeader
     */
    public long getDateHeader(String name) {
        String svalue = (String)headers.get(name.toLowerCase());
        if (svalue == null) {
            addTrace("getDateHeader(" + name + ") = -1");
            return -1;
        }

        // Attempt to parse the header in the various legal formats.
        for (int i = 0; i < dateParsers.length; i++) {
            try {
                long result =  dateParsers[i].parse(svalue).getTime();
                addTrace("getDateHeader(" + name + ") = " + result);
                return result;
            } catch (ParseException e) {
                // Ignore, try another.
            }
        }
        // Its not legal.
        throw new IllegalArgumentException(svalue);
    }

    /**
     * Record call and return cached value.
     *
     * @see javax.servlet.http.HttpServletRequest#getHeaderNames
     */
    public Enumeration getHeaderNames() {
        Vector v = new Vector();
        Enumeration e = headers.keys();
        while (e.hasMoreElements())
            v.addElement(e.nextElement());
        if (v.size() == 0)
            addTrace("getHeaderNames() = {}");
        else {
            addTrace("getHeaderNames() = {");
            for (int i=0; i<v.size(); i++)
                addTrace("&nbsp;&nbsp;&nbsp;&nbsp;" +
                         (String) v.elementAt(i));
            addTrace("}");
        }
        return headers.keys();
    }

    /**
     * Gets the array of cookies found in this request.
     *
     * @return
     *   The array of cookies found in this request.
     * @see javax.servlet.http.HttpServletRequest#getCookies
     */
    public Cookie[] getCookies() {
        if (cookies.length == 0)
            addTrace("getCookies() = []");
        else {
            addTrace("getCookies() = [");
            for (int i=0; i<cookies.length; i++)
                addTrace("&nbsp;&nbsp;&nbsp;&nbsp;" +
                         HttpCookieParser.formatResponseCookie(cookies[i]));
            addTrace("]");
        }
        return cookies;
    }

    /**
     * Returns the number of bytes associated with the response.
     */
    public int getTotalBytes() {
	// X22
        return -1;  //FIXME
    }

    /**
     * Get or create a session.
     * @return
     *    The session associated with this request or null if create was false
     * and no valid session is associated with this request.
     * @see javax.servlet.http.HttpServletRequest#getSession
     */
    public HttpSession getSession(boolean create) {
	// X22
        throw new Error("Servlet sessions are not currently supported");
    }

    /**
     * Gets the session id specified with this request.
     *
     * @return
     *    The session id specified by this request, or null if the request did
     *    not specify a session id
     * @see isRequestedSessionIdValid
     * @see javax.servlet.http.HttpServletRequest#getRequestedSessionId
     */
    public String getRequestedSessionId() {
	// X22
        throw new Error("Servlet sessions are not currently supported");
    }

    /**
     * Checks whether this request is associated with a session that is valid
     * in the current session context.
     *
     * @return
     *   <CODE>True</CODE> if this request is assocated with a session that
     * is valid in the current session context.
     * @see javax.servlet.http.HttpServletRequest#isRequestedSessionIdValid
     */
    public boolean isRequestedSessionIdValid() {
	// X22
        throw new Error("Servlet sessions are not currently supported");
    }

    /**
     * Checks whether the session id specified by this request came in as a
     * cookie.
     *
     * @return
     *  True if the session id specified by this request came in as a cookie; false otherwise
     * @see javax.servlet.http.HttpServletResponse#isRequestedSessionIdFromCookie
     */
    public boolean isRequestedSessionIdFromCookie() {
	// X22
        throw new Error("Servlet sessions are not currently supported");
    }

    /**
     * Checks whether the session id specified by this request came in as part
     * of the URL. (The requested session may not be the one returned by the
     * getSession method.)
     *
     * @return
     *   True if the session id specified by the request for this session came
     * in as part of the URL; false otherwise
     * @see javax.servlet.http.HttpServletResponse#isRequestedSessionIdFromUrl
     */
    public boolean isRequestedSessionIdFromUrl() {
	// X22
        throw new Error("Servlet sessions are not currently supported");
    }

    /**
     * Returns a buffered reader for reading text in the request body.
     *
     * @see javax.servlet.ServletRequest#getReader
     * @exception IOException
     * On other I/O related errors.
     */
    public BufferedReader getReader() throws IOException {
        BufferedReader result = null;
        if (request != null)
            result = request.getReader();
        addTrace("getReader() = " + result);
        return result;
    }

    /**
     * Returns the character set encoding for the input of this request.
     *
     * @see javax.servlet.ServletRequest#getCharacterEncoding
     */
    public String getCharacterEncoding() {
        addTrace("getCharacterEncoding() = " + characterEncoding);
        return characterEncoding;
    }


    public boolean isUserInRole(String role) {
	// X22
	throw new Error("not implemented");
    }

    public Locale getLocale() {
	// X22
	throw new Error("not implemented");
    }

    public java.security.Principal getUserPrincipal() {
	// X22
	throw new Error("not implemented");
    }

    public Enumeration getLocales() {
	// X22
	throw new Error("not implemented");
    }

    public String getContextPath() {
	// X22
	throw new Error("not implemented");
    }

    public Enumeration getHeaders(String name) {
	// X22
	throw new Error("not implemented");
    }

    public void removeAttribute(String name) {
	// X22
        if (request != null)
            request.removeAttribute(name);
        addTrace("removeAttribute(" + name + ")");
    }

    public boolean isSecure() {
	// X22
	throw new Error("not implemented");
    }

    public boolean isRequestedSessionIdFromURL() {
	// X22
	throw new Error("not implemented");
    }

    public HttpSession getSession() {
	// X22
	throw new Error("not implemented");
    }

    public Enumeration getAttributeNames() {
	// X22
        Enumeration result = null;
        if (request != null)
            result = request.getAttributeNames();
        addTrace("getAttributeNames() = " + result);
        return result;
    }

    public void setAttribute(String name, Object value) {
	// X22
        if (request != null)
            request.setAttribute(name, value);
        if (!name.startsWith("Enhydra.Recorder"))
            addTrace("setAttribute(" + name + ", " + value + ")");
    }

    public RequestDispatcher getRequestDispatcher(String name) {
	// X22
	throw new Error("not implemented");
    }

}

