/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: ResponseRecord.java,v 1.4.2.2.2.1 2000/10/19 17:59:10 jasona Exp $
 */

package org.enhydra.servlet.debug;

import java.io.IOException;
import java.io.PrintWriter;
import java.text.SimpleDateFormat;
import java.util.Hashtable;
import java.util.Vector;
import java.util.TimeZone;
import java.util.Date;
import java.util.Enumeration;
import java.util.Locale;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.ServletOutputStream;
import javax.servlet.http.Cookie;
import com.lutris.http.*;
import com.lutris.http.debug.*;
import org.enhydra.servlet.filter.*;


/**
 * HttpServletResponse filter that generate a record of the response that was
 * made.  Once the transaction is complete, this object contains the
 * information.
 *
 * @version  $Revision: 1.4.2.2.2.1 $
 * @author   Mark Diekhans
 * @see com.lutris.filter.ransactionFilter
 * @see javax.servlet.ServletResponse
 * @see javax.servlet.http.HttpServletResponse
 */
public class ResponseRecord extends ResponseFilter implements HttpServletResponse, DebugResponse {

    /**
     * Object for formatting dates.
     */
    private static final SimpleDateFormat dateFormatter =
        new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss zzz");

    /**
     * If not null, the actual response object that will handle the response.
     * Set to null when complete() is called.
     */
    private HttpServletResponse response;

    /**
     *  Should we capture the data written to the socket?
     */
    private boolean saveResponseData;

    /**
     *  If we are to capture the data written to the socket, save it here.
     *  An OutputStreamRecord will call append() on this.
     */
    protected StringBuffer responseData;

    /**
     *  If we are to capture the data written to the socket, this is the
     *  OutputStreamRecord to return when asked for the output stream.
     */
    private OutputStreamRecord outputStreamRecord;

    /**
     *  If we are to capture the data written to the socket, this is the
     *  PrintWriterRecord to return when asked for the PrintWriter.
     */
    private PrintWriterRecord printWriterRecord;

    /**
     * Transaction record that this object is associated with.
     */
    protected ServletTransactionRecord transactionRecord;

    /*
     * Copy of all data that was set.
     */
    private int contentLength = -1;
    private String contentType;
    private int statusCode = -1;
    private String statusMsg = "<i>Not set</i>";
    private Hashtable headers = new Hashtable();
    private Vector cookies = new Vector();



    /**
     * Class constructor.
     */
    static {
	dateFormatter.setTimeZone(TimeZone.getTimeZone("GMT"));
    }

    /**
     * Constructor.  Wrap the object around a single response and record
     * infomation about it.  The information is stored in this object
     * for future access.
     *
     * @param originalResponse The real HttpServletResponse, which all calls
     *  will be passed on to.
     * @param originalResponse Should we save a copy of all data written
     *  to the socket? This could eat up alot of memory.
     * @see javax.servlet.ServletResponse
     * @see javax.servlet.http.HttpServletResponse
     */
    protected ResponseRecord(HttpServletResponse originalResponse,
                             boolean saveResponseData) {
	super(originalResponse);
        response = originalResponse;
        this.saveResponseData = saveResponseData;
    }

    /**
     * Flag the response as completed.  This any summarization to occur
     * and the object to disassociated with the original response object
     * to help catch errors.
     */
    protected void completed() {
        response = null;
    }

    /**
     * Add a string to the trace buffer.
     */
    protected void addTrace(String str) {
        if (transactionRecord != null) {
            transactionRecord.addTrace(str);
        }
    }

    /**
     * Special header processing.  Must remember content length
     * due to Tomcat integration problems.
     */
    private void processHeader(String name, String value) {
        headers.put(name.toLowerCase(), value);
        if (name.equalsIgnoreCase("content-type")) {
            contentType = value;
        } else if (name.equalsIgnoreCase("content-length")) {
            try {
                contentLength = Integer.parseInt(value);
            } catch (NumberFormatException except) {
                // Ignore bad size
            }
        } else if (name.equalsIgnoreCase("set-cookie")) {

            Cookie cookie = HttpCookieParser.parseCookieString(value);
            cookies.addElement(cookie);
        }
    }

    /**
     * Special header processing.  Must remember content length
     * due to Tomcat integration problems.
     */
    private void processHeader(String name, int value) {
        String strValue = Integer.toString(value);
        headers.put(name.toLowerCase(), strValue);
        if (name.equalsIgnoreCase("content-type")) {
            contentType = strValue;
        } else if (name.equalsIgnoreCase("content-length")) {
            try {
                contentLength = value;
            } catch (NumberFormatException except) {
                // Ignore bad size
            }
        }
    }

    /**
     * Record call and pass on to actual response object..
     *
     * @see javax.servlet.ServletResponse#setContentLength
     */
    public void setContentLength(int len) {
        addTrace("setContentLength(" + len + ")");
        contentLength = len;
        response.setContentLength(len);
    }

    /**
     * Record call and pass on to actual response object..
     *
     * @see javax.servlet.ServletResponse#setContentType
     */
    public void setContentType(String type) {
        addTrace("setContentType(" + type + ")");
        contentType = type;
        if (saveResponseData && (responseData == null)) {
            responseData = new StringBuffer();
            if (!type.toLowerCase().startsWith("text/")) {
                /*
                 * Since the data will not be text based, don't save
                 * a copy of it. Set the initial contents to be this
                 * explantion message.
                 */
                responseData.append("Unable to display data of type ");
                responseData.append(type);
                responseData.append(" (only text/* is shown).");
            }
        }
        response.setContentType(type);
    }

    /**
     * Setup for recording output.
     */
    private boolean outputRecordSetup() {
        /*
         * Is this the type of request where we want to save the output?
         */
        boolean isText = ((contentType != null) &&
                          (contentType.toLowerCase().startsWith("text/")));

        /*
         * Be sure there is a place to store the data, if needed.
         */
        if (saveResponseData && (responseData == null)) {
            responseData = new StringBuffer();
            if (!isText) {
                /*
                 * Since the data will not be text based, it won't
                 * be saved. Set the initial contents to be an
                 * explantion message.
                 */
                if (contentType != null) {
                    responseData.append("Unable to display data of type ");
                    responseData.append(contentType);
                    responseData.append(" (only text/* is shown).");
                } else {
                    /*
                     * They are supposed to call setContentType()
                     * before calling this function.
                     */
                    responseData.append("Mime type was not specified " +
                                        "before output began.");
                }
            }
        }
        return isText;
    }

    /**
     * Returns a wrapper around the real output stream. This wrapper
     * traces the calls to <CODE>print()</CODE>, <CODE>println()</CODE>,
     * and <CODE>write()</CODE>. Also, if <CODE>saveResponseData</CODE>
     * was true when passed to the
     * constructor, and <CODE>setContentType()</CODE> has been called
     * with a mime type that starts with "text/", then a copy of the actuall
     * data written to the output stream will be saved in this object's
     * responseData StringBuffer. The decision of whether or not to
     * save a copy of the data is made the first time this function is
     * called.
     *
     * @exception IOException See ServletResponse documentation
     * @see javax.servlet.ServletResponse#getOutputStream
     * @see org.enhydra.servlet.debug.OutputStreamRecord
     */
    public ServletOutputStream getOutputStream() throws IOException {
        ServletOutputStream os = response.getOutputStream();
        addTrace("getOutputStream() = " + os);

        /*
         * If this is the first call, create the OutputStreamRecord.
         */
        if (outputStreamRecord == null) {
            boolean isText = outputRecordSetup();
            outputStreamRecord = new OutputStreamRecord(this,
                                                        (saveResponseData && isText),
                                                        os);
        }

        /*
         * Return the wrapper.
         */
        return outputStreamRecord;
    }

    /**
     * Record call and pass on to actual response object..
     *
     * @see javax.servlet.http.HttpServletResponse#containsHeader
     */
    public boolean containsHeader(String name) {
        boolean ch = (headers.get(name.toLowerCase()) != null);
        addTrace("containsHeader(" + name + ") = " + ch);
        return ch;
    }

    /**
     * Record call and pass on to actual response object..
     *
     * @see javax.servlet.http.HttpServletResponse#setStatus
     */
    public void setStatus(int sc, String sm) {
        addTrace("setStatus(" + sc + ", " + sm + ")");
        statusCode = sc;
        statusMsg = sm;
        response.setStatus(sc, sm);
    }

    /**
     * Record call and pass on to actual response object.
     *
     * @see javax.servlet.http.HttpServletResponse#setStatus
     */
    public void setStatus(int sc) {
        addTrace("setStatus(" + sc + ")");
        statusCode = sc;
        statusMsg = getStatusString(sc);
        response.setStatus(sc);
    }

    /**
     * Record call and pass on to actual response object.
     *
     * @see javax.servlet.http.HttpServletResponse#setHeader
     */
    public void setHeader(String name, String value) {
        addTrace("setHeader(" + name + ", " + value + ")");
        processHeader(name, value);
        response.setHeader(name, value);
    }

    /**
     * Record call and pass on to actual response object.
     *
     * @see javax.servlet.http.HttpServletResponse#setIntHeader
     */
    public void setIntHeader(String name, int value) {
        addTrace("setIntHeader(" + name + ", " + value + ")");
        processHeader(name, value);
        response.setIntHeader(name, value);
    }

    /**
     * Record call and pass on to actual response object.
     *
     * @see javax.servlet.http.HttpServletResponse#setDateHeader
     */
    public void setDateHeader(String name, long date) {
        addTrace("setDateHeader(" + name + ", " + date + ")");
        headers.put(name.toLowerCase(), dateFormatter.format(new Date(date)));
        response.setDateHeader(name, date);
    }

    /**
     * Record call and pass on to actual response object.
     *
     * @exception IOException See HttpServletResponse documentation
     * @see javax.servlet.http.HttpServletResponse#sendError
     */
    public void sendError(int sc, String msg) throws IOException {
        addTrace("sendError(" + sc + ", " + msg + ")");
        statusCode = sc;
        statusMsg = msg;
        response.sendError(sc, msg);
    }

    /**
     * Record call and pass on to actual response object.
     *
     * @exception IOException See HttpServletResponse documentation
     * @see javax.servlet.http.HttpServletResponse#sendError
     */
    public void sendError(int sc) throws IOException {
        addTrace("sendError(" + sc + ")");
        statusCode = sc;
        statusMsg = getStatusString(sc);
        response.sendError(sc);
    }

    /**
     * Record call and pass on to actual response object.
     *
     * @see javax.servlet.http.HttpServletResponse#sendRedirect
     * @exception IOException See HttpServletResponse documentation
     */
    public void sendRedirect(String location) throws IOException {
        addTrace("sendRedirect(" + location + ")");
        statusCode = SC_OK;
        statusMsg = getStatusString(statusCode);
        headers.put("location", location);
        response.sendRedirect(location);
    }

    /**
     * Adds the specified cookie to the response. It can be called multiple
     * times to set more than one cookie.
     *
     * @param The <CODE>Cookie</CODE> to return to the client
     * @see javax.servlet.http.HttpServletResponse#addCookie
     */
    public void addCookie(Cookie cookie) {
        addTrace("addCookie(" +
                 HttpCookieParser.formatResponseCookie(cookie) + ")");
        cookies.addElement(cookie);
        response.addCookie(cookie);
    }

    /**
     * Returns the size of the response entity data, or -1 if not known.
     */
    public int getContentLength() {
        return contentLength;
    }

    /**
     * Returns the Internet Media Type of the response entity data, or null
     * if not known.
     */
    public String getContentType() {
        return contentType;
    }

    /**
     * Returns the status code for the response, or -1 if not set.
     */
    public int getStatusCode() {
        return statusCode;
    }

    /**
     * Returns the status message for the response.
     */
    public String getStatusMessage() {
        return statusMsg;
    }

    /**
     * Returns an array of cookies to be set by this response.
     *
     * @return The array of cookies found in this response.
     */
    public Cookie[] getCookies() {
        Cookie[] retCookies = new Cookie[cookies.size()];
        cookies.copyInto(retCookies);
        return retCookies;
    }

    /**
     * Returns the values of the specified header for the response as
     * an array of strings, or a 0 length array if the named header does
     * not exist.
     *
     * @param name the case-insensitive header field name
     */
    public String getHeader(String name) {
        return (String)headers.get(name.toLowerCase());
    }

    /**
     * Returns an array of strings representing the header names
     * for this response. This does not have return cookies - use
     * getCookies().
     */
    public Enumeration getHeaderNames() {
        return headers.keys();
    }

    /**
     * Returns the number of bytes associated with the response.
     */
    public int getTotalBytes() {
        addTrace("getTotalBytes()");
        return -1;
    }

    /**
     * Encodes the specified URL by including the session ID in it, or, if
     * encoding is not needed, returns the URL unchanged.
     *
     * @param url The url to be encoded.
     * @return The encoded URL if encoding is needed; the unchanged URL
     * otherwise.
     * @see javax.servlet.http.HttpServletResponse#encodeUrl
     */
    public String encodeUrl(String url) {
        String result = null;
        if (response != null)
           result = response.encodeUrl(url);
        addTrace("encodeUrl(" + url + ") = " + result);
        return result;
    }


    /**
     * Encodes the specified URL for use in the sendRedirect method or, if
     * encoding is not needed, returns the URL unchanged.
     *
     * @param url The url to be encoded.
     * @return The encoded URL if encoding is needed; the unchanged URL
     *  otherwise.
     * @see javax.servlet.http.HttpServletResponse#encodeRedirectUrl
     */
    public String encodeRedirectUrl(String url) {
        String result = null;
        if (response != null)
            result = response.encodeRedirectUrl(url);
        addTrace("encodeRedirectUrl(" + url + ") = " + result);
        return result;
    }

    /**
     * Returns a wrapper around the real PrintWriter. This wrapper
     * traces the calls to the output methods.
     * If <CODE>saveResponseData</CODE>
     * was true when passed to the
     * constructor, and <CODE>setContentType()</CODE> has been called
     * with a mime type that starts with "text/", then a copy of the actual
     * data written to the PprintWriter will be saved in this object's
     * responseData StringBuffer. The decision of whether or not to
     * save a copy of the data is made the first time this function is
     * called.
     *
     * @exception IOException See ServletResponse documentation
     * @see javax.servlet.ServletResponse#getPrintWriter
     * @see org.enhydra.servlet.debug.PrintWriterRecord
     */
    public PrintWriter getWriter() throws IOException {
        PrintWriter pw = null;
        if (response != null)
            pw = response.getWriter();
        addTrace("getWriter() = " + pw);

        /*
         * If this is the first call, create the PrintWriterRecord.
         */
        if (printWriterRecord == null) {
            boolean isText = outputRecordSetup();

            printWriterRecord = new PrintWriterRecord(this,
                                                      (saveResponseData && isText),
                                                      pw);
        }
        /*
         * Return the wrapper.
         */
        return printWriterRecord;
    }

    /**
     * Returns the character set encoding for the input of this request.
     *
     * @see javax.servlet.ServletRequest#getCharacterEncoding
     */
    public String getCharacterEncoding() {
        String result = response.getCharacterEncoding();
        addTrace("getCharacterEncoding() = " + result);
        return result;
    }

    /**
     * Returns the data written to the net, or null if this feature
     * was not enabled (see the constructor).
     */
    public String getResponseData() {
        if (saveResponseData && (responseData != null))
            return responseData.toString();
        else
            return null;
    }


    /**
     * Get the status string associated with a status code.
     *
     * @param sc The status code.
     * @return string describing this status code.
     */
    private static String getStatusString(int sc) {
        switch (sc) {
        case -1:
            return "<i>Not set</i>";
        case SC_ACCEPTED:
            return "Accepted";
        case SC_BAD_GATEWAY:
            return "Bad Gateway";
        case SC_BAD_REQUEST:
            return "Bad Request";
        case SC_CREATED:
            return "Created";
        case SC_FORBIDDEN:
            return "Forbidden";
        case SC_INTERNAL_SERVER_ERROR:
            return "Internal Server Error";
        case SC_MOVED_PERMANENTLY:
            return "Moved Permanently";
        case SC_MOVED_TEMPORARILY:
            return "Moved Temporarily";
        case SC_NO_CONTENT:
            return "No Content";
        case SC_NOT_FOUND:
            return "Not Found";
        case SC_NOT_IMPLEMENTED:
            return "Method Not Implemented";
        case SC_NOT_MODIFIED:
            return "Not Modified";
        case SC_OK:
            return "OK";
        case SC_SERVICE_UNAVAILABLE:
            return "Service Temporarily Unavailable";
        case SC_UNAUTHORIZED:
            return "Authorization Required";
        default:
            return "Response";
        }
    }

    public Locale getLocale() {
        addTrace("getLocale()");
        return response.getLocale();
    }

    public void flushBuffer() throws IOException {
        addTrace("flushBuffer()");
        response.flushBuffer();
    }

    public String encodeURL(String url) {
        addTrace("encodeURL(" + url + ")");
        return response.encodeURL(url);
    }

    public boolean isCommitted() {
        addTrace("isCommitted()");
        return response.isCommitted();
    }

    public int getBufferSize() {
        addTrace("getBufferSize()");
        return response.getBufferSize();
    }

    public void reset() {
        addTrace("reset()");
        response.reset();
    }

    public String encodeRedirectURL(String url) {
        addTrace("encodeRedirectURL(" + url + ")");
        return response.encodeRedirectURL(url);
    }

    public void addIntHeader(String name, int value) {
        addTrace("addIntHeader(" + name  + "," + value + ")");
        response.addIntHeader(name, value);
    }

    public void addDateHeader(String name, long value) {
        addTrace("addDateHeader(" + name + "," + value  + ")");
        response.addDateHeader(name, value);
    }

    public void addHeader(String name, String value) {
        addTrace("addHeader(" + name + "," + value + ")");
        processHeader(name, value);
        response.addHeader(name, value);

    }

    public void setBufferSize(int size) {
        addTrace("setBufferSize(" + size + ")");
        response.setBufferSize(size);
    }

    public void setLocale(Locale locale) {
        addTrace("setLocale(" + locale + ")");
        response.setLocale(locale);
    }

}
