/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: ServletRecordQueue.java,v 1.4.12.1 2000/10/19 17:59:10 jasona Exp $
 */

package org.enhydra.servlet.debug;

import java.util.Enumeration;


/*
 * Queue of transaction records.  Designed only for use with the DebugManager,
 * no synchronization is done here.
 * 
 * N.B. Currently finding an entry by id is done by sequentially searching
 * the list.  This was chosen over a hash table under the assumption that
 * additions are very frequent while find by id is infrequent.
 *
 * @version	$Revision: 1.4.12.1 $
 * @author	Mark Diekhans
 */
public class ServletRecordQueue {

    /**
     * Maximum size of queue.
     */
    private int maxSize;

    /**
     * Current size of queue.
     */
    private int currentSize = 0;

    /**
     * Queue head of transaction records.
     */
    private ServletTransactionRecord head = null;

    /**
     * Queue tail of transaction records.
     */
    private ServletTransactionRecord tail = null;

    /**
     * Sequential transaction id to assign.
     */
    private int nextTransactionId = 0;

    
    /**
     * Construct a new queue.
     *
     * @param queueSize maximum number of servlet transaction records that
     *  can be queued at a given time.
     */
    protected ServletRecordQueue(int queueSize) {
        maxSize = queueSize;
    }
  

    /**
     *  Get the current number of items in the queue.
     *
     *  @return
     *    The current number of records in the queue.
     */ 
    public int size() {
        return currentSize;
    }

 
    /**
     * Add a record to the queue; if this forces a record out, it is
     * returned. 
     *
     * @param transactionRecord the record to add to the queue.
     * @return the transaction record that was removed from the
     *  queue, or <CODE>null</CODE> if an entry was not removed.
     */
    protected synchronized ServletTransactionRecord addRecord(
            ServletTransactionRecord transactionRecord) {
        ServletTransactionRecord deleted;

        transactionRecord.id = nextTransactionId++;

        // Add to tail
        if (head == null) {
            head = transactionRecord;
        } else {
            tail.next = transactionRecord;
        }
        tail = transactionRecord;
        
        // Increment size or remove head.
        if (currentSize < maxSize) {
            currentSize++;
            deleted = null;
        } else {
            deleted = head;
            head = head.next;
        }
        transactionRecord.next = null;
        return deleted;
    }

    /**
     * Clear all entries from the queue.
     */
    protected synchronized void clear() {
        currentSize = 0;
        head = null;
        tail = null;
    }

    /**
     * Find an entry in the queue based on tranaction record id.
     * 
     * @param transactionId unique number identifying the desired transaction
     *  record.
     * @return a transaction record or <CODE>null</CODE> if the transaction
     *  is no longer in the queue.
     */
    protected synchronized ServletTransactionRecord
	findById(int transactionId) {

        ServletTransactionRecord scan;
        for (scan = head; scan != null; scan = scan.next) {
            if (scan.getId() == transactionId) {
                return scan;
            }
        }
        return null;
    }

    /**
     * FIX: I need javadoc
     *
     * @param callback
     *   FIX: ???
     */
    protected synchronized void dumpContents(ServletRecordCallback callback) {
        ServletTransactionRecord scan;
        for (scan = head; scan != null; scan = scan.next)
            callback.transactionRecordAddEvent(scan, null);
    }


    /**
     *  Get the current contents of the queue. 
     *
     *  @return
     *    An Enumeration that lists the current contents of the queue.
     */
    public synchronized Enumeration getContents() {
        return new QueueEnumerator(head);
    }



    /**
     *  This is a helper class. When asked for the contents of the queue
     *  it simply walks down the linked list of queue entries.
     */
    private class QueueEnumerator implements Enumeration {
        private ServletTransactionRecord current;
        QueueEnumerator(ServletTransactionRecord head) {
            current = head;
        }
        public boolean hasMoreElements() {
            return (current != null);
        }
        public Object nextElement() {
            ServletTransactionRecord result = current;
            if (current != null)
                current = current.next;
            return result;
        }
    }


}

