/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: ServletTransactionRecord.java,v 1.4.12.1 2000/10/19 17:59:10 jasona Exp $
 */

package org.enhydra.servlet.debug;

/**
 * Transaction record that describes a single servlet transaction.
 *
 * @version	$Revision: 1.4.12.1 $
 * @author	Mark Diekhans
 */
public class ServletTransactionRecord {

    /**
     * Linked used for managing the record queue.
     */
    protected ServletTransactionRecord next = null;

    /**
     * Unique id of this transaction.
     */
    protected int id = -1;

    /**
     * The name of the servlet that was sent this request.
     */
    private String servletName;

    /**
     * Object used to record the request.
     */
    private RequestRecord requestRecord;

    /**
     * Object user to record the response.
     */
    private ResponseRecord responseRecord;

    /**
     * The exception or error that the servlet threw.
     */
    private Throwable error;

    /**
     * Trace string.
     */
    private StringBuffer trace = new StringBuffer();

    /**
     * An HTML representaion of the session data before the request
     * was processed.
     */
    private String preSession;

    /**
     * An HTML representaion of the session data after the request
     * was processed.
     */
    private String postSession;

    /**
     * Is tracing enabled?
     */
    private boolean traceEnabled = true;

    /**
     * Construct a new servlet transaction record.
     *
     * @param requestRecord object used to record the request.
     * @param responseRecord object user to record the response.	
     */
    protected ServletTransactionRecord(String servletName,
                                       RequestRecord requestRecord,
                                       ResponseRecord responseRecord) {
        this.servletName = servletName;
        this.requestRecord = requestRecord;
        this.responseRecord = responseRecord;
    }

    /**
     * Set the exception or error that the servlet threw.
     */
    protected void setError(Throwable error) {
        this.error = error;
    }

    /**
     * Get the id of this object  Id is not set until the object is added to
     * the queue.
     *
     * @return the transaction record id or a value less than zero if the
     *  record has not be added to the queue.
     */
    public int getId() {
        return id;
    }

    /**
     * Get the name of the servlet that was sent this request.
     *
     * @return the name of this transaction's servlet.
     */
    public String getServletName() {
        return servletName;
    }

    /**
     * Get request record.
     *
     * @return the request record object.
     */
    public RequestRecord getRequestRecord() {
        return requestRecord;
    }

    /**
     * Get response record.
     *
     * @return the response record object.
     */
    public ResponseRecord getResponseRecord() {
        return responseRecord;
    }

    /**
     * Add a string to the trace log.
     *
     * @param str string to add to the trace.
     */
    protected void addTrace(String str) {
    // FIXME:  This should be more structured.
        if (traceEnabled) {
            trace.append(str);
            trace.append("\n");
        }
    }

    /**
     * Stop tracing. Calls to <CODE>addTrace()</CODE> will be ignored.
     */
    protected void stopTracing() {
        traceEnabled = false;
    }


    /**
     * Get the trace log.
     *
     * @return trace string of newline seperated entries.
     */
    public String getTrace() {
        return trace.toString();
    }



    /**
     *  Save a string (html) description of the session data before the
     *  request was processed. This call is optional, as session data
     *  only applies when the servlet being debugged is an Enhydra servlet.
     * 
     *  @param desc
     *    A string description of the session data. May contain HTML tags.
     */
    public void setPreSessionString(String desc) {
        preSession = desc;
    }



    /**
     *  Get the string description of the session data before the request
     *  was processed. This is optional data, if setPreSessionString() was
     *  never called, the result will be null. 
     *
     *  @return
     *    A string (possibly with HTML tags) describing the session data
     *    before the request, or null if never set or the servlet is not
     *    an Enhydra servlet.
     */
    public String getPreSessionString() {
        return preSession;
    }



    /**
     *  Save a string (html) description of the session data after the
     *  request was processed. This call is optional, as session data
     *  only applies when the servlet being debugged is an Enhydra servlet.
     * 
     *  @param desc
     *    A string description of the session data. May contain HTML tags.
     */
    public void setPostSessionString(String desc) {
        postSession = desc;
    }



    /**
     *  Get the string description of the session data after the request
     *  was processed. This is optional data, if setPostSessionString() was
     *  never called, the result will be null. 
     *
     *  @return
     *    A string (possibly with HTML tags) describing the session data
     *    after the request, or null if never set or the servlet is not
     *    an Enhydra servlet.
     */
    public String getPostSessionString() {
       return postSession;
    }


}
