/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: FilterManager.java,v 1.4.12.1 2000/10/19 17:59:10 jasona Exp $
 */

package org.enhydra.servlet.filter;
 
import java.util.Hashtable;
import java.util.Enumeration;


/**
 *  This class maintains a set of Filters. Each filter
 *  is associated with an identification string (symbolic name). 
 *  This ID string is used to retrieve a filter. <P>
 *
 *  This class does not monitor the creation of filters. 
 *  If you create one, you are responsible for adding it to this class. 
 *  These public methods are the only things that modify the set of 
 *  filters managed by this class. 
 *
 *  @see org.enhydra.servlet.filter.Filter
 *  @author Andy John
 *  @author Shawn McMurdo
 */
public class FilterManager {

    private Hashtable filters;
    private Hashtable descriptions;

    /**
     *    Create a new, empty, FilterManager.
     */
    public FilterManager () {
        filters = new Hashtable();
        descriptions = new Hashtable();
    }



    /**
     *    Add a Filter to the set. It will be referred to by the
     *   given identifier string. If a filter already being stored has the
     *   same filterID, the new filter will replace the old filter. 
     *
     *   @param filterID The name to use to refer to the filter.
     *   @param filter The Filter to store.
     *   @param description A human readable description of this filter.
     */
    public synchronized void add(String filterID, Filter filter,
	    String description) {

        filters.put(filterID, filter);
        descriptions.put(filterID, description);
    }



    /**
     *   Look up a Filter by its identifier string.
     *
     *   @param filterID The symbolic name to look up.
     *   @return The Filter, or null if the name is not found.
     */
    public synchronized Filter get(String filterID) {
        return (Filter) filters.get(filterID);
    }       



    /**
     *   Look up a filter's description by its identifier string.
     *
     *   @param filterID The symbolic name to look up.
     *   @return The human readable description,
     *   or null if the name is not found.
     */
    public synchronized String getDescription(String filterID) {
        return (String) descriptions.get(filterID);
    }       



    /**
     * Get all the names of the Filters currently being stored.
     *
     * @return All the filterIDs
     */
    public synchronized String[] getIDs() {
        synchronized (filters) {
            int num = filters.size();
            String[] results = new String[num];
            Enumeration e = filters.keys();
            for (int i=0; i<num; i++) 
                results[i] = (String) e.nextElement();
            return results;
        }
    }



    /**
     *  Remove a Filter from the set. Silently ignores errors.
     *
     * @param filterID The name of the Filter to remove.
     */
    public synchronized void delete(String filterID) {
        filters.remove(filterID);
        descriptions.remove(filterID);
    }


    /**
     *  Generate a unique filter id. Given the base name, a number is 
     *  appended onto the end. The resulting string is guarenteed to
     *  not currently exist as a filter id. If the base name is not
     *  already being used, it is returned unchanged.
     *
     * @param baseName The initial name to use.
     * @return A filter ID not currently in use.
     */
    public synchronized String getUniqueID(String baseName) {
        if ((baseName == null) || (baseName.equals("")))
            baseName = "filter";
        String[] ids = getIDs();
        String name = baseName;
        int counter = 0;
        while (true) {
            boolean found = false;
            for (int i=0; i<ids.length; i++) {
                if (ids[i].equals(name)) {
                    found = true;
                    break;
                }
            }
            if (!found)
                return baseName;
            name = baseName + "_" + counter++;
        }
    }
             
}


