/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: ResponseFilter.java,v 1.4.2.3.2.1 2000/10/19 17:59:10 jasona Exp $
 */

package org.enhydra.servlet.filter;

import java.io.IOException;
import java.io.PrintWriter;
import java.util.Locale;
import java.util.Enumeration;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.Cookie;
import javax.servlet.ServletOutputStream;
import org.apache.tomcat.core.ResponseAdapter;
import org.apache.tomcat.util.*;

/**
 * Base class that is the base class for a filter, allowing it to be
 * accessed as a ResponseAdapter.
 *
 *  @author Andy John
 *  @author Shawn McMurdo
 */
abstract public class ResponseFilter implements HttpServletResponse, ResponseAdapter {
    /**
     * Output stream, saved on first access.
     */
    ServletOutputStream outputStream = null;

    /**
     *  The real HttpServletResponse that all calls will be passed on to.
     */
    protected HttpServletResponse myResponse;

    /**
     *  Constructor. You must pass in a HttpServletResponse object. This is
     *  the object that the calls will be passed on to.
     *
     *  @param originalResponse The real HttpServletResponse to route the calls
     *  to.
     *  @see javax.servlet.ServletResponse
     *  @see javax.servlet.http.HttpServletResponse
     */
    public ResponseFilter(HttpServletResponse originalResponse) {
        myResponse = originalResponse;
    }


    /**
     *  This call is passed on to the original HttpServletResponse object.
     *  For a description of what this call does, see the ServletResponse
     *  documentation.
     *
     *  @see javax.servlet.ServletResponse
     */
    public void setContentLength (int len) {
        myResponse.setContentLength(len);
    }


    /**
     *  This call is passed on to the original HttpServletResponse object.
     *  For a description of what this call does, see the ServletResponse
     *  documentation.
     *
     *  @see javax.servlet.ServletResponse
     */
    public void setContentType (String type) {
        myResponse.setContentType(type);
    }


    /**
     *  This call is passed on to the original HttpServletResponse object.
     *  For a description of what this call does, see the ServletResponse
     *  documentation.
     *
     *  @exception IOException See ServletResponse documentation
     *  @see javax.servlet.ServletResponse
     */
    public ServletOutputStream getOutputStream () throws IOException {
        return myResponse.getOutputStream();
    }

    //----------------------------------------------------------------


    /**
     *  This call is passed on to the original HttpServletResponse object.
     *  For a description of what this call does, see the HttpServletResponse
     *  documentation.
     *
     *  @see javax.servlet.http.HttpServletResponse
     */
    public boolean containsHeader(String name) {
        return myResponse.containsHeader(name);
    }


    /**
     *  This call is passed on to the original HttpServletResponse object.
     *  For a description of what this call does, see the HttpServletResponse
     *  documentation.
     *
     *  @see javax.servlet.http.HttpServletResponse
     */
    public void setStatus(int sc, String sm) {
        myResponse.setStatus(sc, sm);
    }


    /**
     *  This call is passed on to the original HttpServletResponse object.
     *  For a description of what this call does, see the HttpServletResponse
     *  documentation.
     *
     *  @see javax.servlet.http.HttpServletResponse
     */
    public void setStatus(int sc) {
        myResponse.setStatus(sc);
    }


    /**
     *  This call is passed on to the original HttpServletResponse object.
     *  For a description of what this call does, see the HttpServletResponse
     *  documentation.
     *
     *  @see javax.servlet.http.HttpServletResponse
     */
    public void setHeader(String name, String value) {
        myResponse.setHeader(name, value);
    }


    /**
     *  This call is passed on to the original HttpServletResponse object.
     *  For a description of what this call does, see the HttpServletResponse
     *  documentation.
     *
     *  @see javax.servlet.http.HttpServletResponse
     */
    public void setIntHeader(String name, int value) {
        myResponse.setIntHeader(name, value);
    }


    /**
     *  This call is passed on to the original HttpServletResponse object.
     *  For a description of what this call does, see the HttpServletResponse
     *  documentation.
     *
     *  @see javax.servlet.http.HttpServletResponse
     */
    public void setDateHeader(String name, long date) {
        myResponse.setDateHeader(name, date);
    }


    /**
     *  This call is passed on to the original HttpServletResponse object.
     *  For a description of what this call does, see the HttpServletResponse
     *  documentation.
     *
     *  @exception IOException See HttpServletResponse documentation
     *  @see javax.servlet.http.HttpServletResponse
     */
    public void sendError(int sc, String msg) throws IOException {
        myResponse.sendError(sc, msg);
    }


    /**
     *  This call is passed on to the original HttpServletResponse object.
     *  For a description of what this call does, see the HttpServletResponse
     *  documentation.
     *
     *  @exception IOException See HttpServletResponse documentation
     *  @see javax.servlet.http.HttpServletResponse
     */
    public void sendError(int sc) throws IOException {
        myResponse.sendError(sc);
    }


    /**
     *  This call is passed on to the original HttpServletResponse object.
     *  For a description of what this call does, see the HttpServletResponse
     *  documentation.
     *
     *  @see javax.servlet.http.HttpServletResponse
     *  @exception IOException See HttpServletResponse documentation
     */
    public void sendRedirect(String location) throws IOException {
        myResponse.sendRedirect(location);
    }

    /**
     * This call is passed on to the original HttpServletResponse object.
     *
     * @see javax.servlet.http.HttpServletResponse#addCookie
     */
    public void addCookie(Cookie cookie) {
        myResponse.addCookie(cookie);
    }

    /**
     * This call is passed on to the original HttpServletResponse object.
     *
     * @see javax.servlet.http.HttpServletResponse#encodeUrl
     */
    public String encodeUrl(String url) {
        return myResponse.encodeUrl(url);
    }

    /**
     * This call is passed on to the original HttpServletResponse object.
     *
     * @see javax.servlet.http.HttpServletResponse#encodeRedirectUrl
     */
    public String encodeRedirectUrl(String url) {
        return myResponse.encodeRedirectUrl(url);
    }

    /**
     * This call is passed on to the original HttpServletResponse object.
     *
     * @see javax.servlet.http.HttpServletResponse#getWriter
     * @exception  IOException
     * On other errors.
     */
    public PrintWriter getWriter() throws IOException {
        return myResponse.getWriter();
    }

    /**
     * This call is passed on to the original HttpServletResponse object.
     *
     * @see javax.servlet.http.HttpServletResponse#getCharacterEncoding
     */
    public String getCharacterEncoding() {
        return myResponse.getCharacterEncoding();
    }


    public Locale getLocale() {
	// X22
	return myResponse.getLocale();
    }

    public void flushBuffer() throws IOException {
	// X22
	myResponse.flushBuffer();
    }

    public String encodeURL(String url) {
	// X22
	return myResponse.encodeURL(url);
    }

    public boolean isCommitted() {
	// X22
	return myResponse.isCommitted();
    }

    public int getBufferSize() {
	// X22
	return myResponse.getBufferSize();
    }

    public void reset() {
	// X22
	myResponse.reset();
    }

    public String encodeRedirectURL(String url) {
	// X22
	return myResponse.encodeRedirectURL(url);
    }

    public void addIntHeader(String name, int value) {
	// X22
	myResponse.addIntHeader(name, value);
    }

    public void addDateHeader(String name, long value) {
	// X22
	myResponse.addDateHeader(name, value);
    }

    public void addHeader(String name, String value) {
	// X22
	myResponse.addHeader(name, value);
    }

    public void setBufferSize(int size) {
	// X22
	myResponse.setBufferSize(size);
    }

    public void setLocale(Locale locale) {
	// X22
	myResponse.setLocale(locale);
    }

    // X22 ResponseAdapter
    // XXX This one or multiple addHeader?
    // Probably not a big deal - but an adapter may have
    // an optimized version for this one ( one round-trip only )
    public void addMimeHeaders(MimeHeaders headers) throws IOException {
	int size = headers.size();
        for (int i = 0; i < size; i++) {
            MimeHeaderField h = headers.getField(i);
            addHeader(h.getName(), h.getValue());
        }
    }

    // X22 ResponseAdapter
    /** Signal that we're done with a particular request, the
	server can go on and read more requests or close the socket
    */
    public void endResponse() throws IOException {
    }

    // X22 ResponseAdapter
    /** Signal that we're done with the headers, and body will follow.
	The adapter doesn't have to maintain state, it's done inside the engine
    */
    public void endHeaders() throws IOException {
    }

    // X22 ResponseAdapter
    public ServletOutputStream getServletOutputStream() throws IOException {
        if (outputStream == null) {
            outputStream = getOutputStream();
        }
        return outputStream;
    }

    // X22 ResponseAdapter
    public void recycle() {
    }

    /** Write a chunk of bytes. Should be called only from ServletOutputStream implementations,
     *	No need to implement it if your adapter implements ServletOutputStream.
     *  Headers and status will be written before this method is executed.
     */
    public void doWrite(byte[] buffer, int pos, int count) throws IOException {
        if (outputStream == null) {
            outputStream = getOutputStream();
        }
        outputStream.write(buffer, pos, count);
    }
}

