/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: StandardLoggingServiceInterceptor.java,v 1.4.2.2.2.1 2000/10/19 17:59:10 jasona Exp $
 */

package org.enhydra.servlet.filter;

import java.io.*;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.text.FieldPosition;
import java.util.Date;
import java.util.TimeZone;
import javax.servlet.Servlet;
import javax.servlet.ServletException;
import javax.servlet.ServletConfig;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import com.lutris.logging.LogChannel;
import org.apache.tomcat.core.ServiceInterceptor;
import org.apache.tomcat.core.Context;

/**
 *   This is part of the BasicTransactionFilter. This is a filter that
 *   logs information in a standard format so that tools like Web
 *   Analyzer can easily parse the data.  Every call is passed through
 *   to the real Servlet.
 *
 *   @see org.enhydra.servlet.filter.StandardLoggingFilter
 *   @see org.enhydra.servlet.filter.StandardLoggingResponse
 *   @author Andy John
 *   @author Kent Henneuse
 *   @author Shawn McMurdo
 */
public class StandardLoggingServiceInterceptor implements ServiceInterceptor {

    // Data from the constructor.
    private String logFile;

    // Data that is used for Multiserver Logging.
    private LogChannel logChannel;
    private int logLevel;
    private long transactionId;

    // Variables needed to make the date formatting nice
    private SimpleDateFormat dateFmt;
    private FieldPosition fp;

    /**
     *  Create a new StandardLoggingServiceInterceptor.
     *
     *  @param logFile The log file for logging service() calls.
     */
    public StandardLoggingServiceInterceptor(String logFile) {

	this.logFile = logFile;
	dateFmt = new SimpleDateFormat("[dd/MMM/yyyy:HH:mm:ss ");
	fp = new FieldPosition(DateFormat.DAY_OF_WEEK_FIELD);
    }


    /**
     *  Service a request.
     *  This will be passed through to the real Servlet.
     *  After the call returns, <CODE>logData()</CODE> will be called on
     *  the  StandardLoggingResponse passed in to the constructor.
     *
     *  @param req The request object.
     *  @param res The response object.
     */
    public void preInvoke(Context context, Servlet servlet,
	    HttpServletRequest req, HttpServletResponse res) {

	// Log the information about the request
	if (logChannel != null && logChannel.isEnabled(logLevel)) {
	    logRequestToLBSLog(req);
 	}
    }

    public void postInvoke(Context context, Servlet servlet,
	    HttpServletRequest req, HttpServletResponse res) {

	StandardLoggingResponse slr = (StandardLoggingResponse)
		req.getAttribute(
		"Enhydra.StandardLoggingFilter.standardLoggingResponse");

	// Log the information about the response
	if (logChannel != null && logChannel.isEnabled(logLevel)) {
	    if (slr != null) {
	        logResponseToLBSLog(slr);
	    }
	}

	// Log request/response pair to an access log
	if (logFile != null) {
	    try {
		if (slr != null) {
	            logAccesses(req, slr);
		}
	    } catch (IOException ioe) {
		// don't care
	    }
	}
    }

    /**
     * Set the information that will be needed to make logging to the
     * Multiserver log possible.
     *
     * @param logChannel the Log Channel of the Multiserver log
     * @param logLevel the Log Level of the log option REQUEST
     * @param id the transactionId of the request/response pair
     */
    public void setLBSLogInformation(LogChannel logChannel, int logLevel,
				     long id) {
	this.logChannel = logChannel;
	this.logLevel = logLevel;
	this.transactionId = id;
    }

    /**
     * Log data about the request to Multiserver
     *
     * @param req the original request that was sent to the server.
     */
    private void logRequestToLBSLog(HttpServletRequest req) {
	String data = transactionId + " " + req.getRemoteHost() + ": " +
	    req.getMethod() + " " + req.getRequestURI();
	String q = req.getQueryString();
	if (q != null)
	    data += "?" + q;
	logChannel.write(logLevel, data);
    }

    private void logResponseToLBSLog(StandardLoggingResponse res) {
	StringBuffer data = new StringBuffer();
	data.append(transactionId);
	data.append(" Response: code = ");
	data.append(res.getStatus());
	if (res.getContentType() != null) {
	    data.append(", type = ");
	    data.append(res.getContentType());
	}
	if (res.wasContentLengthSet()) {
	    data.append(", content length = ");
	    data.append(res.getContentLength());
	}
	if (res.getAction() != null) {
	    data.append(", ");
	    data.append(res.getAction());
	}
	logChannel.write(logLevel, data.toString());
    }

    /**
     * Logs the information about both the request and response to an access
     * log file.
     *
     * @param request the HTTP request that was issued to the server
     * @param response the HTTP response that will be sent back after the data
     *    has been loged.
     */
    private void logAccesses(HttpServletRequest request,
	    StandardLoggingResponse response) throws IOException {

	FileWriter outFile = new FileWriter(logFile, true);
	BufferedWriter out = new BufferedWriter(outFile);
	Date now;
        StringBuffer data = new StringBuffer();
	String holdStr;
	String noData = "- ";
        char space = ' ';

	holdStr = request.getRemoteAddr();
	if (holdStr == null)
	    data.append(noData);
	else {
	    data.append(holdStr);
	    data.append(space);
	}

	holdStr = request.getServerName();
	if (holdStr == null)
	    data.append(noData);
	else {
	    data.append(holdStr);
	    data.append(space);
	}

	holdStr = request.getRemoteUser();
	if (holdStr == null)
	    data.append(noData);
	else {
	    data.append(holdStr);
	    data.append(space);
	}

	now = new Date();
	dateFmt.format(now, data, fp);
	data.append(findGMTOffset());
	data.append("] ");

	holdStr = request.getMethod();
	data.append("\"" + holdStr + " /");
	holdStr = request.getPathInfo();
	if (holdStr != null)
	    data.append(holdStr);
	holdStr = request.getProtocol();
	data.append(" " + holdStr + "\" ");

	holdStr = Integer.toString(response.getStatus());
	data.append(holdStr);
	data.append(space);

	if (response.wasContentLengthSet() == false)
	    data.append(noData);
	else {
	    data.append(response.getContentLength());
	    data.append(space);
	}

	holdStr = request.getHeader("Referer");
	if (holdStr != null) {
	    data.append(holdStr);
	    data.append(space);
	}
	else
	    data.append(noData);

	holdStr = request.getHeader("User-Agent");
	if (holdStr != null) {
	    data.append(holdStr);
	    data.append(space);
	}
	else
	    data.append(noData);

	holdStr = data.toString();
	out.write(holdStr, 0, holdStr.length());
	out.newLine();
	out.close();
	outFile.close();
    }

    private String findGMTOffset() {
	TimeZone tz = TimeZone.getDefault();
	int rawOffset = tz.getRawOffset()/ (10 * 60 * 60);
	String offset;

	if (rawOffset < 0) {
	    rawOffset *= -1;
	    if (rawOffset < 1000)
		offset = "-0" + rawOffset;
	    else
		offset = Integer.toString(-1 * rawOffset);
	} else {
	    if (rawOffset < 1000)
		offset = "+0" + rawOffset;
	    else
		offset = "+" + rawOffset;
	}
	return offset;
    }
}

