/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 * $Id: ServletManager.java,v 1.18.2.18.2.2 2000/12/12 01:35:13 peterj Exp $
 */

package org.enhydra.servlet.servletManager;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.Date;
import java.util.Vector;

import java.net.*;

import javax.servlet.Servlet;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import com.lutris.logging.Logger;
import com.lutris.logging.LogChannel;
import com.lutris.http.MimeType;
import com.lutris.classloader.MultiClassLoader;
import com.lutris.util.Config;
import com.lutris.util.ExceptionUtils;
import com.lutris.appserver.server.httpPresentation.servlet.HttpPresentationServlet;
import org.enhydra.servlet.connectionMethods.*;
import com.lutris.multiServer.MultiServer;
import org.enhydra.servlet.ServletContainer;
import org.enhydra.servlet.filter.FilterManager;
import org.enhydra.servlet.filter.Filter;
import com.lutris.appserver.server.session.*;
import org.enhydra.servlet.tomcat.*;
import org.enhydra.i18n.ResManager;

import org.apache.tomcat.core.*;
import org.apache.tomcat.util.URLUtil;


/**
 *   This class loads Servlet classes, creates instances of Servlets,
 *  caches the Servlets, and returns pointers to them. The Servlets
 *  are refered to by an identifer string (a symbolic name). <P>
 *
 *  First Servlets are registered via <CODE>add()</CODE>. Then they are
 *  started via <CODE>start()</CODE>. Then they are available for
 *  use via <CODE>get()</CODE>. <P>
 *
 *  The ServletManager will call <CODE>init()</CODE> on the Servlet.
 *  It will also call <CODE>destoroy()</CODE>.
 *  Other classes which use Servlets returned by <CODE>get()</CODE> must
 *  not call these two methods. When <CODE>get()</CODE> returns a Servlet,
 *  it is ready to accept <CODE>service()</CODE> calls. <P>
 *
 *  @see javax.servlet.Servlet
 *  @see org.enhydra.servlet.servletManager.ServletStatus
 *  @author Andy John
 */
public class ServletManager implements Runnable {


    /**
     * Attribute key to store session manager in context.
     */

    public static final String  SESSION_MANAGER_KEY = "org.enhydra.session.manager";

    /*
     *  How much statistics to store.
     */
    static private final int RPMBufferSize = 60 * 24; // One day's worth.

    /*
     *  The server named exposed to the user
     */
    static private final String SERVER_INFO_DISPLAY = "Enhydra Multiserver";

    /*
     * The mapping of contexts stored by servletId and the urp prefix
     */
    private Hashtable contextByKey = new Hashtable();

    /*
     * Mapping of servlet id to filter ids.
     */
    private Hashtable filters = new Hashtable();

    // All the data about a particular Servlet.
    class TableEntry {
        String          className;      // Name of Servlet class to instantiate
        String[]        classPath;      // Diretories, .zip, or .jar files.
        String          docRoot;        // Where it's files on disk are.
        Properties      initArgs;       // Servlet initialization string
        String          description;    // Human readable, for admin app
        int             refCount;       // How many times it was asked for.
        LogChannel      logChannel;     // The Servlet logs to this.
        Servlet         servlet;        // The actual Servlet, or null.
        MultiClassLoader classLoader;   // The loader for this Servlet.
        Date            createTime;     // When it was instantiated, or null.

        int             RPM_counter;    // How gets in current minute.
        long            RPM_startTime;  // When current minute started.
        double          RPM_current;    // The req/min for last minute.
        double          RPM_max;        // The maximum req/min so far.
        Date            RPM_maxTime;    // When the max req/min happened.
        /*
         *  Keep a circular buffer of RPM values. The current series is
         *  stored in the range historyStart..(historyStart+historyNum-1).
         *  The next free slot to add to is historyStart+historyNum.
         *  Since this is a circular buffer, remember to wrap around to the
         *  beginning if you go off the end of the array.
         */
        double          historyRPM[];   // The request-per-minute values.
        long            historyWhen[];  // The time when each value applied.
        int             historyStart;   // The start of the series.
        int             historyNum;     // The number of stats int the buffer.

        boolean         isWar;          // The context is a war
        Vector          paths;          // The url prefix's of the context.

        boolean         started;        // The context is Running
        boolean         runOnInit;      // The context should be started at init time.

        // The following are used only by a WAR

        String          defaultSessionTimeout;
        String          isWARExpanded;
        String          isWARValidated;
        String          isInvokerEnabled;
        String          isWorkDirPersistent;

        // used to record the config info for the session manager.
        Config          sessionConfig;
    }


    //
    class SingleContextServer implements Server {

        SingleContextServer() {
        }

        private Context context;

        void setContext(Context context) {
            this.context = context;
        }

        // The server interface can find a context by path
        // Tomcat assumes that the context manager will only have
        // contexts with unique url prefixes. If Enhydra has multiple
        // channels with the same url this function can fail to find the
        // expected context. However if the servlet is asking for a url
        // from withing the same context it will work.
        //
        // N.B. this function will not always return the expected context.
        //
        public Context getContextByPath(String startPath) {
            Context foundContext = null;
            if (startPath.equals("/")) {
                startPath = "";
            }
            String path = startPath;
            // first try the current context.
            boolean done = false;
            while (!done && foundContext == null) {
                if (context.getPath().equals(path)) {
                    foundContext = context;
                }
                else {
                    // didn't find it strip off the end and start again.
                    int i = path.lastIndexOf('/');
                    if (i > -1 ) {
                        path = path.substring(0,i);
                    }
                    done = (path.length() <= 1) || (i == -1);
                }
            }
            if (foundContext == null) {
                // This is the part that may not return the right context
                // if there is more than one that match the url.
                // So warn the user!

                String resKey = "Looking for context by path \"{0}\"";
                String msg = rez.format(resKey,startPath);

                msg = rez.format("Looking for context by path \"{0}\"",startPath);
                context.getLogChannel().write(Logger.WARNING, msg);
                foundContext = lookuptContextByPath(startPath);
                if (foundContext == null) {
                    resKey = "No Context found that matches path \"{0}\"";
                    msg = rez.format(resKey,startPath);
                    context.getLogChannel().write(Logger.WARNING, msg);
                }
                else {
                    // found a context but warn that it may not be correct.
                    resKey = "Found context path is \"{0}\"";
                    msg = rez.format(resKey,context.getPath());
                    context.getLogChannel().write(Logger.WARNING, msg);
                }
            }
            return foundContext;
        }


        // The serverinfo string passed down to the ServletContext
        // This string is visible to the servlet via the ServletContext
        public String getServerInfo() {
            return getServerInfoString();
        }

    }

    // The logging channel for the ServletMananger write to.
    private LogChannel logChannel;

    // The set of all Servlets contained by this object.
    private Hashtable myTable;

    // This is used by the ServletContext object indirectly passed to
    // the Servlet's init() method.
    private MimeType myMime = new MimeType();

    // The thread that monitors the request per minute stats every
    // sixty seconds.
    private  Thread RPM_thread;

    // A flag used to tell the above thread when to die.
    private boolean RPM_stayAlive;


    // some default values taken from the context
    private static String defaultSessionTimeOutDefault;
    private static String isWARExpandedDefault;
    private static String isWARValidatedDefault;
    private static String isInvokerEnabledDefault;
    private static String isWorkDirPersistentDefault;

    // The directory that the jsp work files will be built in.
    private String workDir;

    public void setWorkDir(String workDir) {
        this.workDir = workDir;
        String msg = rez.format("Work dir base set to \"{0}\"",workDir);
        logChannel.write(Logger.DEBUG, msg);
    }

    //
    ResManager rez;


    /**
     *    Create a new ServletManager, initially containing no Servlets.
     */
    public ServletManager () {

        myTable = new Hashtable();
        /*
         *    Where to log messages to.
         */
        logChannel = Logger.getCentralLogger().getChannel("Multiserver");
        /*
         *   Start up the monitoring thread. It will do a little work
         *   once a minute.
         */
        RPM_thread = new Thread(this);
        RPM_thread.setDaemon(true); // VM won't wait to die.
        RPM_stayAlive = true;
        RPM_thread.start();

        // Find the default values for these settings
        // They are set within the context and might not be specified in the
        // conf file.
        //

        contextManager = new ContextManager();

	String docRoot = "/";
	URL docRootURL = getURLFromPath(docRoot);

	defaultContext = contextManager.addContext(docRoot, docRootURL);
	contextManager.setDefaultContext(defaultContext);
        defaultContext.setLogChannel(logChannel);
	//defaultContext.init();

        defaultSessionTimeOutDefault = "" + defaultContext.getSessionTimeOut();
        isWARExpandedDefault         = "" + defaultContext.isWARExpanded();
        isWARValidatedDefault        = "" + defaultContext.isWARValidated();
        isInvokerEnabledDefault      = "" + defaultContext.isInvokerEnabled();
        isWorkDirPersistentDefault   = "" + defaultContext.isWorkDirPersistent();

        // Gets a string  manager wrapped around org.enhydra.servlet.Res
        rez = ResManager.getResManager(this.getClass());

    }


    /**
     * The Tomcat ContextManager used to handle servlet requests.
     */
     private ContextManager contextManager;

    /**
     * The default context
     */

     private Context defaultContext;

    /**
     * Shut down the ServletManager. This stops all the Servlets, then
     * deletes all the Servlets, then kills the statistics gathering
     * thread. There is no way to restart the thread.
     */
    public synchronized void destroy() {
        /*
         * First stop all the Servlets. This will call their
         * destroy() methods, and give them a chance to shut down
         * gracefully.
         */
        Enumeration e = myTable.keys();
        while (e.hasMoreElements()) {
            String id = (String) e.nextElement();
            try {
                TableEntry t = (TableEntry) myTable.get(id);
                if (t != null && t.started) {
                   stop(id);
                }
            } catch (ServletException se) {
                String msg = rez.format("Failed to stop {0}.",id);
                logChannel.write(Logger.DEBUG, msg, se);
            }
        }
        /*
         *  Now delete all the Servlets. This will remove them from the
         *  table, and prevent them from being restarted.
         */
        e = myTable.keys();
        while (e.hasMoreElements()) {
            String id = (String) e.nextElement();
            try {
                delete(id);
            } catch (ServletException se) {
                String msg = rez.format("Failed to delete {0}.",  id);
                logChannel.write(Logger.DEBUG, msg, se);
            }
        }
        /*
         *  Now stop the request per minute monitoring thread.
         *  If we leave this thread alive, the virtual machine will never
         *  exit.
         */
        RPM_stayAlive = false;
        RPM_thread.interrupt();
    }


    /**
     *  This function is for internal use only. Do not call this method.
     *  When the request per minute monitoring thread starts, this is
     *  the code it starts executing. The thread sleeps, waking up once
     *  a minute to compute the current request per minute stats, and the
     *  maximum request per minute stats.
     */
    public void run() {
        // Keep sleeping till this many milliseconds have gone by.
        int snooze = 60000;
        // logChannel.write(Logger.DEBUG, "Starting request/min thread.");
        while (RPM_stayAlive) {
            long start = System.currentTimeMillis();
            while (RPM_stayAlive) {
                try {
                    Thread.sleep(snooze);
                } catch (InterruptedException ie) {
                }
                if (!RPM_stayAlive)
                    return;
                long now = System.currentTimeMillis();
                if ((now - start) >= snooze)
                    break;
             }
             // logChannel.write(Logger.DEBUG, "Updating request/min stats.");
             synchronized(this) {
                 long now = System.currentTimeMillis();
                 Enumeration e = myTable.keys();
                 while (e.hasMoreElements()) {
                     String servletID = (String) e.nextElement();
                     TableEntry t = (TableEntry) myTable.get(servletID);
                     double minutes = (double)(now - t.RPM_startTime) / 60000.0;
                     t.RPM_current = (double)t.RPM_counter / minutes;
                     if (t.RPM_current > t.RPM_max) {
                         t.RPM_max = t.RPM_current;
                         t.RPM_maxTime = new Date();
                     }
                     t.RPM_counter = 0;
                     t.RPM_startTime = now;
                     /*
                      *  Add the current RPM to the circular buffer.
                      */
                     int nextFree = t.historyStart + t.historyNum;
                     if (nextFree >= RPMBufferSize)
                         nextFree -= RPMBufferSize;
                     t.historyRPM[nextFree] = t.RPM_current;
                     t.historyWhen[nextFree] = now;
                     t.historyNum++;
                     if (t.historyNum > RPMBufferSize) {
                         // Overflow; just stomped on an old value.
                         t.historyStart++;
                         if (t.historyStart >= RPMBufferSize)
                             t.historyStart = 0;
                         t.historyNum = RPMBufferSize;
                     }
                 }
             }
         }
    }




    /**
     *  Register a new Servlet with the ServletManager. You pass in
     *  everything needed to find and create the Servlet, but it is not
     *  created until <CODE>start()</CODE> is called on it.
     *  Requests for the Servlet
     *  inbetween the time it is added and started will return null.
     *  After this call, the name passed in as servletID should be used
     *  to refer to the Servlet.<P>
     *
     *  Each element of classPath specifies a source of classes, either
     *  a .jar file, a .zip file, or a directory tree. Each of these sources
     *  has a set of class masks associated with it, initially null.
     *  If a source's set of class masks
     *  is null (the normal condition), then classes are loaded normally.
     *  If the set is not null, then it contains a set of class masks.
     *  Each class mask is a string.
     *  Before a source is searched to see if it contains
     *  a given class, the name of the class is compared with the class masks.
     *  Only if the requested class name begins with one of the class masks
     *  will it be loaded from the source. This lets you specify, for example,
     *  a jar file, but limit access to a specific class, or a set of classes
     *  that start with a common string. The syntax for specifing class
     *  masks is (on Unix): <CODE>pathname:mask,mask,...,mask</CODE>.
     *  If no masks are specified (the normal usage), then any class found
     *  via the pathname is used. Be sure to use
     *  <CODE>java.io.File.pathSeparatorChar</CODE> between the pathname
     *  and the mask list (this will be : on Unix, and ; on Windows and
     *  Macintosh). If the class mask does not specifiy a full class name,
     *  you may want to end it with a period. For example, if your mask was
     *  <CODE>com.lutris</CODE> it would allow access to
     *  <CODE>com.lutris.TheClass</CODE> as well as
     *  <CODE>com.lutrisMisc.UnwatedClass</CODE>. If the mask was
     *  <CODE>com.lutris.</CODE> the class <CODE>UnwantedClass</CODE>
     *  would not be loaded from this source.
     *
     *  @param servletID The identifier string used to refer to this
     *  Servlet.
     *  @param className The name of the class to instantiate. It must
     *  implement the Servlet interface.
     *  @param classPath An array of directories, .zip files or .jar files.
     *  The Servlet will be loaded by it's own class loader, which will
     *  look in these locations first. If a class is not found, the system
     *  class loader will be used (which uses the CLASSPATH Java started
     *  the MultiServer with).
     *  @param docRoot The root of the Servlet's filesystem on disk?
     *  @param initArgs The initial arguments to the Servlet. These will be
     *  made available to the Servlet's <CODE>init()</CODE> method.
     *  These names and values are accessed
     *  via the ServletConfig object passed to the Servlet's
     *  <CODE>init()</CODE> method.
     *  @param description A human readable description of this Servlet.
     *  @exception ServletException If there is an error.
     *  @see javax.servlet.Servlet
     */
    public synchronized void add(String servletID, String className,
                  String[] classPath, String docRoot, Properties initArgs,
                  boolean runOnInit, String description,boolean isWar) throws ServletException {
        TableEntry t = new TableEntry();

        // may not be set by the admin app
        if (docRoot == null) {
            docRoot = "/";
        }
        t.className = className;
        t.classPath = classPath;
        t.docRoot = docRoot;
        t.initArgs = initArgs;
        t.description = description;
        t.refCount = 0;
        t.servlet = null;
        t.classLoader = null;
        t.createTime = null;
        t.RPM_counter = 0;
        t.RPM_startTime = System.currentTimeMillis();
        t.RPM_current = -1.0;
        t.RPM_max = -1.0;;
        t.RPM_maxTime = null;
        t.historyRPM = new double[RPMBufferSize];
        t.historyWhen = new long[RPMBufferSize];
        t.historyStart = 0; // The start of the series.
        t.historyNum = 0;   // The buffer starts off empty.
        t.logChannel = Logger.getCentralLogger().getChannel(servletID);
        t.isWar = isWar;
        t.paths = new Vector();
        t.started = false;
        t.runOnInit = runOnInit;

        t.defaultSessionTimeout=null;
        t.isWARExpanded=null;
        t.isWARValidated=null;
        t.isInvokerEnabled=null;
        t.isWorkDirPersistent=null;

        myTable.put(servletID, t);
    }



    /**
     * Build a key out of the servletid and the path if the path is null
     * this is the base context that will be stored by the servletid only.
     * the separator between the id and the path is some string that can never
     * be in the id. e.g. a space.
     *
     */
    private String buildKey(String id,String path) {
       if (path == null) {
            // this is the base context set the path to something that will never be
            // in a url we will be able to lookup the base context by just the
            // servlet id and this string.
            path = " ? "; // something that can never be a url
       }
       else if (path.equals("/")) {
            // contexts at "/" are stored as ""
            path = "";
       }
       StringBuffer sb = new StringBuffer(80);
       sb.append(id);
       sb.append(" - ");  // something that can never be in an id
       sb.append(path);
       return sb.toString();
    }

    private void putContext(String id,String path,Context context) {
        if (path != null) {
            // notify the user about the new context
            Context baseContext = getContext(id,null);
            String msg;
            if (baseContext == null || path.equals(baseContext.getPath())) {
                msg = rez.format("Application {0} has url prefix \"{1}\".",id,path);
	        logChannel.write(Logger.INFO,msg);
            }
            else {
                msg = rez.format("Application {0} has an extra url prefix \"{1}\".",id,path);
   	        logChannel.write(Logger.WARNING,msg);
            }

        }
        contextByKey.put(buildKey(id,path),context);

    }


    private void removeContext(String id,String path) {
        if (path == null) {
            Enumeration enum = contextByKey.keys();
            while (enum.hasMoreElements()) {
                String key = (String) enum.nextElement();
                if (key.startsWith(id)) {
                    contextByKey.remove(key);
                }
            }
        }
        else {
            contextByKey.remove(buildKey(id,path));
        }
    }

    private Context getContext(String id,String path) {
        return (Context) contextByKey.get(buildKey(id,path));
    }



    /**
     * Record the url prefix that will be used by the context
     * The url path for the context is set when the channel is defined.
     * This is after the application is defined in the .conf file.
     *
     *  @param servletID  The name of the context
     *  @param urlPrefix  The start of the URL that leads to this context
     *
     */
    public void  recordContextPath(String servletID,String urlPrefix)
                     throws ConnectionMethodException  {

        urlPrefix = urlPrefix.trim();

        TableEntry t = (TableEntry) myTable.get(servletID);
        if (t != null) {
            // check if the prefix is already there if it is we can reuse it
            int i=0;
            while (i != t.paths.size()) {
                String path = (String) t.paths.elementAt(i);
                if (path.equalsIgnoreCase(urlPrefix)) {
                    return;
                }
                else {
                    if (t.isWar) {
                        String msg = "Web Application {0} already has a URL prefix of {1}. Cannot add new Prefix {2}.";
                        throw new ConnectionMethodException(rez.format(msg,servletID,path,urlPrefix));
                    }
                }
                i++;
            }
            t.paths.addElement(urlPrefix);
            if (t.started) {

                // need to add the extra context.
                Context baseContext = getContext(servletID,null);
                Context context = createContext(servletID,urlPrefix);
                context.copyContext(baseContext);
                putContext(servletID,urlPrefix,context);
            }
        }
     }


    /**
     * A channel has been closed remove the context.
     *
     *  @param servletID  The name of the context
     *  @param urlPrefix  The start of the URL that leads to this context
     *
     */
    public void  removeContextPath(String servletID,String urlPrefix) {
        urlPrefix = urlPrefix.trim();

        TableEntry t = (TableEntry) myTable.get(servletID);
        if (t != null) {
            int i=0;
            while (i < t.paths.size()) {
                String path = (String) t.paths.elementAt(i);
                if (path.equalsIgnoreCase(urlPrefix)) {
                    t.paths.removeElementAt(i);
                }
                i++;
            }
            if (t.started) {
                if (t.isWar) {
                    try {
                        stop(servletID);
                    }
                    catch (Exception ex) {
                        logChannel.write(Logger.ERROR,
                            rez.format("Error removing channel for app {0}.",
                            servletID,ex));
                    }
                }
                else {
                    // what ? Context context = getContext(servletID,urlPrefix);
                }
            }

        }
    }


    /**
     * These are values that only apply to a WAR
     *
     *  @param servletID  The id of the Context
     *  @param defaultSessionTimeout
     *  @param isWARExpanded
     *  @param isWARValidated
     *  @param isInvokerEnabled
     *  @param isWorkDirPersistent
     *
     */
    public void  recordWarOnlySettings(String servletID,
                                       String defaultSessionTimeout,
                                       String isWARExpanded,
                                       String isWARValidated,
                                       String isInvokerEnabled,
                                       String isWorkDirPersistent ) {

        // These defaults are discovered at init time they may
        // not be set in the conf file.
        if (defaultSessionTimeout == null || defaultSessionTimeout.length()==0)
          defaultSessionTimeout = defaultSessionTimeOutDefault;

        if (isWARExpanded == null || isWARExpanded.length()==0)
          isWARExpanded = isWARExpandedDefault;

        if (isWARValidated == null || isWARValidated.length()==0)
          isWARValidated = isWARValidatedDefault;

        if (isInvokerEnabled == null || isInvokerEnabled.length()==0)
          isInvokerEnabled = isInvokerEnabledDefault;

        if (isWorkDirPersistent == null || isWorkDirPersistent.length()==0)
          isWorkDirPersistent = isWorkDirPersistentDefault;

        TableEntry t = (TableEntry) myTable.get(servletID);
        if (t != null) {
          t.defaultSessionTimeout=defaultSessionTimeout;
          t.isWARExpanded=isWARExpanded;
          t.isWARValidated=isWARValidated;
          t.isInvokerEnabled=isInvokerEnabled;
          t.isWorkDirPersistent=isWorkDirPersistent;
        }
     }


    /**
     *  The context may have configuration information relating to the
     * sessionManager
     *
     *  @param servletID The id of the context.
     *
     */
    public void  recordSessionConfig(String servletID,
                                     Config sessionConfig) {
        TableEntry t = (TableEntry) myTable.get(servletID);
        if (t != null) {
          t.sessionConfig = sessionConfig;
        }
    }



    /**
     * After all the configuration information has been collected
     * it's safe to start all contexts that have been marked as runOnInit
     * in the multiserver configuration file
     *
     *  @exception Exception If there is an error.
     */
    public void startAllContexts() throws ServletException {

        String[] servletIds = null;
        int      i=0;
        try {
          servletIds = getServletIDs();
          while ( i != servletIds.length) {
            ServletStatus ss = getStatus(servletIds[i]);
            if (ss.runOnInit) {
                // if one application fails keep trying the others
                // If the admin starts up the user may be able to fix
                // the problem with GUI rather than editing the conf file.
                try {
                    start(servletIds[i]);
                }
                catch (ServletException se) {
                    String resKey = "Failed to start context {0}.";
                    String msg = rez.format(resKey,  servletIds[i] );
                    logChannel.write(Logger.ERROR, msg);
                }
            }
            i++;
          }
        } catch (Exception e) {
            String resKey = "Error starting Context {0}.";
            String msg = rez.format(resKey,  servletIds[i] );
            logChannel.write(Logger.DEBUG, msg, e);
            throw new ServletException(msg,e);
        }
    }




    // Given the servletID build the context. This method is called when the
    // context is started. If the context is a WAR the context is allowed to
    // read the web.xml file.
    private Context buildContext(String servletID) throws ServletException {

	//logChannel.write(Logger.DEBUG, "func: buildContext(\"" + servletID + "\")");

        ServletStatus servletStatus = getStatus(servletID);
  	URL url = null;
        try {
            url = URLUtil.resolve(servletStatus.docRoot);
        } catch (MalformedURLException mue) {
            String resKey = "Failed to resolve docRoot {0}";
            String msg = rez.format(resKey,  servletStatus.docRoot);
	    logChannel.write(Logger.ERROR,msg,mue);
            throw new ServletException(msg);
	}

        if (servletStatus.paths.size() == 0 ) {
            String resKey = "No path set for context {0}. Make sure there is a valid connection for the app.";
            String msg = rez.format(resKey,  servletID);
	    logChannel.write(Logger.ERROR,msg);
            throw new ServletException(msg);
        }

        Context baseContext = null;
        int i=0;
        while (i != servletStatus.paths.size()) {
            String path = (String) servletStatus.paths.elementAt(i);
            // If the path is "/" or nothing set in to default "".
            // For example if the context path is "/context", a url of "/context/path"
            // can be broken into "/context" and "/path" by removing the context path.
            // If the context path is "/" and we have a url of "/path"
            // it must be stored as "" so that the same operation results in "" and "/path"
            if (path.equals("") || path.equals("/")) {
                path = org.apache.tomcat.core.Constants.Context.Default.Path;
            }
            Context context = createContext(servletID, path);
            initContext(context,baseContext,path,url,servletID);
            if (i==0) {
                baseContext = context;
            }
            i++;
        }
        applyFilters(servletID);
        return baseContext;
    }

    /**
     * Create the context. (duh?)
     */
    private Context createContext(String servletID,String path) {
        SingleContextServer server = new SingleContextServer();
        Context context = new Context(server,path);
        server.setContext(context);
        context.setLogChannel(Logger.getCentralLogger().getChannel(servletID));
        return context;
    }


    private void applyFilters(String servletID) {
	// apply any filters

	ServletContainer sc = (ServletContainer)
		MultiServer.getService("ServletContainer");
	FilterManager fm = sc.getFilterManager();
	Vector fv = (Vector) filters.get(servletID);
	if (fv != null) {
	    Enumeration fe = fv.elements();
	    while (fe.hasMoreElements()) {
		String filterID = (String) fe.nextElement();
		Filter f = fm.get(filterID);
		f.registerInterceptors(servletID);
	    }
	}

    }

    private void initContext(Context context,Context baseContext,
                            String path,URL url,String servletID ) {

       	logChannel.write(Logger.DEBUG,
             rez.format("initContext() servletID={0} context={1} path={2} url={3}.",
             servletID ,context.toString(), path, url));

        String resKey;
        String msg;

        ServletStatus servletStatus = getStatus(servletID);
        putContext(servletID,path,context);

        if (baseContext == null) {
            // This is the baseContext
            putContext(servletID,null,context);
            context.setDocumentBase(url);
        }
        else {
            // copy the internal info from the baseContext
            context.copyContext(baseContext);
        }

        // Now init the context. If it is a war we can call init() which will
        // load and read the web.xml file. Otherwise we need to read the config
        // information from the enhydra app and feed it into the context.

        if (servletStatus.isWar) {
            // The work dir name needs to come from the config file.
            context.setWorkDir(workDir + java.io.File.separator + servletID,
                        servletStatus.IsWorkDirPersistent());
            context.setInvokerEnabled(servletStatus.IsInvokerEnabled());
            context.setIsWARExpanded(servletStatus.IsWARExpanded());
            context.setIsWARValidated(servletStatus.IsWARValidated());

            context.init();
            int timeOut=0;
            try {
               timeOut = Integer.parseInt(servletStatus.defaultSessionTimeOut);
            }
            catch (NumberFormatException nfe) {
                logChannel.write(Logger.ERROR,nfe.getLocalizedMessage(),nfe);
            }
            logChannel.write(Logger.DEBUG,
                rez.format("Context time out is ",new Integer(timeOut)));
            context.setSessionTimeOut(timeOut);
        }
        else {

            // where on the disk we do da work.
            context.setDocumentBase(url);

            // This context will have only one servlet, mapping it to "/" sets it
            // up as the 'default' servlet. Every request that is directed to the
            // context will invoke this servlet.
            context.getContainer().addServlet(servletID,  servletStatus.className, null);
            context.getContainer().addMapping(servletID,"/");

            // Now we have to set the initParameters of the context
            // copy these out of the property names of the property file  we have
            // now

            Hashtable initParams = new Hashtable();
            Enumeration argNames = servletStatus.initArgs.propertyNames();
            while (argNames.hasMoreElements()) {
                String name = (String) argNames.nextElement();
                String value = servletStatus.initArgs.getProperty(name);
                initParams.put(name,value);
            }
            context.getContainer().setServletInitParams(servletID, initParams) ;

            // a war might have a diffirent base and doc url we assume thay are the same
            // set the servlet base to root.

	    URL docRootUrl = getURLFromPath("/");
            context.getContainer().setServletBase(docRootUrl);

            // For a standalone servlet put the classpaths from the config file
            // into the context container.
            if (!servletStatus.className.equals(ServletContainer.ENYHDRA_APP_CLASSNAME)) {
                int len = servletStatus.classPath.length;
                int i=0;
                while (i!=len) {
                    context.getContainer().addClassPath(servletStatus.classPath[i]);
                    i++;
                }
            }

         }
    }

    /**
     * Adds the given filter to the list of filters to be applied
     * to the given servlet when the Context is built.
     *
     * @param servletID a String id for the Context (Servlet).
     * @param filterID a String id for the Filter to apply.
     */
    public void addFilter(String servletID, String filterID) {
	Vector fv = (Vector) filters.get(servletID);
	if (fv == null) {
	    Vector tmpv = new Vector();
	    filters.put(servletID, tmpv);
	    fv = tmpv;
	}
        if (fv.indexOf(filterID) == -1) {
	    fv.addElement(filterID);
	    // check to see if the context has already been built
	    if (getContext(servletID,null) != null) {
	        ServletContainer sc = (ServletContainer)
		        MultiServer.getService("ServletContainer");
	        FilterManager fm = sc.getFilterManager();
	        Filter f = fm.get(filterID);
	        f.registerInterceptors(servletID);
	    }
        }
    }

    public void removeFilter(String servletID, String filterID) {
	Vector fv = (Vector) filters.get(servletID);
	if (fv != null) {
	    fv.removeElement(filterID);
	    // check to see if the context has already been built
            if (getContext(servletID,null) != null) {
	        ServletContainer sc = (ServletContainer)
		        MultiServer.getService("ServletContainer");
	        FilterManager fm = sc.getFilterManager();
	        Filter f = fm.get(filterID);
	        f.unregisterInterceptors(servletID);
	    }
	}
    }



    /**
     * Service the request. The servletManager is taking over the role that the
     * ContextManager plays in Tomcat. So this routine is a close copy of the
     * ContextManager's <CODE>service()</CODE>. The main diffirence is that
     * contexts can be found ny name rather than by path. Which allows us to
     * have two contexts with the same urlprefix but diffirent channel's
     * e.g. diffirent ports.
     *
     *  @param channel  The channel that called us
     *  @param rrequest
     *  @param rresponse
     */
    public void service(Channel channel,Request rrequest, Response rresponse )
        throws Exception {
       // Get the context using the servletId
        String servletId = channel.getServletID();
        String urlPrefix = channel.getURLPrefix();
        Context context = getContext(servletId,urlPrefix);

        if (context == null) {
           String resKey = "Failed to find context for {0}.";
           String msg = rez.format(resKey,  servletId);
           logChannel.write(Logger.ERROR, msg);

           resKey = "The application {0} on path \"{1}\" has not been started.";
           msg = rez.format(resKey,servletId,urlPrefix);
           logChannel.write(Logger.ERROR, msg);

           throw new Exception(msg);
        }
        TableEntry t = (TableEntry)myTable.get(servletId);
        t.refCount++;
        t.RPM_counter++;

	// resolve the server that we are for
	String path = rrequest.getRequestURI();

	// final fix on response & request
	//		rresponse.setServerHeader(server.getServerHeader());

        if (urlPrefix.equals("/")) {
            urlPrefix = "";
        }
        String servletPath = path.substring(0, urlPrefix.length());
	String pathInfo = path.substring(urlPrefix.length(),path.length());

        // set the pathInfo with the full path param. At this point the
        // pathInfo may hold the sessionID. Later the pathInfo will be
        // recalculated from the requestURI by the Context
        rrequest.setPathInfo(pathInfo);

        rrequest.setRequestURI(path);


        // dont like this cast can we always assume that the request will be an EnhydraRequest?
        // add a ref to Context to point to the sessionmanager
        if (rrequest instanceof EnhydraRequest) {
            EnhydraRequest er = (EnhydraRequest) rrequest;

            SessionManager sessionManager = (SessionManager)
                context.getAttribute(SESSION_MANAGER_KEY);
            er.setSessionManager(sessionManager);

            er.obtainSessionId();
        }
        else {
            String resKey = "Could not set session manager unexpected request type for {0}.";
            String msg = rez.format(resKey,  servletId);
            logChannel.write(Logger.ERROR, msg);
            return;
        }

	try {
	    rrequest.setResponse(rresponse);
	    rresponse.setRequest(rrequest);

	    // XXX
	    //    return if an error was detected in processing the
	    //    request line
	    if (rresponse.getStatus() >= 400) {
		rresponse.finish();
		rrequest.recycle();
		rresponse.recycle();
		return;
	    }


	    //    don't do headers if request protocol is http/0.9
	    if (rrequest.getProtocol() == null) {
		rresponse.setOmitHeaders(true);
	    }

	    // do it
	    context.handleRequest(rrequest, rresponse);

	    // finish and clean up
	    rresponse.finish();

	    // protocol notification
	    rresponse.endResponse();

	} catch (org.enhydra.servlet.ServletIOException sioe) {
            // Connection reset by peer.
         //   sioe.printStackTrace();
 	} catch (Exception e) {
            String resKey = "Exception while handeling request \"{0}\" for servlet {1}.";
            String msg = rez.format(resKey,  rrequest.getRequestURI(),servletId);
            logChannel.write(Logger.ERROR, msg,e);
	}
    }



    /**
     *  Force the instantiation of a Servlet. It will be created, and it's
     *  <CODE>init()</CODE> method will be called.
     *
     *  @param servletID Which Servlet to start.
     *  @exception ServletException If there is an error.
     */
    public synchronized void start(String servletID) throws ServletException {
        String resKey;
        String msg = rez.format("Starting servlet/application {0} ",servletID);
        logChannel.write(Logger.INFO, msg );

        TableEntry t = (TableEntry) myTable.get(servletID);
        if (t == null) {
            resKey = "Attempt to start illegal servlet id {0}.";
            msg = rez.format(resKey, servletID);
            logChannel.write(Logger.WARNING, msg );

            resKey = "Unable to start. Servlet id {0} not found";
            msg = rez.format(resKey, servletID );
            throw new ServletException(msg);
        }

        if (t.started) {
            resKey = "Attempt to start already started servlet id {0}.";
            msg = rez.format(resKey, servletID );
            logChannel.write(Logger.WARNING, msg);

            resKey = "Unable to start. Servlet ID {0} not found";
            msg = rez.format(resKey,   servletID );
            throw new ServletException(msg);
        }
        // The context can't be built untill the URL path has been provided.
        if (t.isWar && t.paths.size() == 0) {
            resKey = "Channel must be created before starting Web Application: {0}.";
            msg = rez.format(resKey,   servletID );

            logChannel.write(Logger.WARNING, msg);
            throw new ServletException(msg);
        }

        Context tContext = buildContext(servletID);
        t.started = true;
        t.servlet = tContext.getContainer().getDefaultServlet();
        if (!t.isWar) {
            tContext.setupLoadableServlet(servletID);
            t.servlet = tContext.getContainer().getDefaultServlet();
            // If it's an enhydra force it up.
            try {
                startEnhydraApp(servletID,t);
            } catch (ServletException se) {
                t.started = false;
                removeContext(servletID,null);
                throw se;
            }
        }

        SessionManager sessionManager = null;
        // ensureappisrunning is now called by the init() of the servlet.


        if (t.servlet instanceof HttpPresentationServlet) {
	    HttpPresentationServlet hps = (HttpPresentationServlet) t.servlet;
            sessionManager = hps.getApplication().getSessionManager();
	}


        if (sessionManager == null) {
           // create new one ???
            Config   sessionConfig = t.sessionConfig;
            if (sessionConfig == null) {
                sessionConfig = new Config();
            }
            // If the class loader was not set by the enhyda app use the
            // context classLoader
            ClassLoader classLoader = (ClassLoader) t.classLoader;
            if (classLoader == null) {
                classLoader =  (ClassLoader) tContext.getAttribute("org.apache.tomcat.classloader");
            }
            try {
                StandardSessionManager sm =
                    new StandardSessionManager(classLoader, sessionConfig,
                            tContext.getLogChannel());
                // the context timeout is in minutes
                int timeout = tContext.getSessionTimeOut();
//                if (timeout != -1) {
                    timeout*=60;
//                }
                sm.setMaxSessionIdleTime(timeout);

                msg = rez.format("Session timeout set to {0} ",
                    ""+sm.getMaxSessionIdleTime());
                logChannel.write(Logger.DEBUG, msg);
                sessionManager = sm;
            }
            catch (Exception ex) {
                resKey = "Unable to start servlet {0}. SessionManager failed";
                msg = rez.format(resKey,  servletID);
                logChannel.write(Logger.ERROR, msg,ex);
                t.started = false;
            }
        }

        if (sessionManager != null) {
            tContext.setAttribute(SESSION_MANAGER_KEY,sessionManager);
        }

    }


    private void startEnhydraApp(String servletID, TableEntry t) throws ServletException {
            String resKey;
            String msg;
            if (t.classLoader == null) {
                t.classLoader = new MultiClassLoader(t.classPath, t.logChannel);
            }

            Class c = null;
            try {
                c = t.classLoader.loadClass(t.className);
            } catch (ClassNotFoundException e) {
                resKey = "Unable to start servlet {0}. Class {1} not found.";
                msg = rez.format(resKey,    servletID, t.className);
                logChannel.write(Logger.ERROR, msg);
                t.classLoader = null;

                resKey = "Unable to load class {0}.";
                msg = rez.format(resKey,   t.className);
                throw new ServletException(msg);
            }


            /*
            Object o = null;
            try {
                o = c.newInstance();
            } catch (Exception e) {
                resKey = "Unable to start servlet {0}. Unable to create instance {1}.";
                msg = rez.format(resKey,   servletID, e.toString());
                logChannel.write(Logger.ERROR, msg);
                t.classLoader = null;

                resKey = "Unable to instantiate class {0}.";
                msg = rez.format(resKey,  servletID);
                throw new ServletException(msg);
            }
            if (c.!(o instanceof javax.servlet.http.HttpServlet)) {
                resKey = "Unable to start servlet {0}. Class {1} is not an HttpServlet.";
                msg = rez.format(resKey, servletID, t.className);
                logChannel.write(Logger.ERROR, msg);
                t.classLoader = null;
                throw new ServletException(msg);
            }

            EnhydraServletContext context =
                             new EnhydraServletContext(servletID,
                                                       t.logChannel,
                                                       this,
                                                       t.initArgs,
                                                       t.docRoot,
                                                       t.classLoader);
            EnhydraServletConfig config =  new EnhydraServletConfig(t.initArgs,
                    context);
            config.setServletName(servletID);

            t.servlet = (Servlet) o;
            try {
                t.servlet.init(config);
            } catch (ServletException e) {
                resKey = "Unable to start servlet {0}. Error in init {1}";
                msg = rez.format(resKey,  servletID,e.toString());
                logChannel.write(Logger.ERROR, msg,e);

                try {
                    t.servlet.destroy();
	        } finally {
                    t.classLoader = null;
                    t.servlet = null;
                    throw(e);
	        }
            } catch (RuntimeException re) {
                resKey = "Unable to start servlet {0}. Error in init {1}";
                msg = rez.format(resKey,  servletID,  re.toString());
                logChannel.write(Logger.ERROR, msg);

        	try {
                    t.servlet.destroy();
	        } finally {
                    t.classLoader = null;
                    t.servlet = null;
                    throw(re);
	        }
            }
*/
          // X22
            t.refCount = 0;
            t.createTime = new Date();
            t.RPM_counter = 0;
            t.RPM_startTime = System.currentTimeMillis();
            t.RPM_current = -1.0;
            t.RPM_max = -1.0;;
            t.RPM_maxTime = null;
            t.started = true;

    }

    /**
     *  Stop a Servlet. This calls the Servlet's <CODE>destroy()</CODE> method,
     *  then discards the Servlet (making it available for garbage
     *  collection). After calling this, you may call <CODE>start()</CODE> to
     *  cause a new instance of the Servlet to be created.
     *
     *  @param servletID Which Servlet to stop.
     *  @exception ServletException If there is an error.
     */
    public synchronized void stop(String servletID) throws ServletException {

        String resKey = "Stopping servlet/application {0}";
        String msg = rez.format(resKey,  servletID);
        logChannel.write(Logger.INFO, msg);

        TableEntry t = (TableEntry) myTable.get(servletID);
        if (t == null) {
            resKey = "Attempt to stop illegal servlet id {0}.";
            msg = rez.format(resKey,  servletID);
            logChannel.write(Logger.WARNING, msg);

            resKey = "Unable to stop. Servlet ID {0} not found.";
            msg = rez.format(resKey,  servletID);
            throw new ServletException(msg);
        }

        if (!t.started) {
            resKey = "Attempt to stop already stopped servlet id {0}.";
            msg = rez.format(resKey,  servletID);

            logChannel.write(Logger.WARNING, msg);

            resKey = "Unable to stop. Servlet ID {0} already stopped.";
            msg = rez.format(resKey,  servletID);
            throw new ServletException(msg);
        }


        Context context = getContext(servletID,null);
        t.started = false;
        removeContext(servletID,null);

        if (!t.isWar) {
          // this will also shut it down

            try {
                if (t.servlet != null)
                    t.servlet.destroy();
            } finally {
                /*
                 *  If destroy throws anything, we will still clean up the table.
                 */
                t.classLoader = null; // It's classes will be reloaded next time.
                t.createTime = null;
                t.refCount = 0;
                t.servlet = null; // No more references, will be garbage collected.
            }
        }
        context.shutdown();
    }


    /**
     *  Fetch a Servlet.
     *  If the identifier string can not be found, an exception will be
     *  thrown. If the Servlet has not been started, or if it failed to
     *  start up, then null will be returned. The counters will be incremented
     *  to reflect this access. <P>
     *  This is the prefered version of <CODE>get()</CODE>. Only call the
     *  other version if you need to get the Servlet and you will not
     *  be calling it's <CODE>service()</CODE> method.
     *
     *  @param servletID Which Servlet to get.
     *  @return The specified Servlet.
     *  @exception ServletException If the name was not found.
     */
    public synchronized Servlet get(String servletID) throws ServletException {
        return get(servletID, true);
    }

    /**
     *  Fetch a Servlet, with optional control over the counter.
     *  This is exactly like <CODE>get(String servletID)</CODE>,
     *  except when counter is false, in which case the total references
     *  counter, and the requests per minute counter are not incremented. <BR>
     *  <B>This should only be used if you are not going to call
     *  <CODE>service()</CODE> on the Servlet.</B> It is assumed that the
     *  number of <CODE>get(String servletID)</CODE> is exactly the same
     *  as the number of times <CODE>service()</CODE> was called on the
     *  Servlet.
     *
     *  @param servletID Which Servlet to get.
     *  @param counter Should the counters be updated? This should only
     *  be false if you are not going to call <CODE>service()</CODE> on
     *  the returned Servlet.
     *  @return The specified Servlet.
     *  @exception ServletException If the name was not found.
     */
    public synchronized Servlet get(
               String servletID, boolean counter) throws ServletException {
        TableEntry t = (TableEntry) myTable.get(servletID);
        if (t == null) {
            // logChannel.write(Logger.WARNING, "Attempt to get illegal" +
            //     " servlet id " + servletID + ".");

            String resKey = "Unable to get. Servlet ID {0} not found.";
            String msg = rez.format(resKey,  servletID);
            throw new ServletException(msg);

        }
        if (counter && (t.servlet != null)) {
            t.refCount++;
            t.RPM_counter++;
        }

        Context context = getContext(servletID,null);
        return context.getContainer().getDefaultServlet();
        //return t.servlet;

      }


    /**
     *  Get the current status of a Servlet.
     *
     *  @param servletID Which Servlet to get the status of.
     *  @return A ServletStatus object describing the Servlet, or null
     *  if the name was not found.
     *  @see org.enhydra.servlet.servletManager.ServletStatus
     */
    public synchronized ServletStatus getStatus(String servletID) {
        TableEntry t = (TableEntry) myTable.get(servletID);
        if (t == null)
            return null;
        ServletStatus ss = new ServletStatus();
        ss.initArgs = t.initArgs;
        ss.description = t.description;
        ss.running = t.started;
        ss.refCount = t.refCount;
        ss.docRoot = t.docRoot;
        ss.createTime = t.createTime;
        ss.className = t.className;
        ss.classPath = t.classPath;
        ss.logChannel = t.logChannel;
        ss.currentReqPerMinute = t.RPM_current;
        ss.maxReqPerMinute = t.RPM_max;
        ss.maxReqPerMinuteTime = t.RPM_maxTime;
        ss.isWar = t.isWar;
        ss.runOnInit = t.runOnInit;
        ss.paths = t.paths;


        // Now the war stuff
        ss.defaultSessionTimeOut = t.defaultSessionTimeout;
        ss.isWARExpanded = t.isWARExpanded;
        ss.isWARValidated = t.isWARValidated;
        ss.isInvokerEnabled = t.isInvokerEnabled;
        ss.isWorkDirPersistent = t.isWorkDirPersistent;

        return ss;
    }



    /**
     *  Get the current request-per-minute statistics of a Servlet.
     *  This returns a copy of all the stored data! In the future we
     *  should return a sub-range of the data.
     *
     *  @param servletID Which Servlet to get the status of.
     *  @return A ServletStatistics object all the historical
     *  request-per-minute data for the Servlet, or null
     *  if the name was not found.
     *  @see org.enhydra.servlet.servletManager.ServletStatistics
     */
    public synchronized ServletStatistics getStatistics(String servletID) {
        TableEntry t = (TableEntry) myTable.get(servletID);
        if (t == null)
            return null;
        ServletStatistics ss = new ServletStatistics();
        ss.RPMValues = new double[t.historyNum];
        ss.RPMTimes = new long[t.historyNum];
        int i = t.historyStart;
        int n = 0;
        while (n < t.historyNum) {
            ss.RPMValues[n] = t.historyRPM[i];
            ss.RPMTimes[n] = t.historyWhen[i];
            i++;
            if (i >= RPMBufferSize)
                i = 0;
            n++;
        }
        return ss;
    }


    /**
     *  Resets the maxmimum requests per minute to the current value,
     *  and the time of the maximum to now.
     *
     *  @param servletID Which Servlet to reset the maximum request
     *  per minute stats for.
     *  @exception ServletException If there is an error, or if the Servlet
     *  ID is not found.
     */
    public synchronized void resetStats(
                               String servletID) throws ServletException {
        TableEntry t = (TableEntry) myTable.get(servletID);
        if (t == null) {

            String resKey = "Attempt to reset stats for illegal servlet id {0}.";
            String msg = rez.format(resKey,  servletID);

            logChannel.write(Logger.WARNING, msg);

            resKey = "Unable to reset stats. Servlet {0}  not found.";
            msg = rez.format(resKey,  servletID);

            throw new ServletException(msg);

        }
        t.RPM_max = t.RPM_current;
        t.RPM_maxTime = new Date();
    }



    /**
     *  Unregister a Servlet from the ServletManager.
     *  <CODE>stop()</CODE> is called on the Servlet,
     *  then it is removed from the set
     *  of Servlets. The name servletID is no longer valid.
     *
     *  @param servletID Which Servlet to delete.
     *  @exception ServletException If there is an error, or if the Servlet
     *  ID is not found.
     */
    public synchronized void delete(String servletID) throws ServletException {
        TableEntry t = (TableEntry) myTable.get(servletID);
        if (t == null) {

            String resKey = "Attempt to delete illegal servlet id {0}.";
            String msg = rez.format(resKey,  servletID);
            logChannel.write(Logger.WARNING, msg);

            resKey = "Unable to delete. Servlet {0} not found.";
            msg = rez.format(resKey,  servletID);
            throw new ServletException(msg);
        }
        if (t.servlet != null)
            stop(servletID);
        myTable.remove(servletID);
    }


    /**
     *  Get the names of all the Servlets in the ServletManager.
     *
     *  @return All the serverIDs currently in use.
     */
    public synchronized String[] getServletIDs() {
        int num = myTable.size();
        String[] results = new String[num];
        Enumeration e = myTable.keys();
        for (int i = 0; i<num; i++)
            results[i] = (String) e.nextElement();
        return results;
    }


    /**
     *  Get the names of all the Servlets in the ServletManager.
     *
     *  @return A vector containing the serverIDs currently in use.
     */
    public synchronized Enumeration getServletNames() {
        return myTable.keys();
    }


    /**
     *  Reset the reference count for a particular Servlet.
     *  This is the count of how many times the Servlet was asked for
     *  by <CODE>get()</CODE>.
     *
     *  @param servletID Which Servlet's counter to reset.
     *  @exception ServletException If there is an error, or if the Servlet
     *  ID is not found.
     *  @see org.enhydra.servlet.servletManager.ServletStatus
     */
    public synchronized void resetRefCount(String servletID)
                                               throws ServletException {
        TableEntry t = (TableEntry) myTable.get(servletID);
        if (t == null) {

            String resKey = "Attempt to reset count for illegal servlet id {0} .";
            String msg = rez.format(resKey,servletID);
            logChannel.write(Logger.WARNING, msg);

            resKey = "Unable to reference count. Servlet {0} not found.";
            msg = rez.format(resKey,servletID);
            throw new ServletException(msg);
        }
        t.refCount = 0;
    }



    /**********************************************************************
     * These methods are for the ServletConfig and ServletContext objects.
     **********************************************************************/

    /**
     *  This is for internal use only.
     *  It is part of the Servlet initalization.
     */
    protected synchronized String getMimeType(String file) {
        return myMime.getType(file);
    }

    /*
     *  Convert the a comma-separated string into a Properties
     *  object. Not currently used. Might be usefull in the future.
     */
    private synchronized static Properties parsePropertyString(
                                                    String properties) {
        Properties theProps = new Properties();
        if (properties == null)
            return theProps;
        StringTokenizer tokens = new StringTokenizer(properties, ",");
        while (tokens.hasMoreTokens()) {
            StringTokenizer t = new StringTokenizer(tokens.nextToken(), "=");
            String name = t.nextToken();
            String value = "";
            if (t.hasMoreTokens())
                value = t.nextToken();
            theProps.put(name, value);
        }
        return theProps;
    }


    // The Server  interface

    // The ServletManager implements the Server interface so that it can act
    // as the context manager of the Contexts.




    public Context lookuptContextByPath(String path) {
       boolean done = false;
       while (!done) {
         Enumeration e =  contextByKey.elements();
         while (e.hasMoreElements()) {
            Context context = (Context) e.nextElement();
            if (context.getPath().equals(path)) {
              return context;
            }
         }
         // didn't find it strip off the end and start again.
         int i = path.lastIndexOf('/');
         if (i > -1 ) {
            path = path.substring(0,i);
         }
         done = (path.length() <= 1) || (i == -1);
        }

       return defaultContext;
    }

    /**
     * Get server info string to pass to the contexts
     */
    public String getServerInfoString() {
        return SERVER_INFO_DISPLAY;
    }


    /**
     * Return the Context for the given servlet id.
     *
     * @param id Serlvet id string.
     * @return the Context that handles the servlet.
     */
    public Context getContextByServletID(String id) {
        return (Context) contextByKey.get(buildKey(id,null));
    }


    /**
     *  helper function to convert a string fie name to a url
     *
     * @param path a path to a directory on the disk amy be /
     */

    private URL getURLFromPath(String path) {
	URL url = null;
        try {
            url = URLUtil.resolve(path);
        } catch (MalformedURLException mue) {
            // One of thoese should never happ things
            logChannel.write(Logger.ERROR,
                rez.format("Can't create url for {0}",path),
                mue);
        }

        return url;
    }
}

