/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: ServletStatus.java,v 1.8.12.1 2000/10/19 17:59:10 jasona Exp $
 */

package org.enhydra.servlet.servletManager;

import java.util.*;
import com.lutris.logging.LogChannel;

/**
 *  This class describes the current status of a Servlet being stored in
 *  the ServletManager. Instances of this class are returned by a
 *  ServletManager.
 *
 *  @see org.enhydra.servlet.servletManager.ServletManager
 *  @see javax.servlet.Servlet;
 *  @author Andy John
 */
public class ServletStatus {

    /**
     *  Protected contructor. Only a ServletManager should create
     *  instances of this class.
     *
     *  @see org.enhydra.servlet.servletManager.ServletManager
     */
    protected ServletStatus() {
    }


    /**
     *  The initialization string that either will be or was used to
     *  initialize the Servlet. If running is true, the Servlet has
     *  been created and initialized.
     *
     *  @see org.enhydra.servlet.servletManager.ServletManager
     *  @see java.util.Properties
     */
    public Properties initArgs;


    /**
     *  A human readable string describing the Servlet.
     *
     *  @see org.enhydra.servlet.servletManager.ServletManager
     */
    public String description;

    /**
     *  The class name of the Servlet. For Enhydra applications
     *  this will be
     *  <CODE>org.enhydra.Servlet</CODE>.
     */
    public String className;

    /**
     *  The Servlet's root on the filesystem?
     *
     *  @see org.enhydra.servlet.servletManager.ServletManager
     */
    public String docRoot;


    /**
     *   The additional classpath used for this Servlet.
     */
    public String[] classPath;


    /**
     *  True if the Servlet currently exists. False if it does not
     *  exist. After calling get() or start(), the Servlet will exist,
     *  so this will be true. After calling stop() the Servlet will no
     *  longer exist, so this will be false.
     *
     *  @see org.enhydra.servlet.servletManager.ServletManager
     */
    public boolean running;

    /**
     *  True if the configuration information for the conterxt
     *  says to start at init time
     *
     *  @see org.enhydra.servlet.servletManager.ServletManager
     */
    public boolean runOnInit;


    /**
     *  The log channel the Servlet writes to. If the Servlet call the
     *  <CODE>log()</CODE> method of it's ServletContext, the string
     *  ends up being written to this channel.
     */
    public LogChannel logChannel;


    /**
     *  The number of times that this Servlet was asked for by get().
     *
     *  @see org.enhydra.servlet.servletManager.ServletManager
     */
    public int refCount;


    /**
     *  The time when the Servlet was instantiated, and <CODE>init()</CODE>
     *  was called. This will have happend as a result of either a
     *  <CODE>get()</CODE> or a <CODE>start()</CODE> code.
     *  If <CODE>running</CODE> is null, then the Servlet is not currently
     *  instantiated, in which case this will be null also.
     */
    public Date createTime;


    /**
     * The number of requests per minute made to this Servlet for the
     * most recent completed minute. This statistic can be up to 59 seconds
     * out of date. This is reset when the Servlet is started.
     */
    public double currentReqPerMinute;


    /**
     * The highwater mark for <CODE>currentReqPerMinute</CODE>. This is the
     * largest value of <CODE>currentReqPerMinute</CODE> seen since the
     * Servlet was started. This is reset when the Servlet is started.
     */
    public double maxReqPerMinute;


    /**
     * The time at which the <CODE>maxReqPerMinute</CODE> occured.
     * This is reset when the Servlet is started.
     */
    public Date maxReqPerMinuteTime;

    /**
     * If this servlet is a war
     *
     */
    public boolean isWar;


    /**
     * The URL prefix used by the context
     *
     */
    public Vector paths;


    /**
     * The default session timeout used by a war
     *
     */
    public String defaultSessionTimeOut;

    /**
     * The isWARExpanded used by a war
     *
     */
    public String isWARExpanded;

    /**
     * The default session timeout used by a war
     *
     */
    public String isWARValidated;

    /**
     * The isWARValidated used by a war
     *
     */
    public String isInvokerEnabled;

    /**
     * The isInvokerEnabled used by a war
     *
     */
    public String isWorkDirPersistent;


    public boolean IsWARExpanded()
    {
        return isWARExpanded != null && isWARExpanded.equals("true");
    }

    public boolean IsWARValidated()
    {
        return isWARValidated != null && isWARValidated.equals("true");
    }

    public boolean IsInvokerEnabled()
    {
        return isInvokerEnabled != null && isInvokerEnabled.equals("true");
    }

    public boolean IsWorkDirPersistent()
    {
        return isWorkDirPersistent != null && isWorkDirPersistent.equals("true");
    }

}

