/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: FileServerServlet.java,v 1.4.12.1 2000/10/19 17:59:10 jasona Exp $
 */

package org.enhydra.servlet.servlets;

import java.io.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;
import com.lutris.logging.*;
import com.lutris.http.MimeType;
import org.enhydra.servlet.connectionMethods.http.*;
import org.enhydra.servlet.servletManager.EnhydraServletContext;


/**
 * This class serves files from HTTP request. Currently only a 
 * limited Http GET request will be processed. CGI arguments will
 * not be processed. The initial parameter 'index=something.html' can
 * for the default html page. If this parameter is not set then it
 * will default to "index.html".
 *
 * @version 1.8, 3/98
 * @auther Mark Sanguinetti
 */

public class FileServerServlet 
    extends HttpServlet {

    /**
     * Buffer size for writeing one stream to another.
     */
    private static final int buffer_size = 1024*1;

    /**
     * Html text to send clent for a file not found.
     */
    private final static String NOT_FOUND_MSG = 
        "<TITLE>Not Found</TITLE><H1>Not Found</H1> " +
        "The requested object does not exist on this Enhydra Multiserver. " +
        "The link you followed is either outdated, inaccurate, " +
        "or the server has been instructed not to let you have it.";

    /**
     * Html text to send client when Access to file denied.
     */
    private final static String ACCESS_DENIED_MSG = 
        "<TITLE>Access Denied</TITLE><H1>Access Denied</H1> " +
        "Access is not allowed to the requested file. ";

    /**
     *
     */
    private final static String SC_MOVED_PERMANENTLY_MSG = 
        "<TITLE>Moved Permanently</TITLE><H1>Moved Permanently</H1> " +
        "Resouce has been moved";

    /**
     * Default html page. 
     */
    private String defaultPage = null;

    /**
     * The logging channel.
     */
    private LogChannel logChannel;

    /**
     * The Sevelet context.
     */
    private ServletContext myContext;

    /**
     * The document root for the file server and is 
     * specified in servlet context.
     */
    private Path docRoot = null; 

    /**
     * Number of file served by this servlet.
     */
    private long numServed = 0;

    /**
     * Class for determing the mime type of a file
     */
    private MimeType mimetypes = new MimeType();
                                       

    /**
     * Mime type extension for executing a cgi script in thr Doc Root.
     */
    private String cgiType = null;

    /**
     * This initializes the servlet and sets the variables docRoot and
     * log logging options.
     *
     * @param config The servlet configuration object.
     * @exception If an exception occured during initialztion  
     */
    public void init(ServletConfig config) 
        throws ServletException {
        super.init(config);
        myContext = config.getServletContext();

        // FIX: check for docRoot == null
        docRoot = new Path(myContext.getRealPath(""));

        // get defaultPage form parameters
        defaultPage = config.getInitParameter("index");
        if (defaultPage == null) {
            defaultPage = "index.html";
        }

        // cgi extension, if any. 
        cgiType = config.getInitParameter("cgi");

        //Try to get the logging channel, if we are being run under
        // the MultiServer.
        if (myContext instanceof
            org.enhydra.servlet.servletManager.EnhydraServletContext) {
            EnhydraServletContext lbsContext = 
                (EnhydraServletContext) myContext;
            logChannel = lbsContext.getLogChannel();
        }
        logInfo("Starting File Server Servlet with docRoot " + docRoot.getFilePath());
    }

    /**
     * This method will process a limited http GET request for a file. 
     * The method will serve files, but will not process cgi arguments.
     * 
     * @param req encapsulates the request to the servlet. 
     * @param res encapsulates the response from the servlet. 
     * @exception if the request could not be handled.
     * @exception if detected when processing.
     */
    public void doGet (HttpServletRequest req, HttpServletResponse res)
	throws ServletException, IOException {

        /* FIX: Path object is used because the file object doesn't
         * fix multiple file separator char and also problems with
         * converting URL path to a File path with platform specific
         * file separator char.
         */

        String pathInfo = req.getPathInfo();
        String path = docRoot.appendFilePath(pathInfo);
        File file = new File(path);
        if (isAuth(file.getPath())) {
            // check for cgi
            if (cgiType != null) {
                String scriptPath = getScriptPath(path);
                if (scriptPath != null) {
                    // process cgi
                    String scriptPathInfo = getScriptPathInfo(path, scriptPath);
                    String scriptName = getScriptName(docRoot.getUrlPath(), scriptPath);
                    CgiProcessor cgi = new CgiProcessor();
                    cgi.processCgiRequest(req, res, scriptPath, scriptPathInfo, scriptName);
                } else {
                    // process file request
                    processFileRequest(file, req, res);
                }
            } else {                
                // process file request
                processFileRequest(file, req, res);
            }    
        } else {
            // path containing '..' is not authorized
            res.sendError(res.SC_NOT_FOUND, NOT_FOUND_MSG);
            logInfo("FileServerServlet: File " + file + " not Found.");
        }
    }

    /**
     *
     */
    private String getScriptName(String docRoot, String scriptPath) {

        return scriptPath.substring(docRoot.length() + 1);
    }

    /**
     *
     */
    private String getScriptPathInfo(String path, String scriptPath) {

        int index = path.indexOf(scriptPath);
        if (index == -1) {
            return "";
        }
        return path.substring(index + scriptPath.length());
    }

    /**
     *
     */
    private String getScriptPath(String path) {

        String cgi = "." + cgiType;
        int index = path.indexOf(cgi);
        if (index == -1) {
            return null;
        }
        return path.substring(0, index + cgi.length());
    }


    /**
     *
     */
    private void processFileRequest(File file, HttpServletRequest req, HttpServletResponse res) throws IOException {

        if (file.exists()) {
            // file exists
            if (file.isDirectory()) {
                // send dir. index page of list dir.
                processDirectory(req, res);
            } else {
                // send requested file
                sendPage(file, res);
            }
        } else {
            // file doesn't exist
            res.sendError(res.SC_NOT_FOUND, NOT_FOUND_MSG);
            logInfo("FileServerServlet: File " + file + " not Found.");
        }
    }

    /**
     *
     */
    private void processDirectory(HttpServletRequest req, HttpServletResponse res) 
        throws IOException {

        String pathInfo = req.getPathInfo();
        if (pathInfo == null) {
            pathInfo = "/";
        }

        Path path = new Path(docRoot.getUrlPath(), pathInfo); 
        if (pathInfo.endsWith("/")) {
            String indexPath = path.appendFilePath(defaultPage);
            File indexFile = new File(indexPath);
            if (indexFile.exists()) {
                // send index file
                sendPage(indexFile, res);
            } else {
                // list directory
                File dir = new File(path.getFilePath());
                listDirectory(dir, req, res);
            }
        } else {
            sendDirectoryRedirect(req, res);
        }
    }

    /**
     * 
     */
    private void sendDirectoryRedirect(HttpServletRequest req, HttpServletResponse res) 
        throws IOException {

        String location = req.getRequestURI() + "/";
        res.setHeader("Location", location); 
        res.sendError(res.SC_MOVED_PERMANENTLY, SC_MOVED_PERMANENTLY_MSG);
    } 

    /**
     * Reconstruct the URL for this request.
     *
     * @param request to the servlet.
     * @return url. 
     */
    String getUrl(HttpServletRequest request) { 

        String host = request.getScheme() + "://";
        host += request.getServerName();
        int port;
        if ((port = request.getServerPort()) != 80) {
            host += ":" + port;
        }
        return request.getRequestURI();
    }

    /**
     * Generates an html listing of a directory.
     *
     * FIX: should create a temp file, set a content-length aheadr
     * and output like a normal file.
     *
     * @param root Directory of interest.
     * @param req
     * @param res
     */
    private void listDirectory(File root, HttpServletRequest req, HttpServletResponse res) 
        throws IOException {

        res.setContentType("text/html");
        ServletOutputStream out = res.getOutputStream();

        String vPath = (req.getPathInfo() == null ? "/" : req.getPathInfo());
        String vParent = (vPath.equals("/") ? "" : "../" );
        out.println("<html>");
        out.println("<head><title>Index of " + vPath + "</title></head>");
        out.println("<body>");
        out.println("<h1>Index of " + vPath + "</h1>\n<hr>");
        out.println("<pre>");
        out.println("<a href=\"" + vParent + "\">Parent Directory</a>");

        String href = null;
        String files[] = root.list();
        String path = root.getPath();
        for (int i = 0; i < files.length; i++) {
            File f = new File(path + "/" + files[i]);
            if (f.isDirectory()) {
                href = "<a href=\"" + files[i] + "/\">" + files[i] + "</a>";
            } else {
                href = "<a href=\"" + files[i] + "\">" + files[i] + "</a>";
            }            
            out.println(href);
        }
        out.println("</pre>");
        out.println("</body>");
        out.println("</html>");
    }   
        
    /**
     * Test if is authorized. A path is authorized if
     * it does not contain a '..'. The browser should resolve
     * the '..'.
     *
     * @param path requested path
     * @return true if path is authorized
     */
    private boolean isAuth(String path) {

        if (path.indexOf("..") == -1) {
            return true;
        } 
        return false;
    }

    /**
     *
     *
     */
    private void sendPage(File file, HttpServletResponse res) 
        throws IOException { 

        String filetype = mimetypes.getType(file.getName());
        res.setContentType(filetype);
        Long len = new Long(file.length());
        res.setHeader("Content-Length", len.toString());
        ServletOutputStream out = res.getOutputStream();
        FileInputStream in = new FileInputStream(file);
        writeFile(out, in);
        in.close();
        numServed++;
    }

    /**
     * Http POST command is currently not supported and indicates 
     * this sending a message to the client.
     *
     * @param req encapsulates the request to the servlet. 
     * @param res encapsulates the response from the servlet. 
     * @exception if the request could not be handled.
     * @exception if detected when processing.
     */
    public void doPost (HttpServletRequest req, HttpServletResponse res)
	throws ServletException, IOException {
        res.sendError(res.SC_NOT_IMPLEMENTED);
        logInfo("FileServerServlet: Access Denied.");
    }

    /**
     * Returns a string containing servlet infoemation including name, 
     * and docRoot. 
     *
     * @return Servlet informatioon.
     */
    public String getServletInfo() {
        return "A FileServerServlet, with doc root " + docRoot.getFilePath();
    }

    /**
     *  Extra status info, shown by the MultiServer admin status page.
     */
    public String toHtml() {
        return "Serving files from the directory " + docRoot.getFilePath() + ".<BR>\n" +
               numServed + " files served.";
    }

    /**
     * Transfers to one stream to another via a buffer. In this case
     * the input stream is a file to be transferred and the output
     * stream is servlet output stream.   
     * 
     * @param out The destination stream.
     * @param in The source stream.
     * @exception if an error occured during IO operations. 
     */
    private void writeFile(OutputStream out, InputStream in)
        throws IOException {
    
        byte buffer[] = new byte[buffer_size];
        while (true) {
            int n = in.read(buffer);
            if (n == -1) break;
            out.write(buffer, 0, n);
        }
    }

    /**
     * Write standard (non error) information to the log. If logChannel 
     * is not null then the information will be logged to the lbs 
     * logged channel defined in lbsContext otherwise the information 
     * will be logged to the standard servlet log. 
     *
     * @param msg The message to be logged. 
     */
    private void logInfo(String msg) {
        if (logChannel != null)
            logChannel.write(Logger.INFO, msg);
        else
            myContext.log(msg);
    }

    /**
     * Write error information to the log. If logChannel is not null 
     * then the information will be logged to the lbs logged channel 
     * defined in lbsContext otherwise the information will be logged to 
     * the standard servlet log. 
     *     
     *@param err The error message to be logged.
     *@param e The exception to be logged. 
     */
    private void logError(String err, Exception e) {
        if (logChannel != null)
            logChannel.write(Logger.ERROR, err, e);
        else
            myContext.log(e, err);
    }
}
