/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: JarServerServlet.java,v 1.3.12.1 2000/10/19 17:59:10 jasona Exp $
 */

package org.enhydra.servlet.servlets;

import java.io.*;
import java.util.*;
import java.util.zip.*;
import javax.servlet.*;
import javax.servlet.http.*;
import com.lutris.logging.*;
import com.lutris.http.MimeType;
import org.enhydra.servlet.connectionMethods.http.*;
import org.enhydra.servlet.servletManager.EnhydraServletContext;


/**
 * This class serves files from jar/zip files for HTTP requests.
 * Currently only a 
 * limited Http GET request will be processed. CGI arguments will
 * not be processed.  The initArg "jarFile" sets the name of the jar
 * or zip file to serve files from.
 *
 * @version $Revision: 1.3.12.1 $
 * @auther Mark Sanguinetti
 * @auther Shawn McMurdo
 */
public class JarServerServlet extends HttpServlet {

    /**
     * Buffer size for writing one stream to another.
     */
    private static final int buffer_size = 1024*1;

    /**
     * Html text to send client for a file not found.
     */
    private final static String NOT_FOUND_MSG = 
        "<TITLE>Not Found</TITLE><H1>Not Found</H1> " +
        "The requested object does not exist on this Enhydra Multiserver. " +
        "The link you followed is either outdated, inaccurate, " +
        "or the server has been instructed not to let you have it.";

    /**
     * Html text to send client when Access to file denied.
     */
    private final static String ACCESS_DENIED_MSG = 
        "<TITLE>Access Denied</TITLE><H1>Access Denied</H1> " +
        "Access is not allowed to the requested file. ";

    /**
     *
     */
    private final static String SC_MOVED_PERMANENTLY_MSG = 
        "<TITLE>Moved Permanently</TITLE><H1>Moved Permanently</H1> " +
        "Resource has been moved";

    /**
     * The logging channel.
     */
    private LogChannel logChannel;

    /**
     * The Servlet context.
     */
    private ServletContext myContext;

    /**
     * The jar file for the file server.  It is
     * specified in the servlet initArgs.
     */
    private String jarFileName = null; 
    private ZipFile jarFile = null; 

    /**
     * Number of files served by this servlet.
     */
    private long numServed = 0;

    /**
     * Class for determining the mime type of a file
     */
    private MimeType mimetypes = new MimeType();
                                       

    /**
     * This initializes the servlet and gets the jarFile parameter.
     *
     * @param config The servlet configuration object.
     * @exception If an exception occured during initialztion  
     */
    public void init(ServletConfig config) 
        throws ServletException {
        super.init(config);
        myContext = config.getServletContext();

        jarFileName = config.getInitParameter("jarFile");
	try {
	    jarFile = new ZipFile(jarFileName);
	} catch (IOException ioe) {
	    throw new ServletException(ioe.getMessage());
	}

        //Try to get the logging channel, if we are being run under
        // the MultiServer.
        if (myContext instanceof EnhydraServletContext) {
            EnhydraServletContext lbsContext = 
                (EnhydraServletContext) myContext;
            logChannel = lbsContext.getLogChannel();
        }
        logInfo("Starting Jar Server Servlet with jarFile " + jarFileName);
    }

    /**
     * This method will process a limited http GET request for a file. 
     * The method will serve files, but will not process cgi arguments.
     * 
     * @param req encapsulates the request to the servlet. 
     * @param res encapsulates the response from the servlet. 
     * @exception if the request could not be handled.
     * @exception if detected when processing.
     */
    public void doGet(HttpServletRequest req, HttpServletResponse res)
	    throws ServletException, IOException {

        String pathInfo = req.getPathInfo();
        ZipEntry jarEntry = jarFile.getEntry(pathInfo);
	if (isAuth(pathInfo)) {
            processFileRequest(jarEntry, req, res);
        } else {
            // path containing '..' is not authorized
            res.sendError(res.SC_NOT_FOUND, NOT_FOUND_MSG);
            logInfo("JarServerServlet: File " + pathInfo + " not Found.");
        }
    }

    /**
     *
     */
    private void processFileRequest(ZipEntry file, HttpServletRequest req,
	    HttpServletResponse res) throws IOException {

        if (file != null) {
            // file exists
            if (file.isDirectory()) {
                // don't do directories
                res.sendError(res.SC_NOT_FOUND, NOT_FOUND_MSG);
                logInfo("JarServerServlet: File " + file + " not Found.");
            } else {
                // send requested file
                sendPage(file, res);
            }
        } else {
            // file doesn't exist
            res.sendError(res.SC_NOT_FOUND, NOT_FOUND_MSG);
            logInfo("JarServerServlet: File " + file + " not Found.");
        }
    }

    /**
     * Test if path is authorized. A path is authorized if
     * it does not contain a '..'. The browser should resolve
     * the '..'.
     *
     * @param path requested path
     * @return true if path is authorized
     */
    private boolean isAuth(String path) {

        if (path.indexOf("..") == -1) {
            return true;
        } 
        return false;
    }

    /**
     *
     *
     */
    private void sendPage(ZipEntry file, HttpServletResponse res) 
            throws IOException { 

        String filetype = mimetypes.getType(file.getName());
        res.setContentType(filetype);
        Long len = new Long(file.getSize());
        res.setHeader("Content-Length", len.toString());
        ServletOutputStream out = res.getOutputStream();
        InputStream in = jarFile.getInputStream(file);
        writeFile(out, in);
        in.close();
        numServed++;
    }

    /**
     * Http POST command is currently not supported and indicates 
     * this sending a message to the client.
     *
     * @param req encapsulates the request to the servlet. 
     * @param res encapsulates the response from the servlet. 
     * @exception if the request could not be handled.
     * @exception if detected when processing.
     */
    public void doPost (HttpServletRequest req, HttpServletResponse res)
	    throws ServletException, IOException {
        res.sendError(res.SC_NOT_IMPLEMENTED);
        logInfo("JarServerServlet: Access Denied.");
    }

    /**
     * Returns a string containing servlet information including name, 
     * and jarFile. 
     *
     * @return Servlet information.
     */
    public String getServletInfo() {
        return "A JarServerServlet, with jar file " + jarFileName;
    }

    /**
     *  Extra status info, shown by the MultiServer admin status page.
     */
    public String toHtml() {
        return "Serving files from the jar " + jarFileName + ".<BR>\n" +
               numServed + " files served.";
    }

    /**
     * Transfers from one stream to another via a buffer. In this case
     * the input stream is a file to be transferred and the output
     * stream is servlet output stream.   
     * 
     * @param out The destination stream.
     * @param in The source stream.
     * @exception if an error occured during IO operations. 
     */
    private void writeFile(OutputStream out, InputStream in)
            throws IOException {
    
        byte buffer[] = new byte[buffer_size];
        while (true) {
            int n = in.read(buffer);
            if (n == -1) break;
            out.write(buffer, 0, n);
        }
    }

    /**
     * Write standard (non error) information to the log. If logChannel 
     * is not null then the information will be logged to the multiserver
     * log channel otherwise the information 
     * will be logged to the standard servlet log. 
     *
     * @param msg The message to be logged. 
     */
    private void logInfo(String msg) {
        if (logChannel != null) {
            logChannel.write(Logger.INFO, msg);
        } else {
            myContext.log(msg);
	}
    }

    /**
     * Write error information to the log. If logChannel is not null 
     * then the information will be logged to the multiserver log channel 
     * otherwise the information will be logged to 
     * the standard servlet log. 
     *     
     * @param err The error message to be logged.
     * @param e The exception to be logged. 
     */
    private void logError(String err, Exception e) {
        if (logChannel != null) {
            logChannel.write(Logger.ERROR, err, e);
        } else {
            myContext.log(e, err);
	}
    }
}

