/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: Path.java,v 1.4.12.1 2000/10/19 17:59:10 jasona Exp $
 */

package org.enhydra.servlet.servlets;

import java.io.*;
import java.util.*;

/**
 * Utility class to help manipulate paths. Method cleans up multiple
 * '/' or '\' when constructing the object and prevents adding additional
 * '/' or '\' when appending paths. For the purposes of this class a
 * 'UrlPath' is a path with the path separators character of '/', 
 * regardless of platform; A 'FilePath' is a path with system dependent
 * path separators characters (UNIX - '/' DOS - '\').
 *
 * @version $Revision: 1.4.12.1 $
 * @author Mark Sanguinetti
 */
class Path {

    /*
     * Internal represenation of a path. 
     * This class always stores paths in URL format ('/').
     */
    private String path = null;

    /**
     * Constructs a Path Object based on String p.
     *
     * @param p used to form the path.
     */
    Path(String p) {
        path = cleanUpPath(p);
    }

    /**
     * Constructs a Path Object as by appending p2 to the end
     * of p1.
     *
     * @param p1 the first part of the path.
     * @param p2 the second part of the path.
     */
    Path(String p1, String p2) {
        path = cleanUpPath(p1);
        path = appendPath(p2);
    }

    /**
     * Converts the Path object to 'UrlPath' format. 'UrlPath' is 
     * a path with the path separators character of '/', regardless 
     * of platform.
     * 
     * @return Path in URL format.
     */
    public String getUrlPath() {
        return path; 
    }

    /**
     * Converts the Path object to 'FilePath' format. A 'FilePath' 
     * is a path with system dependent path separators characters 
     * (UNIX - '/' DOS - '\').
     * 
     * @return Path in FilePath format.
     */
    public String getFilePath() {
        return (path.replace('/', File.separatorChar)); 
    }

    /**
     * Appends a path to the end of this object, and returns a 
     * string representing the new Url path without changing the 
     * this object.
     *
     * @param p The path to append to the end of this object.
     * @return The new Url path.
     */
    public String appendUrlPath(String p) {
        return appendPath(p);
    }

    /**
     * Appends a path to the end of this object, and returns a 
     * string representing the new File path without changing the 
     * this object.
     *
     * @param p The path to append to the end of this object.
     * @return The new File path.
     */
    public String appendFilePath(String p) {
        return (appendPath(p).replace('/', File.separatorChar)); 
    }

    /*
     *
     */
    private String appendPath(String p) {

        if (path == null) {
            return null;
        } 
        return (p == null ? path : path + cleanUpPath(p));
    }

    /*
     *
     */
    private String cleanUpPath(String p) {

        // covert to internal represenation of a path ('/' format)
        String tmp = p.replace('\\', '/');
        if (!tmp.startsWith("/")) {
            tmp = "/" + tmp;
        }
        if (!tmp.endsWith("/")) {
            tmp = tmp + "/";
        }

        // parse out info in path and create a cleaned up
        // url path
        StringBuffer sb = new StringBuffer();
        StringTokenizer st = new StringTokenizer(tmp, "/");
        sb.append("/");
        while (st.hasMoreTokens()) {
            String tok = st.nextToken();
            if (st.countTokens() > 0) {
                sb.append(tok);            
                sb.append("/");
            } else if (st.countTokens() == 0) {
                sb.append(tok);
            } else {
                // should never happen
                System.err.println("ERROR:Path.cleanUpPath()");
            }
        }
        return sb.toString();
    }

}
