// Copyright 2000 Sims Computing, Inc.
// Licensed under the GNU Lesser General Public License.

package com.simscomputing.testbed;

import com.simscomputing.SoftwareFaultException;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;

/**
Provides access to private and protected constructors, methods, and
fields in arbitrary classes. Inspired by a similar class written by
Creig Smith <creigs@acm.org>.

@author David Sims
@version $Revision: 1.1.1.1 $ $Date: 2000/02/21 21:22:33 $ */
public class ClearBox {
  /**
  This class is just a collection of static methods. No need to instantiate it.
  */
  private ClearBox() {
  } // constructor

  /**
  Helper method for recursively finding private and protected fields.
  */
  private static Field findField(final Class c, final String fieldName)
      throws NoSuchFieldException {

    // make sure the class is still legitimate
    if (c == null) {
      throw new NoSuchFieldException();
    } // if

    try {
      return c.getDeclaredField(fieldName);
    } // try
    catch (NoSuchFieldException e) {
      // try again on the parent
      return findField(c.getSuperclass(), fieldName);
    } // catch
  } // findField()

  /**
  Finds a private, protected, or public field in the specified object.
  */
  public static Object getField(Object object, String fieldName) throws NoSuchFieldException {
    try {
      final Class c = object.getClass();
      final Field field = findField(c, fieldName);

      // the following call skirts private/protected access
      // protection unless the security manager prevents it
      field.setAccessible(true);

      return field.get(object);
    } // try
    catch (IllegalAccessException e) {
      // due to the setAccessible() call above, should not reach here
      throw new SoftwareFaultException("access should be legal", e);
    } // catch
  } // getField()

  /**
  Finds the class for a private, protected, or public field in the
  specified object.
  */
  public static Class getFieldClass(Object object, String fieldName)
      throws NoSuchFieldException {
    final Class c = object.getClass();
    final Field field = findField(c, fieldName);

    // the following call skirts private/protected access
    // protection unless the security manager prevents it
    field.setAccessible(true);

    return field.getType();
  } // getFieldClass()

  /**
  Sets a private, protected, or public field in the specified object.

  @param object the object whose field is being set
  @param fieldName the name of the field to set
  @param newValue the value that the field receives
  @throws NoSuchFieldException if fieldName does not exist
  */
  public static void setField(final Object object,
                              final String fieldName,
                              final Object newValue)
      throws NoSuchFieldException {
    try {
      final Field field = findField(object.getClass(), fieldName);

      // the following call skirts private/protected access
      // protection unless the security manager prevents it
      field.setAccessible(true);
      
      field.set(object, newValue);
    } // try
    catch (final IllegalAccessException e) {
      // due to the setAccessible() call above, should not reach here
      throw new SoftwareFaultException("access should be legal", e);
    } // catch
  } // setField()

  /**
  Instantiates an object with a private or protected constructor.
  */
  public static Object createNewInstance(String className, Class[] constructorParameterTypes, 
                                         Object[] constructorParameters) 
    throws ClassNotFoundException, NoSuchMethodException,
    SecurityException, InstantiationException, InvocationTargetException,
    IllegalAccessException {
    Class newClass = ClearBox.class.getClassLoader().loadClass(className);
    Constructor constructor = newClass.getDeclaredConstructor(constructorParameterTypes);

    // the following call skirts private/protected access
    // protection unless the security manager prevents it
    constructor.setAccessible(true);
    
    return constructor.newInstance(constructorParameters);
  } // createNewInstance()

  /**
  Helper method for recursively finding private, protected,
  and public methods.
  */
  private static Method findMethod(final Class c, final String methodName,
                                   final Class[] methodParameterTypes)
    throws NoSuchMethodException {
      try {
      return c.getDeclaredMethod(methodName, methodParameterTypes);
    } // try
    catch (NoSuchMethodException e) {
      // try again on the parent
      return findMethod(c.getSuperclass(), methodName, methodParameterTypes);
    } // catch
  } // findMethod()

  /**
  Finds a private, protected, or public method in the specified object.
  */
  public static Method getMethod(Object object, String methodName, Class[] methodParameterTypes)
    throws ClassNotFoundException, NoSuchMethodException,
    SecurityException, InstantiationException, InvocationTargetException, IllegalAccessException {
//      try {
      final Class c = object.getClass();
      final Method method = findMethod(c, methodName, methodParameterTypes);

      // the following call skirts private/protected access
      // protection unless the security manager prevents it
      method.setAccessible(true);

      return method;
//      } // try
//      catch (IllegalAccessException e) {
//        // due to the setAccessible() call above, should not reach here
//        throw new SoftwareFaultException("access should be legal", e);
//      } // catch
  } // getMethod()
} // class ClearBox
