// Java Test Bed: tools for testing Java classes
// Copyright 1999 Sims Computing
// You may use this software any way you like. Use at your own risk.

package com.simscomputing.testbed;

import com.simscomputing.SoftwareFaultException;
import com.simscomputing.testbed.NoSuchTestException;
import com.simscomputing.util.Debugger;

import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.SortedMap;
import java.util.TreeMap;

/**
Users of the Test Bed typically subclass this class and put their test
cases in the subclass. Of special note are the setupCollection(),
teardownCollection(), setup(), and teardown() methods. The first is
called before any test cases in a Collection class are exercised. The
second is called after all such test cases are exercised. Finally,
setup() and teardown() are called before and after each test case.

<p>These methods allow the programmer to perform initialization and
cleanup.

@author David Sims
@version $Revision: 1.1.1.1.2.1 $ $Date: 2000/03/31 18:23:31 $
*/
public class TestCaseCollection {//implements Runnable {
  /** The prefix for method names that represent test cases. */
  public static final String TEST_CASE_PREFIX = "runTest";

  private List /* TestCaseCollection */ relatedTestCaseCollections = new ArrayList();
  /** Method names of test cases that are marked for execution. */ 
  private ArrayList /* String */ filtered = new ArrayList();
  private ArrayList /* Object[2] */ marked = new ArrayList();  
//  private TestResultCollection results = new TestResultCollection();
//  private int methodCount;
//    private TesterListener exerciseListener;
  private StringBuffer testCaseInProgress = new StringBuffer();
  private boolean stopped = false;

  public TestCaseCollection() {
//      relatedTestCaseCollections.add(this);
  } // constructor

  /**
  Returns the number of test cases in this Test Case Collection and
  all related Test Case Collections.
  */
  public int getNumberOfTestCases() {
    return examineMethods(true, null, null, 0);
  } // getNumberOfTestCases()

  public void stop() {
    stopped = true;

    TestCaseCollection tcc;
    for (int index = 0; index < getRelatedTestCaseCollections().size(); ++index) {
      tcc = (TestCaseCollection) getRelatedTestCaseCollections().get(index);
      tcc.stop();
    } // for index
  } // stopTestRun;
  
  public boolean wasStopped() {
    return stopped;
  }

  public void reset() throws TesterException {
//    results = new TestResultCollection();
    relatedTestCaseCollections = new ArrayList();
    stopped = false;
  } // reset()

  public TestResultCollection exercise(final boolean runNow,
                                       final TesterListener exerciseListener)
      throws TesterException {
    final int methodCount = examineMethods(runNow, null, exerciseListener, 0);
    return exercise(runNow, exerciseListener, new TestResultCollection(), methodCount);
  } // exercise()

  /**
  Exercises the TestCaseCollection. Progress is monitored via the
  optional TesterListener interface. Returns the results of the tests.

  @return TestResultCollection the results of the tests
  */
  public TestResultCollection exercise(final boolean runNow,
                                       final TesterListener exerciseListener,
                                       final TestResultCollection r,
                                       final int methodCount)
      throws TesterException {
    reset();
    markRelated();

//      this.exerciseListener = exerciseListener;

// moved to the exercise() method above
//    final int methodCount = examineMethods(null, exerciseListener, 0);

//    examineMethods(results, exerciseListener, methodCount);
    TestResultCollection trc = new TestResultCollection();
    examineMethods(runNow, r, exerciseListener, methodCount);
    return r; // results;
  } // exercise()

  private int examineMethods(final boolean runNow,
                             final TestResultCollection testResultCollection,
                             final TesterListener exerciseListener,
                             final int totalTests) {
    int numberOfMethods = 0;

    // store all test results here
//      TestResultCollection testResultCollection = new TestResultCollection();

//      // store all testable methods found here
//      ArrayList /* ClassAndMethodPair */ testableMethods = new ArrayList();

    // the test results from each of the testable methods
//      SortedMap /* TestResult */ testResults = new TreeMap();

    // for each method starting with "runTest" in each class in 'relatedTestCaseCollections', 
    // call that method and collect the results

    List relatedTccs = new ArrayList();
    relatedTccs.add(this);

    /** Begin D Lamy code block **/
    // Grab the default streams so we can revert later
    PrintStream defaultOut = System.out;
    PrintStream defaultErr = System.err;
    /** End D Lamy code block **/

    for (int index = 0; index < relatedTccs.size(); ++index) {
      final Object object = relatedTccs.get(index);
      final Class cl = object.getClass();
//        if (testResultCollection != null) {
//          System.out.println("related class name = " + cl.getName());
//          System.out.println("toString: " + relatedTccs.get(index).toString());
//        } // if
      final TestCaseCollection exercisee = (TestCaseCollection) relatedTccs.get(index);
  		final Method[] methods = exercisee.getTestCaseMethods();

//        System.out.println("exercisee = " + exercisee);

      // before each TestCaseCollection runs, call setupCollection()
      // to allow the collection to setup. However, the Collection's
      // constructor does this job just fine. But, this method is here
      // for completeness.

      // fixme: maybe should catch Throwable here too. Could come out of setup()
      // and friends.

      // this 'if' means we're not just counting methods, we're running the test cases
      boolean callTeardown = false;
      if (testResultCollection != null) {
        callTeardown = true;
        setupCollection();
      } // if

      for (int jndex = 0; !wasStopped() && jndex < methods.length; ++jndex) {
//          System.out.println("method = " + methods[jndex]);
//        Class returnType = methods[jndex].getReturnType();
//        Class[] parameters = methods[jndex].getParameterTypes();

        // see if this method is a test case
        if (isMethodTestCase(methods[jndex])) {
//        if (methods[jndex].getName().startsWith(TEST_CASE_PREFIX) &&
//            returnType.toString().equals("void") &&
//            parameters.length == 0) {

//            System.out.println(getName() + ": bump MC on " + methods[jndex].getName());
          ++numberOfMethods;

          if (testResultCollection != null) {
              // exercise this test case
              exerciseTestCase(runNow,
                               methods[jndex], testResultCollection,
                               exerciseListener,
                               numberOfMethods - 1,
                               totalTests,
                               exercisee);
          } // if
        } // if
      } // for jndex

      /** Begin D Lamy code block **/
      // Reset the system out and err streams back to the
      // defaults
      System.setOut(defaultOut);
      System.setErr(defaultErr);
      /** End D Lamy code block **/

      // after each TestCaseCollection runs, call teardownCollection()
      // to allow the collection to clean up

      // fixme: maybe should catch Throwable here too. Could come out of setup()
      // and friends.

      // this 'if' means we're not just counting methods, we're running the test cases
      if (callTeardown) {
        teardownCollection();
      } // if

      for (int kndex = 0;
           kndex < exercisee.getRelatedTestCaseCollections().size();
           ++kndex) {
//          System.out.println("************** # related tccs = " +
//                             exercisee.getRelatedTestCaseCollections().size());

        TestCaseCollection tcc = (TestCaseCollection) exercisee.getRelatedTestCaseCollections().get(kndex);
        if (tcc != this) {
          numberOfMethods += tcc.examineMethods(runNow, testResultCollection, exerciseListener, totalTests);
        } // if
      } // for kndex

    } // for index

    // note: make sure this final notification is always sent!

    // send one final progress notification
    if (testResultCollection != null && exerciseListener != null &&
        runNow) {
      TesterProgress tp;

      // since the first two arguments are the same,
      // the progress bar is guaranteed to be removed
      tp = new TesterProgress(totalTests, //testResultCollection.getNumberOfTests(),
                              totalTests,
                              "",
                              (TestResult[]) testResultCollection.getClonedTestResults());
      exerciseListener.progressMade(tp);
    } // if

    return numberOfMethods;
  } // examineMethods()

//      // collect methods that start with 'runTest'
//      collectRelatedTestCaseMethods(this, testableMethods);
//      final List collections = getrelatedTccs();
//      for (int index = 0; index < collections.size(); ++index) {
//        collectRelatedTestCaseMethods( (TestCaseCollection) collections.get(index), testableMethods);
//      } // for index

//      // variables used for timing the execution of the test methods
//      long startTime;
//      long endTime;
//      boolean result;
//      long executionTime;
//      TestResult testResult;
//      String testName;
//      Method method;
//      String className;
//      HashMap testCaseCollectionObjects = new HashMap();
//      Throwable failureException;

//      // run each of the collected test methods
//      for (int index = 0; index < testableMethods.size(); ++index) {
//        failureException = null;

//        // the following assignments are here to appease the compile
//        executionTime = 0;
//        result = false;
//        startTime = 0;
//        try {
//  				class = ((ClassAndMethodPair) testableMethods.get(index)).class;
//  				method = ((ClassAndMethodPair) testableMethods.get(index)).method;
//  				Object object = testCaseCollectionObjects.get(className);
//  				if (object == null) {
//  					object = Class.forName(className).newInstance();
//  					testCaseCollectionObjects.put(className, object);
//  				} // if
//          startTime = System.currentTimeMillis();
//          method.invoke(object, null);
//          endTime = System.currentTimeMillis();
//          executionTime = endTime - startTime;
//  				result = true;
//        } // try
//        catch (ClassNotFoundException e) {
//          Debugger.assert(false, "should not reach here");
//        } // catch
//        catch (IllegalAccessException e) {
//          Debugger.assert(false, "should not reach here 2");
//        } // catch
//        catch (InstantiationException e) {
//          Debugger.assert(false, "should not reach here 3");
//        } // catch
//        catch (InvocationTargetException e) {
//          endTime = System.currentTimeMillis();
//          executionTime = endTime - startTime;
//          result = false;
//  				failureException = e.getTargetException();
//        } // catch

//        // store the results of the test
//        testName = ((ClassAndMethodPair) testableMethods.get(index)).class.getName() + "." +
//  				((ClassAndMethodPair) testableMethods.get(index)).method.getName();
//        testResult = new TestResult(testName, result, executionTime, failureException);
//        testResultCollection.add(testResult);
//      } // for index




  /**
  Returns a list of TestCaseCollection objects that are related to
  this TestCaseCollection. Several TestCaseCollection objects can be
  grouped together to form a kind of <em>test suite</em>.
  */
  public List /* TestCaseCollection */ getRelatedTestCaseCollections() {
    return relatedTestCaseCollections;
  } // getRelatedTestCaseCollections()

  /**
  Adds a new TestCaseCollection to the list of related
  TestCaseCollections for this particular TestCaseCollection.
  */
  public void addTestCaseCollection(TestCaseCollection testCaseCollection) {
//      throws ClassCastException, ClassNotFoundException, IllegalAccessException,
//  		InstantiationException {
    Debugger.checkPrecondition(testCaseCollection != null, "testCaseCollection must not be null");

//      Object object = Class.forName(testCaseCollection).newInstance();
    relatedTestCaseCollections.add(testCaseCollection);
  } // addTestCaseCollection()

//    /**
//  		 Collects methods that start with 'runTest' from the specified class name and inserts
//  		 found methods int 'testableMethods'. A reference to the 'testableMethods' parameter
//  		 is returned.

//  		 @param String the name of the class that extends TestCaseCollection and contains test cases
//  		 @param ArrayList the list of methods and class names found are added to this list
//  		 @return a reference to the (possibly) modified 'testableMethods' object
//    */
//    private ArrayList /* Object[] */ collectRelatedTestCaseMethods(TestCaseCollection testCaseCollection,
//  																																 ArrayList /* ClassAndMethodPair */ testableMethods) {

//  		Class returnType;
//  		Class[] parameters;
//  		Method[] methods = testCaseCollection.getClass().getDeclaredMethods();

//  		for (int index = 0; index < methods.length; ++index) {
//  			returnType = methods[index].getReturnType();
//  			parameters = methods[index].getParameterTypes();

//  			if (methods[index].getName().startsWith("runTest") &&
//  					returnType.toString().equals("void") &&
//  					parameters.length == 0) {
//  				ClassAndMethodPair pair = new ClassAndMethodPair(methods[index],
//                                                           testCaseCollection.getClass().getName());
//  				testableMethods.add(pair);
//  			} // if
//  		} // for index

//      return testableMethods;
//    } // collectRelatedTestCaseMethods()

  public static void assert(boolean condition, String message) throws TestFailedException {
    if (!condition) {
      throw new TestFailedException(message);
    } // if
  } // assert()

  public String getName() {
    return "<TestCaseCollection>";
  } // getName()

  /**
  Exercises a single test case.
  */
  protected void exerciseTestCase(final boolean runNow,
                                  final Method testCaseMethod,
                                  final TestResultCollection testResultCollection,
                                  final TesterListener exerciseListener,
                                  final int testedSoFar,
                                  final int totalTests,
                                  final TestCaseCollection exercisee) {

    if (!runNow) {
      // don't actually run the test case
      final String testCaseName = testCaseMethod.getName();
      final char[] shortTestCaseName =
        new char[testCaseName.length() - TEST_CASE_PREFIX.length()];
      testCaseName.getChars(TEST_CASE_PREFIX.length(), testCaseName.length(), shortTestCaseName, 0);

      TestResult testResult = new TestResult(exercisee,
                                             new String(shortTestCaseName));
      synchronized (testResultCollection) {
        testResultCollection.add(testResult);
      } // synchronized
      return;
    } // if

    long startTime = System.currentTimeMillis();
    long executionTime = 0;
    long endTime = 0;
    boolean result = true;
    Throwable failureException = null;

    synchronized (testCaseInProgress) {
      testCaseInProgress.replace(0, testCaseInProgress.length(), testCaseMethod.getName());
    } // synchronized
    if (exerciseListener != null) {
      TesterProgress tp;

//      System.out.println("tested so far = " + testedSoFar);

      tp = new TesterProgress(testedSoFar, //testResultCollection.getNumberOfTests(),
                              totalTests,
                              testCaseInProgress.toString(),
                              (TestResult[]) testResultCollection.getClonedTestResults());
      exerciseListener.progressMade(tp);
    } // if

    /** Begin D Lamy code block **/
    ByteArrayOutputStream outStream = null;
    ByteArrayOutputStream errStream = null;
    /** End D Lamy code block **/
    int executingTeardown = 0;
    try {
      // Begin D Lamy code block
      /*
              Redirect system.out and system.err to these streams for
              display in the testbed GUI later on
              */
      outStream = new ByteArrayOutputStream();
      errStream = new ByteArrayOutputStream();
      System.setOut(new PrintStream(outStream, false));
      System.setErr(new PrintStream(errStream, false));
      // end D Lamy code block
      executingTeardown = 1;
      setup();
      startTime = System.currentTimeMillis();
      testCaseMethod.invoke(exercisee, null);
      endTime = System.currentTimeMillis();
      executingTeardown = 2;
      teardown();
      executionTime = endTime - startTime;
    } // try
    catch (IllegalAccessException e) {
      throw new SoftwareFaultException("should not reach here");
    } // catch
    catch (InvocationTargetException e) {
      endTime = System.currentTimeMillis();
      executionTime = endTime - startTime;
      result = false;
      failureException = e.getTargetException();
    } // catch

    // fixme: maybe should catch Throwable here too. Could come out of setup()
    // and friends.

    finally {
      synchronized (testCaseInProgress) {
        testCaseInProgress.delete(0, testCaseInProgress.length());
      } // synchronized
      // if an exeception is thrown out of the test case, we may
      // still need to call teardown()
      if (executingTeardown == 1) {
        teardown();
      } // if
    } // finally

    // store the results of the test
    String testCaseCollectionName = exercisee.getClass().getName();
    String testCaseName = testCaseMethod.getName();
    // fixme
    char[] shortTestCaseName = new char[testCaseName.length() - TEST_CASE_PREFIX.length()];
    testCaseName.getChars(TEST_CASE_PREFIX.length(), testCaseName.length(), shortTestCaseName, 0);

    /** Begin D Lamy code block modification **/
    TestResult testResult = new TestResult(exercisee,
                                           new String(shortTestCaseName),
                                           result,
                                           executionTime,
                                           failureException,
                                           outStream,
                                           errStream);    
    /** End D Lamy code block modification **/
    synchronized (testResultCollection) {
      testResultCollection.add(testResult);
    } // synchronized

    if (exerciseListener != null) {
      TesterProgress tp;

      tp = new TesterProgress(testedSoFar, //testResultCollection.getNumberOfTests(),
                              totalTests,
                              testCaseInProgress.toString(),
                              (TestResult[]) testResultCollection.getClonedTestResults());
      exerciseListener.progressMade(tp);
    } // if

    // Allow chance to write output to file.
    logStdOut(testCaseCollectionName, 
	      new String(shortTestCaseName),
	      outStream);

    logStdError(testCaseCollectionName, 
		new String(shortTestCaseName),
		errStream);
  } // exerciseTestCase()

  /**
  Determines whether the specified method has the proper
  signature for a test case.
  */
  protected boolean isMethodTestCase(final Method method) {
    final Class returnType = method.getReturnType();
    final Class[] parameters = method.getParameterTypes();

    return (method.getName().startsWith(TEST_CASE_PREFIX) &&
        returnType.toString().equals("void") &&
        parameters.length == 0);
  } // isMethodTestCase()

  /**
  Returns the relevant test case methods from this test case
  collection. 
  */
  protected Method[] getTestCaseMethods() {
    Method[] result = getClass().getDeclaredMethods();

    if (filtered.size() == 0) {
//      System.out.println("!!!! no filtering done in " + getClass().getName());
      return result;
    } // if

    final ArrayList filteredResult = new ArrayList();
    String filteredMethodName;
    Method method;
    for (int i = 0; i < filtered.size(); ++i) {
      for (int j = 0; j < result.length; ++j) {
        filteredMethodName = (String) filtered.get(i);
        if (result[j].getName().equals(filteredMethodName)) {
//          System.out.println("!!!! filtering done on " + filteredMethodName);
          filteredResult.add(result[j]);
        } // if
      } // for j
    } // for i

    // finally, convert the method list to an array
    result = new Method[filteredResult.size()];
    for (int i = 0; i < result.length; ++i) {
      result[i] = (Method) filteredResult.get(i);
    } // for i

    return result;
  } // getTestCaseMethods()

  public void clearMarkedMethods() {
    filtered.clear();
    marked.clear();

    TestCaseCollection tcc;
    for (int index = 0; index < getRelatedTestCaseCollections().size(); ++index) {
      tcc = (TestCaseCollection) getRelatedTestCaseCollections().get(index);
      tcc.clearMarkedMethods();
    } // for index
  } // clearMarkedMethods()

  public void markForTesting(final String testCaseCollectionClassName,
                             final String testCaseName) {
    Object[] strings = new Object[2];
    strings[0] = testCaseCollectionClassName;
    strings[1] = testCaseName;
    marked.add(strings);
  } // markForTesting()

  private void markRelated(String testCaseCollectionClassName, String testCaseName) {

      if (getClass().getName().equals(testCaseCollectionClassName)) {
        filtered.add(TEST_CASE_PREFIX + testCaseName);
//        System.out.println("marked: <" + TEST_CASE_PREFIX + testCaseName +
//                           "> in " + getClass().getName());
//        System.out.println("filtered size = " + filtered.size());
      } // if
      else {
        TestCaseCollection tcc;
        for (int index = 0; index < getRelatedTestCaseCollections().size(); ++index) {
          tcc = (TestCaseCollection) getRelatedTestCaseCollections().get(index);
          tcc.markRelated(testCaseCollectionClassName, testCaseName);
//fixme        tcc.markForTesting(testCaseCollectionClassName,
//                           testCaseName);

        } // for index
      } // else

  } // markRelated()

  private void markRelated() {
    for (int i = 0; i < marked.size(); ++i) {
      Object[] strings = (Object[]) marked.get(i);
      String testCaseCollectionClassName = (String) strings[0];
      String testCaseName = (String) strings[1];

//      System.out.println("markRelated: class name = " + testCaseCollectionClassName +
//                         ", test cae name = " + testCaseName);

      if (getClass().getName().equals(testCaseCollectionClassName)) {
        filtered.add(TEST_CASE_PREFIX + testCaseName);
//        System.out.println("marked: <" + TEST_CASE_PREFIX + testCaseName +
//                           "> in " + getClass().getName());
//        System.out.println("filtered size = " + filtered.size());
      } // if
      else {
        TestCaseCollection tcc;
        for (int index = 0; index < getRelatedTestCaseCollections().size(); ++index) {
          tcc = (TestCaseCollection) getRelatedTestCaseCollections().get(index);
          tcc.markRelated(testCaseCollectionClassName, testCaseName);
//fixme        tcc.markForTesting(testCaseCollectionClassName,
//                           testCaseName);

        } // for index
      } // else
    } // for i

  } // markForTesting()

  /**
  Called before each test case runs.
  */
  public void setup() {
    // do nothing
  } // setup()

  /**
  Called after each test case runs.
  */
  public void teardown() {
    // do nothing
  } // teardown()

    /**
     * Called after each test case is run.  Allows the opertunity for
     * a test case to log all standard output.
     **/
    public void logStdOut(String testCaseCollectionName, 
			  String shortTestCaseName,
			  ByteArrayOutputStream outStream) {
	// Do nothing
    }

    /**
     * Called after each test case is run.  Allows the opertunity for
     * a test case to log all standard error.
     **/
    public void logStdError(String testCaseCollectionName, 
			    String shortTestCaseName,
			    ByteArrayOutputStream errStream) {
	// Do nothing
    }

  /**
  Called before any test case in a collection class runs.
  */
  public void setupCollection() {
    // do nothing
  } // setupCollection()

  /**
  Called after all test cases in a collection class have run.
  */
  public void teardownCollection() {
    // do nothing
  } // teardownCollection()
} // class TestCaseCollection
