/*
 * Enhydra Java Application Server Project
 *
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific terms governing rights and limitations
 * under the License.
 *
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 *
 * Contributor(s):
 *
 */
package org.enhydra.tool.codegen;

// ToolBox imports

import org.enhydra.tool.ToolBoxInfo;
import org.enhydra.tool.common.FileUtil;
import org.enhydra.tool.common.PathHandle;
import org.enhydra.tool.common.Replacement;
import org.enhydra.tool.common.ReplacementSet;
import org.enhydra.tool.common.ResUtil;
import org.enhydra.tool.common.ToolException;
import org.enhydra.tool.common.TemplateFile;
import org.enhydra.tool.common.event.TemplateListener;
import org.enhydra.tool.configure.ConfigTool;

// Standard imports

import java.io.File;
import java.io.FileFilter;
import java.util.Properties;
import java.util.ResourceBundle;
import java.util.Vector;

/**
 * A partial implementation of the Generator interface that works with
 * template files. The predefined generators are based on this class.
 *
 */
abstract public class TemplateDrivenGenerator implements Generator,
        TemplateListener, Constants {
    public static ResourceBundle       res =
        ResourceBundle.getBundle ("org.enhydra.tool.codegen.Res");      // nores

    //

    private Properties          properties = new Properties ();
    private File                templateDirectory = null;
    private File                rootDir = null;
    private Vector              fileVector = new Vector ();
    private ReplacementSet      replacementSet = null;
    private OptionSet           optionSet = null;
    private FileFilter          toolFilter = null;
    private boolean             overwrite = false;
    private boolean             echo = true;
    private boolean             swing = true;

    //

    abstract public TemplateFile[] preprocessTemplates (TemplateFile[] temps);
    abstract public File[] preprocessCopyList (File[] files);
    abstract public FileFilter getTemplateFilter ();
    abstract public FileFilter getCopyFilter ();
    abstract public String getDestinationPath () throws GeneratorException;
    abstract public void initReplacementSet () throws GeneratorException;
    abstract public Replacement[] createReplacementsForDirectory (PathHandle path)
            throws GeneratorException;

    /**
     * Use the display name for the string representation of this class
     * so that a list model can map directly into generators.
     *
     * @return
     * The display name.
     */
    public String toString () {
        return getDisplayName ();
    }

    /**
     * Get properties used to customize CodeGen. This may contain
     * information on the template directory location and
     * persisted generation options.
     *
     * @return
     * Properties passed from the CodeGen class.
     */
    public Properties getProperties () {
        return properties;
    }

    /**
     * Set CodeGen customization properties. The CodeGen
     * class calls this method after it instantiates a
     * generator class.
     *
     * @param properties
     * Properties passed from the CodeGen class.
     */
    public void setProperties (Properties p) throws GeneratorException {
        StringBuffer    buf = new StringBuffer ();
        String          path = new String ();

        properties = p;

        buf.append (ToolBoxInfo.getEnhydraPath ());
        buf.append (File.separator);
        buf.append (DIR_TOOL);
        buf.append (File.separator);
        buf.append (DIR_CODEGEN);
        buf.append (File.separator);
        buf.append (getCommandName ());

        path = FileUtil.toCurrentPath (buf.toString ());
        templateDirectory = new File (path);

        if (!templateDirectory.isDirectory ()) {
            throw new GeneratorException (ResUtil.format (res.getString ("Invalid_template_0"),
                    templateDirectory));
        }

        // read in persisteted option values

        GeneratorOption[]       options = getOptionSet ().toArray ();

        for (int i = 0; i < options.length; i++) {
            String      name = options[i].getName ();
            String      value = options[i].getValue ();

            value = properties.getProperty (getCommandName () + '.' + name,
                                            value);

            if (value != null) {
                if (options[i].isRequired ()) {
                    if (value.trim ().length () > 0) {
                        getOptionSet ().lookup (name).setValue (value);
                    }
                } else {
                    getOptionSet ().lookup (name).setValue (value);
                }
            }
        }
    }

    /**
     * Get the location of the template files.
     *
     * @returns
     * The directory containing template files.
     */
    public File getTemplateDirectory () {
        return templateDirectory;
    }

    public boolean isOverwrite () {
        return overwrite;
    }

    public void setOverwrite (boolean b) {
        overwrite = b;
    }

    public boolean isEcho () {
        return echo;
    }

    public void setEcho (boolean b) {
        echo = b;
    }

    public boolean isSwing () {
        return swing;
    }

    public void setSwing (boolean b) {
        swing = b;
    }

    /**
     * Get the replacement operations to run against the templates. If this is
     * null or has a zero length, no replacements are performed during generate().
     *
     * @returns
     * Array of placeholders to search for and the
     * strings use for replacement.
     */
    public ReplacementSet getReplacementSet () {
        return replacementSet;
    }

    /**
     * Set the replacement operations to run against the templates. If this is
     * not called, called prior to generate(), no replacements are performed.
     *
     * @param replacements
     * Array of placeholders to search for and the
     * strings use for replacement.
     */
    public void setReplacementSet (ReplacementSet replacementSet) {
        this.replacementSet = replacementSet;
    }

    /**
     * Get generator option set
     *
     * @return
     * An option set that you can use to read and write generator
     * options.
     */
    public OptionSet getOptionSet () {
        return optionSet;
    }

    public void setOptionSet (OptionSet set) {
        optionSet = set;
    }

    /**
     * Generate application files from a set of templates. This implementation of
     * the generate method uses the replacements array to perform a search and
     * replace on generated file names as well as there contents.
     *
     * @return
     * An array of the files that were generated.
     *
     * @throws GenerationException
     * Thrown if unable to generate all the files. Note that generation will fail
     * if the template directory has not been set.
     */
    public File[] generate () throws GeneratorException {
        initReplacementSet ();

        File[]  files = new File[0];

        rootDir = new File (getDestinationPath ());
        toolFilter = getTemplateFilter ();
        fileVector = new Vector ();

        if (templateDirectory == null) {
            throw new GeneratorException (res.getString ("Template_null"));
        }

        try {
            recurseFolders (templateDirectory, rootDir);
        } catch (ToolException e) {
            String      mess = new String ();

            mess = ResUtil.format (res.getString ("Failed_to_copy"),
                                   templateDirectory, rootDir);

            throw new GeneratorException (e, mess);
        }

        storeProperties ();

        files = (File[]) fileVector.toArray (files);

        fileVector.removeAllElements ();

        return files;
    }

    // /
    // / PRIVATE METHODS
    // /

    /**
     * Method declaration
     *
     *
     */
    public void storeProperties () {

        // persist options

        GeneratorOption[]       options = getOptionSet ().toArray ();

        for (int i = 0; i < options.length; i++) {
            if (options[i].isPersistent ()) {
                properties.setProperty (getCommandName () + '.'
                                        + options[i].getName (), options[i].getValue ());
            }
        }

        // save

        try {
            ToolBoxInfo.storeProperties (properties);
        } catch (ToolException e) {

            // non fatal, just print the stack trace.

            e.printStackTrace ();
        }
    }

    /**
     * Method declaration
     *
     *
     * @param sourceDir
     * @param destDir
     *
     * @exception java.io.IOException
     */
    private void recurseFolders (File source, File dest)
            throws ToolException, GeneratorException {
        File[]          toolList = new File[0];
        File[]          copyList = new File[0];
        File[]          output = new File[0];
        ConfigTool      tool = null;
        String          destPath = null;

        tool = new ConfigTool ();

        tool.addTemplateListener (this);
        tool.setReplacements (replacementSet.toArray ());
        tool.setOverwrite (isOverwrite ());
        tool.setEcho (isEcho ());
        tool.setSwing (isSwing ());

        toolList = source.listFiles (toolFilter);
        copyList = source.listFiles (getCopyFilter ());
        destPath = dest.getAbsolutePath ();
        destPath = tool.lineSearchAndReplace (destPath);
        dest = new File (destPath);

        // copy non-templates

        if (copyList != null && copyList.length >= 1) {
            copyList = preprocessCopyList (copyList);

            for (int i = 0; i < copyList.length; i++) {
                if (copyList[i].isDirectory ()) {
                    File        nextDest = null;

                    nextDest = new File (dest, copyList[i].getName ());

                    recurseFolders (copyList[i], nextDest);
                } else {
                    dest.mkdirs ();

                    File        outCopy = null;

                    outCopy = FileUtil.copyFile (copyList[i], dest);

                    fileVector.addElement (outCopy);
                }
            }
        }

        // process templates

        if (toolList != null && toolList.length >= 1) {
            tool.initTemplates (toolList, dest, TYPE_TEMPLATE);
            tool.setTemplates (preprocessTemplates (tool.getTemplates ()));

            if (tool.getTemplates ().length >= 1) {
                Replacement[]   pathReplacements = new Replacement[0];
                String          outPath = new String ();

                if (tool.getTemplates ()[0].getOutput ().isDirectory ()) {
                    outPath =
                        tool.getTemplates ()[0].getOutput ().getAbsolutePath ();
                } else {
                    outPath =
                        tool.getTemplates ()[0].getOutput ().getParent ();
                }

                pathReplacements =
                    createReplacementsForDirectory (PathHandle.createPathHandle (outPath));

                tool.setReplacements (pathReplacements);

                output = tool.createOutput ();
                overwrite = tool.isOverwrite ();

                for (int i = 0; i < output.length; i++) {
                    fileVector.addElement (output[i]);
                }

                for (int i = 0; i < toolList.length; i++) {
                    if (toolList[i].isDirectory ()) {
                        File    nextDest = null;

                        nextDest = new File (dest, toolList[i].getName ());

                        recurseFolders (toolList[i], nextDest);
                    }
                }
            }
        }
    }

}

