/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id$
 */

package org.enhydra.xml.dom;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.enhydra.xml.xmlc.XMLObject;
import org.enhydra.xml.lazydom.LazyDocument;

/**
 * Various static methods the do simple operations on the DOM.
 */
public final class DOMOps {
    /**
     * Prevent instantiation.
     */
    private DOMOps() {
    }

    /**
     * Get the owner document object for a node.  This gets the actual
     * document.  If a XMLC document class instance the contained document is
     * returned.
     */
    public static Document getDocument(Node node) {
        if (node instanceof XMLObject) {
            // Get real document from XMLC.
            return ((XMLObject)node).getDocument();
        } else if (node instanceof Document) {
            return (Document)node;
        } else {
            return node.getOwnerDocument();
        }
    }

    /**
     * Get the real DOM node, bypassing an XMLC-generated container. If a XMLC
     * document class instance the contained document is returned.  Otherwise
     * the node is returned.
     */
    public static Node getActualNode(Node node) {
        if (node instanceof XMLObject) {
            return ((XMLObject)node).getDocument();
        } else {
            return node;
        }
    }

    /**
     * Replace a node with one imported from another document.
     * @param srcNode The node to clone and insert 
     * @param destNode The node to be replaced
     * @return The new node that replaces the destination node.
     */
    public static Node replaceNode(Node srcNode, Node destNode) {
        Node node = destNode.getOwnerDocument().importNode(srcNode, true);
        Node parent = destNode.getParentNode(); 
        if (parent != null){
            parent.replaceChild(node, destNode);
        }
        return node;
    }

    /**
     * Count the number of children in a node has.
     */
    public static int countChildren(Node node) {
        int cnt = 0;
        for (Node child = node.getFirstChild(); child != null;
             child = child.getNextSibling()) {
            cnt++;
        }
        return cnt;
    }

    /**
     * Determine if a document is a LazyDOM instance document.  
     * @return true if its a LazyDOM instance, false if its a LazyDOM template
     *  or not a LazyDOM at all.
     */
    public static boolean isLazyDOMInstance(Document doc) {
        return ((doc instanceof LazyDocument) && !((LazyDocument)doc).isTemplateNode());
    }
}
