/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: XMLFormatter.java,v 1.8.2.1.4.7 2000/12/20 22:54:18 markd Exp $
 */

package org.enhydra.xml.io;

import org.apache.xml.serialize.XMLSerializer;
import org.apache.xml.serialize.DOMSerializer;
import org.apache.xml.serialize.OutputFormat;
import java.io.Writer;
import java.io.IOException;
import java.io.OutputStream;
import org.w3c.dom.*;

/*
 * FIXME: 
 *  - Add option to encode entity references from the DTD.
 *  - namespaces
 *  - writespace
 *  - Need to handle entities for 8bit encodings
 */

/**
 * Formatter for outputting a HTML DOM as a HTML text document.
 */
final class XMLFormatter extends BaseDOMFormatter implements Formatter {
    /**
     * Default XML encoding.
     */
    private static final String DEFAULT_XML_ENCODING = "UTF-8";

    /**
     * XML version (only 1.0 supported, as thats all there is now).
     */
    private static final String XML_VERSION = "1.0";

    /**
     * Table use to optimized checking for characters that should be
     * represented as entity references.
     */
    private static final boolean[] fEntityQuickCheck
        = new boolean[MAX_ENTITY_QUICK_CHECK_CHAR+1];

    /**
     * Static constructor.
     */
    static {
        fEntityQuickCheck['<'] = true;
        fEntityQuickCheck['>'] = true;
        fEntityQuickCheck['"'] = true;
        fEntityQuickCheck['\''] = true;
        fEntityQuickCheck['&'] = true;
    }

    /**
     * Constructor.
     */
    public XMLFormatter(Node node,
                        OutputOptions outputOptions,
                        boolean forPreFormatting) {
        super(node, outputOptions, forPreFormatting, DEFAULT_XML_ENCODING, fEntityQuickCheck);
    }

    /**
     * Get the default OutputOptions for a document formatter with
     * this formatter.
     */
    static OutputOptions getDefaultOutputOptions() {
        OutputOptions options = new OutputOptions();
        options.setXmlEncoding(DEFAULT_XML_ENCODING);
        return options;
    }

    /**
     * @see BaseDOMFormatter#getCharacterEntity
     */
    protected final String getCharacterEntity(char textChar) {
        switch (textChar) {
	case '<':
            return "lt";
	case '>':
            return "gt";
	case '"':
            return "quot";
	case '\'':
            return "apos";
	case '&':
            return "amp";
        default:
            return null;
        }
    }

    /**
     * Write the XML header.
     */
    private void writeXMLHeader(Document document) throws IOException {
        fOut.write("<?xml version=\"");
        fOut.write(XML_VERSION);
        fOut.write("\"");

        // Encoding
        if (!fOptions.getOmitEncoding()) {
            fOut.write(" encoding=\"");
            fOut.write(getXmlEncoding());
            fOut.write('"');
        }

        //FIXME: this always cause expansion of DocumentType...
        // Standalone
        if ((fPublicId == null) && (fSystemId == null)) {
            fOut.write(" standalone=\"yes\"");
        }

        fOut.write("?>");
        writeln();
    }

    /**
     * Handler called for Document nodes; creates the XML file header.
     * @see DOMTraversal.Handler#handleDocument
     */
    public void handleDocument(Document document) throws IOException {
        if (!fOptions.getOmitXMLHeader()) {
            writeXMLHeader(document);
        }
        fTraverser.processDocumentType(document);
        fTraverser.processChildren(document);
    }

    /**
     * Handler called for Document nodes; writes the DOCTYPE specification.
     * @see DOMTraversal.Handler#handleDocumentType
     */
    public void handleDocumentType(DocumentType documentType) throws IOException {
        if (fOptions.getOmitDocType()) {
            return; // Don't output DOCTYPE.
        }
        String internalSubset = documentType.getInternalSubset();
        if ((fPublicId == null) && (fSystemId == null)
            && (internalSubset == null)) {
            return; // No DOCTYPE
        }
        Element docElement = fDocument.getDocumentElement();
        if (docElement == null) {
            throw new XMLIOError("Document has DocumentType, with out having a root element"); 
        }

        fOut.write("<!DOCTYPE ");
        fOut.write(docElement.getTagName());
        
        if (fPublicId != null) {
            // Public requires system.
            if (fSystemId == null) {
                throw new XMLIOError("No SYSTEM id to accompany PUBLIC id: "
                                     + fPublicId);
            }
            fOut.write(" PUBLIC \"");
            fOut.write(fPublicId);
            fOut.write("\" \"");
            fOut.write(fSystemId);
            fOut.write("\"");
        } else if (fSystemId != null) {
            fOut.write(" SYSTEM \"");
            fOut.write(fSystemId);
            fOut.write("\"");
        }
        
        if (internalSubset != null) {
            writeln();
            fOut.write(" [");
            fOut.write(internalSubset);
            fOut.write("]");
        }
        fOut.write('>');
        writeln();
    }

    /**
     * Handler called for DocumentFragment nodes; just process children
     * @see DOMTraversal.Handler#handleDocumentFragment
     */
    public void handleDocumentFragment(DocumentFragment documentFragment) {
        fTraverser.processChildren(documentFragment);
    }

    /**
     * Handler called for Attr nodes.
     * @see DOMTraversal.Handler#handleAttr
     */
    public void handleAttr(Attr attr) throws IOException {
        fOut.write(' ');
        fOut.write(attr.getName());
        writeAttributeValue(attr);
    }

    /**
     * Write an element open tag.
     */
    private void writeOpenTag(Element element, 
                              String tagName,
                              boolean hasChildren) throws IOException {
        String formattedTag = null;
        if (fUsePreFormattedElements && (element instanceof PreFormattedText)) {
            formattedTag = ((PreFormattedText)element).getPreFormattedText();
        }
        if (formattedTag != null) {
            fOut.write(formattedTag);
            fPreFormattedElementCount++;
        } else {
            fOut.write('<');
            fOut.write(tagName);
            fTraverser.processAttributes(element);
            if (!hasChildren) {
                fOut.write('/');
            }
            fOut.write('>');
            fDynamicFormattedElementCount++;
        }
    }

    /**
     * Handler called for Element nodes.
     * @see DOMTraversal.Handler#handleElement
     */
    public void handleElement(Element element) throws IOException {
        String tagName = element.getTagName();
        boolean hasChildren = element.hasChildNodes();

        writeOpenTag(element, tagName, hasChildren);

        // Childern and close not done when preformatting
        if (hasChildren && !fPreFormatMode) {
            fTraverser.processChildren(element);

            fOut.write("</");
            fOut.write(tagName);
            fOut.write('>');
        }
    }

    /**
     * Handler called for ProcessingInstruction nodes.
     * @see DOMTraversal.Handler#handleProcessingInstruction
     */
    public void handleProcessingInstruction(ProcessingInstruction pi) throws IOException {
        fOut.write("<?");
        fOut.write(pi.getTarget());
        String data = pi.getData();
        if (data != null) {
            fOut.write(' ');
            fOut.write(data);
        }
        fOut.write("?>");
    }

    /**
     * Handler called for CDATASection nodes.
     * @see DOMTraversal.Handler#handleCDATASection
     */
    public void handleCDATASection(CDATASection cdata) throws IOException {
        //FIXME: Should we handle embedded ]]>???
        fOut.write("<![CDATA[");
        fOut.write(cdata.getNodeValue());
        fOut.write("]]>");
    }
}
