/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: EnhydraXMLCLogger.java,v 1.1.2.1 2000/11/16 09:08:42 markd Exp $
 */

package org.enhydra.xml.xmlc;

import com.lutris.logging.LogChannel;
import com.lutris.logging.LogWriter;
import com.lutris.logging.Logger;
import java.io.PrintWriter;

/**
 * XMLC logger interface to the enhydra logger.
 */
public class EnhydraXMLCLogger implements XMLCLogger {
    /**
     * XMLC log level.  May be used by various factory implementations.
     */
    public static final String XMLC_LOG_LEVEL = "XMLC";

    /**
     * XMLC debug log level.  May be used by various factory implementations.
     */
    public static final String XMLC_DEBUG_LOG_LEVEL = "XMLC_DEBUG";

    /**
     * The log channel.
     */
    private LogChannel fLogChannel;

    /**
     * Log writer for XMLC informative logging or null if not enabled.
     */
    private LogWriter fInfoWriter;

    /**
     * Log writer for XMLC debug logging or null if not enabled.
     */
    private LogWriter fDebugWriter;

    /**
     * Log writer for XMLC error logging or null if no logging.
     */
    private LogWriter fErrorWriter;

    /**
     * Constructor
     *
     * @param logChannel The log channel to use to log events and errors.  If
     *  null, no logging is done.  Level XMLC is used to log reloading and
     *  recompilation.  Level ERROR is used for errors.
     */
    public EnhydraXMLCLogger(LogChannel logChannel) {
        if (logChannel != null) {
            fLogChannel = logChannel;
            if (logChannel.isEnabled(XMLC_LOG_LEVEL)) {
                fInfoWriter = logChannel.getLogWriter(XMLC_LOG_LEVEL);
            }
            if (logChannel.isEnabled(XMLC_DEBUG_LOG_LEVEL)) {
                fDebugWriter = logChannel.getLogWriter(XMLC_DEBUG_LOG_LEVEL);
            }
            fErrorWriter = logChannel.getLogWriter(Logger.ERROR);
        }
    }

    /**
     * Get the log channel, or null if there is none.
     */
    public LogChannel getLogChannel() {
        return fLogChannel;
    }

    /**
     * Determine if info logging is enabled.
     */
    public boolean infoEnabled() {
        return (fInfoWriter != null);
    }

    /**
     * Get the info logging writer.
     */
    public PrintWriter getInfoWriter() {
        return fInfoWriter;
    }

    /**
     * Login an info message.
     */
    public void logInfo(String msg) {
        if (fInfoWriter != null) {
            fInfoWriter.println(msg);
        }
    }

    /**
     * Login an info message with exception.
     */
    public void logInfo(String msg,
                        Throwable except) {
        if (fInfoWriter != null) {
            fInfoWriter.println(msg, except);
        }
    }

    /**
     * Determine if error logging is enabled.
     */
    public boolean errorEnabled() {
        return (fErrorWriter != null);
    }

    /**
     * Get the error logging writer.
     */
    public PrintWriter getErrorWriter() {
        return fErrorWriter;
    }

    /**
     * Login an error message.
     */
    public void logError(String msg) {
        if (fErrorWriter != null) {
            fErrorWriter.println(msg);
        }
    }

    /**
     * Login an error message with exception.
     */
    public void logError(String msg,
                         Throwable except) {
        if (fErrorWriter != null) {
            fErrorWriter.println(msg, except);
        }
    }

    /**
     * Determine if debug logging is enabled.
     */
    public boolean debugEnabled() {
        return (fDebugWriter != null);
    }

    /**
     * Get the debug logging writer.
     */
    public PrintWriter getDebugWriter() {
        return fDebugWriter;
    }

    /**
     * Login an debug message.
     */
    public void logDebug(String msg) {
        if (fDebugWriter != null) {
            fDebugWriter.println(msg);
        }
    }

    /**
     * Login an debug message with exception.
     */
    public void logDebug(String msg,
                         Throwable except) {
        if (fDebugWriter != null) {
            fDebugWriter.println(msg, except);
        }
    }
}
