/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id$
 */

package org.enhydra.xml.xmlc.dom.generic;

import org.enhydra.xml.xmlc.XMLCError;
import org.enhydra.xml.xmlc.codegen.JavaLang;
import org.enhydra.xml.dom.DOMOps;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Map;
import java.util.Iterator;
import java.lang.Comparable;
import org.w3c.dom.*;

/*
 * FIXME: If children were moved to methods that create multiple children
 * of a single node (using the mechanism for chaining, only adding a
 * lastChild argument, it should create better packed methods.
 */

//FIXME: Need better overview.

/**
 * Class to calculate which elements are to have build methods created for
 * them based on trying to optimize the amount of code (create-cost) in each
 * build methods without overflowing the maximum create-cost.
 * <P> 
 * The advantage of this object determining where build methods are created in
 * a seperate traversal from creating the build methods is that the
 * determination is best down from the bottom up, which creating the methods
 * is done top-down.  The seperation also keeps the code cleaner and easier to
 * understand.  The cost is building an intermediate table.
 */
final class BuildMethodMappings {
    /**
     * Create-cost to call a build-method.
     */
    public static final int BUILD_METHOD_CALL_CREATE_COST = 1;

    /**
     * Create-cost to create an element node.  Slightly higher due
     * to possibility of initializing access methods.
     */
    public static final int CREATE_ELEMENT_NODE_CREATE_COST = 2;

    /**
     * Maximum creation-cost that a single build method can handle.
     */
    private int fMaxCreateCostPerBuildMethod;

    /*
     * Table of node record.
     */
    private HashMap fNodeRecordTable = new HashMap();

    /**
     * Temporary table used for sorting child elements.  This table
     * never used recursively, so one can be reused for efficient.
     */
    private ArrayList fTmpElementSort = new ArrayList();

    /**
     * Node record.
     */
    private class Record implements Comparable {
        /**
         * The node we are associated with.
         */
        public Node fNode;

        /**
         * The minimum cost to create this node in a method.  This includes
         * non-element children and attribute costs, with element-children
         * costs being the cost to call another build method.
         */
        public int fMinMethodCost;

        /**
         * The total cost to create this node and its children in this build
         * method.  It includes calls to other build methods, but not the cost
         * in other build methods.
         */
        public int fMethodTotalCost;
        
        /**
         * The total cost to create this node and its children.  This
         * includes cost in other build methods, but excludes the cost
         * of calls to other build methods.  Its the cost to create the
         * tree to this point.
         */
        public int fTreeTotalCost;

        /**
         * Should this element be at the root of a build method?
         */
        public boolean fMethodRoot;

        /**
         * Should the children be created in a chained method?
         */
        public boolean fChainedChildrenMethods;

        /**
         * Constructor.  Initialize base costs.
         */
        public Record(Node node) {
            fNode = node;
            if (node instanceof Document) {
                fMinMethodCost = 0;
            } else if (fNode instanceof Element) {
                fMinMethodCost = CREATE_ELEMENT_NODE_CREATE_COST;
            } else {
                fMinMethodCost = 1;
            }
            fMethodTotalCost = fMinMethodCost;
            fTreeTotalCost = fMinMethodCost;

            // Add self to table
            fNodeRecordTable.put(node, this);
        }

        /**
         * Sort by method cost in descending order.  Allows this to be
         * this object ot be used as key and value in a SortedMap.
         */
        public int compareTo(Object elem2) {
            int cost2 = ((Record)elem2).fMethodTotalCost;
            return ((fMethodTotalCost > cost2) ? -1 : ((fMethodTotalCost < cost2) ? 1 : 0));
        }

        /**
         * Get string description for debugging.
         */
        public String toString() {
            return "Node-rec: " + JavaLang.simpleClassName(fNode)
                + " methodRoot=" + fMethodRoot
                + " chained=" + fChainedChildrenMethods
                + " minMethod=" + fMinMethodCost
                + " methodTotal=" + fMethodTotalCost
                + " treeTotal=" + fTreeTotalCost;
        }
    }

    /**
     * Constructor.
     */
    public BuildMethodMappings(int maxCreateCostPerBuildMethod,
                               Document document) {
        fMaxCreateCostPerBuildMethod = maxCreateCostPerBuildMethod;
        calculateTreeCosts(document);
    }

    /**
     * Get the cost of creating a single node based on its type.  This does
     * not provide cumulative costs, its based on type only.
     */
    public static int getNodeTypeCost(Node node) {
        if (node instanceof Element) {
            return BuildMethodMappings.CREATE_ELEMENT_NODE_CREATE_COST;
        } else {
            return 1;
        }
    }

    /**
     * Get a record.
     */
    private Record getRecord(Node node) {
        Record rec = (Record)fNodeRecordTable.get(node);
        if (rec == null) {
            throw new XMLCError("BUG: record not found: " + node);
        }
        return rec;
    }

    /**
     * Convert a child to being build in another element, if this will reduce
     * the cost of building the parent.  Adjust the size of the parent.
     */
    private void moveChildIntoMethod(Record parent,
                                     Record child) {
        if ((!child.fMethodRoot)
            && (child.fMethodTotalCost > BUILD_METHOD_CALL_CREATE_COST)) {
            parent.fMethodTotalCost -= child.fMethodTotalCost;
            parent.fMethodTotalCost += BUILD_METHOD_CALL_CREATE_COST;
            child.fMethodRoot = true;
        }
    }

    /**
     * Move element children into other build methods until the cost of
     * this element is below the maximum.  Sorting to move largest first
     * results in fewer methods being created.
     */
    private void adjustElementBuild(Record rec) {
        // Build a descending sorted list of element children.
        fTmpElementSort.clear();
        for (Node child = rec.fNode.getFirstChild(); 
             (child != null) && (rec.fMethodTotalCost >= fMaxCreateCostPerBuildMethod);
             child = child.getNextSibling()) {
            if (child instanceof Element) {
                Record childRec = getRecord(child);
                fTmpElementSort.add(childRec);
            }
        }

        // Move largest-first into their own build methods until this
        // method is no longer full.
        Collections.sort(fTmpElementSort);
        int numChildren = fTmpElementSort.size();
        for (int idx = 0; 
             (idx < numChildren) && (rec.fMethodTotalCost >= fMaxCreateCostPerBuildMethod);
             idx++) {
            moveChildIntoMethod(rec, (Record)fTmpElementSort.get(idx));
        }
        if (rec.fMethodTotalCost > fMaxCreateCostPerBuildMethod) {
            throw new XMLCError("BUG: adjustElementBuild failed to adjust method cost below max: " + rec);
        }
    }

    /**
     * Adjust an element record, if nececssary, to keep it from overflowing
     * the maximum build method create-cost.  This must be called *before* the
     * parent record is created.
     */
    private void adjustElementRecord(Record rec) {
       if (rec.fMinMethodCost > fMaxCreateCostPerBuildMethod) {
            // Create a chained method.
            rec.fMethodRoot = true;
            rec.fChainedChildrenMethods = true;
       } else if (rec.fMethodTotalCost >= fMaxCreateCostPerBuildMethod) {
            // Split children into other build methods until this one fits
            adjustElementBuild(rec);
        }
    }

    /**
     * Sum a record with one of its children.
     */
    private void sumRecords(Record nodeRec,
                            Record childRec) {
        // Min cost is based on if node could be but in seperate method
        if (childRec.fNode instanceof Element) {
            nodeRec.fMinMethodCost += BUILD_METHOD_CALL_CREATE_COST;
        } else {
            nodeRec.fMinMethodCost += childRec.fMethodTotalCost;
        }
        // Cost is based on if node is in seperate method
        if (!childRec.fMethodRoot) {
            nodeRec.fMethodTotalCost += childRec.fMethodTotalCost;
        }
        // Complete tree cost
        nodeRec.fTreeTotalCost += childRec.fTreeTotalCost;
    }

    /**
     * Calculate the costs of an element and its element's attributes.
     */
    private Record calculateElementCosts(Node element) {
        Record rec = new Record(element);
        NamedNodeMap attrList = element.getAttributes();
        if (attrList != null) {
            for (int idx = 0; idx < attrList.getLength(); idx++) {
                Attr attr = (Attr)attrList.item(idx);
                if (attr.getSpecified()) {
                    Record attrRec = calculateTreeCosts(attr);
                    sumRecords(rec, attrRec);
                }
            }
        }
        return rec;
    }

    /**
     * Traverse the tree, calculating costs and determining where build methods
     * should be created.  This builds the table of node records.
     * @return The node record.
     */
    private Record calculateTreeCosts(Node node) {
        Record rec;
        if (node instanceof Element) {
            rec = calculateElementCosts(node);
        } else {
            rec = new Record(node);
        }
        
        // depth-fist traversal, summing cost of child elements;
        for (Node child = node.getFirstChild(); child != null;
             child = child.getNextSibling()) {
            sumRecords(rec, calculateTreeCosts(child));
        }
        if (node instanceof Element) {
            adjustElementRecord(rec);
        }
        return rec;
    }

    /**
     * Should a new method be created at the root of this element.
     */
    public boolean isMethodRoot(Node element) {
        return getRecord(element).fMethodRoot;
    }

    /**
     * Should the children be created in a chained method?
     */
    public boolean useChainedChildrenMethods(Node node) {
        return getRecord(node).fChainedChildrenMethods;
    }

    /**
     * Get the minimum method cost of a node.
     */
    public int getMinMethodCost(Node node) {
        return getRecord(node).fMinMethodCost;
    }

    /**
     * Get the method total cost of a node.
     */
    public int getMethodTotalCost(Node node) {
        return getRecord(node).fMethodTotalCost; 
    }

    /**
     * Get the tree total cost of a node.
     */
    public int getTreeTotalCost(Node node) {
        return getRecord(node).fTreeTotalCost; 
    }

    /**
     * Get a string repersentation of an entry for debugging.
     */
    public String toString(Node node) {
        return getRecord(node).toString(); 
    }
}
