/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id$
 */

package org.enhydra.xml.xmlc.dom.generic;

import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.xmlc.codegen.*;
import org.enhydra.xml.xmlc.dom.*;
import org.enhydra.xml.xmlc.compiler.ElementInfo;
import org.enhydra.xml.xmlc.compiler.ElementTable;
import org.enhydra.xml.dom.DOMOps;
import java.util.ArrayList;
import java.util.TreeMap;
import java.util.HashMap;
import java.util.Map;
import java.util.Iterator;
import java.lang.Comparable;
import org.w3c.dom.*;


//FIXME: add building of doctype, maybe can't be done generically

/**
 * Class to generate code to build the document tree using only the W3C DOM
 * interface.  The document building code is generated in several functions to
 * avoid the JVM limit on maximum method size.  Since it is not possible to
 * determine the amount of bytecode generate, a metic is defined called
 * <em>create-cost</em>.  A unit of create-cost is approximately the cost
 * to create a node and append it to it's parent.  The limit on create-cost
 * is determined experimentally.
 * <P>
 * The following methods can be overridden to provide DOM
 * implementation-specific code generation:
 * <UL>
 * <LI> genNodeCreate - Generates code to create a node.
 * </UL>
 */
public class DOMBuilderGenerator {
    /** Name of argument to method containing parent node. */
    protected static final String PARENT_NODE_ARG = "parentNode";

    /* Name of argument to method containing document */
    protected static final String DOCUMENT_ARG = "document";

    /* Name of variable containing document element (only created if needed) */
    private static final String DOCUMENT_ELEMENT_VAR = "$docElement";

    /** Maximum creation-cost that a single build method can handle. */
    private int fMaxCreateCostPerBuildMethod;

    /**
     * Type of document argument of build methods.  Allows derived
     * DOMs to use a more specific argument type to avoid casting.
     */
    private String fDocumentArgClassName;

    /** Name of method to create. */
    private String fMethodName;

    /** Source document. */
    private Document fDocument;

    /** Object used to create DOM nodes. */
    private NodeCreateGenerator fNodeCreateGenerator;

    /** Table of element information. */
    private ElementTable fElementTable;

    /**
     * Object that determines where new methods will be created to
     * prevent method overflow.
     */
    private BuildMethodMappings fBuildMethodMappings;

    /** Class being created. */
    private JavaClass fDocClass;

    /**
     * Generator of access methods, needed to generate access method
     * initialization
     */
    private AccessorGenerator fAccessorGenerator;

    /**
     * Should the methods be static.
     */
    private boolean fStaticMethods;

    /**
     * Class that represents a single subdocument building method.
     * <P>
     * A build method build either a subtree or a set of nodes with the
     * same parent. This is required because the stack for building the tree
     * is contained in variables in the build methods.  The same-parent
     * is used when all of the immediate children of a node will not fit
     * under the limit, these objects are refered to as being chained to
     * the object that created the parent.
     */
    private class BuildMethodGenerator {
        /** Table of tmp node variable names. */
        private VarNames fNodeVarNames = new VarNames("Node", "$node");

        /* Table of tmp Element node variable names. */
        private VarNames fElementVarNames = new VarNames("Element", "$elem");

        /* Table of tmp attribute variable names. */
        private VarNames fAttrVarNames = new VarNames("Attr", "$attr");

        /* Has document element variable been referenced? */
        private boolean fReferencedDocElement;

        /* The java code for the build metho */
        private JavaMethod fMethod;

        /** Body of current method being generated. */
        private JavaCode fBody;

        /** Next number to for generating submethod names. */
        private int fNextSubMethodId;

        /** If this is a chained entry, this is the chain level. */
        private int fChainLevel;

        /** The current cost of this method. */
        private int fCurrentMethodCost;

        /**
         * Constructor. Create a method to generate a subtree and its
         * children.
         *
         * @param methodName The name for the build method.  A unique
         *  suffix is append to the name for each sub-method generated.
         */
        public BuildMethodGenerator(String methodName) throws XMLCException {
            createJavaMethod(methodName);
        }

        /**
         * Generate a method to create DOM nodes.  This is the entry point,
         * it may recursively create more methods.
         * @param root The root of the subtree.  This maybe the preexisting
         *  Document node.
         * @param docVarName The variable containing the document, if not null.
         * @param parentVarName The variable containing the parent, if not null.
         * @param body Append a call to this build method here, if not null.
         */
        public void createMethod(Node root,
                                 String docVarName,
                                 String rootVarName,
                                 JavaCode body) throws XMLCException {
            if (body != null) {
                body.addln(fMethod.getName() + "(" + docVarName + ", " + rootVarName + ");");
            }

            processNode(PARENT_NODE_ARG, root, 0);
            createVariables();

            if (fCurrentMethodCost > fMaxCreateCostPerBuildMethod) {
                throw new XMLCError("BUG: maximum number of nodes per method exceeded");
            }
        }
        
        /**
         * Create children of a node using chaining.  This is the entry when
         * creating more chained nodes.
         */
        private void createChainedChildrenMethod(int chainLevel,
                                                 Node parent,
                                                 Node firstChild,
                                                 String docVarName,
                                                 String parentVarName,
                                                 JavaCode body) throws XMLCException {
            fChainLevel = chainLevel;
            body.addln(fMethod.getName() + "(" + docVarName + ", " + parentVarName + ");");
            processChildrenChaining(parent, firstChild, PARENT_NODE_ARG, 0);
            createVariables();
        }

        /**
         * Get temporary variable name for a node. Special handling for
         * Elements; allow for attribute setting without requiring (slower)
         * type casting in generate code.
         */
        private String getNodeVarName(Node node,
                                      int level) {
            if (node instanceof Element) {
                return fElementVarNames.getVarName(level);
            } else {
                return fNodeVarNames.getVarName(level);
            }
        }

        /**
         * Get the name of the next subtree method to build.
         */
        private String getNextSubTreeMethodName() {
            return fMethod.getName() + "_" + fNextSubMethodId++;
        }

        /**
         * Determine how much space is left in the method, in create-cost units.
         */
        private int spaceLeft() {
            int left = (fMaxCreateCostPerBuildMethod - fCurrentMethodCost);
            return (left < 0) ? 0 : left;
        }

        /**
         * Create a new build method rooted at an element.
         */
        private void createElementBuildMethod(String elementVarName,
                                              Element element) throws XMLCException {
            BuildMethodGenerator buildMethod
                = new BuildMethodGenerator(getNextSubTreeMethodName());
            buildMethod.createMethod(element, DOCUMENT_ARG, elementVarName,
                                     fBody);
            fCurrentMethodCost += BuildMethodMappings.BUILD_METHOD_CALL_CREATE_COST;
        }
        
        /**
         * Generate code to create an element and its attributes in this
         * method.  Create calls to initialize access methods.
         */
        private void processElement(String elementVarName,
                                    Element element,
                                    int level) throws XMLCException {
            processNodeAttrs(elementVarName, element, level);
            ElementInfo elementInfo = fElementTable.getElementInfo(element);
            if (elementInfo.getNumAccessMethods() > 0) {
                fAccessorGenerator.createAccessMethodInit(elementInfo,
                                                          elementVarName,
                                                          fBody);
            }
        }

        /**
         * Generate code to to initialize an attribute.
         */
        private void processAttr(String elementVarName,
                                 Attr attr,
                                 int level) throws XMLCException {
            String attrVarName = fAttrVarNames.getVarName(level);
            if (attr.getSpecified()) {
                // Create attribute node
                fNodeCreateGenerator.genNodeCreate(DOCUMENT_ARG, attrVarName, attr, fBody);
                fBody.addln(elementVarName + ".setAttributeNode(" + attrVarName + ");");
                fCurrentMethodCost += BuildMethodMappings.getNodeTypeCost(attr);

                // Create attribute children.
                processChildren(attrVarName, attr, level+1);
            }
        }

        /**
         * Generate code to to initialize an element's attributes.
         */
        private void processNodeAttrs(String elementVarName,
                                      Node node,
                                      int level) throws XMLCException {
            NamedNodeMap attrList = node.getAttributes();
            if (attrList != null) {
                for (int idx = 0; idx < attrList.getLength(); idx++) {
                    processAttr(elementVarName, (Attr)attrList.item(idx), level);
                }
            }
        }

        /**
         * Determine if a node is a child of the document and needs to be
         * inserted before the document variable.
         * FIXME: This will not work if we got to precreating
         */
        private boolean preceedsDocElement(Node node) {
            if (node.getParentNode() == fDocument) {
                Element docElement = fDocument.getDocumentElement();
                // Scan following siblings for document element
                for (Node sibling = node.getNextSibling(); sibling != null;
                     sibling = sibling.getNextSibling()) {
                    if (sibling == docElement) {
                        return true;  // doc element after
                    }
                }
            }
            return false;        
        }

        /**
         * Generate code to create a node.
         * @return The generated variable name that will contain the node.
         */
        private String createNode(String parentVarName,
                                  Node node,
                                  int level) throws XMLCException {
            String nodeVarName = getNodeVarName(node, level);
            fBody.addln();
            fNodeCreateGenerator.genNodeCreate(DOCUMENT_ARG, nodeVarName,
                                               node, fBody);
            // Add to parent (unless pre-existing Element)
            if (!fNodeCreateGenerator.isPreCreatedNode(node)) {
                if (preceedsDocElement(node)) {
                    fBody.addln(parentVarName + ".insertBefore(" + nodeVarName
                                + ", " + DOCUMENT_ELEMENT_VAR + ");");
                    fReferencedDocElement = true;
                } else {
                    fBody.addln(parentVarName + ".appendChild(" + nodeVarName + ");");
                }
            }
            fCurrentMethodCost += BuildMethodMappings.getNodeTypeCost(node);
            return nodeVarName;
        }

        /**
         * Process a node, generating code to recreate it.
         */
        private void processNode(String parentVarName,
                                 Node node,
                                 int level) throws XMLCException {
            if ((level != 0) && (node instanceof Element)
                && fBuildMethodMappings.isMethodRoot(node)) {
                // Make this element root of a method
                createElementBuildMethod(parentVarName, (Element)node);
            } else if (node == fDocument) {
                // Document already created and in arg, just handle children
                processChildren(PARENT_NODE_ARG, node, level+1);
            } else {
                // Build in this method
                String nodeVarName = createNode(parentVarName, node, level);
                if (node instanceof Element) {
                    processElement(nodeVarName, (Element)node, level);
                }
                processChildren(nodeVarName, node, level+1);
            }
        }
        
        /**
         * Process children of a parent node while chaining build methods
         * together.  This fills in the current build method and then
         * chains to another.
         */
        private void processChildrenChaining(Node parent,
                                             Node firstChild,
                                             String parentVarName,
                                             int level) throws XMLCException {
            // Fit as many as we an into this method.
            Node child = firstChild; 
            while ((child != null) && (fBuildMethodMappings.getMethodTotalCost(child) < spaceLeft())) {
                processNode(parentVarName, child, level);
                child = child.getNextSibling();
            }
            
            // If there are any left, create another method.
            if (child != null) {
                BuildMethodGenerator nextBuildMethod
                    = new BuildMethodGenerator(fMethod.getName() + "c");
                nextBuildMethod.createChainedChildrenMethod(fChainLevel+1,
                                                            parent, child,
                                                            DOCUMENT_ARG,
                                                            parentVarName,
                                                            fBody);
                fCurrentMethodCost += BuildMethodMappings.BUILD_METHOD_CALL_CREATE_COST;
            }
        }

        /**
         * Create children of the node.
         */
        private void processChildren(String parentVarName,
                                     Node parent,
                                     int level) throws XMLCException {
            if (fBuildMethodMappings.useChainedChildrenMethods(parent)) {
                processChildrenChaining(parent, parent.getFirstChild(),
                                        parentVarName, level);
            } else {
                for (Node child = parent.getFirstChild(); child != null;
                     child = child.getNextSibling()) {
                    processNode(parentVarName, child, level);
                }
            }
        }

        /**
         * Create the method that build this subtree.
         */
        private void createJavaMethod(String methodName) {
            JavaParameter docArg
                = new JavaParameter(DOCUMENT_ARG,
                                    fDocumentArgClassName,
                                    (String)null);
            JavaParameter parentArg
                = new JavaParameter(PARENT_NODE_ARG,
                                    Node.class.getName(),
                                    (String)null);
            fMethod = new JavaMethod(methodName, "void",
                                     (JavaModifiers.PRIVATE | (fStaticMethods ? JavaModifiers.STATIC : 0)),
                                     new JavaParameter[] {docArg, parentArg},
                                     new String[] {"Create a subtree of the document."});
            fDocClass.addMethod(fMethod);
            fBody = fMethod.getCode();
        }

        /**
         * Fill in the variables for this method.
         */
        private void createVariables() {
            fNodeVarNames.insertVarDefs(fBody);
            fElementVarNames.insertVarDefs(fBody);
            fAttrVarNames.insertVarDefs(fBody);
            if (fReferencedDocElement) {
                fBody.addVars("Element " + DOCUMENT_ELEMENT_VAR + " = " + DOCUMENT_ARG + ".getDocumentElement();");
            }

        }
    }

    /**
     * Constructor.  Build the methods.
     */
    public DOMBuilderGenerator(String methodName,
                               Document document,
                               String documentArgClassName,
                               NodeCreateGenerator nodeCreateGenerator,
                               AccessorGenerator accessorGenerator,
                               ElementTable elementTable,
                               JavaClass docClass,
                               int maxCreateCostPerBuildMethod,
                               boolean staticMethods) throws XMLCException {
        fMethodName = methodName;
        fDocument = document;
        fDocumentArgClassName = documentArgClassName;
        fNodeCreateGenerator = nodeCreateGenerator;
        fAccessorGenerator = accessorGenerator;
        fElementTable = elementTable;
        fDocClass = docClass;
        fMaxCreateCostPerBuildMethod = maxCreateCostPerBuildMethod;
        fStaticMethods = staticMethods;

        fBuildMethodMappings = new BuildMethodMappings(fMaxCreateCostPerBuildMethod,
                                                       document);

        BuildMethodGenerator buildMethod
            = new BuildMethodGenerator(fMethodName);
        buildMethod.createMethod(document, null, null, null);
    }

    /**
     * Create a call to the top level method that was generated.
     */
    public void createMethodCall(String docVarName,
                                 JavaCode body) {
        body.addln(fMethodName + "(" + docVarName + ", " + docVarName + ");");
    }
}
