/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id$
 */

package org.enhydra.xml.xmlc.dom.lazydom;

import org.enhydra.xml.xmlc.XMLCError;
import org.enhydra.xml.xmlc.XMLCException;
import org.enhydra.xml.xmlc.codegen.JavaLang;
import org.enhydra.xml.xmlc.codegen.JavaCode;
import org.enhydra.xml.xmlc.dom.*;
import org.enhydra.xml.xmlc.dom.generic.NodeCreateGenerator;
import org.enhydra.xml.io.PreFormattedText;
import org.enhydra.xml.lazydom.*;
import org.w3c.dom.*;
import org.w3c.dom.html.HTMLDocument;

/**
 * DOM node creation class that overrides methods in the generic version to
 * assign node ids.
 */
class LazyDOMNodeCreateGenerator extends NodeCreateGenerator {
    /**
     * If not null, the object to preformat node text.
     */
    private PreFormatter fPreFormatter;

    /**
     * Constructor.
     */
    public LazyDOMNodeCreateGenerator(XMLCDocument xmlcDoc,
                                      PreFormatter preFormatter) {
        super(xmlcDoc);
        fPreFormatter = preFormatter;
    }

    /**
     * Get the string constant for the preformatted text for a node, or "null"
     * if node should not be associated with the node.
     */
    private String getPreFormattedText(Node node) {
        String text = null;
        if ((node instanceof PreFormattedText) && (fPreFormatter != null)) {
            text = fPreFormatter.preFormatNode(node);
        }
        return JavaLang.createStringConst(text);
    }

    /**
     * Generate Attr creation.
     */
    private void genAttrCreate(Attr attr,
                               int nodeId,
                               String docVarName,
                               JavaCode body) {
        if (attr.getNamespaceURI() == null) {
            body.addln(docVarName + ".createTemplateAttribute("
                + JavaLang.createStringConst(attr.getName()) + ", "
                + nodeId + ");");
        } else {
            body.addln(docVarName + ".createTemplateAttributeNS("
                       + JavaLang.createStringConst(attr.getNamespaceURI()) + ", "
                + JavaLang.createStringConst(attr.getName()) + ", "
                + nodeId + ");");
        }
    }

    /**
     * Generate Element creation.
     */
    private void genElementCreate(Element element,
                                  int nodeId,
                                  String docVarName,
                                  String varName,
                                  JavaCode body) {
        if (isPreCreatedNode(element)) {
            String castVar = "((" + LazyElement.class.getName() + ")"
                + varName + ")";
            body.addln(docVarName + ".getDocumentElement();");
            body.addln(castVar + ".makeTemplateNode(" + nodeId + ");");
            body.addln(castVar + ".setPreFormattedText("
                       + getPreFormattedText(element) + ");");
        } else if (element.getNamespaceURI() == null) {
            body.addln(docVarName + ".createTemplateElement("
                       + JavaLang.createStringConst(element.getTagName()) + ", "
                       + nodeId + ", " + getPreFormattedText(element) + ");");
        } else {
            body.addln(docVarName + ".createTemplateElementNS("
                       + JavaLang.createStringConst(element.getNamespaceURI()) + ", "
                       + JavaLang.createStringConst(element.getTagName()) + ", "
                       + nodeId + ", " + getPreFormattedText(element) + ");");
        }
    }

    /**
     * Generate Notation creation.
     */
    private void genNotationCreate(Notation notation,
                                   int nodeId,
                                   String docVarName,
                                   JavaCode body) {
        body.addln(docVarName + ".createNotation("
                   + JavaLang.createStringConst(notation.getNodeName()) + ", "
                   + JavaLang.createStringConst(notation.getPublicId()) + ", "
                   + JavaLang.createStringConst(notation.getSystemId()) + ", "
                   + nodeId + ");");
    }

    /**
     * Generate Entity creation.
     */
    private void genEntityCreate(Entity entity,
                                 int nodeId,
                                 String docVarName,
                                 JavaCode body) {
        body.addln(docVarName + ".createTemplateEntity("
                   + JavaLang.createStringConst(entity.getNodeName()) + ", "
                   + JavaLang.createStringConst(entity.getPublicId()) + ", "
                   + JavaLang.createStringConst(entity.getSystemId()) + ", "
                   + JavaLang.createStringConst(entity.getNotationName()) + ", "
                   + nodeId + ");");
    }

    /**
     * Generate ProcessingInstruction creation.
     */
    private void genProcessInstructionCreate(ProcessingInstruction procInstr,
                                             int nodeId,
                                             String docVarName,
                                             JavaCode body) {
        body.addln(docVarName + ".createTemplateProcessingInstruction("
                   + JavaLang.createStringConst(procInstr.getNodeName()) + ", "
                   + JavaLang.createStringConst(procInstr.getData()) + ", "
                   + nodeId + ");");
    }

    /**
     * Generate code to construct a node.  Returned string is an R-value
     * to use in an assignment.  
     * @see NodeCeateGenerator#getNodeCreate(Node)
     */
    public void genNodeCreate(String docVarName,
                              String varName,
                              Node node,
                              JavaCode body) throws XMLCException {
        int nodeId = ((LazyNode)node).getNodeId();
        body.add(varName + " = ");

        switch (node.getNodeType()) {
        case Node.ATTRIBUTE_NODE:
            genAttrCreate((Attr)node, nodeId, docVarName, body);
            break;

        case Node.CDATA_SECTION_NODE:
            body.addln(docVarName + ".createTemplateCDATASection("
                       + JavaLang.createStringConst(((CDATASection)node).getData()) + ", "
                       + nodeId + ");");
            break;

        case Node.COMMENT_NODE:
            body.addln(docVarName + ".createTemplateComment("
                       + JavaLang.createStringConst(((Comment)node).getData()) + ", "
                       + nodeId + ");");
            break;

        case Node.ELEMENT_NODE:
            genElementCreate((Element)node, nodeId, docVarName, varName, body);
            break;

        case Node.ENTITY_NODE:
            genEntityCreate((Entity)node, nodeId, varName, body);
            break;

        case Node.ENTITY_REFERENCE_NODE:
            body.addln(docVarName + ".createTemplateEntityReference("
                       + JavaLang.createStringConst(((EntityReference)node).getNodeName()) + ", "
                       + nodeId + ");");
            break;

        case Node.NOTATION_NODE:
            genNotationCreate((Notation)node, nodeId, varName, body);
            break;

        case Node.PROCESSING_INSTRUCTION_NODE:
            genProcessInstructionCreate((ProcessingInstruction)node,
                                        nodeId, docVarName, body);
            break;

        case Node.TEXT_NODE:
            body.addln(docVarName + ".createTemplateTextNode("
                       + JavaLang.createStringConst(((Text)node).getData()) + ", "
                       + nodeId + ", " + getPreFormattedText(node) + ");");
            break;

        case Node.DOCUMENT_FRAGMENT_NODE:
        case Node.DOCUMENT_NODE:
        case Node.DOCUMENT_TYPE_NODE:
            throw new XMLCError("Creating node of this type not supported: "
                                + node.getClass().getName());

        default:
            throw new XMLCError("Unknown node type: " + node.getNodeType());
        }
    }
}
