/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id$
 */

package org.enhydra.xml.xmlc.metadata;

import java.io.File;
import java.io.PrintWriter;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.HashMap;
import org.w3c.dom.*;
import org.apache.xerces.dom.DocumentImpl;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.enhydra.xml.xmlc.XMLCException;
import org.enhydra.xml.xmlc.XMLCError;
import org.enhydra.xml.io.ErrorReporter;
import org.enhydra.xml.io.DOMParser;
import org.enhydra.xml.io.OutputOptions;
import org.enhydra.xml.io.DOMFormatter;
import org.enhydra.xml.io.XMLEntityResolver;
import org.enhydra.xml.dom.DOMInfo;
import org.enhydra.xml.io.InputSourceOps;

//FIXME: modify Xerces to take a already created document class.
//FIXME: use DOMImplementation

/**
 * Metadata XML document associated with a Document being parsed.  Data maybe
 * set from the command line or read from an XML file (or both).
 */
public class MetaDataDocument extends DocumentImpl {
    /**
     * Compile-time debugging.
     */
    private static final boolean DEBUG = false;

    /**
     * Schema attributes and values.
     * FIXME: standardize all of this someplace
     */
    private static final String[][] SCHEMA_ATTRS = {
        {"xmlns:xsi", "http://www.w3.org/1999/XMLSchema-instance"},
        {"xsi:noNamespaceSchemaLocation", "file:/org/enhydra/xml/xmlc/metadata/xmlc-1.0.xsd"}
    };

    /**
     * Extension for XMLC metadata files.
     */
    public static final String METADATA_FILE_EXTENSION = ".xmlc";

    /**
     * Table of tag name to element class.
     */
    private static HashMap tagClassTable = new HashMap();

    /**
     * Signature used to locate constructor of element class.
     */
    private static final Class[] elemClassConstructorSign
        = new Class[]{Document.class};

    /**
     * Static constructor.
     */
    static {
        // Initialize tag table.
        tagClassTable.put(JavaCompilerSection.TAG_NAME, JavaCompilerSection.class);
        tagClassTable.put(CompileOptions.TAG_NAME,  CompileOptions.class);
        tagClassTable.put(DeleteElement.TAG_NAME, DeleteElement.class);
        tagClassTable.put(DOMEdits.TAG_NAME, DOMEdits.class);
        tagClassTable.put(DocumentClass.TAG_NAME, DocumentClass.class);
        tagClassTable.put(DocumentSection.TAG_NAME, DocumentSection.class);
        tagClassTable.put(ElementDef.TAG_NAME, ElementDef.class);
        tagClassTable.put(HTMLAttr.TAG_NAME, HTMLAttr.class);
        tagClassTable.put(HTMLTag.TAG_NAME, HTMLTag.class);
        tagClassTable.put(HTMLTagSet.TAG_NAME, HTMLTagSet.class);
        tagClassTable.put(HTMLCompatibility.TAG_NAME, HTMLCompatibility.class);
        tagClassTable.put(HTMLSection.TAG_NAME, HTMLSection.class);
        tagClassTable.put(Implements.TAG_NAME, Implements.class);
        tagClassTable.put(JavacOption.TAG_NAME, JavacOption.class);
        tagClassTable.put(ParserSection.TAG_NAME, ParserSection.class);
        tagClassTable.put(TagClass.TAG_NAME, TagClass.class);
        tagClassTable.put(URLRegExpMapping.TAG_NAME, URLRegExpMapping.class);
        tagClassTable.put(URLMapping.TAG_NAME, URLMapping.class);
        tagClassTable.put(XCatalog.TAG_NAME, XCatalog.class);
        tagClassTable.put(MetaData.TAG_NAME, MetaData.class);
    }

    /**
     * Parse a XMLC metadata file.
     *
     * @param inputSource Specification of file to parse.
     * @param reporter Object use to report errors.
     * @param classLoader If not null, class loader to use in resolving
     *  entities to be found on the classpath.
     */
    static public MetaDataDocument parseMetaData(InputSource inputSource,
                                                 ErrorReporter reporter,
                                                 ClassLoader classLoader) 
        throws SAXException, IOException, XMLCException {

        DOMParser parser = new DOMParser();
        XMLEntityResolver resolver = new XMLEntityResolver();
        if (DEBUG) {
            resolver.setDebugWriter(new PrintWriter(System.err, true));
        }

        if (classLoader != null) {
            resolver.addClassLoader(classLoader);
        }
        resolver.setDefaultResolving();
        parser.setEntityResolver(resolver);

        parser.setErrorHandler(reporter);
        parser.setDocumentClassName(MetaDataDocument.class.getName());

        MetaDataDocument document = (MetaDataDocument)parser.parse(inputSource);
        if (reporter.getErrorCnt() > 0) {
            throw new SAXException("XMLC metadata file parse failed");
        }
        if (DEBUG) {
            DOMInfo.printTree("Document after parse: " + InputSourceOps.getName(inputSource),
                              document, System.err);
        }

        document.completeModifications();

        if (DEBUG) {
            DOMInfo.printTree("Document after completeModifications: " + InputSourceOps.getName(inputSource),
                              document, System.err);
        }
        return document;
    }

    /**
     * Construct an empty instace of a metadata document.
     */
    public MetaDataDocument() {
    }

    /**
     * Ensure that the document element has the Schema attributes
     */
    private void ensureSchemaAttrs() {
        Element docElement = getDocumentElement();
        for (int i = 0; i < SCHEMA_ATTRS.length; i++) {
            if (docElement.getAttributeNode(SCHEMA_ATTRS[i][0]) == null) {
                docElement.setAttribute(SCHEMA_ATTRS[i][0], SCHEMA_ATTRS[i][1]);
            }
        }
    }

    /**
     * Serialize this object to a XML file.
     */
    public void serialize(File file) throws IOException {
        ensureSchemaAttrs();
        if (DEBUG) {
            DOMInfo.printTree("Document before write: " + file,
                              this, System.err);
        }
        OutputOptions outputOptions = new OutputOptions();
        outputOptions.setPrettyPrinting(true);
        outputOptions.setPreserveSpace(false);
        DOMFormatter domFormatter = new DOMFormatter(outputOptions);

        domFormatter.write(this, file);
    }
    
    /**
     * Serialize this object to a XML file using the default name.
     */
    public void serialize() throws IOException {
        DocumentClass documentClass = getMetaData().getDocumentClass();
        if (documentClass == null) {
            throw new XMLCError("No DocumentClass metadata object; can't determine default name");
        }

        String fileName;
        if (documentClass.getGenerate().generateInterface()) {
            fileName = documentClass.getUnqualInterfaceName() +  METADATA_FILE_EXTENSION;
        } else {
            fileName = documentClass.getUnqualClassName() +  METADATA_FILE_EXTENSION;
        }
        
        serialize(new File(getMetaData().getCompileOptions().getPackageOutputDir(), fileName));
    }

    /**
     * Get the root MetaData object, which must exist.
     */
    public MetaData getMetaData() {
        MetaData root = (MetaData)getDocumentElement();
        if (root == null) {
            root = (MetaData)createElement(MetaData.TAG_NAME);
            appendChild(root);
        }
        return root;
    }
    
    /**
     * Create a XMLC metadata document given its class.
     */
    public Element createElement(Class elementClass) {
        try {
            Constructor constr = elementClass.getConstructor(elemClassConstructorSign);
            return (Element)constr.newInstance(new Object[]{this});
        } catch (Exception except) {
            throw new XMLCError(except);
        }
    }

    /**
     * Create a XMLC metadata document element.
     */
    public Element createElement(String tagName) throws DOMException {
        Class elementClass = (Class)tagClassTable.get(tagName);
        if (elementClass == null) {
            throw new XMLCError("Invalid tag name: \"" + tagName + "\"");
        }
        return createElement(elementClass);
    }

    /**
     * Create a XMLC metadata document element. in a namespace.
     */
    public Element createElementNS(String namespaceURI,
                                   String qualifiedName) {
	if ((namespaceURI == null) || (namespaceURI.length() == 0)) {
	    return createElement(qualifiedName);
	} else {
            // FIXME: this should work in a namespace.
            throw new XMLCError("createElementNS not implemented");
        }
    }

    /**
     * Complete modifications to DOM.  This should be called after a set of
     * modifcations have been completed (it is called after a read).  It is
     * used to synchronize data in different areas of the DOM.
     */
    public void completeModifications() throws XMLCException {
        getMetaData().completeModifications();
    }

    /**
     * Merge another MetaData document into this document.
     */
    public void merge(MetaDataDocument srcDocument) {
        getMetaData().mergeElement(srcDocument.getMetaData());
    }
}
