/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: ClassEntry.java,v 1.2.2.2.4.2 2000/11/16 09:08:42 markd Exp $
 */

package org.enhydra.xml.xmlc.reloading;

import org.enhydra.xml.xmlc.XMLCError;
import org.enhydra.xml.xmlc.XMLCLogger;
import java.io.PrintWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.net.URL;

// FIXME: could keep a point to factory, instead of recompiler + resource loader.
// Maybe reloading factory should just implement some of this instead of having
// it here.

/**
 * Object used to hold information about a XMLC-generate class and perform
 * operations assocaited with this class.  This manages only the actually
 * generated class, not a derived class.  This object also serves as a lock
 * for operations on the class.
 */
class ClassEntry {
    /**
     * Class object for originally loaded class.
     */
    private Class fOriginalClass;

    /**
     * File associated with originally loaded class.
     */
    private File fOriginalClassFile;

    /**
     * Most current version of the class.
     */
    private Class fCurrentClass;

    /**
     * The last-modified time when the class was loaded.
     */
    private long fCurrentClassLoadLastModified;

    /**
     * File associated with currenty loaded class.
     */
    private File fCurrentClassFile;

    /**
     * ClassLoader wrapper used to load this class.
     */
    private ResourceLoader fResourceLoader;

    /**
     * Handler used to check for out-of-date classes and used to
     * recompile them; null if just reloading.
     */
    private RecompilationHandler fRecompilationHandler;

    /**
     * Logging interface.
     */
    private XMLCLogger fLogger;

    /**
     * Constructor.  
     *
     * @param xmlcDocClass XMLC generated document class.
     * @param resourceLoader ClassLoader wrapper used to load this class.
     * @param recompilationHandler Object to handle checking for out-of-date
     *  class files and recompiling classes.  Maybe null if only reloading
     *  is implemented.
     * @param logger XMLC logger.
     */
    public ClassEntry(Class xmlcDocClass,
                      ResourceLoader resourceLoader,
                      RecompilationHandler recompilationHandler,
                      XMLCLogger logger) {
        fOriginalClass = xmlcDocClass;
        fCurrentClass = xmlcDocClass;
        fResourceLoader = resourceLoader;
        fRecompilationHandler = recompilationHandler;
        fLogger = logger;

        // Make sure class was loaded by the class loader associated
        // with the resource loader.
        if (fOriginalClass.getClassLoader() != fResourceLoader.getClassLoader()) {
            throw new XMLCReloadingException("Class loader for XMLC document class is not the same as the one associated with the ResourceLoader: \""
                                             + fOriginalClass.getName() + "\""
                                             + ", doc=" + fOriginalClass.getClassLoader()
                                             + ", rsc=" + fResourceLoader.getClassLoader());
        }

        // Get the File from the class loader.
        try {
            fOriginalClassFile = getClassFile(fOriginalClass);
            fCurrentClassFile = fOriginalClassFile;
            fCurrentClassLoadLastModified = getLastModified(fOriginalClassFile);
        } catch (Exception except) {
            // This should never be able to happen, as we already have
            // the class
            throw new XMLCReloadingException("Unable to find file for class \""
                                             + fOriginalClass.getName() + "\"", except);
        }
    }

    /**
     * Convert a URL to a File.
     */
    private File urlToFile(URL url) {
        String protocol = url.getProtocol();
        if (!(protocol.equals("") || protocol.equals("file"))) {
            throw new XMLCReloadingException("XMLC reloading only operates on file: URLs: \""
                                             + url + "\"");
        }
        return new File(url.getFile());
    }

    /**
     * Get the File object for a class file.
     */
    private File getClassFile(Class classObj) throws ClassNotFoundException {
        String fileName = "/" + classObj.getName().replace('.', '/') + ".class";
        ClassLoader loader = classObj.getClassLoader();
        URL classURL;
        if (loader == null) {
            classURL = ClassLoader.getSystemResource(fileName);
        } else {
            classURL = loader.getResource(fileName);
        }
        if (classURL == null) {
            throw new ClassNotFoundException("Can't get class file URL for class: " 
                                             + classObj.getName());
        }
        return urlToFile(classURL);
    }

    /**
     * Get the last modified time from a file, generate an exception
     * if it doesn't exist.
     */
    private long getLastModified(File file) throws FileNotFoundException {
        long time = file.lastModified();
        if (time == 0L) {
            throw new FileNotFoundException(file.getPath());
        }
        return time;
    }

    /**
     * Get the class object for originally loaded class.
     */
    public Class getOriginalClass() {
        return fOriginalClass;
    }

    /**
     * Get the File associated with originally loaded class.
     */
    public File getOriginalClassFile() {
        return fOriginalClassFile;
    }

    /**
     * Get the current version of the class.
     */
    public Class getCurrentClass() {
        return fCurrentClass;
    }

    /**
     * Get File associated with currenty loaded class.
     */
    public File getCurrentClassFile() {
        return fCurrentClassFile;
    }

    /**
     * Check class modification time.
     */
    private boolean classOutOfDate() {
        long fileLastModified;
        try {
            fileLastModified = getLastModified(fCurrentClassFile);
        } catch (FileNotFoundException except) {
            throw new XMLCReloadingException(except);
        }
        
        if (fLogger.debugEnabled()) {
            fLogger.logDebug("class is "
                             + ((fileLastModified > fCurrentClassLoadLastModified) ? "out-of-date" : "current")
                             + " with class file:\n"
                             + "class file: "
                             + fileLastModified + ": "
                             + fCurrentClassFile.getAbsolutePath() + "\n"
                             + "class: " + fCurrentClassLoadLastModified + ": "
                             + fCurrentClass.getName());
        }
        return (fileLastModified > fCurrentClassLoadLastModified);
    }

    /**
     * Load the class in a new classloader.  The requested class and it's
     * inner classes are the only class this loader will load.  All others are
     * passed to the main class loader.
     * @return true if the class was actually reloaded.
     */
    private synchronized boolean reloadClass() {
        // synchronized: check for out-of-date again
        if (!classOutOfDate()) {
            return false;
        }
        String className = fOriginalClass.getName();
        if (fLogger.infoEnabled()) {
            fLogger.logInfo("Reloading " + className);
        }
        
        // Create class loader that will hold this class.
        ClassLoader selectiveLoader
            = fResourceLoader.createSelectiveClassLoader(className);

        // Load in this class loader and get File object.
        try {
            fCurrentClass = selectiveLoader.loadClass(className);
            fCurrentClassFile = getClassFile(fCurrentClass);
        } catch (ClassNotFoundException except) {
            //FIXME: chained exception should handle this
            if (except.getException() != null) {
                throw new XMLCReloadingException(except.getException());
            } else {
                throw new XMLCReloadingException(except);
            }
        }
        
        if (fCurrentClass == null) {
            throw new XMLCReloadingException("selective class loader returned null: " + className
                                             + ", selective loader: " + selectiveLoader.getClass());
        }
        if (fCurrentClass == fOriginalClass) {
            throw new XMLCReloadingException("selective class loader returned old class: " + className
                                             + ", selective loader: " + selectiveLoader.getClass());
        }
        return true;
    }
    
    /**
     * Recompile the class with synchronization.
     * @return true if the class was actually recompiled.
     */
    private synchronized boolean recompileClass() {
        // synchronized: check for out-of-date again
        if (!fRecompilationHandler.isOutOfDate(this)) {
            return false;
        }

        if (fLogger.infoEnabled()) {
            fLogger.logInfo("Recompiling " + fOriginalClass.getName());
        }
        fRecompilationHandler.recompile(this);

        // Check that we actually did something
        if (fRecompilationHandler.isOutOfDate(this)) {
            throw new XMLCReloadingException("recompilation didn't updated class file "
                                             + fOriginalClass.getName());
        }
        return true;
    }

    /**
     * Get the current version of the class if the originally loaded class is
     * out-of-date.  The current class, either the original class or a
     * previously reloaded class will be check to determine if its current.
     * If a recompilation handler is set, then the class file is compared to
     * its source file modification time.  If its out of date, the class
     * will be recompiled. If the loaded class is out-of-date relative to the
     * class file, it is loaded in its own class loaded and returned.  If an
     * updated class is has previously been loaded, it will be returned.
     * This class only synchronizes if necessary to update the class.
     *
     * @param factory The factory the entry is associated with.
     * @return The current class, either the original class or the new one
     *  loaded in its own class loader.
     * @exception XMLCReloadingException Thrown on failure to recompiler or
     *  reload.
     */
    public Class getCurrentClass(XMLCReloadingFactory factory) {
        // Recompile if class file is out-of-date with source file.
        if ((fRecompilationHandler != null)
            && fRecompilationHandler.isOutOfDate(this)) {
            if (recompileClass()) {
                factory.incrRecompileCount();
            }
        }

        // Reload class if its out-of-date
        if (classOutOfDate()) {
            if (reloadClass()) {
                factory.incrReloadCount();
            }
        }
        return fCurrentClass;
    }
}
