/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: Recompiler.java,v 1.1.8.2 2000/11/16 09:08:42 markd Exp $
 */

package org.enhydra.xml.xmlc.reloading;

import org.enhydra.xml.xmlc.XMLCException;
import org.enhydra.xml.xmlc.XMLCLogger;
import org.enhydra.xml.xmlc.compiler.*;
import org.enhydra.xml.xmlc.codegen.JavaLang;
import org.enhydra.xml.xmlc.dom.XMLCDocument;
import org.enhydra.xml.xmlc.metadata.*;
import org.enhydra.xml.io.ErrorReporter;
import org.enhydra.xml.io.XMLEntityResolver;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.io.InputStream;
import java.io.IOException;
import java.io.File;
import java.net.URL;
import java.util.Vector;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * Class to recompile a XMLC-generated class.
 */
class Recompiler {
    /**
     * Factory we are associated with.
     */
    private XMLCReloadingFactory fFactory;

    /**
     * Information about the class.
     */
    private ClassEntry fClassEntry;

    /**
     * Source file being recompiled.
     * FIXME: lot of moving back and forth between file and string (maybe just inputsource);
     */
    private File fSrcFile;

    /**
     * Logging interface.
     */
    private XMLCLogger fLogger;

    /**
     * Error report for collecting errors.
     */
    //FIXME: Should not use stderr, but need to combine sax exception with 
    //this below. System.err
    ErrorReporter errorReporter = new ErrorReporter();

    /**
     * Constructor.
     */
    public Recompiler(ClassEntry classEntry,
                      File srcFile,
                      XMLCReloadingFactory factory) {
        fClassEntry = classEntry;
        fSrcFile = srcFile;
        fFactory = factory;
        fLogger = fFactory.getLogger();
    }

    /**
     * Get the meta-data file path.  The name the interface name for
     * class with .xmlc extension.
     */
    private String metaDataFileName(Class currentClass) throws XMLCReloadingException {
        // Convert class to file path, drop Impl and add extension.
        String fileName = currentClass.getName().replace('.', '/');
        if (!fileName.endsWith(DocumentClass.IMPLEMENTATION_SUFFIX)) {
            throw new XMLCReloadingException("Class name doesn't end in \""
                                            + DocumentClass.IMPLEMENTATION_SUFFIX
                                            + "\"");
        }
        //FIXME: puts extension in const.
        return fileName.substring(0, fileName.length()-4) + ".xmlc";
    }

    /**
     * Load the XMLC meta-data file.
     * FIXME: This should be part of metadata package..
     */
    private MetaData loadMetaData()
        throws IOException, XMLCException,
               XMLCReloadingException, SAXException {

        // Load file throught class loader.
        Class currentClass = fClassEntry.getCurrentClass();
        String fileURL = XMLEntityResolver.CLASSPATH_PROTOCOL_PREFIX
            + metaDataFileName(currentClass);
        MetaDataDocument metaDataDoc
            = MetaDataDocument.parseMetaData(new InputSource(fileURL),
                                             errorReporter,
                                             fFactory.getResourceLoader().getClassLoader());
        return metaDataDoc.getMetaData();
    }

    /**
     * Get the CLASSPATH string.
     */
    private String getClassPathString() {
        ResourceLoader loader = fFactory.getResourceLoader();
        StringBuffer strPath = new StringBuffer();
        String pathSepar = "";

        // Resource loader will return the full classpath.
        URL[] classPath = loader.getClassPath();
        for (int idx = 0; idx < classPath.length; idx++) {
            strPath.append(pathSepar);
            strPath.append(classPath[idx].getFile());
            pathSepar = File.pathSeparator;
        }
        return strPath.toString();
    }

    /**
     * Create a new -classpath argument to javac using the current
     * classpath.
     */
    private void setClassPath(MetaData metaData) {
        String opt = "-classpath";

        JavaCompilerSection compilerSection = metaData.getJavaCompilerSection();
        // Remove existing -classpath option.
        compilerSection.deleteJavacOption(opt);
        JavacOption javacOption = compilerSection.addJavacOption();
        javacOption.setName(opt);
        javacOption.setValue(getClassPathString());
    }

    /**
     * Parse a file path into an array of its components.
     */
    private String[] parseFileName(File file) {
        //FIXME: Use JDK 1.2 get*File() methods
        File absFile = new File(file.getAbsolutePath());
        Vector pathVec = new Vector();
        
        while (absFile != null) {
            pathVec.addElement(absFile.getName());
            String parent = absFile.getParent();
            if (parent == null) {
                absFile = null;
            } else {
                absFile = new File(absFile.getParent());
            }
        }
        String[] path = new String[pathVec.size()];
        
        int idx = 0;
        for (int vidx = pathVec.size()-1; vidx >= 0; vidx--) {
            path[idx++] = (String)pathVec.elementAt(vidx);
        }
        return path;
    }

    /**
     * Convert a class name and class file path to a destination
     * directory for the java compile.
     */
    private File getDestDir(Class classObj,
                            File srcFile) {
        String[] className = JavaLang.parseClassName(classObj.getName());
        String[] path = parseFileName(new File(srcFile.getAbsolutePath()));  //FIXME: use 1.2 get*File

        if (path.length < className.length) {
            throw new XMLCReloadingException("Class name has more elements that its directory, can form a destDir: \""
                                            + classObj.getName() + "\" and \"" + srcFile.getAbsolutePath()
                                            + "\"");
        }
        
        File destDir = new File("");
        int numElements = path.length - className.length;
        for (int idx = 0; idx < numElements; idx++) {
            destDir = new File(destDir, path[idx]);
        }
        return destDir;
    }

    /**
     * Load the meta-data and set parameters.
     */
    private MetaData getMetaData()
            throws IOException, XMLCException, XMLCReloadingException, SAXException {
        MetaData metaData = loadMetaData();
        DocumentClass documentClass = metaData.getDocumentClass();
        CompileOptions compileOptions = metaData.getCompileOptions();

        // Path to source file
        compileOptions.setInputDocument(fSrcFile.getPath());
        
        // Root for class output; also use as tmp dir for generated source.
        File destDir = getDestDir(fClassEntry.getOriginalClass(),
                                  fClassEntry.getOriginalClassFile());
        compileOptions.setClassOutputRoot(destDir.getPath());
        compileOptions.setSourceOutputRoot(destDir.getPath());

        // Must generate delegates and implementation of interface.
        documentClass.setDelegateSupport(true);
        documentClass.setGenerate(GenerateType.IMPLEMENTATION);

        // Must use current CLASSPATH
        setClassPath(metaData);

        metaData.getDocument().completeModifications();
        return metaData;
    }

    /**
     * Run compilation steps.
     */
    private void doCompile(MetaData metaData,
                           ErrorReporter errorReporter)
        throws IOException, XMLCException { 

        Parse parser = new Parse(errorReporter, null);
        XMLCDocument xmlcDoc = parser.parse(metaData);

        EditDOM domEditor = new EditDOM(metaData);
        domEditor.edit(xmlcDoc);

        ClassGenerator classGen = new ClassGenerator(metaData, xmlcDoc, null);
        classGen.generateJavaSource(null);

        try {
            Javac javac = new Javac();
            javac.setThrowErrorMsgs(true);
            javac.compile(metaData, errorReporter, fLogger.getDebugWriter());
        } finally {
            if (!metaData.getCompileOptions().getKeepGeneratedSource()) {
                metaData.getDocumentClass().getJavaClassSource().delete();
            }
        }
    }
     
    /**
     * Generate and throw a XMLCReloadingException, including error
     * output, if any.
     * @param cause Exception that cause the failer, or null if being
     *  thrown because of errorReport errors.
     */
    private void throwAutoCompException(Throwable cause) throws XMLCReloadingException {
        String msg = "Recompile of XMLC generated class: "
            + fClassEntry.getCurrentClass().getName() 
            + " failed:";

        XMLCReloadingException except = new XMLCReloadingException(msg, cause);
        if (errorReporter.getErrorCnt() > 0) {
            except.setErrorOutput(errorReporter.getOutput());
        }
        throw except;
    }

    /**
     * Compile the source file.
     */
    public void compile() throws XMLCReloadingException {
        // Parse metadata and then compile.  Both can output errors to
        // the error report, which we want to include in the exception.

        MetaData metaData = null;
        try {
            metaData = getMetaData();
        } catch (Exception except) {
            throwAutoCompException(except);
        }
        if (errorReporter.getErrorCnt() > 0) {
            throwAutoCompException(null);
        }

        try {
            doCompile(metaData, errorReporter);
        } catch (Exception except) {
            throwAutoCompException(except);
        }
        if (errorReporter.getErrorCnt() > 0) {
            throwAutoCompException(null);
        }
    }
}
