/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: StandardRecompilationHandler.java,v 1.1.8.1 2000/11/16 09:08:42 markd Exp $
 */

package org.enhydra.xml.xmlc.reloading;

import java.util.HashMap;
import java.io.PrintWriter;
import java.io.FileNotFoundException;
import java.io.File;
import java.net.URL;
import org.enhydra.xml.xmlc.XMLObject;
import org.enhydra.xml.xmlc.XMLCError;
import org.enhydra.xml.xmlc.XMLCLogger;

/**
 * Standard recompilation handler.  This searches the class path for a source
 * file using the name compiled into the class.  The source file must exist
 * when the class is first loaded.  A XMLC options file must exist same
 * directory as the class file and having the same basename as the class.
 * This file contains the information needed to recompile the class.
 */
public class StandardRecompilationHandler implements RecompilationHandler {
    /**
     * Table mapping current class object to a File object.
     */
    private HashMap fSrcFiles = new HashMap();

    /** 
     * Factory we are associated with.
     */
    private XMLCReloadingFactory fFactory;

    /**
     * Logging interface.
     */
    private XMLCLogger fLogger;

    /**
     * Tmp directory for recompiler.
     */
    private File fTmpDir;

    /**
     * Bind to a factory.  This is done so that loggers, etc
     * can be grabbed from the factory.
     *
     * @param factory Factory that is creating the object.  Useful
     *  for getting logging object.
     */
    public void init(XMLCReloadingFactory factory) {
        fFactory = factory;
        fLogger = fFactory.getLogger();
    }

    /**
     * Get the CLASSPATH-relative file name for the source file assocaited
     * with a class.  This uses reflection to obtain from a compiled-in
     * constant.
     */
    private String getRelativeSourceFileName(Class currentClass) {
        String name = (String)
            XMLCReloadingFactory.getClassConstant(currentClass,
                                                 XMLObject.XMLC_SOURCE_FILE_FIELD_NAME);
        if (name == null) {
            throw new XMLCReloadingException(XMLObject.XMLC_SOURCE_FILE_FIELD_NAME
                                             + " is null in class " + currentClass);
        }
        return name;
    }

    /**
     * Search for source file that and add to table.
     */
    private synchronized File findSourceFile(Class currentClass)
        throws XMLCReloadingException {

        File srcFile = (File)fSrcFiles.get(currentClass);
        if (srcFile != null) {
            // Added while not synchronizing
            return srcFile;
        }

        // Search for the source file with the class loader.
        String srcPath = getRelativeSourceFileName(currentClass);
        ClassLoader loader = fFactory.getResourceLoader().getClassLoader();
        URL srcURL = loader.getResource(srcPath);
        if (srcURL == null) {
            // Not found.
            throw new XMLCReloadingException("Source file not found on classpath for "
                                             + currentClass.getName() + ": " + srcPath);
        }
        srcFile = new File(srcURL.getFile());
        
        // Got it.
        fSrcFiles.put(currentClass, srcFile);
        return srcFile;
    }

    /**
     * Get the source file, search if needed.  Avoids synchronization unless
     * needed.
     */
    private File getSourceFile(Class currentClass) throws XMLCReloadingException {
        File srcFile = (File)fSrcFiles.get(currentClass);
        if (srcFile != null) {
            return srcFile;
        }
        // Load it..
        return findSourceFile(currentClass);
    }

    /**
     * @see RecompilationHandler#isOutOfDate
     */
    public boolean isOutOfDate(ClassEntry classEntry)
            throws XMLCReloadingException {

        File srcFile = getSourceFile(classEntry.getCurrentClass());
        if (srcFile == null) {
            return false;  // not found
        }
        File currentClassFile = classEntry.getCurrentClassFile();
        boolean isOutOfDate =
            (srcFile.lastModified() > currentClassFile.lastModified());
        if (fLogger.debugEnabled()) {
            fLogger.logDebug("class file is " + (isOutOfDate ? "out-of-date" : "current")
                             + " with source:\n"
                             + "source file: "
                             + srcFile.lastModified() + ": "
                             + srcFile.getAbsolutePath() + "\n"
                             + "class file: "
                             + currentClassFile.lastModified() + ": "
                             + currentClassFile.getAbsolutePath());
        }
        return isOutOfDate;
    }

    /**
     * @see StandardRecompilationHandler#recompile
     */
    public void recompile(ClassEntry classEntry)
            throws XMLCReloadingException {
        File srcFile = getSourceFile(classEntry.getCurrentClass());
        if (srcFile == null) {
            throw new XMLCError("recompiled called when class is know not to exist");
        }

        if (fLogger.debugEnabled()) {
            fLogger.logDebug("Recompiling out-of-date class: " + classEntry.getCurrentClass());
        }
        Recompiler recompiler = new Recompiler(classEntry, srcFile, fFactory);
        recompiler.compile();
    }
}
