/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id$
 */

package tests.org.enhydra.xml.xmlc;

import java.util.ArrayList;
import java.lang.reflect.Method;
import tests.org.enhydra.xml.XMLTestException;

/**
 * Check the attributes of a class.
 *
 * The command line is: <BR>
 * <CODE>
 *  CheckClass [options] classname
 * </CODE>
 *
 * The following options are supported:
 * <UL>
 * <LI> -instanceof classname - Verify that the object is an instanceof this
 *      class. This maybe repeated.
 * <LI> -not-instanceof classname - Verify that the object is not an
 *      instanceof this class. This maybe repeated.
 * <LI> -method name - Check if method name is defined (check for any method
 *     of this name regardless of signature.
 * </UL>
 */
public class CheckClass {
    /**
     * Usage message.
     */
    private static final String USAGE = "wrong # args: CheckClass [options] class";

    // Parsed from command line.
    private ArrayList instanceOfList = new ArrayList();
    private ArrayList notInstanceOfList = new ArrayList();
    private ArrayList methodList = new ArrayList();
    
    /*
     * Load a class.
     */
    private Class loadClass(String name) {
        try {
            return Class.forName(name);
        } catch (ClassNotFoundException except) {
            throw new XMLTestException(except);
        }
    }

    /*
     * Validate instanceof.
     */
    private void checkInstanceOf(Class testClass) {
        for (int idx = 0; idx < instanceOfList.size(); idx++) {
            Class checkClass = loadClass((String)instanceOfList.get(idx));
            if (!checkClass.isAssignableFrom(testClass)) {
                TestUtils.fatal(testClass + " is not an instance of " + checkClass);
            }
        }
    }

    /*
     * Validate not instanceof.
     */
    private void checkNotInstanceOf(Class testClass) {
        for (int idx = 0; idx < notInstanceOfList.size(); idx++) {
            Class checkClass = loadClass((String)notInstanceOfList.get(idx));
            if (checkClass.isAssignableFrom(testClass)) {
                TestUtils.fatal(testClass + " should not be an instance of " + checkClass);
            }
        }
    }

    /*
     * Validate method.
     */
    private void checkMethods(Class testClass) {
        Method[] methods = testClass.getMethods();
        for (int idx = 0; idx < methodList.size(); idx++) {
            String methodName = (String)methodList.get(idx);
            int midx;
            for (midx = 0; midx < methods.length; midx++) {
                if (methods[midx].getName().equals(methodName)) {
                    break;
                }
            }
            if (midx >= methods.length) {
                TestUtils.fatal(testClass + " does not have a method " + methodName);
            }
        }
    }

    /*
     * Test entry point.
     */
    public void run(String[] args) throws Throwable {
        int argIdx = 0;

        while ((argIdx < args.length) && (args[argIdx].startsWith("-"))) {
            if (args[argIdx].equals("-instanceof")) {
                TestUtils.checkOptArg(1, argIdx, args);
                instanceOfList.add(args[++argIdx]);
            } else if (args[argIdx].equals("-not-instanceof")) {
                TestUtils.checkOptArg(1, argIdx, args);
                notInstanceOfList.add(args[++argIdx]);
            } else if (args[argIdx].equals("-method")) {
                TestUtils.checkOptArg(1, argIdx, args);
                methodList.add(args[++argIdx]);
            } else {
                TestUtils.fatal("Error: invalid option \"" + args[argIdx]
                                + "\", expected one of "
                                + " -instanceof, "
                                + " -not-instanceof, "
                                + " -method");
            }
            argIdx++;
        }

        int argsLeft = args.length-argIdx;
        if (argsLeft != 1) {
            TestUtils.fatal(USAGE);
        }

        Class testClass = loadClass(args[argIdx]);
        checkInstanceOf(testClass);
        checkNotInstanceOf(testClass);
        checkMethods(testClass);
    }

    public static void main(String[] args) throws Throwable {
        try {
            new CheckClass().run(args);
        } catch (Throwable error) {
            TestUtils.handleError(error);
        }
    }
}
