/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id$
 */

package tests.org.enhydra.xml.xmlc;

import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.dom.DOMInfo;
import org.enhydra.xml.dom.DOMStats;
import org.enhydra.xml.io.*;
import java.io.*;
import java.util.Vector;
import org.w3c.dom.*;
import org.w3c.dom.html.HTMLDocument;
import org.enhydra.xml.io.DOMFormatter;
import org.enhydra.xml.io.OutputOptions;
import org.enhydra.xml.io.Encoding;

/**
 * Simplistic runner for XMLObjects. Generates a formatted document.
 *
 * The command line is: <BR>
 * <CODE>
 *  RunObject [options] objclass [output]
 * </CODE>
 *
 * The following options are supported:
 * <UL>
 * <LI> -dump file - Dump the DOM to the specified file.
 * <LI> -instanceof classname - Verify that the object is an instanceof this class.
 *       This maybe repeated.
 * <LI> -use-factory - Use a simplistic factor to get an instance of this object.
 * <LI> -reversetext - Reverse the text in all the text nodes.  Use to generate
 *      files to use in recompilation testing.  SPAN IDs are also preserved.
 * <LI> -encoding encoding - XML/HTML encoding.  This will be used for output
 *      and will be checked as the value of getEncoding().
 * <LI> -omit-encoding-header - Omit the encoding header (as used for
 *      WML).
 * <LI> -omit-attribute-char-entity-refs - Don't include entity references in
 *      attribute values.
 * <LI> -check-encoding encoding - Check that the encoding value matches.
 *      A blank argument checks for encoding being null.
 * <LI> -check-mime-type type - MIME type to check against value in document.
 *      A blank argument checks for encoding being null.
 * <LI> -use-to-document - Force the use of toDocument() method to create a
 *      document.  Normally, toDocument() is used only with HTML 
 *      with no special test cases.  This is used to test other cases.
 * <LI> -publicid id - Set override public id in output options; spaces can
 *      be expressed as `@', to solved shell parsing problems
 * <LI> -systemid id - Set override system id in output options.
 * </UL>
 */
public class RunObject {
    /**
     * Usage message.
     */
    private static final String USAGE = "wrong # args: RunObject [options] objclass [output]";

    // Parsed from command line.
    private static String dumpFile;
    private static String docClassName;
    private static boolean useFactory;
    private static boolean reverseText;
    private static Vector instanceOfList = new Vector();
    private static String xmlcObjClass;
    private static String outFile;
    private static String xmlEncoding;
    private static String javaEncoding;
    private static boolean omitEncodingHeader;
    private static String checkEncoding;
    private static String checkMIMEType;
    private static boolean useToDocument;
    private static String publicId;
    private static String systemId;
    private static boolean omitAttributeCharEntityRefs;

    /*
     * Create the test object.
     */
    private static XMLObject createTestObject()
        throws ClassNotFoundException, InstantiationException, IllegalAccessException {
        XMLObject xmlObj;
        if (useFactory) {
            XMLCFactory factory = new XMLCStdFactory(null, null);
            // Append Impl to interface name to get class.
            Class xmlcInterface = Class.forName(xmlcObjClass);
            xmlObj = factory.create(xmlcInterface);
        } else {
            Class objClass = Class.forName(xmlcObjClass);
            xmlObj = (XMLObject)objClass.newInstance();
        }
        // Factory bug check:
        if (xmlObj == null) {
            throw new XMLCError("Null object created for: \""
                                + xmlcObjClass + "\"");
        }
        return xmlObj;
    }

    /*
     * Validate instanceof.
     */
    private static void checkInstanceOf(XMLObject xmlObj)
        throws ClassNotFoundException {
        for (int idx = 0; idx < instanceOfList.size(); idx++) {
            String name = (String)instanceOfList.elementAt(idx);
            Class checkClass = Class.forName(name);
            if (!checkClass.isInstance(xmlObj)) {
                TestUtils.fatal("object is not an instance of " + name);
            }
        }
    }

    /**
     * Reverse all text in a DOM.
     */
    private static void reverseDOMText(Node node) {
        if ((node instanceof Text)&& !(node instanceof CDATASection)) {
            Text text = (Text)node;
            String data = text.getData();
            StringBuffer rdata = new StringBuffer(data.length());
            for (int idx = data.length()-1; idx >= 0; idx--) {
                rdata.append(data.charAt(idx));
            }
            text.setData(rdata.toString());
        } else {
            for (Node child = node.getFirstChild(); child != null;
                 child = child.getNextSibling()) {
                reverseDOMText(child);
            }
        }
    }
 

    /*
     * Validate document class.
     */
    private static void checkDocumentClass(XMLObject xmlObj)
        throws ClassNotFoundException {
        if (docClassName != null) {
            Class docClass = Class.forName(docClassName);
            if (!docClass.isInstance(((XMLObjectImpl)xmlObj).getDocument())) {
                TestUtils.fatal("document is not an instance of " + docClassName);
            }
        }
    }

    /*
     * Validate encoding.
     */
    private static void checkEncoding(XMLObject xmlObj) {
        if (checkEncoding != null) {
            if (checkEncoding.length() == 0) {
                if (xmlObj.getEncoding() != null) {
                    TestUtils.fatal("document encoding is \""
                                    + xmlObj.getEncoding()
                                    + "\", should be null");
                }
            } else {
                if (!checkEncoding.equalsIgnoreCase(xmlObj.getEncoding())) {
                    TestUtils.fatal("document encoding is \""
                                    + xmlObj.getEncoding()
                                    + "\", should be \""
                                    + checkEncoding + "\"");
                }
            }
        }
    }

    /*
     * Validate MIME type.
     */
    private static void checkMIMEType(XMLObject xmlObj) {
        if (checkMIMEType != null) {
            if (checkMIMEType.length() == 0) {
                if (xmlObj.getMIMEType() != null) {
                    TestUtils.fatal("document MIME type is \""
                                    + xmlObj.getMIMEType()
                                    + "\", should be null");
                }
            } else {
                if (!checkMIMEType.equalsIgnoreCase(xmlObj.getMIMEType())) {
                    TestUtils.fatal("document MIME type is \""
                                    + xmlObj.getMIMEType()
                                    + "\", should be \""
                                    + checkMIMEType + "\"");
                }
            }
        }
    }

    /**
     * Output an XML document. 
     */
    private static void outputDocument(XMLObject xmlObj)
        throws IOException {

        // Use toDocument() behavior for HTML files with nothing special.
        // This drops span tag ids which can't be done if reverseText
        // is set. 
        boolean plainHtml = (xmlObj instanceof HTMLDocument)
            && (!reverseText) 
            && (javaEncoding == null)
            && (checkEncoding == null)
            && (publicId == null)
            && (systemId == null)
            && (!omitEncodingHeader)
            && (!omitAttributeCharEntityRefs);
        

        if (plainHtml || useToDocument) {
            TestUtils.outputDocument(xmlObj, outFile);
        } else {
            //If anything is added here, update plainHtml assignment
            String docEncoding = xmlObj.getEncoding();
            OutputOptions options = new OutputOptions();
            if (javaEncoding != null) {
                options.setJavaEncoding(javaEncoding);
            } else {
                options.setXmlEncoding(docEncoding);
            }
            if (publicId != null) {
                options.setPublicId(publicId);
            }
            if (systemId != null) {
                options.setSystemId(systemId);
            }
            options.setOmitEncoding(omitEncodingHeader);
            options.setOmitAttributeCharEntityRefs(omitAttributeCharEntityRefs);
            TestUtils.outputDocument(xmlObj, outFile, options);
        }
    }

    /**
     * Dump DOM information to the specified output file.
     */
    private static void dumpInfo(XMLObject xmlObj,
                                 String outFile) throws IOException {
        File parent = new File(outFile).getParentFile();
        if (parent != null) {
            parent.mkdirs();
        }
        OutputStream outStream = new FileOutputStream(outFile);
        //FIXME: need to be able to specify the encoding:
        PrintWriter out = new PrintWriter(new OutputStreamWriter(outStream, "UTF8"));
        if (xmlObj.getEncoding() != null) {
            out.println("Document encoding: " + xmlObj.getEncoding());
        }
        DOMInfo.printTree("DOM hierarchy", xmlObj,
                          DOMInfo.PRINT_ATTR_DETAILS, out);
        DOMStats.printStats("DOM statistics", xmlObj, DOMStats.SIMPLE_CLASS_NAMES, out);
        out.close();
    }

    /*
     * Test entry point.
     */
    public static void main(String[] args) throws Throwable {
        int argIdx = 0;

        while ((argIdx < args.length) && (args[argIdx].startsWith("-"))) {
            if (args[argIdx].equals("-dump")) {
                TestUtils.checkOptArg(1, argIdx, args);
                dumpFile = args[++argIdx];
            } else if (args[argIdx].equals("-reversetext")) {
                reverseText = true;
            } else if (args[argIdx].equals("-instanceof")) {
                TestUtils.checkOptArg(1, argIdx, args);
                instanceOfList.addElement(args[++argIdx]);
            } else if (args[argIdx].equals("-docclass")) {
                TestUtils.checkOptArg(1, argIdx, args);
                docClassName = args[++argIdx];
            } else if (args[argIdx].equals("-use-factory")) {
                useFactory = true;
            } else if (args[argIdx].equals("-encoding")) {
                TestUtils.checkOptArg(1, argIdx, args);
                xmlEncoding = args[++argIdx];
            } else if (args[argIdx].equals("-omit-encoding-header")) {
                omitEncodingHeader = true;
            } else if (args[argIdx].equals("-omit-attribute-char-entity-refs")) {
                omitAttributeCharEntityRefs = true;
            } else if (args[argIdx].equals("-check-encoding")) {
                TestUtils.checkOptArg(1, argIdx, args);
                checkEncoding = args[++argIdx];
            } else if (args[argIdx].equals("-check-mime-type")) {
                TestUtils.checkOptArg(1, argIdx, args);
                checkMIMEType = args[++argIdx];
            } else if (args[argIdx].equals("-use-to-document")) {
                useToDocument = true;
            } else if (args[argIdx].equals("-publicid")) {
                TestUtils.checkOptArg(1, argIdx, args);
                publicId = args[++argIdx].replace('@', ' ');
            } else if (args[argIdx].equals("-systemid")) {
                TestUtils.checkOptArg(1, argIdx, args);
                systemId = args[++argIdx];
            } else {
                TestUtils.fatal("Error: invalid option \"" + args[argIdx]
                                + "\", expected one of "
                                + "-dump, -reversetext, -instanceof, "
                                + "-docclass, -use-factory, "
                                + "-encoding, -omit-encoding-header, "
                                + "-omit-attribute-char-entity-refs, "
                                + "-check-encoding, "
                                + "-check-mime-type, "
                                + "-use-to-document "
                                + "-public-id "
                                + "-system-id");
            }
            argIdx++;
        }
        int argsLeft = args.length-argIdx;
        if ((argsLeft < 1) || (argsLeft > 2)) {
            TestUtils.fatal(USAGE);
        }
        if (xmlEncoding != null) {
            javaEncoding = Encoding.xml2JavaEncoding(xmlEncoding);
            if (javaEncoding == null) {
                TestUtils.fatal("invalid XML/HTML encoding: \""
                                + xmlEncoding + "\"");
            }
        }
        
        xmlcObjClass = args[argIdx];
        outFile = (argsLeft > 1) ? args[argIdx+1] : null;
        
        try {
            XMLObject xmlObj = createTestObject();

            checkInstanceOf(xmlObj);
            checkDocumentClass(xmlObj);
            checkEncoding(xmlObj);
            checkMIMEType(xmlObj);
            if (reverseText) {
                reverseDOMText(xmlObj);
            }

            if (dumpFile != null) {
                dumpInfo(xmlObj, dumpFile);
            }
            if (outFile != null) {
                outputDocument(xmlObj);
            }
        } catch (Throwable error) {
            TestUtils.handleError(error);
        }
        
    }
}
