/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id$
 */

package tests.org.enhydra.xml.xmlc;

import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.xmlc.html.*;
import org.enhydra.xml.xmlc.EnhydraXMLCLogger;
import com.lutris.logging.LogChannel;
import com.lutris.logging.StandardLogger;
import com.lutris.logging.Logger;
import com.lutris.util.FatalExceptionError;
import org.enhydra.xml.io.DOMFormatter;
import org.enhydra.xml.io.OutputOptions;
import java.io.*;
import java.util.Vector;
import java.lang.reflect.*;
import org.w3c.dom.*;

/**
 * Utility methods used in testing.
 */
public class TestUtils {
    /**
     * Seconds to sleep to allow for system time to advance.  This takes
     * makes sure that we sleep longer than the resolution of the clock
     * used for files.
     * FIXME: Find out the scoop on the resoultion.
     */
    private static final double TOUCH_DELAY = 1.0;

    // Should any logging be done for debugging
    private static final boolean ENABLE_DEBUG_LOGGING = true;

    // Should CLASSLOAD be included in debug logging
    private static final boolean ENABLE_CLASSLOAD_LOGGING = true;
    
    // Log constants.
    public static final String TEST_FACILITY = "test";
    public static final String TEST_LOG_LEVEL = "TEST";
    public static final String CLASSLOAD_LOG_LEVEL = "CLASSLOAD";

    // Standard verbose logging set.
    private static final String[] VERBOSE_LOG_SET = {
        "ERROR", "ALERT", TEST_LOG_LEVEL,
        EnhydraXMLCLogger.XMLC_LOG_LEVEL,
        EnhydraXMLCLogger.XMLC_DEBUG_LOG_LEVEL,
    };

    /**
     * Set to true to dump log messages to stderr.  Useful
     * for intermixing log messages with hacked-in printlns.
     */
    private static final boolean DEBUG_LOG_STDERR = false;

    // Log file.
    private static final String LOG_FILE = "autocomp.loader.log";

    /**
     * Test log channel.
     */
    private static LogChannel logChannel;

    /**
     * Merge two string arrays.
     */
    private static String[] merge(String[] strs1,
                                  String[] strs2) {
        String[] strs = new String[strs1.length+strs2.length];
        int i = 0;
        for (int j = 0; j < strs1.length; j++) {
            strs[i++] = strs1[j];
        }
        for (int j = 0; j < strs2.length; j++) {
            strs[i++] = strs2[j];
        }
        return strs;
    }

    /**
     * Create the test logger 
     */
    private static void createLogger() {
        File logFile;
        String[] fileLogFlags = new String[0];
        String[] stderrLogFlags = new String[] {
            "ERROR", "ALERT"
        };

        if (ENABLE_DEBUG_LOGGING) {
            logFile = new File(LOG_FILE);
            fileLogFlags = merge(fileLogFlags, VERBOSE_LOG_SET);
            if (ENABLE_CLASSLOAD_LOGGING) {
                fileLogFlags = merge(fileLogFlags,
                                     new String[]{CLASSLOAD_LOG_LEVEL});
            }
            if (DEBUG_LOG_STDERR) {
                stderrLogFlags = merge(stderrLogFlags, VERBOSE_LOG_SET);
            }
        } else {
            logFile = new File("/dev/null");
        }

        StandardLogger logger = new StandardLogger(false);
        try {
            // Deleting here causes file to endup empty, so just zero.
            new FileOutputStream(logFile).close();
            logger.configure(logFile, fileLogFlags, stderrLogFlags);
        } catch (IOException ioe) {
            throw new FatalExceptionError(ioe);
        }
        logChannel = logger.getChannel(TEST_FACILITY);
    }

    /**
     * Get the test logger, creating if it doesn't already exist.
     */
    public static LogChannel getLogChannel() {
        if (logChannel == null) {
            createLogger();
        }
        return logChannel;
    }

    /**
     * Add a note to the log file.
     */
    public static void log(String msg) {
        if (logChannel != null) {
            logChannel.write(TEST_LOG_LEVEL, msg);
        }
    }

    /**
     * Add a note and exception to the log file.
     */
    public static void log(String msg, Throwable except) {
        if (logChannel != null) {
            logChannel.write(TEST_LOG_LEVEL, msg, except);
        }
    }

    /*
     * Print an error and exit.
     */
    public static void handleError(Throwable err) {
        if (err instanceof ExceptionInInitializerError) {
            err = ((ExceptionInInitializerError)err).getException();
        }
        System.err.println(err.toString());
        err.printStackTrace();
        System.exit(2);
    }

    /**
     * Print an error message and exit.
     */
    public static void fatal(String msg) {
        System.err.print("Error: ");
        System.err.println(msg);
        System.exit(1);
    }

    /**
     * Check for there being sufficient arguments to an option.
     * Call before advancing index.
     */
    public static void checkOptArg(int numRequired,
                                   int argIdx,
                                   String[] args) {
        if (argIdx + numRequired >= args.length) {
            TestUtils.fatal("Option " + args[argIdx] + " requires "
                            + numRequired + " argument(s)");
        }
    }

    /*
     * Output a document. Options can be null for defaults, same
     * with encoding.
     */
    public static void outputDocument(XMLObject xmlObj,
                                      String outFile,
                                      OutputOptions options)
        throws IOException {
        
        if (xmlObj == null) {
            throw new XMLCError("attempt to output null document");
        }
        OutputStream outStream = new FileOutputStream(outFile);
        if (options == null) {
            // Nothing special, test toDocument().
            PrintWriter out = new PrintWriter(outStream);
            out.println(xmlObj.toDocument());
            out.close();
        } else {
            DOMFormatter formatter = new DOMFormatter(options);
            formatter.write(xmlObj, outStream);
            outStream.close();
        }
    }

    /*
     * Output a document
     */
    public static void outputDocument(XMLObject xmlObj,
                                      String outFile)
        throws IOException {
        outputDocument(xmlObj, outFile, null);
    }

    /**
     * Touch a file by reading and rewriting one byte.
     */
    public static void touchFile(File file) 
            throws XMLCException {
        // Wait, just to let file clock get ahead.
        sleep(TOUCH_DELAY);

        try {
            RandomAccessFile raFile = new RandomAccessFile(file, "rw");
            int fileByte = raFile.read();
            if (fileByte < 0) {
                throw new XMLCException("unexpected EOF on " + file);
            }
            raFile.seek(0);
            raFile.write(fileByte);
            raFile.close();
        } catch (FileNotFoundException except) {
            throw new XMLCException("error touching " + file,
                                    except);
        } catch (IOException except) {
            throw new XMLCException("error touching " + file,
                                    except);
        }
    }

    /**
     * Copy a file.
     */
    public static void copyFile(File src, File dest)
        throws IOException {
        byte[] buf = new byte[2048];

        sleep(TOUCH_DELAY);

        InputStream in = new BufferedInputStream(new FileInputStream(src));
        try {
            OutputStream out = new BufferedOutputStream(new FileOutputStream(dest));
            try {
                int cnt;
                while ((cnt = in.read(buf)) >= 0) {
                    out.write(buf, 0, cnt);
                }
            } finally {
                out.close();
            }
        } finally {
            in.close();
        }
    }

    /**
     * Delay for the specified number of seconds.
     */
    public static void sleep(double secs) {
        try {
            Thread.currentThread().sleep((long)(secs*1000));
        } catch (InterruptedException except) {
            throw new FatalExceptionError(except);
        }
    }

    /**
     *  Get class name and hashCode.
     */
    private static String getClassIdent(Class cl) {
        return cl.getName() + " " + Integer.toString(cl.hashCode(), 16);
    }

    /**
     * Dump object information for debugging.
     */
    public static void dumpObjInfo(String msg,
                                   Object obj) {
        Class cl;
        System.err.println("==== " + msg);
        if (obj == null) {
            System.err.println("  Object is null");
        } else {
            if (obj instanceof Class) {
                cl = (Class)obj;
                System.err.println("  Object is class: " + getClassIdent(cl));
            } else {
                cl = obj.getClass();
                System.err.println("  Object is instance of class: " +  getClassIdent(cl));
            }
            Class[] classes = cl.getInterfaces();
            for (int i = 0; i < classes.length; i++) {
                System.err.println("    I=" + getClassIdent(classes[i]));
            }
        }
    }
}
