/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id$
 */

package tests.org.enhydra.xml.xmlc.devtools;

import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.xmlc.compiler.*;
import org.enhydra.xml.xmlc.commands.options.*;
import org.enhydra.xml.xmlc.dom.DOMInfo;
import org.enhydra.xml.xmlc.dom.XMLCDocument;
import org.enhydra.xml.io.ErrorReporter;
import com.lutris.util.ChainedThrowable;
import org.w3c.dom.*;
import org.xml.sax.SAXException;
import java.io.*;
import org.enhydra.xml.io.OutputOptions;
import org.enhydra.xml.io.DOMFormatter;

/**
 * Tool for parsing files and analysing the resulting DOM.
 * <PRE>
 *   dom-analysis input
 * </PRE>
 */
public class DOMAnalysis {
    /*
     * Object containing the parsed parameters.
     */
    private CompileParams params;

    /**
     * Enable verbose output.
     */
    private boolean verbose = false;

    /*
     * Document being compiled.
     */
    private Document document;

    /*
     * Error output.
     */
    ErrorReporter errorReporter;

    /*
     * Trace and verbose output.
     */
    private PrintWriter traceOut = new PrintWriter(System.err, true);

    /*
     * XML document management object.
     */
    private XMLCDocument xmlcDoc;

    /*
     * Print out a stack trace with special handling of SAX exceptions.
     */
    private void printStackTrace(Throwable except) {
        except.printStackTrace();

        // If this is a chained exception, then walk the stack
        // looking for a non-chained exception.
        while ((except != null) && (except instanceof ChainedThrowable)) {
            except = ((ChainedThrowable)except).getCause();
        }
        if ((except != null) && (except instanceof SAXException)) {
            Exception except2 = ((SAXException)except).getException();
            if (except2 != null) {
                printStackTrace(except2);
            }
        }
    }
    
    /*
     * Handle an exception, printing the approriate error
     * message.
     */
    public void printException(Throwable except) {
        if (except instanceof XMLCException) {
            System.err.println("Error: " + except.toString());
            if (verbose) {
                printStackTrace(except);
            }

            // If this is result of a java.lang exception, print stack.
            Throwable cause = ((XMLCException)except).getCause();
            if ((!verbose) && (cause != null)
                && cause.getClass().getName().startsWith("java.lang.")) {
                printStackTrace(cause);
            }
        } else if ((except instanceof FileNotFoundException)
                   || (except instanceof IOException)) {
            System.err.println("Error: " + except.toString());
            if (verbose) {
                printStackTrace(except);
            }
        } else {
            System.err.println("Error: " + except.toString());
            printStackTrace(except);
        }
        System.exit(1);
    }

    /*
     * Parse the page into the DOM and perform various checks and edits.
     */
    private void parsePage()
        throws XMLCException, IOException {
        Parse parser = new Parse(errorReporter,
                                 verbose ? traceOut : null);
        xmlcDoc = parser.parse(params);
        document = xmlcDoc.getDocument();
    }

    /*
     * Parse arguments and analyize the page.
     */
    public void doRun(String[] args)
        throws XMLCException, IOException {
        
        // Parse command line
        CompileParams params = new CompileParams();

        int argIdx = 0;
        while ((argIdx < args.length) && (args[argIdx].startsWith("-"))) {
        }
        if (argIdx != args.length-1) {
            System.err.println("wrong # args: dom-analysis [options] document");
            System.exit(1);
        }
        params.setSourceDocument(new File(args[argIdx]));
        errorReporter = new ErrorReporter(new PrintWriter(System.err, true));
        parsePage();
    }

    /*
     * Parse arguments and compile the page.  Print errors that
     * are thrown.
     */
    private void run(String[] args) {
        try {
            doRun(args);
        } catch (Exception except) {
            printException(except);
        }
        
    }

    /*
     * Program entry point.
     */
    public static void main(String[] args) {
        DOMAnalysis analyzer = new DOMAnalysis();
        analyzer.run(args);
    }
}
