/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id$
 */

package tests.org.enhydra.xml.xmlc.html;

import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.xmlc.html.*;
import org.enhydra.xml.dom.DOMInfo;
import java.io.*;
import java.net.URLEncoder;
import java.util.*;
import org.w3c.dom.*;
import org.w3c.dom.html.*;

/**
 * Test case, based on the shopping demo, that does some manipulation
 * of the DOM, including cloning and inserting.
 */
public class ShelfTest {
    /**
     * Compile-time debug flag.
     */
    private static boolean DEBUG = false;

    /**
     * XMLC factory object.
     */
    private static XMLCFactory factory;
    
    /**
     * Class name of page (compiled with -generate both), name
     * allows different parsers
     */
    private static String pageClass;

    /**
     * Root of similated Inventory.
     */
    private static Inventory inventoryRoot;

    /**
     * Base name for optional dom dumps.
     */
    private static String domDumpPrefix;

    /**
     * Current request number for DOM dumps.
     */
    private static int requestCnt;

    /**
     * URL to reference ourselve.
     */
    private static final String SHELF_URL = "Shelf.po";

    /**
     * URL to reference the item presentation object.
     */
    private static final String ITEM_URL = "Item.po";

    /**
     * Image to use for open categories.
     */
    private static final String OPEN_CAT_IMAGE = "../media/true.gif";
    private static final String CLOSED_CAT_IMAGE = "../media/false.gif";

    /**
     * The catagory id of the button that was pressed or NULL if
     * one was not pressed.
     */
    private Long buttonCatId;

    /**
     * Table of category ids that are open.  Each entry contains a Long
     * object.
     */
    private Hashtable openCatIds = new Hashtable();

    /**
     * CGI argument containing all open category ids.  If null,
     * then none are open.
     */
    private String openCatCgiArg;
    
    /**
     * Shelf HTML object.
     */
    private ShelfHTML shelfHtml;

    /**
     * Pointers to category prototype entries.
     */
    private HTMLParagraphElement sampleCatBlock;
    private HTMLAnchorElement sampleCatAnchor;
    private HTMLImageElement sampleCatImg;
    private Text sampleCatText;

    /**
     * Pointers to item prototype entries.
     */
    private HTMLParagraphElement sampleItemBlock;
    private HTMLAnchorElement sampleItemAnchor;
    private Text sampleItemText;

    /**
     * Remove prototype table and save pointers to the prototype
     * entries.  For easy of use, prototypes are edited and then
     * cloned.
     */
    private void prepareHtml() {
        shelfHtml = (ShelfHTML)factory.create(pageClass);

        // Save pointers to prototypes, remove ids as they shouldn't be duplicated.
        sampleCatBlock = shelfHtml.getElementSampleCatBlock();
        sampleCatAnchor = shelfHtml.getElementSampleCatAnchor();
        sampleCatImg = shelfHtml.getElementSampleCatImg();
        sampleCatText = XMLCUtil.getFirstText(shelfHtml.getElementSampleCatText());

        sampleItemBlock = shelfHtml.getElementSampleItemBlock();
        sampleItemAnchor = shelfHtml.getElementSampleItemAnchor();
        sampleItemText = XMLCUtil.getFirstText(shelfHtml.getElementSampleItemText());

        // Remove children of prototype list.
        Node listNode = shelfHtml.getElementList();
        Node child = listNode.getFirstChild();
        while (child != null) {
            Node nextChild = child.getNextSibling();
            listNode.removeChild(child);
            child = nextChild;
        }
        
    }

    /**
     * Create a new HTML entry for a category.
     */
    private void makeCategoryHtml(Inventory category, Node parent,
                                  boolean isOpen, int level) {
        if (DEBUG) {
            System.err.println("makeCategotyHtml enter: " + level);
        }
        long id = category.getObjectId();

        String url = SHELF_URL + "?button=" + id;
        if (openCatCgiArg != null) {
            url += "&" + openCatCgiArg;
        }
        sampleCatAnchor.setHref(url);
        if (isOpen) {
            sampleCatImg.setSrc(OPEN_CAT_IMAGE);
        } else {
            sampleCatImg.setSrc(CLOSED_CAT_IMAGE);
        }
        sampleCatText.setData(category.getDesc());
        sampleCatBlock.setClassName("level" + level);
        parent.appendChild(sampleCatBlock.cloneNode(true));
        if (DEBUG) {
            System.err.println("makeCategotyHtml exit: " + level);
        }
    }

    /**
     * Create a new HTML entry for an item.
     */
    private void makeItemHtml(Inventory item, Node parent, int level) {
        if (DEBUG) {
            System.err.println("makeItemHtml enter: " + level);
        }
        String url = ITEM_URL + "?ref=../contents/" + item.getRef()
            + "&add=" + item.getObjectId();
        sampleItemAnchor.setHref(url);
        sampleItemAnchor.setTarget("item");
        sampleItemText.setData(item.getDesc());
        sampleItemBlock.setClassName("level" + level);
        parent.appendChild(sampleItemBlock.cloneNode(true));
        if (DEBUG) {
            System.err.println("makeItemHtml exit: " + level);
        }
    }

    /**
     * Recursively process a category or item entry, adding it to the DOM.
     * Sets node class using level for CSS.
     */
    private void processInventoryEntry(Inventory entry, Node parent, int level) {
        if (entry.isCategory()) {
            boolean isOpen = openCatIds.contains(new Long(entry.getObjectId()));
            makeCategoryHtml(entry, parent, isOpen, level);
            if (isOpen) {
                Vector entries = entry.getChildren();
                for (int idx = 0; idx < entries.size(); idx++) {
                    processInventoryEntry((Inventory)entries.elementAt(idx), parent, level+1);
                }
            }
        } else {
            makeItemHtml(entry, parent, level);
        }
    }

    /**
     * Generate the DOM entries for the current state of the inventory.
     */
    private void processInventory() {
        processInventoryEntry(inventoryRoot,
                              shelfHtml.getElementList(), 1);
    }

    /**
     * Parse CGI arguments into fields.
     */
    private void parseArgs(String button,
                           String open) {

        // Parse button id.
        if (button != null) {
            buttonCatId = new Long(button);
        }
        
        if (open != null) {
            StringTokenizer tokens = new StringTokenizer(open, ",");
            while (tokens.hasMoreTokens()) {
                String token = tokens.nextToken();
                try {
                    Long id = new Long(token);
                    openCatIds.put(id, id);
                } catch (NumberFormatException except) {
                    // Ignore, probably a bogus URL.
                }
            }
        }
    }

    /**
     * Modify the open category list based on the currenmt button.
     * Also generate CGI argument for the open categories.
     */
    private void adjustOpenCategories() {
        if (buttonCatId != null) {
            // Adjust the open entry for the current button.
            if (openCatIds.containsKey(buttonCatId)) {
                openCatIds.remove(buttonCatId);
            } else {
                openCatIds.put(buttonCatId, buttonCatId);
            }
        }

        // Generate CGI arguments.
        StringBuffer buf = new StringBuffer();
        Enumeration openIds = openCatIds.keys();

        while (openIds.hasMoreElements()) {
            if (buf.length() > 0) {
                buf.append(",");
            }
            buf.append(((Long)openIds.nextElement()).toString());
        }
        if (buf.length() > 0) {
            openCatCgiArg = "open=" + buf.toString();
        }
    }

    /**
     * Run a simulated request.
     */
    private void runRequest(String button,
                            String open) throws IOException {
        parseArgs(button, open);
        prepareHtml();
        adjustOpenCategories();
        processInventory();
        if (domDumpPrefix != null) {
            PrintWriter out = new PrintWriter(new FileWriter(domDumpPrefix + "." + requestCnt + ".dom"));
            try {
                DOMInfo.printTree("Shell HTML for request " + requestCnt,
                                  shelfHtml, out);
            } finally {
                out.close();
            }
        }
    }

    /**
     * Create an object and simulate a request.
     */
    private static void request(String button,
                                String open) throws IOException {
        new ShelfTest().runRequest(button, open);
        requestCnt++;
    }

    /**
     * Playback recorded http requests.
     */
    private static void playback() throws IOException {
        request(null, null);
        request("0", null);
        request("2", "0");
        request("2", "2,0");
        request("7", "0");
        request("8", "7,0");
        request("9", "8,7,0");
        request("7", "9,8,7,0");
        request("7", "9,8,0");
        request("8", "9,8,7,0");
        request("7", "9,7,0");
        request("3", "9,0");
        request("4", "9,3,0");
        request("1", "9,4,3,0");
        request("2", "9,4,3,1,0");
        request("1", "9,4,3,2,1,0");
        request("0", "9,4,3,2,0");
        request("0", "9,4,3,2");
        request("2", "9,4,3,2,0");
        request("4", "9,4,3,0");
        request("3", "9,3,0");
        request("1", "9,0");
        request("1", "9,1,0");
        request("0", "9,0");
        request("0", "9");
        request("7", "9,0");
        request("3", "9,7,0");
        request("2", "9,7,3,0");
        request("1", "9,7,3,2,0");
        request("0", "9,7,3,2,1,0");
        request("0", "9,7,3,2,1");
        request("1", "9,7,3,2,1,0");
        request("2", "9,7,3,2,0");
        request("5", "9,7,3,0");
        request("2", "9,7,5,3,0");
        request("7", "9,7,5,3,2,0");
        request("2", "9,5,3,2,0");
    }
    
    /**
     * Build test inventory (build from tracing demo program).
     */
    private static Inventory buildInventory() {
        Inventory entry0, entry1, entry2, entry3;
        entry0 = new Inventory(true, 0, "Store Directory", "Main Menu");
        entry1 = new Inventory(true, 1, "Accessories", "Accessories");
        entry0.addChild(entry1);
        entry2 = new Inventory(false, 5, "Assorted Balls", "pages/accessories_balls.html");
        entry1.addChild(entry2);
        entry2 = new Inventory(false, 6, "Gloves", "pages/accessories_gloves.html");
        entry1.addChild(entry2);
        entry2 = new Inventory(false, 7, "Practice Balls", "pages/accessories_practiceballs.html");
        entry1.addChild(entry2);
        entry1 = new Inventory(true, 2, "Clubs", "Clubs");
        entry0.addChild(entry1);
        entry2 = new Inventory(false, 8, "Graphite Irons", "pages/clubs_irons_graphite.html");
        entry1.addChild(entry2);
        entry2 = new Inventory(false, 9, "Black Hawk Putter", "pages/clubs_putter_1.html");
        entry1.addChild(entry2);
        entry2 = new Inventory(false, 10, "Steel Irons", "pages/clubs_irons_steel.html");
        entry1.addChild(entry2);
        entry2 = new Inventory(false, 11, "Basic Putter", "pages/clubs_putter_2.html");
        entry1.addChild(entry2);
        entry2 = new Inventory(false, 12, "Grey Hawk Putter", "pages/clubs_putter_3.html");
        entry1.addChild(entry2);
        entry2 = new Inventory(false, 13, "Side Waggle Putter", "pages/clubs_putter_4.html");
        entry1.addChild(entry2);
        entry2 = new Inventory(false, 14, "2 Tone Putter", "pages/clubs_putter_5.html");
        entry1.addChild(entry2);
        entry2 = new Inventory(false, 15, "Graphite Wood", "pages/clubs_woods_graphite.html");
        entry1.addChild(entry2);
        entry2 = new Inventory(false, 16, "Steel Wood Driver", "pages/clubs_woods_steel.html");
        entry1.addChild(entry2);
        entry2 = new Inventory(false, 17, "Wood Woods", "pages/clubs_woods_woods.html");
        entry1.addChild(entry2);
        entry1 = new Inventory(true, 3, "Men's Apparel", "MensApparel");
        entry0.addChild(entry1);
        entry2 = new Inventory(true, 4, "Shirts", "Shirts");
        entry1.addChild(entry2);
        entry3 = new Inventory(false, 22, "Classic Golf Shirt", "pages/mens_shirt_1.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 23, "Pocketed Golf Shirt", "pages/mens_shirt_2.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 24, "No-Pocket Polo", "pages/mens_shirt_3.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 25, "Striped Polo", "pages/mens_shirt_4.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 26, "Long Sleeve ", "pages/mens_shirt_5.html");
        entry2.addChild(entry3);
        entry2 = new Inventory(true, 5, "Shoes", "Shoes");
        entry1.addChild(entry2);
        entry3 = new Inventory(false, 27, "Golf-Classic Series", "pages/mens_shoes_1.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 28, "Talon Golf Shoe", "pages/mens_shoes_2.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 29, "Sport Golf Shoe", "pages/mens_shoes_3.html");
        entry2.addChild(entry3);
        entry2 = new Inventory(true, 6, "Sweaters", "Sweaters");
        entry1.addChild(entry2);
        entry3 = new Inventory(false, 30, "Pretty-N-Pink Sweater", "pages/mens_sweater_1.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 31, "Stylish Sweater", "pages/mens_sweater_2.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 32, "Supplex Wind Shirt", "pages/mens_sweater_3.html");
        entry2.addChild(entry3);
        entry1 = new Inventory(true, 7, "Women's Apparel", "WomensApparel");
        entry0.addChild(entry1);
        entry2 = new Inventory(true, 8, "Shirts", "Shirts");
        entry1.addChild(entry2);
        entry3 = new Inventory(false, 33, "Short Sleeve Polo", "pages/womens_shirt_1.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 34, "Striped Polo Jersey", "pages/womens_shirt_2.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 35, "Fancy Collar Jersey", "pages/womens_shirt_3.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 36, "Polo Jersey", "pages/womens_shirt_4.html");
        entry2.addChild(entry3);
        entry2 = new Inventory(true, 9, "Shorts", "Shorts");
        entry1.addChild(entry2);
        entry3 = new Inventory(false, 39, "Microfiber Twill", "pages/womens_shorts_1.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 40, "Twill II", "pages/womens_shorts_2.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 41, "Plaid Twill", "pages/womens_shorts_3.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 42, "Khaki Shorts", "pages/womens_shorts_4.html");
        entry2.addChild(entry3);
        return entry0;
    }

    /**
     * Entry:
     *   ShelfTest pageClass [domDumpPrefix]
     */
    public static void main(String args[]) throws IOException {
        if ((args.length < 1) || (args.length > 2)) {
            System.err.println("wrong # args ShelfTest pageClass [domDumpPrefix]");
            System.exit(1);
        }
        inventoryRoot = buildInventory();
        factory = new XMLCStdFactory(ClassLoader.getSystemClassLoader(),
                                     null);
        pageClass = args[0];
        if (args.length > 1) {
            domDumpPrefix = args[1];
        }
        playback();
    }
}

/**
 * Entry in the inventory tree.
 */
class Inventory {
    private boolean fIsCategory;
    private long fObjectId;
    private String fDesc;
    private String fRef;
    private Vector children;

    /*
     * Constructor
     */ 
    public Inventory(boolean isCategory,
                     long objectId,
                     String desc,
                     String ref) {
        fIsCategory = isCategory;
        fObjectId = objectId;
        fDesc = desc;
        fRef = ref;
    }

    public void addChild(Inventory child) {
        if (children == null) {
            children = new Vector();
        }
        children.add(child);
    }

    public Vector getChildren() {
        return children;
    }

    public boolean isCategory() {
        return fIsCategory;
    }

    public long getObjectId() {
        return fObjectId;
    }

    public String getDesc() {
        return fDesc;
    }

    public String getRef() {
        return fRef;
    }
}
