/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id$
 */

package tests.org.enhydra.xml.xmlc.ssi;

import org.enhydra.xml.xmlc.misc.SSIReader;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import java.io.IOException;
import java.io.Writer;
import java.io.FileWriter;
import java.io.PrintWriter;

/**
 * SSIRead [-readsize chars] infile outfile
 *
 * Read file with SSIReader and write results.  A read size of 1 uses single
 * character reads.
 */
public class SSIRead {
    /** 
     * Print various debugging info.  Some tests will not pass if this
     * is true, as it prints stack on IOExceptions.
     */
    private static boolean DEBUG = false;

    /** Buffer size to use if not specified */
    private static final int DEFAULT_BUF_SIZE = 13;

    /**
     * Read file with SSIReader.
     */
    private void ssiCopyBuffered(String inFile,
                                 PrintWriter out,
                                 int readSize) throws IOException {
        char[] buf = new char[readSize];
        SSIReader ssiReader = new SSIReader(new InputSource(inFile));
        int readLen;
        try {
            do {
                readLen = ssiReader.read(buf, 0, readSize);
                if (readLen > 0) {
                    out.write(buf, 0, readLen);                    
                }
            } while (readLen != -1);
        } finally {
            ssiReader.close();
        }
    }

    /**
     * Read file with SSIReader, one char at a tupe
     */
    private void ssiCopyChar(String inFile,
                             PrintWriter out) throws IOException {
        int ch;
        SSIReader ssiReader = new SSIReader(new InputSource(inFile));
        try {
            do {
                ch = ssiReader.read();
                if (ch >= 0) {
                    out.write(ch);
                }
            } while (ch != -1);
        } finally {
            ssiReader.close();
        }
    }

    /**
     * Read file with various different buffer sizes.
     */
    private void ssiTest(String inFile,
                         String outFile,
                         int readSize) throws IOException {
        PrintWriter out = new PrintWriter(new FileWriter(outFile), true);
        try {
            if (readSize == 1) {
                ssiCopyChar(inFile, out);
            } else {
                ssiCopyBuffered(inFile, out, readSize);
            }
        } finally {
            out.close();
        }
    }
    
    /**
     * Entry.
     */
    static public void main(String[] args) {
        int readSize = -1;
        int argIdx = 0;
        while ((argIdx < args.length) && args[argIdx].startsWith("-")) {
            if (args[argIdx].equals("-readsize")) {
                readSize = Integer.parseInt(args[++argIdx]);
            } else {
                System.err.println("Invalid option: " + args[argIdx]);
            }
            argIdx++;
        }
        if (readSize < 0) {
            readSize = DEFAULT_BUF_SIZE;
        }

        if (args.length - argIdx != 2) {
            System.err.println("Wrong # args: SSIRead -onechar [-readsize chars] infile outfile");
            System.exit(1);
        }


        String inFile = args[argIdx++];
        String outFile = args[argIdx++];
        try {
            new SSIRead().ssiTest(inFile, outFile, readSize);
        } catch (IOException except) {
            // For I/O failuers, just write the message so it can be compare
            System.err.println(except.toString());
            if (DEBUG) {
                except.printStackTrace();
            }
            System.exit(1);
        } catch (Exception except) {
            except.printStackTrace();
            System.exit(1);
        }
    }
}
